/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_dump.h"

#include "../filebrowser.h"
#include "bup_menus.h"
#include "bup_functions.h"

#include "bup_indirect_dump.h"

/*
 * Include group import routines, in order to list
 * contents of file after dumping it.
 */
#include "bup_import_group.h"

/* In order to report which version of
 * Pseudo Saturn Kai was used in order to dump.
 */
#include "../release.h"


/**
 * Include save data manager global stuff :
 * //extern int cdrom_available;
 * //void bup_satcom_init(void);
 * //void bup_anim_starfield(void);
**/
#include "main.h"



/**
 * Note : report uses RomFs decompression buffer in order
 *        to work :
 *        ROMFS_COMP_CHUNK (8*1024)
 *        unsigned char _romfs_decomp_buffer[ROMFS_COMP_CHUNK];
 *        Hence it may not work correctly when
 *        trying to access files from RomFs while
 *        doing dump operation.
 *        The point above is currently safe, so I don't
 *        plan to change memory usage since it saves few KBs
 *        of memory usage.
**/
#include <sc_romfs.h>

/**
 * bup_report_init
 *
 * Setup path to dump report file and initializes its buffer.
**/
void bup_report_init(unsigned char dev_id, char* path_fmt, ...)
{
    char* path = (char*)(_romfs_decomp_buffer) + sizeof(unsigned char);
    char* str = (char*)(_romfs_decomp_buffer) + sizeof(unsigned char) + SC_PATHNAME_MAXLEN;
    va_list argptr;

    /* Setup path to dump report. */
    *((unsigned char*)_romfs_decomp_buffer) = dev_id;
    va_start(argptr, path_fmt);
    vsprintf(path, path_fmt, argptr);
    va_end(argptr);

    /* Initialize dump report. */
    str[0] = '\0';
}

/**
 * bup_report_delete
 *
 * Delete dump report file.
**/
void bup_report_delete(void)
{
    unsigned char dev_id = *((unsigned char*)_romfs_decomp_buffer);
    char* path = (char*)(_romfs_decomp_buffer) + sizeof(unsigned char);

    scf_delete(dev_id, path);
}

/**
 * bup_report_flush
 *
 * Flush dump report informations to path specified in bup_report_init function.
**/
void bup_report_flush(void)
{
    unsigned char dev_id = *((unsigned char*)_romfs_decomp_buffer);
    char* path = (char*)(_romfs_decomp_buffer) + sizeof(unsigned char);
    char* str = (char*)(_romfs_decomp_buffer) + sizeof(unsigned char) + SC_PATHNAME_MAXLEN;

    int len;

    len = strlen(str);

    if(len != 0)
    {
        scf_append((void*)str, (unsigned long)len, dev_id, path);
        str[0] = '\0';
    }
}

/**
 * bup_report_append
 *
 * Append specified message to dump report buffer.
 * When dump report buffer is nearly full, it is automatically
 * flushed to path specified in bup_report_init function.
**/
void bup_report_append(char *string, ...)
{
    char* str = (char*)(_romfs_decomp_buffer) + sizeof(unsigned char) + SC_PATHNAME_MAXLEN;
    va_list argptr;
    int len;

    /* Append report string to buffer. */
    len = strlen(str);
    va_start(argptr, string);
    vsnprintf(str+len, 256, string, argptr);
    va_end(argptr);

    /* Log report string. It doesn't hurts, and is useful for debug. */
    scd_logout("[dump]%s", str+len); sc_check();

    /* Don't forget to append line break. */
    len = strlen(str);
    str[len++] = '\r';
    str[len++] = '\n';
    str[len  ] = '\0';

    /* Automatically flush to media when buffer is near full. */
    if(len >= (ROMFS_COMP_CHUNK - SC_PATHNAME_MAXLEN - 256 - (8*sizeof(unsigned char))))
    {
        bup_report_flush();
    }
}




/**
 * bup_bootcart_saves_available
 *
 * Check if boot cartridge ROM contains save data or not.
 * Example : saves stored by or Action Replay or Memory Card Plus etc.
 *
 * Returns 1 if save(s) could be found, zero else.
**/
dump_file_info_t _df_bootcart;
int bup_bootcart_saves_available(void)
{
    int i;
    unsigned char* buff = (unsigned char*)BUP_BUFFER_ADR;
    unsigned long dump_size = 256*1024;
    void* dump_address = (void*)0x22000000;

    /* Initialize the buffer where the dump data
     * to parse will be stored. Group dump data
     * parse may read outside of buffer, so it is
     * safer to clear it before filling it.
     *
     * The "outside of buffer access" problem concerns
     * the first 100-ish bytes of the dump data, so this
     * won't crash the Saturn.
     *
     * After that, copy dump data to memory in order to parse it.
     */
    memset(buff, 0x00, BUP_BUFFER_LEN-sizeof(sc_global_t));

    /* Using memcpy fills output buffer with null characters
     * when copying from internal backup memory (0x*0180000).
     *
     * Maybe this is due to access restriction when accessing
     * internal backup memory ? (8 bit access is OK, larger
     * access doesn't works, etc)
     *
     * Don't know what happens on real hardware (tested under
     * Yabause only), but anyway let's use the for loop below
     * in order to copy the data to LRAM.
     */
    for(i=0; i<(int)dump_size; i++)
    {
        buff[i] = ((unsigned char*)(dump_address))[i];
    }

    dump_file_info_t* df = &_df_bootcart;
    memset((void*)df, 0, sizeof(dump_file_info_t));

    /* Specify location of data to to parse. */
    df->header_only = 0;
    df->ptr = (unsigned char*)BUP_BUFFER_ADR;
    df->datalen = dump_size;

    int ret = 0;

    if(bup_import_is_group_file(df))
    {
        /* Parse success : verify that at least one save is available. */
        import_group_t* imp = &_import_group_buffer;

        if(imp->ulSaveCount > 0)
        {
            ret = 1;
        }
    }

    return ret;
}


/**
 * bup_cartridge_get_dump_size
 *
 * Guess dump size for cartridge
 * (support both BRAM and ROM cases)
 * and prompt user about which size to use for dump :
 *  - Detected Size
 *  - Large (4MB) : should be enough for all variations of backup cartridges.
 *  - Full (16MB) : just in case of somebody wants to dump everything on his cartridge.
 *
 * Returns zero if user canceled size selection menu,
 * else returns dump size.
**/
unsigned long bup_cartridge_get_dump_size(void* address, unsigned long minsize, unsigned long maxsize)
{
    unsigned char* ptr = (unsigned char*)address;
    unsigned long size = maxsize;

    /* Calculate dump size. */
    while((size>>1) >= minsize) 
    {
        int can_reduce = 0;

        //scd_logout("size=0x%08X (%5d KB) half=0x08X: %02X %02X %02X %02X", 
        //        size, size>>10, 
        //        ptr+(size>>1), 
        //        ptr[(size>>1)+0], 
        //        ptr[(size>>1)+1], 
        //        ptr[(size>>1)+2], 
        //        ptr[(size>>1)+3]);

        if(memcmp(ptr, ptr+(size>>1), 16) == 0)
        {
            /* Mirror data : don't need to dump
            * second half of the data.
            */
            can_reduce = 1;
        }
        else
        {
            /* If all nibbles in second half
             * of the memory to dump are identical, 
             * it should be safe to dump first half
             * of the memory.
             *
             * Why comparing nibbles ? Because cartridge
             * port is 16 bits wide, hence comparing data
             * at same width helps in discarding cases when
             * cartridge returns "constant, but non-zero, 
             * nor-FFh" dummy data.
             */
            unsigned short* cmp_ptr  = (unsigned short*)(ptr+(size>>1));
            unsigned long   cmp_size = (size>>1) / sizeof(unsigned short);
            unsigned short  cmp_dat;

            /* Initialize data to compare first first nibble
             * of second half of data to dump.
             */
            cmp_dat = *cmp_ptr;
            cmp_ptr++;
            cmp_size--;

            can_reduce = 1;
            do
            {
                if((*cmp_ptr) != cmp_dat)
                {
                    can_reduce = 0;
                    break;
                }

                cmp_ptr++;
                cmp_size--;
            } while(cmp_size != 0);
        }

        if(can_reduce)
        {
            size = size>>1;
        }
        else
        {
            break;
        }
    }


    /* Show detected data length, and allow user
     * to specify length of data to dump.
     */
    conio_printf( 2, 4, COLOR_WHITE, "* Detected raw Size:        KB     *");
    conio_printf(23, 4, COLOR_YELLOW, "%5d", size >> 10);

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, "Dump Size Selection");
    menu_set_help(mset, "A/C: Confirm, B: Main Menu");
    menu_set_pos(mset, -1/*x0*/, 6*CONIO_CHAR_H/*y0*/, 30/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    menu_add_item(mitems, "     Detected Size     ", 0/*tag*/);
    if(minsize <= (256*1024))
    {
        menu_add_item(mitems, "  Small Dump  (256 KB) ", 1/*tag*/);
    }
    if(minsize <= (1024*1024))
    {
        menu_add_item(mitems, " Normal Dump  (  1 MB) ", 2/*tag*/);
    }
    if(minsize <= (4*1024*1024))
    {
        menu_add_item(mitems, "  Large Dump  (  4 MB) ", 3/*tag*/);
    }
    if(minsize <= (16*1024*1024))
    {
        menu_add_item(mitems, "   Full Dump  ( 16 MB) ", 4/*tag*/);
    }

    if((((unsigned long)address) & 0x0FFFFFFF) == 0x02000000)
    {
        /* Extra item for some specific boot cartridges. */
        menu_add_item(mitems, "UrawazaDataro (128 KB) ", 8/*tag*/);
        menu_add_item(mitems, "Magic Card V2 ( 32 KB) ", 9/*tag*/);
    }

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    /* Erase detected data length. */
    conio_printf( 2, 4, COLOR_TRANSPARENT, empty_line);

    /* Return dump size as "zero" in
     * the case user canceled size selection.
     */
    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        size = 0;
    }

    /* If requested, force dump size to the one selected by user. */
    if(menu_selected >= 1)
    {
        switch(menu_selected)
        {
        case(1):
            size = 256*1024;
            break;
        case(2):
            size = 1*1024*1024;
            break;
        case(3):
            size = 4*1024*1024;
            break;
        case(4):
            size = 16*1024*1024;
            break;

        case(8):
            /* Special case for Urawaza Dataro. */
            size = 128*1024;
            break;
        case(9):
            /* Special case for Magic Card V2. */
            size = 32*1024;
            break;
        }
    }

    return size;
}


/**
 * writable_media_is_available.
 *
 * Check if writable media is available or not.
 * Used in normal dump feature as well as in indirect dump feature.
 *
 * This check is required in order to display some
 * error message when trying to dump something while
 * there isn't any place where to dump data.
 * Typical example : port-less Action Replay, etc.
**/
int writable_media_is_available(void)
{
    /* Verify if writable media is/are available
     * If not, display error message and return to main menu.
     *
     * Note : some media require menu display in order to be
     *        detected (example : PC link), hence media detection
     *        needs to be performed after initializing SatCom and
     *        displaying any menu.
     */
    sc_device_t devlist[SC_DEVICES_CNT];
    int dev_total_cnt = scf_dev_list(devlist);
    int writable_media = 0;
    int id;

    for(id=0; id<dev_total_cnt; id++)
    {
        if(devlist[id].read_only == 0)
        {
            writable_media = 1;
            break;
        }
    }

    return writable_media;
}

/**
 * writable_media_unavailable_error.
 *
 * Show some error stating that writable media couldn't be found.
**/
void writable_media_unavailable_error(void)
{
    conio_printf(2, 4, COLOR_ERROR, "*          ~~~ Error ~~~           *");
    conio_printf(2, 5, COLOR_ERROR, "* Writable media not found !       *");
    conio_printf(2, 6, COLOR_WHITE, "*                                  *");
    conio_printf(2, 7, COLOR_WHITE, "* You need a writable media such   *");
    conio_printf(2, 8, COLOR_WHITE, "* as PC link or SD card in order   *");
    conio_printf(2, 9, COLOR_WHITE, "* to perform save data dump.       *");

    wait_for_start(NULL);
}


/**
 * bup_kana_to_ut8.
 *
 * Convert and append kana character code to UTF-8 encoding.
 *
 * Note : one kana character takes three bytes when converted to UTF-8.
**/
unsigned char _kana_utf8_tbl[8*8/*row * col*/ * 3/*bytes*/] = 
{
     ' ', '\0', '\0', 0xEF, 0xBD, 0xA1, 0xEF, 0xBD, 0xA2, 0xEF, 0xBD, 0xA3, 0xEF, 0xBD, 0xA4, 0xEF, 0xBD, 0xA5, 0xEF, 0xBD, 0xA6, 0xEF, 0xBD, 0xA7, //  ｡｢｣､･ｦｧ
    0xEF, 0xBD, 0xA8, 0xEF, 0xBD, 0xA9, 0xEF, 0xBD, 0xAA, 0xEF, 0xBD, 0xAB, 0xEF, 0xBD, 0xAC, 0xEF, 0xBD, 0xAD, 0xEF, 0xBD, 0xAE, 0xEF, 0xBD, 0xAF, // ｨｩｪｫｬｭｮｯ
    0xEF, 0xBD, 0xB0, 0xEF, 0xBD, 0xA7, 0xEF, 0xBD, 0xA8, 0xEF, 0xBD, 0xA9, 0xEF, 0xBD, 0xAA, 0xEF, 0xBD, 0xAB, 0xEF, 0xBD, 0xB6, 0xEF, 0xBD, 0xB7, // ｰｧｨｩｪｫｶｷ
    0xEF, 0xBD, 0xB8, 0xEF, 0xBD, 0xB9, 0xEF, 0xBD, 0xBA, 0xEF, 0xBD, 0xBB, 0xEF, 0xBD, 0xBC, 0xEF, 0xBD, 0xBD, 0xEF, 0xBD, 0xBE, 0xEF, 0xBD, 0xBF, // ｸｹｺｻｼｽｾｿ
    0xEF, 0xBE, 0x80, 0xEF, 0xBE, 0x81, 0xEF, 0xBE, 0x82, 0xEF, 0xBE, 0x83, 0xEF, 0xBE, 0x84, 0xEF, 0xBE, 0x85, 0xEF, 0xBE, 0x86, 0xEF, 0xBE, 0x87, // ﾀﾁﾂﾃﾄﾅﾆﾇ
    0xEF, 0xBE, 0x88, 0xEF, 0xBE, 0x89, 0xEF, 0xBE, 0x8A, 0xEF, 0xBE, 0x8B, 0xEF, 0xBE, 0x8C, 0xEF, 0xBE, 0x8D, 0xEF, 0xBE, 0x8E, 0xEF, 0xBE, 0x8F, // ﾈﾉﾊﾋﾌﾍﾎﾏ
    0xEF, 0xBE, 0x90, 0xEF, 0xBE, 0x91, 0xEF, 0xBE, 0x92, 0xEF, 0xBE, 0x93, 0xEF, 0xBE, 0x94, 0xEF, 0xBE, 0x95, 0xEF, 0xBE, 0x96, 0xEF, 0xBE, 0x97, // ﾐﾑﾒﾓﾔﾕﾖﾗ
    0xEF, 0xBE, 0x98, 0xEF, 0xBE, 0x99, 0xEF, 0xBE, 0x9A, 0xEF, 0xBE, 0x9B, 0xEF, 0xBE, 0x9C, 0xEF, 0xBE, 0x9D, 0x22, '\0', '\0', 0xEF, 0xBE, 0x9F  // ﾘﾙﾚﾛﾜﾝ"ﾟ
};

int bup_kana_to_ut8(char* str, unsigned char c)
{
    int ret;

    if(c == '\0')
    {
        /* Terminate string. */
        str[0] = '\0';
        ret = 1;
    }
    else if(c < ' ')
    {
        /* Reserved character, converted to space. */
        str[0] = ' ';
        ret = 1;
    }
    else if(c < 0x80)
    {
        /* Regular ASCII character. */
        str[0] = char2pchar(c);
        ret = 1;
    }
    else
    {
        if((c < (0x80+32)) || (c >= (0x80+96)))
        {
            /* Undefined extended character : let's
             * consider as a space character.
             */
            str[0] = ' ';
            ret = 1;
        }
        else
        {
            /* Kana converted from table. */
            memcpy(str, _kana_utf8_tbl + ((c - (0x80+32)) * 3), 3);
            if(str[1] == '\0')
            {
                /* "Kana" converted to ASCII code. */
                ret = 1;
            }
            else
            {
                ret = 3;
            }
        }
    }

    //str[ret] = '\0';

    return ret;
}






/**
 * bup_print_comment.
 *
 * Sanitize save comment.
 *
 * `convert_utf8' parameter values :
 *   - 0 : keep as-is extended characters above 128, usually Japanese kanas.
 *   - else : convert extended characters in UTF-8 character code.
**/
char _bup_comment_string[(3*10) + 2];
char* bup_print_comment(unsigned char* comment, int convert_utf8)
{
    char* ret = _bup_comment_string;
    int i;

    memset(_bup_comment_string, '\0', sizeof(_bup_comment_string));

    if(convert_utf8)
    {
        int pos = 0;
        int null_char = 0;
        for(i=0; i<10; i++)
        {
            unsigned char c = comment[i];

            if(c == '\0')
            {
                null_char = 1;
            }

            if(null_char)
            {
                ret[pos] = ' ';
                pos++;
            }
            else
            {
                pos += bup_kana_to_ut8(ret + pos, c);
            }
        }
    }
    else
    {
        int null_char = 0;
        for(i=0; i<10; i++)
        {
            unsigned char c = comment[i];

            if(c == '\0')
            {
                null_char = 1;
            }

            if(null_char)
            {
                c = ' ';
            }
            else
            {
                if(c < ' ')
                {
                    c = ' ';
                }
            }

            ret[i] = c;
        }
    }

    return ret;
}


/* Prepare and write individual dump report file. */
void bup_do_individual_report(unsigned char dev_id, char* path, char* real_dump_path, BupDir* tb, unsigned char* bup_dat)
{
    char report_path[SC_PATHNAME_MAXLEN];

    strcpy(report_path, path);
    report_path[strlen(report_path)-3] = 'T';
    report_path[strlen(report_path)-2] = 'X';
    report_path[strlen(report_path)-1] = 'T';
    bup_report_init(dev_id, "%s", report_path);

    /* If exist, delete report file from previous dump. */
    bup_report_delete();

    bup_report_append("--- Saturn Save Data Individual Dump ---");
    struct tm* t = sc_tim_smpcpoll();
    bup_report_append("Dump date: %4d/%02d/%02d %02d:%02d:%02d", 
        1900+t->tm_year, 1+t->tm_mon, t->tm_mday, 
        t->tm_hour, t->tm_min, t->tm_sec);
    char save_name[12];
    memset(save_name, 0, sizeof(save_name));
    memcpy(save_name, tb->filename, 11);
    bup_report_append("Save Name   : %s", save_name);
    bup_report_append("Save Comment: %s", bup_print_comment(tb->comment, 1/*convert_utf8*/));

    /* Convert and output save data date. */
    BupDate bupDate;
    BUP_GetDate(tb->date, &bupDate);
    bup_report_append("Save Date   : %04d/%02d/%02d (%s) %02d:%02d", 
        1980 + bupDate.year, 
        bupDate.month, 
        bupDate.day, 
        bup_get_weekday(bupDate.week), 
        bupDate.time, 
        bupDate.min);

    bup_report_append("Dump Length : %d bytes (%d KB)", tb->datasize, (tb->datasize)>>10);
    bup_report_append("CRC:0x%08X", crc32_calc((const void *)bup_dat, tb->datasize));
    bup_report_append("Dump To:[%s]%s", scf_dev_get_long_name(dev_id), real_dump_path);
    bup_report_append("Dumped with Pseudo Saturn Kai V%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);

    bup_report_flush();
}


/*
 * Omake for SSF users : convert and dump to SSF format.
 * char* path_prefix must be a valid folder location.
 * Example : /SAVES/<date>/{BRAM|CART|FDD1|FDD2}_<ID>
 * And from it, "/SSF/<name>.BIN" is appended.
 *
 * Also, it writes the save file in BUP format directly
 * into SAVES -> LATEST folder so that latest save can be
 * easily accessed after that.
 * Because of write speed limitation, this feature is
 * disabled for Gamer's Cartridge and Fenrir devices.
 */
void bup_save_extra_format(unsigned char dev_id, char* real_dump_path, vmem_bup_header_t* bup_hdr, unsigned char* bup_dat, int ssf_format)
{
    int oversave = 1;

    if(dev_id == SC_FILE_EXTRA)
    {
        oversave = 0;
    }
    else if(dev_id == SC_FILE_SDCARD)
    {
        if(sc_extdev_getid() == SC_EXT_SATCART)
        {
            oversave = 0;
        }
    }

    if(oversave)
    {
        char path2[32];

        memset(path2, '\0', sizeof(path2));
        strcpy(path2, "/SAVES/LATEST");

        if(scf_mkdir(dev_id, path2))
        {
            int k;
            char* filename = path2 + 13/*strlen("/SAVES/LATEST")*/;

            /* Append slash and save data name. */
            filename[0] = '/'; filename++;
            memcpy(filename, bup_hdr->dir.filename, sizeof(bup_hdr->dir.filename)-1);

            for(k=0; k<(sizeof(bup_hdr->dir.filename)-1); k++)
            {
                if(filename[k] <= ' ')
                {
                    filename[k] = '_';
                }
            }
            strcat(filename, BUP_DATA_EXT);

            scf_write(bup_hdr, sizeof(vmem_bup_header_t), dev_id, path2);
            scf_append(bup_dat, bup_hdr->dir.datasize, dev_id, path2);
            /* Extra : write dump report too. */
            bup_do_individual_report(dev_id, path2, real_dump_path, &bup_hdr->dir, bup_dat);
        }
    }


    if(ssf_format)
    {
        /* SSF header format :
         *  11 bytes : comment
         *   1 byte  : empty
         *   4 bytes : date/time (SSF-specific ?)
         */
        unsigned char ssf_header[16];

        /* Fill SSF header.
         * Specification about date/time format are unclear, 
         * so date is set to zero. Majority of games don't
         * use it, so it shouldn't cause problems.
         */
        memset(ssf_header, 0, sizeof(ssf_header));
        memcpy(ssf_header+0, bup_hdr->dir.comment, 11); ssf_header[10] = '\0';
        ssf_header[11] = bup_hdr->dir.language;

        /* Set SSF save file name and path. */
        char ssf_path[SC_PATHNAME_MAXLEN];
        strcpy(ssf_path, real_dump_path);
        strcat(ssf_path, "/SSF");

        if(scf_mkdir(dev_id, ssf_path))
        {
            /* Set SSF save file name from save name. */
            char save_name[12];
            memcpy(save_name, bup_hdr->dir.filename, 12); save_name[11] = '\0';

            strcat(ssf_path, "/");
            strcat(ssf_path, save_name);
            strcat(ssf_path, ".BIN");

            scf_write(ssf_header, sizeof(ssf_header), dev_id, ssf_path);
            scf_append(bup_dat, bup_hdr->dir.datasize, dev_id, ssf_path);
        }
    }
}



void bup_dump_individual(unsigned char dev_id, char* path)
{
    menu_settings_t* mset = &_menu_settings;
    int i, j, k;
    unsigned long vmem_save_id = 0;

    /* Keep previous status of the menu in a dedicated structure. */
    bup_list_menucontext_t menu_context;
    memset(&menu_context, 0, sizeof(bup_list_menucontext_t));

    /* "Suffix" starts from save device name.
     * It changes from a device to another.
     */
    char* path_suffix = path + strlen(path);
    char real_dump_path[SC_PATHNAME_MAXLEN];

    /* Select save data to dump. */
    while(bup_list_exec(mset, BUP_SELECT_MULTIPLE, &menu_context))
    {
        int success_cnt = 0;
        int failure_cnt = 0;

        /* Prepare informations used to display progress bar. */
        sc_device_t dev;
        scf_dev_status(&dev, dev_id);
        int fast_write = (dev.speed_info & SC_DEVFLAG_FASTWRITE ? 1 : 0);

        unsigned long savecnt_total    = 0;
        unsigned long savecnt_current  = 0;
        unsigned long datasize_total   = 0;
        unsigned long datasize_current = 0;
        for(i=0; i<BUP_PARTITION_CNT; i++)
        {
            BupPartInfo* bup = _bupPartInfo+i;

            if(_bupPartInfo[i].part_cnt != 0)
            {
                for(j=0; j<BUP_DIR_SIZE; j++)
                {
                    if(_bup_list_selection[bup->part_id][j])
                    {
                        BupDir* tb = _bupDir[bup->part_id] + j;
                        savecnt_total++;
                        datasize_total += tb->datasize;
                    }
                }
            }
        }

        for(i=0; i<BUP_PARTITION_CNT; i++)
        {
            BupPartInfo* bup = _bupPartInfo+i;
            char* filename_ptr = NULL;

            if(_bupPartInfo[i].part_cnt != 0)
            {
                /* If there is something to dump in this device, 
                 * prepare display and create dump folder.
                 */
                for(j=0; j<BUP_DIR_SIZE; j++)
                {
                    if(_bup_list_selection[bup->part_id][j])
                    {
                        int folder_id;
                        int folder_available = 0;
                        for(folder_id=1; folder_id<1000; folder_id++)
                        {
                            sprintf(path_suffix, "/%s_%03d", bup->part_abbr_name, folder_id);

                            /* (Re)Create folder candidate where to dump saves. */
                            scf_mkdir(dev_id, path);

                            /* Verify that folder doesn't contains saves from previous dump. */
                            unsigned long entries_cnt = scf_filelist(_filelist, FILE_BROWSER_MAXCOUNT, 0/*list_offset*/, 0/*list_sort*/, BUP_DATA_EXT/*filter*/, dev_id, path);
                            if(scf_filecount(_filelist, entries_cnt) == 0)
                            {
                                folder_available = 1;
                                break;
                            }
                        }

                        /* Display some error message if destination folder couldn't be found.
                         * That happens after dumping same thing 999 times in one day ...
                         */
                        if(folder_available == 0)
                        {
                            conio_printf( 2, 4, COLOR_ERROR , "*          ~~~ Error ~~~           *");
                            conio_printf( 2, 5, COLOR_ERROR , "* Destination folder not found !   *");

                            conio_printf( 2, 6, COLOR_WHITE , "* Media:                           *");
                            conio_printf(11, 6, COLOR_YELLOW, "%s", scf_dev_get_long_name(dev_id));

                            path_suffix[0] = '\0';
                            conio_printf( 2, 7, COLOR_WHITE , "* ->                               *");
                            conio_printf( 7, 7, COLOR_YELLOW, "%s", path);

                            wait_for_start(NULL);
                            return;
                        }

                        conio_printf(2,  6, COLOR_SILVER, "*                                  *");
                        conio_printf(2,  7, COLOR_WHITE , "* Source data information :        *");
                        conio_printf(2,  8, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2,  9, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, 10, COLOR_TRANSPARENT, empty_line);

                        /* Setup pointer where to append dump file name. */
                        filename_ptr = path + strlen(path);

                        break;
                    }
                }

                /* If selected, dump save data. */
                for(j=0; j<BUP_DIR_SIZE; j++)
                {
                    if(_bup_list_selection[bup->part_id][j])
                    {
                        BupDir* tb = _bupDir[bup->part_id] + j;
                        char filename[11+1];
                        Sint32 ret;
                        int err = 0;

                        /* Display progress bar according to size of data dumped so far.
                         *
                         * Display method depends of write speed profile of the
                         * device where data are written :
                         *  - "Slow" device (ex: Fenrir)   : display progress on
                         *    a data size basis.
                         *  - "Fast" device (ex: Satiator) : display progress on
                         *    a file count basis.
                         * Purpose of this trick is to make progress bar display
                         * fitting to actual write progress, assuming that fast
                         * devices spend majority of time on creating files
                         * to then instantly write data there.
                         */
                        if(fast_write)
                        {
                            display_progress((savecnt_current * 2) + 1, savecnt_total * 2);
                        }
                        else
                        {
                            display_progress(datasize_current + (tb->datasize / 2), datasize_total);
                        }
                        savecnt_current++;
                        datasize_current += tb->datasize;

                        memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                        filename[sizeof(filename)-1] = '\0';
                        conio_printf(2, 8, COLOR_SILVER, "* dev:%s, name:%s       *", bup->part_abbr_name, filename);

                        vmem_bup_header_t* bup_hdr = (vmem_bup_header_t*)_temp_buff;
                        unsigned char*     bup_dat = ((unsigned char*)bup_hdr) + sizeof(vmem_bup_header_t);

                        /* Prepare save data header (Vmem format). */
                        memset((void*)bup_hdr, 0, sizeof(vmem_bup_header_t));
                        bup_hdr->save_id = vmem_save_id++;
                        memcpy((void*)(bup_hdr->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);
                        memcpy((void*)(&(bup_hdr->dir)), (void*)tb, sizeof(BupDir));
                        bup_hdr->date = bup_get_saturn_date();

                        /* Read save data to temporary buffer. */
                        bup_selpart(bup);
                        dbg_print("BUP_Read(part=%d@%d, \"%s\") ...", bup->device_id, bup->device_part_id, filename);
                        ret = BUP_Read((Uint32)bup->device_id, (Uint8*)filename, (volatile Uint8*)bup_dat);
                        dbg_print("BUP_Read(part=%d@%d, \"%s\") = %d", bup->device_id, bup->device_part_id, filename, ret);
                        display_buplib_error(ret);
                        if(ret != 0)
                        {
                            err = 1;
                        }

                        /* Display contents of data we are currently processing. */
                        bup_display_contents(tb, bup_dat);

                        /* Cleanup save file name, in order to
                         * prepare dump file name.
                         */
                        for(k=0; k<(sizeof(filename)-1); k++)
                        {
                            if(filename[k] <= ' ')
                            {
                                filename[k] = '_';
                            }
                        }
                        sprintf(filename_ptr, "/%s" BUP_DATA_EXT, filename);

                        /* Write backup data to media. */
                        unsigned long write_len;
                        strcpy(real_dump_path, path);
                        write_len = scf_write(bup_hdr, sizeof(vmem_bup_header_t) + tb->datasize, dev_id, real_dump_path);
                        if(write_len != (sizeof(vmem_bup_header_t) + tb->datasize))
                        {
                            err = 1;
                        }

                        if(!err)
                        {
                            /* If save data write success, write dump
                             * report and then save in SSF format.
                             */
                            bup_do_individual_report(dev_id, path, real_dump_path, tb, bup_dat);

                            /* Extra : dump in SSF format too. */
                            filename_ptr[0] = '\0';
                            strcpy(real_dump_path, path);
                            bup_save_extra_format(dev_id, real_dump_path, bup_hdr, bup_dat, 1/*ssf_format*/);
                        }

                        /* Update error count if needed. */
                        if(err)
                        {
                            failure_cnt++;
                        }
                        else
                        {
                            success_cnt++;
                        }

                        /* Display success/failure counts. */
                        conio_printf(2, 9, COLOR_WHITE, "* Success:%3d, Fail:%3d file(s)    *", success_cnt, failure_cnt);
                    }
                }
            } // if(_bupPartInfo[i].part_cnt != 0)
        } // for(i=0; i<BUP_PARTITION_CNT; i++)

        conio_printf(2, 10, COLOR_GREEN, "* Operation ended.                 *");

        if(!wait_for_start("Start: Additional Export, B: Return"))
        {
            break;
        }

        /* Don't forget to clear the display before doing another export. */
        for(i=0; i<21; i++)
        {
            conio_printf(2, 4+i, COLOR_TRANSPARENT, empty_line);
        }
    }
}



void bup_dump_exit(void)
{
    clear_screen();
}


char _dump_path[SC_PATHNAME_MAXLEN];
unsigned char _dump_default_dev_id = SC_FILE_INVALID;

void bup_dump(void)
{
    int i;

    /* First, initialize SatCom. */
    bup_satcom_init();

    /* If indirect dump header found in internal or FDD backup memory, process it.
     * Else, zero is returned and we can continue to normal dump menu.
     */
    if(bup_idump_mainmenu() == BUP_IDUMP_EXIT_MAINMENU)
    {
        bup_dump_exit();
        return;
    }

    /* Check if saves can be extracted from cartridge ROM.
     * Note : this requires at least one writable media to be available.
     */
    int writable_media = writable_media_is_available();
    int boot_saves_available = 0;
    if(writable_media)
    {
        boot_saves_available = bup_bootcart_saves_available();
    }

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;

    while(1)
    {
        /* Re-check for the availability of writable media.
         * This (for example) is handy in the case the user forgot
         * to insert a SD card until entering this menu.
         *
         * Note : as detection of saves in cartridge ROM takes time, 
         *        it is done only during initial display of this menu.
         */
        writable_media = writable_media_is_available();

        /* Verify if cartridge BRAM is available or not. */
        int cart_bram = 0;
        unsigned char* memc_ptr = (unsigned char*)0x24000000;
        if((memc_ptr[0] == 0xFF) && (memc_ptr[1] ==  'B')
        && (memc_ptr[2] == 0xFF) && (memc_ptr[3] ==  'a')
        && (memc_ptr[4] == 0xFF) && (memc_ptr[5] ==  'c')
        && (memc_ptr[6] == 0xFF) && (memc_ptr[7] ==  'k'))
        {
            cart_bram = 1;
        }

        /* Verify if cartridge boot ROM is available or not. */
        int boot_cart = 1;
        unsigned char* bootc_ptr = (unsigned char*)0x22000000;
        if((bootc_ptr[0] == 0xFF) && (bootc_ptr[1] == 0xFF)
        && (bootc_ptr[2] == 0xFF) && (bootc_ptr[3] == 0xFF))
        {
            boot_cart = 0;
        }

        do
        {
            menu_reset_settings(mset);
            menu_set_callback_vblank(mset, bup_anim_starfield);

            menu_set_title(mset, "Save Data Export");
            menu_set_help(mset, "A/C: Confirm, B: Main Menu");
            menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 30/*w*/, 0/*h*/, 1/*cols*/);
            menu_set_erase_on_exit(mset, 1);
            menu_set_features(mset, MENU_FEATURES_TEXTCENTER | MENU_FEATURES_CONDENSED_TITLE | MENU_FEATURES_TALL_PROPORTIONAL);

            menu_reset_items(mitems);
            menu_add_item(mitems, "Select Save(s) to Export"         ,  0/*tag*/);
            if(writable_media)
            {
                menu_add_item(mitems, "Select Save(s) from Cartridge ROM", (boot_saves_available ? 1 : MENU_TAG_GREYOUT)/*tag*/);
            }
            menu_add_item(mitems, "Export Saturn Backup RAM"         , 10/*tag*/);
            menu_add_item(mitems, "Export Cartridge Boot ROM"        , (boot_cart       ? 20 : MENU_TAG_GREYOUT)/*tag*/);
            menu_add_item(mitems, "Export Cartridge Backup RAM"      , (cart_bram       ? 30 : MENU_TAG_GREYOUT)/*tag*/);
            if((cart_bram == 0) || (boot_cart == 0))
            {
                menu_add_item(mitems, "Force Cart Detection", 80/*tag*/);
            }

            /* Show first bytes of BRAM/boot cartridge contents. */
            conio_printf(2, 15, COLOR_WHITE, "Cartridge CS0 contents :");
            conio_printf(2, 20, COLOR_WHITE, "Cartridge CS1 contents :");
            for(i=0; i<4; i++)
            {
                conio_printf(2, 16+i, COLOR_SILVER, "%02X %02X %02X %02X %02X %02X %02X %02X  %c%c %c%c %c%c %c%c", 
                    bootc_ptr[8*i+0], bootc_ptr[8*i+1], bootc_ptr[8*i+2], bootc_ptr[8*i+3], 
                    bootc_ptr[8*i+4], bootc_ptr[8*i+5], bootc_ptr[8*i+6], bootc_ptr[8*i+7], 
                    char2pchar(bootc_ptr[8*i+0]), char2pchar(bootc_ptr[8*i+1]), char2pchar(bootc_ptr[8*i+2]), char2pchar(bootc_ptr[8*i+3]), 
                    char2pchar(bootc_ptr[8*i+4]), char2pchar(bootc_ptr[8*i+5]), char2pchar(bootc_ptr[8*i+6]), char2pchar(bootc_ptr[8*i+7]));

                conio_printf(2, 21+i, COLOR_SILVER, "%02X %02X %02X %02X %02X %02X %02X %02X  %c%c %c%c %c%c %c%c", 
                    memc_ptr[8*i+0], memc_ptr[8*i+1], memc_ptr[8*i+2], memc_ptr[8*i+3], 
                    memc_ptr[8*i+4], memc_ptr[8*i+5], memc_ptr[8*i+6], memc_ptr[8*i+7], 
                    char2pchar(memc_ptr[8*i+0]), char2pchar(memc_ptr[8*i+1]), char2pchar(memc_ptr[8*i+2]), char2pchar(memc_ptr[8*i+3]), 
                    char2pchar(memc_ptr[8*i+4]), char2pchar(memc_ptr[8*i+5]), char2pchar(memc_ptr[8*i+6]), char2pchar(memc_ptr[8*i+7]));
            }

            menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

            /* Clear cartridge's CS0/1 contents display */
            for(i=0; i<10; i++)
            {
                conio_printf(2, 15+i, COLOR_TRANSPARENT, empty_line);
            }

            if(mset->exit_code == MENU_EXIT_CANCEL)
            {
                bup_dump_exit();
                return;
            }

            if(menu_selected == 80)
            {
                /* Force detection of boot/memory cartridges : this is required
                 * in order to dump cartridges with corrupted header, or ROM cartridge
                 * with boot header filled with FFs.
                 */
                cart_bram = 1;
                boot_cart = 1;
                menu_selected = 90;
            }

        } while((menu_selected == 80) || (menu_selected == 90));

        /* Show operation title now.
         * It is not displayed before showing the menu simply because
         * "Save Data Export" is already displayed in menu title.
         */
        conio_printf(2,  3, COLOR_WHITE, "*         Save Data Export         *");

        /* Look for media where to write on.
         * If trying to dump something else than cartridge and writable
         * media not found, display error message and return to main menu.
         *
         * Note : Error is not displayed when trying to dump cartridge
         *        BRAM or ROM because it's possible to indirect dump it.
         *
         * Note : Error message is displayed in the following cases :
         *         - Individual dump when backup memory cartridge is not detected
         *         - Whole dump of internal backup memory
         *        In other cases (backup or boot cartridge), indirect dump
         *        feature can help to export save data or ROM contents so
         *        consequently error message is not displayed.
         */
        writable_media = writable_media_is_available();

        if((writable_media == 0)
        && ((menu_selected == 10) || ((menu_selected ==  0) && ((_whole_part_list & (1 << PART_ID_CART)) == 0x00))))
        {
            writable_media_unavailable_error();

            bup_dump_exit();
            continue;
        }


        if(menu_selected == 1)
        {
            /* Select save(s) to export from cartridge ROM contents. */
            bup_export_from_group_menu(&_df_bootcart);

            bup_dump_exit();
            continue;
        }


        /* Set dump address and length according to
         * device selected for dump.
         */
        void* dump_address = NULL;
        unsigned long dump_size = 0;

        if(menu_selected != 0)
        {
            /* 1: Dump Internal BRAM.
             * 2: Dump Boot Cartridge.
             * 3: Dump Cartridge BRAM.
             *
             * Except internal BRAM, where size to dump doesn't changes, 
             * it is necessary to "guess" size of data to dump, and
             * additionally to allow user to specify size to dump.
             */
            if(menu_selected == 10)
            {
                dump_address = (void*)0x20180000;
                dump_size = 64*1024;
            }
            else if(menu_selected == 20)
            {
                dump_address = (void*)0x22000000;
                dump_size = bup_cartridge_get_dump_size(dump_address, 128*1024, 256*1024);
            }
            else //if(menu_selected == 30)
            {
                dump_address = (void*)0x24000000;
                dump_size = bup_cartridge_get_dump_size(dump_address, 1024*1024, 16*1024*1024);
            }

            /*
             * Zero bytes to dump ?
             * This happens when user cancels dump size selection, 
             * so let's abort dump operation in this case.
             */
            if(dump_size == 0)
            {
                bup_dump_exit();
                continue;
            }
        }


        /* If trying to dump thing other than BRAM, but no writable media found, 
         * prompt user if he's OK to start indirect dump of the cartridge.
         */
        if((menu_selected != 10) && (writable_media == 0))
        {
            int fdd_available = (_whole_part_list & (1 << PART_ID_FDD1) ? 1 : 0);
            int individual_dump = (menu_selected == 0 ? 1 : 0);

            bup_idump_initialmenu(dump_address, individual_dump, dump_size, fdd_available);

            /* Whatever the user input, it is necessary to return to main menu :
             *  - Indirect dump initiated : return to main menu.
             *  - Indirect dump declined : no writable media available, hence
             *                             processing normal dump doesn't makes sense.
             */
            bup_dump_exit();
            continue;
        }


        /* Writable device found, so can continue
         * to do backup data dump.
         */
        int tmp = browser_select_device(FB_FLAG_WRITABLE_DEVICES, &_dump_default_dev_id);
        if(tmp < 0)
        {
            bup_dump_exit();
            continue;
        }

        /* Prepare path where to dump save data :
         * Internal BRAM full dump : /SAVES/<date>/BRAM_<ID>.BIN
         * Cart BRAM full dump : /SAVES/<date>/CART_<ID>.BIN
         * Boot Cart dump : /SAVES/<date>/BOOT_<ID>.BIN
         * Individual saves : /SAVES/<date>/{BRAM|CART|FDD1|FDD2}_<ID>/NAME.BUP
         *
         * <ID> is a 3 digits number between 001 and 999.
         */
        unsigned char dev_id = (unsigned char)tmp;
        char* path = _dump_path;
        memset((void*)path, '\0', SC_PATHNAME_MAXLEN);

        unsigned long ret;

        strcat(path, "/SAVES");
        ret = scf_mkdir(dev_id, path);
        if(!ret)
        {
            bup_dump_exit();
            return;
        }

        struct tm* t = sc_tim_smpcpoll();
        sprintf(path+strlen(path), "/%04d%02d%02d", t->tm_year+1900, t->tm_mon+1, t->tm_mday);
        ret = scf_mkdir(dev_id, path);
        if(!ret)
        {
            bup_dump_exit();
            return;
        }


        if(menu_selected == 0)
        {
            /* Select saves and write them individually to media.
             * Typical usage is when dumping saves from device not
             * mapped in memory such as FDD.
             */
            bup_dump_individual(dev_id, path);
        }
        else //if((menu_selected == 10) || (menu_selected == 20) || (menu_selected == 30))
        {
            /* Set filename where to dump memory contents. */
            char* path_suffix = path + strlen(path);

            int file_id;
            int file_available = 0;
            for(file_id=1; file_id<1000; file_id++)
            {
                /* 1: Dump Internal BRAM.
                 * 2: Dump Boot Cartridge.
                 * 3: Dump Cartridge BRAM.
                 */
                if(menu_selected == 10)
                {
                    sprintf(path_suffix, "/BRAM_%03d.BIN", file_id);
                }
                else if(menu_selected == 20)
                {
                    sprintf(path_suffix, "/BOOT_%03d.BIN", file_id);
                }
                else //if(menu_selected == 30)
                {
                    sprintf(path_suffix, "/CART_%03d.BIN", file_id);
                }

                /* Check file availability. Don't overwrite if already exist. */
                if(scf_size(dev_id, path) == 0)
                {
                    file_available = 1;
                    break;
                }
            }

            /* Display some error message if destination folder is full.
             * That happens after dumping same thing 999 times in one day ...
             */
            if(file_available == 0)
            {
                conio_printf( 2, 4, COLOR_ERROR , "*          ~~~ Error ~~~           *");
                conio_printf( 2, 5, COLOR_ERROR , "* Destination folder full !        *");

                conio_printf( 2, 6, COLOR_WHITE , "* Media:                           *");
                conio_printf(11, 6, COLOR_YELLOW, "%s", scf_dev_get_long_name(dev_id));

                path_suffix[0] = '\0';
                conio_printf( 2, 7, COLOR_WHITE , "* ->                               *");
                conio_printf( 7, 7, COLOR_YELLOW, "%s", path);

                wait_for_start(NULL);
                bup_dump_exit();
                continue;
            }

            /* If possible, shrink the contents of data to dump when it is mapped
             * as 8 bits memory on a 16 bits data bus.
             */
            unsigned char* shrink_address = (unsigned char*)dump_address;
            unsigned long shrink_size = dump_size;
            int is_shrinked = 0;

            if((dump_size/2) <= (BUP_BUFFER_LEN/*-sizeof(sc_global_t)*/))
            {/* Data after shrink can fit into RAM. */

                int dummy_byte_used = 1;
                for(i=0; i<(int)dump_size; i+=2)
                {
                    if(((unsigned char*)dump_address)[i+0] != 0xFF)
                    {
                        dummy_byte_used = 0;
                        break;
                    }
                }

                if(dummy_byte_used)
                {/* Backup memory contains FFh on every even address -> can shrink it. */
                    shrink_address = (unsigned char*)BUP_BUFFER_ADR;
                    shrink_size = dump_size / 2;
                    is_shrinked = 1;

                    for(i=0; i<(int)shrink_size; i++)
                    {
                        shrink_address[i] = ((unsigned char*)dump_address)[(2*i)+1];
                    }
                }
            }

            conio_printf( 2, 4, COLOR_WHITE , "* Media:                           *");
            conio_printf(11, 4, COLOR_YELLOW, "%s", scf_dev_get_long_name(dev_id));

            conio_printf( 2, 5, COLOR_WHITE , "* ->                               *");
            conio_printf( 7, 5, COLOR_YELLOW, "%s", path);

            conio_printf( 2, 6, COLOR_WHITE, "*                                  *");
            conio_printf( 2, 6, COLOR_WHITE, "* Data Length:       KB            *");
            conio_printf(17, 6, COLOR_YELLOW, "%5d", shrink_size >> 10);

            conio_printf(2 , 7, COLOR_WHITE, "* Writing. Please wait ...         *");

            /* Setup dump report. */
            char report_path[SC_PATHNAME_MAXLEN];

            strcpy(report_path, path);
            report_path[strlen(report_path)-3] = 'T';
            report_path[strlen(report_path)-2] = 'X';
            report_path[strlen(report_path)-1] = 'T';
            bup_report_init(dev_id, "%s", report_path);

            /* If exist, delete report file from previous dump. */
            bup_report_delete();

            bup_report_append("--- Saturn Save Data Dump ---");
            struct tm* t = sc_tim_smpcpoll();
            bup_report_append("Dump date: %4d/%02d/%02d %02d:%02d:%02d", 
                1900+t->tm_year, 1+t->tm_mon, t->tm_mday, 
                t->tm_hour, t->tm_min, t->tm_sec);
            bup_report_append("Start Address:0x%08X", dump_address);
            bup_report_append("Dump File Size:%d bytes (%d KB)", shrink_size, shrink_size>>10);
            if(is_shrinked)
            {
                bup_report_append("Note: dummy bytes at even address discarded during dump process.");
            }
            bup_report_append("CRC:0x%08X", crc32_calc((const void *)shrink_address, shrink_size));

            /* Dump data to media. */
            unsigned long dump_len;
            dump_len = scf_write(shrink_address, shrink_size, dev_id, path);

            bup_report_append("Dump To:[%s]%s", scf_dev_get_long_name(dev_id), path);
            bup_report_append("Dumped with Pseudo Saturn Kai V%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);

            if(dump_len != shrink_size)
            {
                conio_printf(2, 7, COLOR_ERROR, "*          ~~~ Error ~~~           *");
                conio_printf(2, 8, COLOR_ERROR, "* File write failure !             *");

                /* Dump failure : end of report. */
                bup_report_append("Error : File write failure !");
                bup_report_flush();

                wait_for_start(NULL);
                bup_dump_exit();
                continue;
            }
            else
            {
                conio_printf(2, 7, COLOR_WHITE, "* Data dump success !              *");

                bup_report_append("Data dump success.");

                /* If dump fits in memory, parse its saves list to report file. */
                unsigned char* buff = (unsigned char*)BUP_BUFFER_ADR;
                int dump_fit_in_ram = (dump_size <= (BUP_BUFFER_LEN/*-sizeof(sc_global_t)*/) ? 1 : 0);

                if(dump_fit_in_ram)
                {
                    /* Initialize the buffer where the dump data
                     * to parse will be stored. Group dump data
                     * parse may read outside of buffer, so it is
                     * safer to clear it before filling it.
                     *
                     * The "outside of buffer access" problem concerns
                     * the first 100-ish bytes of the dump data, so this
                     * won't crash the Saturn.
                     *
                     * After that, copy dump data to memory in order to parse it.
                     */
                    memset(buff, 0x00, BUP_BUFFER_LEN-sizeof(sc_global_t));

                    /* Using memcpy fills output buffer with null characters
                     * when copying from internal backup memory (0x*0180000).
                     *
                     * Maybe this is due to access restriction when accessing
                     * internal backup memory ? (8 bit access is OK, larger
                     * access doesn't works, etc)
                     *
                     * Don't know what happens on real hardware (tested under
                     * Yabause only), but anyway let's use the for loop below
                     * in order to copy the data to LRAM.
                     */
                    for(i=0; i<(int)dump_size; i++)
                    {
                        buff[i] = ((unsigned char*)(dump_address))[i];
                    }

                    /* Prepare display area for saves list. */
                    for(i=0; i<17; i++)
                    {
                        conio_printf( 2, 8+i, COLOR_WHITE, blank_line);
                    }

                    dump_file_info_t df_dat;
                    dump_file_info_t* df = &df_dat;
                    memset((void*)df, 0, sizeof(dump_file_info_t));

                    /* Specify location of data to to parse. */
                    df->header_only = 0;
                    df->ptr = (unsigned char*)BUP_BUFFER_ADR;
                    df->datalen = dump_size;

                    if(bup_import_is_group_file(df))
                    {
                        /* Parse success !
                         * Show saves count and list on screen and dump report.
                         */
                        import_group_t* imp = &_import_group_buffer;

                        conio_printf( 2,  9, COLOR_WHITE , "* Saves Count:                     *");
                        conio_printf(17,  9, COLOR_YELLOW, "%d", imp->ulSaveCount);
                        bup_report_append("Saves Count: %d", imp->ulSaveCount);

                        /* Report saves list. */
                        for(i=0; i<(int)(imp->ulSaveCount); i++)
                        {
                            group_save_t* save = imp->SaveData+i;
                            BupDate bupDate;

                            /* Show first saves on screen. */
                            if(i < 30)
                            {
                                int x0 = ((i%2) == 0 ? 4 : 20);
                                int y0 = 10 + i/2;
                                conio_printf(x0, y0, COLOR_GRAY, "%s", save->tb->filename);
                            }

                            BUP_GetDate(save->tb->date, &bupDate);
                            bup_report_append("|Name:%s, Comment:%s, Size:%6d bytes, Date:%04d/%02d/%02d (%s) %02d:%02d", 
                                save->tb->filename, 
                                bup_print_comment(save->tb->comment, 1/*convert_utf8*/), 
                                save->tb->datasize, 
                                1980 + bupDate.year, 
                                bupDate.month, 
                                bupDate.day, 
                                bup_get_weekday(bupDate.week), 
                                bupDate.time, 
                                bupDate.min);
                        }

                        /* Dump success : end of report. */
                        bup_report_flush();

                        /* ---------------------------------------------- */
                        /* Write each saves in /SAVES/LATEST folder.      */
                        for(i=0; i<(int)(imp->ulSaveCount); i++)
                        {
                            group_save_t* save = imp->SaveData+i;

                            /* Show first saves on screen. */
                            if(i < 30)
                            {
                                int x0 = ((i%2) == 0 ? 4 : 20);
                                int y0 = 10 + i/2;
                                conio_printf(x0, y0, COLOR_WHITE, "%s", save->tb->filename);
                            }

                            /* Clean up buffer where to temporary store save data. */
                            memset((void*)_extract_buff, 0, EXTRACT_BUFFER_LEN);
                            vmem_bup_header_t* bup_hdr = (vmem_bup_header_t*)_extract_buff;
                            unsigned char*     bup_dat = ((unsigned char*)_extract_buff) + sizeof(vmem_bup_header_t);

                            /* Extract data to temporary buffer. */
                            ParseResult res;
                            res = extract_group_save_data(imp, save, bup_dat);
                            if(res == ParseOk)
                            {
                                BupDir* tb = save->tb;

                                /* Prepare save data header (Vmem format). */
                                memset((void*)bup_hdr, 0, sizeof(vmem_bup_header_t));
                                bup_hdr->save_id = i;
                                memcpy((void*)(bup_hdr->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);
                                memcpy((void*)(&(bup_hdr->dir)), (void*)tb, sizeof(BupDir));
                                bup_hdr->date = bup_get_saturn_date();

                                bup_save_extra_format(dev_id, path, bup_hdr, bup_dat, 0/*ssf_format*/);
                            }
                        }
                        /* Write each saves in /SAVES/LATEST folder.      */
                        /* ---------------------------------------------- */
                    }
                    else
                    {
                        conio_printf( 2,  9, COLOR_WHITE , "* No saves found in dump file.     *");

                        /* Dump success : end of report. */
                        bup_report_flush();
                    }
                }
                else
                {
                    /* Dump success : end of report. */
                    bup_report_flush();
                }

                wait_for_start(NULL);
            }
        }

        bup_dump_exit();
    } /* while(1), for transition between menus. */
}
