/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_import.h"

#include "../filebrowser.h"
#include "bup_menus.h"
#include "bup_functions.h"

/**
 * Include save data manager global stuff :
 * //extern int cdrom_available;
 * //void bup_satcom_init(void);
 * //void bup_anim_starfield(void);
**/
#include "main.h"



/* Function common for both single and group import modules. */
int bup_is_valid_filename(unsigned char* filename)
{
    int i;

    /* Discard save starting with space character, 
     * but allow saves with some spaces.
     * Purpose 1 : support NBA Jam saves, which contains spaces in its name.
     * Purpose 2 : discard save with name filled with only space character.
     */
    if(filename[0] == ' ')
    {
        return 0;
    }

    /* Exception for a game whose designers didn't read
     * enough SEGA save data naming standards.
     */
    if(memcmp(filename, "CongoSave_", 10) == 0)
    {
        return 1;
    }

    for(i=0; i<11; i++)
    {
        unsigned char c = filename[i];
        if(((c >= 'A') && (c <= 'Z'))
        || ((c >= '0') && (c <= '9'))
        ||  (c == '_')
        ||  (c == ' '))
        {
            /* Valid save data filename character.
             * Note : space character doesn't complies with Sega
             *        standards, but it is used in NBA Jam saves ...
             *        Name : "NBA JAM SAT", comment : "Save Games"
             */
        }
        else
        {
            /* Special tweak for Swagman, whose saves are named "SWAGMAN.DAT" */
            if((i == 7) && (c == '.'))
            {
                continue;
            }

            return 0;
        }
    }

    return 1;
}

/* Function common for both single and group import modules. */
int bup_filename_is_greater(BupDir* tb1, unsigned long id1, BupDir* tb2, unsigned long id2)
{
    int ret = (id1 > id2 ? 1 : 0);
    int i;
    char c1, c2;

    for(i=0; i<11; i++)
    {
        c1 = sc_tolower(tb1->filename[i]);
        c2 = sc_tolower(tb2->filename[i]);

        if(c1 != c2)
        {
            ret = (c1 > c2 ? 1 : 0);

            /* Put underscore at the bottom
             * of the list, as Saturn BIOS does.
             */
            if((c1 == '_') || (c2 == '_'))
            {
                ret = 1-ret;
            }
            break;
        }
    }

    return ret;
}

/* Function common for both single and group import modules. */
unsigned long bup_get_saturn_date(void)
{
    BupDate sat_date;
    unsigned long saturn_time;
    struct tm* t;

    /* Poll SMPC time. */
    t = sc_tim_smpcpoll();

    /* Convert date to Saturn date. */
    sc_bup_conv_date(t, &sat_date);

    saturn_time = BUP_SetDate(&sat_date);

    return saturn_time;
}


/* Buffer where to extract save data from dump file. */
unsigned char _extract_buff[EXTRACT_BUFFER_LEN];


#include "bup_import_single.h"
#include "bup_import_group.h"

void bup_import(void)
{
    dump_file_info_t df_dat;
    dump_file_info_t* df = &df_dat;


    /* First, initialize SatCom. */
    bup_satcom_init();

    /* Initialize each import modules. */
    bup_import_group_init();
    bup_import_single_init();


    /* Initial path = CD-ROM or SD card or Satiator, "/SAVES" directory. */
    memset((void*)df, 0, sizeof(dump_file_info_t));

    df->dev_id = (cdrom_available ? SC_FILE_CDROM : SC_FILE_SDCARD);
    {
        /* Put priority on on SD card from top cartridge
         * like Gamer's Cartridge or Wasca etc.
         * 
         * Reason is that user is more likely to remove top
         * cartridge than Satiator. Maybe this would require
         * to add some setting to disable SD card access from
         * SH-2 on top cartridge, so that user could use both
         * expansion RAM and store their saves on Satiator if
         * they prefer their data to be organized like that.
         */
        sc_device_t dev[2];
        scf_dev_status(dev+0, SC_FILE_SDCARD  );
        scf_dev_status(dev+1, SC_FILE_SATIATOR);

        if((!dev[0].available) && (dev[1].available))
        {
            df->dev_id = SC_FILE_SATIATOR;
        }
    }

    strcpy(df->path, "/SAVES/DUMMY");

    /* In the case of SDM for Fenrir, let's initialize
     * settings accordingly.
     *
     * Note : as Fenrir file I/O is currently restricted to SD card
     *        root folder, default path is changed too.
     */
    if(_fenrir_sdm)
    {
        df->dev_id = SC_FILE_EXTRA;
        strcpy(df->path, "/DUMMY");
    }
    else
    {
        /* In the case SAVES folder is empty or doesn't exists, 
         * set default path to root folder.
         * This case may happen when try to import a save without
         * having exported anything beforehand.
         */
        unsigned long entries_cnt = scf_filelist(_filelist, FILE_BROWSER_MAXCOUNT, 0/*list_offset*/, 0/*list_sort*/, NULL/*filter*/, df->dev_id, "/SAVES");
        if(scf_filecount(_filelist, entries_cnt) == 0)
        {
            strcpy(df->path, "/DUMMY");
        }
    }


    while(1)
    {
        char curr_sel_file[SC_FILENAME_MAXLEN];

        /* Retrieve previous filename from full path, 
         * in order to default select it in file browser.
         */
        strncpy(curr_sel_file, scf_get_filename(df->path), SC_FILENAME_MAXLEN);
        curr_sel_file[SC_FILENAME_MAXLEN-1] = '\0';

        /* Remove file name from path in the case
         * returning from previous saves import.
         */
        scf_parent_path(df->path);

        conio_printf(2, 3, COLOR_WHITE, "*         Save Data Import         *");
        /* Select file from where to import. */
        int ret = browser_select_file(
                    &(df->dev_id), df->path, curr_sel_file, 
                    FB_FLAG_LIST_FOLDERS | FB_FLAG_BROWSE_DEVICE | FB_FLAG_BROWSE_FOLDER, 
                    DUMP_FILEEXT_SINGLE "," DUMP_FILEEXT_GROUP, 
                    "A: Select File, C/Start: Scan Folder"/* help message. */);
        if(ret == 0)
        {
            return;
        }

        /* Decide if should scan whole folder or import only selected file. */
        int scan_folder = (ret == FB_SELECT_A ? 0 : 1);

        conio_printf(2,  4, COLOR_WHITE, "* Save data file information :     *");

        df->datalen = scf_size(df->dev_id, df->path);
        df->ptr = (unsigned char*)BUP_BUFFER_ADR;
        int can_parse = 1;

        /* File size check. */
        if(df->datalen == 0)
        {
            conio_printf(2, 4, COLOR_TRANSPARENT, empty_line);
            conio_printf(2, 5, COLOR_ERROR, "*          ~~~ Error ~~~           *");
            conio_printf(2, 6, COLOR_ERROR, "*        Can't open file !         *");

            conio_printf(2, 8, COLOR_WHITE, "  Dev : %s", scf_dev_get_long_name(df->dev_id));
            conio_printf(2, 9, COLOR_WHITE, "  Path: %s", df->path);

            wait_for_start(NULL);

            can_parse = 0;
        }
        if(df->datalen > BUP_BUFFER_LEN)
        {
            conio_printf(2, 4, COLOR_TRANSPARENT, empty_line);
            conio_printf(2, 5, COLOR_ERROR, "*          ~~~ Error ~~~           *");
            conio_printf(2, 6, COLOR_ERROR, "* This utility doesn't support     *");
            conio_printf(2, 7, COLOR_ERROR, "* files larger than 1MB.           *");

            wait_for_start(NULL);

            can_parse = 0;
        }


        if(can_parse)
        {
            /* Todo : if <same name>.txt file found, display its contents. */


            /* Clear read buffer before reading dump file.
             * This is required because remains from group dump file may
             * cause incorrect detection as group format when checking
             * the format (group or individual) of a save small enough.
             *
             * Yeah, improving things on group format detection side
             * (for example by verifying input data size) should be
             * be better, but clearing buffer is good, and also fixes
             * incorrect detection problem, so let's keep group format
             * detection thing as-is.
             *
             * Also, it would be better to clear all buffer
             * (BUP_BUFFER_LEN bytes) before reading dump file, but
             * this makes debugger invisible from PC, so let's clear it
             * at a reasonable size.
             */
            memset(df->ptr, 0x00, BUP_BUFFER_LEN-sizeof(sc_global_t));

            /* Read file to import, and try to guess its format. */
            df->header_only = 0;
            df->datalen = scf_read(df->dev_id, df->path, 0/*offset*/, BUP_BUFFER_LEN, (void*)df->ptr);
            conio_printf(2,  5, COLOR_WHITE, "* File size: %7d bytes         *", df->datalen);

            /*
             * In the special case when file is in .BIN extension, 
             * bup_import_is_group_file function will first verify 
             * if specified file is a group data save by analyzing
             * its contents.
             * And if this fails, then file will be processed as a
             * save data in SSF format.
             */
            int ret = 0;
            if(bup_import_is_group_file(df))
            {
                ret = bup_import_group_menu(df);
            }
            else
            {
                if(bup_import_is_single_file(df, NULL/*tb*/))
                {
                    ret = bup_import_single_menu(df, scan_folder);
                }
            }

            if(ret == 0)
            {
                int i;
                unsigned char* data = df->ptr;

                conio_printf(2, 15, COLOR_ERROR, "*          ~~~ Error ~~~           *");
                conio_printf(2, 16, COLOR_ERROR, "* Unrecognized dump file !         *");
                conio_printf(2, 17, COLOR_ERROR, "*                                  *");

                for(i=0; i<5; i++)
                {
                    conio_printf(2, 18+i, COLOR_SILVER, "%02X %02X %02X %02X %02X %02X %02X %02X  %c%c %c%c %c%c %c%c", 
                        data[8*i+0], data[8*i+1], data[8*i+2], data[8*i+3], 
                        data[8*i+4], data[8*i+5], data[8*i+6], data[8*i+7], 
                        char2pchar(data[8*i+0]), char2pchar(data[8*i+1]), char2pchar(data[8*i+2]), char2pchar(data[8*i+3]), 
                        char2pchar(data[8*i+4]), char2pchar(data[8*i+5]), char2pchar(data[8*i+6]), char2pchar(data[8*i+7]));
                }

                wait_for_start(NULL);
            }
        }

        clear_screen();
    }
}
