/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef BUP_IMPORT_GROUP_H
#define BUP_IMPORT_GROUP_H

/**
 * Functions in order to import save data from :
 *  - BRAM dump file.
 *  - Official Memory Cart dump file.
 *  - Memory Cart Plus cartridge dump file.
 *
 * TODO:
 *  - Action Replay cartridge dump file support.
**/
#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>
#include "bup_import_dump.h"


/**
 * File extensions allowed for cartridge dump files.
 * .dat and .fla extension are used by USB Data Link software.
 * .bcr and .bkr extension are used by mednafen emulator.
**/
#define DUMP_FILEEXT_GROUP ".bin,.dmp,.dat,.fla,.bcr,.bkr"



/**
 * Structures and constants definitions for saves
 * group data parse routines.
 *
 * (Re-used in dump modules in order to list saves
 * from dumped data.)
**/
#include "bup_menus.h"

typedef struct _group_save_t
{
    /* Description string for use in menu. */
    char* desc;

    /* Save header. */
    // Uint8	filename[12];	/* t@C */
    // Uint8	comment[11];	/* Rg */
    // Uint8	language;	/* Rǧ */
    // Uint32	date;		/* ^CX^v */
    // Uint32	datasize;	/* f[^TCYiBytej */
    // Uint16	blocksize;	/* f[^TCYiubNj */
    BupDir* tb;

    //QByteArray DateTimeRaw;
    unsigned long ulBlocks;
    unsigned long ulStartCluster;

    char cCounter;

    /* Global data that need to be refreshed by caling ParseSat
     * function before accessing save informations.
     */
    unsigned long ulSATSize;
    short* SAT;

    /* RLE data start pointer (Memory Cart Plus mode only). */
    unsigned char* rle_stt;
} group_save_t;


typedef enum
{
    UNRECOGNIZED_IO_MODE, 
    BACKUPRAM_IO_MODE, 
    MEMCARTPLUS_IO_MODE, 
} FileIOMode;


typedef struct _import_group_t
{
    /* Pointer to begining of dump data. */
    unsigned char* dump_ptr;

    /* Dump file format. */
    FileIOMode IOMode;
    int iClusterSize;

    unsigned long ulSaveCount;
    group_save_t SaveData[BUP_DIR_SIZE];

    /* Global data, shared for each saves. */
#define CLUSTER_MAXCNT 4096
    short SAT[CLUSTER_MAXCNT];
} import_group_t;

/* Buffer for everything related to save import. */
extern import_group_t _import_group_buffer;


typedef enum
{
    ParseOk              = 0,
    ParseNotAHeader      = 1,
    ParseSigNotFound     = 2, 
    ParseDataSizeError   = 3, 
    ParseRleError        = 4, 
    ParseInvalidFilename = 5, 
} ParseResult;


/**
 * extract_group_save_data
 *
 * Extract specified save from dump to specified buffer.
 *
 * Return ParseOk if operation success, other value else.
**/
ParseResult extract_group_save_data(import_group_t* imp, group_save_t* save, unsigned char* buff);


/**
 * bup_import_group_init
 *
 * Initialize internal stuff.
 * Please call on program startup.
**/
void bup_import_group_init(void);

/**
 * bup_import_is_group_file
 *
 * Check if specified file is a cartridge dump file or not.
 *
 * Returns 1 if cartridge dump, zero else.
**/
int bup_import_is_group_file(dump_file_info_t* df);


/**
 * bup_import_group_menu
 *
 * List all saves in specified cartridge dump file, 
 * and import the ones selected by user.
 *
 * Returns 1 if valid dump file, zero else.
**/
int bup_import_group_menu(dump_file_info_t* df);


/**
 * bup_export_from_group_menu
 *
 * List all saves in specified cartridge dump file, 
 * and allow to export them to SD card or ODE etc.
 *
 * Note : that's a mix between "save data import from group file"
 *        and "save data individual dump" features.
 *
 * Returns 1 if something was exported, zero else.
**/
int bup_export_from_group_menu(dump_file_info_t* df);


#endif // BUP_IMPORT_GROUP_H

