/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef BUP_IMPORT_PLANETWEB_H
#define BUP_IMPORT_PLANETWEB_H

/**
 * Functions in order to import save data from :
 *  - Planetweb files.
 *  - SSF files.
 *  - Comms Link files.
 *  - Vmem files.
**/
#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>
#include "bup_import_dump.h"

/* Definitions for vmem format. */
#include "../vmem/dev_cart.h"

/**
 * File extensions allowed for single save files.
**/
#define DUMP_FILEEXT_PLANETWEB ".PWB"
#define DUMP_FILEEXT_RAW       ".RAW"
#define DUMP_FILEEXT_SSF       ".SSF"
#define DUMP_FILEEXT_COMMS     ".CMS"
#define DUMP_FILEEXT_VMEM      BUP_DATA_EXT

/**
 * Real file extension for SSF save data.
 * Unfortunately it collides with file
 * extension for other save data format,
 * hence clarification from user may be needed
 * in order to be sure about save data format.
**/
#define DUMP_FILEEXT_REALSSF   ".BIN"


#define DUMP_FILEEXT_SINGLE DUMP_FILEEXT_PLANETWEB "," DUMP_FILEEXT_RAW "," DUMP_FILEEXT_SSF "," DUMP_FILEEXT_REALSSF "," DUMP_FILEEXT_COMMS "," DUMP_FILEEXT_VMEM


/**
 * Header size for all single save formats.
 *
 * Planet Web: No header
 *
 * SSF: 16 bytes header
 *  - Data[ 0 -  9] = Comment
 *  - Data[     10] = Unused
 *  - Data[     11] = Language
 *  - Data[14 - 16] = Date/Time
 *
 * Comms Link: 34 bytes header
 *  - Data[ 0 -  3] = Start header (80 00 00 00)
 *  - Data[ 4 - 14] = Name (without terminating null character)
 *  - Data[     15] = Language
 *  - Data[16 - 25] = Comment (without terminating null character)
 *  - Data[26 - 29] = Date/Time
 *  - Data[30 - 33] = Data size
 *  - Data[33~    ] = alloc table, then data (+ 4 bytes header every 64 bytes)
 *
 * Vmem: BupDir as-is and few other informations. (64 bytes)
 *       See ../vmem/dev_file.h for details.
**/
#define HDRSIZE_PLANETWEB 0
#define HDRSIZE_SSF       16
#define HDRSIZE_COMMS     34
#define HDRSIZE_VMEM      sizeof(vmem_bup_header_t)
/* "Header" of reserved file such as BIOS dump.
 * It must be small to not slowdown quick folder scan
 * and large enough to allow identification of reserved file.
 * A good tradeoff is 512 bytes because it's the size of
 * one SD card sector and smaller than one CD-ROM sector.
 */
#define HDRSIZE_RESERVED 0x200

#define HDRSIZE_SINGLE MAX(MAX(MAX(MAX(HDRSIZE_PLANETWEB, HDRSIZE_SSF), HDRSIZE_COMMS), HDRSIZE_VMEM), HDRSIZE_RESERVED)


/**
 * bup_import_single_init
 *
 * Initialize internal stuff.
 * Please call on program startup.
**/
void bup_import_single_init(void);


/**
 * bup_import_is_bin_file
 *
 * Check if specified file is a planetweb dump file or not.
 *
 * Returns 1 if planetweb dump, zero else.
**/
int bup_import_is_single_file(dump_file_info_t* df, BupDir* tb);


/**
 * bup_import_single_menu
 *
 * List all planetweb saves in specified folder, 
 * and import the ones selected by user.
 *
 * Returns 1 if valid dump file, zero else.
**/
int bup_import_single_menu(dump_file_info_t* df, int scan_folder);



#endif // BUP_IMPORT_PLANETWEB_H

