/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef BUP_INDIRECT_DUMP_H
#define BUP_INDIRECT_DUMP_H

/**
 * Cartridge Indirect Dump main functions.
 *
 * How it works:
 *  - When cartridge to dump is inserted and booting Save Data Manager
 *    from CD-ROM, cartridge is partially copied to internal backup memory
 *    or backup FDD.
 *  - When cartridge with modern media support (link with PC, SD card, etc) is
 *    inserted, partial copy is saves to media.
 *  - When all parts are copied to media, they are merged into a final dump file.
 *
 * Supported Hardware:
 *  - Can dump backup cartridge or boot cartridge.
 *  - Supports following media :
 *     - Gamer's Cartridge (SD card)
 *     - USB dev cart (USB connectivity to PC)
 *     - Action Replay + USB Data Link (USB connectivity to PC)
 *
 * Notes:
 *  - PC link save support requires to add link selection
 *    menu within Save Data Manager.
**/
#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>
#include "bup_import_dump.h"



// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

/**
 * bup_idump_hdr_t structure :
 * Header data for indirect dump data stored to backup device.
 * Header data is stored in a save separate of dump data.
 *
 * It contains :
 *  - Information of whole data to dump
 *  - Information of current dump
**/
typedef struct _bup_idump_hdr_t
{
    /*
     * Complete dump information :
     * Start address and length of memory region to dump.
     * Also, contains CRC of this memory region, in order
     * to avoid bad dump due to cartridge incorrectly inserted.
     * (Usual CRC32 is used to compute CRC value)
     */
    unsigned long cpl_start_addr;
    unsigned long cpl_length;
    unsigned long cpl_crc32;

    /*
     * Partial dump information :
     * Start address and length of memory region dumped.
     * LZF compressed length is also indicated.
     * Note: address is updated after dumping block to media.
     * Note: lengths are updated after compressing block to backup memory.
     */
    unsigned long prt_start_addr;
    unsigned long prt_length; /* Decompressed length. */
    unsigned long prt_lzf_length; /* Compressed length + size (4 bytes), total for each chunks. */

    /* Dump date, used in order to set dump path. */
    unsigned short year;
    unsigned char  month;
    unsigned char  day;

    /*
     * Individual save data information.
     * Used when indirect dump of user-specified save from cartridge.
     * (Devices other than cartridge aren't supported since they can be directly dumped.)
     * Note : BupDir structure size is 34+2 padding bytes.
     */
    BupDir tb;

    /*
     * Padding for data alignment.
     */
    unsigned char padding[4+1];

    /*
     * Individual/full dump information.
     *  - Zero : dump cartridge from cpl_start_addr, cpl_length bytes.
     *  - Else : dump cartridge save specified in tb structure.
     */
    unsigned char individual_dump;


    /*
     * Folder ID.
     * Used in order to allow multiple dumps on same day.
     */
    unsigned short folder_id;

    /*
     * BUP operation error code.
     * Stored when compressing from cartridge to backup memory, 
     * and written to log file when decompressing to SD card.
     */
    long bup_error_code;

    /*
     * Dump block ID.
     * Initialized at zero, and incremented after each partial dump.
     */
    unsigned char block_id;

    /*
     * Indirect dump save device ID, where dump data is temporarily stored.
     *  DEVICE_ID_BRAM (0) : Built-in memory
     *  DEVICE_ID_DISC (2) : FDD
     * (First partition of FDD is always used.)
     */
    unsigned char idump_dev_id;

    /*
     * Indirect dump status :
     * Indicates status of indirect dump data.
     *  IDUMP_STATUS_INBRAM : Saved in backup device, waiting to be saved to media.
     *  IDUMP_STATUS_READY  : Ready to copy data to dump to backup device.
     */
#define IDUMP_STATUS_READY  1
#define IDUMP_STATUS_INBRAM 2
    unsigned char idump_status;

    /*
     * Device ID where to save partial dumps.
     * If invalid value set ((unsigned char)-1), prompt user
     * where to which media where to do partial dumps.
     *
     * Value is set to IDUMP_MEDIA_UNDEF on first dump, and
     * defined on next step, after user changes cartridge from
     * the one to dump from (example : Memory Cart, Action Replay, etc)
     * to the one to dump to (example : Gamer's Cartridge).
     */
#define IDUMP_MEDIA_UNDEF 0xFF
    unsigned char media_dev_id;
} bup_idump_hdr_t;

// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif

/**
 * Save data read buffer used when dumping specified save :
 * Use the last 256KB of LRAM
 *  BUP_BUFFER_ADR (0x00200000) (<- whole LRAM)
 *  BUP_BUFFER_LEN (1024*1024)
 * Note : LRAM is also used when compressing save data, but
 *        it's not possible theses two buffers overlap since
 *        single save data maximum size shouldn't be larger
 *        than 256KB.
**/
#define IDUMP_SAVE_READ_BUFFER (BUP_BUFFER_ADR + BUP_BUFFER_LEN - (256*1024))

/**
 * Indirect save backup names.
 *  - IDUMP_HDR_SAVE_NAME : dump header informations
 *  - IDUMP_DAT_SAVE_NAME : dump data itself
**/
#define IDUMP_HDR_SAVE_NAME "_CARTDMPHDR"
#define IDUMP_DAT_SAVE_NAME "_CARTDMPDAT"

/**
 * Indirect save folder/file names.
**/
#define IDUMP_FILE_NAME   "CARTDUMP.BIN"
#define IDUMP_REPORT_NAME "CARTDUMP.TXT"


/**
 * bup_idump_search
 *
 * Verify if indirect dump is present or not.
 *
 * If found, return pointer to dump header structure.
 * If not found, return NULL.
**/
//bup_idump_hdr_t* bup_idump_search(void);


/**
 * bup_idump_initialmenu
 *
 * Show indirect dump initial menu :
 *  - Explain what indirect dump is, and warns user that some modern
 *    cartridge is required in order to do it.
 *  - Save first block of data to dump to backup device
 *
 * BupDir* tb parameter :
 *  - NULL : Dump from specified address and length.
 *  - Else : Dump specified save from cartridge.
 * 
 * Note : dump_size parameter is suggested size to dump, and
 *        indirect dump feature may choose a different size of data to dump.
 *
 * Must be called when cartridge dump is requested, but no writable media is found.
**/
void bup_idump_initialmenu(void* dump_address, int individual_dump, unsigned long dump_size, int fdd_available);


/**
 * bup_idump_mainmenu
 *
 * Show indirect dump main menu, allowing following operations :
 *  - Save current partial dump to media
 *  - Dump next block to backup device
 *  - Abort indirect dump
 *
 * Must be called at the begining of Save Data Import menu.
 *
 * Returns BUP_IDUMP_EXIT_MAINMENU (1) if need to exit to Save Data Manager main menu.
 * (In case of partial dump processed correctly)
 *
 * Returns BUP_IDUMP_EXIT_DUMPMENU (0) if need to show normal dump menu.
 * (In case of partial dump not found, etc)
**/
#define BUP_IDUMP_EXIT_DUMPMENU 0
#define BUP_IDUMP_EXIT_MAINMENU 1
int bup_idump_mainmenu(void);


#endif // BUP_INDIRECT_DUMP_H
