/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_menus.h"

#include "../soft_reset.h"
#include "../vdp.h"

#include "main.h"

/* Stuff for extra features. */
#include "extra_menu.h"
#include "flasher.h"


/* Declaration of menu shared variables.
 * And declaration of function to init satcom library.
 */
#include "main.h"

menu_items_t    _menu_items;
menu_settings_t _menu_settings;

/* Backup library RAM (8KB).
 * One unsigned long element is declared in order
 * to verify eventual buffer overflow on vmem side.
 */
unsigned long _bupLibrary[4096+1];

/* Backup library work RAM (8KB).
 * One unsigned long element is declared in order
 * to verify eventual buffer overflow on vmem side.
 */
unsigned long _bupRamWork[2048+1];

BupConfig   _bupConf[BUP_DEVICE_CNT];        /* Up to 3 devices.    */
BupPartInfo _bupPartInfo[BUP_PARTITION_CNT]; /* Up to 4 partitions. */

BupDir    _bupDir [BUP_PARTITION_CNT][BUP_DIR_SIZE];
Sint32    _bupDirN[BUP_PARTITION_CNT];
char      _bup_desc[BUP_PARTITION_CNT][BUP_DIR_SIZE][BUP_DESC_LEN];

int _whole_part_cnt  = 0;
int _whole_part_list = PART_NONE;


int _bup_list_selection[BUP_PARTITION_CNT][BUP_DIR_SIZE];
const char* _bup_part_abbr_names[BUP_PARTITION_CNT] = {
    "BRAM", 
    "CART", 
    "FDD1", 
    "FDD2"
};
int _bup_partinfo_id = 0;
int _bup_dat_id = 0;

/* Buffer in LRAM for copy/move/whatever temporary use.
 * Pointer initialized in buplib_init function.
 */
unsigned char* _temp_buff;


const char* _weekday_names[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
const char* bup_get_weekday(unsigned char weekday)
{
    return _weekday_names[weekday % 7];
}


void bup_selpart(BupPartInfo* bup)
{
    Sint32 ret;

    /* Other devices don't support multiple partitions, so
     * Only select partition when FDD is used.
     *
     * Also, selecting partition is probably (didn't verified) required even
     * if the floppy has only one partition, because user can swap floppy
     * from a two partitions one to a single partition one, and who knows
     * what FDD does in that case ...
     */
    if(bup->device_id == DEVICE_ID_DISC)
    {
        dbg_print("BUP_SelPart(part=%d@%d) ...", bup->device_id, bup->device_part_id);
        ret = BUP_SelPart(bup->device_id, bup->device_part_id);
        dbg_print("BUP_SelPart(part=%d@%d) = %d", bup->device_id, bup->device_part_id, ret);
    }
}



void exit_menu(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected, firmware_available;

    /* Check for Satiator availability. */
    sc_device_t dev;
    scf_dev_status(&dev, SC_FILE_SATIATOR);

    /* Check if boot cartridge is available or not. */
    firmware_available = (memcmp((void*)(0x02000000), (void*)"SEGA SEGASATURN ", 16) == 0 ? 1 : 0);

    /* Pre-select the most suitable exit method according to hardware setup :
     *  - Boot cartridge : "Return to Firmware" = system reset
     *  - Satiator : system reset
     *  - Other : exit to multiplayer
     *
     * The "s_reset_to_satiator" option for Satiator is not used because it
     * sometimes works, some other times doesn't.
     */
    menu_selected = ((dev.available ? 3 : firmware_available ? 2 : 0));

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, "Exit Menu");
    menu_set_help(mset, "A/C: Confirm, B: Main Menu");
    menu_set_pos(mset, -1/*x0*/, -1/*y0*/, 25/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    menu_add_item(mitems, "Exit to Multiplayer", 0/*tag*/);

    if(firmware_available)
    {
        menu_add_item(mitems, "Return to Firmware", 2/*tag*/);
    }
    else
    {
        if(dev.available)
        {
            /* The most reliable way to exit to Satiator menu is to do a
             * system reset, however let's display that as "reset to Satiator"
             * in order to indicate what the action is doing from user POV.
             */
            menu_add_item(mitems, "Reset to Satiator", 3/*tag*/);
        }
        else
        {
            menu_add_item(mitems, "System Reset", 3/*tag*/);
        }
    }

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Don't exit. */
    }
    else if(menu_selected == 0)
    { /* Return to multiplayer. */
        pseudo_exit();
    }
    // else if(menu_selected == 1)
    // { /* Return to Satiator menu. */
    //     s_reset_to_satiator();
    // }
    else if(menu_selected == 2)
    { /* Return to Pseudo Saturn Kai. */
#if 0 // Experimental direct return to Pseudo Saturn Kai. Not working super well ...
        void (*func_ptr)(void);
        func_ptr = (void(*)(void))(0x02000100);
        (*func_ptr)();
#else
        pseudo_sysres();
#endif
    }
    else //if(menu_selected == 3)
    { /* System reset. */
        pseudo_sysres();
    }
}

int wait_for_start(char* msg)
{
    pad_t* pad;

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
    if(msg)
    {
        conio_printf(2, 25, COLOR_WHITE, "%s", msg);
    }
    else
    {
        conio_printf(2, 25, COLOR_WHITE, "Press Start to continue");
    }

    /* Wait until user releases start button. */
    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        /* Animate starfield. */
        bup_anim_starfield();

        /* Poll joypad */
        pad = pad_read();
        if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START)) == 0)
        {
            break;
        }
    }


    while (1)
    {
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        pad = pad_read();
        if(pad->delta[0] & (PAD_A|PAD_B|PAD_C|PAD_START))
        {
            break;
        }

        /* Animate starfield. */
        bup_anim_starfield();
    }

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);

    return (pad->delta[0] & PAD_B ? 0 : 1);
}


int display_msg(int err, char *string, ...)
{
    va_list argptr;
    char buff[LINE_LEN];

    /* Print the log message to the memory. */
    va_start(argptr, string);
    vsnprintf(buff, sizeof(buff), string, argptr); buff[sizeof(buff)-1] = '\0';
    va_end(argptr);

    /* No error happened. */
    if(err == 0)
    {
        /* Clear error display */
        conio_printf(2, 24, COLOR_TRANSPARENT, "%s", empty_line);
//wait_for_start(NULL); // TMP for debug purpose
        return 1;
    }

    conio_printf(2, 24, COLOR_ERROR, "%s", empty_line);
    conio_printf(2, 24, COLOR_ERROR, "%s", buff);
wait_for_start(NULL);//TMP

    return 0;
}

int display_buplib_error(Sint32 err)
{
    const char* error_msg[9] = 
    {
        "Unknown Error Code   ", 
        "BUP_NON              ", 
        "BUP_UNFORMAT         ", 
        "BUP_WRITE_PROTECT    ", 
        "BUP_NOT_ENOUGH_MEMORY", 
        "BUP_NOT_FOUND        ", 
        "BUP_FOUND            ", 
        "BUP_NO_MATCH         ", 
        "BUP_BROKEN           "
    };

    /* No error happened. */
    if(err == 0)
    {
        /* Clear error display */
        return display_msg(err, "");
    }

    /* Error with known error message. */
    if((err >= 1) && (err <= 8))
    {
        return display_msg(err, "Error[%d]: %s", (int)err, error_msg[err]);
    }

    /* Error with unknown error message. */
    return display_msg(err, "Error[%d]: %s", (int)err, error_msg[0]);
}

void display_progress_ex(int n, int total, int row)
{
    /* Progress parameters sanity check. */
    if(total <= 0)
    {
        total = 0;
    }
    if(n > total)
    {
        n = total;
    }

    /* Display progress bar. */
    conio_drawprogress(
        2*CONIO_CHAR_W, row*CONIO_CHAR_H, 
        36*CONIO_CHAR_W, 1*CONIO_CHAR_H, 
        total, n);
}

int _dbg_print_use = 0;
void dbg_print_enable(void)
{
    _dbg_print_use = 1;
}

void dbg_print(char *string, ...)
{
    va_list argptr;
    char buff[LINE_LEN];
    pad_t* pad;
    int cntr = 0;

    if(_dbg_print_use)
    {
        /* Print the log message to the memory. */
        va_start(argptr, string);
        vsnprintf(buff, sizeof(buff), string, argptr); buff[sizeof(buff)-1] = '\0';
        va_end(argptr);

        conio_printf(2, 25, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), empty_line);
        conio_printf(2, 25, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), buff);


        /* Wait until user releases start button. */
        while(1)
        {
            WaitForVBLANKIn();
            WaitForVBLANKOut();

            conio_printf(37, 25, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "%c", '0' + (cntr++)%9);

            pad = pad_read();
            if((pad->new[0] & PAD_START) == 0)
            {
                break;
            }
        }

        while (1)
        {
            WaitForVBLANKIn();
            WaitForVBLANKOut();

            conio_printf(37, 25, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "%c", '0' + (cntr++)%9);

            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                break;
            }
        }
    }
}


/* Save data list callback function.
 * Display additional informations about save data.
 */
void bup_display_infos(int id, int disp_row_id)
{
    char tmp[LINE_LEN];

    if(MENU_TAG_VALID(id) && (id < _bupDirN[_bupPartInfo[_bup_partinfo_id].part_id]))
    {
        BupDir* tb = _bupDir[_bupPartInfo[_bup_partinfo_id].part_id] + id;
        BupDate bupDate;

        memset((void*)tmp, '\0', sizeof(tmp));
        memcpy((void*)tmp, tb->filename, 11);
        conio_printf(2, 21, COLOR_SILVER, "Filename: %s", tmp);

        conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE/*tb->language*/, 2*CONIO_CHAR_W, 22*CONIO_CHAR_H, COLOR_SILVER, "Comment : %s, Lang : %3d", bup_print_comment(tb->comment, 0/*convert_utf8*/), tb->language);

        conio_printf(2, 23, COLOR_SILVER, "Size : %6d, %5d blk", tb->datasize, tb->blocksize);

        /* Convert and display save data date. */
        BUP_GetDate(tb->date, &bupDate);
        conio_printf(2, 24, COLOR_SILVER, "Date : %04d/%02d/%02d (%s) %02d:%02d", 
            1980 + bupDate.year, 
            bupDate.month, 
            bupDate.day, 
            bup_get_weekday(bupDate.week), 
            bupDate.time, 
            bupDate.min);
    }
    else
    {
        conio_printf(2, 21, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 22, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 23, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
    }
}


/* Internal function for bup_list_exec. */
int bupdir_compare_name(BupDir* tb1, BupDir* tb2)
{
    int i;
    for(i=0; i<12; i++)
    {
        if(tb1->filename[i] > tb2->filename[i])
        {
            return 1;
        }
        if(tb1->filename[i] < tb2->filename[i])
        {
            return -1;
        }
    }
    return 0;
}

sel_stats_t _list_sel_stats;

void bup_list_str_callback(int item_id, char** str_ptr, int* str_len_ptr, int* codepage)
{
    if((item_id >= 0) && (item_id < _bupDirN[_bupPartInfo[_bup_partinfo_id].part_id]))
    {
        //BupDir* tb = _bupDir[_bupPartInfo[_bup_partinfo_id].part_id] + item_id;

        //*str_ptr = (char*)(tb->filename);
        //*str_len_ptr = 12;

        *str_ptr = _bup_desc[_bupPartInfo[_bup_partinfo_id].part_id][item_id];
        *str_len_ptr = -1/*auto length*/;
        /* Using tb->language for codepage may cause problems. */
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
    else
    {
        *str_ptr = "???";
        *str_len_ptr = -1/*auto length*/;
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
}


int bup_list_exec(menu_settings_t* mset, int sel_type, bup_list_menucontext_t* menu_ctx)
{
    int i, j, k;
    /* Highlighted item ID in list. */
    int menu_current[BUP_PARTITION_CNT];

    /* Initialize menu internals. */
    menu_reset_settings(mset);
    menu_set_callback_redraw(mset, bup_display_infos);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    /* Reset selection, highlight, etc status.
     * and list up saves in each partitions.
     */
    memset((void*)_bupDir, 0, sizeof(BupDir)*BUP_PARTITION_CNT*BUP_DIR_SIZE);
    for(i=0; i<BUP_PARTITION_CNT; i++)
    {
        menu_current[i] = (menu_ctx ? menu_ctx->current_save_id[i] : 0);
        for(j=0; j<BUP_DIR_SIZE; j++)
        {
            _bup_list_selection[i][j] = 0;
        }
        _bupDirN[i] = 0;
    }

    /* Reset selection stats. */
    memset((void*)&_list_sel_stats, 0, sizeof(sel_stats_t));

    /* Retrieve contents of all save partitions. */
    for(i=0; i<_whole_part_cnt; i++)
    {
        BupPartInfo* bup = _bupPartInfo+i;

        if(_bupPartInfo[i].part_cnt != 0)
        {
            Sint32 ret;

            /* List save data files in each devices. */
            bup_selpart(bup);
            dbg_print("#3 BUP_Dir(part=%d@%d) (%d %d) ...", bup->device_id, bup->device_part_id, i, _whole_part_cnt);
            ret = BUP_Dir(bup->device_id, (Uint8 *)"", BUP_DIR_SIZE, _bupDir[bup->part_id]);
            dbg_print("#4 BUP_Dir(part=%d@%d) = %d", bup->device_id, bup->device_part_id, ret);

            /* Contents cleanup of all entries */
            for(j=0; j<ret; j++)
            {
                BupDir* tb = _bupDir[bup->part_id] + j;

                /* Add terminating null character to strings. */
                tb->filename[11] = '\0';
                tb->comment [10] = '\0';
            }

            /* No error when reading directory list ? */
            if(ret >= 0)
            {
                /* If more than 2 entries, sort entries in ASCII order. */
                if(ret >= 2)
                {
                    for(j=0; j<ret-1; j++)
                    {
                        for(k=j+1; k<ret; k++)
                        {
                            BupDir* tb1 = _bupDir[bup->part_id] + j;
                            BupDir* tb2 = _bupDir[bup->part_id] + k;
                            if(bupdir_compare_name(tb1, tb2) > 0)
                            {
                                BupDir tmp;
                                memcpy((void*)&tmp, (void*)tb1 , sizeof(BupDir));
                                memcpy((void*)tb1 , (void*)tb2 , sizeof(BupDir));
                                memcpy((void*)tb2 , (void*)&tmp, sizeof(BupDir));
                            }
                        }
                    }
                }

                /* Set save description for use in selection menu. */
                for(j=0; j<ret; j++)
                {
                    BupDir* tb = _bupDir[bup->part_id] + j;

                    char* desc = _bup_desc[bup->part_id][j];
                    memset((void*)desc, '\0', BUP_DESC_LEN);

                    snprintf
                    (
                        desc, BUP_DESC_LEN, "%s:%s", 
                        tb->filename, 
                        bup_print_comment(tb->comment, 0/*convert_utf8*/)
                    );
                }
            }
            else
            {
                /* Not enough room to store all save data.
                 * No support for this, but at least display
                 * some error message.
                 */
                dbg_print("part=%d@%d -> Dir=%d !!!", bup->part_id, bup->device_part_id, ret);
                display_buplib_error(42);
            }

            /* remember number of files. */
            _bupDirN[bup->part_id] = (ret < 0 ? 0 : ret);
        } // if(_bupPartInfo[i].part_cnt != 0)
    }
    dbg_print("read directories end !");


    /* Restore initial partition to display.
     *
     * If select save within one device only, force
     * initial display to it.
     *
     * And, force initial display to non-empty save
     * device when device selection is allowed.
     */
    _bup_partinfo_id = (menu_ctx ? menu_ctx->partinfo_id : 0);
    if((sel_type == BUP_SELECT_SINGLE_CART) || (sel_type == BUP_SELECT_SINGLE_BRAM))
    {
        int target_part_id = (sel_type == BUP_SELECT_SINGLE_CART ? PART_ID_CART : PART_ID_BRAM);
        _bup_partinfo_id = target_part_id;

        /* If target partition not found (this may happen with
         * cartridge), emergency exit.
         */
        if(_bupPartInfo[_bup_partinfo_id].part_id != target_part_id)
        {
            return 0;
        }

        /* If target partition doesn't contains any save, emergency exit. */
        if(_bupDirN[_bupPartInfo[_bup_partinfo_id].part_id] == 0)
        {
            return 0;
        }
    }
    else
    {
        /* Force initial display to non-empty save device. */
        for(i=0; i<_whole_part_cnt; i++)
        {
            if(_bupDirN[_bupPartInfo[_bup_partinfo_id].part_id] > 0)
            {
                break;
            }

            _bup_partinfo_id++;
            if(_bup_partinfo_id > (_whole_part_cnt-1))
            {
                _bup_partinfo_id = 0;
            }
        }
    }


    while(1)
    {
        char title[LINE_LEN];

        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 28/*w*/, 13/*h*/, 1/*cols*/);

        sprintf(title, "Save Data List [%s]", _bup_part_abbr_names[_bupPartInfo[_bup_partinfo_id].part_id]);
        menu_set_title(mset, title);
        if(sel_type == BUP_SELECT_MULTIPLE)
        {
            menu_set_help(mset, "A/C: Confirm, B: Exit, L/R: Change save device, X: (Un)select one, Y: (Un)select all");
        }
        else if(sel_type == BUP_SELECT_SINGLE)
        {
            menu_set_help(mset, "A/C: Confirm, B: Exit, L/R: Change save device");
        }
        else //if((sel_type == BUP_SELECT_SINGLE_CART) || (sel_type == BUP_SELECT_SINGLE_BRAM))
        {
            menu_set_help(mset, "A/C: Confirm, B: Exit");
        }
        menu_set_erase_on_exit(mset, 1);
        menu_set_callback_getstr(mset, bup_list_str_callback);
        if((sel_type == BUP_SELECT_SINGLE_CART) || (sel_type == BUP_SELECT_SINGLE_BRAM))
        {
            menu_set_features(mset, MENU_FEATURES_TEXTCENTER);
        }
        else
        {
            menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER);
        }

        menu_current[_bupPartInfo[_bup_partinfo_id].part_id]
            = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, 
                    _bupDirN[_bupPartInfo[_bup_partinfo_id].part_id], 
                    mset, 
                    menu_current[_bupPartInfo[_bup_partinfo_id].part_id], 
                    sel_type == BUP_SELECT_MULTIPLE ? _bup_list_selection[_bupPartInfo[_bup_partinfo_id].part_id] : NULL);

        /* Store selected item to global variable. */
        _bup_dat_id = menu_current[_bupPartInfo[_bup_partinfo_id].part_id];

        /* If available, update context data. */
        if(menu_ctx)
        {
            menu_ctx->partinfo_id = _bup_partinfo_id;
            menu_ctx->current_save_id[_bupPartInfo[_bup_partinfo_id].part_id] = menu_current[_bupPartInfo[_bup_partinfo_id].part_id];
        }
        if((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
        || (mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER))
        {
            /* Change device with shoulder Left/Right buttons.
             * And, automatically skip device/partition that are empty : this was
             * initially added in order to avoid a bug in menu module. The bug is
             * now fixed, but this feature is kept, because it's actually quite convenient.
             */
            int direction = (mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER ? -1 : 1);

            for(i=0; i<_whole_part_cnt; i++)
            {
                _bup_partinfo_id += direction;
                if(_bup_partinfo_id < 0)
                {
                    _bup_partinfo_id = _whole_part_cnt-1;
                }
                if(_bup_partinfo_id > (_whole_part_cnt-1))
                {
                    _bup_partinfo_id = 0;
                }

                if(_bupDirN[_bupPartInfo[_bup_partinfo_id].part_id] > 0)
                {
                    break;
                }
            }
        }
        else if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            clear_screen();

            return 0;
        }
        else
        {
            for(i=0; i<BUP_PARTITION_CNT; i++)
            {
                for(j=0; j<BUP_DIR_SIZE; j++)
                {
                    if(_bup_list_selection[i][j])
                    {
                        _list_sel_stats.sel_cnt[i]++;
                        _list_sel_stats.sel_cnt_total++;
                    }
                }
            }
            for(i=0; i<BUP_PARTITION_CNT; i++)
            {
                if(_list_sel_stats.sel_cnt[i] != 0)
                {
                    _list_sel_stats.dev_list |= (1 << i);
                    _list_sel_stats.dev_cnt++;
                }
            }

            if(sel_type == BUP_SELECT_MULTIPLE)
            {
                conio_printf(2, 4, COLOR_SILVER, "* You have selected %4d element(s)*", _list_sel_stats.sel_cnt_total);
                conio_printf(2, 5, COLOR_SILVER, "*  Selection is over %d device(s).  *", _list_sel_stats.dev_cnt);
            }

            /* Don't forget to clear bottom of screen. */
            bup_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);

            return 1;
        }
    }

    return 1;
}

/* Extra (debug) menu enable flag. */
int _bup_extramenu_use = 0;

int main_menu_exec(menu_settings_t* mset, menu_items_t* mitems, int init_tag)
{
    /* 0:main menu, 1:sub menu. */
    int main_menu_page = 0;
    int i;

    int menu_selected;

    clear_screen();

    while(1)
    {
        int row = 17;

        menu_reset_settings(mset);
        menu_reset_items(mitems);

        menu_set_callback_vblank(mset, bup_anim_starfield);

        /* Don't auto-erase menu, because we will
         * do this manually after that when it's needed.
         */
        menu_set_erase_on_exit(mset, 0);
        menu_set_title(mset, "Save Data Manager");

        /* Disable sub menu when running as Fenrir Save Data Manager
         * or in limited mode, because nearly all the functions there
         * are either not available, or their usage doesn't make sense.
         */
        if((_fenrir_sdm) || (_limited_sdm))
        {
            menu_set_help(mset, "A/C: Enter, B: Exit");
        }
        else
        {
            menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER);
            menu_set_help(mset, "A/C: Enter, B: Exit, L/R: Toggle menus");
        }

        /* Check if Satiator/Gamer's Cartridge/etc/Fenrir provides
         * a write access to its mass storage media.
         */
        int enable_biosode = 0;
        if(SC_FILE_IS_VALID(sc_get_sdcard_devid()))
        {
            enable_biosode = 1;
        }
        else
        {
            sc_device_t dev;
            scf_dev_status(&dev, SC_FILE_EXTRA);
            if((dev.available) && (!dev.read_only))
            {
                enable_biosode = 1;
            }
        }

        if(main_menu_page == 0)
        {
            int firmware_available = (memcmp((void*)(0x02000000), (void*)"SEGA SEGASATURN ", 16) == 0 ? 1 : 0);

            /* If flasher feature available, 
             * and boot cartridge inserted, 
             * show how to access flasher utility.
             */
            if((cdrom_available) && (firmware_available))
            {
                conio_printf(2, row++, COLOR_YELLOW, "Cartridge Flasher Help :            ");
                conio_printf(2, row++, COLOR_WHITE , " Please press R shoulder button in  ");
                conio_printf(2, row++, COLOR_WHITE , " order to show sub menu, then       ");
                conio_printf(2, row++, COLOR_WHITE , " select \"Firmware Flasher\" entry.   ");
                conio_printf(2, row++, COLOR_WHITE , empty_line);
                conio_printf(2, row++, COLOR_WHITE , empty_line);
                conio_printf(2, row++, COLOR_WHITE , empty_line);
                conio_printf(2, row++, COLOR_WHITE , empty_line);
            }
            else
            {
                /* Else, don't forget to clear display. */
                for(i=0; i<8; i++)
                {
                    conio_printf(2, row++, COLOR_TRANSPARENT, empty_line);
                }
            }

            /* Show main menu items. */
            menu_add_item(mitems, "     Copy     ", TAG_COPY      ); menu_add_item(mitems, "     Move     ", TAG_MOVE      );
            menu_add_item(mitems, "    Delete    ", TAG_DELETE    ); menu_add_item(mitems, "   Hex-Edit   ", TAG_HEXEDIT   );
            menu_add_item(mitems, "      Device M", MENU_TAG_TITLE); menu_add_item(mitems, "anagement     ", MENU_TAG_TITLE);
            menu_add_item(mitems, "    Format    ", TAG_FORMAT    ); menu_add_item(mitems, "     Info     ", TAG_INFO      );
            menu_add_item(mitems, "            Ot", MENU_TAG_TITLE); menu_add_item(mitems, "her           ", MENU_TAG_TITLE);

            const int import_disable = 0;
            menu_add_item(mitems, "      Save Dat", (import_disable ? MENU_TAG_GREYOUT : TAG_IMPORT)); menu_add_item(mitems, "a Import      ", (import_disable ? MENU_TAG_GREYOUT : TAG_IMPORT));

            /* Disable Save Data Dump item during limited mode, because this
             * focus only in copying saves from CD-ROM to Saturn or device.
             */
            int dump_disable = (_limited_sdm ? 1 : 0);
            menu_add_item(mitems, "      Save Dat", (dump_disable ? MENU_TAG_GREYOUT : TAG_DUMP)); menu_add_item(mitems, "a Export      ", (dump_disable ? MENU_TAG_GREYOUT : TAG_DUMP));

            if(enable_biosode)
            {
                /* Extra feature for Fenrir or other extension
                 * device : allow to dump BIOS to SD card.
                 * 
                 * Menu features only one extra line, so in the case
                 * both "BIOS Dump" and "Dezaemon 2 SGM" features are
                 * available, then let's split this line in two.
                 */
                if(_deza_sgm_available)
                {
                    menu_add_item(mitems, "  BIOS Dump   ", TAG_BIOSODE);
                    menu_add_item(mitems, "  Deza2 SGM   ", TAG_DEZASGM);
                }
                else
                {
                    menu_add_item(mitems, "         BIOS ", TAG_BIOSODE   ); menu_add_item(mitems, "Dump          ", TAG_BIOSODE   );
                }
            }
            else
            {
                if(_deza_sgm_available)
                {
                    menu_add_item(mitems, "       Dezaemo", TAG_DEZASGM   ); menu_add_item(mitems, "n 2 SGM       ", TAG_DEZASGM   );
                }
            }

            menu_add_item(mitems, "           Cre", TAG_CREDITS   ); menu_add_item(mitems, "dits          ", TAG_CREDITS   );

            menu_selected = init_tag;
            if(!(_whole_part_list & ((1 << PART_ID_CART) | (1 << PART_ID_FDD1) | (1 << PART_ID_FDD2))))
            {
                /* In the case only one device is available, 
                 * disable copy and move features.
                 * There's no need to modify menu_selected initial
                 * value : it is done in menu_start function.
                 */
                menu_disable_item(mitems, TAG_COPY);
                menu_disable_item(mitems, TAG_MOVE);
            }
            menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 28/*w*/, (_deza_sgm_available || enable_biosode ? 10 : 9)/*h*/, 2/*cols*/);
        }
        else
        {
            /* Enable extra stuff by using menu keys combo. */
            menu_set_callback_keycombo(mset, NULL/*no callback*/, 0x2E8AFCD4);

            /* Display _bupConf contents on screen. */
            clear_screen();
            conio_printf(2, row++, COLOR_YELLOW, "BUP_Init results :");
            for(i=0; i<BUP_DEVICE_CNT; i++)
            {
                conio_printf(2, row++, COLOR_WHITE, " | [%d] partition=%d, unit_id=%d", i, _bupConf[i].partition, _bupConf[i].unit_id);
            }

            conio_printf(2, row++, COLOR_YELLOW, "Partitions infos :");
            for(i=0; i<BUP_PARTITION_CNT; i++)
            {
                BupPartInfo* bup = _bupPartInfo+i;

                conio_printf(2, row, COLOR_WHITE, " | ");
                conio_printf(((i % 2) == 0 ? 5 : 20), row, COLOR_WHITE, "[%d] %d@%d cnt:%d", 
                    i, 
                    bup->device_id, bup->device_part_id, 
                    bup->part_cnt);
                if((i % 2) == 1)
                {
                    row++;
                }
            }

            /* Show sub menu items. */
            menu_add_item(mitems, "       Start Homebrew       ", TAG_HOMEBREW);
            if(cdrom_available)
            {
                menu_add_item(mitems, "      Firmware Flasher      ", TAG_FLASHER);
            }
            if(_bup_extramenu_use)
            {
                menu_add_item(mitems, "     *** Extra Stuff ***    ", TAG_EXTRA);
            }
            menu_add_item(mitems, "      Debug Information     ", TAG_DBGINFO);
            menu_add_item(mitems, "   Cartridge ROM HexView    ", TAG_ROMHEXVIEW);
            menu_add_item(mitems, "     Commslink Transfer     ", TAG_COMMSLINK);
            menu_add_item(mitems, "            Exit            ", TAG_EXIT);

            /* Must set same dimensions as in main menu, 
             * but columns count can be different.
             */
            menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 28/*w*/, (_deza_sgm_available || enable_biosode ? 10 : 9)/*h*/, 1/*cols*/);
            menu_selected = TAG_HOMEBREW;
        }

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            /* If main menu, show exit menu.
             * If sub menu, return back to main menu.
             */
            if(main_menu_page == 0)
            {
                clear_screen();
                exit_menu();
            }
            else
            {
                main_menu_page = 0;
            }
        }
        else
        {
            if((mset->exit_code == MENU_EXIT_PREVPAGE_ARROW)
            || (mset->exit_code == MENU_EXIT_NEXTPAGE_ARROW)
            || (mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
            || (mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER))
            {
                main_menu_page = (main_menu_page ? 0 : 1);
            }
            else if(mset->exit_code == MENU_EXIT_KEYCOMBO)
            {
                /* Enable/disable extra menu when key combo is entered. */
                _bup_extramenu_use = (_bup_extramenu_use ? 0 : 1);
            }
            else
            {
                clear_screen();
                if(menu_selected == TAG_FLASHER)
                {
                    flasher_exec();

                    /* Restore application title to Save Data Manager one. */
                    disp_app_title();

                    /* Re-init BUP library.
                     * It is needed in order to initialize
                     * SatCom debugger on vmem side.
                     */
                    buplib_init();
                }
                else if(menu_selected == TAG_HOMEBREW)
                {
                    start_homebrew();

                    /* Re-init BUP library.
                     * It is needed in order to initialize
                     * SatCom debugger on vmem side.
                     */
                    buplib_init();
                }
                else if(menu_selected == TAG_EXTRA)
                {
                    extra_menu();

                    /* Re-init BUP library.
                     * It is needed in order to initialize
                     * SatCom debugger on vmem side.
                     */
                    buplib_init();
                }
                else if(menu_selected == TAG_DBGINFO)
                {
                    dbginfo_show();

                    /* Re-init BUP library.
                     * It is needed in order to initialize
                     * SatCom debugger on vmem side.
                     */
                    buplib_init();
                }
                else if(menu_selected == TAG_ROMHEXVIEW)
                {
                    cartrom_hexview();
                }
                else if(menu_selected == TAG_COMMSLINK)
                {
                    do_commslink();
                }
                else if(menu_selected == TAG_EXIT)
                {
                    exit_menu();
                }
                else
                {
                    return menu_selected;
                }
            }
        }
    }

    /* Dummy return statement, code never reaches here. */
    return TAG_EXIT;
}


int overwrite_check_menu(menu_settings_t* mset, menu_items_t* mitems, int init_tag)
{
    int menu_selected;

    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, "Overwrite ?");
    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_add_item(mitems, "      No      ", TAG_OVERWRITE_NO    );
    menu_add_item(mitems, "      Yes     ", TAG_OVERWRITE_YES   );
    menu_add_item(mitems, "    Always    ", TAG_OVERWRITE_ALWAYS);

    menu_set_pos(mset, -1/*x0*/, 10*CONIO_CHAR_H/*y0*/, 14/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);

    menu_selected = init_tag;
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return TAG_OVERWRITE_NO;
    }

    return menu_selected;
}


int format_check_menu(menu_settings_t* mset, menu_items_t* mitems)
{
    int menu_selected;

    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, "Format device ?");
    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_add_item(mitems, "      Cancel     ", TAG_FORMAT_CANCEL);
    menu_add_item(mitems, "      Format     ", TAG_FORMAT_YES   );

    menu_set_pos(mset, -1/*x0*/, 10*CONIO_CHAR_H/*y0*/, 17/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);

    menu_selected = TAG_FORMAT_CANCEL;
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return TAG_FORMAT_CANCEL;
    }

    return menu_selected;
}


int devfull_check_menu(menu_settings_t* mset, menu_items_t* mitems)
{
    int menu_selected;

    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, "Device full !");
    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_add_item(mitems, "       Cancel      ", TAG_DEVFULL_CANCEL  );
    menu_add_item(mitems, "  Continue Anyway  ", TAG_DEVFULL_CONTINUE);

    menu_set_pos(mset, -1/*x0*/, 10*CONIO_CHAR_H/*y0*/, 19/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);

    menu_selected = TAG_DEVFULL_CONTINUE;
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return TAG_DEVFULL_CANCEL;
    }

    return menu_selected;
}


int confirm_menu_exec(menu_settings_t* mset, menu_items_t* mitems, char* title)
{
    int menu_selected;

    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, title);
    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_add_item(mitems, "     Yes      ", 0/*tag*/);
    menu_add_item(mitems, "      No      ", 1/*tag*/);

    menu_set_pos(mset, -1/*x0*/, 8*CONIO_CHAR_H/*y0*/, 28/*w*/,  0/*h*/, 2/*cols*/);
    menu_set_erase_on_exit(mset, 1);

    menu_selected = 1; /* Initial item is "No" */
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if((mset->exit_code == MENU_EXIT_CANCEL)
    || (menu_selected == 1                 ))
    {
        return 0;
    }

    return 1;
}


int bup_partition_select(menu_settings_t* mset, menu_items_t* mitems, char* title, int dev_use, int full_dev_list, BupPartInfo** bup)
{
    int menu_selected;
    int i;

    /* Initialize output parameter. */
    *bup = _bupPartInfo+PART_ID_BRAM;

    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, title);
    menu_set_help(mset, "A/C: Select device, B: Cancel");

    if(full_dev_list)
    {
        /* Display all devices (including unavailable ones) in menu. */
        menu_add_item(mitems, "    Internal Storage    ", (dev_use & (1 << PART_ID_BRAM) ? PART_ID_BRAM : MENU_TAG_GREYOUT)/*tag*/);
        menu_add_item(mitems, "        Cartridge       ", (dev_use & (1 << PART_ID_CART) ? PART_ID_CART : MENU_TAG_GREYOUT)/*tag*/);
        menu_add_item(mitems, "   FDD, Partition 1/2   ", (dev_use & (1 << PART_ID_FDD1) ? PART_ID_FDD1 : MENU_TAG_GREYOUT)/*tag*/);
        menu_add_item(mitems, "   FDD, Partition 2/2   ", (dev_use & (1 << PART_ID_FDD2) ? PART_ID_FDD2 : MENU_TAG_GREYOUT)/*tag*/);
    }
    else
    {
        /* Fill menu with available devices only. */
        for(i=0; i<BUP_PARTITION_CNT; i++)
        {
            if(_bupPartInfo[i].part_cnt)
            {
                switch(_bupPartInfo[i].part_id)
                {
                case(PART_ID_BRAM):
                    menu_add_item(mitems, "    Internal Storage    ", _bupPartInfo[i].part_id/*tag*/);
                    if(!(dev_use & (1 << PART_ID_BRAM)))
                    {
                        menu_disable_item(mitems, _bupPartInfo[i].part_id);
                    }

                    break;

                case(PART_ID_CART):
                    menu_add_item(mitems, "        Cartridge       ", _bupPartInfo[i].part_id/*tag*/);
                    if(!(dev_use & (1 << PART_ID_CART)))
                    {
                        menu_disable_item(mitems, _bupPartInfo[i].part_id);
                    }
                    break;

                case(PART_ID_FDD1):
                    menu_add_item(mitems, "   FDD, Partition 1/2   ", _bupPartInfo[i].part_id/*tag*/);
                    if(!(dev_use & (1 << PART_ID_FDD1)))
                    {
                        menu_disable_item(mitems, _bupPartInfo[i].part_id);
                    }
                    break;

                default:
                case(PART_ID_FDD2):
                    menu_add_item(mitems, "   FDD, Partition 2/2   ", _bupPartInfo[i].part_id/*tag*/);
                    if(!(dev_use & (1 << PART_ID_FDD2)))
                    {
                        menu_disable_item(mitems, _bupPartInfo[i].part_id);
                    }
                }
            }
        }
    }

    /* Emergency exit if no device listed in menu. */
    if(mitems->item_count == 0)
    {
        return 0;
    }

    /* Set initial item as first available one. */
    menu_selected = 0;
    for(i=0; i<BUP_PARTITION_CNT; i++)
    {
        if(dev_use & (1 << i))
        {
            menu_selected = i;
            break;
        }
    }

    menu_set_pos(mset, -1/*x0*/, 7*CONIO_CHAR_H/*y0*/, 25/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return 0;
    }

    /* Set output parameter.
     *
     * When undetected partition is selected, returning pointer
     * to its partition structure doesn't makes sense, hence a "dummy"
     * partition structure (the last one in the array, which is free
     * when there is/are undetected partitions) is filled with selected
     * partition informations.
     * This tweak is not very elegant, but should work without troubles
     * since it usually not happens.
     *
     * Note : unavailable devices can be listed by setting the parameters :
     *  - dev_use to PART_ALL
     *  - full_dev_list to 1
     *
     * Remark : When available partition is selected, care need to be taken
     * that target partition information is not necessarily in partition array
     * from selected ID, hence the for loop below.
     * Example : BRAM detected, CART not detected, FDD detected
     *           -> partition information for FDD1 is stored just after BRAM.
     */
    *bup = NULL;
    for(i=0; i<_whole_part_cnt; i++)
    {
        if((_bupPartInfo+i)->part_id == menu_selected)
        {
            /* Detected partition selected. */
            *bup = _bupPartInfo+i;

            break;
        }
    }

    if(*bup == NULL)
    {
        /* Undetected partition selected. */
        *bup = _bupPartInfo + (BUP_PARTITION_CNT-1);
        memset(*bup, 0, sizeof(BupPartInfo));

        /* Set device ID (0:BRAM, 1:CART, 2~:FDD). */
        (*bup)->device_id = (menu_selected >= DEVICE_ID_DISC ? DEVICE_ID_DISC : menu_selected);

        /* Set partition ID. */
        (*bup)->part_id = menu_selected;

        /* Unrecognized device, so set partition count to zero. */
        (*bup)->part_cnt = 0;

        /* Set partition ID with device : always zero, 
         * except "1" when FDD2 partition is selected.
         */
        (*bup)->device_part_id = (menu_selected == PART_ID_FDD2 ? 1 : 0);

        /* Set abbreviated name (4 characters) for current partition. */
        (*bup)->part_abbr_name = _bup_part_abbr_names[(*bup)->part_id];
    }

    return 1;
}


