/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#ifndef _BUP_MENUS_H_
#define _BUP_MENUS_H_

#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>

#include "../menu.h"
#include "../conio.h"

/* Stuff for import/dump functions. */
#include "bup_import.h"
#include "bup_dump.h"

/* Dezaemon 2 Save Game Manager. */
#include "deza_sgm.h"


/* Menu items and settings. */
extern menu_items_t    _menu_items;
extern menu_settings_t _menu_settings;

/* Save data library related global variables.
 * Only first partition is supported.
 */
extern unsigned long _bupLibrary[4096+1];
extern unsigned long _bupRamWork[2048+1];

/**
 * BupConf structure :
 * Defined on BUP library side.
 *
 * Contains information for each backup devices :
 *  - ID=0 : BRAM
 *  - ID=1 : CART
 *  - ID=2 : DISK
 *
 * Note : information for each device is stored at constant ID.
 *        So when cartridge is not detected, but FDD is then
 *        zeroed information is found in bupConf array at ID=1.
**/
#define BUP_DEVICE_CNT 3
extern BupConfig _bupConf[BUP_DEVICE_CNT];

/**
 * BupPartInfo structure :
 * Used in order to store information
 * for each backup partition available.
 * It is different from BupConfig structure for
 * example because one BupConfig structure may
 * bring several partitions when a FDD is used.
 *
 * Currently only up to 4 partitions are supported :
 *  - BRAM : 1 partition
 *  - CART : 0~1 partition
 *  - DISK : 0~2 partition(s)
 *
 * (Naming of this structure is different to usual
 * naming in order to look like similar "BupConfig"
 * structure naming used in BUP library.)
**/
typedef struct _BupPartInfo {

    /*
     * Device ID :
     *  DEVICE_ID_BRAM (0) : Built-in memory
     *  DEVICE_ID_CART (1) : Memory cartridge
     *  DEVICE_ID_DISC (2) : FDD
     *
     * Used in nearly all BUP_* functions in order
     * to specify which device to access.
     */
    unsigned char device_id;

    /*
     * Partition ID :
     *  PART_ID_BRAM (0) : Built-in memory
     *  PART_ID_CART (1) : Memory cartridge
     *  PART_ID_FDD1 (2) : FDD, partition 1
     *  PART_ID_FDD2 (3) : FDD, partition 2
     *
     * Used in menus in order to differentiate each partitions.
     */
    unsigned char part_id;

    /*
     * Partition count.
     *  0: No unavailable. Device unavailable.
     *  1: One partition available.
     *  2: Two partitions available (FDD only).
     *
     * Copied from parent BupConfig structure.
     */
    unsigned char part_cnt;

    /*
     * Partition ID within device.
     *  0: First partition.
     *  1: Second partition (FDD only).
     *
     * Used by BUP_SelPart when accessing partitions in FDD.
     */
    unsigned char device_part_id;

    /*
     * Abbreviated name (4 characters) for current partition.
     */
    const char* part_abbr_name;
} BupPartInfo;

/* ID for each backup device. */
#define DEVICE_ID_BRAM 0
#define DEVICE_ID_CART 1
#define DEVICE_ID_DISC 2

/* ID for each backup partition. */
#define PART_ID_BRAM 0
#define PART_ID_CART 1
#define PART_ID_FDD1 2
#define PART_ID_FDD2 3

/* Maximum partition count. */
#define BUP_PARTITION_CNT 4
extern BupPartInfo _bupPartInfo[BUP_PARTITION_CNT];


#define BUP_DIR_SIZE 400
extern BupStat   _bupStat[BUP_PARTITION_CNT];
extern BupDir    _bupDir [BUP_PARTITION_CNT][BUP_DIR_SIZE];
extern Sint32    _bupDirN[BUP_PARTITION_CNT];

/* Temporary buffer to handle save data filenames. */
#define BUP_DESC_LEN (32)
extern char _bup_desc[BUP_PARTITION_CNT][BUP_DIR_SIZE][BUP_DESC_LEN];


/*
 * Bitfield and associed count related to
 * backup partitions available.
 */
#define PART_NONE (0)
#define PART_ALL  ((1 << (PART_ID_BRAM)) | (1 << (PART_ID_CART)) | (1 << (PART_ID_FDD1)) | (1 << (PART_ID_FDD2)))
extern int _whole_part_cnt;
extern int _whole_part_list;

/* Global variables to handle selection status in list. */
extern int     _bup_list_selection[BUP_PARTITION_CNT][BUP_DIR_SIZE];
/* Abbreviated name (4 characters) for each partitions. */
extern const char* _bup_part_abbr_names[BUP_PARTITION_CNT];

/* Current partition informations structure ID
 * (0~_whole_part_cnt-1).
 *
 * This is NOT partition ID : partition ID
 * must be retrieved as described below :
 * _bupPartInfo[_bup_partinfo_id].part_id
 */
extern int _bup_partinfo_id;

/* Current backup data ID, when only single item
 * can be selected.
 */
extern int _bup_dat_id;


/**
 * bup_get_weekday
 * Return name of weekday from code stored in BupDate function.
 * Note #1 : weekday code = Sunday:0 - Saturday:6.
 * Note #2 : name is returned as a 3 letters string.
**/
const char* bup_get_weekday(unsigned char weekday);


/* Buffer to temporary store save data
 * Whole LRAM can be used, so save data > 1MB
 * is not supported.
 *
 * Since BIOS doesn't support saves larger than 256KB, 
 * this 1MB limitation shouldn't be a problem.
 */
extern unsigned char* _temp_buff;
#define TEMP_BUFF_ADR (0x00200000)
#define TEMP_BUFF_LEN (1024*1024)


/**
 * bup_selpart
 * Wrapper for BUP_SelPart function.
 * Needs to be called before call of other BUP functions.
**/
void bup_selpart(BupPartInfo* bup);




/**
 * exit_menu
 * Show exit menu.
**/
void exit_menu(void);

/**
 * wait_for_start.
 * Prompt user to push start button.
 * If msg parameter is set to NULL, 
 * defaut prompt message is displayed.
 *
 * Returns 1 if A/C/Start button pushed, zero else.
**/
int wait_for_start(char* msg);


/**
 * display_msg
 * Deneric display of message/error strings.
 * Return 1 if no error, zero else.
**/
int display_msg(int err, char *string, ...);

/**
 * display_buplib_error
 * Display save data library related errors.
 * Return 1 if no error, zero else.
**/
int display_buplib_error(Sint32 err);


/**
 * display_progress_ex/display_progress
 * Display progress bar.
 * Example at 42% : display_progress(42, 100).
**/
void display_progress_ex(int n, int total, int row);
#define BUP_PROGRESS_ROW 25
#define display_progress(_NUM_, _TOTAL_) display_progress_ex(_NUM_, _TOTAL_, BUP_PROGRESS_ROW)


/**
 * dbg_print.
 * Display debug message at the bottom of the screen.
 *
 * By default, messages are not displayed. It is required to call
 * dbg_print_enable function in order to enable debug messages display.
**/
void dbg_print_enable(void);
void dbg_print(char *string, ...);


/**
 * bup_list_exec
 * Display save data list for all devices, 
 * prompt user for save selection, (multiple
 * selection possible), then display details
 * about selection contents.
 *
 * Single/Multiple selection can be switched with
 *  sel_type parameter :
 *  BUP_SELECT_SINGLE      : single selection
 *  BUP_SELECT_MULTIPLE    : multiple selection
 *  BUP_SELECT_SINGLE_CART : single selection from cartridge only
 *  BUP_SELECT_SINGLE_BRAM : single selection from internal memory only
 *
 * "bup_list_menucontext_t* menu_ctx" input/output parameter
 * is optional and is used in order to restore/save
 * menu context information in order to restore menu
 * from a previous state.
 * Structure needs to be zero-memsetted on first usage.
 * If not used, parameter can be set to NULL.
 *
 * Return 0 if user cancels selection, 1 else.
 * Fill _list_sel_stats global variable with stats
 * about selection contents.
**/
typedef struct _sel_stats_t
{
    /* Number of items selected per devices. */
    int sel_cnt[BUP_PARTITION_CNT];

    /* Number of items selected on all devices. */
    int sel_cnt_total;

    /* Number of devices containing selected items. */
    int dev_cnt;
    /* Bitfield for use with DEVICE_* values. */
    int dev_list;
} sel_stats_t;
extern sel_stats_t _list_sel_stats;

/* Context data used in order to restore backup selection menu. */
typedef struct _bup_list_menucontext_t
{
    /* ID of partition currently selected. */
    int partinfo_id;

    /* Save ID currently selected. */
    int current_save_id[BUP_PARTITION_CNT];
} bup_list_menucontext_t;

#define BUP_SELECT_SINGLE      0
#define BUP_SELECT_MULTIPLE    1
#define BUP_SELECT_SINGLE_CART 2
#define BUP_SELECT_SINGLE_BRAM 3
int bup_list_exec(menu_settings_t* mset, int sel_type, bup_list_menucontext_t* menu_ctx);


/**
 * main_menu_exec
 * Display main menu and ask user about something to do.
 * Possible return values are #defined in TAG_ values.
**/
#define TAG_COPY        0
#define TAG_MOVE        1
#define TAG_DELETE      2
#define TAG_HEXEDIT     3
#define TAG_FORMAT      4
#define TAG_INFO        5
#define TAG_IMPORT      6
#define TAG_DUMP        7
#define TAG_BIOSODE     8
#define TAG_DEZASGM     9
#define TAG_CREDITS    10

/* Sub menu tag values. */
#define TAG_HOMEBREW   50
#define TAG_FLASHER    51
#define TAG_EXTRA      52
#define TAG_DBGINFO    53
#define TAG_ROMHEXVIEW 54
#define TAG_COMMSLINK  55

#define TAG_EXIT       90
int main_menu_exec(menu_settings_t* mset, menu_items_t* mitems, int init_tag);


/**
 * overwrite_check_menu
 * Ask overwrite yes/no/always to user.
 * Possible return values are #defined in TAG_OVERWRITE_ values.
**/
#define TAG_OVERWRITE_NO     0
#define TAG_OVERWRITE_YES    1
#define TAG_OVERWRITE_ALWAYS 2
int overwrite_check_menu(menu_settings_t* mset, menu_items_t* mitems, int init_tag);


/**
 * format_check_menu
 * Ask if need to format device or not to user.
 * Possible return values are #defined in TAG_FORMAT_ values.
**/
#define TAG_FORMAT_CANCEL 0
#define TAG_FORMAT_YES    1
int format_check_menu(menu_settings_t* mset, menu_items_t* mitems);


/**
 * devfull_check_menu
 * Ask if need to cancel or continue anyway when "not enough memory" error happened.
 * Possible return values are #defined in TAG_DEVFULL_ values.
**/
#define TAG_DEVFULL_CANCEL   0
#define TAG_DEVFULL_CONTINUE 1
int devfull_check_menu(menu_settings_t* mset, menu_items_t* mitems);


/**
 * confirm_menu_exec
 * Prompt yes/no to user.
 *
 * Return 0 if user cancels, 1 else.
**/
int confirm_menu_exec(menu_settings_t* mset, menu_items_t* mitems, char* title);


/**
 * bup_partition_select
 * Allow user to select backup partition.
 *
 * Selected partition is returned in "BupPartInfo** bup" output parameter.
 *
 * Return 0 if user cancels, 1 else.
**/
int bup_partition_select(menu_settings_t* mset, menu_items_t* mitems, char* title, int dev_use, int full_dev_list, BupPartInfo** bup);


#endif /* _BUP_MENUS_H_ */

