/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <sc_saturn.h>

#include "deza_screens.h"

/* Inclusion of DZ2SGM stuff. */
#include "deza_sgm.h"

/* Buffer declaration :
 *  extern unsigned char* _temp_buff;
 *  #define TEMP_BUFF_ADR (0x00200000)
 *  #define TEMP_BUFF_LEN (1024*1024)
 */
#include "bup_menus.h"

/**
 * Include save data manager global stuff :
 *  - void clear_screen(void)
**/
#include "main.h"


/**
 * Notes about RAM usage by DZ2SGM screenshot module :
 *  - High Work RAM is used a work RAM for compressed data.
 *  - Low Work RAM is basically used as buffer for recently read screenshot data.
 *
 * Low work RAM memory management :
 *  - Allow to backup/restore logo image data.
 *  - Remember compressed data for latest screenshots.
 *  - Allow to retrieve compressed screenshot data, if available.
 *
 * Low work RAM definitions :
 *  - extern unsigned char* _temp_buff;
 *  - #define TEMP_BUFF_ADR (0x00200000)
 *  - #define TEMP_BUFF_LEN (1024*1024)
 *
 * Low work RAM memory map :
 *  00200000 - 00222FFF [  140KB] : Logo image (320 x 224, RGB555, tiled)
 *  00223000 - 002FFBFF [  883KB] : Latest screenshot buffer area (circular buffer)
 *  002FFC00 - 002FFFF7 [ 1016 B] : Reserved for future use.
 *  002FFFF8 - 002FFFFF [    8 B] : Debugger global data (pointer, then magic number).
 *
 *
 * High work RAM is used in order to store compressed data just after reading :
 *  - #define EXTRACT_BUFFER_LEN MAX(256*1024, SGM_SCR_BUFFER_LEN)
 *  - extern unsigned char _extract_buff[];
 * This is used in order to read compressed screenshot, hence maximum size
 * is 165KB + few bytes for LZF related header, which consequently fits
 * into this buffer.
 *
 * Remark : due to circular nature of buffer in Low Work RAM, it is necessary to first
 *          read compressed data to High Work RAM, and then copy it (sometimes continously, 
 *          sometimes in two times if near boundary of buffer) to circular buffer.
**/
#define SGM_LRAM_LOGOBACKUP_PTR  ((unsigned char*)TEMP_BUFF_ADR)
#define SGM_LRAM_LOGOBACKUP_LEN  (LOGO_WIDTH * LOGO_HEIGHT * sizeof(unsigned short))

#define SGM_LRAM_SCREENSBUFF_PTR (SGM_LRAM_LOGOBACKUP_PTR + SGM_LRAM_LOGOBACKUP_LEN)
#define SGM_LRAM_SCREENSBUFF_LEN (TEMP_BUFF_LEN - SGM_LRAM_LOGOBACKUP_LEN - 1024)




/**
 * Screenshot buffering module.
 * 882KB in LRAM are used for keeping screenshot compressed data in memory :
 *  - SGM_LRAM_SCREENSBUFF_PTR (SGM_LRAM_LOGOBACKUP_PTR + SGM_LRAM_LOGOBACKUP_LEN)
 *  - SGM_LRAM_SCREENSBUFF_LEN (TEMP_BUFF_LEN - SGM_LRAM_LOGOBACKUP_LEN - 1024)
 *
 * It is implemented by using array in order to keep screenshot informations, and
 * circular buffer in order to keep screenshot data itself.
 *
 * The module keeps buffering screenshots even after leaving screenshot page, so
 * that it's possible to return back to a previously viewed entry without having to
 * re-read screenshot file(s) previously viewed.
 *
 * Because non-contingous nature of circular buffer, it's not possible to do
 * processing (such as data decompression), so that compressed screenshot needs
 * to be first decompressed (which is done on HRAM), and then processed.
**/

// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

typedef struct _scrbuff_entry_t
{
    /* Screenshot ID : it is unique for all screenshots
     * in Dezaemon 2 Save Game Manager :
     *  - bit 24-31 : SGM ID
     *  - bit 16-23 : Game ID
     *  - bit  8-16 : unused (set to zero)
     *  - bit  0- 7 : Screenshot ID
     */
    unsigned long scr_id;
#define DZ2SCR_GET_ID(_ENTRY_, _SCR_ID_) ((((_ENTRY_)->sgm_id) << 24) | (((_ENTRY_)->id) << 16) | ((_SCR_ID_) << 0))

    /* Screenshot compressed data informations :
     *  - Offset within circular buffer.
     *  - Size.
     */
    unsigned long circbuff_offset;
    unsigned long datalen;


    /* CRC usage flag. It is used for debug
     * purpose only, to be sure that correct
     * data is extracted from circular buffer.
     */
//#define SCRBUFF_USE_CRC

#ifdef SCRBUFF_USE_CRC
    /* CRC, in order to verify integrity
     * when taking from circular buffer.
     */
    unsigned long crc32;
#endif // SCRBUFF_USE_CRC
} SC_ALIGN_4_BYTES(scrbuff_entry_t);


typedef struct _scrbuff_data_t
{
    /* Circular buffer informations.
     * These are offsets within circular buffer, not pointers.
     */
    unsigned long circbuff_rd_ptr;
    unsigned long circbuff_wr_ptr;

    /* Screenshot entries informations :
     *  - ID of latest entry filled so far.
     *  - Entries count.
     */
    int latest_id;
    int entries_cnt;

    /* Informations for each entries.
     *
     * Entries list size is set in order to get
     * a good balance between HRAM usage
     * (64 entries * 16 bytes = 1KB), time to process
     * it when it is full, and typical count of images
     * that can fit in LRAM.
     * Typical size of compressed image is between 20KB
     * and 80KB, vs 883KB on HRAM. So there's no need to
     * set a significantly large count for entries list.
     */
#define SCRBUFF_MAXCNT 64
    scrbuff_entry_t entries[SCRBUFF_MAXCNT];
} SC_ALIGN_4_BYTES(scrbuff_data_t);


// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


/* Screenshot buffering data :
 *  - Pointers, sizes, etc (few KBs) is declared here (HRAM)
 *  - BUffer itself (800+ KB) is declared on LRAM.
 */
scrbuff_data_t _scrbuff_dat;


/**
 * Add screenshot data to buffer.
**/
void scrbuff_put(deza2_entry_t* e, int entry_scr_id, unsigned char* data_ptr, unsigned long data_len)
{
    scrbuff_data_t* scrbuff = &_scrbuff_dat;
    int i, id;
    scrbuff_entry_t* s = NULL;

    unsigned long scr_id = DZ2SCR_GET_ID(e, entry_scr_id);

    /* Verify that screenshot is not already registered. */
    id = scrbuff->latest_id;
    for(i=0; i<scrbuff->entries_cnt; i++)
    {
        s = scrbuff->entries+id;
        if(id == 0)
        {
            id = SCRBUFF_MAXCNT-1;
        }
        else
        {
            id--;
        }

        if(s->scr_id == scr_id)
        {
            /* Entry already registered, hence no need
             * to re-register it.
             */
            return;
        }
    }


    /* Remove enough oldest entries as necessary
     * in order to add current entry.
     */
    while(1)
    {
        /* Verify circular buffer free space. */
        unsigned long freespace = SGM_LRAM_SCREENSBUFF_LEN;
        if(scrbuff->entries_cnt != 0)
        {
            if(scrbuff->circbuff_wr_ptr >= scrbuff->circbuff_rd_ptr)
            {
                freespace = SGM_LRAM_SCREENSBUFF_LEN - scrbuff->circbuff_wr_ptr + scrbuff->circbuff_rd_ptr - 1;
            }
            else
            {
                freespace = scrbuff->circbuff_rd_ptr - scrbuff->circbuff_wr_ptr - 1;
            }
        }

        if(freespace >= data_len)
        {
            /* If enough space in circular buffer,
             * there's no need to delete oldest entries.
             */
            break;
        }

        /* Select oldest entry, and free it. */
        id = (scrbuff->latest_id + (SCRBUFF_MAXCNT - scrbuff->entries_cnt) ) % SCRBUFF_MAXCNT;
        s = scrbuff->entries+id;
        scrbuff->circbuff_rd_ptr = (scrbuff->circbuff_rd_ptr + s->datalen) % SGM_LRAM_SCREENSBUFF_LEN;

        /* Since circular buffer is larger than maximum size of
         * compressed screenshot, there's no need to worry about
         * getting negative entries count after decrement.
         */
        memset(s, 0, sizeof(scrbuff_entry_t));
        scrbuff->entries_cnt--;
    }


    /* Create new entry. */
    if(scrbuff->entries_cnt < SCRBUFF_MAXCNT)
    {
        scrbuff->entries_cnt++;
    }
    id = (scrbuff->latest_id+1) % SCRBUFF_MAXCNT;
    scrbuff->latest_id = id;

    s = scrbuff->entries+id;
    memset(s, 0, sizeof(scrbuff_entry_t));
    s->scr_id = scr_id;
    s->datalen = data_len;
    s->circbuff_offset = scrbuff->circbuff_wr_ptr;
#ifdef SCRBUFF_USE_CRC
    s->crc32 = crc32_calc((const void *)data_ptr, s->datalen);
#endif // SCRBUFF_USE_CRC

    /* Copy data into circular buffer. */
    unsigned long offs = scrbuff->circbuff_wr_ptr;
    unsigned char* ptr = (unsigned char*)SGM_LRAM_SCREENSBUFF_PTR;

    if((offs + s->datalen) <= SGM_LRAM_SCREENSBUFF_LEN)
    { /* Contingous data split in circular buffer. */
        memcpy(ptr+offs, data_ptr, s->datalen);
    }
    else
    { /* Data split around circular buffer end. */
        unsigned long len1, len2;
        len1 = SGM_LRAM_SCREENSBUFF_LEN - offs;
        len2 = s->datalen - len1;
        memcpy(ptr+offs, data_ptr, len1);
        memcpy(ptr, data_ptr+len1, len2);
    }

    /* Update offset in cicrcular buffer. */
    scrbuff->circbuff_wr_ptr = (scrbuff->circbuff_wr_ptr + s->datalen) % SGM_LRAM_SCREENSBUFF_LEN;
}


/**
 * Try to retrieve screenshot data from buffer.
 *
 * At most SGM_SCR_EXTRACTBUFF_LEN bytes are required on output
 * buffer, which is around 165 KB
 * (352*240 * 2 bytes per pixel) -> 165 KB.
 * Plus SC_COMP_HEADER_MAXSIZE (5 bytes), rounded to 512 bytes.
 *
 * Returns : 
 *  - Zero if not available.
 *  - Size of retrieved data, when succesfully retrieved.
**/
unsigned long scrbuff_get(deza2_entry_t* e, int entry_scr_id, unsigned char* dst_buff)
{
    scrbuff_data_t* scrbuff = &_scrbuff_dat;
    unsigned long ret = 0;
    int i, id;
    scrbuff_entry_t* s = NULL;

    /* Log circular buffer contents (debug purpose). */
    for(i=0; i<SCRBUFF_MAXCNT; i++)
    {
        s = scrbuff->entries+i;
    }


    unsigned long scr_id = DZ2SCR_GET_ID(e, entry_scr_id);

    /* Look for requested screenshot in buffered entries list. */
    id = scrbuff->latest_id;
    for(i=0; i<scrbuff->entries_cnt; i++)
    {
        s = scrbuff->entries+id;
        if(id == 0)
        {
            id = SCRBUFF_MAXCNT-1;
        }
        else
        {
            id--;
        }

        if(s->scr_id == scr_id)
        {
            /* Entry found. No need to search further.
             * Copy it from circular buffer to output buffer.
             */
            unsigned long offs = s->circbuff_offset;
            unsigned char* ptr = (unsigned char*)SGM_LRAM_SCREENSBUFF_PTR;

            if((offs + s->datalen) <= SGM_LRAM_SCREENSBUFF_LEN)
            { /* Contingous data split in circular buffer. */
                memcpy(dst_buff, ptr+offs, s->datalen);
            }
            else
            { /* Data split around circular buffer end. */
                unsigned long len1, len2;
                len1 = SGM_LRAM_SCREENSBUFF_LEN - offs;
                len2 = s->datalen - len1;
                memcpy(dst_buff, ptr+offs, len1);
                memcpy(dst_buff+len1, ptr, len2);
            }

#ifdef SCRBUFF_USE_CRC
            /* Verify CRC. */
            unsigned long crc32 = crc32_calc((const void *)dst_buff, s->datalen);
            if(crc32 != s->crc32)
            {
                /* CRC mismatch ?! This shouldn't happen. */
                conio_title_printf(2, 3, COLOR_ERROR, "Screenshot CRC ERROR !");
                wait_for_start(NULL);
            }
#endif // SCRBUFF_USE_CRC

            ret = s->datalen;
            break;
        }
    }

    return ret;
}





void deza_screens_init(void)
{
    scrbuff_data_t* scrbuff = &_scrbuff_dat;

    /* Initialize screenshot data buffering stuff. */
    memset(scrbuff, 0, sizeof(scrbuff_data_t));
    scrbuff->latest_id = SCRBUFF_MAXCNT - 1;
}


int deza_screens_disable_logo(void)
{
    int ret = logo_is_enabled();

    /* Wait for V-Blank. */
    WaitForVBLANKIn();
    WaitForVBLANKOut();

    /* Backup logo data from VDP1 to LRAM. */
    memcpy(SGM_LRAM_LOGOBACKUP_PTR, LOGO_DATAPTR, SGM_LRAM_LOGOBACKUP_LEN);

    /* Disable logo display. */
    logo_set_enable(0/*enable_flag*/);

    /* Redraw starfield (and only starfield) in order
     * to clear display of entries icons.
     * Reason : deza_screens_show function reads thumbnails file
     *          from CDROM/SD card, which takes few seconds in
     *          order to complete.
     */
    bup_anim_starfield();

    return ret;
}


void deza_screens_restore_logo(int logo_enable_flag)
{
    /* Wait for V-Blank. */
    WaitForVBLANKIn();
    WaitForVBLANKOut();

    /* Temporarily hide logo while restoring its bitmap data. */
    logo_set_enable(0);
    bup_anim_starfield();

    WaitForVBLANKIn();
    WaitForVBLANKOut();

    /* Restore logo data from LRAM to VDP1. */
    memcpy(LOGO_DATAPTR, SGM_LRAM_LOGOBACKUP_PTR, SGM_LRAM_LOGOBACKUP_LEN);

    /* Restore logo display flag to its previous state. */
    logo_set_enable(logo_enable_flag);

    /* Fill VDP1 commands list with starfield + logo (if enabled). */
    bup_anim_starfield();
}


int deza_screens_show(unsigned char dev_id, deza2_entry_t* entry)
{
    unsigned long datalen;
    int row = 4;
    pad_t* pad;
    int color_anim_id = 0;

    /* Verify that there are thumbnail(s) available. */
    if(entry->screenshot_cnt == 0)
    {
        conio_printf(2, row++, COLOR_ERROR , "*          ~~~ Error ~~~           *");
        conio_printf(2, row++, COLOR_ERROR , "* No screenshot available !        *");

        wait_for_start(NULL);
        clear_screen();

        /* Return to SGM list. */
        return 1;
    }

    /* Read screenshots thumnails. */
    sprintf(_deza_file_path, SGM_BASE_PATH "SGM%d/%03d/" SGM_SCRTHB_FILE, entry->sgm_id, entry->id);

    datalen = scf_read(dev_id, _deza_file_path, 0/*offset*/, SGM_ENTRY_THBMAXLEN, SGM_ENTRY_THBPTR);

    if(datalen == 0)
    {
        conio_printf(2, row++, COLOR_ERROR , "*          ~~~ Error ~~~           *");
        conio_printf(2, row++, COLOR_ERROR , "* Thumbnails file not found !      *");
        conio_printf(2, row++, COLOR_WHITE , blank_line);
        conio_printf(2, row++, COLOR_WHITE , "* File Path:                       *");
        conio_printf(2, row  , COLOR_WHITE , blank_line);
        conio_printf(4, row++, COLOR_YELLOW, "%s", _deza_file_path);

        wait_for_start(NULL);
        clear_screen();

        /* Return to SGM list. */
        return 1;
    }



    /* Wait until user releases pad buttons. */
    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        bup_anim_starfield();

        /* Poll joypad */
        pad = pad_read();

        if((pad->new[0] & (PAD_UP|PAD_DOWN|PAD_LEFT|PAD_RIGHT|PAD_A|PAD_B|PAD_C|PAD_X|PAD_Y|PAD_Z|PAD_START|PAD_LBUTTON|PAD_RBUTTON)) == 0)
        {
            break;
        }

    }


    /* Screenshots selection and display main loop :
     *  - Show thumbnails for each screenshots, 
     *    and allow to select screenshot to display.
     *  - Allow to zoom and scroll screenshot display.
     *
     * Keys assignment :
     *  - Return to SGM list with Z or B button
     *  - Scroll screenshot with arrow keys
     *  - Select thumbnails with L/R shoulder buttons
     *  - Change zoom factor with X/Y buttons
     *  - Z button : show help screen
     */
    int thb_id = 0;
    int thb_prev_id = -1;
    int thb_list_offset = 0;
    int scr_cnt = entry->screenshot_cnt;

    /* Definitions for screenshot (THB_DISP_*) :
     *  - Display top-left coordinates.
     *  - Display space between thumbnails.
     *  - Maximum thumbnails count displayable on screen.
     *
     * Definitions for thumbnails list (THB_DISP_*) :
     *  - Same stuff as for screenshot.
     */
#define THB_DISP_VERTICAL

#ifdef THB_DISP_VERTICAL // Screenshot and thumbnails vertical display

    /* Small gap between two thumbnails. */
#   define THB_DISP_SPACE     4
    /* Screenshot display top-left coordinates. */
#   define SCREEN_DISP_X0     SGM_LIST_X0
#   define SCREEN_DISP_Y0     (3*(CONIO_CHAR_H) + THB_DISP_SPACE - 1)
    /* Thumnails display left coordinate : displayed on screen right side. */
#   define THB_DISP_X0        (320 - SGM_LIST_X0 - SGM_THUMB_WIDTH)
    /* Screenshot display dimensions. */
#   define SCREEN_DISP_WIDTH  (THB_DISP_X0 - THB_DISP_SPACE - SCREEN_DISP_X0)
#   define SCREEN_DISP_HEIGHT (224 - SCREEN_DISP_Y0 - THB_DISP_SPACE - (3*CONIO_CHAR_H))
    /* The maximum count of thumbnails displayable simultaneously. */
#   define THB_DISP_MAXCNT    ((SCREEN_DISP_HEIGHT) / (SGM_THUMB_HEIGHT + THB_DISP_SPACE))
    /* Thumbnails display top coordinate : center justified with screenshot display area. */
#   define THB_DISP_Y0        (SCREEN_DISP_Y0 + (THB_DISP_SPACE/2) +  (SCREEN_DISP_HEIGHT - (THB_DISP_MAXCNT * (SGM_THUMB_HEIGHT + THB_DISP_SPACE))) / 2)

#else // Screenshot and thumbnails horizontal display
#   define SCREEN_DISP_X0     SGM_LIST_X0
#   define SCREEN_DISP_Y0     (SGM_LIST_Y0 + SGM_THUMB_WIDTH + THB_DISP_SPACE)
#   define SCREEN_DISP_WIDTH  (320 - (2*SCREEN_DISP_X0))
#   define SCREEN_DISP_HEIGHT (224 - SCREEN_DISP_Y0 - (2*CONIO_CHAR_H))

#   define THB_DISP_X0 SGM_LIST_X0
#   define THB_DISP_Y0 SGM_LIST_Y0
#   define THB_DISP_SPACE 4
#   define THB_DISP_MAXCNT ((320 - (2*THB_DISP_X0)) / (SGM_THUMB_WIDTH + THB_DISP_SPACE))

#endif


    /* Allow x.5 zoom factor. */
//#define ZOOMFACTOR_UNIT 2
#define ZOOMFACTOR_UNIT 16

    /* By default, set zoom factor to "1:1".
     * And, center image display when zoom factor
     * other than "Zoom All" is used.
     */
    int zoom_factor = ZOOMFACTOR_UNIT;
    int prev_zoom_factor = ZOOMFACTOR_UNIT;
    int scr_disp_x0 = (SGM_SCRSHOT_WIDTH  - SCREEN_DISP_WIDTH ) / 2;
    int scr_disp_y0 = (SGM_SCRSHOT_HEIGHT - SCREEN_DISP_HEIGHT) / 2;


    int i;
    /* L/R shoulder release counter, used to give a small delay
     * before loading screenshot, hence smooth scrolling within thumbnails.
     */
    int lr_release_cnt = 0;

    while(1)
    {
        /* Poll joypad */
        pad = pad_read();

        /* Quit program ? */
        if((pad->new[0] & PAD_A) && (pad->new[0] & PAD_B) && (pad->new[0] & PAD_C) && (pad->new[0] & PAD_START))
        { /* Exit to multiplayer. */
            /* Exit to multiplayer. */
            (**(void(**)(void))0x600026C)();
        }

        if((!(pad->new[0] & PAD_B)) && (pad->old[0] & PAD_B))
        { /* Exit screenshot view ? */
            /* Return to SGM list. */
            clear_screen();
            return 1;
        }

        if(pad->delta[0] & PAD_START)
        { /* Show help. */

            clear_screen();

            int row = 4;
            for(i=row; i<25; i++)
            {
                conio_printf(2, row++, COLOR_WHITE , blank_line);
            }
            row = 4;

            conio_printf(4, row++, COLOR_YELLOW, "   Dezaemon2 Save Game Manager   ");
            conio_printf(4, row++, COLOR_YELLOW, "     Screenshot Gallery Help     ");
            row++;
            row++;
            conio_printf(4, row++, COLOR_WHITE , " Arrows : Scroll display         ");
            row++;
            conio_printf(4, row++, COLOR_WHITE , " X/Y    : Change zoom factor     ");
            row++;
            conio_printf(4, row++, COLOR_WHITE , " Z      : Unused                 ");
            row++;
            conio_printf(4, row++, COLOR_WHITE , " B      : Return to games list   ");
            row++;
            conio_printf(4, row++, COLOR_WHITE , " L/R    : Select screenshot      ");
            row++;
            conio_printf(4, row++, COLOR_WHITE , " Start  : Show this help screen  ");

            wait_for_start(NULL);

            /* Wait until user releases pad buttons. */
            while(1)
            {
                /* Wait for V-Blank. */
                WaitForVBLANKIn();
                WaitForVBLANKOut();

                bup_anim_starfield();

                /* Poll joypad */
                pad = pad_read();

                if((pad->new[0] & (PAD_UP|PAD_DOWN|PAD_LEFT|PAD_RIGHT|PAD_A|PAD_B|PAD_C|PAD_X|PAD_Y|PAD_Z|PAD_START|PAD_LBUTTON|PAD_RBUTTON)) == 0)
                {
                    break;
                }

            }

            clear_screen();
        }

        if(((pad->old[0] & (PAD_X|PAD_Y)) == 0)
        &&  (pad->new[0] & (PAD_X|PAD_Y)))
        { /* Change zoom factor. */
#if 1
            int zf_list[] = {
                0,                      /* Zoom All  */
                1*ZOOMFACTOR_UNIT,      /* Zoom x1   */
                (3*ZOOMFACTOR_UNIT)/2,  /* Zoom x1.5 */
                2*ZOOMFACTOR_UNIT,      /* Zoom x2   */
                //(5*ZOOMFACTOR_UNIT)/2,  /* Zoom x2.5 */
                //3*ZOOMFACTOR_UNIT,      /* Zoom x3   */
                //(7*ZOOMFACTOR_UNIT)/2,  /* Zoom x3.5 */
                //4*ZOOMFACTOR_UNIT       /* Zoom x4   */
            };
#else // Test zoom point stuff
            int zf_list[] = {
                ( 4*ZOOMFACTOR_UNIT)/4,
                ( 5*ZOOMFACTOR_UNIT)/4,
                ( 6*ZOOMFACTOR_UNIT)/4,
                ( 7*ZOOMFACTOR_UNIT)/4,
                //( 8*ZOOMFACTOR_UNIT)/4,
                ( 9*ZOOMFACTOR_UNIT)/4,
                (10*ZOOMFACTOR_UNIT)/4,
                (11*ZOOMFACTOR_UNIT)/4,
                //(12*ZOOMFACTOR_UNIT)/4,
                //(13*ZOOMFACTOR_UNIT)/4,
                (14*ZOOMFACTOR_UNIT)/4,
                //(15*ZOOMFACTOR_UNIT)/4,
                (16*ZOOMFACTOR_UNIT)/4,
            };
#endif
#define ZFLIST_CNT (sizeof(zf_list)/sizeof(int))
            int zf_id = 0;
            for(i=0; i<ZFLIST_CNT; i++)
            {
                if(zf_list[i] == zoom_factor)
                {
                    zf_id = i;
                    break;
                }
            }

            /* Y:next zoom, X:previous zoom. */
            if(pad->new[0] & PAD_Y)
            {
                zf_id = (zf_id+1) % ZFLIST_CNT;
            }
            else
            {
                zf_id = (zf_id + ZFLIST_CNT-1) % ZFLIST_CNT;
            }
            zoom_factor = zf_list[zf_id];

            if(zoom_factor > 0)
            {
                /* Set zoom center point to center of display.
                 *
                 * Some geometry and maths in order to explain the conversion formulas below :
                 * Illustration of display areas after changing zoom factor from zf to zf' :
                 *
                 * Origin
                 * +------------------------------+==============++
                 * |                              |              ||
                 * |                  <-- dw -->  |              ||
                 * |    x0*zf         +--------+  |              ||
                 * |----------------->|        |  |              ||
                 * |                  |        |  |              ||
                 * |                  +--------+  |  <-- dw -->  ||
                 * +------------------------------+  +========+  ||
                 * |        x0'*zf'                  |        |  ||
                 * |-------------------------------->|        |  ||
                 * |                                 +========+  ||
                 * +=============================================++
                 *
                 * Where :
                 *  - zf  =      zoom_factor / ZOOMFACTOR_UNIT
                 *  - zf' = prev_zoom_factor / ZOOMFACTOR_UNIT
                 *  - dw  = SCREEN_DISP_WIDTH
                 *
                 * Since display center point is invariant, we get :
                 *  x0'*zf' + dw/2 = (x0*zf + dw/2) * zf' / zf
                 * So after simplification we get :
                 *  x0' = x0 + ((zf' - zf) * dw) / (2 * zf * zf')
                 *
                 * Similar formula applies to height computation, and
                 * since denominator doesn't changes between X and Y
                 * computations, keeping it common to both X/Y would
                 * speed up processing, but since this is done only
                 * when changing zoom factor, processing is kept in
                 * its readable shape.
                 */
                int tmp1, tmp2;

                tmp1 = (zoom_factor - prev_zoom_factor) * ZOOMFACTOR_UNIT * SCREEN_DISP_WIDTH;
                tmp2 = 2 * prev_zoom_factor * zoom_factor;
                tmp1 = tmp1 / tmp2;
                scr_disp_x0 += tmp1;

                tmp1 = (zoom_factor - prev_zoom_factor) * ZOOMFACTOR_UNIT * SCREEN_DISP_HEIGHT;
                tmp2 = 2 * prev_zoom_factor * zoom_factor;
                tmp1 = tmp1 / tmp2;
                scr_disp_y0 += tmp1;

                prev_zoom_factor = zoom_factor;
            }
        }

        /* Check keys for thumbnail selection. */
        if(pad->delta[0] & PAD_LBUTTON)
        {
            if(thb_id != 0)
            {
                thb_id--;
                if(thb_id < thb_list_offset)
                {
                    thb_list_offset = thb_id;
                }
            }
        }
        if(pad->delta[0] & PAD_RBUTTON)
        {
            if(thb_id != (scr_cnt-1))
            {
                thb_id++;
                if(thb_id >= (thb_list_offset+THB_DISP_MAXCNT))
                {
                    thb_list_offset = thb_id - (THB_DISP_MAXCNT-1);
                }
            }
        }

        /* Update release counter for L/R shoulder buttons. */
        if((pad->new[0] & (PAD_LBUTTON|PAD_RBUTTON)) == 0)
        {
            if(lr_release_cnt < 1000)
            {
                lr_release_cnt++;
            }
        }
        else
        {
            lr_release_cnt = 0;
        }

        /* If needed, read screenshot data from CDROM/SD card. */
        if((thb_prev_id < 0) /* <- Initial read. */
        || ((lr_release_cnt > 10) && (thb_id != thb_prev_id)))
        {
            thb_prev_id = thb_id;

            /* Read screenshot file. */
            sprintf(_deza_file_path, SGM_BASE_PATH "SGM%d/%03d/Z%03d.LZF", entry->sgm_id, entry->id, thb_id+1);
#define SGM_SCR_READBUFF_LEN    ((352 * 240 * sizeof(unsigned short)) + 512)
#define SGM_SCR_EXTRACTBUFF_LEN SGM_SCR_READBUFF_LEN

            unsigned char* read_buff = _extract_buff;
            unsigned long read_datalen = SC_COMP_HEADER_MAXSIZE + SGM_SCRSHOT_WIDTH * SGM_SCRSHOT_HEIGHT * sizeof(unsigned short);
            memset(read_buff, 0, read_datalen);

            /* Try to retrieve data from buffer. */
            unsigned long buffer_read_datalen;
            buffer_read_datalen = scrbuff_get(entry, thb_id, read_buff);

            if(buffer_read_datalen == 0)
            {
                /* If screenshot not available in buffer, then
                 * it seems we have to read it from CDROM or SD card.
                 */
                read_datalen = scf_read(dev_id, _deza_file_path, 0/*offset*/, read_datalen, (void*)read_buff);
            }
            else
            {
                read_datalen = buffer_read_datalen;
            }


            /* Bufferize compressed screenshot data. */
            scrbuff_put(entry, thb_id, read_buff, read_datalen);

            int valid_data = 0;
            unsigned char* decomp_buff = _extract_buff + SGM_SCR_READBUFF_LEN;
            unsigned long decomp_datalen = 0;
            if(read_datalen != 0)
            {
                /* Decompress screenshot file. */
                decomp_datalen = decomp_exec((unsigned char*)read_buff, read_datalen, decomp_buff);

                /* Verify if decompressed size is as expected or not. */
                if(decomp_datalen == (SGM_SCRSHOT_WIDTH * SGM_SCRSHOT_HEIGHT * sizeof(unsigned short)))
                {
                    valid_data = 1;
                }
            }

            if(valid_data)
            {
                /* Blit the decompressed screenshot data to VDP1 VRAM.
                 * Note : decompressing directly from LRAM to VDP1 VRAM
                 *        causes display problems because decompression
                 *        duration doesn't fits in a single V-Blank.
                 *        As a workaround, image is decompressed from LRAM
                 *        LRAM, and then copied to VDP1 VRAM.
                 */
                memcpy(SGM_SCRSHOT_DATAPTR, decomp_buff, decomp_datalen);
            }
            else
            {
                /* If screenshot file couldn't be read, 
                 * then show some dummy pattern instead.
                 */
                decomp_datalen = SGM_SCRSHOT_WIDTH * SGM_SCRSHOT_HEIGHT * sizeof(unsigned short);
                memset(SGM_SCRSHOT_DATAPTR, 0xFF, decomp_datalen);
            }
        }

        if(zoom_factor > 0)
        { /* Scroll image display (not available in "Zoom All" mode) */
            if(pad->new[0] & PAD_UP)
            {
                int d = (pad->old[0] & pad->delta[0] & PAD_UP ? 2 : 1);
                scr_disp_y0 -= d;
            }
            if(scr_disp_y0 < 0)
            {
                /* Clamp display offset. */
                scr_disp_y0 = 0;
            }

            if(pad->new[0] & PAD_DOWN)
            {
                int d = (pad->old[0] & pad->delta[0] & PAD_DOWN ? 2 : 1);
                scr_disp_y0 += d;
            }
            if(((scr_disp_y0*zoom_factor)/ZOOMFACTOR_UNIT + SCREEN_DISP_HEIGHT) >= ((SGM_SCRSHOT_HEIGHT * zoom_factor) / ZOOMFACTOR_UNIT))
            {
                /* Clamp display offset to maximum value. */
                scr_disp_y0 = (SGM_SCRSHOT_HEIGHT - 1) - ((SCREEN_DISP_HEIGHT * ZOOMFACTOR_UNIT) / zoom_factor);
            }

            if(pad->new[0] & PAD_LEFT)
            {
                int d = (pad->old[0] & pad->delta[0] & PAD_LEFT ? 2 : 1);
                scr_disp_x0 -= d;
            }
            if(scr_disp_x0 < 0)
            {
                /* Clamp display offset. */
                scr_disp_x0 = 0;
            }

            if(pad->new[0] & PAD_RIGHT)
            {
                int d = (pad->old[0] & pad->delta[0] & PAD_RIGHT ? 2 : 1);
                scr_disp_x0 += d;
            }
            if(((scr_disp_x0*zoom_factor)/ZOOMFACTOR_UNIT + SCREEN_DISP_WIDTH) >= ((SGM_SCRSHOT_WIDTH * zoom_factor)) / ZOOMFACTOR_UNIT)
            {
                /* clamp display offset to maximum value. */
                scr_disp_x0 = (SGM_SCRSHOT_WIDTH - 1) - ((SCREEN_DISP_WIDTH * ZOOMFACTOR_UNIT) / zoom_factor);
            }
        }


        /* Update cursor color animation. */
        color_anim_id = (color_anim_id + 1) % MENU_COLOR_ANIM_CNT;



        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        /* 
         * Start VDP1 related display.
         *
         * Don't use bup_anim_starfield function, which contains :
         *  - sf_animate();
         *  - sf_draw();
         *  - sf_stop();
         * Thumbnails are added to VDP1 commands after displaying
         * starfield, hence calling sf_stop function before that
         * is prohibited.
         */
        sf_animate(0/*mesh_enable*/);
        sf_draw();

        /* Display SGM screenshot stuff. */
        unsigned short* ptr;
        int disp_w, disp_h, x0, y0;

        /* Display screenshots thumbnails. */
        for(i=0; i<THB_DISP_MAXCNT; i++)
        {
            int id = thb_list_offset+i;
#ifdef THB_DISP_VERTICAL
            /* Display thumbnails vertically. */
            disp_w = SGM_THUMB_WIDTH;
            disp_h = SGM_THUMB_HEIGHT;
            x0 = THB_DISP_X0;
            y0 = THB_DISP_Y0 + i*(disp_h + THB_DISP_SPACE);
#else
            /* Display thumbnails horizontally. */
            disp_w = SGM_THUMB_WIDTH;
            disp_h = SGM_THUMB_HEIGHT;
            x0 = THB_DISP_X0 + i*(disp_w + THB_DISP_SPACE);
            y0 = THB_DISP_Y0;
#endif

            /* Verify if the thumbnail can be displayed or not. */
            if(id >= scr_cnt)
            {
                break;
            }

            /* Set pointer to current thumbnail and display it. */
            ptr = (unsigned short*)SGM_ENTRY_THBPTR;
            ptr += id * SGM_THUMB_WIDTH * SGM_THUMB_HEIGHT;

            vdp1_add_distorted_sprite(
                x0       , y0       , 
                x0+disp_w, y0       , 
                x0+disp_w, y0+disp_h, 
                x0       , y0+disp_h, 
                VDP_SPRITE_SIZE(SGM_THUMB_WIDTH, SGM_THUMB_HEIGHT), 
                (((unsigned long)ptr) & 0xFFFFF)>>3, 
                0/*grad*/, 
                0/*mesh_use*/);

            if(id == thb_id)
            {
                /* Add some selection square around the currently highlighted entry. */
                vdp1_add_polyline(
                    x0        - 1, y0        - 1, 
                    x0+disp_w + 1, y0        - 1, 
                    x0+disp_w + 1, y0+disp_h + 1, 
                    x0        - 1, y0+disp_h + 1, 
                    _menu_color_animation[color_anim_id]);
            }
        }



        if(zoom_factor > 0)
        { /* Display zoomed screenshot. */
            disp_w = SCREEN_DISP_WIDTH;
            disp_h = SCREEN_DISP_HEIGHT;
            x0 = SCREEN_DISP_X0;
            y0 = SCREEN_DISP_Y0;

            /* Display square around screenshot display area. */
            vdp1_add_polyline(
                x0        - 1, y0        - 1, 
                x0+disp_w + 1, y0        - 1, 
                x0+disp_w + 1, y0+disp_h + 1, 
                x0        - 1, y0+disp_h + 1, 
                COLOR(31, 31, 31));

            ptr = (unsigned short*)SGM_SCRSHOT_DATAPTR;
            ptr += scr_disp_y0 * SGM_SCRSHOT_WIDTH;
            ptr += scr_disp_x0;

            /* Display screenshot within specific range. */
            //vdp1_add_user_clip(x0, y0, x0+disp_w, y0+disp_h);
            vdp1_add_system_clip(x0+disp_w, y0+disp_h);

            disp_w = (SGM_SCRSHOT_WIDTH  * zoom_factor) / ZOOMFACTOR_UNIT;
            disp_h = (SGM_SCRSHOT_HEIGHT * zoom_factor) / ZOOMFACTOR_UNIT;
            vdp1_add_distorted_sprite(
                x0       , y0       , 
                x0+disp_w, y0       , 
                x0+disp_w, y0+disp_h, 
                x0       , y0+disp_h, 
                VDP_SPRITE_SIZE(SGM_SCRSHOT_WIDTH, SGM_SCRSHOT_HEIGHT), 
                (((unsigned long)ptr) & 0xFFFFF)>>3, 
                0/*grad*/, 
                0/*mesh_use*/);

            // /* Show rectangle around zoom center point (debug purpose). */
            // disp_w = 8;
            // disp_h = 8;
            // x0 = SCREEN_DISP_X0 + (SCREEN_DISP_WIDTH/2) - (disp_w/2);
            // y0 = SCREEN_DISP_Y0 + (SCREEN_DISP_HEIGHT/2) - (disp_h/2);
            // vdp1_add_polyline(
            //     x0        - 1, y0        - 1, 
            //     x0+disp_w + 1, y0        - 1, 
            //     x0+disp_w + 1, y0+disp_h + 1, 
            //     x0        - 1, y0+disp_h + 1, 
            //     COLOR(15, 15, 31));
        }
        else
        { /* Display screenshot in "Zoom All" mode. */

            if(((SCREEN_DISP_WIDTH * SGM_SCRSHOT_HEIGHT) / SGM_SCRSHOT_WIDTH) <= SCREEN_DISP_HEIGHT)
            {
                disp_w = SCREEN_DISP_WIDTH;
                disp_h = (SCREEN_DISP_WIDTH * SGM_SCRSHOT_HEIGHT) / SGM_SCRSHOT_WIDTH;
                x0 = SCREEN_DISP_X0;
                y0 = SCREEN_DISP_Y0 + ((SCREEN_DISP_HEIGHT - disp_h) / 2);
            }
            else
            {
                disp_w = (SCREEN_DISP_HEIGHT * SGM_SCRSHOT_WIDTH) / SGM_SCRSHOT_HEIGHT;
                disp_h = SCREEN_DISP_HEIGHT;
                x0 = SCREEN_DISP_X0 + ((SCREEN_DISP_WIDTH - disp_w) / 2);
                y0 = SCREEN_DISP_Y0;
            }

            /* Display square around screenshot display area. */
            vdp1_add_polyline(
                x0        - 1, y0        - 1, 
                x0+disp_w + 1, y0        - 1, 
                x0+disp_w + 1, y0+disp_h + 1, 
                x0        - 1, y0+disp_h + 1, 
                COLOR(31, 31, 31));

            ptr = (unsigned short*)SGM_SCRSHOT_DATAPTR;
            vdp1_add_distorted_sprite(
                x0       , y0       , 
                x0+disp_w, y0       , 
                x0+disp_w, y0+disp_h, 
                x0       , y0+disp_h, 
                VDP_SPRITE_SIZE(SGM_SCRSHOT_WIDTH, SGM_SCRSHOT_HEIGHT), 
                (((unsigned long)ptr) & 0xFFFFF)>>3, 
                0/*grad*/, 
                0/*mesh_use*/);
        }

        /* Terminate VDP1 display. */
        sf_stop();

        /* Display some infos on last line:
         * "123456789012345678901234567890123456"
         * "Name:Areichalke Zoom:x1.5 Start:Help"
         */
        //conio_printf(2, 25, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "TEST123456789");
        conio_printf( 2, 25, COLOR_WHITE, "Name:");
        conio_printf( 7, 25, COLOR_YELLOW, entry->comment);

        conio_printf(18, 25, COLOR_WHITE, "Zoom:");
        char zoom_str[4+1];
        memset(zoom_str, '\0', sizeof(zoom_str));
        if(zoom_factor <= 0)
        {
            sprintf(zoom_str, "ALL ");
        }
        else
        {
            sprintf(zoom_str, "x%d.%d", zoom_factor/ZOOMFACTOR_UNIT, ((10*zoom_factor)/ZOOMFACTOR_UNIT) % 10);
        }
        for(i=0; i<(sizeof(zoom_str)-1); i++)
        {
            if(zoom_str[i] == '\0')
            {
                zoom_str[i] = ' ';
            }
        }
        conio_printf(23, 25, COLOR_YELLOW, zoom_str);

        conio_printf(28, 25, COLOR_WHITE , "Start:");
        conio_printf(34, 25, COLOR_YELLOW, "Help");

        /* Do I/O check. */
        sc_check();
    }


    return 1;
}

