/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef DEZA_SGM_H
#define DEZA_SGM_H

#include <stdio.h>


/* Inclusion of definitions shared with PC. */
#include "deza_list.h"


/* VDP1 related definitions :
 *  - Start address for current screenshot (shared with logo)
 *  - Maximum data length for all thumbnails (150KB for 16 x 16 thumbnails x max 300 entries)
 *  - Start address for all thumbnails
 *  - Maximum data length for current entry's thumbnails (50KB for 16 x 16 thumbnails x max 100 screenshots per entry)
 *  - Start address for current entry's thumbnails
 */
#include "../starfield.h"
#include "../logo.h"
#define SGM_SCRSHOT_DATAPTR     LOGO_DATAPTR

#define SGM_LIST_THBMAXLEN      (SGM_THUMB_WIDTH * SGM_THUMB_HEIGHT * sizeof(unsigned short) * SGM_MAXENTRIES)
#define SGM_LIST_THBPTR         (unsigned short*)(((unsigned char*)LOGO_GRADPTR) - SGM_LIST_THBMAXLEN)

#define SGM_ENTRY_THBMAXLEN     (SGM_THUMB_WIDTH * SGM_THUMB_HEIGHT * sizeof(unsigned short) * SGM_SCRSHOT_MAXCNT)
#define SGM_ENTRY_THBPTR        (unsigned short*)(((unsigned char*)SGM_LIST_THBPTR) - SGM_ENTRY_THBMAXLEN)

/*
 * Constants for SGM entry selection menu.
 *  - Large size, for currently selected item.
 *  - Small size, for items in far neighboring of currently selected item.
 *  - Display half count for thumbnails.
 *    Example : if displaying 9 thumbnails, then half count is 4.
 *  - Space between two thmbnails, in pixels.
 */
#define SGM_LIST_X0     16
#define SGM_LIST_Y0     (4*CONIO_CHAR_H)
#define SGM_LIST_COLCNT 28
#define SGM_LIST_ROWCNT 13


/*
 * Thumbnails display constants in SGM entries list :
 *  - Large size, for currently selected item.
 *  - Small size, for items in far neighboring of currently selected item.
 *    (currently commented-out, because this isn't really fancy)
 *  - Display half count for thumbnails.
 *    Example : if displaying 9 thumbnails, then half count is 4.
 *  - Space between two thmbnails, in pixels.
 *  - Display start coordinate, in pixels.
 */
#define SGM_THUMBL_WIDTH   24
#define SGM_THUMBL_HEIGHT  24
//#define SGM_THUMBF_WIDTH   SGM_THUMB_WIDTH
//#define SGM_THUMBF_HEIGHT  SGM_THUMB_HEIGHT

#define SGM_THUMB_DISP_HALFCNT ((SGM_LIST_ROWCNT * MENU_CHAR_H + CONIO_CHAR_H) / SGM_THUMB_WIDTH)
#define SGM_THUMB_DISP_YSPACE  4

#define SGM_LIST_THUMB_X0 (320 - SGM_LIST_X0 - SGM_THUMBL_WIDTH)


/**
 * deza_is_sgm_available
 *
 * Check if data for Dezaemon 2 Save Game Manager is available or not.
 *
 * It checks for availability in several places in the following order :
 *  1. CD-ROM, physical or emulated by ODE.
 *  2. Files copied in SD card's DEZA2 folder.
 *  3. Iso file in SD card root folder.
 *       - dz2sgm.iso : name for future version of DZ2SGM ?
 *         ^^^ if not found, then vvv
 *       - pskai_sdm_flasher.iso : DZ2SGM3/4 iso file name.
 *
 * Choice of the best place to install Dezaemon 2 SGM depends of hardware setup.
 * For example :
 *  A. Gamer's Cartridge = best choice (2) because it avoids to change CD-ROM
 *     between Dezaemon 2 and SGM, and also because (2) is faster than (3).
 *  B. Saturn with ODE installed = best choice (1), assuming that ODE makes
 *     things simple disc image selection between Dezaemon 2 and SGM.
 *  C. Cart with SD card access faster than Gamer's Cartridge = best choice (3)
 *     because individual copy of all the files in DEZA2 folder takes a super
 *     long time.
 *
 * Returns : 
 *  - 1 if SGM data available.
 *  - Zero if SGM data not available.
 *
 * Note : consistency of SGM data isn't checked, so it may return 1
 *        in the case SGM data is partially missing or corrupted.
**/
int deza_is_sgm_available(void);

/**
 * deza_sgm_dev_id
 *
 * Returns device ID where to access Dezaemon 2 SGM files.
 * Possible values :
 *  - SC_FILE_CDROM  : CD-ROM
 *  - SC_FILE_SDCARD : SD card (Gamer's Cartridge)
 *
 * Note : this function can't be called before `deza_is_sgm_available' function.
**/
unsigned char deza_sgm_dev_id(void);

/**
 * deza_sgm_is_override
 *
 * Indicates if CD-ROM access is overridden with access on SD card.
 *
 * Returns :
 *  - 0    : access via CD-ROM or ODE
 *  - else : access via iso file on SD card
**/
int deza_sgm_is_override(void);

/**
 * deza_sgm
 *
 * Dezaemon 2 Save Game Manager main function.
**/
void deza_sgm(void);


#endif // DEZA_SGM_H
