/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
/*
 * Sega Saturn cartridge flash tool
 * by Anders Montonen, 2012
 *
 * Original software by ExCyber
 * Graphics routines by Charles MacDonald
 *
 * Creative Commons Attribution-ShareAlike 3.0 Unported (CC BY-SA 3.0)
 */
#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>

#include "main.h"
#include "app_title.h"

#include "../vdp.h"
#include "../conio.h"

#include "bup_menus.h"
#include "bup_functions.h"
#include "../credits.h"
#include "../debug_console.h"
#include "extra_menu.h"
#include "../ip_header.h"
#include "../soft_reset.h"
#include "../satiator_debug.h"
#include "../starfield.h"
#include "../logo.h"


/* Commslink related definitions. */
commslink_status_t _commslink_stat = {0};

/* Fenrir I/O related. */
#include "fenrir_fileio_wrapper.h"
#include "../pskai_settings.h"
extern sc_settings_t _satcom_settings;

#include "../romfs_location.h"

#include "../release.h"

int cdrom_available = 1;
int _deza_sgm_available = 0;
int _fenrir_sdm = 0;
int _limited_sdm = 0;

/* Global variable for satcom library. */
unsigned long _sc_workram[SC_WORKRAM_U32SIZE];
int _satcom_initialized = 0;

/* If set, disable SatCom initialization on startup.
 * Purpose : keep focus on logs from vmem.
 */
int _vmem_dbg_mode = 0;


/* Display application title at the top of the screen. */
void disp_app_title(void)
{
    /* Change title according to vmem debug/release mode, 
     * in order not to forget which build mode is on ...
     */
    if(_fenrir_sdm)
    {
        conio_title_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), " Fenrir Save Data Manager -- V%d.%03d ", RELEASE_ID/1000, RELEASE_ID%1000);
    }
    else
    {
        if(_vmem_dbg_mode)
        {
            conio_title_printf(2, 2,                           COLOR_ERROR, "    Vmem Debug Utility  -- V%d.%03d   ", RELEASE_ID/1000, RELEASE_ID%1000);
        }
        else
        {
            conio_title_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), " Saturn Save Data Manager -- V%d.%03d ", RELEASE_ID/1000, RELEASE_ID%1000);
        }
    }
}


/* Init SatCom stuff. */
void bup_satcom_init(void)
{
    sc_initcfg_t init_cfg;

    /* Small tweak in order to continue receiving logs from vmem.
     * Comment-out the line below if you prefer to get logs from Save Data Manager.
     * Note : this tweak disables PC file access from Save Data Manager.
     */
//#define RCV_VMEM_LOGS
#ifdef RCV_VMEM_LOGS
    sc_global_t global_tmp;
    sc_global_t* gl = (sc_global_t*)SC_GLOBAL_DATA_ADR;
    memcpy(&global_tmp, gl, sizeof(sc_global_t));
#endif

    if(_satcom_initialized == 0)
    {
        /* Init SatCom. */
        sc_initcfg_reset(&init_cfg, _sc_workram);
        /* Note : this may be executed from cartridge, 
         * so using CDROM may cause problems.
         * But, it shouldn't do anything bad as long as
         * CDROM isn't accessed.
         */
        init_cfg.flags = (cdrom_available ? 0/*default*/ : SC_INITFLAG_NOCDROM);

        /* Small hack to prevent from initializing Satiator
         * when Save Data Manager is loaded from Gamer's Cartridge ROM.
         *
         * This is required because it looks like trying to initialize
         * Satiator on an unmodded Saturn is messing CD-ROM authentication.
         */
        if(memcmp(((ip_bin_t*)0x02000000)->prd_nb, (void*)"PSKAI_F   ", 10) == 0)
        {
            init_cfg.flags |= SC_INITFLAG_NOSATIATOR;
        }


        // /* Similarly to the hack above, let's disable Satiator access
        //  * when running as SDM for Fenrir. (Maybe both together work, 
        //  * maybe not ... don't know what may happen in that kind of
        //  * exotic configuration)
        //  */
        // if(_fenrir_sdm)
        // {
        //     init_cfg.flags |= SC_INITFLAG_NOSATIATOR;
        // }


//#define SHOW_CONSOLE

#ifdef SHOW_CONSOLE
        #define TEST_CONS_Y0 6
        #define TEST_CONS_HEIGHT (25-TEST_CONS_Y0)
        cons_enable(TEST_CONS_Y0, TEST_CONS_HEIGHT, 1/*pause_enable*/);
        init_cfg.logscreen_callback = debug_console;
#endif // SHOW_CONSOLE

#if SATIATOR_DIAGNOSTIC_ENABLE == 1
        /* SatCom initialization + Satiator disagnostic menu. */
        satiator_diagnostic_menu(&init_cfg);
#else
        sc_init_cfg(&init_cfg);
#endif

#ifdef SHOW_CONSOLE
    wait_for_start(NULL);
    cons_disable();
#endif // SHOW_CONSOLE

        /* Specify where RomFs data is located. */
        sc_romfs_setdata((unsigned char*)PSKAI_ROMFS_ADDRESS);

        /* Prevent from initializing SatCom library twice, 
         * in order to speedup program execution.
         */
        _satcom_initialized = 1;
    }

#ifdef RCV_VMEM_LOGS
    memcpy(gl, &global_tmp, sizeof(sc_global_t));
#endif
}


/* Used in menus when satcom library is enabled. */
void background_vblank_clbk(void)
{
    bup_anim_starfield();
}


/* Animate starfield on each vblank. */
void bup_anim_starfield(void)
{
    int transparent_logo = logo_is_transparent();
    int logo_mesh_use = logo_is_mesh_enabled();

    /* Update logo and starfield animations. */
    logo_animate();
    sf_animate(logo_mesh_use/*mesh_enable*/);

    /* When opaque logo is used, show it on
     * background, and starfield over it.
     */
    if(!transparent_logo)
    {
        logo_draw();
    }

    sf_draw();

    /* When transparent logo is used, show it
     * on foreground, so that starfield is displayed
     * only on transparent regions.
     */
    if(transparent_logo)
    {
        logo_draw();
    }

    /* Terminate VDP1 drawing. */
    sf_stop();

    if(_satcom_initialized)
    {
        sc_check();
    }

    /* Do Commslink related IO. */
    comms_link_check(&_commslink_stat);
}

void clear_screen(void)
{
    int i;

    /* Clear lines 3~25. */
    for(i=3; i<26; i++)
    {
        conio_printf(2, i, COLOR_TRANSPARENT, empty_line);
    }
}


/* Used after executing features requiring access to CD-ROM or SD card. */
void buplib_init(void)
{
    int i;

    /* Initialize pointer to LRAM buffer. */
    _temp_buff = (unsigned char*)TEMP_BUFF_ADR;

    /* Clear each buffers used by BUP_Init function. */
    memset(_bupConf, 0, sizeof(BupConfig)*BUP_DEVICE_CNT);
    memset(_bupLibrary, 0, sizeof(unsigned long)*4096);
    memset(_bupRamWork, 0, sizeof(unsigned long)*2048);

    /* Write dummy value just after each RAM areas, 
     * in order to verify they are not overwritten by vmem.
     */
    _bupLibrary[4096] = 0xCAFE1616;
    _bupRamWork[2048] = 0xCAFE0808;

    dbg_print("#1 BUP_Init=0x%08X", BUP_Init);
    BUP_Init((volatile unsigned long *)_bupLibrary, _bupRamWork, _bupConf);
    dbg_print("#2 BUP_Init Finish.");

    /* Convert informations for all 3 backup devices, in device order
     * to partition information for available partitions (up to 4 partitions, less if partition not detected).
     *
     * Input  : BupConfig   _bupConf[BUP_DEVICE_CNT];
     * Output : BupPartInfo _bupPartInfo[BUP_PARTITION_CNT];
     *
     * Also, add some tweaks regarding FDD detection. (see below for details)
     */
    _whole_part_list = PART_NONE;
    _whole_part_cnt = 0;
    memset(_bupPartInfo, 0, sizeof(BupPartInfo)*BUP_PARTITION_CNT);

    /* First, add internal memory and cartridge memory informations.
     *
     * Theses are usually detected as follow :
     * Case 1: BRAM, no memory cart :
     *  [0] partition=1, unit_id=1
     *  [1] partition=1, unit_id=0
     * Case 1: BRAM, memory cart (or Gamer's Cartridge) :
     *  [0] partition=1, unit_id=1
     *  [1] partition=1, unit_id=2
     */
    for(i=0; i<2; i++)
    {
        if(_bupConf[i].unit_id != 0)
        {
            /* Set device ID (0:BRAM, 1:CART). */
            _bupPartInfo[_whole_part_cnt].device_id = i;

            /* Set partition ID to same value as device ID for BRAM and CART. */
            _bupPartInfo[_whole_part_cnt].part_id = i;

            /* Set partition count to "1", because (AFAIK), 
             * BRAM and CART doesn't support multiple partitions.
             * Consequently, partition ID within device is set to zero.
             */
            _bupPartInfo[_whole_part_cnt].part_cnt = 1;
            _bupPartInfo[_whole_part_cnt].device_part_id = 0;

            /*
             * Set abbreviated name (4 characters) for current partition.
             */
            _bupPartInfo[_whole_part_cnt].part_abbr_name = _bup_part_abbr_names[_bupPartInfo[_whole_part_cnt].part_id];

            _whole_part_list |= (1 << _bupPartInfo[_whole_part_cnt].part_id);
            _whole_part_cnt++;
        }
    }

    /* Add FDD detection, which seems to be detected as below :
     *  partition=2, unit_id=0
     * While it is expected to be detected as partition=2, and unit_id=2 ...
     *
     * Also, discard partitions with ID > 2.
     * (AFAIK, theses aren't supported by FDD operation disc.)
     */
    if(_bupConf[DEVICE_ID_DISC].partition != 0)
    {
        for(i=0; i<2; i++)
        {
            /* Set device ID to DEVICE_ID_DISC. */
            _bupPartInfo[_whole_part_cnt].device_id = DEVICE_ID_DISC;

            /* Set partition ID (2:FDD_part1, 3:FDD_part2). */
            _bupPartInfo[_whole_part_cnt].part_id = PART_ID_FDD1+i;

            /* Set partition count to "1" or "2" according to
             * detected partition count.
             * Also, set partition ID within device to "0" or "1".
             */
            _bupPartInfo[_whole_part_cnt].part_cnt = (_bupConf[DEVICE_ID_DISC].partition == 1 ? 1 : 2);
            _bupPartInfo[_whole_part_cnt].device_part_id = i;

            /*
             * Set abbreviated name (4 characters) for current partition.
             */
            _bupPartInfo[_whole_part_cnt].part_abbr_name = _bup_part_abbr_names[_bupPartInfo[_whole_part_cnt].part_id];

            _whole_part_list |= (1 << _bupPartInfo[_whole_part_cnt].part_id);
            _whole_part_cnt++;
        }
    }

    dbg_print("#2 BUP_Init END [%d][%02X]", _whole_part_cnt, _whole_part_list);
    /* This should never happen, but if no save device found, emergency stop. */
    if(_whole_part_cnt == 0)
    {
        conio_printf(2, 17, COLOR_ERROR, "No save device detected !!!");
        wait_for_start(NULL);
    }
}


void _main(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int tag;

    /* Retrieve vmem debug mode. */
    _vmem_dbg_mode = (BUP_PARAM_STRUCT->vmem_dbg_mode == 1 ? 1 : 0);

    /* Indicate SatCom library as "not initialized yet". */
    _satcom_initialized = 0;

    /* Check if was loaded from CD-ROM or not. */
    cdrom_available = (memcmp(((ip_bin_t*)0x06002000)->title, (void*)BUP_APP_TITLE_STR, BUP_APP_TITLE_LEN) == 0 ? 1 : 0);

    /* Initialize screenshot gallery related internals. */
    deza_screens_init();

    /* Init logo and starfield internals. */
    sf_init((cdrom_available ? 0 : 1)/*multicolor*/);
    logo_init(LOGO_APPTYPE_BUPMNGR);

    vdp_init();
    conio_init();
    pad_init();

    /* Reset Commslink I/O status. */
    memset(&_commslink_stat, 0, sizeof(commslink_status_t));

    /* Init internals for soft reset routine. */
    soft_reset_init();

    /* Because of interrupt based pad status acquisition, display of
     * debug messages on screen can no longer be enabled with shortcut
     * button push on startup.
     * As a consequence, application rebuild with the line below uncommented
     * is required to enable this feature.
     */
    //dbg_print_enable();

    /* Init save data library. */
    buplib_init();

    /* Load settings from internal backup memory. */
    pskai_settings_read();

    /* Verify if Fenrir access should be enabled or not. */
    if(memcmp(((ip_bin_t*)0x06002000)->title, "Fenrir Loader   Kai Edition   ON", 32) == 0)
    {
        /* In the case of running Fenrir Save Data Manager, locally
         * enable Fenrir ODE in settings.
         * This is needed because some processing common with Pseudo
         * Saturn Kai may check for correctness of ODE type.
         */
        pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
        set->bram.od_type = SET_ODTYPE_FENRIR;

        /* Add Fenrir I/O to exiting ones in satcom_lib. */
        fenrir_media_register();

        /* Indicate we are running Save Data Manager for Fenrir. */
        _fenrir_sdm = 1;
    }

    /* Enable limited mode when version in IP header is set to "9.999".
     * This is used to repack Save Data Manager with extra save data
     * and other stuff removed so that it can be redistributed easily.
     */
    if(memcmp(((ip_bin_t*)0x06002000)->version+1, "9.999", 5) == 0)
    {
        _limited_sdm = 1;
    }

    /* Display application title at the top of the screen. */
    disp_app_title();

    /* Set initial action to "copy".
     *
     * In limited mode, save data import is initially selected
     * because this mode is used for repacking and distributing saves.
     */
    tag = (_limited_sdm ? TAG_IMPORT : TAG_COPY);

    if(_vmem_dbg_mode)
    {
        /* When debugging vmem, SatCom library is not initialized on startup, so that
         * debug messages from vmem aren't hidden by logs from Save Data Manager.
         *
         * Consequently, don't (can't) check for availability of Dezaemon 2 SGM.
         */
        _deza_sgm_available = 0;
    }
    else
    {
        /* Check for availablility of Dezaemon 2 SGM.
         * This will be used in order to tune main menu items.
         *
         * Note : deza_is_sgm_available initializes SatCom library, 
         *        so that it is not needed to initialize it here.
         */
        _deza_sgm_available = deza_is_sgm_available();
    }

    /* Specifiy to credits module if
     * Dezaemon 2 mode can be enabled or not.
     */
    credits_set_dz2sgm_mode((_deza_sgm_available ? 1 : 0));


    /* Load logo stuff in similar way as done in Pseudo Saturn Kai.
     * If custom logo couldn't be loaded (SD card not available, 
     * custom logo file not set, etc), then default logo is loaded instead.
     *
     * Loading logo requires SatCom library to be initialized, and even
     * for default logo (because it is compressed), so don't load any logo
     * when SatCom isn't initialized. This happens in vmem debug mode only.
     */
    if(_satcom_initialized)
    {
        logo_load(_logo_load_params+LOGO_CUSTOM_LOAD_PARAMS);
    }


    /* Show Dezaemon 2 SGM on startup when it is loaded from CD-ROM.
     *
     * Save Data Manager is released in both "Pseudo Saturn Kai installer"
     * and "Dezaemon 2 SGM" flavors, so it makes sense to automatically show
     * DZ2SGM since user can select between which CD-ROM to boot.
     *
     * On the other hand, swapping SD cards (for example between a blank one
     * and another one with DZ2SGM files installed) isn't convenient, and
     * also causes problem when Save Data Manager is used for cartridge internal
     * purpose, so that this automatic DZ2SGM show on startup is restricted
     * to CD-ROM version only.
     */
    if(_deza_sgm_available)
    {
        /* CD-ROM access can be either from :
         *  - Physical CD-ROM drive, or ODE
         *  - Via file on SD card = "override access"
         * In the latter case, Save Data Manager may be used to copy/delete/etc
         * saves on Gamer's Cartridge so that automatic launch of Dezaemon 2 SGM
         * on startup is restricted to CD-ROM drive (or ODE) only.
         */
        if((deza_sgm_dev_id() == SC_FILE_CDROM)
        && (!deza_sgm_is_override()))
        {
            deza_sgm();
            buplib_init();
        }
    }

    while(1)
    {
        /* Display main menu. */
        tag = main_menu_exec(mset, mitems, tag);

        if((tag == TAG_COPY)
        || (tag == TAG_MOVE))
        {
            bup_copymove(tag == TAG_MOVE ? 1 : 0);
        }
        else if(tag == TAG_DELETE)
        {
            bup_delete();
        }
        else if(tag == TAG_HEXEDIT)
        {
            bup_hexedit();
        }
        else if(tag == TAG_FORMAT)
        {
            bup_format();
        }
        else if(tag == TAG_INFO)
        {
            bup_devinfo();
        }
        else if(tag == TAG_DUMP)
        {
            bup_dump();

            /* Re-init BUP library.
             * It is needed in order to initialize
             * SatCom debugger on vmem side.
             */
            buplib_init();
        }
        else if(tag == TAG_BIOSODE)
        {
            bios_to_ode();

            /* Re-init BUP library.
             * It is needed in order to initialize
             * SatCom debugger on vmem side.
             */
            buplib_init();
        }
        else if(tag == TAG_DEZASGM)
        {
            deza_sgm();

            /* Re-init BUP library.
             * It is needed in order to initialize
             * SatCom debugger on vmem side.
             */
            buplib_init();
        }
        else if(tag == TAG_IMPORT)
        {
            bup_import();

            /* Re-init BUP library.
             * It is needed in order to initialize
             * SatCom debugger on vmem side.
             */
            buplib_init();
        }
        else if(tag == TAG_CREDITS)
        {
            /* Show credits screen.
             * Initializing SatCom (and then
             * re-initializing BUP library)
             * is done for backup purpose
             * as well as for reading extra
             * image files from CD-ROM.
             */
            bup_satcom_init();

            credits(bup_anim_starfield);

            buplib_init();
        }
    }
}
