/*  Copyright 2007,2013 Theo Berkau

    This file is part of Iapetus.

    Iapetus is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Iapetus is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Iapetus; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

// The endian conversion macros
#define SWAP_U16(_VAL_) ((((_VAL_) >> 8) & 0x00FF) | (((_VAL_) << 8) & 0xFF00))
#define SWAP_U32(_VAL_) ((((_VAL_) >> 24) & 0x000000FF) | \
						 (((_VAL_) <<  8) & 0x00FF0000) | \
						 (((_VAL_) >>  8) & 0x0000FF00) | \
						 (((_VAL_) << 24) & 0xFF000000))


#if 1 // Define that in the case strcasecmp is not available

int tolower(int c)
{
  return ((c >= 'A') && (c <= 'Z')) ? (c) - 'A' + 'a' : c;
}

int strcasecmp(const char *a, const char *b)
{
    int ca, cb;

    do
    {
        ca = * (unsigned char *)a;
        cb = * (unsigned char *)b;

        ca = tolower(ca);
        cb = tolower(cb);

        a++;
        b++;
    } while (ca == cb && ca != '\0');

    return ca - cb;
}

int strncasecmp(const char *s1, const char *s2, size_t n)
{
    unsigned char u1, u2;

    for(; n != 0; --n)
    {
        u1 = (unsigned char) *s1++;
        u2 = (unsigned char) *s2++;

        if(tolower(u1) != tolower(u2))
        {
            return tolower(u1) - tolower(u2);
        }

        if(u1 == '\0')
        {
            return 0;
        }
    }

    return 0;
}

#endif // strcasecmp implementation


#include "cdfs.h"

//////////////////////////////////////////////////////////////////////////////

FILE* iso_fp = NULL;

int cd_iso_open(char* iso_path)
{
    if(iso_fp)
    {
        fclose(iso_fp);
        iso_fp = NULL;
    }

    iso_fp = fopen(iso_path, "rb");

    return (iso_fp ? IAPETUS_ERR_OK : IAPETUS_ERR_FILENOTFOUND);
}

int cd_iso_close(void)
{
    if(iso_fp)
    {
        fclose(iso_fp);
        iso_fp = NULL;
    }

    return IAPETUS_ERR_OK;
}

int cd_read_sector(void *buffer, unsigned long FAD, int sector_size, unsigned long num_bytes)
{
    printf("cd_read_sector FAD:%u\n", FAD);

    if(!iso_fp)
    {
        printf("+ Error cd_read_sector iso not opened\n");
        return IAPETUS_ERR_FILENOTFOUND;
    }

    /* Support only iso files, which have their sectors being contiguous. */
    if(sector_size == SECT_2048)
    {
        unsigned long sector = FAD - 150;
        fseek(iso_fp, sector * 2048, SEEK_SET);
        fread((void*)buffer, 1, num_bytes, iso_fp);
    }
    else
    {
        printf("+ Error cd_read_sector sector size invalid\n");
        return IAPETUS_ERR_FILENOTFOUND;
    }

    return IAPETUS_ERR_OK;
}


//////////////////////////////////////////////////////////////////////////////

//#define USE_XA_REC

#ifdef USE_XA_REC
typedef struct
{
    unsigned short group_id;
    unsigned short user_id;
    unsigned short attributes;
    unsigned short signature;
    unsigned char file_number;
    unsigned char reserved[5];
} xa_rec_struct;
#endif // USE_XA_REC

typedef struct
{
    unsigned char record_size;
    unsigned char xa_record_size;
    unsigned long lba;
    unsigned long size;
    unsigned char date_year;
    unsigned char date_month;
    unsigned char date_day;
    unsigned char date_hour;
    unsigned char date_minute;
    unsigned char date_second;
    unsigned char gmt_offset;
    unsigned char flags;
    unsigned char file_unit_size;
    unsigned char interleave_gap_size;
    unsigned short volume_sequence_number;
    unsigned char name_length;
    char name[32];
#ifdef USE_XA_REC
    xa_rec_struct xa_record;
#endif // USE_XA_REC
} dir_rec_struct;

unsigned char *dir_tbl;
unsigned char *sect_buf;
int sect_buffered;
int dir_tbl_size;
int root_lba;
int root_size;

//////////////////////////////////////////////////////////////////////////////

void copy_dir_record(unsigned char *buffer, dir_rec_struct *dir_rec)
{
    unsigned char *temp_pointer;

    temp_pointer = buffer;

    memcpy(&dir_rec->record_size, buffer, sizeof(dir_rec->record_size));
    buffer += sizeof(dir_rec->record_size);

    memcpy(&dir_rec->xa_record_size, buffer, sizeof(dir_rec->xa_record_size));
    buffer += sizeof(dir_rec->xa_record_size);

    buffer += sizeof(dir_rec->lba);
    memcpy(&dir_rec->lba, buffer, sizeof(dir_rec->lba));
    dir_rec->lba = SWAP_U32(dir_rec->lba);
    buffer += sizeof(dir_rec->lba);

    buffer += sizeof(dir_rec->size);
    memcpy(&dir_rec->size, buffer, sizeof(dir_rec->size));
    dir_rec->size = SWAP_U32(dir_rec->size);
    buffer += sizeof(dir_rec->size);

    dir_rec->date_year = buffer[0];
    dir_rec->date_month = buffer[1];
    dir_rec->date_day = buffer[2];
    dir_rec->date_hour = buffer[3];
    dir_rec->date_minute = buffer[4];
    dir_rec->date_second = buffer[5];
    dir_rec->gmt_offset = buffer[6];
    buffer += 7;

    dir_rec->flags = buffer[0];
    buffer += sizeof(dir_rec->flags);

    dir_rec->file_unit_size = buffer[0];
    buffer += sizeof(dir_rec->file_unit_size);

    dir_rec->interleave_gap_size = buffer[0];
    buffer += sizeof(dir_rec->interleave_gap_size);

    buffer += sizeof(dir_rec->volume_sequence_number);
    memcpy(&dir_rec->volume_sequence_number, buffer, sizeof(dir_rec->volume_sequence_number));
    dir_rec->volume_sequence_number = SWAP_U16(dir_rec->volume_sequence_number);
    buffer += sizeof(dir_rec->volume_sequence_number);

    dir_rec->name_length = buffer[0];
    buffer += sizeof(dir_rec->name_length);

    memset(dir_rec->name, 0, sizeof(dir_rec->name));
    memcpy(dir_rec->name, buffer, dir_rec->name_length);
    buffer += dir_rec->name_length;

    // handle padding
    buffer += (1 - dir_rec->name_length % 2);

#ifdef USE_XA_REC
    memset(&dir_rec->xa_record, 0, sizeof(dir_rec->xa_record));

    // Sadily, this is the best way I can think of for detecting XA records
    if ((dir_rec->record_size - (buffer - temp_pointer)) == 14)
    {
        memcpy(&dir_rec->xa_record.group_id, buffer, sizeof(dir_rec->xa_record.group_id));
        buffer += sizeof(dir_rec->xa_record.group_id);

        memcpy(&dir_rec->xa_record.user_id, buffer, sizeof(dir_rec->xa_record.user_id));
        buffer += sizeof(dir_rec->xa_record.user_id);

        memcpy(&dir_rec->xa_record.attributes, buffer, sizeof(dir_rec->xa_record.attributes));
        buffer += sizeof(dir_rec->xa_record.attributes);

        memcpy(&dir_rec->xa_record.signature, buffer, sizeof(dir_rec->xa_record.signature));
        buffer += sizeof(dir_rec->xa_record.signature);

        memcpy(&dir_rec->xa_record.file_number, buffer, sizeof(dir_rec->xa_record.file_number));
        buffer += sizeof(dir_rec->xa_record.file_number);

        memcpy(dir_rec->xa_record.reserved, buffer, sizeof(dir_rec->xa_record.reserved));
        buffer += sizeof(dir_rec->xa_record.reserved);
    }
#endif // USE_XA_REC
}

//////////////////////////////////////////////////////////////////////////////

int cdfs_init(unsigned char *work_dir_tbl, int size)
{
    int ret;
    dir_rec_struct dir_rec;

    dir_tbl = work_dir_tbl;
    dir_tbl_size = size;
    if (size < 4096)
    {
        return IAPETUS_ERR_INVALIDARG;
    }

    sect_buf = work_dir_tbl+2048;
    sect_buffered = 0;

    // Read in lba 16
    if ((ret = cd_read_sector(dir_tbl, 166, SECT_2048, 2048)) != IAPETUS_ERR_OK)
    {
printf("Error LBA 16 read failure !\n");
        return ret;
    }

    copy_dir_record(dir_tbl+0x9C, &dir_rec);

    // Ok, now we have the root directory(that's good enough for now)
    root_lba = dir_rec.lba;
    root_size = (dir_rec.size / 2048);

printf("root_lba : %d (%08X)\n", root_lba, root_lba);
printf("root_size : %d (%08X)\n", root_size, root_size);

    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

int cdfs_open(char *path, file_struct *file)
{
    char *p;
    char dir_name[32];
    dir_rec_struct dir_rec;
    unsigned long lba;
    unsigned long size;
    int sectors_left;
    unsigned char *work_buffer;
    int done=0;
    int ret;

    /* Clear informations that may remain from a
     * previous CD-ROM file operation.
     * Example : open existing file, then open missing file.
     * (I know that return codes are here for that.)
     */
    memset((void*)file, 0, sizeof(file_struct));

    // Starting with the root directory, go through each subdirectory,
    // read the directory table, and find the next subdirectory. Once we're
    // in the correct level, parse through the table and find the file.

    if ((ret = cd_read_sector(dir_tbl, 150+root_lba, SECT_2048, 2048)) != IAPETUS_ERR_OK)
    {
        return ret;
    }

    lba = root_lba + 1;
    sectors_left = root_size - 1;
    work_buffer = dir_tbl;

    if (path == NULL)
    {
        int i;
        // Get first file(I may remove this feature yet)
        for (i = 0; i < 3; i++)
        {
            if (work_buffer[0] == 0)
            {
                return IAPETUS_ERR_FILENOTFOUND;
            }

            copy_dir_record(work_buffer, &dir_rec);
            work_buffer += dir_rec.record_size;
        }

        // We're done.
        file->lba = dir_rec.lba;
        file->size = dir_rec.size;
        file->sect_pos = 0;
        file->pos = 0;

        return IAPETUS_ERR_OK;
    }

    /* Skip initial slash, if exist. */
    if(path[0] == '/')
    {
        path++;
    }

    while(!done)
    {
        if ((p = strchr(path, '/')) == NULL)
        {
            // It's just the filename now
            done = 1;
            size = strlen(path);
        }
        else
        {
            size = p - path;
        }

        strncpy(dir_name, path, size);
        dir_name[size] = '\0';
        if(strlen(path) > (size+1))
        {
            path += size+1;
        }
        else
        {
            path += strlen(path);
        }
        if((size == 0) || ((*path) < ' '))
        {
            done = 1;
        }

        // Now that we've got the current subdirectory's name, let's find its
        // table.
        for (;;)
        {
            int i;

            copy_dir_record(work_buffer, &dir_rec);
            work_buffer += dir_rec.record_size;

            /* Remove trailing ";" character, if exist. */
            for(i=0; i<32; i++)
            {
                if(dir_rec.name[i] == ';')
                {
                    dir_rec.name[i] = '\0';
                }
            }

            if (strcasecmp(dir_rec.name, dir_name) == 0 ||
                strncasecmp(dir_rec.name, dir_name, strlen(dir_name)) == 0)
            {
                lba = dir_rec.lba;
                sectors_left = dir_rec.size / 2048;
                break;
            }

            // If record size of the next entry is zero, it means we're at the end
            // of the directory record data in the current sector
            if (work_buffer[0] == 0)
            {
                // Let's see if we can read in another sector yet
                if (sectors_left > 0)
                {
                    // Read in new sector
                    if ((ret = cd_read_sector(dir_tbl, 150+lba, SECT_2048, 2048)) != IAPETUS_ERR_OK)
                    {
                        return ret;
                    }
                    lba++;
                    sectors_left--;
                    work_buffer = dir_tbl;
                }
                else
                {
                    // We can't, let's bail
                    return IAPETUS_ERR_FILENOTFOUND;
                }
            }
        }

        if (done)
        {
            break;
        }

        // Ok, we've found the next directory table, time to read it
        if ((ret = cd_read_sector(dir_tbl, 150+lba, SECT_2048, 2048)) != IAPETUS_ERR_OK)
        {
            return ret;
        }
        lba++;
        sectors_left--;
        work_buffer = dir_tbl;
    }

    // Ok, time to fill out the file structure
    file->lba = lba;
    file->size = dir_rec.size;
    file->sect_pos = 0;
    file->pos = 0;

    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

int cdfs_filelist(scdf_file_entry_t* list_ptr, unsigned long list_count, unsigned long list_offset, char* path, unsigned long* file_count_ptr)
{
    char *p;
    char dir_name[32];
    dir_rec_struct dir_rec;
    unsigned long lba;
    unsigned long size;
    int sectors_left;
    unsigned char *work_buffer;
    int done=0;
    int ret;
    long file_index = 0;
    long list_index = 0;

    /* Initialize output parameters. */
    if(file_count_ptr)
    {
        *file_count_ptr = 0;
    }
    if(list_ptr)
    {
        memset((void*)list_ptr, 0, list_count*sizeof(scdf_file_entry_t));
    }

    // Starting with the root directory, go through each subdirectory,
    // read the directory table, and find the next subdirectory. Once we're
    // in the correct level, parse through the table and find the file.

    if ((ret = cd_read_sector(dir_tbl, 150+root_lba, SECT_2048, 2048)) != IAPETUS_ERR_OK)
    {
        return ret;
    }

    lba = root_lba + 1;
    sectors_left = root_size - 1;
    work_buffer = dir_tbl;

    /* Skip initial slash, if exist. */
    if(path[0] == '/')
    {
        path++;
    }

    while(!done)
    {
        if ((p = strchr(path, '/')) == NULL)
        {
            /* Directory change ended. */
            //done = 1;
            size = strlen(path);
        }
        else
        {
            size = p - path;
        }

        if((size == 0) || ((*path) < ' '))
        {
            /* Directory change ended. */
            done = 1;
            size = 0;
        }
        else
        {

            strncpy(dir_name, path, size);
            dir_name[size] = '\0';
            if(strlen(path) > (size+1))
            {
                path += size+1;
            }
            else
            {
                path += strlen(path);
            }
        }

        // Now that we've got the current subdirectory's name, let's find its
        // table.
        for (;;)
        {
            int i;
            copy_dir_record(work_buffer, &dir_rec);
            work_buffer += dir_rec.record_size;

            /* Remove trailing ";" character, if exist. */
            for(i=0; i<32; i++)
            {
                if(dir_rec.name[i] == ';')
                {
                    dir_rec.name[i] = '\0';
                }
            }

            if(done)
            {
                if(dir_rec.name[0] < ' ')
                {
                    /* Skip "." and ".." directories. */
                }
                else
                {
                    if((list_ptr)
                    && (list_count != 0)
                    && (file_index >= list_offset)
                    && (file_index < (list_offset+list_count)))
                    {
                        list_ptr[list_index].flags = FILEENTRY_FLAG_ENABLED;
                        if(dir_rec.flags & 0x02)
                        {
                            list_ptr[list_index].flags |= FILEENTRY_FLAG_FOLDER;
                        }
                        strncpy(list_ptr[list_index].filename, dir_rec.name, SC_FILENAME_MAXLEN);
                        list_ptr[list_index].filename[SC_FILENAME_MAXLEN-1] = '\0';
                        list_ptr[list_index].file_index = file_index;
                        list_index++;
                    }
                    file_index++;
                }
            }
            else
            {
                if (strcasecmp(dir_rec.name, dir_name) == 0 ||
                    strncasecmp(dir_rec.name, dir_name, strlen(dir_name)) == 0)
                {
                    lba = dir_rec.lba;
                    sectors_left = dir_rec.size / 2048;
                    break;
                }
            }

            // If record size of the next entry is zero, it means we're at the end
            // of the directory record data in the current sector
            if (work_buffer[0] == 0)
            {
                // Let's see if we can read in another sector yet
                if (sectors_left > 0)
                {
                    // Read in new sector
                    if ((ret = cd_read_sector(dir_tbl, 150+lba, SECT_2048, 2048)) != IAPETUS_ERR_OK)
                    {
                        return ret;
                    }
                    lba++;
                    sectors_left--;
                    work_buffer = dir_tbl;
                }
                else
                {
                    break;
                    // We can't, let's bail
                    //return IAPETUS_ERR_FILENOTFOUND;
                }
            }
        }

        if (done)
        {
            break;
        }

        // Ok, we've found the next directory table, time to read it
        if ((ret = cd_read_sector(dir_tbl, 150+lba, SECT_2048, 2048)) != IAPETUS_ERR_OK)
        {
            return ret;
        }
        lba++;
        sectors_left--;
        work_buffer = dir_tbl;
    }

    /* Update output parameter. */
    if(file_count_ptr)
    {
        *file_count_ptr = file_index;
    }

    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

int cdfs_seek(file_struct *file, int offset, int seek_type)
{
    switch(seek_type)
    {
    case CDFS_SEEK_SET:
        file->sect_pos = offset / 2048;
        file->pos = offset % 2048;
        break;
    case CDFS_SEEK_CUR:
        file->sect_pos = file->sect_pos + ((file->pos + offset) / 2048);
        file->pos = (file->pos + offset) % 2048;
        break;
    case CDFS_SEEK_END:
        file->sect_pos = (file->size - 1 - offset) / 2048;
        file->pos = (file->size - 1 - offset) % 2048;
        break;
    default:
        return IAPETUS_ERR_INVALIDARG;
    }

    // Change the CD read position here if need be
    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

int cdfs_read(unsigned char *buffer, int size, int num, file_struct *file)
{
    int ret;

    // Only sector-aligned transfers supported for now
    if (file->pos == 0)
    {
        // Straight sectors reads. Nice and fast
        if ((ret = cd_read_sector(buffer, 150+file->lba+file->sect_pos, SECT_2048, size * num)) == IAPETUS_ERR_OK)
        {
            file->sect_pos += (size * num);
        }
        return ret;
    }
    else
    {
        return IAPETUS_ERR_UNSUPPORTED;
    }

    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

int cdfs_close(file_struct *file)
{
    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////


unsigned long _cd_workbuff[4096/sizeof(unsigned long)];

int _folder_depth = 0;
void extract_iso_folder(char* base_out, char* base_path, char* folder)
{
    int i;
    char out_folder[MAX_PATH] = {'\0'};
    char path[MAX_PATH] = {'\0'};

    unsigned long list_offset = 0;
    unsigned long file_count = 0;
    int items_count = 0;

    // if(_folder_depth > 10)
    // {
    //     return;
    // }
    _folder_depth++;

    strcpy(out_folder, base_out);
    if(folder[0] != '\0')
    {
        strcat(out_folder, "\\");
        strcat(out_folder, folder);
    }

    //char folder_wk[MAX_PATH];
    //sprintf(folder_wk, "%s\\%s", out_folder, list[i].filename);
    //CreateDirectoryA(folder_wk, NULL);
    CreateDirectoryA(out_folder, NULL);

    strcpy(path, base_path);
    strcat(path, folder);
    strcat(path, "/");

    do
    {
#define DIR_LIST_COUNT 10
        scdf_file_entry_t list[DIR_LIST_COUNT] = {'\0'};
        file_count = 0;
        cdfs_filelist(list, DIR_LIST_COUNT, list_offset, path, &file_count);

        /* Get number if items in list. */
        items_count = (int)(file_count - list_offset);
        if(items_count > DIR_LIST_COUNT)
        {
            items_count = DIR_LIST_COUNT;
        }

        for(i=0; i<items_count; i++)
        {
            unsigned long size = 0;
            int is_file = 0;

            if((list[i].flags & (FILEENTRY_FLAG_ENABLED | FILEENTRY_FLAG_FOLDER)) == FILEENTRY_FLAG_ENABLED)
            {
                file_struct cdrom_file;
                char filename_wk[256];

                sprintf(filename_wk, "%s%s", path, list[i].filename);
                //printf("File open :\"%s\"\n", filename_wk);

                if(cdfs_open(filename_wk, &cdrom_file) == IAPETUS_ERR_OK)
                {
                    is_file = 1;
                    size = cdrom_file.size;

                    if(size < (700*1024*1024))
                    {
                        unsigned char* buff = (unsigned char*)malloc(size * sizeof(unsigned char));
                        if(buff)
                        {
                            if(cdfs_read(buff, 1, size, &(cdrom_file)) == IAPETUS_ERR_OK)
                            {
                                char filename_out[MAX_PATH];
                                FILE* fp;

                                sprintf(filename_out, "%s\\%s", out_folder, list[i].filename);
                                fp = fopen(filename_out, "wb");
                                if(fp)
                                {
                                    fwrite(buff, size, 1, fp);
                                    fclose(fp);
                                }
                            }

                            free(buff);
                        }
                    }


                }

                cdfs_close(&cdrom_file);
            }
            else
            {
                //char folder_wk[MAX_PATH];
                //sprintf(folder_wk, "%s\\%s", out_folder, list[i].filename);
                //CreateDirectoryA(folder_wk, NULL);

                printf("D:%d, extract_iso_folder(out_folder:%s, path:%s, list[i].filename:%s)\n", _folder_depth, out_folder, path, list[i].filename);
                extract_iso_folder(out_folder, path, list[i].filename);
            }

            printf("[%2d/%2d] Flags:%02X, size:%6u, Name:\"%s\"\n", list_offset+i+1, items_count, list[i].flags, size, list[i].filename);
        }
        list_offset += items_count;
    } while(items_count != 0);

    printf("file_count:%u\n", file_count);

    _folder_depth--;
}


void extract_whole_iso(char* iso_path, char* base_out)
{
    char* base_path = (char*)"";
    char* folder = (char*)"";

    _folder_depth = 0;

    cd_iso_open(iso_path);

    /* Create base output directory. */
    CreateDirectoryA(base_out, NULL);

    /* Specify which iso file to read. */
    cdfs_init((unsigned char *)_cd_workbuff, sizeof(_cd_workbuff));

    extract_iso_folder(base_out, base_path, folder);

    cd_iso_close();
}

//////////////////////////////////////////////////////////////////////////////
