/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h>

#include "autodump.h"
#include "filebrowser.h"

/*
 * Definition of .BUP file header :
 *  - vmem_bup_header_t
 */
#include "vmem/dev_cart.h"

/*
 * Definition of indirect dump saves names :
 *  - IDUMP_HDR_SAVE_NAME : header save name.
 *  - IDUMP_DAT_SAVE_NAME : compressed data save name.
 */
#include "BackupMngr/bup_indirect_dump.h"

/*
 * Definition Pseudo Saturn Kai internal backup
 *  - General settings (BRAMSET_NAME      ).
 *  - Extra settings   (BRAM_EXTRASET_NAME).
 */
#include "pskai_settings.h"


#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

/* Autodump status for a single save.
 * Stores automatically dumped save name and its CRC.
 * Allows to dump save only when created or modified rather
 * than doing it on every firmware startup.
 *
 * Supports up to AUTODUMP_MAX (FILE_BROWSER_MAXCOUNT -> 500) saves.
 * Note : If more than AUTODUMP_MAX saves in autodump folder, then
 *        saves not in status buffer will be dumped on each firmware startup.
 */
#define AUTODUMP_MAX (FILE_BROWSER_MAXCOUNT)
typedef struct _autodump_entry_t
{
    /* Save name, null terminated. */
    char name[12];

    /* Save data CRC. */
    unsigned long crc32;

    /* Save revision ID (number of times it was written to SD card). */
    unsigned long save_id;
} autodump_entry_t;

#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


/* Address defines for use by autodump function internal use.
 * 
 * AUTODUMP_DIR_BUFF     : BRAM directory list, max BUP_DIR_SIZE (400) BupDir entries.
 * AUTODUMP_SAVE_BUFF    : save data read buffer, max 32KB (around 29KB in fact), also space for .BUP file header (vmem_bup_header_t) is added.
 * AUTODUMP_LISTR_BUFF   : autodump saves list buffer, AUTODUMP_MAX (500) autodump_entry_t entries
 * AUTODUMP_LISTC_BUFF   : autodump saves list buffer, AUTODUMP_MAX (500) * sizeof(autodump_entry_t) + SC_COMP_HEADER_MAXSIZE bytes
 * 
 * All workrams are enclosed by the following definitions :
 * AUTODUMP_WORKRAM_STT  : start address of first workram
 * AUTODUMP_WORKRAM_SIZE : length in bytes of all workrams
 */
#define AUTODUMP_WORKRAM_STT  (0x06004000)

#define AUTODUMP_DIR_BUFF     (AUTODUMP_WORKRAM_STT)
#define AUTODUMP_SAVE_BUFF    (AUTODUMP_DIR_BUFF     + BUP_DIR_SIZE*sizeof(BupDir))
#define AUTODUMP_LISTR_BUFF   (AUTODUMP_SAVE_BUFF    + sizeof(vmem_bup_header_t) + 32*1024)
#define AUTODUMP_LISTC_BUFF   (AUTODUMP_LISTR_BUFF   + AUTODUMP_MAX*sizeof(autodump_entry_t))

#define AUTODUMP_WORKRAM_SIZE ((AUTODUMP_LISTC_BUFF + AUTODUMP_MAX*sizeof(autodump_entry_t)+SC_COMP_HEADER_MAXSIZE) - AUTODUMP_WORKRAM_STT)


/* BUP library internals.
 * These are used outside of autodump, 
 * hence can't be declared on HRAM.
 */
unsigned long _bupLibrary[4096+1];
unsigned long _bupRamWork[2048+1];


/* FDD detection global variable.
 * Used in start game menu as well as when starting Save Data Manager.
 */
int _fdd_detected = 0;


void bram_autodump_init(void)
{
    BupConfig bupConf[BUP_DEVICE_CNT];

    /* Init backup library. */
    BUP_Init((volatile unsigned long *)_bupLibrary, _bupRamWork, bupConf);

    /* Check for FDD availability. */
    _fdd_detected = (bupConf[DEVICE_ID_DISC].partition != 0 ? 1 : 0);
}


unsigned long bup_get_saturn_date(void)
{
    BupDate sat_date;
    unsigned long saturn_time;
    struct tm* t;

    /* Poll SMPC time. */
    t = sc_tim_smpcpoll();

    /* Convert date to Saturn date. */
    sc_bup_conv_date(t, &sat_date);

    saturn_time = BUP_SetDate(&sat_date);

    return saturn_time;
}


void bram_autodump(int sdcard_available, int autodump_enable)
{
    Sint32 ret;
    BupDir* saves_list = (BupDir*)AUTODUMP_DIR_BUFF;
    int i;

    /* 
     * Don't do anything if SD card is not available, because
     * autodump can't be performed in that case.
     */
    if(!sdcard_available)
    {
        return;
    }

    /* 
     * If autodump is not enabled, don't go any further.
     * Note : it's not possible to test this before, because
     *        FDD detection status flag needs to be set
     *        for use later in vmem setup menu.
     */
    if(!autodump_enable)
    {
        return;
    }

    /* List-up saves in internal backup memory. */
    ret = BUP_Dir(DEVICE_ID_BRAM, (Uint8 *)"", BUP_DIR_SIZE, saves_list);
    int bram_saves_cnt = (ret >= 0 ? ret : BUP_DIR_SIZE);


    /* First, create the autodump folder if necessary.
     * And if available, decompress and read autodump list.
     */
    char* autodump_stat_file = "/SAVES/AUTODUMP/LIST.TBL";
    autodump_entry_t* list = (autodump_entry_t*)AUTODUMP_LISTR_BUFF;
    unsigned char* list_comp_buff = (unsigned char*)AUTODUMP_LISTC_BUFF;
    unsigned long list_comp_len;

    list_comp_len = scf_size(SC_FILE_SDCARD, autodump_stat_file);
    if((list_comp_len == 0)
    || (list_comp_len > (AUTODUMP_MAX*sizeof(autodump_entry_t) + SC_COMP_HEADER_MAXSIZE)))
    {
        /* List not available or invalid : create folders and initialize list. */
        scf_mkdir(SC_FILE_SDCARD, "/SAVES");
        scf_mkdir(SC_FILE_SDCARD, "/SAVES/AUTODUMP");
        memset(list, 0, AUTODUMP_MAX*sizeof(autodump_entry_t));
    }
    else
    {
        scf_read(SC_FILE_SDCARD, autodump_stat_file, 0/*offset*/, list_comp_len, list_comp_buff);

        /* Verify list decompressed length : if not
         * at expected length then discard it.
         */
        unsigned long decom_len = decomp_getsize(list_comp_buff, list_comp_len);

        if(decom_len != (AUTODUMP_MAX*sizeof(autodump_entry_t)))
        {
            /* Size of list after decompression not as expected : reset list. */
            memset(list, 0, AUTODUMP_MAX*sizeof(autodump_entry_t));
        }
        else
        {
            /* Expected size : use decompressed list. */
            decomp_exec(list_comp_buff, list_comp_len, (unsigned char*)list);
        }
    }

    /* Count saves autodumped so far. */
    int autodump_saves_cnt = 0;
    for(i=0; i<AUTODUMP_MAX; i++)
    {
        if(list[i].name[0] == '\0')
        {
            break;
        }
        autodump_saves_cnt++;
    }

    /* If no saves in BRAM, but saves in autodump folder, it's very
     * likely time to restore saves automatically dumped so far.
     * So let's show some help in order to restore theses saves.
     */
    if(bram_saves_cnt == 0)
    {
        if(autodump_saves_cnt != 0)
        {
            conio_printf(2,  3, COLOR_YELLOW, "    --- Backup RAM Cleared ?! ---   ");

            conio_printf(2,  5, COLOR_WHITE , "There's some good and bad news.");
            conio_printf(2,  6, COLOR_WHITE , "The bad news : your Saturn's backup");
            conio_printf(2,  7, COLOR_WHITE , "               RAM is empty.");
            conio_printf(2,  8, COLOR_WHITE , "The good news : it's possible to");
            conio_printf(2,  9, COLOR_WHITE , "                recover some saves.");

            conio_printf(2, 11, COLOR_WHITE , "If you want to recover your saves, ");
            conio_printf(2, 12, COLOR_WHITE , "please select \"Save Data Manager\"");
            conio_printf(2, 13, COLOR_WHITE , "from main menu, then select \"Save");
            conio_printf(2, 14, COLOR_WHITE , "Data Import\" feature, and finally");
            conio_printf(2, 15, COLOR_WHITE , "import saves from AUTODUMP folder.");

            conio_printf(2, 17, COLOR_WHITE , "Overview of autodumped saves :");
            for(i=0; i<autodump_saves_cnt; i++)
            {
                /* "In case of" save name terminating null character clear. */
                list[i].name[11] = '\0';
            }
            /* Show name of the first saves in autodump list. */
            for(i=0; i<8; i++)
            {
                conio_printf( 2, 18+i, COLOR_YELLOW, "%s", list[3*i+0].name);
                if((3*i+0 + 1) >= autodump_saves_cnt) break;

                conio_printf(14, 18+i, COLOR_YELLOW, "%s", list[3*i+1].name);
                if((3*i+1 + 1) >= autodump_saves_cnt) break;

                conio_printf(26, 18+i, COLOR_YELLOW, "%s", list[3*i+2].name);
                if((3*i+2 + 1) >= autodump_saves_cnt) break;
            }

            while(1)
            {
                if(pad_read()->delta[0] & (PAD_A|PAD_B|PAD_C|PAD_X|PAD_Y|PAD_Z|PAD_START))
                {
                    break;
                }
            }

            clear_screen();
        }

        /* No save in BRAM, so no need to go further. */
        memset((void*)AUTODUMP_WORKRAM_STT, 0x00, AUTODUMP_WORKRAM_SIZE);
        return;
    }

    /*
     * If there's potentially not enough room in autodump list, 
     * clear it in order to list all saves currently in BRAM.
     *
     * This typically happens after autodumping many different saves
     * copied (or created, imported, etc), then deleted : theses
     * deleted saves remain in autodump folder as well as in its list, 
     * hence the need to purge list from time to time.
     */
    if((autodump_saves_cnt+bram_saves_cnt) > AUTODUMP_MAX)
    {
        memset(list, 0, AUTODUMP_MAX*sizeof(autodump_entry_t));
        autodump_saves_cnt = 0;
    }


    /* Verify each saves in internal memory  */
    int list_modified = 0;
    for(i=0; i<bram_saves_cnt; i++)
    {
        /* Emergency exit if A-C, X-Z or Start is pushed */
        if(pad_read()->new[0] & (PAD_A|PAD_B|PAD_C|PAD_X|PAD_Y|PAD_Z|PAD_START))
        {
            break;
        }

        BupDir* tb = saves_list + i;

        /* Save filename/comment sanity check. */
        tb->filename[11] = '\0';
        tb->comment[10] = '\0';

        /* Skip indirect dump saves. */
        if((memcmp(tb->filename, IDUMP_HDR_SAVE_NAME, 11) == 0)
        || (memcmp(tb->filename, IDUMP_DAT_SAVE_NAME, 11) == 0))
        {
            continue;
        }

        /* Read save and compute its CRC. */
        vmem_bup_header_t* save_header = (vmem_bup_header_t*)AUTODUMP_SAVE_BUFF;
        unsigned char* save_data = (unsigned char*)(AUTODUMP_SAVE_BUFF+sizeof(vmem_bup_header_t));

        memset(save_data, 0, tb->datasize);
        BUP_Read(DEVICE_ID_BRAM, tb->filename, save_data);

        unsigned long crc = crc32_calc((const void *)save_data, tb->datasize);

        /* Look for save with same name and CRC in list. */
        int j;
        int autodump_add_save = 1;
        int autodump_save_id = -1;
        for(j=0; j<autodump_saves_cnt; j++)
        {
            autodump_entry_t* ad = list+j;
            if((memcmp(ad->name, tb->filename, 11) == 0))
            {
                /* Save name found in list : remember its ID in list
                 * in order to update its CRC if necessary.
                 */
                autodump_save_id = j;
                if(ad->crc32 == crc)
                {
                    /* Both save name and CRC matched : no
                     * need to autodump this save.
                     */
                    autodump_add_save = 0;
                    break;
                }
            }
        }

        scd_logout("[%3d/%3d] name:%s, add_save:%d, save_id:%d", 
            i, bram_saves_cnt, 
            tb->filename, 
            autodump_add_save, 
            autodump_save_id);

        if(autodump_add_save)
        {
            /* Display information about current save and progress bar.
             *
             * However, don't display anything when processing
             * Pseudo Saturn Kai settings, because this may be
             * modified frequently (for example when using ODE), 
             * and size is small (one block), hence shouldn't
             * take a lot of time when writing to SD card.
             *
             * This is also done with extra settings file, which
             * is updated when modifying custom loader list, or
             * custom cheat codes.
             */
            if((memcmp(tb->filename, BRAMSET_NAME      , 11) != 0)
            && (memcmp(tb->filename, BRAM_EXTRASET_NAME, 11) != 0))
            {
                conio_printf( 2, 10, COLOR_WHITE , "     ~~~ Save Data AutoDump ~~~     ");
                conio_printf( 2, 11, COLOR_WHITE , "Name:");
                conio_printf( 8, 11, COLOR_YELLOW, "%s", tb->filename);
                conio_printf( 2, 12, COLOR_WHITE , "Size:           ");
                if(tb->datasize < 2048)
                {
                    conio_printf( 8, 12, COLOR_YELLOW, "%4d bytes", tb->datasize);
                }
                else
                {
                    conio_printf( 8, 12, COLOR_YELLOW, "%2d KB", tb->datasize>>10);
                }
                conio_drawprogress(
                    2*CONIO_CHAR_W, 13*CONIO_CHAR_H, 
                    36*CONIO_CHAR_W, 1*CONIO_CHAR_H, 
                    bram_saves_cnt, i);
                conio_printf(2, 15, COLOR_WHITE, "Hold B button to Skip AutoDump");
            }

            /* Need to add/append save in list and to write
             * save data (.BUP format) to autodump folder.
             *
             * Also, don't forget to check that list is not full.
             * This check is not mandatory because of autodump list
             * automatic purge, but adding it doesn't hurts.
             */
            if(autodump_save_id < 0)
            {
                /* Not in list : append save informations, or
                 * overwrite last save information if list is full.
                 */
                if(autodump_saves_cnt >= AUTODUMP_MAX)
                {
                    autodump_save_id = autodump_saves_cnt-1;
                }
                else
                {
                    autodump_save_id = autodump_saves_cnt;
                    autodump_saves_cnt++;
                }

                /* Clear save ID (revision ID) when adding save
                 * to list.
                 * When updating save in list, this ID is incremented, 
                 * so that it's possible to verify the number of times the
                 * save was updated.
                 * Only latest revision of the save is kept in SD card, 
                 * so this information is not very useful, but it's better
                 * than always setting it to zero.
                 */
                list[autodump_save_id].save_id = 0;
            }

            /* Add entry to list. */
            autodump_entry_t* ad = list+autodump_save_id;
            memcpy(ad->name, tb->filename, 12);
            ad->crc32 = crc;
            ad->save_id++;

            /* Indicate that list was modified. */
            list_modified = 1;

            /* Set informations in header. */
            memset(save_header, 0, sizeof(vmem_bup_header_t));
            memcpy((void*)(save_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);
            memcpy((void*)(&(save_header->dir)), (void*)tb, sizeof(BupDir));
            save_header->date = bup_get_saturn_date();
            save_header->save_id = ad->save_id;

            /* Setup path to save file on SD card. */
            char save_path[SC_PATHNAME_MAXLEN];
            int k;
            for(k=0; k<11; k++)
            {
                char c = tb->filename[k];
                if(((c >= '0') && (c <= '9'))
                || ((c >= 'A') && (c <= 'Z'))
                || (c == '_'))
                {
                    /* Valid save filename character. */
                }
                else
                {
                    /* Invalid save filename character : replace with underscore. */
                    if((c >= 'a') && (c <= 'z'))
                    {
                        tb->filename[k] = 'A' + (c - 'a');
                    }
                    else
                    {
                        tb->filename[k] = '_';
                    }
                }
                
            }
            sprintf(save_path, "/SAVES/AUTODUMP/%s.BUP", tb->filename);

            /* Write to SD card. */
            scf_write(save_header, sizeof(vmem_bup_header_t) + tb->datasize, SC_FILE_SDCARD, save_path);
        }
    }

    /* If needed, save autodump list to SD card. */
    if(list_modified)
    {
        /* Compress list. */
        list_comp_len = comp_exec((unsigned char*)list, AUTODUMP_MAX*sizeof(autodump_entry_t), list_comp_buff);

        /* (Over)write list. */ 
        scf_write(list_comp_buff, list_comp_len, SC_FILE_SDCARD, autodump_stat_file);
    }

    /* Clear screen and workrams. */
    clear_screen();
    memset((void*)AUTODUMP_WORKRAM_STT, 0x00, AUTODUMP_WORKRAM_SIZE);
}
