/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "autoset.h"

#ifdef SC_SATURN
#   include <sc_saturn.h>

    /* Custom loader module integration. */
#   include "custom_loader.h"
#endif // SC_SATURN


/**
 * Internal definitions for game loader, vmem, etc automatic settings.
**/

/*
 * Dummy CRC value.
 * It is used in order to define settings for games not in
 * list (working fine with default settings), or to stuff CRC
 * list in CRC <-> settings table.
 *
 * As long as the value doesn't collides with CRC from an
 * existing game, everything is OK, so let's hope that
 * simple zero value won't cause collision problems.
 */
#define AUTOSET_DUMMY_CRC (IPHEADER_DUMMY_CRC)

/*
 * CRC count per automatic settings entry.
 * Some games may have several variations in IP header, typically
 * regarding region or version. And rather than defining same
 * automatic settings entries for each variations of the game, it
 * is wiser to allow multiple definitions of CRC in a same entry.
 *
 * Let's first try with a "not so large, not so small" value now, 
 * and adjust it when populating the list.
 */
#define AUTOSET_CRC_CNT 7


typedef struct _loader_settings_t
{
    /*
     * IP header CRC values.
     *
     * If first CRC value is set to AUTOSET_DUMMY_CRC, then entry
     * is considered as terminating one for the automatic settings
     * table, and consequently :
     *  - entries after the terminating one won't be checked.
     *  - terminating entry will be used for CRC not defined in the table.
     *    (-> terminating entry = default settings entry)
     *
     * If entry has less than AUTOSET_CRC_CNT CRCs, then unused CRC
     * definitions need to be set to AUTOSET_DUMMY_CRC and placed at
     * the end of CRCs list.
     * Example : {crc1, crc2, dummy, dummy, ..., dummy}
     */
    unsigned long crc[AUTOSET_CRC_CNT];

    /*
     * Settings informations.
     */
    unsigned long set;
} loader_settings_t;


/**
 * Default automatic settings.
 * Used for games not found in custom/floppy tables :
 *  - Can use both CWX/JHL loader
 *  - Can use vmem for both memory cartridge and FDD
 *  - Game don't uses FDD
 *  - Game don't uses/requires RAM expansion
**/
loader_settings_t _autoset_default_settings =
{
    {AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC}, 
    AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
    | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
};

/* Another special setting, used when retrieving custom loader settings. */
loader_settings_t _autoset_custom_settings;


/**
 * Custom automatic settings table.
 *
 * Used for games requiring a specific loader and/or
 * not compatible with vmem.
 *
 * Also, it contains specific settings for games accessing
 * Floppy Disk Drive.
**/
#include "BackupMngr/app_title.h"
loader_settings_t _autoset_custom_tbl[] =
{
    {
        /* Settings for games using Floppy Disk Drive.
         * Automatic settings are similar to default ones, except that it
         * is indicated that the game use FDD, so that it is possible to
         * prompt user to select between using real FDD or vmem FDD when
         * Gamer's Cartridge and FDD are simultaneously in use.
         *
         * When automatic loader is selected, vFDD is preferred. In the case
         * it is needed to use FDD with theses games, then manually selecting
         * CWX loader should display FDD/vFDD selection menu.
         *
         * List of games accessing FDD :
         * - Culdcept
         * - Dezaemon 2
         * - Eiyuu Shigan: Gal ACT Heroism
         * - Game Basic for Sega Saturn
         * - Panzer Dragoon Saga (JP, US and EU versions)
         * - Sangokushi Koumeiden
         * ( http://segaretro.org/Saturn_Floppy_Drive )
         *
         * Note : Panzer Dragoon Saga requires custom loader settings, 
         *        hence stored in separate entries.
         *
         * Update on January 2019 : the following games were added on
         * Sega Retro page above, so that autoset is updated accordingly.
         * (Special thanks to privateye for informing me about the update !)
         *  - Mouri Motonari: Chikai no Sanshi
         *  - Pad Nifty
         *  - Pad Nifty 1.1 & Habitat II
         *  - Sega Saturn-you Word Processor Set (EGWORD)
         *
         * Update on February 2019 : that both of these games support
         * the Saturn Floppy Drive for saving high score data:
         *  - Midway Presents Arcade's Greatest Hits
         *  - Midway Presents Arcade's Greatest Hits: The Atari Collection 1 (US and EU versions)
         * (Theses games weren't released in USA and Europe only, so that's
         * surprising to see them supporting a peripheral released only in Japan !)
         *
         * Update on October 2019 : added CRC for EGWORD V1.001.
         * Update on December 2019 : added back CRC for EGWORD V1.021 (I don't know why it was deleted, sorry).
         *
         * September 2020 update : more games are added thanks to the
         * informations from the page below.
         * https://www.segasaturnshiro.com/2020/07/20/secrets-of-the-floppy-disk-drive/
         * Special thanks to privateye for updating this compatibility list !
         * (*) : same IP header CRC as the previous game.
         *
         * Crusader: No Remorse
         *  - Crusader - No Remorse    (Europe)
         *  - Crusader - No Remorse    (France)    (*)
         *  - Crusader - No Remorse    (Germany)   (*)
         *  - Crusader - No Remorse    (USA)       (*)
         * 
         * NBA Action
         *  - NBA Action               (Europe)
         *  - NBA Action               (USA)
         * 
         * NHL All-Star Hockey
         *  - NHL All-Star Hockey      (Europe)
         *  - NHL All-Star Hockey      (USA)
         * 
         * Shockwave Assault
         *  - Shockwave Assault        (Europe)
         *  - Shockwave Assault        (USA)       (*)
         */
        {
            /* Culdcept */
            0xCF79E31F, // V1.004
            0x1A66D07D, // V2.000

            /* Dezaemon 2 */
            0x56A77451, 

            /* Eiyuu Shigan: Gal ACT Heroism */
            0x49F4B547, 

            /* Game Basic for Sega Saturn */
            0x53265516, 

            /* Sangokushi Koumeiden */
            0xA5AA7E19, 

            AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
    }, 
    {
        {
            /*
             * |ip_crc :48E9D7BB CRC[48E9D7BB]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-76
             * |prd_nb :T-7646G
             * |version:V1.000
             * |date   :19970801
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JF
             * |title  :MOURIMOTONARI CHIKAINOSANSHI
             */
            0x48E9D7BB, 

            /*
             * |ip_crc :B3939230 CRC[B3939230]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :610-6276
             * |version:V1.003
             * |date   :19960620
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :J
             * |title  :SEGA SATURN     FDD OPERATOR
             */
            0xB3939230, 

            /*
             * |ip_crc :542268AB CRC[542268AB]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-76
             * |prd_nb :T-7626G
             * |version:V1.001
             * |date   :19961024
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JFK
             * |title  :EGWORD
             */
            0x542268AB, 

            /*
             * |ip_crc :38835A7A CRC[38835A7A]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-76
             * |prd_nb :T-7626G
             * |version:V1.021
             * |date   :19970227
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JFK
             * |title  :EGWORD
             */
            0x38835A7A, 

            /*
             * |ip_crc :6FA8B405 CRC[6FA8B405]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-76
             * |prd_nb :T-7641G
             * |version:V2.000
             * |date   :19970619
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JFKHU
             * |title  :EGWORD
             */
            0x6FA8B405, 

            /*
             * |ip_crc :611A5AF8 CRC[611A5AF8]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :GS-7101
             * |version:V1.007
             * |date   :19960701
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JKDF
             * |title  :PAD NIFTY
             */
            0x611A5AF8, 

            /*
             * |ip_crc :69E2EC03 CRC[69E2EC03]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :GS-7109
             * |version:V1.003
             * |date   :19971003
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JKDFM
             * |title  :PAD NIFTY/HABITAT2
             */
            0x69E2EC03
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
    }, 
    {
        {
            /*
             * |CRC : 0xC932675A
             * | 0000:SEGA SEGASATURN 
             * | 0010:SEGA TP T-97    
             * | 0020:T-9703H   V1.000
             * | 0030:19961118CD-1/1  
             * | 0040:U               
             * | 0050:J               
             * | 0060:WILLIAMS ARCADES
             * | 0070: GREATEST HITS  
             * | 0080:                
             * | 0090:                
             * | 00A0:                
             * | 00B0:                
             * | 00C0:                
             */
            0xC932675A, 

            /*
             * |CRC : 0x32E801C8
             * | 0000:SEGA SEGASATURN 
             * | 0010:SEGA TP T-254   
             * | 0020:T-25413H  V1.000
             * | 0030:19971212CD-1/1  
             * | 0040:E               
             * | 0050:JAM             
             * | 0060:MIDWAY PRESENTS 
             * | 0070:ARCADES GREATEST
             * | 0080:HITS            
             * | 0090:THE ATARI COLLEC
             * | 00A0:TION VOLUME 1   
             * | 00B0:                
             * | 00C0:                
             */
            0x32E801C8, 

            /*
             * |CRC : 0xCF5E472D
             * | 0000:SEGA SEGASATURN 
             * | 0010:SEGA TP T-97    
             * | 0020:T-9706H   V1.000
             * | 0030:19970411CD-1/1  
             * | 0040:U               
             * | 0050:JAM             
             * | 0060:MIDWAY PRESENTS 
             * | 0070:ARCADES GREATEST
             * | 0080:HITS            
             * | 0090:THE ATARI COLLEC
             * | 00A0:TION VOLUME 1   
             * | 00B0:                
             * |00C0:                
             */
            0xCF5E472D, 

            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
    }, 
    {
        {
            /* Crusader: No Remorse
             *  - Crusader - No Remorse    (Europe)
             *  - Crusader - No Remorse    (France)    (*)
             *  - Crusader - No Remorse    (Germany)   (*)
             *  - Crusader - No Remorse    (USA)       (*)
             * 
             * [Crusader - No Remorse (USA).img]
             * |ip_crc :CC1EDF8A CRC[CC1EDF8A]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-50
             * |prd_nb :T-5014H
             * |version:V1.000
             * |date   :19961203
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :J
             * |title  :CRUSADER
             */
            0xCC1EDF8A, 

            /* NBA Action
             *  - NBA Action               (Europe)
             *  - NBA Action               (USA)
             * 
             * [NBA Action (Europe).img]
             * |ip_crc :0FF72312 CRC[0FF72312]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :MK81103-50
             * |version:V1.026
             * |date   :05231996
             * |dev_id :CD-1/1
             * |area   :E
             * |pers   :JT
             * |title  :   NBA ACTION  BASKETBALL
             * 
             * ------------------------------------------
             * [NBA Action (USA).img]
             * |ip_crc :0FD00948 CRC[0FD00948]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :MK-81103
             * |version:V1.026
             * |date   :05151996
             * |dev_id :CD-1/1
             * |area   :TU
             * |pers   :JT
             * |title  :   NBA ACTION  BASKETBALL
             */
            0x0FF72312, 
            0x0FD00948, 

            /* NHL All-Star Hockey
             *  - NHL All-Star Hockey      (Europe)
             *  - NHL All-Star Hockey      (USA)
             * 
             * [NHL All-Star Hockey (Europe).img]
             * |ip_crc :6276A289 CRC[6276A289]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :MK-8100250
             * |version:V1.001
             * |date   :09111995
             * |dev_id :CD-1/1
             * |area   :E
             * |pers   :JT
             * |title  :  NHL ALL-STAR   HOCKEY
             * 
             * ------------------------------------------
             * [NHL All-Star Hockey (USA) (2S).img]
             * |ip_crc :C5A519B5 CRC[C5A519B5]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :MK-81002
             * |version:V1.072
             * |date   :08251995
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :JT
             * |title  :  NHL ALL-STAR   HOCKEY
             */
            0x6276A289, 
            0xC5A519B5, 

            /* Shockwave Assault
             *  - Shockwave Assault        (Europe)
             *  - Shockwave Assault        (USA)       (*)
             * 
             * [Shockwave Assault (USA).img]
             * |ip_crc :52E92182 CRC[52E92182]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-50
             * |prd_nb :T-5005H
             * |version:V1.000
             * |date   :19960628
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :JA
             * |title  :SHOCKWAVE   ASSAULT
             */
            0x52E92182, 

            AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
    }, 

    {
        /* Settings for Save Data Manager.
         * According to usage, this application may require access to FDD, 
         * or virtual memory cartridge/FDD.
         *
         * Examples :
         *  - Manage saves on virtual memory cartridge : must select vFDD
         *  - Dump internal backup memory to SD card: both are OK
         *  - Indirect dump via FDD : require to enable FDD when decompressing
         *    dumped data from FDD to SD card.
         *
         * For theses reasons, FDD/vFDD selection menu is displayed even when
         * "automatic" loader is selected.
         */
        {
            /* Save Data Manager */
            BUP_APP_IPHDR_CRC, 

            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
        | AUTOSET_VFDD_SELECTMENU
    }, 

    {
        /* Settings for games not compatible with Floppy Disk Drive.
         * Special thanks to SEGA Saturn, Shiro! and privateye for gathering
         * this compatibility information !
         * https://www.segasaturnshiro.com/2020/07/20/secrets-of-the-floppy-disk-drive/
         * 
         * - Madden NFL 97
         * 
         * Compatible with vCart and this article also indicates that it is
         * semi-compatible with a real FDD. However, I confirm that you cannot
         * perform load/save operations with vFDD in the following versions:
         *   - Madden NFL 97 (Europe)
         *   - Madden NFL 97 (USA)
         */
        {
            /* [Madden NFL 97 (Europe).img]
             * |ip_crc :6BF37284 CRC[6BF37284]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-50
             * |prd_nb :T-5010H
             * |version:V1.000
             * |date   :19961001
             * |dev_id :CD-1/1
             * |area   :E
             * |pers   :JT
             * |title  :MADDEN NFL '97
             * 
             * ------------------------------------------
             * [Madden NFL 97 (USA).img]
             * |ip_crc :715FAC8F CRC[715FAC8F]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-50
             * |prd_nb :T-5010H
             * |version:V1.000
             * |date   :19960920
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :JT
             * |title  :MADDEN NFL '97
             */
            0x6BF37284, 
            0x715FAC8F, 

            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE /*| AUTOSET_VFDD_COMPATIBLE*/
    }, 

    {
        {
           /*
            * Alien Trilogy
            * Version: V1.000, Date: 19960720, Country: USA
            * Report files:#1, #2, #3
            * Comment: Stuck at SEGA logo, laser looks like its trying to read something, 
            *          then kicks you back to multiplayer screen. Patching the ISO with
            *          Pseudo Saturn patcher utility made the game playable. 
            *
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            *
            * IP Header informations :
            * CRC : 0x154FE022
            *  0000:SEGA SEGASATURN 
            *  0010:SEGA TP T-81    
            *  0020:T-8113H   V1.000
            *  0030:19960720CD-1/1  
            *  0040:U               
            *  0050:J               
            *  0060:ALIEN TRILOGY   
            *  0070:                
            *  0080:                
            *  0090:                
            *  00A0:                
            *  00B0:                
            *  00C0:                
            * 
            * [Alien Trilogy (Europe) (En,Fr,Es,It) (7S) (Track 01).bin]
            * |ip_crc :253A0B97 CRC[253A0B97]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8113H-50
            * |version:V1.000
            * |date   :19960729
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :ALIEN TRILOGY
            * 
            * ------------------------------------------
            * [Alien Trilogy (Europe) (En,Fr,Es,It) (Beta) (Track 01).bin]
            * |ip_crc :15B8005D CRC[15B8005D]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-999
            * |prd_nb :T-99901G
            * |version:V1.000
            * |date   :19950418
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :GAME SAMPLE
            * 
            * ------------------------------------------
            * [Alien Trilogy (Japan) (Track 01).bin]
            * |ip_crc :CCDB89CB CRC[CCDB89CB]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8113G
            * |version:V1.001
            * |date   :19960716
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :ALIEN TRILOGY
            */
            0x154FE022, 0x253A0B97, 0x15B8005D, 0xCCDB89CB, 

           /* 2020/09/29 : 4MB hacked version of Samurai Spirits RPG requires
            * JHL loader to boot. Unhacked version does not require JHL but it
            * will be forced to use the same (JHL) loader since hacked version
            * uses exactly the same header as the original one.
            *
            * ------------------------------------------
            * [Shinsetsu_Samurai_Spirits_Bushidou_Retsuden_HACKED_BY_YZB_20200819.img]
            * |ip_crc :8AFC075F CRC[8AFC075F]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-31
            * |prd_nb :T-3112G
            * |version:V1.001
            * |date   :19970520
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :SAMURAI SPIRITS BUSIDO RETSUDEN
            */
            0x8AFC075F, 

            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * ATgOX - Assault Rigs
            * Version: V1.000, Date: 19970417, Country: JPN
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Black screen after Sega logo, which also happens with swap trick, 
            * and previous Pseudo Saturn versions, Game needs to be patched to work properly
            * due to protection. Patching the ISO with Pseudo Saturn patcher utility made
            * the game playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            */
            0x4CD8AD07, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * Batman Forever The Arcade Game
            * Version: V1.000, Date: 19961008, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher
            * utility in order to make it playable. 
            * 
            * 2019/05/26 Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [Batman Forever - The Arcade Game (Japan).img]
            * |ip_crc :5011C129 CRC[5011C129]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8118G
            * |version:V1.000
            * |date   :19961031
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :BATMAN COINOP
            */
            0x8C37FB5F/*USA*/, 0xCD572605/*EUR*/, 0x5011C129/*JPN*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * JvR@WFl[V@`QW@EƋRm` - Capcom Generation ~Dai-2-shuu Makai to Kishi~
            * Version: V1.003, Date: 19980824, Country: JPN
            * Both loader and Virtual Memory Cart worked on: [224B752C, NTSC]
            * Report file:#1
            *
            * IP Header informations :
            * CRC : 0x15044BA0
            *  0000:SEGA SEGASATURN 
            *  0010:SEGA TP T-12    
            *  0020:T-1233G   V1.003
            *  0030:19980824CD-1/1  
            *  0040:J               
            *  0050:J               
            *  0060:CAPCOM GENERATIO
            *  0070:N VOL.2         
            *  0080:                
            *  0090:                
            *  00A0:                
            *  00B0:                
            *  00C0:                
            *
            * [vmem_debug] bup_conf_ptr=0x002FFFF0 [BupConfig*=0x00000000 0x060491A0] context=0x00000000
            * [vmem]bup_init END (unit_id=1 2 0) (workram=0x06049220, libram=0x060C0000)
            * [vmem]bup_dir(device=0, filename="CAPGENE_002", tbsize=1) END ret=0
            * [vmem]bup_dir(device=1, filename="CAPGENE_002", tbsize=1) END ret=1
            * [vmem_debug] bup_conf_ptr=0x002FFFF0 [BupConfig*=0x00000000 0x060491A0] context=0x00000000
            * [vmem]bup_init END (unit_id=1 2 0) (workram=0x06049220, libram=0x060C0000)
            * [vmem]bup_dir(device=0, filename="CAPGENE_002", tbsize=1) END ret=0
            * [vmem]bup_dir(device=1, filename="CAPGENE_002", tbsize=1) END ret=1
            * [vmem_debug] bup_conf_ptr=0x002FFFF0 [BupConfig*=0x00000000 0x060491A0] context=0x00000000
            * [vmem]bup_init END (unit_id=1 2 0) (workram=0x06049220, libram=0x060C0000)
            * [vmem]bup_read(device=1, filename="CAPGENE_002") END ret=0
            * (snip)
            * [vmem]bup_init END (unit_id=1 2 0) (workram=0x06049220, libram=0x060F8000)
            * [vmem]bup_read(device=1, filename="CAPGENE_002") END ret=0
            * [feedback][Game Compatibility][Smooth:Booted fine, everything OK][color:green]
            * [feedback][SD Memory Cart    ][Incompatible:Can't play game (freeze, etc)][color:red]
            * [feedback][Amount Played     ][A little bit:Played it but not a lot][color:yellow]
            *
            * Compatibility memo :
            *  According to presence of log output from vmem, game can boot with CWX loader, 
            *  but it seems that enabling vmem makes the game not working correctly, or that
            *  something after boot makes game not compatible with CWX loader.
            *
            * TODO : try to fix vmem compatibility with this game
            */
            0x15044BA0, 

            /* 2020/09/30 :  'All-Star Baseball '97 Featuring Frank Thomas'
             * and 'Frank Thomas Big Hurt Baseball' are not compatible with
             * Virtual Memory Cart.
             * 
             * [All-Star Baseball '97 Featuring Frank Thomas (USA).img]
             * |ip_crc :A2DF6DC6 CRC[A2DF6DC6]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-81
             * |prd_nb :T-8150H
             * |version:V1.000
             * |date   :19970502
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :JT
             * |title  :ALLSTAR_BASEBALL_97
             */
            0xA2DF6DC6, 

            /* [Frank Thomas Big Hurt Baseball (Europe).img]
             * |ip_crc :71AA8214 CRC[71AA8214]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-81
             * |prd_nb :T-8138H-50
             * |version:V1.000
             * |date   :19960519
             * |dev_id :CD-1/1
             * |area   :E
             * |pers   :JT
             * |title  :FRANK THOMAS BIG HURT BASEBALL
             * 
             * ------------------------------------------
             * [Frank Thomas Big Hurt Baseball (Japan).img]
             * |ip_crc :B09AB25A CRC[B09AB25A]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-81
             * |prd_nb :T-8111G
             * |version:V1.003
             * |date   :19960620
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :JT
             * |title  :FRANK THOMAS BIG HURT BASEBALL
             * 
             * ------------------------------------------
             * [Frank Thomas Big Hurt Baseball (USA).img]
             * |ip_crc :42289998 CRC[42289998]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-81
             * |prd_nb :T-8138H
             * |version:V1.000
             * |date   :19960519
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :JT
             * |title  :FRANK THOMAS BIG HURT BASEBALL
             */
            0x71AA8214, 
            0xB09AB25A, 
            0x42289998, 

            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        //| AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

#if 0 // 2019/05/26 Reported working with CWX loader after patching
    {
        {
           /*
            * Center Ring Boxing
            * Version: V1.02U, Date: 19950929, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            */
            0x319E6885, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
#endif // 0

    {
        {
           /*
            * fXgNVE_[r[ - Destruction Derby
            * Version: V3.200, Date: 19960802, Country: JPN
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable.
            * 
            * 2019/05/06 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [Destruction Derby (Europe).img]
            * |ip_crc :50D51E46 CRC[50D51E46]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-113
            * |prd_nb :T-11303H
            * |version:V2.900
            * |date   :19960728
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JS
            * |title  :DESTRUCTION DERBY
            */
            0x28E71575/*JPN*/, 0x50D51E46/*EUR*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

#if 0 // 2019/05/09 Reported working with CWX loader without the need of patching
    {
        {
           /*
            * fW^s{[lNm~R - Digital Pinball Necronomicon
            * Version: V1.002, Date: 19961011, Country: JPN
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Having a hard time to make this game boot with CWX Loader, 
            * gives an error, much like Skull Fang did in Pseudo 0.082 for me, takes
            * some tries but it does boot and plays fine, JHL Loader works everytime !
            * tried many, many times.
            */
            0xBE584539, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
#endif // 0

    {
        {
           /*
            * Discworld
            * 
            * 2020/08/29 : following versions are incompatible
            *              with CWX loader in RmenuKai :
            *  - Discworld (Japan) 
            *  - Terry Pratchett's Discworld (Europe) 
            * 
            * [Discworld (Japan).img]
            * |ip_crc :B3712D8C CRC[B3712D8C]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-205
            * |prd_nb :T-20502G
            * |version:V1.000
            * |date   :19961009
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :JM
            * |title  :DISCWORLD
            * 
            * ------------------------------------------
            * [Terry Pratchett's Discworld (Europe) (En,Fr,De,Es,It).img]
            * |ip_crc :E6C8E701 CRC[E6C8E701]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-113
            * |prd_nb :T-11302H
            * |version:V5.230
            * |date   :19960718
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JM
            * |title  :DISCWORLD
            */
            0xB3712D8C, 0xE6C8E701, 

           /*
            * Discworld II - Missing, presumed... !?
            * Version: V1.007, Date: 19970710, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo
            * Saturn patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            */
            0xD063565D, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * ciEH[Y - DJ Wars
            *
            * 2019/05/26 Reported working with CWX loader after patching
            * 2019/05/28 Reported as crashing to a black screen after the
            *            SEGA logo when vmem is enabled
            *
            * [DJ Wars (Japan) (Demo) (Track 1).bin]
            * |ip_crc :71055082 CRC[71055082]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-188
            * |prd_nb :6106762
            * |version:V0.004
            * |date   :19971104
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :DJ WARS
            * 
            * [DJ Wars (Japan) (Rev A) (10M) (Track 1).bin]
            * |ip_crc :257F67D9 CRC[257F67D9]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-188
            * |prd_nb :T-18807G
            * |version:V1.005
            * |date   :19971113
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :DJ WARS
            */
            0x71055082, 0x257F67D9, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE
        //| AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * Machine Head
            * Version: V1.000, Date: 19960920, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [Blam! Machinehead (Europe).img]
            * |ip_crc :2904F3FF CRC[2904F3FF]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-115
            * |prd_nb :T-11505H50
            * |version:V1.000
            * |date   :19960830
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :BLAM! MACHINEHEAD
            * 
            * ------------------------------------------
            * [Blam! Machinehead (Japan).img]
            * |ip_crc :4C4212E5 CRC[4C4212E5]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-70
            * |prd_nb :T-7015G
            * |version:V1.000
            * |date   :19970117
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :BLAM MACHINEHEAD
            */
            0xD05F14C4/*USA*/, 0x2904F3FF/*EUR*/, 0x4C4212E5/*JPN*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * Mechwarrior 2
            * Version: V1.000, Date: 19970314, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            * 
            * Note : with CWX loader it gets stuck on SEGA logo screen.
            */
            0xE95E95C1, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
    {
        {
           /*
            * Mortal Kombat II
            * Version: V1.000, Date: 19951229, Country: USA
            * Report files:#1, #2, #3, #4, #5
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * IP Header informations :
            * CRC : 0x78A37A80
            *  0000:SEGA SEGASATURN 
            *  0010:SEGA TP T-81    
            *  0020:T-8103H   V1.000
            *  0030:19951229CD-1/1  
            *  0040:U               
            *  0050:J               
            *  0060:MORTAL KOMBAT II
            *  0070:                
            *  0080:                
            *  0090:                
            *  00A0:                
            *  00B0:                
            *  00C0:                
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            * 
            * Note : With CWX loader it will crash to a black screen after the Probe Entertainment logo.
            */
            0x78A37A80, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * NBA Jam Extreme
            * Version: V1.000, Date: 19961027, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [NBA Jam Extreme (Europe).img]
            * |ip_crc :8798B380 CRC[8798B380]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8120H-50
            * |version:V1.000
            * |date   :19961111
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JT
            * |title  :NBA JAM EXTREME
            * 
            * ------------------------------------------
            * [NBA Jam Extreme (Japan).img]
            * |ip_crc :A6D68C38 CRC[A6D68C38]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8122G
            * |version:V1.000
            * |date   :19961210
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :JT
            * |title  :NBA JAM EXTREME
            */
            0xB3DCAE6B/*USA*/, 0x8798B380/*EUR*/, 0xA6D68C38/*JPN*/, 

           /*
            * NBA Live 97
            * 2020/08/29 : following games only work with JHL loader in RmenuKai :
            * - NBA Live 97 (Europe)
            * - NBA Live 97 (USA)       (same IP CRC as Europe)
            * 
            * [NBA Live 97 (USA).img]
            * |ip_crc :14E422DE CRC[14E422DE]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-50
            * |prd_nb :T-5015H
            * |version:V1.000
            * |date   :19970228
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :J
            * |title  :NBA LIVE 97
            */
            0x14E422DE, 


           /*
            * NBA Live 98
            * Version: V1.000, Date: 19971122, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * 2020/08/29 : following games only work with JHL loader in RmenuKai :
            * - NBA Live 98 (Europe) 
            * - NBA Live 98 (USA)        (same IP CRC as Europe)
            */
            0x6AC0EFE5, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * NBA Jam Tournament Edition
            * Version: V1.000, Date: 19951003, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [NBA Jam - Tournament Edition (Europe) (Track 1).bin]
            * |ip_crc :A95D6DA1 CRC[A95D6DA1]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8102H-50
            * |version:V1.000
            * |date   :19951019
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :NBA JAM TOURNAMENT EDITION
            * 
            * ------------------------------------------
            * [NBA Jam - Tournament Edition (Japan) (Track 1).bin]
            * |ip_crc :26B71A36 CRC[26B71A36]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8102G
            * |version:V1.000
            * |date   :19951102
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :NBA JAM TOURNAMENT EDITION
            */
            0xEB018C55/*USA*/, 0xA95D6DA1/*EUR*/, 0x26B71A36/*JPN*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * Panzer Dragoon Saga CD1
            * Version: V1.000, Date: 19980318, CD-1/4, Country: USA
            * Both loader and Virtual Memory Cart worked on: [224B752C, NTSC]
            * Report file:#1
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * IP Header informations :
            * CRC : 0xCC0DA2F3
            *  0000:SEGA SEGASATURN 
            *  0010:SEGA ENTERPRISES
            *  0020:MK-81307  V1.000
            *  0030:19980318CD-1/4  
            *  0040:U               
            *  0050:AEFJK           
            *  0060:J:AZEL PANZER DR
            *  0070:AGOON RPG       
            *  0080:U:PANZER DRAGOON
            *  0090: SAGA           
            *  00A0:                
            *  00B0:                
            *  00C0:                
            * 
            * Panzer Dragoon Saga CD1
            * Version: V1.000, Date: 19980309, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            *
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0xCC0DA2F3/*USA*/, 0xC2554E94/*JPN*/, 0xF3D06731/*EUR*/, 

           /*
            * Panzer Dragoon Saga CD2
            * Version: V1.000, Date: 19980309, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Panzer Dragoon Saga CD2
            * Version: V1.000, Date: 19980318, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0x81076143/*USA*/, 0x8F5F8D24/*JPN*/, 0xBEDAA481/*EUR*/, 
            AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
    }, 
    {
        {
           /*
            * Panzer Dragoon Saga CD3
            * Version: V1.000, Date: 19980309, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Panzer Dragoon Saga CD3
            * Version: V1.000, Date: 19980318, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0x0CD1DDEC/*USA*/, 0x0289318B/*JPN*/, 0x330C182E/*EUR*/, 

           /*
            * Panzer Dragoon Saga CD4
            * Version: V1.000, Date: 19980309, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Panzer Dragoon Saga CD4
            * Version: V1.000, Date: 19980318, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0x1B12E623/*USA*/, 0x154A0A44/*JPN*/, 0x24CF23E1/*EUR*/, 
            AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
        | AUTOSET_USE_FLOPPYDISK /* <- FDD/vFDD selection menu. */
    }, 
    {
        {
           /*
            * Panzer Dragoon Zwei
            * Version: V1.000, Date: 19960308, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game freezes after first FMV when using CWX loader.
            * JHL Loader plays the game normally. 
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            *
            * Note (2019/05/09) : this game doesn't needs to be patched.
            *
            * Update (202/03/19) : this game needs to be patched to work on
            *  systems with certain motherboard revisions. It was reported
            *  to work unpatched on PAL VA5 and PAL VA13 hardware revisions, 
            *  but doesn't boots without patch on some other systems.
            * Details : https://www.segasaturnshiro.com/2019/10/14/burning-saturn-discs-the-ultimate-guide/ , section "9. Panzer Dragoon Zwei/Saga Only"
            * Succesful testing reports after patching the disc image :
            *  - https://old.reddit.com/r/SegaSaturn/comments/emr2ce/issues_playing_certain_backups/
            *  - https://old.reddit.com/r/SegaSaturn/comments/ev269v/cant_seem_to_get_panzer_dragoon_saga_running/ffxx0i0/
            */
            0x4CF55C64/*USA*/, 0xA257544E/*JPN*/, 0xC518C054/*EUR*/, 0xEB324F78/*PROTO*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
    {
        {
           /*
            * Scud - The Disposable Assassin
            * Version: V1.015, Date: 19970128, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn patcher
            * utility in order to make it playable. (Not tested without patching.) 
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0x4CA7D336, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

#if 0 // 2019/10/27 Vmem reported working (could record and replay ghost data on Gamer's Cartridge with Pseudo Saturn Kai v6.315), so that entry for this game is disabled
    {
        {
           /*
            * ZK[E`sIVbvEvX - Sega Rally Championship Plus
            * Version: V1.000, Date: 19960826, Country: JPN
            * Both loader and Virtual Memory Cart worked on: [224B752C, NTSC]
            * Report files:#1, #2
            * 
            * IP Header informations :
            * CRC : 0x27B906B3
            *  0000:SEGA SEGASATURN 
            *  0010:SEGA ENTERPRISES
            *  0020:GS-9116   V1.000
            *  0030:19960826CD-1/1  
            *  0040:JT              
            *  0050:JASEDX          
            *  0060:SEGA RALLY CHAMP
            *  0070:IONSHIP PLUS    
            *  0080:                
            *  0090:                
            *  00A0:                
            *  00B0:                
            *  00C0:                
            * 
            * Incorrect CRC ?! Because peripheral code is not the official one.
            * Maybe it was modified by region patcher program or something like that.
            * // CRC : 0xCA508310
            * //  0000:SEGA SEGASATURN 
            * //  0010:SEGA ENTERPRISES
            * //  0020:GS-9116   V1.000
            * //  0030:19960826CD-1/1  
            * //  0040:JTUBKAEL        
            * //  0050:JS              
            * //  0060:SEGA RALLY CHAMP
            * //  0070:IONSHIP PLUS    
            * //  0080:                
            * //  0090:                
            * //  00A0:                
            * //  00B0:                
            * //  00C0:                
            * 
            * Compatibility memo :
            *  Incompatibility is probably caused by the game trying to access
            *  Netlink registers ? Theses are probably mapped at the same place
            *  of Gamer's Cartridge ROM (CS0), or SD card related registers (CS1).
            *  Disabling vmem probably won't fix anything, so either some "Hide CS0
            *  or CS1" feature is required on cartridge side (require PCB/firmware
            *  modification), or using cheat code in order to patch Nelink registers
            *  to a safe (unused) location.
            *
            * TODO : Check behavior with normal Action Replay (or USB dev cart) : if
            * game can boot, then it is definitely trying to access something on CS1.
            */
            0x27B906B3, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        //| AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
#endif // 0

#if 0 // 2019/05/06 Reported working with CWX loader without the need of patching
    {
        {
           /*
            * Shinobi-X
            * Version: V1.000, Date: 19950803, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Euro version freezes in main menu, and US version
            * shows black screen after SEGA logo. 
            */
            0x2BA9136D, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
#endif // 0

    {
        {
           /*
            * Skeleton Warriors
            * Version: V1.001, Date: 19960206, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable.  
            * 
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            * 
            * Compatibility memo :
            *  It seems that this game re-authenticates itself, 
            *  probably making it harder to boot with loader other
            *  than JHL.
            * 
            * October 2019 update : added CRC for PAL version of the game.
            *  |ip_crc :FBB8C9B9 CRC[FBB8C9B9]
            *  |hwr_id :SEGA SEGASATURN
            *  |man_id :SEGA TP T-70
            *  |prd_nb :T-7018H-50
            *  |version:V1.000
            *  |date   :19960522
            *  |dev_id :CD-1/1
            *  |area   :E
            *  |pers   :J
            *  |title  :SKELETON WARRIORS
            */
            0x7B1E1E28, 0xFBB8C9B9, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
    {
        {
           /*
            * Tunnel B1
            * Version: V1.001, Date: 19961216, Country: USA
            * Report files:#1, #2, #3
            * 
            * IP Header informations :
            * CRC : 0x8161C8C1
            *  0000:SEGA SEGASATURN 
            *  0010:SEGA TP T-81    
            *  0020:T-8144H   V1.001
            *  0030:19961216CD-1/1  
            *  0040:U               
            *  0050:J               
            *  0060:TUNNEL B1       
            *  0070:                
            *  0080:                
            *  0090:                
            *  00A0:                
            *  00B0:                
            *  00C0:                
            *
            * And PAL version too :
            * |ip_crc :0B0C3592 CRC[0B0C3592]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-164
            * |prd_nb :T-16414H50
            * |version:V1.000
            * |date   :19961107
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :TUNNEL B1
            *
            * Compatibility memo :
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0x8161C8C1, 0x0B0C3592, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * Warcraft II - The Dark Saga
            * Version: V1.000, Date: 19970722, Country: EUR
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [WarCraft II - The Dark Saga (Japan).img]
            * |ip_crc :DF9A4CCA CRC[DF9A4CCA]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-106
            * |prd_nb :T-10623G
            * |version:V1.000
            * |date   :19970927
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :WARCRAFT II
            */
            0xEB56E2F8/*EUR*/, 0xD8D4F974/*USA*/, 0xDF9A4CCA/*JPN*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * WipEout
            * Version: V1.002, Date: 19960411, Country: USA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * |ip_crc :31D8B2E7 CRC[31D8B2E7]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-186
            * |prd_nb :T-18601H
            * |version:V1.002
            * |date   :19960411
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :JS
            * |title  :WIPEOUT
            * 
            * ------------------------------------------
            * [WipEout (Europe).img]
            * |ip_crc :3E2474F4 CRC[3E2474F4]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-113
            * |prd_nb :T-11301H
            * |version:V3.700
            * |date   :19960304
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JS
            * |title  :WIPEOUT
            * 
            * ------------------------------------------
            * [WipEout (Japan) (2M).img]
            * |ip_crc :B8A87BC1 CRC[B8A87BC1]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-186
            * |prd_nb :T-18603G
            * |version:V3.900
            * |date   :19960313
            * |dev_id :CD-1/1
            * |area   :JT
            * |pers   :JS
            * |title  :WIPEOUT
            */
            0x31D8B2E7/*USA*/, 0x3E2474F4/*EUR*/, 0xB8A87BC1/*JPN*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

#if 0 // 2019/05/26 Reported working with CWX loader after patching
    {
        {
           /*
            * Genso Suikoken
            * Version: V1.006, Date: 19980701, Country: JPN
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo Saturn
            * patcher utility in order to make it playable. 
            * 
            * Compatibility memo :
            *  Works with patch on JHL loader only - opening FMV will only
            *  play between the first 1 to 10 seconds before glitching, but
            *  can still be skipped with Start button, even after glitching
            *  out. Will not run without patch.
            */
            0x7A778836/*JPN*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
#endif // 0

    {
        {
           /*
            * C4-2007 package
            * The Rockin'-B All Stars
            * 
            * |ip_crc :F47FBDF3 CRC[F47FBDF3]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP ROCKIN-B
            * |prd_nb :T-060680
            * |version:V1.000
            * |date   :19941124
            * |dev_id :CD-1/1
            * |area   :JTUE
            * |pers   :J
            * |title  :Released by The Rockin'-B for SEGA Saturn. Go towww.rockin-b.de for further SEGAconsole dev stuff.
            * 
            * The Rockin'-B All Stars\Atlas.iso
            * |ip_crc :F9F1CA64 CRC[F9F1CA64]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA THEROCKIN-B
            * |prd_nb :MINESWPR
            * |version:V1.000
            * |date   :20030831
            * |dev_id :CD-1/1
            * |area   :JTUE
            * |pers   :J
            * |title  :Rockin-Bs MineSweeper for SEGA Saturn/segaXtreme coding contest 2003
            * 
            * C4_2007_CD\1_dont_burn_this_file.iso
            * |ip_crc :F47FBDF3 CRC[F47FBDF3]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP ROCKIN-B
            * |prd_nb :T-060680
            * |version:V1.000
            * |date   :19941124
            * |dev_id :CD-1/1
            * |area   :JTUE
            * |pers   :J
            * |title  :Released by The Rockin'-B for SEGA Saturn. Go towww.rockin-b.de for further SEGAconsole dev stuff.
            * 
            * Comment:
            *  There's no need to pseudo pach the image, but JHL loader seems
            *  required to correctly load it.
            *  Unfortunately, nearly all productions from Rockin-'B use the same
            *  IP header, making all of them JHL loader dependent.
            *  As a consequence, theses won't be playable on some grey Japanese
            *  Saturn models.
            * 
            * Compatibility memo :
            *  It was reported not working when using a Phoebe ODE, as
            *  described below :
            *  This application represents quite a specific edge-case, 
            *  because it also doesn't load through Rmenu and even fails
            *  to load if you place it in folder 01 on the Phoebe SD card.
            *  Note that this application requires the JHL loader. If
            *  you use the CWX loader it appears to be compatible at first
            *  but, when you attempt to load one of the games or apps, 
            *  the Saturn resets to the multiplayer screen.
            *
            *  Game is not compatible with CWX loader, and usage of
            *  JHL loader with pseudo patched game is mandatory.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0xF47FBDF3, 0xF9F1CA64, 0xF47FBDF, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 


   /*
    * Certain titles are incompatible with vmem.
    * Unless stated otherwise, they all crash to a black screen after the SEGA logo when vmem is enabled:
    *
    * Not compatible with CWX loader or vmem :
    *  - Sega Flash Vol.7 (Europe)
    *  - Flash Sega Saturn Vol.24 (Japan)
    *  - Flash Sega Saturn Vol.25 (Japan)
    *  - Flash Sega Saturn Vol.26 (Japan)
    *  - Flash Sega Saturn Vol.27 (Japan)
    *  - Flash Sega Saturn Vol.28 (Japan)
    *  - Flash Sega Saturn Vol.29 (Japan)
    *  - Flash Sega Saturn Vol.30 (Japan)
    *  - Flash Sega Saturn Vol.31 (Japan)
    *  - Flash Sega Saturn Vol.32 (Japan)
    *  - Mogitate SegaSaturn: Special Movie (Japan)
    *
    * Compatible with CWX loader :
    *  - Arcade Gears 3: Wonder 3 (Japan)
    *  - Manx TT Superbike (Europe) - The games boots up successfully but if you don't press start on the title screen, it will freeze on a white screen when it tries to play the opening cinema. It will also freeze when attempting to play a race, after the 'GET READY TO RACE' screen has appeared.
    *  - Manx TT Superbike (Japan)
    *  - Manx TT Super Bike (USA)
    *  - Sega Ages Memorial Selection Vol.1 (Japan)  - This sometimes crashes to a black screen with vmem enabled.
    *  - The House of the Dead (Europe)
    *  - The House of the Dead (Japan | Satakore release)
    *  - The House of the Dead (Japan | Standard release)
    *  - The House of the Dead (USA)
    *  - Touge King the Spirits (Japan)
    *  - Tryrush Deppy (Japan) - When vmem is enabled and an RGB connection is used, it crashes to a black screen when exiting the Options menu.
    *  - Virtua Cop (Europe)
    *  - Virtua Cop (Japan)
    *  - Virtua Cop (USA)
    *  - Wipeout 2097 (Europe)
    *  - Wipeout XL (Japan) 
    *
    * ------------------
    * (Games not compatible with vmem, added 2019/04/17)
    *  - Linkle Liver Story - When prompted to select a save destination, selecting the Cartridge RAM option will cause the game to crash. vmem via USB sometimes works however but it's not super stable.
    *  - Ninpen Manmaru - Attempting to load a game and selecting the Cartridge RAM option will make the green light on the Gamer's Cartridge will blink every 0.5~ seconds and controller input is hardly recognised, which makes it almost impossible to load an existing save file 
    *  - Resident Evil (USA & Europe) / Biohazard (Japan) - Attempting to load a game from either internal or external memory will just take back to the intro cinema. When selecting the Load Game option from the main menu, the game will also crash sometimes. Save files can be loaded with vmem disabled.
    *
    * ------------------
    * Games added 2019/12/09 : NHL Powerplay '96 and Virtua Fighter :
    *  - Virtua Fighter requires JHL loader.
    *  - NHL Powerplay '96 requires pseudo patching, is incompatible with vmem, and has more consistent performance with CWX loader.
    *     Boot from burned CD   : 80% boot chance with CWX loader vs ~100% with JHL loader
    *     Boot from Rhea/Phoebe : 80% boot chance with CWX loader vs ~50% with JHL loader
    *
    *
    *
    * |ip_crc :FB54E944 CRC[FB54E944]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :6106288G
    * |version:V1.000
    * |date   :19980320
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :J
    * |title  :FLASH SEGASATURNPAL VOL.7
    * 
    * |ip_crc :B1F62D63 CRC[B1F62D63]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616624
    * |version:V1.000
    * |date   :19971119
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.24
    * 
    * |ip_crc :F4163244 CRC[F4163244]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616625
    * |version:V1.001
    * |date   :19980109
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.25
    * 
    * |ip_crc :54DBDD67 CRC[54DBDD67]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616626
    * |version:V1.000
    * |date   :19980203
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.26
    * 
    * |ip_crc :D8241254 CRC[D8241254]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616627
    * |version:V1.001
    * |date   :19980305
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.27
    * 
    * |ip_crc :3EB474A9 CRC[3EB474A9]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616628
    * |version:V1.003
    * |date   :19980403
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.28
    * 
    * |ip_crc :0CD0C6D9 CRC[0CD0C6D9]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616629
    * |version:V1.000
    * |date   :19980421
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.29
    * 
    * |ip_crc :242B071B CRC[242B071B]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616630
    * |version:V1.001
    * |date   :19980527
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.30
    * 
    * |ip_crc :5753D245 CRC[5753D245]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616631
    * |version:V1.002
    * |date   :19980626
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.31
    * 
    * |ip_crc :DB3D8B92 CRC[DB3D8B92]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :610616632
    * |version:V1.001
    * |date   :19980721
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :FLASH SEGASATURNVOL.32
    * 
    * |ip_crc :6D56A3BD CRC[6D56A3BD]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-261
    * |prd_nb :T-26107G
    * |version:V1.002
    * |date   :19980123
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :WONDER3
    * 
    * |ip_crc :58053AA2 CRC[58053AA2]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81210
    * |version:V1.000
    * |date   :19970217
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :JSE
    * |title  :MANX TT SUPER BIKE
    * 
    * |ip_crc :93B2D7F0 CRC[93B2D7F0]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9102
    * |version:V1.001
    * |date   :19970218
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :JSE
    * |title  :MANX TT SUPER BIKE
    * 
    * |ip_crc :7BE0A7E7 CRC[7BE0A7E7]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81210
    * |version:V1.001
    * |date   :19970423
    * |dev_id :CD-1/1
    * |area   :UJT
    * |pers   :JSE
    * |title  :MANX TT SUPER BIKE
    * 
    * |ip_crc :3B9A53E3 CRC[3B9A53E3]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :6106929
    * |version:V1.000
    * |date   :19980422
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :J
    * |title  :MOGITATE        SEGASATURN      DEEP FEAR &     SHADOWS OF THE  TUSK            SPECIAL MOVIE
    * 
    * |ip_crc :FF7A8D97 CRC[FF7A8D97]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9135
    * |version:V1.002
    * |date   :19970126
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :J
    * |title  :MEMORIALSELECTION VOL1
    * 
    * |ip_crc :4D41D9EB CRC[4D41D9EB]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81802
    * |version:V1.001
    * |date   :19980311
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :JEG
    * |title  :THE HOUSE OF THE DEAD
    * 
    * |ip_crc :29B5E9A3 CRC[29B5E9A3]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9173
    * |version:V1.003
    * |date   :19980225
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :JEG
    * |title  :THE HOUSE OF THE DEAD
    * 
    * |ip_crc :7B667E07 CRC[7B667E07]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9173
    * |version:V1.003
    * |date   :19980521
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :JEG
    * |title  :THE HOUSE OF THE DEAD
    * 
    * |ip_crc :EFD118D1 CRC[EFD118D1]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81802
    * |version:V1.000
    * |date   :19980311
    * |dev_id :CD-1/1
    * |area   :U
    * |pers   :JEG
    * |title  :THE HOUSE OF THE DEAD
    * 
    * |ip_crc :60C91334 CRC[60C91334]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-144
    * |prd_nb :T-14401G
    * |version:V1.003
    * |date   :19951010
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :JS
    * |title  :Touge KING The Spirits
    * 
    * |ip_crc :5CEFCC3F CRC[5CEFCC3F]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-213
    * |prd_nb :T-21302G
    * |version:V1.000
    * |date   :19961010
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :TRYRUSH DEPPY
    * 
    * |ip_crc :22EE51C5 CRC[22EE51C5]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81015
    * |version:V1.004
    * |date   :19951108
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :JMG
    * |title  :VIRTUA COP
    * 
    * |ip_crc :CD8D1C5F CRC[CD8D1C5F]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9060
    * |version:V1.000
    * |date   :19951012
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :JMG
    * |title  :VIRTUA COP
    * 
    * |ip_crc :9C5BFF7D CRC[9C5BFF7D]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81015
    * |version:V1.001
    * |date   :19951017
    * |dev_id :CD-1/1
    * |area   :TU
    * |pers   :JMG
    * |title  :VIRTUA COP
    * 
    * |ip_crc :43D8E7D1 CRC[43D8E7D1]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-113
    * |prd_nb :T-11308H-5
    * |version:0 V1.0
    * |date   :19970704
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :JSE
    * |title  :WIPEOUT-2
    * 
    * |ip_crc :3D6ECC45 CRC[3D6ECC45]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-186
    * |prd_nb :T-18619G
    * |version:V1.000
    * |date   :19970922
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :JSE
    * |title  :WIPEOUT-2
    * 
    * ------------------
    * (Games not compatible with vmem, added 2019/04/17)
    * |ip_crc :79B9FC6C CRC[79B9FC6C]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9055 V1
    * |version:.05
    * |date   :19960220
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :J
    * |title  :LINKLE LIVER STORY
    * 
    * |ip_crc :60096B86 CRC[60096B86]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-355
    * |prd_nb :T-35502G
    * |version:V1.001
    * |date   :19971107
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :NINPEN MANMARU
    * 
    * |ip_crc :1E85D941 CRC[1E85D941]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-12
    * |prd_nb :T-1219G
    * |version:V1.002
    * |date   :19970624
    * |dev_id :CD-1/1
    * |area   :J
    * |pers   :J
    * |title  :BIO HAZARD
    * 
    * |ip_crc :14C84A4F CRC[14C84A4F]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81092
    * |version:V1.002
    * |date   :19970806
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :J
    * |title  :RESIDENT EVIL
    * 
    * |ip_crc :8595D448 CRC[8595D448]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-12
    * |prd_nb :T-1221H
    * |version:V1.002
    * |date   :19970804
    * |dev_id :CD-1/1
    * |area   :U
    * |pers   :J
    * |title  :RESIDENT EVIL
    * 
    * ------------------
    * (Games not compatible with vmem, added 2019/12/09)
    * 
    * |ip_crc :FD121C99 CRC[FD121C99]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-70
    * |prd_nb :T-7013H-50
    * |version:V4.000
    * |date   :19961121
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :JT
    * |title  :NHL POWERPLAY '96
    * 
    * |ip_crc :79102D35 CRC[79102D35]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA TP T-70
    * |prd_nb :T-7012G
    * |version:V4.000
    * |date   :19961126
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :JT
    * |title  :NHL POWERPLAY '96
    * 
    * ------------------
    * Games not compatible with CWX loader, added 2019/12/09
    * 
    * |ip_crc :5FD72A91 CRC[5FD72A91]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK_8100550
    * |version:V1.000
    * |date   :19950425
    * |dev_id :CD-1/1
    * |area   :E
    * |pers   :J
    * |title  :VIRTUA FIGHTER
    * 
    * |ip_crc :E6209E7E CRC[E6209E7E]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :GS-9001
    * |version:V1.000
    * |date   :19941005
    * |dev_id :CD-1/1
    * |area   :JT
    * |pers   :J
    * |title  :VIRTUA FIGHTER
    * 
    * |ip_crc :B33598FC CRC[B33598FC]
    * |hwr_id :SEGA SEGASATURN
    * |man_id :SEGA ENTERPRISES
    * |prd_nb :MK-81005
    * |version:V1.000
    * |date   :19950317
    * |dev_id :CD-1/1
    * |area   :BKUT
    * |pers   :J
    * |title  :VIRTUA FIGHTER
    */
    {
        {
            /* Flash Sega Saturn entries until volume 29. */
            0xFB54E944, 0xB1F62D63, 0xF4163244, 0x54DBDD67, 0xD8241254, 0x3EB474A9, 0x0CD0C6D9
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
    }, 
    {
        {
            /* Remaining of the Flash Sega Saturn entries. */
            0x242B071B, 0x5753D245, 0xDB3D8B92, 
            /* Mogitate SegaSaturn: Special Movie (Japan) */
            0x3B9A53E3, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
    }, 
    {
        {
            /* Manx TT Superbike. */
            0x58053AA2, 0x93B2D7F0, 0x7BE0A7E7, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
    }, 
    {
        {
           /*  - Sega Ages Memorial Selection Vol.1 (Japan)  - This sometimes crashes to a black screen with vmem enabled.
            *  - The House of the Dead (Europe)
            *  - The House of the Dead (Japan | Satakore release)
            *  - The House of the Dead (Japan | Standard release)
            *  - The House of the Dead (USA)
            *  - Touge King the Spirits (Japan)
            */
            0xFF7A8D97, 0x4D41D9EB, 0x29B5E9A3, 0x7B667E07, 0xEFD118D1, 0x60C91334, 
            AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
    }, 
    {
        {
           /*  - Tryrush Deppy (Japan) - When vmem is enabled and an RGB connection is used, it crashes to a black screen when exiting the Options menu.
            *  - Virtua Cop (Europe)
            *  - Virtua Cop (Japan)
            *  - Virtua Cop (USA)
            *  - Wipeout 2097 (Europe)
            *  - Wipeout XL (Japan) 
            */
            0x5CEFCC3F, 0x22EE51C5, 0xCD8D1C5F, 0x9C5BFF7D, 0x43D8E7D1, 0x3D6ECC45, 
            AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
    }, 
    {
        {
           /* ------------------
            * (Games not compatible with vmem, added 2019/04/17)
            *  - Linkle Liver Story
            *  - Ninpen Manmaru
            *  - Resident Evil (USA & Europe) / Biohazard (Japan)
            */
            0x79B9FC6C, 0x60096B86, 0x1E85D941, 0x14C84A4F, 0x8595D448, 

           /* ------------------
            * (Game not compatible with vmem, added 2019/10/22)
            *  - Mizubaku Daibouken
            *    https://segaxtreme.net/threads/pseudo-saturn-kai-official-thread.24037/page-18
            *     |I have noticed, that in Mizubaku Daibouken there are
            *     |occasional graphical glitches if I start the game through
            *     |pseudo saturn kai. I have a phoebe installed in my saturn
            *     |and if I start the same image through the regual rmenu, 
            *     |there are no glitches. They don't seem to be random, as
            *     |they appear in specific spots as for example at the end
            *     |of the 2nd stage, as seen in the picture below.
            *     |I'm using a gamer cart with version 6.314 installed.
            *    -> This may be caused by the game writing to memory
            *       reserved to backup library ?
            *       Whatever it may be, this game uses very small
            *       save data memory, hence making it being supported by 
            *       vmem is not worth the hassle.
            *
            *    [Mizubaku Daibouken (Japan) (Track 01).bin]
            *     |ip_crc :2F789C88 CRC[2F789C88]
            *     |hwr_id :SEGA SEGASATURN
            *     |man_id :SEGA TP T-199
            *     |prd_nb :T-19910G
            *     |version:V1.001
            *     |date   :19980910
            *     |dev_id :CD-1/1
            *     |area   :J
            *     |pers   :J
            *     |title  :MIZUBAKU DAIBOUKEN
            */
            0x2F789C88, 
            AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
    }, 
    {
        {
           /*
            * Games requiring JHL loader, added 2019/05/26.
            *
            * [3D Mission Shooting - Finalist (Japan) (Track 01).bin]
            * |ip_crc :33B0CD74 CRC[33B0CD74]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-185
            * |prd_nb :T-18511G
            * |version:V1.000
            * |date   :19970619
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :FINALIST
            * 
            * [MechWarrior 2 - 31st Century Combat - Arcade Combat Edition (Europe) (Track 01).bin]
            * |ip_crc :C31F3560 CRC[C31F3560]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-130
            * |prd_nb :T-13004H
            * |version:V1.000
            * |date   :19970519
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JAE
            * |title  :MECHWARRIOR2
            * 
            * [MechWarrior 2 (Japan) (Track 01).bin]
            * |ip_crc :48B0B18E CRC[48B0B18E]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-234
            * |prd_nb :T-23406G
            * |version:V1.002
            * |date   :19970725
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :MECHWARRIOR2
            * 
            * [Mortal Kombat II (Europe) (Track 1).bin]
            * |ip_crc :D7B1F205 CRC[D7B1F205]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8103H-50
            * |version:V1.000
            * |date   :19960102
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :MORTAL KOMBAT II
            * 
            * [Mortal Kombat II Kanzenban (Japan) (4S) (Track 1).bin]
            * |ip_crc :389A4B17 CRC[389A4B17]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8107G
            * |version:V1.000
            * |date   :19960214
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :J
            * |title  :MORTAL KOMBAT II
            * 
            * Compatibility memo :
            *  Games are not compatible with CWX loader, and usage of
            *  JHL loader is mandatory. Theses games require Pseudo patching.
            *  Game is marked as compatible with virtual memory cart, which is
            *  currently not working with JHL loader : it is set for eventual
            *  future revision supporting virtual memory cartridge on JHL loader.
            */
            0x33B0CD74, 0xC31F3560, 0x48B0B18E, 0xD7B1F205, 0x389A4B17, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 


    {
        {
           /*
            * Game requiring JHL loader, added 2019/09/23.
            * 
            * [Pebble Beach Golf Links (Europe) (3S) (Track 1).bin]
            * |ip_crc :428B2074 CRC[428B2074]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA ENTERPRISES
            * |prd_nb :MK-81101
            * |version:V.001
            * |date   :19950427
            * |dev_id :CD-1/1
            * |area   :AEL
            * |pers   :J
            * |title  :Pebble Beach Golf Links
            * 
            * [Pebble Beach Golf Links (USA) (Track 1).bin]
            * |ip_crc :DC0E8BF7 CRC[DC0E8BF7]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA ENTERPRISES
            * |prd_nb :MK-81101
            * |version:V.001
            * |date   :19950407
            * |dev_id :CD-1/1
            * |area   :UKBT
            * |pers   :J
            * |title  :Pebble Beach Golf Links
            * 
            * [Pebble Beach Golf Links - Stadler ni Chousen (Japan) (Track 1).bin]
            * |ip_crc :7325B39C CRC[7325B39C]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA ENTERPRISES
            * |prd_nb :GS-9006
            * |version:V1.000
            * |date   :19941227
            * |dev_id :CD-1/1
            * |area   :JT
            * |pers   :J
            * |title  :Pebble Beach Golf Links
            */
            0x428B2074, 0xDC0E8BF7, 0x7325B39C, 

           /*
            * Game requiring JHL loader, added 2020/07/15.
            * 
            * ------------------------------------------
            * [Bubble Bobble also featuring Rainbow Islands (Europe) (Track 1).bin]
            * |ip_crc :6458C58E CRC[6458C58E]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8131H-50
            * |version:V1.000
            * |date   :19960812
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :BUBBLE BOBBLE
            * 
            * ------------------------------------------
            * [Bubble Bobble also featuring Rainbow Islands (USA) (Track 1).bin]
            * |ip_crc :57DADE02 CRC[57DADE02]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8131H
            * |version:V1.000
            * |date   :19960812
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :J
            * |title  :BUBBLE BOBBLE
            */
            0x6458C58E, 0x57DADE02, 

           /*
            * NHL Powerplay '96 (USA) 
            * 2020/08/29 : Reported only working with JHL loader in RmenuKai
            * 
            * [NHL Powerplay '96 (USA) (Track 1).bin]
            * |ip_crc :5F223A49 CRC[5F223A49]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-070
            * |prd_nb :T-07013H
            * |version:V1.000
            * |date   :19960605
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :JAT
            * |title  :NHL Power Play '96
            */
            0x5F223A49, 
            AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 
    {
        {
            /* NHL Powerplay '96 : incompatible with vmem and overall works best with CWX. */
            0xFD121C99, 
            0x79102D35, 
            //0x5F223A49, // <- USA version requires JHL loader

            /* Dark Savior : reported incompatible with vmem (2020/07/20).
             * CRC : 0xE747C8CC
             *  0000:SEGA SEGASATURN 
             *  0010:SEGA ENTERPRISES
             *  0020:MK-81304  V1.000
             *  0030:19961119CD-1/1  
             *  0040:JTU             
             *  0050:JAE             
             *  0060:DARK SAVIOR     
             */
            0xE747C8CC, // USA
            0x13C9CEB0, // EUR
            0x5DC5DF09, // JPN (v0.5)
            0xADE59255, // JPN (v1.0)
            AUTOSET_DUMMY_CRC
        }, 
        AUTOSET_CWX_LDR_COMPATIBLE | AUTOSET_JHL_LDR_COMPATIBLE
        /*| AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE*/
    }, 
    {
        {
            /* Virtua Fighter : require JHL loader.
             * (Vmem compatibility is indicated, but it can't be used
             * as JHL loader isn't supporting it)
             */
            0x5FD72A91, 
            0xE6209E7E, 
            0xB33598FC, 

            /* Herc's Adventures (USA) : require JHL loader.
             * (with the CWX loader, it crashes to a black screen after the SEGA logo)
             *
             * [Herc's Adventures (USA) (RE) (Track 01).bin]
             * |ip_crc :4AB9BAE5 CRC[4AB9BAE5]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-230
             * |prd_nb :T-23001H
             * |version:V1.002
             * |date   :19970530
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :J
             * |title  :HERC'S ADVENTURES
             */
            0x4AB9BAE5, 

            /* Spot Goes To Hollywood (EUR/JPN/USA) : all require JHL loader.
             *
             * [Spot Goes to Hollywood (Europe) (Track 01).bin]
             * |ip_crc :4853EE10 CRC[4853EE10]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-70
             * |prd_nb :T-7001H-50
             * |version:V1.400
             * |date   :19961206
             * |dev_id :CD-1/1
             * |area   :E
             * |pers   :J
             * |title  :SPOT GOES TO HOLLYWOOD
             * 
             * [Spot Goes to Hollywood (Japan) (Track 01).bin]
             * |ip_crc :E37FA858 CRC[E37FA858]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-70
             * |prd_nb :T-7014G
             * |version:V1.000
             * |date   :19961107
             * |dev_id :CD-1/1
             * |area   :JT
             * |pers   :J
             * |title  :SPOT GOES TO HOLLYWOOD
             * 
             * [Spot Goes to Hollywood (USA) (Track 01).bin]
             * |ip_crc :1745BC7A CRC[1745BC7A]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA TP T-70
             * |prd_nb :T-7001H
             * |version:V1.000
             * |date   :19961211
             * |dev_id :CD-1/1
             * |area   :U
             * |pers   :J
             * |title  :SPOT GOES TO HOLLYWOOD
             */
            0x4853EE10, 0xE37FA858, 0x1745BC7A
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * 2020/08/29 : following games only work with JHL loader in RmenuKai :
            *  - FIFA Soccer 96 (Europe)
            *  - FIFA Soccer 96 (USA)   (same IP CRC as Europe)
            *  
            *  - FIFA 97 (Europe)
            *  - FIFA 97 (France)       (same IP CRC as Europe)
            *  - FIFA 97 (Germany)      (same IP CRC as Europe)
            *  - FIFA Soccer 97 (USA)
            *
            * Comment: Games need to have their image files patched with
            * Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * [FIFA Soccer 96 (USA).img]
            * |ip_crc :1A596C8C CRC[1A596C8C]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-500
            * |prd_nb :T-5003H
            * |version:V1.000
            * |date   :19951201
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :J
            * |title  :FIFA SOCCER '96
            * 
            * [FIFA Soccer 97 (USA).img]
            * |ip_crc :EEA9C3ED CRC[EEA9C3ED]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-50
            * |prd_nb :T-5017H
            * |version:V1.001
            * |date   :19970228
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :J
            * |title  :FIFA 97
            * 
            * ------------------------------------------
            * [FIFA 97 (Europe) (En,Fr,De,Es,It,Sv).img]
            * |ip_crc :A06E5DC4 CRC[A06E5DC4]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP 69
            * |prd_nb :T-5017H
            * |version:V1.000
            * |date   :19970216
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :J
            * |title  :FIFA 97         
            */
            0x1A596C8C, 0xEEA9C3ED, 0xA06E5DC4,

           /*
            * FIFA Rumbo al Mundial 98
            * Version: V1.000, Date: 19971128, Country: EUR SPA
            * Remark: test performed with Pseudo Saturn Kai lite.
            * Comment: Game needs to have its image file patched with Pseudo
            * Saturn patcher utility in order to make it playable. 
            * 
            * 2019/05/26 : Reported working with CWX loader after patching
            * 2020/08/29 : following versions are incompatible
            *              with CWX loader in RmenuKai :
            *  - FIFA - Die WM-Qualifikation 98 (Germany)  (same IP CRC as France)
            *  - FIFA - En Route pour la Coupe du Monde 98 (France)
            *  - FIFA - Road to World Cup 98 (Europe)      (same IP CRC as France)
            *  - FIFA - Road to World Cup 98 (Italy)       (same IP CRC as France)
            *  - FIFA - Rumbo al Mundial 98 (Spain)        (same IP CRC as France)
            *  - FIFA - Road to World Cup 98 (USA)
            * 
            * [FIFA - Road to World Cup 98 (USA).img]
            * |ip_crc :77139D92 CRC[77139D92]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP 69
            * |prd_nb :T-5025H
            * |version:V1.000
            * |date   :19971128
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :JT
            * |title  :FIFA - ROAD TO  WORLD CUP 98
            */
            0x4491861E/*EUR*/, 0x77139D92/*USA*/, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * Impact Racing
            * 2020/08/29 : following games only work with JHL loader :
            *  - Impact Racing (Europe)
            *  - Impact Racing (Europe Demo) (same IP CRC as Europe) 
            *  - Impact Racing (Japan)
            *  - Impact Racing (USA)
            * 
            * ------------------------------------------
            * [Impact Racing (Europe).img]
            * |ip_crc :03B91FE6 CRC[03B91FE6]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-60
            * |prd_nb :T-6010H-50
            * |version:V1.200
            * |date   :19961004
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JA
            * |title  :IMPACT RACING
            * 
            * ------------------------------------------
            * [Impact Racing (Japan).img]
            * |ip_crc :2F403E6E CRC[2F403E6E]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-73
            * |prd_nb :T-7307G
            * |version:V1.002
            * |date   :19970416
            * |dev_id :CD-1/1
            * |area   :J
            * |pers   :JA
            * |title  :IMPACT RACING
            * 
            * ------------------------------------------
            * [Impact Racing (USA).img]
            * |ip_crc :463710C1 CRC[463710C1]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-81
            * |prd_nb :T-8139H
            * |version:V1.000
            * |date   :19960801
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :JA
            * |title  :IMPACT RACING
            */
            0x03B91FE6, 0x2F403E6E, 0x463710C1,
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        {
           /*
            * NHL 98
            * 2020/08/29 : following games only work with JHL loader in RmenuKai :
            * - NHL 98 (Europe) 
            * - NHL 98 (USA) 
            * 
            * Comment: Games need to have their image files patched with
            * Pseudo Saturn patcher utility in order to make it playable. 
            * 
            * [NHL 98 (Europe).img]
            * |ip_crc :E56EF297 CRC[E56EF297]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-69
            * |prd_nb :T-5026H-50
            * |version:V1.000
            * |date   :19971215
            * |dev_id :CD-1/1
            * |area   :E
            * |pers   :JT
            * |title  :NHL 98
            * 
            * ------------------------------------------
            * [NHL 98 (USA).img]
            * |ip_crc :D6ECE91B CRC[D6ECE91B]
            * |hwr_id :SEGA SEGASATURN
            * |man_id :SEGA TP T-69
            * |prd_nb :T-5026H
            * |version:V1.000
            * |date   :19971215
            * |dev_id :CD-1/1
            * |area   :U
            * |pers   :JT
            * |title  :NHL 98
            */
            0xE56EF297, 0xD6ECE91B, 
            AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC
        }, 
        /*AUTOSET_CWX_LDR_COMPATIBLE | */AUTOSET_JHL_LDR_COMPATIBLE
        | AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE
    }, 

    {
        /*
         * Terminating entry with dummy loader settings.
         */
        {AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC, AUTOSET_DUMMY_CRC}, 0
    }
};


#ifdef SC_SATURN
/* Current loader settings. */
loader_settings_t* _loader_settings = &_autoset_default_settings;


void autoset_init(void)
{
    /* Set loader settings to default. */
    _loader_settings = &_autoset_default_settings;
}


void autoset_search(ip_bin_t* ip)
{
    unsigned long crc = ip_header_get_crc(ip);

    /* Initialize loader settings to default. */
    _loader_settings = &_autoset_default_settings;

    /* Look for game in custom loader settings table. */
    loader_settings_t* ret = _autoset_custom_tbl;
    int j;
    int found = 0;

    while(ret->crc[0] != AUTOSET_DUMMY_CRC)
    {
        for(j=0; j<AUTOSET_CRC_CNT; j++)
        {
            if(ret->crc[j] == AUTOSET_DUMMY_CRC)
            {
                break;
            }
            if(ret->crc[j] == crc)
            {
                _loader_settings = ret;
                found = 1;
                break;
            }
        }

        if(found)
        {
            break;
        }

        ret++;
    } // settings loop

    /* Search if settings are registered into custom loader ones. */
    unsigned long cl_ret = cl_ip_crc_search(crc);
    if(CL_SEARCH_FOUND(cl_ret))
    {
        /* Patch automatic settings with informations from custom loader. */
        memcpy(&_autoset_custom_settings, _loader_settings, sizeof(loader_settings_t));
        _loader_settings = &_autoset_custom_settings;

        if(CL_SEARCH_CWX_USE(cl_ret))
        {
            _loader_settings->set |= AUTOSET_CWX_LDR_COMPATIBLE;
        }
        else
        {
            _loader_settings->set &= ~(AUTOSET_CWX_LDR_COMPATIBLE);
            _loader_settings->set |= AUTOSET_JHL_LDR_COMPATIBLE;
        }

        if(CL_SEARCH_VMEM_USE(cl_ret))
        {
            _loader_settings->set |= AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE;
        }
        else
        {
            _loader_settings->set &= ~(AUTOSET_VCART_COMPATIBLE | AUTOSET_VFDD_COMPATIBLE);
        }
    }
}


unsigned long autoset_get_info(void)
{
    return _loader_settings->set;
}

#endif // SC_SATURN
