/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef AUTOSET_H
#define AUTOSET_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ip_header.h"


/**
 * Definitions for game loader, vmem, etc automatic settings.
**/

/*
 * Definition of suitable settings for current game.
 *
 *  - bit 0 : set when game is compatible with CWX loader
 *  - bit 1 : set when game is compatible with JHL loader
 *  - bit 2 : require pseudo patching (usually set when only JHL loader can be used)
 *  - bit 3 : reserved
 *  - bit 4 : set when game is compatible with virtual memory cartridge
 *  - bit 5 : set when game is compatible with virtual FDD
 *  - bit 6 : set when vFDD/FDD selection menu is required (when not set, vFDD is preferred)
 *  - bit 7 : set when game uses floppy disk
 *  - bit 8 : set when game optionally accepts 1MB memory expansion cartridge (= can work without)
 *  - bit 9 : set when game requires 1MB memory expansion cartridge (= can't work without)
 *  - bit10 : set when game optionally accepts 4MB memory expansion cartridge (= can work without)
 *  - bit11 : set when game requires 4MB memory expansion cartridge (= can't work without)
 *  - other : reserved
 *
 * "JHL loader compatible" bit is usually always set, but automatically
 * selected only when game isn't compatible with CWX loader.
 * Why ? Because JHL loader doesn't works at all on some few models
 * of Saturn (example : my grey Japanese test unit).
 *
 * "Virtual memory cart compatible" and "Virtual FDD compatible" are
 * always set together, because (AFAIK) games not using FDD work
 * fine even if a FDD is connected to Saturn.
 * Separating both virtual memory cart and FDD in two bits is just
 * "in case of" there are games behaving incorrectly when FDD is used.
 *
 * Expansion RAM compatibility is separated into "optional" and "mandatory", 
 * so that it may be possible to show some error message before loading a
 * mandatory game while expansion memory isn't detected.
 * Optional bit is here in order to enable expansion RAM without having to
 * beg input or confirmation from user.
 */
#define AUTOSET_CWX_LDR_COMPATIBLE  (1 <<  0)
#define AUTOSET_JHL_LDR_COMPATIBLE  (1 <<  1)
//#define AUTOSET_REQUIRE_PSEUDOPATCH (1 <<  2)

#define AUTOSET_VCART_COMPATIBLE    (1 <<  4)
#define AUTOSET_VFDD_COMPATIBLE     (1 <<  5)

#define AUTOSET_VFDD_SELECTMENU     (1 <<  6)

#define AUTOSET_USE_FLOPPYDISK      (1 <<  7)

#define AUTOSET_1MBRAM_OPTION       (1 <<  8)
#define AUTOSET_1MBRAM_MANDATORY    (1 <<  9)
#define AUTOSET_4MBRAM_OPTION       (1 << 10)
#define AUTOSET_4MBRAM_MANDATORY    (1 << 11)



/**
 * autoset_init
 * 
 * Initialize loader settings to default.
 * Must be called at the beginning of main program.
**/
void autoset_init(void);

/**
 * autoset_search
 * 
 * Look for game loader settings according to game
 * IP header's CRC value.
**/
void autoset_search(ip_bin_t* ip);

/**
 * autoset_get
 * 
 * Loader settings is a bitfield composed with
 * AUTOSET_* values.
**/
unsigned long autoset_get_info(void);



#endif // AUTOSET_H
