/**
 *  Sega Saturn boot stub code for Pseudo Saturn Kai
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h>

#include "cdhack.h"
#include "cdhack_exec.h"


/*
 * iapetus functions required by cd_move_sector_data_cd_auth
 * and jhl_cd_hack functions :
 *  cd_exec_command
 *  cd_get_stat
 *  is_cd_auth
 *  cd_set_sector_size
 *  cd_put_sector_data
 *  cd_end_transfer
 *  cd_reset_selector_all
 *
 * Plus the following sub-functions :
 *  interrupt_set_level_mask
 *  interrupt_get_level_mask
 *  cd_wait_hirq
 *  cd_write_command
 *  cd_read_return_status
 */


void interrupt_set_level_mask(u32 imask)
{
   u32 sr;

   asm("stc sr,%0": "=r"(sr));
   sr &= 0xFFFFFF0F;
   imask <<= 4;
   sr |= imask;
   asm("ldc %0,sr": : "r" (sr));
}

u32 interrupt_get_level_mask(void)
{
   u32 sr;

   asm("stc sr,%0": "=r"(sr));
   sr = (sr & 0xF0) >> 4;

   return sr;
}

/* Called only once in a time critical function, hence inlined. */
//void cd_write_command(cd_cmd_struct *cd_cmd)
//{
//    CDB_REG_CR1 = cd_cmd->CR1;
//    CDB_REG_CR2 = cd_cmd->CR2;
//    CDB_REG_CR3 = cd_cmd->CR3;
//    CDB_REG_CR4 = cd_cmd->CR4;
//}

//void cd_read_return_status(cd_cmd_struct *cd_cmd_rs)
//{
//    cd_cmd_rs->CR1 = CDB_REG_CR1;
//    cd_cmd_rs->CR2 = CDB_REG_CR2;
//    cd_cmd_rs->CR3 = CDB_REG_CR3;
//    cd_cmd_rs->CR4 = CDB_REG_CR4;
//}

int cd_wait_hirq(int flag)
{
    int i;
    u16 hirq_temp;

    for (i = 0; i < 0x240000; i++)
    {
        hirq_temp = CDB_REG_HIRQ;
        if (hirq_temp & flag)
        {
            return TRUE;
        }
    }

    return FALSE;
}

int cd_exec_command(u16 hirq_mask, cd_cmd_struct *cd_cmd, cd_cmd_struct *cd_cmd_rs)
{
//    int old_level_mask;
    u16 hirq_temp;
    u16 cd_status;

    /*
     * iapetus mask interrupts during cd_exec_command only, 
     * but it's safer to do this during whole CD Hack, hence
     * interrupt_*_level_mask functions calls are moved to
     * jhl_cd_hack function.
     * (Thanks to CyberWarriorX for the idea.)
     */
//    // Mask any interrupts, we don't need to be interrupted
//    old_level_mask = interrupt_get_level_mask();
//    interrupt_set_level_mask(0xF);

    hirq_temp = CDB_REG_HIRQ;

    // Make sure CMOK flag is set, or we can't continue
    if (!(hirq_temp & HIRQ_CMOK))
    {
        return IAPETUS_ERR_CMOK;
    }

    // Clear CMOK and any other user-defined flags
    CDB_REG_HIRQ = ~(hirq_mask | HIRQ_CMOK);

//    // Alright, time to execute the command
//    cd_write_command(cd_cmd);
    CDB_REG_CR1 = cd_cmd->CR1;
    CDB_REG_CR2 = cd_cmd->CR2;
    CDB_REG_CR3 = cd_cmd->CR3;
    CDB_REG_CR4 = cd_cmd->CR4;

    // Let's wait till the command operation is finished
    if (!cd_wait_hirq(HIRQ_CMOK))
    {
        return IAPETUS_ERR_TIMEOUT;
    }

//    // Read return data
//    cd_read_return_status(cd_cmd_rs);
    cd_cmd_rs->CR1 = CDB_REG_CR1;
    cd_cmd_rs->CR2 = CDB_REG_CR2;
    cd_cmd_rs->CR3 = CDB_REG_CR3;
    cd_cmd_rs->CR4 = CDB_REG_CR4;

    cd_status = cd_cmd_rs->CR1 >> 8;

    // Was command good?
    if (cd_status == STATUS_REJECT)
    {
        return IAPETUS_ERR_INVALIDARG;
    }
    else if (cd_status & STATUS_WAIT)
    {
        return IAPETUS_ERR_BUSY;
    }

//    // return interrupts back to normal
//    interrupt_set_level_mask(old_level_mask);

    // It's all good
    return IAPETUS_ERR_OK;
}

int cd_set_sector_size(int size)
{
    cd_cmd_struct cd_cmd;
    cd_cmd_struct cd_cmd_rs;

    /*
     * iapetus declares a global variable specifying sector size.
     * This global variable isn't used here, hence code below
     * is commented out.
     * Then, global variable on Pseudo Saturn Kai side is
     * updated to its correct value (SECT_2048) in cdhack_exec.h.
     */
    //cd_sector_size = size;

    cd_cmd.CR1 = 0x6000 | (size & 0xFF);
    cd_cmd.CR2 = size << 8;
    cd_cmd.CR3 = 0x0000;
    cd_cmd.CR4 = 0x0000;

    return cd_exec_command(HIRQ_ESEL, &cd_cmd, &cd_cmd_rs);
}

int cd_put_sector_data(u8 sel_num, u16 num_sectors)
{
    cd_cmd_struct cd_cmd;
    cd_cmd_struct cd_cmd_rs;

    cd_cmd.CR1 = 0x6400;
    cd_cmd.CR2 = 0x0000;
    cd_cmd.CR3 = sel_num << 8;
    cd_cmd.CR4 = num_sectors;

    return cd_exec_command(HIRQ_EHST, &cd_cmd, &cd_cmd_rs);
}

int cd_end_transfer()
{
    int ret;
    cd_cmd_struct cd_cmd;
    cd_cmd_struct cd_cmd_rs;

    cd_cmd.CR1 = 0x0600;
    cd_cmd.CR2 = 0x0000;
    cd_cmd.CR3 = 0x0000;
    cd_cmd.CR4 = 0x0000;

    ret = cd_exec_command(0, &cd_cmd, &cd_cmd_rs);

    CDB_REG_HIRQ = (~HIRQ_DRDY) | HIRQ_CMOK;

    return ret;
}

int cd_get_stat(cd_stat_struct *cd_status)
{
    cd_cmd_struct cd_cmd;
    cd_cmd_struct cd_cmd_rs;
    int ret;

    cd_cmd.CR1 = 0x0000;
    cd_cmd.CR2 = 0x0000;
    cd_cmd.CR3 = 0x0000;
    cd_cmd.CR4 = 0x0000;

    if ((ret = cd_exec_command(0, &cd_cmd, &cd_cmd_rs)) != 0)
    {
        return ret;
    }

    cd_status->status = cd_cmd_rs.CR1 >> 8;
    cd_status->flag = (cd_cmd_rs.CR1 >> 4) & 0xF;
    cd_status->repeat_cnt = cd_cmd_rs.CR1 & 0xF;
    cd_status->ctrl_addr = cd_cmd_rs.CR2 >> 8;
    cd_status->track = cd_cmd_rs.CR2 & 0xFF;
    cd_status->index = cd_cmd_rs.CR3 >> 8;
    cd_status->FAD = ((cd_cmd_rs.CR3 & 0xFF) << 16) | cd_cmd_rs.CR4;

    return IAPETUS_ERR_OK;
}

int is_cd_auth(u16 *disc_type_auth)
{
    cd_cmd_struct cd_cmd;
    cd_cmd_struct cd_cmd_rs;

    cd_cmd.CR1 = 0xE100;
    cd_cmd.CR2 = 0x0000;
    cd_cmd.CR3 = 0x0000;
    cd_cmd.CR4 = 0x0000;

    // If command fails, assume it's not authenticated
    if (cd_exec_command(0, &cd_cmd, &cd_cmd_rs) != IAPETUS_ERR_OK)
    {
        return FALSE;
    }

    if (disc_type_auth)
    {
        *disc_type_auth = cd_cmd_rs.CR2;
    }

    // Disc type Authenticated:
    // 0x00: No CD/Not Authenticated
    // 0x01: Audio CD
    // 0x02: Regular Data CD(not Saturn disc)
    // 0x03: Copied/Pirated Saturn Disc
    // 0x04: Original Saturn Disc
    if (cd_cmd_rs.CR2 != 0x04 && cd_cmd_rs.CR2 != 0x02 && cd_cmd_rs.CR2 != 0x01)
    {
        return FALSE;
    }

    return TRUE;
}

int cd_reset_selector_all()
{
    //return cd_reset_selector(0xFC, 0);
    const u8 reset_flags = 0xFC;
    const u8 sel_num     = 0;
    int ret;

    cd_cmd_struct cd_cmd;
    cd_cmd_struct cd_cmd_rs;

    // Reset Selector Command
    cd_cmd.CR1 = 0x4800 | reset_flags;
    cd_cmd.CR2 = 0x0000;
    cd_cmd.CR3 = (sel_num << 8);
    cd_cmd.CR4 = 0x0000;

    if ((ret = cd_exec_command(HIRQ_EFLS, &cd_cmd, &cd_cmd_rs)) != 0)
    {
        return ret;
    }

    // wait for function to finish
    while (!(CDB_REG_HIRQ & HIRQ_ESEL)) {}

    return IAPETUS_ERR_OK;
}



//////////////////////////////////////////////////////////////////////////////

int cd_move_sector_data_cd_auth(u8 dst_filter, u16 sector_pos, u8 sel_num, u16 num_sectors, int* r)
{
    cd_cmd_struct cd_cmd1;
    cd_cmd_struct cd_cmd2;
    cd_cmd_struct cd_cmd_rs;
    int i;
    cd_stat_struct cd_status;
    u16 auth;

    cd_cmd1.CR1 = 0x6600 | dst_filter;
    cd_cmd1.CR2 = sector_pos;
    cd_cmd1.CR3 = sel_num << 8;
    cd_cmd1.CR4 = num_sectors;

    cd_cmd2.CR1 = 0xE000;
    cd_cmd2.CR2 = 0x0000;
    cd_cmd2.CR3 = 0x0000;
    cd_cmd2.CR4 = 0x0000;

    r[0] = cd_exec_command(0, &cd_cmd1, &cd_cmd_rs);

    /* Clear hirq flags. */
    CDB_REG_HIRQ = ~(HIRQ_DCHG | HIRQ_EFLS);

    /* Authenticate disc. */
    if((r[1] = cd_exec_command(HIRQ_EFLS, &cd_cmd2, &cd_cmd_rs)) != 0)
    {
        return r[1];
    }

    /* Wait till operation is finished. */
    while(!(CDB_REG_HIRQ & HIRQ_EFLS)) {}

    /* Wait until drive has finished seeking. */
    for(;;)
    {
        /* Wait a bit. */
        for(i = 0; i < 100000; i++) { }

        if(cd_get_stat(&cd_status) != 0)
        {
            continue;
        }

        if(cd_status.status == STATUS_PAUSE)
        {
            break;
        }
        else if(cd_status.status == STATUS_FATAL)
        {
            return IAPETUS_ERR_UNKNOWN;
        }
    }

    /* Was Authentication successful ? */
    if(!is_cd_auth(&auth))
    {
        return IAPETUS_ERR_AUTH;
    }

    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

/*
 * Details about what jhl_cd_hack function does :
 * (from : http://assemblergames.com/l/threads/saturn-cd-block-rom-dumped.52419/page-5#post-762819)
 *
 * It pleases me greatly to tell that the analysis of this dump has already borne some fruit, 
 * allowing one of the Saturn's security expectations to be gently violated.
 * 
 * The assumption in question is that burned Saturn discs - ie. those beginning with SEGA SEGASATURN, 
 * but missing the ring wobble or ring data - will be detected as fakes (type 3), and access to their
 * data disallowed. This prevents any SH-2 code - BIOS or cartridge - from booting such a disc.
 * 
 * Using a combination of static analysis of the binary, and tests in my CDB emulator (which is getting
 * quite decent), it's been possible to find a sequence of commands which tricks the CDB into
 * authenticating such a disc as a data disc (type 2). In this case, access to the whole disc is
 * permitted, although the stock Saturn BIOS will still not try and boot it as it's not seen as a
 * Saturn disc (type 4).
 * 
 * This means that it will be possible for Cyber Warrior X's Pseudo Saturn to boot burned discs without
 * needing them to be patched first. It would also be possible for a sufficiently motivated person to
 * build the attack into a modified BIOS (presumably in conjunction with the region mods which are
 * already deployed in this way), essentially embedding the core idea (booting type2 discs) of Pseudo Saturn.
 * 
 * 
 * A description of the attack follows, for the sake of public documentation (sorry, I'm slack about
 * documenting things). For details of how the CDB interfaces with the Saturn please see CWX's docs here:
 * http://wiki.yabause.org/index.php5?title=CDBlock
 * 
 * The CDB runs a multi-tasking operating system. Each task has a priority determining who gets to
 * run when. There are also mutexes, used to protect things like the selector/filter data structures, 
 * which are accessed from many tasks.
 * The CDB is notified of new commands by an interrupt. For simple/fast tasks, the interrupt handler
 * directly calls the relevant code. For slower or more complicated tasks - which may run in the
 * background - the handler instead sends off a message to wake a task to perform the action.
 * Relevant here are tasks 4, 7, C, and D (they're numbered but not contiguously). During the execution
 * of the authentication process, no commands at all are accepted; this limits the possible attack
 * moves to things that can be performed before kicking off auth.
 * 
 * Task 7 is the critical one: filesystem ops and authentication. The former represents a lot
 * of spaghetti code in the task but isn't relevant to us; the latter is what's critical. At the
 * beginning of authentication, the TOC is checked to see if there's a first data track to consider
 * (otherwise it's an audio CD). If so, the first sector is read and checked: if it's a Mode1 sector,
 * the task continues on to read a bunch of sectors, looking for SEGA SEGASATURN - and the various
 * other auth jobs, not relevant here. If Task 7 gets a non-Mode1 sector, it assumes it's a VCD or
 * other data disc.
 * 
 * Looking more closely at how it gets that sector: it sets up one of the CDB selectors, filtering
 * on FAD 150 (the first sector of a disc). It then stores the number of the next sector to be read
 * into that selector's buffer - for example, if there are 10 sectors in the buffer, it will store
 * the number of sector 11, where it'll expect to find this sector.
 * It then requests a read from FAD 150 through the CD mechanism task (task 6), and then looks at
 * the position it stored in the buffer. Whatever's there determines how it will proceed. So, if you
 * can somehow write arbitrary sectors into this selector in between the first part (where it stores
 * the sector number) and the second part (where it reads the sector), you can cause it to see a sector
 * that's not really on the disc - and miss the SEGA SEGASATURN header that may be there.
 * 
 * On to the other actors.
 * Task 4 performs selector-related tasks: commands 0x40, 42, 46, 42, 52, 55. A number of these
 * are capable of modifying the sector count in a selector, but unfortunately, they're all pretty
 * quick. Any attempt to start one of these before the auth process just results in it all happening
 * either before or after the auth process: no good.
 * 
 * Task C is used for Saturn<->CDB DMA: commands 0x61, 62, 63, 64.. That's how the Saturn normally
 * gets its data from the CD; it can also be used in the other direction, putting sectors from the
 * Saturn into CDB selectors. This is really nifty, and higher priority than Task 7 - but unfortunately,
 * it doesn't actually update the selectors until the transfer is done and you call command 0x06,
 * and that executes immediately. So it's a piece of the puzzle but it's not useful for attack.
 * 
 * Our final friend appears in Task D: handling the somewhat mysterious Copy/Move Sector commands
 * 0x65, 66. These don't seem to be used very often (if at all), but their purpose is simply to move
 * or duplicate sectors from a selector into another. They're fully filtered on the way; for example,
 * if you're copying sectors with FAD 149-151 into a selector with FAD 150 filtering active, only the
 * 150 will get through. This has the rather nice property that it will keep running over the top
 * of Task 7 when 7 sleeps waiting for the CD to read that critical first sector - so it's actually
 * possible to write into the necessary sector before the CD drive returns with the real data.
 * 
 * So, putting it all together, the attack is as follows:
 * 
 * 1. Use the Put Sector Data command to put a whole bunch of sectors into the CDB - all with FAD 150,
 *    Mode 2 in their headers.
 * 2. Call End Data Transfer to push them into the selector.
 * 3. Call Copy Sector Data, starting a copy of all those fake sectors into the selector that the
 *    Authenticate Disc command will be using.
 * 4. Immediately (like 15 microseconds later immediately) call the Authenticate Disc command,
 *    specifying the same selector.
 * 
 * The authentication process sees the fake sectors, decides it's a data disc, unlocks read access,
 * and returns Type 2 - data disc. VoilA!!
 * 
 * Technically the process looks very similar to that just described:
 * - set sector size to 2352
 * cmd: 0x6003, 0x0300, 0x0000, 0x0000
 * wait: HIRQ_CMOK
 * - write 150 sectors into selector 1
 * cmd: 0x6400, 0x0000, 0x0100, 0x0096
 * wait: HIRQ_DRDY
 * Then write 150 x 2352 bytes to the transfer register -
 * content is unimportant, except at offset 12 in each sector, you must
 * have the sequence:
 * 00 02 00 02
 * as in, MSF 00:02:00, Mode 2 header
 * - end the transfer
 * cmd: 0x0600, 0x0000, 0x0000, 0x0000
 * wait: HIRQ_EHST
 * - start a sector copy to seltr 0, but not enough secs to overflow the
 * buffer space available
 * cmd: 0x6600, 0x0000, 0x0100, 0x0032
 * - immediately begin auth with seltr 0 (wait for HIRQ_CMOK or about 15
 * microseconds so that the copy command is accepted)
 * cmd: 0xe000, 0x0000, 0x0000, 0x0000
 * 
 * - Wait for HIRQ_EFLS and HIRQ_ECPY, then clear the buffer and go back to business.
 * 
 * The attack has now been tested on real hardware, with many thanks to CWX; expect some news
 * in the Pseudo Saturn thread in due course.
 */
int jhl_cd_hack(u16* type, int* r)
{
    int i, j;

    u32 old_level_mask;

    // Mask any interrupts, we don't need to be interrupted
    old_level_mask = interrupt_get_level_mask();
    interrupt_set_level_mask(0xF);

    if((r[0] = cd_set_sector_size(SECT_2352)) != 0)
    {
        // return interrupts back to normal
        interrupt_set_level_mask(old_level_mask);
        return 1;
    }

    /**
     * 1. Use the Put Sector Data command to put a whole bunch of sectors into the CDB - all with FAD 150,
     *    Mode 2 in their headers.
    **/
    /* Write 150x2352 sectors. */
    if((r[1] = cd_put_sector_data(0, 150)) != 0)
    {
        // return interrupts back to normal
        interrupt_set_level_mask(old_level_mask);
        return 2;
    }

    /* Copy data. */
    for(j = 0; j < 150; j++)
    {
        for(i = 0; i < (2352/4); i++)
            CDB_REG_DATATRNS = 0x00020002;
    }

    /**
     * 2. Call End Data Transfer to push them into the selector.
    **/
    if((r[2] = cd_end_transfer()) != 0)
    {
        // return interrupts back to normal
        interrupt_set_level_mask(old_level_mask);
        return 3;
    }

    while(!(CDB_REG_HIRQ & HIRQ_EHST)) {}

    /**
     * 3. Call Copy Sector Data, starting a copy of all those fake sectors into the selector that the
     *    Authenticate Disc command will be using.
    **/
    r[3] = cd_move_sector_data_cd_auth(0, 0, 0, 50, r+4);

    /**
     * 4. Immediately (like 15 microseconds later immediately) call the Authenticate Disc command,
     *    specifying the same selector.
    **/
    r[6] = is_cd_auth(type);
    //conio_printf(2, 7, COLOR_WHITE, "is_cd_auth = %d, type = %d", ret, *type);

    while(!(CDB_REG_HIRQ & HIRQ_ECPY)) {}

    if((r[7] = cd_end_transfer()) != 0)
    {
        // return interrupts back to normal
        interrupt_set_level_mask(old_level_mask);
        return 7;
    }

    if((r[8] = cd_set_sector_size(SECT_2048)) != 0)
    {
        // return interrupts back to normal
        interrupt_set_level_mask(old_level_mask);
        return 8;
    }

    if((r[9] = cd_reset_selector_all()) != 0)
    {
        // return interrupts back to normal
        interrupt_set_level_mask(old_level_mask);
        return 9;
    }

    // return interrupts back to normal
    interrupt_set_level_mask(old_level_mask);
    return r[9];
}

