/**
 *  Sega Saturn cheat codes for Pseudo Saturn Kai
 *  by cafe-alpha and CyberWarriorX.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CHEATS_DATA_H
#define CHEATS_DATA_H

#include <stdio.h>

/* Inclusion of structure alignment macros. */
#include "../../satcom_lib/sc_common.h"


/**
 * Misc. informations about cheat codes implementation :
 *
 * About vocabulary :
 *  - A "code" is a pair of address and value.
 *    Example : 36055FEC 0003
 *    Upper 4 bits of the address is the command of the code, detailed
 *    a bit later in this document.
 *  - A "cheat" is a combination of one or several code(s), used in order
 *    to modify game behavior.
 *    Cheats are divided into two main groups "master cheat" and "game cheat".
 *    "Game cheat" describes how modify Saturn memory while game is running.
 *    "Master cheat" describes how to setup Saturn internals in order to
 *    execute game cheats while game is running.
 *
 *    To sum up, list handling all cheat codes is organized as described below :
 *    A list of all games
 *    ... each game owning a list of cheats
 *        ... each cheat owning one or several code(s)
 *            ... each code being composed by address and value
 *                ... each address having command set in upper 4 bits
 *
 * ------------------------------------------
 * Data related to cheat codes are splitted into 3 main locations :
 *  - "ROM data", containing address/codes pairs for each games.
 *    Stored uncompressed in ROM, easy to access from cheats codes interrupt.
 *    Note : only codes for user cheats are stored in ROM.
 *           Codes for master cheats are stored in RAM. (See below for details)
 *  - "RAM data", containing games and codes descriptions as well as codes for master cheats.
 *    Stored in main executable (= compressed in ROM, decompressed to
 *    LRAM on firmware startup), so can't be accessed from cheats ROM code handler.
 *  - "configuration data", containing informations about which game is selected, 
 *    which code(s) are enabled, and a cache for use when few codes are enabled.
 *    It also contains a small SH-2 executable in  order to manage cheat codes handler.
 *    Stored in a "safe" place in RAM, not accessed by a majority of games.
 *    As small as possible, in order to avoid conflicts with game.
 *
 * Advantages of this memory arrangement :
 *  - Flash ROM friendly : ROM is not written each time cheats
 *    configuration is modified : changing codes configuration
 *    won't brick the cartridge)
 *  - Small size in flash memory, because RAM data is compressed with
 *    Pseudo Saturn Kai main executable.
 *  - During execution of cheat codes interrupt handler, enabled codes can be retrieved
 *    directly from ROM and hence don't need to be cached in configuration data.
 *
 * Disadvantages of this memory arrangement :
 *  - Modifiying codes list requires to recompile the whole firmware.
 *  - Codes can be enabled for only one game at once.
 *    (Is that a problem ? And this can be improved if really needed)
 *  - Access to cartridge's flash ROM in order to retrieve enabled code data
 *    is rather slow, and hence may cause issues with games sensitive regarding
 *    A-Bus access.
 *
 * ------------------------------------------
 * How cheat codes are managed (overview) :
 *  - From firmware, user selects game on which cheats will be enabled.
 *  - Before booting, firmware starts initialization of configuration data.
 *  - Before booting, firmware calls function in cheat codes ROM in order to
 *    enable master code and finish configuration data initialization.
 *  - When cheats interrupt handler is called, enabled codes are processed, 
 *    and optionally USB connectivity is checked.
 *
 * ------------------------------------------
 * About cheat codes interrupt handler :
 *  - Interrput for cheat codes is set in order to execute from beginning of configuration
 *    data. From there, PC is copied to register in order to be provided as parameter to
 *    cheat codes handler.
 *      1. Save R4 to stack
 *      2. Store PC to R4
 *      3. Call cheat codes handler
 *      4. Restore R4 from stack
 *      5. Execute opcode patched by master code
 *      6. rte
 *
 *
 * ------------------------------------------
 * About the two implementations of cheat codes handler :
 *  - ROM-based cheat codes handler, which is executed directly from cheat
 *    codes ROM area.
 *    This implementation is written in C, and supports all known types of
 *    commands for codes. It also supports USB connectivity.
 *    In this implementation, enabled codes information are read from
 *    cartridge ROM, and consequently RAM usage is reduced.
 *    Demerit of this implementation is to read cartridge ROM and being a bit
 *    slow (because of C language), which may cause issues (typically, missing
 *    graphisms) with some games.
 *      1. Store all registers to interrupt context area, located in stack
 *          * A small tweak is needed in order to store real value of R4
 *            to interrupt context area.
 *      2. Provide pointer to interrupt context area via R5 register
 *      3. Call user function (in C)
 *          3-1. Perform USB connectivity stuff
 *          3-2. If configuration memory is corrupted, emergency exit
 *          3-3. If used, do "D" code check, and exit if it doesn't matches
 *          3-4. Execute "1" and "3" codes.
 *      4. Restore all registers from interrupt context area.
 *      5. rts
 *
 *  - In-RAM cheat codes handler, which is executed directly within configuration
 *    data.
 *    This implementation is written in assembly, and supports only the minimal
 *    set of features : common commands for codes, and no support for USB connectivity.
 *    In this implementation, enabled codes are read from configuration data, which
 *    haves the consequence of being fast (in comparison with ROM-based way of accessing
 *    codes information), but also haves the consequence of supporting a (small) maximum
 *    of codes.
 *    Demerit of this implementation is to have limitation regarding codes count, and
 *    not supporting USB connectivity. Since it is the faster implementation, it is
 *    used by default whenever possible.
 *      1. Store minimal set of registers to stack
 *          * Upper level function saved R4 to stack, so it's not required to save it
 *      2. If used, do "D" code check, and exit if it doesn't matches
 *      3. Execute "1" and "3" codes.
 *      4. Restore registers from stack.
 *      5. rts
 *      Note : codes are read from configuration memory, whose location
 *             is provided via R4 register.
**/

/* Some constants that may vary in future revisions, 
 * generated by CodeGen, hence put in a separate file.
 * Contents :
 *  - CT_MAXCODES_PER_GAME : Maximum code (=address/code pair) count for each games, including master codes.
 *  - CT_MAXCHEATS_PER_GAME : Maximum cheat count for each games, excluding master cheat.
 *  - CT_GAMECNT  : Games count.
 *  - CT_CHEATCNT : Total cheats count.
 *  - CT_CODECNT  : Total codes (address/code pairs) count.
 *  - CT_REVISION : Cheat codes revision number (incremented on each release).
 */
#ifdef PSKAI_LITE_BUILD
#   include "cheats_lite_data_defs.h"
#else // !PSKAI_LITE_BUILD
#   include "cheats_full_data_defs.h"
#endif // !PSKAI_LITE_BUILD



// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif


/* Address/value pair structure definition, 
 * used in ROM data.
 */
typedef struct _ct_code_t
{
    /* Address. */
    unsigned long addr;

    /* Value. */
    unsigned short value;

    /* Commands :
     *  - 0x0F : Patch 16 bits specified at specifed address with code value.
     *           Typical usage : "F6000914 C305" patches address 0x06000914
     *           with 0xC305 value, which is "trapa #5" instruction.
     *           [Master code only]
     *  - 0x0B : Specify address for Action Replay handler code and cached
     *           cheat codes. Default value is 0x060FF000.
     *           [Master code only]
     *
     *  - 0x00 : 16 bit word write at bootup.
     *  - 0x01 : Continuous 16 bit word write.
     *  - 0x03 : Continuous 8 bit byte write with lower 8 bits of value.
     *  - 0x08 : Continuous 2x8 bit bytes write.
     *           Note: Address may not be 16 bits aligned.
     *           Note: Upper 8 bits are written to specified address, and
     *                 lower ones are written to specified address + 1 byte.
     *  - 0x0D : Activate all other codes only when specified address contains
     *           specified 16 bits value.
     */
    unsigned char command;

    /* For 4 bytes alignment. */
    unsigned char unused;
} SC_ALIGN_4_BYTES(ct_code_t);

/* Single cheat code structure definition
 * Contains cheat description, and link to address/value pair(s).
 * used in RAM data.
 */
typedef struct _ct_cheat_t
{
    /* Cheat name. */
    char* name;

    /* Index of first user code
     * in address/code pair list.
     */
    unsigned short first_code_id;

    /* Code count. */
    unsigned char num_codes;

    /* Misc. parameters (bitfield). */
    unsigned char params_bf;
} SC_ALIGN_4_BYTES(ct_cheat_t);


/* Cheat parameters bitfield definitions :
 *  - CT_PARAM_BF_DEFAULT_ON :
 *    Enabled by default in cheats selection menu. For use with
 *    `D' codes, code required by Pseudo Saturn game loader, etc.
 *  - CT_PARAM_BF_GREYOUT :
 *    Show a greyed out line in cheats selection menu. For use
 *    in order to show sub menu in games having codes separated
 *    into several categories. Example : multiple CDROM games, several
 *    sub games in on CDROM (Sonic Jam, Die Hard Trilogy, etc), or
 *    in order to separate codes between two different categories, etc.
 */
#define CT_PARAM_BF_DEFAULT_ON 0x01
#define CT_PARAM_BF_GREYOUT    0x80


/* IP CRC related definitions :
 *  - CT_IPCRC_CNT     : Maximum count of CRCs per game
 *  - CT_IPCRC_INVALID : Dummy CRC used in order to fill CRC list when there
 *                       are less than CT_IPCRC_CNT CRCs for a given game.
 */
#define CT_IPCRC_CNT     8
#define CT_IPCRC_INVALID 0x00000000


/* Single game cheats structure definition
 * Contains game informations, and link to several cheat codes.
 * used in RAM data.
 */
typedef struct _ct_game_t
{
    /* Game name. */
    char* name;

    /* Game IP header CRC. */
    unsigned long ip_crc[CT_IPCRC_CNT];

    /* Code informations from master (`F') code :
     *  - Patch address
     *  - Patch opcode
     */
    unsigned long  master_address;
    unsigned short master_opc;

    /* Count of cheat(s) available for
     * this game.
     * -> at least one.
     */
    unsigned short num_cheats;

    /* Configuration memory location from `B' code.
     * This defines address where to store cheat codes
     * interrupt handler code and codes cache.
     */
    unsigned long  cfg_location;

    /* For 4 bytes alignment, set to zero.
     * Reserved for future use, if any.
     *
     * Note : This can be set from `B' code's upper byte.
     */
    unsigned char unused_u8;

    /* Codes cache capacity, from `B' code.
     * Unit : 4 codes (32 bytes).
     * (0: no cache, 4: up to 4 codes, 2: up to 8 codes, etc)
     *
     * When codes selection exceeds cache capacity,
     * then ROM based cheat codes handler is automatically used.
     *
     * Note : value is defined in `B' code's lower byte, 
     *        and upper byte is reserved for future use.
     */
    unsigned char cache_4_capacity;

    /* Index of first user cheat code
     * in cheats list.
     */
    unsigned short first_cheat_id;
} SC_ALIGN_4_BYTES(ct_game_t);



/**
 * "Enabled codes" bitfield related macros.
 *
 * CT_ENABLED_ULONG_CNT : number of u32 elements required in order to
 *                        manage enabled codes.
 * CT_MARK_ENABLED      : mark specified code as enabled.
 * CT_IS_ENABLED        : verifies if specified code is enabled or not.
 * CT_MARK_DISABLED     : mark specified code as disabled.
 *
 * ID parameter is game's user code ID.
**/
#define CT_ENABLED_ULONG_CNT ((CT_MAXCODES_PER_GAME+31) / 32)
#define CT_MARK_ENABLED(_ENBF_, _ID_)   (_ENBF_)[(_ID_)/32] |= (1 << ((_ID_)%32))
#define CT_IS_ENABLED(_ENBF_, _ID_)    ((_ENBF_)[(_ID_)/32] & (1 << ((_ID_)%32)))
#define CT_MARK_DISABLED(_ENBF_, _ID_)  (_ENBF_)[(_ID_)/32] &= ~(1 << ((_ID_)%32))


/**
 * Size of in-RAM cheat codes handler, in bytes.
 * Code currently takes 34 instructions, so let's set the size
 * to 48 instructions (0x60 bytes) in the case something needs
 * to be fixed or improved in the future.
 *
 * This in-RAM cheat codes handler is located just
 * after cheat codes cache :
 *  |- configuration memory
 *  |   |- interrupt boostrap code
 *  |   |- interrupt boostrap data
 *  |   |- cheat codes misc. data
 *  |- enabled codes cache (8 bytes per code)
 *  |- cheat codes handler <- you are here !
**/
#define CHEATS_INRAM_HANDLER_SIZE (48*sizeof(unsigned short))


/**
 * Maximum supported codes count for in-RAM cheat codes handler.
 *
 * This is the grand maximum value supported by Pseudo Saturn Kai, 
 * which is limited by type and unit of cheat codes cache capaticty.
 * This capacity is defined `B' code's value lower 1 byte, and 
 * its unit is 4 codes, hence this "0x100 * 4 = 1024 codes" value.
 *
 * In order to avoid overwrite of code/data, missing graphics and/or
 * performance issues, some game may have a smaller limit defined
 * in their `B' code.
**/
#define CT_CODES_CACHE_MAX 1024



/* Configuration data structure definition
 * Contains information about game and cheats enabled.
 */
typedef struct _ct_config_t
{
    /* Cheat codes interrupt bootstrap code.
     * Size : 20 instructions, followed with following data :
     *  - cheats handler location (4 bytes)
     *  - configuration memory location (4 bytes)
     */
#define CHEATS_BOOTSTRAP_INST_CNT (20)
#define CHEATS_BOOTSTRAP_IHR_SIZE (CHEATS_BOOTSTRAP_INST_CNT*sizeof(unsigned short))
    unsigned short bootstrap_header[16];

    /* Backup of opcode patched by master code.
     * This will be executed just before exiting cheats.
     * interrupt handler.
     *
     * This opcode needs to be manually set when
     * setting up cheat codes interrupt bootstrap.
     */
    unsigned short master_patch_opc;

    /* Remaining instructions until return from exception.
     * unused nop -> rte/nop, hence 3 instructions (6 bytes).
     */
    unsigned short bootstrap_footer[
                        CHEATS_BOOTSTRAP_INST_CNT
                        -  1/*master_patch_opc*/
                        - 16/*bootstrap_header*/];

    /* End of cheat interrupt boostrap code. */
    /* ------------------------------------------ */

    /* Cheat codes handler location.
     * It is located either in cartridge ROM (for ROM-based cheat
     * codes handler) or configuration memory (for in-RAM cheat
     * codes handler)/
     *
     * This location is read and jumped to by cheat codes interrupt bootstrap.
     *
     * This location needs to be manually set when
     * setting up cheat codes interrupt bootstrap.
     */
    unsigned long ct_hdlr_location;

    /* Configuration memory location.
     *  - Points to a safe place in RAM not used by game.
     *  - Small size : smallest size possible is sizeof(ct_config_t), and
     *    this grows according to number of cheat codes enabled.
     *
     * Typical locations :
     *  - 0x060FF000 : default location.
     *  - 0x06002800 : default alternate location.
     *  - Other locations, rarely used : 0x06030040, 0x060FE000, 0x06062800, etc
     *    Note : alternate location is defined in "B" code.
     *
     * About capacity of typical locations :
     *  - 0x060FF000 : 0x1000 (4096) bytes available for handler and codes cache.
     *                 -> maximum 492 codes.
     *  - 0x06002800 : 1352 bytes available for handler and codes cache.
     *                 -> maximum 157 codes.
     * Note : this is calculated with the following sizes :
     *  - Configuration memory (ct_config_t)               : 0x40 bytes
     *  - In-RAM codes handler (CHEATS_INRAM_HANDLER_SIZE) : 0x60 bytes
     *
     * This location needs to be manually set when setting up cheat codes
     * interrupt bootstrap, and this interrupt bootstrap uses it as parameter
     * to cheat codes handler.
     */
    unsigned long cfg_location;

    /* End of cheat interrupt boostrap data. */
    /* ------------------------------------------ */

    /* Enable code address value. This is defined in "D" code.
     * Two cases for address :
     *  - Zero : don't check enable code
     *  - Nonzero : check enable code
     */
    unsigned long enable_addr;
    unsigned short enable_val;
    unsigned char reserved3[2];

    /* ------------------------------------------ */

    /* Index of game's first user code
     * in codes list stored in ROM.
     */
    unsigned short user_start_id;

    /* Codes count for in-game user cheats. This consequently
     * doesn't includes master code nor startup code(s).
     *
     * This can be either :
     *  - Enabled code maximum count, for cartridge ROM based cheat codes handler.
     *  - Cached codes count, for in-RAM cheat codes handler.
     */
    unsigned short user_codes_cnt;

    /* ------------------------------------------ */

    /* Miscellaneous flags :
     *
     *  - bit0  : In-game USB access
     *  - other : unused, assign with care (see note below)
     *  - bit5  : Fill code cache with dummy value (debug use).
     *  - bit6  : Cartridge ROM-less in-game cheat codes handler
     *  - bit7  : Enable exception handlers
     *
     * Note : bit5-7 are reset when setting up master code, 
     *        hence it is safe to perform USB connectivity
     *        check when debug flag is non-zero.
     */
#define CHEATS_FLAGS_INGAME_USB    0x01
#define CHEATS_FLAGS_CACHEFILL     0x20
#define CHEATS_FLAGS_CARTROM_LESS  0x40
#define CHEATS_FLAGS_ENABLE_EX_IHR 0x80
    unsigned char flags;
    unsigned char padding1;

    /* Counter, used to make green LED blinking when
     * cheat codes interrupt is running.
     *
     * LED is turned on when cycling counter from period to zero.
     * LED is turned off to specified value, so that it's possible
     * to set LED turning on duty and length.
     *
     * Counter:[*****************|.............................]
     *          +Zero (LED ON)   +LED off                      +Period
     *
     * Note : period/duty are adjusted in order to blink every 1~2 seconds
     *        when cheat codes are enabled.
     *        The blink timing varies from a game to another.
     */
#define CHEAT_ACT_LED_PERIOD   0x400
#define CHEAT_ACT_LED_TURN_OFF ((CHEAT_ACT_LED_PERIOD * 20) / 100)
    unsigned short led_counter;

    /* ------------------------------------------ */

    /*
     * Memory requirements when executing cheat codes are very tight, hence 
     * "reserved" memory area at the end of this structure isn't declared.
     * -> Cheat codes config area size is set to 64 bytes.
     */
} SC_ALIGN_4_BYTES(ct_config_t);


/* Version structure definition.
 * For use by ROM code function when returning back version infos.
 */
typedef struct _ct_version_t
{
    /* Input parameter, indicating if ROM-less or not. */
    int is_rom_less;

    /* Version strings, copied from ROM code header. */
    char* version1;
    char* version2;

    /* Build date. */
    char* date;

    /* Extra comment. */
    char* comment;
} SC_ALIGN_4_BYTES(ct_version_t);



/* Structure definition for cheat codes apply parameters.
 *
 * Note : bootstub.h is included in order to retrieve
 *        structure definition of bootstub_params_t structure.
 */
#include "../bootstub/bootstub.h"
typedef struct _ct_apply_params_t
{
    /* Bootstub parameter : used in order to
     * apply cheat startup code(s). It is necessary since
     * game first executable code may be located in
     * temporary RAM location until being copied then
     * executed by bootstub.
     */
    bootstub_params_t* bs_params;


    /* Cheat codes configuration memory.
     * Temporary buffer of configuration memory :
     *  - Initialized on Pseudo Saturn Kai side.
     *  - Copied to appropriate location on cheat codes apply ROM code side.
     */
    ct_config_t cfg;


    /* Master code information :
     *  - Patch address.
     *  - Patch op-code.
     *
     * Theses are retrieved from cheat codes game list, and used
     * in cheats_rcf_apply. Since they are not required when game is
     * running theses informations are specified in this structure
     * rather than configuration memory.
     */
    unsigned long master_address;
    unsigned short master_opc;

    /* For 4 bytes alignment.
     * Reserved for future use, if any.
     */
    unsigned char reserved1[2];


    /* Pointer to the start of cheat codes list.
     *
     * Some implementations of cheat codes use list stored in ROM
     * and some other use list in Pseudo Saturn Kai work RAM.
     * For this reason, location of this list is stored here.
     */
    ct_code_t* codes_list_start;


    /* User (custom) code usage flag :
     *  - Zero : use builtin codes
     *  - Nonzero : use user codes from cache
     */
    int enable_user_codes;


    /* USB operation mode.
     * Specify how how USB connectivity is performed.
     *  - CT_IGUSB_PROMPT    : Select from USB mode selection menu
     *  - CT_IGUSB_DISABLE   : Don't use USB connectivity
     *  - CT_IGUSB_AND_CHEAT : USB connectivity + cheat codes
     *  - CT_IGUSB_ONLY_ROM  : USB connectivity only (no cheat codes) from ROM (*1)
     *  - CT_IGUSB_ONLY_RAM  : USB connectivity only (no cheat codes) from RAM (*2)
     *
     *  (*1) : Don't use cheat codes cache location, 
     *         but slow since executed from ROM
     *  (*2) : Need around 512 bytes available in cheat
     *         codes cache location, but fast since
     *         executed from RAM
     *  (*1) and (*2) : Cheat codes aren't available.
     */
#define CT_IGUSB_DISABLE    0
#define CT_IGUSB_AND_CHEAT  1
#define CT_IGUSB_ONLY_ROM   2
#define CT_IGUSB_ONLY_RAM   3
#define CT_IGUSB_PROMPT     4
#define CT_IGUSB_ITEMSCNT   5
    int ingame_usb;


    /* "Software Reset to BSOD" usage flag :
     *  - Zero : don't modify software reset handler
     *  - Nonzero : redirect software reset handler to BSOD
     *
     * Purpose : provide a way to do USB transfer even when
     *           master code isn't working.
     *           This doesn't allows to do in-game USB transfer, 
     *           but allows to check memory contents after executing
     *           a game, for example in order to search suitable
     *           location for cheat codes configuration location.
     */
    int use_bsod_softres;


    /* "Show configuration contents on startup" usage flag :
     *  - Zero : don't show configuration (normal use)
     *  - Nonzero : show configuration on BSOD screen
     *              when applying cheat codes.
     *
     * Purpose : cheat codes development debug.
     *           This will probably be deleted in a future version.
     */
    int show_cfg_on_apply;


    /* Enabled user codes informations :
     *  - `enabled_codes' contains bitfield specifiying enabled codes.
     *  - `codes_cache' contains enabled codes themselves.
     *
     * `enabled_codes' is used when cartridge ROM based cheats handler is enabled.
     * `codes_cache' is used when in-RAM cheats handler is enabled.
     *
     * Note : `enabled_codes_max_id' is maximum ID for all enabled codes.
     *        For example, if codes #0, #1, #7, #9 are enabled, 
     *        then bitfield is something like this : EEdd dddE dEdd dddd ...
     *        and `enabled_codes_max_id' is set to 9.
     *
     * Note : `enabled_codes' includes startup codes (commands 0, 1 and 3),
     *        but `codes_cache' includes only in-game codes (commands 1 and 3).
     *
     * Enabled codes information is copied just after configuration memory
     * in cheats_rcf_apply function.
     */
    unsigned long* enabled_codes;
    unsigned short enabled_codes_max_id;

    unsigned short codes_cache_cnt;
    ct_code_t* codes_cache;

    /* Codes cache capacity, retrieved from `B' code.
     * Unit : 4 codes (32 bytes).
     * (0: no cache, 4: up to 4 codes, 2: up to 8 codes, etc)
     *
     * This is used in order to fill all codes cache with dummy
     * value in order to verify its capacity is correctly designed.
     */
    unsigned char cache_4_capacity;


    /* If non-zero, disable cheat codes.
     * For debug purpose.
     */
    unsigned char cheats_disable;


    /* For 4 bytes alignment.
     * Reserved for future use, if any.
     */
    unsigned char reserved[2];

} SC_ALIGN_4_BYTES(ct_apply_params_t);



/**
 * USB dev cart I/O check module related stuff.
 * More informations available in usb_only_hdlr_ram.s file.
**/
typedef struct _ct_usb_ram_vectors_t
{
    /* USB dev cart register address, used
     * in order to locate this structure.
     */
#define USB_RAM_MAGIC_FLAGS_REG 0x222CAFE1
    unsigned long usb_flags_reg;
#define USB_RAM_MAGIC_FIFO_REG 0x221A1FA1
    unsigned long usb_fifo_reg;

    /* Pointers to main function, and then 
     * to each sub functions.
     */
    unsigned long usb_dev_cart_check_io_ptr;
    unsigned long send_byte_ptr;
    unsigned long recv_byte_ptr;
    unsigned long recv_dword_ptr;
} SC_ALIGN_4_BYTES(ct_usb_ram_vectors_t);

/**
 * The prototype of  USB dev cart I/O check module @ RAM.
 * Function itself can be found in usb_only_hdlr_ram.s file.
**/
void asm_usb_dev_cart_check_io(void);

/**
 * Maximum size of USB dev cart I/O check module @ RAM.
 * This is used in order to avoid potential infinite loop
 * when looking for ct_usb_ram_vectors_t structure.
**/
#define USB_IOCHECK_RAM_MAXLEN (2*1024)






// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif




/* ROM data declaration. */
extern ct_code_t _ct_codes_data[CT_CODECNT];


/* RAM data declarations. */
extern ct_cheat_t _ct_cheats_data[CT_CHEATCNT];
extern ct_game_t _ct_games_data[CT_GAMECNT];


#endif
