/**
 *  Comms Link transfer for Pseudo Saturn Kai lite
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
**/

#include <sc_saturn.h>
#include "commslink.h"


/**
 * PAR Registers addresses definition.
 *
 * https://segaxtreme.net/threads/usb-downloading-device-for-sega-saturn.14685/page-3#post-154494
 * The flash ID has some tag of correlation to cart type (Datel usually uses
 * SST chips, EMS usually uses Atmel chips; I have no idea about other vendors
 * such as Xinga), but I don't think it's a guarantee.
 *
 * 2021/10/24 : edde_baby cartridges comply with this association between flash
 *              ROM chips and Comms registers mapping.
 *              (It took something like 10 years to actually verify that ...)
**/
//
// Comms status flag (read-only)
// ===
// Datel: 0x22500001
// EMS: 0x22100001
#define PAR_DATEL_FLAG         0x22500001
#define PAR_EMS_FLAG           0x22100001

// Comms byte read (read-only)
// ===
// Datel: 0x22600001
// EMS: 0x22180001
#define PAR_DATEL_READ         0x22600001
#define PAR_EMS_READ           0x22180001

// Comms byte write (write-only)
// ===
// Datel: 0x22400001
// EMS: 0x22080001
#define PAR_DATEL_WRITE        0x22400001
#define PAR_EMS_WRITE          0x22080001

unsigned long _comms_address[3] = { 0 };


void comms_set_type(unsigned char type)
{
    if(type == COMMS_TYPE_DATEL)
    { /* Set address to Datel. */
        _comms_address[0] = PAR_DATEL_FLAG;
        _comms_address[1] = PAR_DATEL_READ;
        _comms_address[2] = PAR_DATEL_WRITE;
    }
    else //if(adr == COMMS_TYPE_EMS)
    { /* Set address to EMS. */
        _comms_address[0] = PAR_EMS_FLAG;
        _comms_address[1] = PAR_EMS_READ;
        _comms_address[2] = PAR_EMS_WRITE;
    }
}


/* Uses code from cgfm2 programs. */
unsigned char comms_exc_byte(unsigned char value)
{
    unsigned char temp;
    volatile unsigned char *flag = (unsigned char *)_comms_address[0];

    /* Wait for comms exchange flag. */
    while(((*flag) & 1) == 0);

    /* Receive data. */
    temp = *((unsigned char *)_comms_address[1]);

    /* Send data. */
    *((unsigned char *)_comms_address[2]) = value;

    return temp;
}
unsigned short comms_exc_short(unsigned short data)
{
    unsigned short rdata;
    rdata  = comms_exc_byte((data >>  8) & 0xFF) <<  8;
    rdata |= comms_exc_byte((data >>  0) & 0xFF);
    return rdata;
}
unsigned long comms_exc_long(unsigned long data)
{
    unsigned long rdata;
    rdata  = comms_exc_byte((data >> 24) & 0xFF) << 24;
    rdata |= comms_exc_byte((data >> 16) & 0xFF) << 16;
    rdata |= comms_exc_byte((data >>  8) & 0xFF) <<  8;
    rdata |= comms_exc_byte((data >>  0) & 0xFF);
    return rdata;
}


void comms_link_check(commslink_status_t* stat)
{
    if(sc_usb_is_port_available())
    {
        /* Avoid interference with USB dev cart. */
        return;
    }

    int type = 0;
    for(type=0; type<2; type++)
    {
        comms_set_type(type == 0 ? COMMS_TYPE_DATEL : COMMS_TYPE_EMS);

        /* ARP Communication data. */
        volatile unsigned char *flag = (unsigned char *)_comms_address[0];
        unsigned long tmp;
        unsigned char command_byte;
        unsigned long address;
        unsigned long length;
        unsigned char exec_flag;
        unsigned char checksum;
        unsigned long rw4_buffer;

        unsigned char* ptr;
        unsigned long i;

        int flag_on = 0;
        /*
         * Update state and exchange data when state becomes 1
         */
        if(*flag & 1)
        {
            flag_on = 1;
        }

        /* No data received -> exit function. */
        if(flag_on != 1)
        {
            continue;
        }

        // LINKDIAG_LOGOUT("DATA IN", 0);

        /* Synchronize with PC. */
        tmp = comms_exc_byte('D');
        // LINKDIAG_LOGOUT(" 0x%02X (%c)", tmp, tmp);
        if (tmp != 'I')
        {
            continue;
        }
        tmp = comms_exc_byte('O');
        // LINKDIAG_LOGOUT(" 0x%02X (%c)", tmp, tmp);
        if (tmp != 'N')
        {
            continue;
        }

        /* Synchronization with PC done : indicate which link type is used. */
        stat->comms_type = (type == 0 ? COMMS_TYPE_DATEL : COMMS_TYPE_EMS);

        /* Read the command byte from the PC. */
        command_byte = comms_exc_byte(0);
        // LINKDIAG_LOGOUT(" CMD:0x%02X", tmp);

        stat->tr_count++;

        switch(command_byte)
        {

        /** From http://cgfm2.emuviews.com/sat/satar.txt :
         * Function $01: Download memory
         * 
         * It does the following:
         * 
         * 1. Write longword in SH-2 register R9 to the PC.
         * 2. Read two longwords, address and length, from the PC.
         * 3. If length is zero, exchange bytes 'O' and 'K', then return.
         *    It does not care what the PC sends back.
         * 4. Exchange data bytes from specified memory addresses until all memory
         *    has been read. The SH-2 can hang if it reads invalid address.
         * 5. Exchange a single byte, which is a checksum.
         * 6. Go to step 2 and repeat.
         * 
         * For a single download, you'd send a length of zero and any address the
         * second time around. The checksum is generated by adding each byte to
         * an 8-bit accumulator. (so it wraps from $FF to $00)
        **/
        case(0x01):
            comms_exc_long(0);

            address = comms_exc_long(0);
            // LINKDIAG_LOGOUT(" ADR:0x%08X", address);

            length  = comms_exc_long(0);
            // LINKDIAG_LOGOUT(" LEN:0x%08X", length);

            while(length != 0)
            {
                if((length == 2) || (length == 4))
                { /* 1,2,4 bytes length data access. */
                    if(length == 2)
                    {
                        unsigned short* ptr_in2  = (unsigned short*)(address & 0xFFFFFFFE);
                        unsigned short* ptr_out2 = (unsigned short*)&rw4_buffer;
                        *ptr_out2 = *ptr_in2;
                    }
                    else if(length == 4)
                    {
                        unsigned long* ptr_in4  = (unsigned long*)(address & 0xFFFFFFFC);
                        unsigned long* ptr_out4 = (unsigned long*)&rw4_buffer;
                        *ptr_out4 = *ptr_in4;
                    }
                    ptr = (unsigned char*)&rw4_buffer;
                }
                else
                { /* Byte-per-byte register access. */
                    ptr = (unsigned char*)address;
                }

                checksum = 0;
                for(i=0; i<length; i++)
                {
                    comms_exc_byte(ptr[i]);
                    checksum += ptr[i];
                }
                comms_exc_byte(checksum);

                stat->snd_count += length;

                address = comms_exc_long(0);
                length  = comms_exc_long(0);
            }
            /* Finished download. */
            comms_exc_byte('O');
            comms_exc_byte('K');
            break;


        /** From http://cgfm2.emuviews.com/sat/satar.txt :
         * Function $08: Write byte to memory address
         * 
         * Reads a longword from the PC (address to write to) and exchanges zero,
         * using the value it receives to write as a byte value to the specified
         * address.
        **/
        case(0x08):
            address = comms_exc_long(0);
            tmp     = comms_exc_byte(0);
            ((unsigned char*)address)[0] = tmp;

            stat->rcv_count += 1;
            break;


        /** From http://cgfm2.emuviews.com/sat/satar.txt :
         * Function $09: Upload (and execute) data
         * 
         * 1. Read longword from the PC (address to load data)
         * 2. Read longword from the PC(data length)
         * 3. Exchange byte (if $01, JSR to load address after loading, else do nothing)
         * 4. Read data bytes from PC until all data was transferred
         * 
         * When transferring data, the first byte exchanged is the value of SH-2
         * register R9, and all subsequent bytes are the previous byte sent to the PC.
         * 
         * Assuming the stack isn't modified, the program called by this function
         * could end with an RTS and return to the PAR software.
        **/
        case(0x09):
            address   = comms_exc_long(0);
            // LINKDIAG_LOGOUT(" ADR:0x%08X", address);
            length    = comms_exc_long(0);
            // LINKDIAG_LOGOUT(" LEN:0x%08X", length);
            exec_flag = comms_exc_byte(0);
            // LINKDIAG_LOGOUT(" EXE:%d", exec_flag);

            if((length == 2) || (length == 4))
            { /* 1,2,4 bytes length data access. */

                /* Receive data to write. */
                ptr = (unsigned char*)&rw4_buffer;
                for(i=0; i<length; i++)
                {
                    ptr[i] = comms_exc_byte((i==0 ? 0 : ptr[i-1]));
                }

                if(length == 2)
                {
                    unsigned short* ptr_in2  = (unsigned short*)ptr;
                    unsigned short* ptr_out2 = (unsigned short*)(address & 0xFFFFFFFE);
                    *ptr_out2 = *ptr_in2;
                }
                else if(length == 4)
                {
                    unsigned long* ptr_in4  = (unsigned long*)ptr;
                    unsigned long* ptr_out4 = (unsigned long*)(address & 0xFFFFFFFC);
                    *ptr_out4 = *ptr_in4;
                }
            }
            else
            {
                /* Receive data to write. */
                ptr = (unsigned char*)address;
                for(i=0; i<length; i++)
                {
                    ptr[i] = comms_exc_byte((i==0 ? 0 : ptr[i-1]));
                }
            }

            /* Execute ? */
            if(exec_flag == 1)
            {
                sc_execute(SC_SOFTRES_ALL, address/*tmp_address*/, length/*length*/, address/*exec_address*/);
            }

            stat->rcv_count += length;
            break;


        /* Unknown/unimplemented function. */
        default:
            break;
        }

    } /* Link type loop. */
}

