/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "credits.h"

#include <sc_saturn.h>

/* Declaration of menu shared variables.
 * And declaration of function to init satcom library.
 */
#include "main.h"

#include "vdp.h"
#include "conio.h"
#include "starfield.h"
#include "logo.h"

int atoi(const char *nptr);

/* Uncomment the line below when doing credits related debug. */
//#define CREDITS_DEBUG_ENABLE


/* Maximum line count for credits display area. */
#define CREDITS_MAXCOLS (64)
#define CREDITS_MAXROWS (23)


/**
 * Credits string syntax help.
 *
 * Line is set by :
 *  - Start character : '|'
 *  - Line comment start character : '#'
 *  - Command string
 *  - String delimiter characters
 *     - '|' : bump line ID
 *     - '~' : stay on same line
 *     - '@' : return to first line
 *  - Credits string itself
 *
 * About credits execution order :
 *  - Execute command(s).
 *  - Display credits string.
 *  - Wait several frame(s).
 *
 * Notes :
 *  - There's no start character required for first line.
 *    (Hence it can't be commented)
 *  - Muliple commands is possible, they need to be
 *    separated with ',' character.
 * 
 * How to :
 *  - Comment out a line :
 *    Set its start character to '#'.
 *    Note : first line can't be commented.
 *  - Set a "commands only" line :
 *    Add "sk" command in its command list.
 *
 * Command string syntax :
 *  - Change colors           : c[<bgcolor>][<fgcolor>]
 *  - Change display delay    : d[<cnt>] (delay between two lines, in frame count)
 *  - One shot wait           : wt[<cnt>] (frame count unit)
 *  - Clear screen            : cls
 *  - Starfield count         : sf_cnt[<cnt>]
 *  - Clear display area      : clr
 *  - Change start line       : sl[<line_id>] (within display area)
 *  - Text justify            : tj[l|c|r] (left, center, right)
 *  - Change font             : cf[<font ID or name>]
 *  - Skip line display       : sk
 *  - No wait until next line : nw
 *  - Set display pattern     : bg_pat[<pattern>]
 *  - Change background image : bg_img[<name>]              (SGM build only)
 *  - Specify DZ2SGM region   : dz_one, dz_stt, dz_end      (SGM build only)
 *  - Set V display area      : vda[<line_stt>][<line_cnt>]
 *  - Set H display area      : hda[<column_stt>][<column_cnt>]
 *  - Draw disp. area rect    : dr
 *  - Auto scroll             : as[0|1]
 * 
 * About line clearing rules :
 *  Full line is cleared when text is centered.
 *  Full line is NOT cleared when text left or right justified.
 *
 * About line ID :
 *  Line ID starts from zero, first line on top of
 *  credits display area, to 22, last line on bottom
 *  of credits display area.
 *
 * Color syntax :
 * - Transparent : "T"
 * - Black : "B", White : "W"
 * - Red : "r", Blue : "b", Green : "g"
 * - Yellow : "y"
 * - Cursor (blinking) : "C"
 * - Other : color ID in palette
**/


#ifdef MAKE_BUPMNGR

int _credits_dz2sgm_mode = 0;
void credits_set_dz2sgm_mode(int mode)
{
    _credits_dz2sgm_mode = mode;
}

#endif // MAKE_BUPMNGR


/* Logo swap stuff.
 * This is currently used only in Save Data Manager, because
 * Pseudo Saturn Kai lite doesn't have enought room in Flash
 * ROM for storing more than one logo image.
 * Pseudo Saturn Kai full edition may store extra logo in
 * flash ROM (for example in RomFs) but this is out the scope
 * of current release.
 */
#ifdef MAKE_BUPMNGR

    /* The logo module, for ... sawpping logos. */
#   include "logo.h"

    /* Location where to store next logo.
     * It is stored in decompressed shape, ready for
     * being blitted to VDP1 RAM.
     *
     * Usage example is for use to do transition with
     * previous image.
     * When blitting the whole image in one pass, then 
     * `logo_load' function is used without transiting
     * via this buffer.
     *
     * Just in case of, its size is designed to handle
     * either raw logo or compressed logo.
     * Compressed logo isn't currently used, but may
     * be necessary to chache in the case compressed
     * image have to be read from CD-ROM or SD card.
     *
     * Note : _extract_buff length is 331KB, in order to
     * store both compressed and uncompressed images for
     * Dezaemon 2 screenshots gallery.
     * Dimensions are 320*224, RGB555, and both buffers
     * have 512 bytes for handling compression header.
     * More information available in deza_screens.h .
     *
     * _extract_buff currently handles decompressed logo
     * and credits script (MAX 48 KB), so the remaining
     * around 170KB is free for any use in future revisions.
     */
#   include "BackupMngr/bup_menus.h"
#   define CRELOGO_DECOMP_ADR ((char*)_extract_buff)
#   define CRELOGO_DECOMP_LEN (LOGO_DATALEN + MAX(SC_COMP_HEADER_MAXSIZE, 512))


    /* Credits logo cache : this is used to store each
     * images used when displaying credits.
     * When converting each images on PC side, pre-tiling
     * and then compression (with LZF algorithm) is done, 
     * and this cache consequently contains compressed data.
     *
     * Cache is filled when entering credits page, so the
     * smaller cache the better since when using CD-ROM
     * access speed is ~150KB/s, and drops to 50~80KB/s
     * when using Gamer's Cartridge SD card.
     * Additionally, last bytes of LRAM (where this cache
     * is located) is used for SatCom debugger global data
     * which needs to be left as-is.
     * As a consequence, cache size is artificially limited
     * to 700KB, and the remaining is currently unused hence
     * available for any future extension.
     *
     * Also, a backup logo image when entering credits is kept
     * so that it doesn't have to be reloaded when exiting credits.
     * This saves loading of file from SD card when custom logo is used.
     * (This logo backup is also done in Dezaemon 2 screenshots Gallery)
     *
     *  00200000 - 00222FFF [  140KB] : (1) Pre-credits logo image (320 x 224, RGB555, tiled)
     *  00223000 - 002D1FFF [  700KB] : credits logo cache (LZF images blob).
     *  002D2000 - 002F4FFF [  140KB] : Image cache  - for any eventual future extension.
     *  002F5000 - 002FFBFF [   43KB] : Unused cache - for any eventual future extension.
     *  002FFC00 - 002FFFF7 [ 1016 B] : Reserved for future use.
     *  002FFFF8 - 002FFFFF [    8 B] : Debugger global data (pointer, then magic number).
     *
     * (1) This is a copy of logo image data from VDP1 RAM when entering
     *     credits screen, so that image can be restored back without
     *     having to read it from RAM or SD card.
     */
#   define CRELOGO_BACKUP_PTR  ((unsigned char*)TEMP_BUFF_ADR)
//#   define CRELOGO_BACKUP_LEN  (LOGO_WIDTH * LOGO_HEIGHT * sizeof(unsigned short))
#   define CRELOGO_BACKUP_LEN  LOGO_DATALEN

#   define CRELOGO_CACHE_ADR  (TEMP_BUFF_ADR + CRELOGO_BACKUP_LEN)
#   define CRELOGO_CACHE_LEN  (700 * 1024)

    /* Unused areas : definition stubs. */
//#   define CRELOGO_UNUSED_IMAGE_ADR (CRELOGO_CACHE_ADR + CRELOGO_CACHE_LEN)
//#   define CRELOGO_UNUSED_IMAGE_LEN  LOGO_DATALEN
//#   define CRELOGO_UNUSED_EXTRA_ADR (CRELOGO_UNUSED_IMAGE_ADR + CRELOGO_UNUSED_IMAGE_LEN)
//#   define CRELOGO_UNUSED_EXTRA_LEN  (43*1024)

    /* Logo parameters structure, for use by logo module
     * when loading compressed logo directly to VDP1 RAM.
     *
     * Example with credits logo with appropriate load parameters
     * set beforehand :
     *  logo_load(&_crelogo_load_params);
     *
     * Another example with default logo by Alien :
     *  logo_load(_logo_load_params+LOGO_DEFAULT_LOAD_PARAMS);
     */
    logo_load_params_t _crelogo_load_params;
#endif // MAKE_BUPMNGR


/* Inclusion of extra background image. */
#ifdef MAKE_PSKAI

#   include "logo.h"

#   define CRELOGO_DECOMP_ADR ((char*)0x06040000)

    logo_load_params_t _crelogo_load_params;
#endif // MAKE_BUPMNGR



/* Inclusion of compressed credits, and also
 * inclusion of buffer where to decompress them.
 */
#ifdef MAKE_PSKAI
#   include "binwrap.h"
#   define CREDITS_DECOMP_LOC ((char*)0x06004000)
#endif
#ifdef MAKE_BUPMNGR
#   include "BackupMngr/binwrap.h"
#   define CREDITS_DECOMP_LOC (((char*)_extract_buff) + CRELOGO_DECOMP_LEN)
#endif



#ifdef MAKE_BUPMNGR

/**
 * crelogo_load_cache
 *
 * Load background images cache from CD-ROM or SD card.
 * This is done only when DZ2SGM mode is enabled, but may be extended
 *  to other usages in the future :
 *   - Save Data Manager : load cache from CD-ROM/SD card, from other
                           file name than DZ2SGM.
 *   - Pseudo Saturn Kai : load cache RomFs = full version only.
**/
void crelogo_load_cache(void)
{
    if(!_credits_dz2sgm_mode)
    {
        return;
    }

    /* Read RomFs file containing background images for credits page. */
    unsigned char dev_id = deza_sgm_dev_id();
    unsigned long datalen = 0;
    char* path = SGM_BASE_PATH "DATA/CREDITS.RFS";

    datalen = scf_read(dev_id, path, 0/*offset*/, CRELOGO_CACHE_LEN, (void*)CRELOGO_CACHE_ADR);

    int romfs_mount = 0;
    if(datalen != 0)
    {
        romfs_mount = sc_romfs_setdata((unsigned char*)CRELOGO_CACHE_ADR);
    }

#ifdef CREDITS_DEBUG_ENABLE
    scd_msgout(6, "crelogo_load_cache datalen:%u romfs_mount:%u", datalen, romfs_mount);
#endif // CREDITS_DEBUG_ENABLE

    if((datalen == 0) || (romfs_mount == 0))
    {
        /* Error : couldn't load or mount cache. */
        memset((void*)CRELOGO_CACHE_ADR, 0, CRELOGO_CACHE_LEN);
        return;
    }
}


/**
 * crelogo_load_image
 *
 * Load a background image from cache.
**/
void crelogo_load_image(char* name)
{
    if(name[0] == '\0')
    {
        /* Special case when logo name is 
         * empty : load Pseudo Saturn Kai logo.
         */
        logo_load(_logo_load_params+LOGO_DEFAULT_LOAD_PARAMS);
    }
    else
    {
        unsigned long datalen;

        datalen = scf_size(SC_FILE_ROMFS, name);

#ifdef CREDITS_DEBUG_ENABLE
        scd_msgout(5, "load_image datalen:%u %s", datalen, name);
#endif // CREDITS_DEBUG_ENABLE

        if(datalen == LOGO_DATALEN)
        {
            /* Read only when a valid file was found. */
            scf_read(SC_FILE_ROMFS, name, 0/*offset*/, datalen, (void*)CRELOGO_DECOMP_ADR);

            /* Image decompressed from RomFs to RAM : load it as a a logo image. */
            memset(&_crelogo_load_params, 0, sizeof(logo_load_params_t));

            _crelogo_load_params.data       = (unsigned char*)CRELOGO_DECOMP_ADR;
            _crelogo_load_params.datalen    = datalen;
            _crelogo_load_params.compressed = 0;
            logo_load(&_crelogo_load_params);

#ifdef CREDITS_DEBUG_ENABLE
            scd_msgout(4, "load_image ended");
#endif // CREDITS_DEBUG_ENABLE
        }
    }
}


#endif // MAKE_BUPMNGR



#ifdef MAKE_PSKAI
    /* Definition of pre-credits logo image (320 x 224, RGB555, tiled)
     * This is similar to the definition for Save Data Manager, but
     * adapted to comply with Pseudo Saturn Kai memory assignment.
     * 
     *  06004000 - 0600FFFF [   48KB] : Decompressed credits script
     *  06010000 - 06032FFF [  140KB] : Pre-credits logo image (320 x 224, RGB555, tiled)
     */
#   define CRELOGO_BACKUP_PTR  ((unsigned char*)0x06010000)
#   define CRELOGO_BACKUP_LEN  LOGO_DATALEN


/**
 * crelogo_load_image
 *
 * Load a background image from cache.
**/
void crelogo_load_image(char* name)
{
    if(name[0] == '\0')
    {
        /* Special case when logo name is 
         * empty : load Pseudo Saturn Kai logo.
         */
        logo_load(_logo_load_params+LOGO_DEFAULT_LOAD_PARAMS);
    }
    else
    {
        memset(&_crelogo_load_params, 0, sizeof(logo_load_params_t));

        _crelogo_load_params.data = (unsigned char*)(&_logo_byuu_dat);
        _crelogo_load_params.datalen = (unsigned long )(&_logo_byuu_end - &_logo_byuu_dat);
        _crelogo_load_params.compressed = 1;

        logo_load(&_crelogo_load_params);
    }
}

#endif // MAKE_PSKAI






char* get_single_command(char* command, char* buff)
{
    char* ptr = command;

    *buff = '\0';
    while((*ptr) != '\0')
    {
        char c = *ptr;

        /* Command delimiter ? */
        if(c == ',')
        {
            ptr++;
            return ptr;
        }

        /* Store current character in output buffer. */
        *buff++ = c;
        *buff = '\0';

        ptr++;
    }

    /* End of string ? Indicate there isn't more command. */
    return NULL;
}

/**
 * get_command_parameter
 *
 * Example :
 *  get_command_parameter("c1[123][456]", 1);
 * Will return "456".
 *
 * Note #1 : global buffer is used, so mutiple usage at once is forbidden.
 * Note #2 : parameter string too long will explode your SH2 CPU.
**/
char _cmd_param_buff[32];
char* get_command_parameter(char* cmd, int param_id)
{
    char* ptr = _cmd_param_buff;
    int i;
    char c;

    memset((void*)ptr, '\0', sizeof(_cmd_param_buff));

    param_id++;
    i = 0;
    do
    {
        c = *cmd;

        /* Update parameter ID
         * Note : garbage between 2 parameters isn't supported :
         * First parameter of "[123]456[789]" will return "123456".
         */
        if(c == '[')
        {
            i++;
        }
        else if(c == ']')
        {
            if(i == param_id)
            {
                break;
            }
        }

        /* If needed, copy current character
         * to output buffer.
         */
        if((c != '[') && (i == param_id))
        {
            *ptr++ = c;
        }

        cmd++;
    } while(c != '\0');

    return _cmd_param_buff;
}


/**
 * get_color_id
 *
 * Return color ID (in palette) from parameter string.
**/
char get_color_id(char* str)
{
    unsigned char col;

    switch(str[0])
    {
    case('T'):
        col = COLOR_TRANSPARENT;
        break;
    case('B'):
        col = COLOR_BLACK;
        break;
    case('W'):
        col = COLOR_WHITE;
        break;
    case('r'):
        col = COLOR_RED;
        break;
    case('b'):
        col = COLOR_BLUE;
        break;
    case('g'):
        col = COLOR_GREEN;
        break;
    case('y'):
        col = COLOR_YELLOW;
        break;
    case('C'):
        col = 15;
        break;
    default:
        col = atoi(str);
    }

    return col;
}


/**
 * credits_clearscreen
 *
 * Clear screen.
 * And, erase rectangle around display area.
 *
**/
void credits_clearscreen(void)
{
    int i;

    for(i=3; i<27; i++)
    {
        conio_printf(2, i, COLOR_TRANSPARENT, empty_line);
    }

    conio_fillrect(
            (2)              * CONIO_CHAR_W  - 1, 
            (3)              * CONIO_CHAR_H  - 0, 
            ((LINE_LEN-1)    * CONIO_CHAR_W) + 2, 
            (CREDITS_MAXROWS * CONIO_CHAR_H) + 1, 
            COLOR_TRANSPARENT);
}



//void credits(void)
//{
//    pad_t* pad = NULL;
//    int x=130;
//    int y=120;
//    int radius=240;
//    int shadow_update;
//
//    /* Few stars and animate shadow during credits. */
//    sf_set_star_count(10);
//    logo_set_pattern(LOGO_PATTERN_SHADOW);
//
//    /* Hide application title. */
//    conio_display_enable(0/*enable*/);
//
//    while(1)
//    {
//        sc_check();
//
//        pad = pad_read();
//
//        shadow_update = 0;
//        if(pad->delta[0] & PAD_A)
//        {
//            logo_shadow(50, 100, 200, 100, 200);
//        }
//        else if(pad->delta[0] & PAD_B)
//        {
//            logo_shadow_disable();
//        }
//        else if(pad->delta[0] & PAD_C)
//        {
//            logo_shadow(50, 100, 250, 200, 280);
//        }
//        else if(pad->delta[0] & PAD_START)
//        {
//            break;
//        }
//        else if(pad->delta[0] & PAD_UP)
//        {
//            y-=3;
//            shadow_update = 1;
//        }
//        else if(pad->delta[0] & PAD_DOWN)
//        {
//            y+=3;
//            shadow_update = 1;
//        }
//        else if(pad->delta[0] & PAD_LEFT)
//        {
//            x-=3;
//            shadow_update = 1;
//        }
//        else if(pad->delta[0] & PAD_RIGHT)
//        {
//            x+=3;
//            shadow_update = 1;
//        }
//        else if(pad->delta[0] & PAD_LBUTTON)
//        {
//            if(radius >= 20)
//            {
//                radius-=10;
//                shadow_update = 1;
//            }
//        }
//        else if(pad->delta[0] & PAD_RBUTTON)
//        {
//            radius+=10;
//            shadow_update = 1;
//        }
//
//        if(shadow_update)
//        {
//            scd_logout("x/y=%d/%d, r=%d", x, y, radius);
//            logo_shadow(x-100, y, x+100, y, radius);
//        }
//
//        logo_animate();
//        logo_draw();
//
//        sf_animate();
//        sf_draw();
//
//        sf_stop();
//
//        sc_wait_vblank();
//    }
//
//
//    // conio_printf(2,  6, COLOR_WHITE, "Copyright 2011-2015 Pseudo Saturn Team");
//    // conio_printf(2,  7, COLOR_WHITE, "http://github.com/cyberwarriorx/pseudosaturn");
//
//
//
//    // wait_for_start("Press Start to go back");
//
//    /* Restore back application title. */
//    conio_display_enable(1/*enable*/);
//
//    /* Restore logo/starfield settings. */
//    logo_set_pattern(LOGO_PATTERN_NOMESH);
//    sf_set_star_count(STARS_MAXCOUNT);
//}
void credits(Fct_menu_callback_vblank callback_vblank)
{
    int i;
    pad_t* pad;

    /* Text color, font ID, etc. */
    unsigned char bgcolor = COLOR_TRANSPARENT;
    unsigned char fgcolor = COLOR_WHITE;
    int cnt, waitcnt;
#define TEXT_JUSTIFY_LEFT   0
#define TEXT_JUSTIFY_CENTER 1
#define TEXT_JUSTIFY_RIGHT  2
    int text_justify = TEXT_JUSTIFY_CENTER;
    int font_id = FONT_DEFAULT;
    int vda_stt = 0;
    int vda_cnt = CREDITS_MAXROWS;
    int hda_stt = 0;
    int hda_cnt = LINE_LEN-1;
    int autoscroll = 0;

    /* Menu settings, used ... for highlight color update. */
    menu_settings_t* mset = &_menu_settings;
    menu_reset_settings(mset);

    /* Setup current credits line and decompress credits text there.
     * Also, character just after last one is voided in order to
     * loop to begin of text.
     */
    unsigned char* credits_comp_ptr = &_credits_dat;
    unsigned long  credits_comp_len = (unsigned long )(&_credits_end - &_credits_dat);

    /* Declaration of credits decompression buffer. */
    char* credits_start_line = CREDITS_DECOMP_LOC;
    char* line = credits_start_line;
    unsigned long credits_len;

    credits_len = decomp_exec(credits_comp_ptr, credits_comp_len, (unsigned char*)credits_start_line);
    credits_start_line[credits_len] = '\0';

    char command_buff[256];
    char cmd[32];
    char disp_buff[CREDITS_MAXCOLS+1];

    /* Few stars during credits. */
    sf_set_star_count(20);

#if 0
    /* Animate shadow during credits. */
    logo_set_pattern(LOGO_PATTERN_SHADOW);
#else
    /* Greyout logo display so that text is visible. */
    logo_set_pattern(LOGOPAT_GREYOUT_SLOW);
#endif

    /* Hide application title (makes sense when using VDP1 font). */
    //conio_display_enable(0/*enable*/);

    /* Backup regular (Pseudo Saturn Kai or user-set)
     * logo so that we can play with VDP1 after that.
     */
    memcpy(CRELOGO_BACKUP_PTR, LOGO_DATAPTR, CRELOGO_BACKUP_LEN);

    /* Force logo display to Pseudo Saturn Kai one */
    logo_load(_logo_load_params+LOGO_DEFAULT_LOAD_PARAMS);
    logo_set_pattern(LOGOPAT_FORCE_GREYOUT);

#ifdef MAKE_BUPMNGR
    /* Load background images from CD-ROM or SD card. */
    crelogo_load_cache();
#endif // MAKE_BUPMNGR

    waitcnt = 10;
    cnt = waitcnt;
    int waitcnt_temp = -1;
    int line_id = -1;

    /* Enabled when between "dz_stt" and "dz_end" keywords. */
    int dz2sgm_region = 0;

    while(1)
    {
        int dz2sgm_skip_line = dz2sgm_region;

        if(dz2sgm_skip_line)
        {
            cnt = 0;
        }
        else
        {
            pad = pad_read();

            if(pad->new[0] == (PAD_A|PAD_B|PAD_C|PAD_START))
            {
                /* Quit program ? */
                sc_exit();
            }

            /* Exit credits page with A/B/C/Start buttons. */
            if(pad->delta[0] & (PAD_A|PAD_B|PAD_C|PAD_START))
            {
                break;
            }

            /* Faster display with L+R shoulder buttons (debug purpose). */
            if((pad->new[0] & (PAD_LBUTTON|PAD_RBUTTON)) == (PAD_LBUTTON|PAD_RBUTTON))
            {
                cnt = 0;
            }
        }

        if(cnt <= 0)
        {
            int disp_row_incr = 1;
            int prev_line_id = line_id;
            int nowait = 0;
            char* ptr;

            /* Extract command string from current line. */
            memset((void*)command_buff, '\0', sizeof(command_buff));
            ptr = line;
            for(i=0; i<sizeof(command_buff); i++)
            {
                char c = ptr[i];
                if(c == '|')
                {
                    /* Goto next line. */
                    disp_row_incr = 1;
                    break;
                }
                else if(c == '~')
                {
                    /* Stay on same line. */
                    disp_row_incr = 0;
                    break;
                }
                else if(c == '@')
                {
                    /* Return to first line. */
                    disp_row_incr = vda_stt - line_id;
                    break;
                }
                command_buff[i] = c;
            }
            command_buff[sizeof(command_buff)-1] = '\0';

#ifdef CREDITS_DEBUG_ENABLE
            scd_msgout(7, "disp_row_incr:%d", disp_row_incr);
#endif // CREDITS_DEBUG_ENABLE

            /* Bump line ID. */
            line_id += disp_row_incr;

            /* Extract credits string from current line. */
            memset((void*)disp_buff, '\0', sizeof(disp_buff));
            ptr = line+i+1;
            for(i=0; i<sizeof(disp_buff); i++)
            {
                char c = ptr[i];
                if((c == '\0') || (c == '\r') || (c == '\n'))
                {
                    break;
                }
                disp_buff[i] = c;
            }
            disp_buff[sizeof(disp_buff)-1] = '\0';
            if(disp_buff[0] == '\0')
            {
                disp_buff[0] = ' ';
                disp_buff[1] = '\0';
            }

#ifdef CREDITS_DEBUG_ENABLE
            scd_logout("disp_buff=\"%s\"", disp_buff);
#endif // CREDITS_DEBUG_ENABLE

            /* Jump to next line. */
            int line_start = 0;
            while(1)
            {
                char c = *line;
                line++;

                if(c == '\0')
                {
                    /* End of credits script : loop back to start. */
                    line = credits_start_line;
                    break;
                }

                if((c == '\r')
                || (c == '\n'))
                {
                    line_start = 1;
                }
                else
                {
                    if(line_start)
                    {
                        if(c == '|')
                        {
                            /* Found next line to display. */
                            c = *line;
                            if(c == '\0')
                            {
                                /* End of credits script : loop back to start. */
                                line = credits_start_line;
                            }
                            break;
                        }

                        line_start = 0;
                    }
                }
            }

            /* Parse command(s). */
            ptr = command_buff;
            do
            {
                ptr = get_single_command(ptr, cmd);
#ifdef CREDITS_DEBUG_ENABLE
                if(ptr != NULL)
                {
                    scd_logout(" cmd=\"%s\", ptr=\"%s\" (0x%08X)", cmd, ptr, ptr);
                }
#endif // CREDITS_DEBUG_ENABLE

                if(memcmp((void*)cmd, (void*)"c[", 2) == 0)
                { /* Change colors           : c[<bgcolor>][<fgcolor>] */
                    if(!dz2sgm_skip_line)
                    {
                        bgcolor = get_color_id(get_command_parameter(cmd, 0));
                        fgcolor = get_color_id(get_command_parameter(cmd, 1));
                    }
                }
                else if(memcmp((void*)cmd, (void*)"d[", 2) == 0)
                { /* Change display delay    : d[<cnt>] */
                    if(!dz2sgm_skip_line)
                    {
                        waitcnt = atoi(get_command_parameter(cmd, 0));
                    }
                }
                else if(memcmp((void*)cmd, (void*)"wt[", 3) == 0)
                { /* Change display delay    : d[<cnt>] */
                    if(!dz2sgm_skip_line)
                    {
                        waitcnt_temp = atoi(get_command_parameter(cmd, 0));
                    }
                }
                else if(memcmp((void*)cmd, (void*)"cls", 3) == 0)
                { /* Clear screen            : cls */
                    if(!dz2sgm_skip_line)
                    {
                        credits_clearscreen();
                        line_id = vda_stt;
                    }
                }
                else if(memcmp((void*)cmd, (void*)"sf_cnt[", 4) == 0)
                { /* Starfield count         : sf_cnt[<cnt>] */
                    if(!dz2sgm_skip_line)
                    {
                        int cnt = atoi(get_command_parameter(cmd, 0));
                        if(cnt > STARS_MAXCOUNT)
                        {
                            cnt = STARS_MAXCOUNT;
                        }
                        sf_set_star_count(cnt);
                    }
                }
                else if(memcmp((void*)cmd, (void*)"clr", 3) == 0)
                { /* Clear display area      : clr */
                    if(!dz2sgm_skip_line)
                    {
                        /* Erase display area.
                         * Keep first line, reserved for main title.
                         * Also, keep border too.
                         */
                        conio_fillrect(
                                (2+hda_stt  )*CONIO_CHAR_W, 
                                (3+vda_stt+1)*CONIO_CHAR_H, 
                                (hda_cnt    *CONIO_CHAR_W), 
                                ((vda_cnt-1)*CONIO_CHAR_H), 
                                COLOR_TRANSPARENT);

                        line_id = vda_stt+1;
                    }
                }
                else if(memcmp((void*)cmd, (void*)"sl[", 3) == 0)
                { /* Change start line       : sl[<line_id>] */
                    if(!dz2sgm_skip_line)
                    {
                        line_id = vda_stt + atoi(get_command_parameter(cmd, 0)) - 1;
                        prev_line_id = line_id;
                        disp_row_incr = 0;
                    }
                }
                else if(memcmp((void*)cmd, (void*)"tj[", 3) == 0)
                { /* Text justify            : tj[l|c|r] */
                    if(!dz2sgm_skip_line)
                    {
                        char* tj = get_command_parameter(cmd, 0);

                        if(tj[0] == 'c')
                        {
                            text_justify = TEXT_JUSTIFY_CENTER;
                        }
                        else if(tj[0] == 'r')
                        {
                            text_justify = TEXT_JUSTIFY_RIGHT;
                        }
                        else //if(text_justify[0] == 'l')
                        {
                            text_justify = TEXT_JUSTIFY_LEFT;
                        }
                    }
                }
                else if(memcmp((void*)cmd, (void*)"cf[", 3) == 0)
                { /* Change font ID. */
                    if(!dz2sgm_skip_line)
                    {
                        char* font_str = get_command_parameter(cmd, 0);

                        if((strcmp(font_str, "amiga12") == 0)
                        || (strcmp(font_str, "A") == 0))
                        {
                            font_id = FONT_AMIGA12;
                        }
                        else if((strcmp(font_str, "harmonica") == 0)
                             || (strcmp(font_str, "H") == 0))
                        {
                            font_id = FONT_HARMONICA;
                        }
                        else if((strcmp(font_str, "microknight") == 0)
                             || (strcmp(font_str, "M") == 0))
                        {
                            font_id = FONT_MICROKNIGHT;
                        }
                        else if((strcmp(font_str, "mosoul") == 0)
                             || (strcmp(font_str, "M") == 0))
                        {
                            font_id = FONT_MOSOUL;
                        }
                        else if((strcmp(font_str, "oldschool") == 0)
                             || (strcmp(font_str, "O") == 0))
                        {
                            font_id = FONTID_SET_PROP(FONT_TANDY);
                        }
                        else
                        {
                            font_id = atoi(font_str);
                        }
                    }
                }
                else if(memcmp((void*)cmd, (void*)"sk", 2) == 0)
                { /* Skip line display       : sk */
                    memset((void*)disp_buff, '\0', sizeof(disp_buff));
                    line_id = prev_line_id;
                    disp_row_incr = 0;
                }
                else if(memcmp((void*)cmd, (void*)"nw", 2) == 0)
                { /* No wait until next line : nw */
                    nowait = 1;
                }
                else if(memcmp((void*)cmd, (void*)"bg_pat[", 7) == 0)
                { /* Set display pattern : bg_pat[<pattern>] */
                    if(!dz2sgm_skip_line)
                    {
                        unsigned long pattern = strtoul(get_command_parameter(cmd, 0), NULL, 16);
                        logo_set_pattern(pattern);
                    }
                }
                else if(memcmp((void*)cmd, (void*)"bg_img[", 7) == 0)
                { /* Change background image : bg_img[<name>] */
                    if(!dz2sgm_skip_line)
                    {
                        char* img_name = get_command_parameter(cmd, 0);
                        crelogo_load_image(img_name);
                    }
                }
#ifdef MAKE_BUPMNGR
                else if(memcmp((void*)cmd, (void*)"dz_", 3) == 0)
                { /* Specify DZ2SGM region : dz_one, dz_stt, dz_end */
                    int dz2sgm_on = 0;

                    if(memcmp((void*)(cmd+3), (void*)"one", 3) == 0)
                    {
                        dz2sgm_on = 1;
                    }
                    else if(memcmp((void*)(cmd+3), (void*)"stt", 3) == 0)
                    {
                        dz2sgm_region = (_credits_dz2sgm_mode ? 0 : 1);
                    }
                    else if(memcmp((void*)(cmd+3), (void*)"end", 3) == 0)
                    {
                        dz2sgm_region = 0;
                    }

                    if((!_credits_dz2sgm_mode)
                    && (dz2sgm_on || dz2sgm_region))
                    {
                        dz2sgm_skip_line = 1;
                        break;
                    }
                }
#endif // MAKE_BUPMNGR
                else if(memcmp((void*)cmd, (void*)"vda[", 3) == 0)
                { /* Set V display area      : vda[<line_stt>][<line_cnt>] */
                    if(!dz2sgm_skip_line)
                    {
                        vda_stt = atoi(get_command_parameter(cmd, 0));
                        vda_cnt = atoi(get_command_parameter(cmd, 1));
                        line_id = vda_stt - 1;
                        prev_line_id = line_id;
                        disp_row_incr = 0;
                    }
                }
                else if(memcmp((void*)cmd, (void*)"hda[", 3) == 0)
                { /* Set H display area      : hda[<column_stt>][<column_cnt>] */
                    if(!dz2sgm_skip_line)
                    {
                        hda_stt = atoi(get_command_parameter(cmd, 0));
                        hda_cnt = atoi(get_command_parameter(cmd, 1));
                    }
                }
                else if(memcmp((void*)cmd, (void*)"dr", 2) == 0)
                { /* Draw disp. area rect    : dr */
                    if(!dz2sgm_skip_line)
                    {
                        conio_drawrect(
                                (2+hda_stt)*CONIO_CHAR_W - 1, 
                                (3+vda_stt)*CONIO_CHAR_H - 1, 
                                (hda_cnt * CONIO_CHAR_W) + 2, 
                                (vda_cnt * CONIO_CHAR_H) + 2, 
                                bgcolor);
                    }
                }
                else if(memcmp((void*)cmd, (void*)"as[", 3) == 0)
                { /* Auto scroll             : as[0|1] */
                    if(!dz2sgm_skip_line)
                    {
                        autoscroll = atoi(get_command_parameter(cmd, 0));
                    }
                }
            } while(ptr != NULL);


#ifdef MAKE_BUPMNGR
            /* Skip dezaemon 2 only contents when SGM mode
             * is not enabled.
             * Skipping is equivalent to "sk" and "nw" commands.
             */
            if(dz2sgm_skip_line)
            {
                memset((void*)disp_buff, '\0', sizeof(disp_buff));
                line_id = prev_line_id;
                disp_row_incr = 0;
                nowait = 1;
                cnt = 0;

                /* Completely skip the drawing and vblank stuff. */
                continue;
            }
#endif // MAKE_BUPMNGR

            /* Prepare frame counter for next line. */
            if(nowait)
            {
                cnt = 0;
            }
            else
            {
                if(waitcnt_temp >= 0)
                {
                    cnt = waitcnt_temp;
                }
                else
                {
                    cnt = waitcnt;
                }

                /* Reset oneshot wait counter value. */
                waitcnt_temp = -1;
            }

            if(!dz2sgm_skip_line)
            {
                /* Display credits current line. */
                WaitForVBLANKIn();
            }

            if(disp_buff[0] != '\0')
            {
#ifdef CREDITS_DEBUG_ENABLE
                scd_logout("DISPLAY");
                scd_logout("|autoscroll:%d", autoscroll);
                scd_logout("|disp_row_incr:%d", disp_row_incr);
                scd_logout("|line_id:%d", line_id);
                scd_logout("|vda_stt[%d] vda_cnt[%d]:%d", vda_stt, vda_cnt, vda_stt + vda_cnt);
#endif // CREDITS_DEBUG_ENABLE

                int scroll_text = 0;
                if(disp_row_incr == 1)
                {
                    if(line_id >= (vda_stt+vda_cnt))
                    {
                        if(autoscroll)
                        {
                            /* If auto scroll, stay on last line. */
                            line_id = vda_stt+vda_cnt-1;
                            scroll_text = 1;
                        }
                        else
                        {
                            /* If not auto scroll, return to top of display area. */
                            line_id = vda_stt;
                        }
                    }
                }

                /* Scroll text ? */
                if(scroll_text)
                {
                    int i;
                    for(i=2; i<vda_cnt; i++)
                    {
                        conio_scroll_up(3+vda_stt+i, (2+hda_stt)*CONIO_CHAR_W, hda_cnt*CONIO_CHAR_W);
                    }
                    int j;
                    for(j=0; j<hda_cnt; j++)
                    {
                        conio_printstr(
                            FONT_MOSOUL, FONT_CODEPAGE_ENGLISH, 
                            (2+hda_stt+j)*CONIO_CHAR_W, 
                            (3+line_id)*CONIO_CHAR_H, 
                            COLOR_BG_FG(bgcolor, bgcolor), " ");
                    }
                }

                int offset;
                if(text_justify == TEXT_JUSTIFY_LEFT)
                {
                    offset = 0;
                }
                else
                {
                    int len = conio_text_width(font_id, FONT_CODEPAGE_ENGLISH, disp_buff, -1/*msg_len*/);
                    offset = (hda_cnt*CONIO_CHAR_W - 1) - len;

                    if(text_justify == TEXT_JUSTIFY_CENTER)
                    {
                        offset = offset/2;

                        // /* Don't clear line when text justify is left or right.
                        //  * But, pre-clear during center justify, in order to prevent
                        //  * from previous letters remainings on screen.
                        //  */
                        // int j;
                        // for(j=0; j<hda_cnt; j++)
                        // {
                        //     conio_printstr(
                        //         FONT_MOSOUL, FONT_CODEPAGE_ENGLISH, 
                        //         (2+hda_stt+j)*CONIO_CHAR_W, 
                        //         (3+line_id)*CONIO_CHAR_H, 
                        //         COLOR_BG_FG(bgcolor, bgcolor), " ");
                        // }
                    }
                }

                /* If background color is not transparent, let's consider
                 * that whole width of display area needs to be filled
                 * with this color.
                 */
                if(bgcolor != COLOR_TRANSPARENT)
                {
                    int j;
                    for(j=0; j<hda_cnt; j++)
                    {
                        conio_printstr(
                            FONT_MOSOUL, FONT_CODEPAGE_ENGLISH, 
                            (2+hda_stt+j)*CONIO_CHAR_W, 
                            (3+line_id)*CONIO_CHAR_H, 
                            COLOR_BG_FG(bgcolor, bgcolor), " ");
                    }
                }

                conio_font_printf(font_id, FONT_CODEPAGE_ENGLISH, 
                        (2+hda_stt)*CONIO_CHAR_W + offset, 
                        (3+line_id)*CONIO_CHAR_H, 
                        COLOR_BG_FG(bgcolor, fgcolor), 
                        "%s", disp_buff);
            }
        }
        else
        {
            cnt--;

            if(!dz2sgm_skip_line)
            {
                WaitForVBLANKIn();
            }
        }

        if(!dz2sgm_skip_line)
        {
            /* Update highlight color. */
            menu_update_color_palette(mset);

            /* Display starfield, and call sc_check. */
            callback_vblank();

            WaitForVBLANKOut();
        }
    }

    /* Paranoia wipe of decompressed credits text.
     * This buffer is shared with other modules, so it is
     * good manner to leave it in same state as before using it.
     */
    memset(credits_start_line, 0, credits_len);

    /* Clear screen. */
    credits_clearscreen();

    /* Restore back application title. */
    conio_display_enable(1/*enable*/);

    /* Restore back Pseudo Saturn Kai logo data. */
    memcpy(LOGO_DATAPTR, CRELOGO_BACKUP_PTR, CRELOGO_BACKUP_LEN);

    /* Restore logo shadow settings. */
    logo_set_pattern(LOGO_DEFAULT_PATTERN);

    /* Restore starfield settings. */
    sf_set_star_count(STARS_MAXCOUNT);
}

