/**
 *  Sega Saturn Cartridge Firmware Flash Utility
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  Made from "Sega Saturn cartridge flash tool", by Anders Montonen.
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h> // For memset,memcpy

#include "vdp2.h"
#include "smpc.h"

#include "vdp.h"
#include "conio.h"

#include "menu.h"

#include "flash.h"

#include "../../satcom_lib/sc_compress.h"
#include "romwrap.h"

#include "../release.h"
#define LINE_LEN (36+1)

#include "../minipseudo/minipseudo_rc.h"


typedef struct _fw_entry
{
    /* Line displayed in menu. */
    char* desc;

    /* Firmware data and length. */
    unsigned char* data_stt;
    unsigned char* data_end;

    /* Text displayed under the menu. */
    char* detail1;
    char* detail2;
    char* detail3;

    /* Fill value for unused area after firmware. */
    unsigned char fill_val;

    /* 0:use firmware as-is, 1:create lite firmware. */
    unsigned char lite;
} fw_entry;

/* The firmware list. */
fw_entry _fw_list[] =
{
#ifdef PSKAI_LITE_BUILD
    { NULL, // "Pseudo Saturn Kai Lite        "
        (unsigned char*)&_fw_pskai_dat, 
        (unsigned char*)&_fw_pskai_end, 
        " Modified version of Pseudo Saturn.  ", 
        " For use with Action Replay, or      ", 
        " or USB Dev Cart.      PSKAI256.BIN  ", 
        0xFF, 1/*lite*/
    }, 
#else
    { NULL, // "Pseudo Saturn Kai             "
        (unsigned char*)&_fw_pskai_dat, 
        (unsigned char*)&_fw_pskai_end, 
        " Modified version of Pseudo Saturn.  ", 
        " For use with cartridge featuring   ", 
        " SD card adapter.         PSKAI.BIN ", 
        0xFF, 0/*lite*/
    }, 
#endif
    { "Cancel Firmware Update", 
        (unsigned char*)&_fw_pskai_dat, /* Dummy.     */
        (unsigned char*)&_fw_pskai_end, /* Dummy too. */
        "                                     ", 
        "                                     ", 
        "                                     ", 
        0xFF, 0/*lite*/
    }, 
    /* Terminating entry. */
    { "", NULL, NULL, "",  "", "", 0x00, 0}
};

/* Flash ROM size for lite firmware. */
#define LITE_FW_ROM_LEN (256*1024)

char* empty_line = "                                     ";

unsigned char char2pchar(unsigned char c)
{
    if(c == 0xFF) return 'X';
    if(c < ' ') return '.';
    if(c > 127) return '?';
    return c;
}

void flash_erase_callback(int page_cnt, int page)
{
    int p;

    if(page_cnt == 0)
    {
        p = 0;
        conio_printf(26, 13, COLOR_WHITE, "      ");
    }
    else
    {
        p = (100*page) / page_cnt;
        conio_printf(26, 13, COLOR_WHITE , "(   %%)");
        conio_printf(27, 13, COLOR_YELLOW, "%02d", p);
    }
}

/* Wait for pad keys release. */
void wait_keys_release(unsigned short key_mask)
{
    pad_t* pad;
    while(1)
    {
        pad = pad_read();
        if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START)) == 0)
        {
            break;
        }
        WaitForVBLANKIn();
        WaitForVBLANKOut();
    }
}

#define ERROR_WRTEST 1
#define ERROR_ERASE  2
#define ERROR_WRITE  3
void DisplayError(Romcheck_t* pCheckResults, int type)
{
    int i, j;
    pad_t* pad;

    /* Display error message. */
    if(type == ERROR_WRTEST)
    {
        /* Error message is already displayed a little above, 
         * and write test failure error screen is a bit crowded, 
         * so let's keep this line blank.
         * Okay, that's messy, but this isn't the first time, 
         * and won't be the last time too !
         */
        conio_printf(2, 17, COLOR_WHITE, empty_line);
    }
    else if(type == ERROR_ERASE)
    {
        conio_printf(2, 17, COLOR_ERROR, "    Erase verification failure !    ");
    }
    else
    {
        conio_printf(2, 17, COLOR_ERROR, "    Write verification failure !    ");
    }

    /* Display offset where error happened. */
    conio_printf(2, 18, COLOR_YELLOW, "Check Error Offset = 0x%08X", pCheckResults->ErrorOffset);

    /* Display hex dump near where error happened and stop everything. */
    conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

    wait_keys_release(PAD_START);
    while(1)
    {
        pad = pad_read();
        if(pad->delta[0] & PAD_START)
        {
            return;
        }

        /* Refresh dump display, in order to spot furtive memory error. */
        conio_printf(2, 19, COLOR_YELLOW, "Dump Start Address = 0x%08X", pCheckResults->pErrorData);
#define CHAR_PER_LINE 8
        for(j=0; j<6; j++)
        {
            /* Display start address. */
            conio_printf(2, 20+j, COLOR_YELLOW, "+%02X:", j*CHAR_PER_LINE);

            for(i=0; i<CHAR_PER_LINE; i++)
            {
                unsigned char c = pCheckResults->pErrorData[j*CHAR_PER_LINE+i];
                conio_printf(2+4+i*3  , 20+j, COLOR_WHITE, "%02X", c);
                conio_printf(2+4+8*3+i, 20+j, COLOR_WHITE, "%c"  , char2pchar(c));
            }
        }

        WaitForVBLANKIn();
        WaitForVBLANKOut();
    }
}

/**
 * DisplayFlashIds
 * Display vendor/device IDs.
 */
#define PROGRESS_STR_CNT 10
char* _progress_str[] = {
    "O    ", 
    ".O   ", 
    " .O  ", 
    "  .O ", 
    "   .O", 
    "    O", 
    "   O.", 
    "  O. ", 
    " O.  ", 
    "O.   "
};
int _progress_id = 0;
int _progress_divider = 0; /* In order to slow down the progress stuff. */
void DisplayFlashIds(FlashDevice_t* pDevice, unsigned short* pVendorID, unsigned short* pDeviceID)
{
    unsigned short color;

    /* Display progress, in order to know if Saturn is still running or not. */
    if((_progress_id < 0)
    || (_progress_id >= PROGRESS_STR_CNT))
    {
        _progress_divider = 0;
        _progress_id = 0;
    }
    else
    {
        _progress_divider++;
        if(_progress_divider >= 5)
        {
            _progress_divider = 0;
            _progress_id = (_progress_id+1) % PROGRESS_STR_CNT;
        }
    }
    conio_printf(2, 6, COLOR_WHITE, "Updating Flash Info [%s]", _progress_str[_progress_id]);

    conio_printf(2   , 4, COLOR_WHITE, "Flash Vendor ID:");
    conio_printf(2+22, 4, COLOR_WHITE, "-");
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    *pVendorID = FlashGetVendorID();
    conio_printf(19, 4, COLOR_YELLOW, "%04X", *pVendorID);
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    conio_printf(2   , 5, COLOR_WHITE, "Flash Device ID:");
    conio_printf(2+22, 5, COLOR_WHITE, "-");
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    *pDeviceID = FlashGetDeviceID();
    conio_printf(19, 5, COLOR_YELLOW, "%04X", *pDeviceID);
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    FillFlashDeviceInfo(*pVendorID, *pDeviceID, pDevice);
    SetFlashEraseCallback(flash_erase_callback, pDevice);

    /* Change color of vendor/device strings according to flashability. */
    switch(pDevice->Flashability)
    {
    default:
    case(FLASH_HOPELESS):
        color = COLOR_ERROR;
        break;

    case(FLASH_MAYBEOK):
        color = COLOR_YELLOW;
        break;

    case(FLASH_RECOGNIZED):
        color = COLOR_GREEN;
        break;
    }
    conio_printf(26, 4, color, (char*)pDevice->pVendorString);
    conio_printf(26, 5, color, (char*)pDevice->pDeviceString);
}

/**
 * ClearFlashIdsProgress
 */
void ClearFlashIdsProgress(void)
{
    conio_printf(2, 6, COLOR_WHITE, empty_line);
}


/* Flash memory related infos. */
FlashDevice_t  device;
unsigned short vendorID;
unsigned short deviceID;

menu_items_t    _menu_items;
menu_settings_t _menu_settings;
int menu_selected = 0;

void display_flash_ids(void)
{
    /* Flash memory : first get infos about flash chips. */
    DisplayFlashIds(&device, &vendorID, &deviceID);
}

void display_fw_infos(int id)
{
    /* Display additional informations about firmware. */
    conio_printf(2, 23, COLOR_WHITE, empty_line);
    if(id >= 0)
    {
        conio_printf(2, 23, COLOR_SILVER, "%s", _fw_list[id].detail1);
    }

    conio_printf(2, 24, COLOR_WHITE, empty_line);
    if(id >= 0)
    {
        conio_printf(2, 24, COLOR_SILVER, "%s", _fw_list[id].detail2);
    }

    conio_printf(2, 25, COLOR_WHITE, empty_line);
    if(id >= 0)
    {
        conio_printf(2, 25, COLOR_SILVER, "%s", _fw_list[id].detail3);
    }
}

void FlasherProcess(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    unsigned short* pRom = (unsigned short*)0x20200000;
#define PROM_LEN (1024*1024)
    unsigned long CmpLen;
    unsigned long RomLen;

    unsigned long page;
    Romcheck_t checkRes;
    pad_t* pad;
    int i;

    /* Flash firmware : first get infos about flash chips. */
    DisplayFlashIds(&device, &vendorID, &deviceID);

    /* Initialize menu internals. */
    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_redraw(mset, display_fw_infos);
    menu_set_callback_vblank(mset, display_flash_ids);

    menu_set_title(mset, "SELECT FIRMWARE");

    /* Indicate version in description strings for latest firmware. */
    char latest_desc[LINE_LEN];

#ifdef PSKAI_LITE_BUILD
    sprintf(latest_desc, "Pseudo Saturn Kai V%d.%03d Lite", RELEASE_ID/1000, RELEASE_ID%1000);
    _fw_list[0].desc = latest_desc;
#else
    sprintf(latest_desc, "Pseudo Saturn Kai V%d.%03d"     , RELEASE_ID/1000, RELEASE_ID%1000);
    _fw_list[0].desc = latest_desc;
#endif

    /* Prepare pointers for displaying informations under the menu. */
    i=0;
    while(_fw_list[i].data_stt != NULL)
    {
        menu_set_item(mitems, i/*id*/, _fw_list[i].desc, i/*tag*/);
        i++;
    }
    menu_set_pos(mset, -1/*x0*/, 60/*y0*/, 32/*w*/, i+1/*h*/, 1/*cols*/);
    /* Clear screen after executing menu. */
    menu_set_erase_on_exit(mset, 1);

    /* By default, select first firmware. */
    menu_selected = 0;
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    /* Erase additional display area. */
    display_fw_infos(-1);
    ClearFlashIdsProgress();

    /* Cancel ? */
    if((mset->exit_code == MENU_EXIT_CANCEL)
    || (menu_selected == 1))
    {
        return;
    }

    /* Test flash memory integrity. */
    if(device.BrickWarning)
    {
        /* Unknown device, but maybe Action Replay Flash chips.
         * Warn user that this program will destroy his ARP before continuing.
         */
        conio_printf(2, 21, COLOR_YELLOW, "Uncommon flash chips! Do you want to");
        conio_printf(2, 22, COLOR_YELLOW, "try with default chip settings ?    ");
        conio_printf(2, 23, COLOR_WHITE , "You can slightly move your cartridge");
        conio_printf(2, 24, COLOR_WHITE , "in order to improve its detection.  ");
        conio_printf(2, 25, COLOR_ERROR , "Please try at your own risk !!!     ");
        conio_printf(2, 26, COLOR_WHITE , "Press Start: exit, A+B+C: continue. ");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            /* Q&D hack : display vendor/device IDs during idle. */
            DisplayFlashIds(&device, &vendorID, &deviceID);

            pad = pad_read();
            if((pad->new[0] & (PAD_A|PAD_B|PAD_C)) == (PAD_A|PAD_B|PAD_C))
            {
                break;
            }
            if(pad->delta[0] & PAD_START)
            {
                /* Flash memory : clear progress display. */
                ClearFlashIdsProgress();

                return;
            }
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }

        /* User decided to continue, so clear disclaimer. */
        conio_printf(2, 21, COLOR_WHITE, empty_line);
        conio_printf(2, 22, COLOR_WHITE, empty_line);
        conio_printf(2, 23, COLOR_WHITE, empty_line);
        conio_printf(2, 24, COLOR_WHITE, empty_line);
        conio_printf(2, 25, COLOR_WHITE, empty_line);
        conio_printf(2, 26, COLOR_WHITE, empty_line);

        /* Flash memory : clear progress display. */
        ClearFlashIdsProgress();
    }

    /* Final check of flash memory integrity. */
    if(device.RomSize == 0 || device.PageSize == 0)
    {
        /* Unknown device.
         * Don't go any further, but update display of vendor/device IDs.
         */
        conio_printf(2, 21, COLOR_ERROR, "Unrecognized flash chips !          ");
        conio_printf(2, 22, COLOR_ERROR, "Reflash process is interrupted.     ");
        conio_printf(2, 23, COLOR_WHITE, "You can slightly move your cartridge");
        conio_printf(2, 24, COLOR_WHITE, "in order to improve its detection.  ");
        conio_printf(2, 25, COLOR_ERROR, "Please try at your own risk !!!     ");
        conio_printf(2, 26, COLOR_WHITE, "Press Start to exit.                ");

        wait_keys_release(PAD_START);
        while(1)
        {
            /* Q&D hack : display vendor/device IDs during idle. */
            DisplayFlashIds(&device, &vendorID, &deviceID);
 
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                /* Flash memory : clear progress display. */
                ClearFlashIdsProgress();

                return;
            }
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }
    }


    /* Set pointer and length of selected firmware. */
    memset((void*)pRom, _fw_list[menu_selected].fill_val, PROM_LEN);
    CmpLen = (unsigned long)(_fw_list[menu_selected].data_end - _fw_list[menu_selected].data_stt);
    RomLen = decomp_exec(
            (unsigned char*)_fw_list[menu_selected].data_stt, 
            CmpLen, 
            (unsigned char*)pRom);

    conio_printf(2, 7, COLOR_WHITE, "%s", _fw_list[menu_selected].desc);
    conio_printf(2, 8, COLOR_WHITE, "Compressed size  : %7u (%4u KB)", CmpLen, CmpLen>>10);
    conio_printf(2, 9, COLOR_WHITE, "Decompressed size: %7u (%4u KB)", RomLen, RomLen>>10);

    /* Verify that firmware was correctly decompressed. */
    if(RomLen == 0)
    {
        conio_printf(2, 21, COLOR_ERROR, "Firmware decompress error !         ");
        conio_printf(2, 22, COLOR_ERROR, "Reflash process is interrupted.     ");

        conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

        wait_keys_release(PAD_START);
        while(1)
        {
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }
    }

    /* Verify that firmware fits into flash memory. */
    if(RomLen > (device.RomSize * sizeof(unsigned short)))
    {
        conio_printf(2, 21, COLOR_ERROR, "Firmware doesnt fit in flash memory!");
        conio_printf(2, 22, COLOR_ERROR, "FW size:%4u KB, Flash size:%4u KB ", 
            RomLen>>10, 
            (device.RomSize * sizeof(unsigned short)) >> 10);
        conio_printf(2, 23, COLOR_ERROR, "Reflash process is interrupted.     ");

        conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

        wait_keys_release(PAD_START);
        while(1)
        {
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }
    }

    /* Warn user from flashing firmware other than
     * Pseudo Saturn Kai full version on Gamer's Cartridge.
     */
    if((FlashGetVendorID() == 0xBFBF)
    && (FlashGetDeviceID() == 0xB7B7)
    && (RomLen != (1024*1024)))
    {
        conio_printf(2, 21, COLOR_YELLOW, "It is recommended to flash Pseudo");
        conio_printf(2, 22, COLOR_YELLOW, "Saturn Kai full firmware on this");
        conio_printf(2, 23, COLOR_YELLOW, "cartridge !");
        conio_printf(2, 24, COLOR_YELLOW, "Press Start button to continue, ");
        conio_printf(2, 25, COLOR_YELLOW, "or B button to cancel operation.");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            pad = pad_read();
            if(pad->new[0] & (PAD_A|PAD_C|PAD_START))
            {
                break;
            }
            if(pad->delta[0] & PAD_B)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }

        /* User decided to continue, so clear warning message. */
        conio_printf(2, 21, COLOR_WHITE, empty_line);
        conio_printf(2, 22, COLOR_WHITE, empty_line);
        conio_printf(2, 23, COLOR_WHITE, empty_line);
        conio_printf(2, 24, COLOR_WHITE, empty_line);
        conio_printf(2, 25, COLOR_WHITE, empty_line);
        conio_printf(2, 26, COLOR_WHITE, empty_line);
    }

    conio_printf(2, 11, COLOR_WHITE, "Press A+B+C to initiate write...");
    conio_printf(2, 12, COLOR_WHITE, "Press Start to cancel.");

    wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);

    while(1)
    {
        pad = pad_read();

        if(pad->delta[0] & PAD_START)
        {
            return;
        }

        if((pad->new[0] & (PAD_A|PAD_B|PAD_C)) == (PAD_A|PAD_B|PAD_C))
        {
            break;
        }
        WaitForVBLANKIn();
        WaitForVBLANKOut();
    }


    /* Look for unused flash page, and test writing on it
     * before erasing and re-writing everything. If write failed
     * on this test flash page, warn user it could brick their
     * cartridge, and prevent from flashing more, so that cartridge
     * still should boot fine after that.
     *
     * Let's assume that the farther in the ROM, the potentially
     * safer for testing the page is. And, look for page containing
     * only 0x0000 or 0xFFFF value, so that botch writing it
     * shouldn't brick the cartridge.
     *
     * Note: first (boot) page is discarded from testing, hence it
     *       is safe to consider that no test page could be found
     *       when test page ID is zero.
     */
    conio_printf( 2, 12, COLOR_WHITE, empty_line);
    conio_printf( 2, 12, COLOR_WHITE, "Write Test:");
    conio_printf( 2, 13, COLOR_WHITE, "Searching...");
    conio_printf( 2, 15, COLOR_WHITE, "Page          of          ");
    conio_printf(19, 15, COLOR_YELLOW, "%08X", device.RomSize / device.PageSize);

    unsigned long testpage_id = 0;
    for(page = (device.RomSize / device.PageSize) - 1; page > 0; page--)
    {
        conio_printf( 7, 15, COLOR_YELLOW, "%08X", page + 1);

        unsigned short* pPageDat = ((unsigned short*)0x22000000) + (page * device.PageSize);
        unsigned short usTestDat = pPageDat[0];
        if((usTestDat == 0x0000) || (usTestDat == 0xFFFF))
        {
            int is_testable_page = 1;
            unsigned long offset = 0;
            for(offset = 1; offset < device.PageSize; offset++)
            {
                if(pPageDat[offset] != usTestDat)
                {
                    is_testable_page = 0;
                    break;
                }
            }
            if(is_testable_page)
            {
                testpage_id = page;
                break;
            }
        }
    }

    /* Erase test page search progress. */
    conio_printf( 2, 15, COLOR_WHITE, empty_line);

    /* Show test page ID. */
    conio_printf(16, 13, COLOR_WHITE , "Page: ");
    conio_printf(22, 13, COLOR_YELLOW, "%08X", testpage_id);


    /* Inform user that write test page couldn't be found.
     * This doesn't prevents from writing flash ROM.
     */
    if(testpage_id == 0)
    {
        conio_printf(2, 21, COLOR_YELLOW, "Could not find unused page where to ");
        conio_printf(2, 22, COLOR_YELLOW, "test flash write ! Please confirm to");
        conio_printf(2, 23, COLOR_YELLOW, "to continue flashing to ROM anyway. ");
        conio_printf(2, 24, COLOR_WHITE , empty_line);
        conio_printf(2, 25, COLOR_WHITE , empty_line);
        conio_printf(2, 26, COLOR_WHITE , "Press Start: exit, A+B+C: continue. ");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            pad = pad_read();
            if((pad->new[0] & (PAD_A|PAD_B|PAD_C)) == (PAD_A|PAD_B|PAD_C))
            {
                break;
            }
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }

        /* Indicate that write test couldn't be done, but
         * user decided to continue anyway.
         */
        conio_printf(14, 12, COLOR_YELLOW, "test page not found     ");
    }
    else
    {
        /* Test write on page found above, and remember this is
         * not some kind of benchmark/will it blend ?/whatever
         * stress test : it would be a pity to destroy the flash
         * just for testing purpose !
         * Instead, fill the test page with enough relevant data
         * for typical test patterns, and erase/write it only once.
         *
         * Note : test page contents are written on RAM, over
         *        decompressed firmware data from offset 256KB.
         *        So it is required to re-read firmware data if
         *        it is larger than 256KB, which concerns only
         *        full firmware = majority of users don't need
         *        re-read and re-decompress firmware data.
         */
        unsigned short* pTestDat = ((unsigned short*)pRom) + (LITE_FW_ROM_LEN / sizeof(unsigned short));

        /* Prepare write test page with random enough data. */
        unsigned long offset = 0;
        for(offset = 0; offset < device.PageSize; offset++)
        {
            unsigned long t1 = offset *  17;
            unsigned long t2 = offset * 101;
            pTestDat[offset] = ((t2 & 0xFF) << 8) | (t1 & 0xFF);
        }
        /* Add data bit test patterns on first 4 data words. */
        pTestDat[0] = 0x5555;
        pTestDat[1] = 0xAAAA;
        pTestDat[2] = 0xA5A5;
        pTestDat[3] = 0x5A5A;

        /* Add "TEST" string on next 4 data words. */
        pTestDat[4] = ('T' << 8) | 't';
        pTestDat[5] = ('E' << 8) | 'e';
        pTestDat[6] = ('S' << 8) | 's';
        pTestDat[7] = ('T' << 8) | 't';

        /* Erase write test flash page. */
        conio_printf( 2, 13, COLOR_WHITE, "Erasing...  ");
        device.EraseFlashPage(&device, testpage_id);

        /* Write flash ROM page with test data. */
        conio_printf( 2, 13, COLOR_WHITE, "Writing...  ");
        device.WriteFlashPage(&device, testpage_id, pTestDat);

        /* Compare written data and test data. */
        conio_printf( 2, 13, COLOR_WHITE, "Comparing...");
        unsigned short* pFlashDat = ((unsigned short*)0x22000000) + (testpage_id * device.PageSize);
        int compare_ok = 1;
        for(offset = 0; offset < device.PageSize; offset++)
        {
            if(pFlashDat[offset] != pTestDat[offset])
            {
                compare_ok = 0;
                break;
            }
        }

        /* End of flash page write test : don't forget
         * to wipe test data in RAM. That concerns
         * users with cartridge flash ROM larger than
         * 256KB wanting to write lite or legacy firmwares, 
         * so basically nobody, but hey this doesn't
         * costs anything to wipe data anyway !
         */
        memset((void*)pTestDat, _fw_list[menu_selected].fill_val, device.PageSize * sizeof(unsigned short));

        if(compare_ok)
        {
            /* Test page write success : allow firmware flash. */
            conio_printf(14, 12, COLOR_YELLOW, "Success (page:%08X) ");
        }
        else
        {
            /* Test page write failure : show some error message
             * and prevent from destroying more the cartridge flash ROM.
             */
            conio_printf(14, 12, COLOR_ERROR, "data compare failure !  ");

            conio_printf(2, 14, COLOR_YELLOW, "Write on flash ROM test page failed!");
            conio_printf(2, 15, COLOR_YELLOW, "Please check cartridge connectivity,");
            conio_printf(2, 16, COLOR_YELLOW, "or try to flash with another Saturn.");

            /* Show hex dump near where write compare failed. */
            memset(&checkRes, 0, sizeof(Romcheck_t));
            checkRes.ErrorOffset = offset*sizeof(unsigned short);
            checkRes.pErrorData  = ((unsigned char*)pFlashDat) + (checkRes.ErrorOffset & ~0xF);
            DisplayError(&checkRes, ERROR_WRTEST);

            /* Prevent from flashing firmware. */
            return;
        }
    }


    /* Clear flash page test write stuff. */
    for(i=13; i<27; i++)
    {
        conio_printf(2, i, COLOR_WHITE, empty_line);
    }


    /* Re-read and decompress firmware data in the case
     * write test data overwrote firmware data in RAM.
     */
    if(RomLen > LITE_FW_ROM_LEN)
    {
        conio_printf(2, 13, COLOR_WHITE, "Re-decompressing ... Please wait.");

        memset((void*)pRom, _fw_list[menu_selected].fill_val, PROM_LEN);
        CmpLen = (unsigned long)(_fw_list[menu_selected].data_end - _fw_list[menu_selected].data_stt);
        RomLen = decomp_exec(
                (unsigned char*)_fw_list[menu_selected].data_stt, 
                CmpLen, 
                (unsigned char*)pRom);

        /* Verify that firmware was correctly decompressed. */
        if(RomLen == 0)
        {
            conio_printf(2, 21, COLOR_ERROR, "Firmware decompress error !         ");
            conio_printf(2, 22, COLOR_ERROR, "Reflash process is interrupted.     ");

            conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

            wait_keys_release(PAD_START);
            while(1)
            {
                pad = pad_read();
                if(pad->delta[0] & PAD_START)
                {
                    return;
                }
                WaitForVBLANKIn();
                WaitForVBLANKOut();
            }
        }

        conio_printf(2, 13, COLOR_WHITE, empty_line);
    }

    /* If creating lite firmware from full one, 
     * fill firmware data over 256KB offset with 0xFF.
     * (Not necessary, but just in case of)
     */
    if(_fw_list[menu_selected].lite)
    {
        memset((void*)(((unsigned char*)pRom) + LITE_FW_ROM_LEN), _fw_list[menu_selected].fill_val, RomLen-LITE_FW_ROM_LEN);
        RomLen = LITE_FW_ROM_LEN;
    }

    /* Individually erase then write each sectors when writing
     * Gamer's Cartridge firmware.
     * Main interest of doing that is to reduce the risk of bricking
     * the cartridge when Saturn freezes or is turned off during
     * firmware update.
     * (Disclaimer : There are still risk of bricking the cartridge, 
     *               so do that at your own risks !)
     *
     * This is not available for other cartridges for the simple
     * reason that "erase whole flash and write each sectors"
     * behavior was tested in depth a long ago and that I'm too
     * tired to find volunteers to re-do that again.
     */
    int flash_minipseudo_first = 0;
    if((FlashGetVendorID() == 0xBFBF)
    && (FlashGetDeviceID() == 0xB7B7)
    && (RomLen == (1024*1024)))
    {
        flash_minipseudo_first = 1;
    }

    /* Okay, let's erase then write cartridge flash ROM ! */
    conio_printf(2, 13, COLOR_WHITE, "Erasing... Please wait.");

    conio_printf(2, 14, COLOR_ERROR, "DO NOT RESET OR TURN OFF YOUR SATURN");
    device.EraseFlash(&device);
    conio_printf(2, 14, COLOR_WHITE, empty_line);

    /* Test flash erase. */
    if(!device.EraseOK(&device, &checkRes))
    {
        DisplayError(&checkRes, ERROR_ERASE);
        return;
    }

    /* Re-write whole flash memory. */
    conio_printf(2, 14, COLOR_WHITE, "Writing...");
    conio_printf(2, 15, COLOR_WHITE, "Page          of          ");
    conio_printf(2, 16, COLOR_ERROR, "DO NOT RESET OR TURN OFF YOUR SATURN");

    conio_printf(19, 15, COLOR_YELLOW, "%08X", device.RomSize / device.PageSize);

    for(page = 0; page < device.RomSize / device.PageSize; page++)
    {
        unsigned long write_page = page;
        if(flash_minipseudo_first)
        {
            /* Flash mini Pseudo Saturn stub first.
             * Purpose : minimize risk to brick the cartridge when flashing it
             * How it works : flash mini Pseudo Saturn just after the first
             *                two flash pages
             */
            unsigned long minips_start_page = (MINIPSEUDO_HDR_START & 0x01FFFFFF) / (device.PageSize * sizeof(unsigned short));
            unsigned long minips_page_count = MINIPSEUDO_MAX_LEN / (device.PageSize * sizeof(unsigned short));
#define BOOT_PAGE_COUNT 2

            if((page >= BOOT_PAGE_COUNT) && (page < (BOOT_PAGE_COUNT + minips_page_count)))
            {
                /* Flash mini Pseudo Saturn just after IP, FSBL etc. */
                write_page = minips_start_page + (page - BOOT_PAGE_COUNT);
            }
            else if((page >= minips_start_page) && (page < (minips_start_page + minips_page_count)))
            {
                /* Flash first Pseudo Saturn Kai flash pages
                 * at the moment of mini Pseudo Saturn.
                 */
                write_page = BOOT_PAGE_COUNT + (page - minips_start_page);
            }
            else
            {
                /* Other pages written in normal order. */
                write_page = page;
            }
        }

        conio_printf(7, 15, COLOR_YELLOW, "%08X", write_page + 1);

        device.WriteFlashPage(&device, write_page,
                              (unsigned short*)(pRom + (write_page * device.PageSize)));
    }

    conio_printf(2, 16, COLOR_WHITE, empty_line);

    /* Test flash write. */
    if(!device.WriteOK(&device, &checkRes, (const unsigned short*)(pRom)))
    {
        DisplayError(&checkRes, ERROR_WRITE);
        return;
    }
    else
    {
        conio_printf(2, 17, COLOR_GREEN, "Write verification success.");
    }

    conio_printf(2, 23, COLOR_WHITE, "Done!");
    conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

    wait_keys_release(PAD_START);
    while(1)
    {
        pad = pad_read();
        if(pad->delta[0] & PAD_START)
        {
            return;
        }
        WaitForVBLANKIn();
        WaitForVBLANKOut();
    }
}


/**
 * System Reset.
**/
/* Acquisition timeout value, unit is poll count. */
#define SC_SYSRES_TIMEOUT (5000*1000)
void smpc_sysres(void)
{
    unsigned long cntr = 0;

    /* Waits for SF to be Reset. */
    do { cntr++; if(cntr > SC_SYSRES_TIMEOUT) return; } while(SF == 1);

    /* Writes Command Code in COMREG. */
    COMREG = 0x0D;

    /* Set SF. */
    SF = 1;

    /* Expected value in OREG31 is 0x0E. */
    if(OREG31 != 0x0E)
    {
        return;
    }

    /* Final wait. */
    while(1) {;}
}



void _main(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    vdp_init();
    conio_init();
    pad_init();

    TVMD = 0x8000;

    conio_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), "  Pseudo Saturn Kai Flasher V%d.%03d  ", RELEASE_ID/1000, RELEASE_ID%1000);

    while(1)
    {
        FlasherProcess();

        /* Clear display. */
        int i;
        for(i=7; i<27; i++)
        {
            conio_printf(2, i, COLOR_WHITE, empty_line);
        }

        /* Display "Do you want to exit ?" menu. */
        menu_reset_settings(mset);
        menu_reset_items(mitems);
        menu_set_title(mset, "Please select action");
        menu_set_item(mitems, 0/*id*/, "   Restart Flasher  ", 0/*tag*/);
        menu_set_item(mitems, 1/*id*/, " Exit to Multiplayer", 1/*tag*/);
        menu_set_item(mitems, 2/*id*/, "    System Reset    ", 2/*tag*/);
        menu_set_pos(mset, -1/*x0*/, 60/*y0*/, 21/*w*/, 4/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);

        menu_selected = 0;
        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        /* Exit to multiplayer ? */
        if(((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 1))
        ||  (mset->exit_code == MENU_EXIT_CANCEL))
        {
            (**(void(**)(void))0x600026C)();
        }

        /* System Reset ? */
        if((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 2))
        {
            smpc_sysres();
        }
    }
}
