/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2014 James Laird-Wah

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <sc_saturn.h>

#include "main.h"
#include "conio.h"
#include "starfield.h"

/* Bootstub ROM code declarations. */
#include "bootstub/bootstub_rom_exec.h"

#include "pskai_flags.h"

/* Extra inclusions for custom loader stuff. */
#include "pskai_settings.h"
#include "custom_loader.h"


static int emulate_bios_loadcd_init(void)
{
    int ret;
    *(uint32_t*)0x6000278 = 0;
    *(uint32_t*)0x600027c = 0;
    if ((ret = cd_abort_file()) != IAPETUS_ERR_OK)
    {
        return ret;
    }
    if ((ret = cd_end_transfer()) != IAPETUS_ERR_OK)
    {
        return ret;
    }
    if ((ret = cd_reset_selector_all()) != IAPETUS_ERR_OK)
    {
        return ret;
    }
    if ((ret = cd_set_sector_size(SECT_2048)) != IAPETUS_ERR_OK)
    {
        return ret;
    }
    return IAPETUS_ERR_OK;
}

static struct region_s
{
    char id;
    const char *key;
} regions[] =
{
    {1,   "JAPAN"},
    {2,   "TAIWAN and PHILIPINES"},
    {4,   "USA and CANADA"},
    {5,   "BRAZIL"},
    {6,   "KOREA"},
    {10,  "ASIA PAL area"},
    {12,  "EUROPE"},
    {13,  "LATIN AMERICA"},
    {0,   NULL}
};

static const char *get_region_string(void)
{
    char region = *(volatile char*)0x20100033;
    struct region_s *r;
    for (r=regions; r->id; r++)
    {
        if (r->id == region)
        {
            return r->key;
        }
    }
    return NULL;
}

int set_image_region(u8 *base)
{
    const char *str = get_region_string();
    if (!str)
    {
        conio_printf(23, 13, COLOR_WHITE, "Invalid region!");
        return -1;
    }

    /* Set region header. */
    memset(base + 0x40, ' ', 0x10);
    base[0x40] = str[0];

    /* Set region footer. */
    char *ptr = (char*)base + 0xe00;
    memset(ptr, ' ', 0x20);
    *(unsigned long*)ptr = 0xa00e0009;
    strcpy(ptr + 4, "For ");
    strcpy(ptr + 8, str);
    char *end = ptr + 4 + strlen(ptr + 4);
    *end = '.';

    conio_printf(23, 13, COLOR_WHITE, "Region: %s", str);
    return 0;
}

#define JHL_IP_HEADER ((ip_bin_t*)0x06002000)

static int emulate_bios_loadcd_read(void)
{
    int ret, i;

    /* Doesn't matter where. */
    u8 *ptr = (u8*)JHL_IP_HEADER;

    ret = cd_read_sector(ptr, 150, SECT_2048, 2048*16);
    if (ret < 0)
    {
        return ret;
    }

    /* Keep a backup of game's IP header. */
    memcpy(&_ip_backup, ptr, sizeof(ip_bin_t));

    /* Display IP header infos to screen. */
    ip_bin_display(&_ip_backup);

    /* BIOS doesn't like PSEUDO discs ;) */
    memcpy(ptr, "SEGA SEGASATURN ", 16);

    ret = set_image_region(ptr);
    if (ret < 0)
    {
        return ret;
    }

    ret = cd_put_sector_data(0, 16);
    if (ret < 0)
    {
        return ret;
    }
    while (!(CDB_REG_HIRQ & HIRQ_DRDY)) {}

    for (i = 0; i < 2048 * 16; i+=4)
    {
        CDB_REG_DATATRNS = *(unsigned long*)(ptr + i);
    }

    if ((ret = cd_end_transfer()) != 0)
    {
        return ret;
    }

    while (!(CDB_REG_HIRQ & HIRQ_EHST)) {}

#if 1 /* Compatible with all hardware/BIOS variations ? */
   // we need to set a flag, but it's in different places in different BIOS versions
   uint16_t *read_flag_a = (uint16_t*)0x06000380;
   uint16_t *read_flag_b = (uint16_t*)0x060003a0;

   uint16_t **read_flag_ptr_a = (uint16_t**)0x2174;
   uint16_t **read_flag_ptr_b = (uint16_t**)0x2a04;

   if (read_flag_a == *read_flag_ptr_a)
   {
      *read_flag_a = 1;
   }
   else if (read_flag_b == *read_flag_ptr_b)
   {
      *read_flag_b = 1;
   }
   //else
   //{
   //    return BOOT_UNRECOGNISED_BIOS;
   //}
#else /* Incompatible with some grey Japanese Saturns. */
    *(unsigned short*)0x60003a0 = 1;
#endif

    return 0;
}


void jhload(unsigned long custom_loader_set)
{
    /* Initial version of JHL loader wasn't compatible
     * with all Saturn hardware variations and as a
     * countermeasure "plain" loader (which call BIOS
     * init/read/boot functions) was used when Saturn
     * was equiped with  ODE.
     * This worked because ODE doesn't cause problem
     * when authenticating a CD-ROM.
     *
     * Now that JHL loader is working with all (?)
     * know Saturn hardware variations, JHL loader
     * is used without consideration of optical disc
     * type. This brings the advantage of not requiring
     * the setting of appropriate region setting on
     * ODE side.
     *
     * Special thanks to jhl for developing and maintaining
     * this loader, and privateye for pointing about the
     * improvement of this loader when ODE is installed !
     *
     * Update : for unknown reason, RmenuKai and Fenrir Loader Kai
     *          load games to a black screen when emulated
     *          functions are used ... but works when Pseudo Saturn
     *          Kai boots from cartridge.
     *          So selection between BIOS or emulated functions
     *          depends on optical disc type and if it is
     *          running in "ROM-less" mode or not.
     */
#if 0
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    /* Use BIOS functions when booting a disc from Rhea/Phoebe. */
    int emufct_use = (set->bram.od_type != SET_ODTYPE_NORMAL ? 0 : 1);
#else
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int emufct_use;
    switch(set->bram.od_type)
    {
    case(SET_ODTYPE_NORMAL):
        emufct_use = 1;
        break;
 
    default:
    case(SET_ODTYPE_RHOEBE):
    case(SET_ODTYPE_FENRIR):
        emufct_use = (PSKAI_IS_ROM_LESS() ? 0 : 1);
        break;
    }
#endif

    int ret;

    //! Prepare for Load CD operation.
    /*!
     *  A Saturn CD may be read and booted using the bios_loadcd_ calls in order init, read, boot.
     *  Negative return values indicate failures.
     *
     *  bios_loadcd_init():
     *      - resets selectors and the usual things
     *      - resets some BIOS state variables
     *      - starts a disc authentication
     */
    conio_printf(2, 20, COLOR_WHITE, "Initialising CD block...");
    if(emufct_use)
    {
        ret = emulate_bios_loadcd_init();
    }
    else
    {
        ret = bios_loadcd_init();
    }
    if(ret < 0)
    {
        conio_printf(2, 20, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), "Initialising CD block: Error %d.", ret);
        for (;;) {}
    }
    conio_printf(25, 20, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), "Done.");

    //! Start reading for Load CD.
    /*!
     *  bios_loadcd_read():
     *      - checks the CD auth result
     *      - starts a Play command for the first 16 disc sectors, into selector 0
     *      - sets some BIOS flags
     */
    conio_printf(2, 21, COLOR_WHITE, "Reading CD data...");
    if(emufct_use)
    {
        ret = emulate_bios_loadcd_read();
    }
    else
    {
        ret = bios_loadcd_read();
    }
    if(ret < 0)
    {
        conio_printf(2, 21, COLOR_WHITE, "Reading CD data: Error %d.", ret);
        for (;;) {}
    }
    conio_printf(19, 21, COLOR_WHITE, "Done.");

    /* It seems it's the right time to update custom
     * loader settings in the case JHL loader is manually used.
     */
    if(custom_loader_set != CL_SET_INVALID)
    {
        cl_setting_update(ip_header_get_crc(&_ip_backup), custom_loader_set);
        pskai_extraset_write();
    }

    /* Last steps before booting the game, so it's
     * time to flush temporary report
     * informations to SD card.
     */
    compat_flushbuffer();

    //conio_printf(2, 22, COLOR_WHITE, "Booting CD...");

    /* Reset VDP registers. */
    soft_reset(SC_SOFTRES_EACH_VDPS);


    //! Attempt to boot loaded disc
    /*!
     *  Must not be called until the disc read is complete.
     *  bios_loadcd_boot():
     *      - reads the sector data from the CDB to 0x06002000
     *      - checks the SEGA SEGASATURN header
     *      - checks the IP size, security code
     *      - checks the region
     *      - boots the disc if checks pass
     *
     *  Return values include:
     *      -1 - bad header
     *      -4 - bad security code
     *      -8 - bad region
     *      +1 - not a Saturn disc
     */
#if 1
    /* Execute bootstub for JHL loader.
     * Calling bootstub will clear LRAM, 
     * then call bios_loadcd_boot function.
     */
    bootstub_params_t* params = BOOTSTUB_PARAMS_ADDRESS;
    memset((void*)params, 0, sizeof(bootstub_params_t));
    params->rom_less = (rc_bootstub_is_romcode_available() ? 0 : 1);
    rc_bootstub_exec(params->rom_less);
#else
    do
    {
        ret = bios_loadcd_boot();
    } while(ret == 1);

    /* if we came back, that's an error!
     * error -8: bad region
     * error -4: bad security code
     * error -1: bad headers
     */
    
    /* Re-init VDP registers, and other stuff. */
    init_iapetus();
    vdp_init();
    pad_init();
    conio_init();
    sf_init(1/*multicolor*/);
    logo_set_enable(0/*enable_flag*/);

    conio_printf(2, 22, COLOR_WHITE, "Booting CD: Error %d.", ret);

    while(1)
    {
        pad_t* pad;
        background_vblank_clbk();

        pad = pad_read();
        if((pad->delta[0] & PAD_START)
        || (pad->delta[0] & PAD_A    ))
        {
            break;
        }
    }
#endif
}
