/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "logo.h"

#include <stdio.h>
#include <string.h>
#include <sc_saturn.h>

#include "conio.h"
#include "main.h"
#include "release.h"
#include "starfield.h"
#include "vdp.h"

/* Inclusion of default logo data. */
#include "binwrap.h"

/* Uncomment the line below when doing logo related debug. */
//#define LOGO_DEBUG_ENABLE


/* Luminosity -> color table. */
unsigned short tbl_lum2col[32] =
{
    COLOR( 0,  0,  0), 
    COLOR( 1,  1,  1), 
    COLOR( 2,  2,  2), 
    COLOR( 3,  3,  3), 
    COLOR( 4,  4,  4), 
    COLOR( 5,  5,  5), 
    COLOR( 6,  6,  6), 
    COLOR( 7,  7,  7), 
    COLOR( 8,  8,  8), 
    COLOR( 9,  9,  9), 
    COLOR(10, 10, 10), 
    COLOR(11, 11, 11), 
    COLOR(12, 12, 12), 
    COLOR(13, 13, 13), 
    COLOR(14, 14, 14), 
    COLOR(15, 15, 15), 
    COLOR(16, 16, 16), 
    COLOR(17, 17, 17), 
    COLOR(18, 18, 18), 
    COLOR(19, 19, 19), 
    COLOR(20, 20, 20), 
    COLOR(21, 21, 21), 
    COLOR(22, 22, 22), 
    COLOR(23, 23, 23), 
    COLOR(24, 24, 24), 
    COLOR(25, 25, 25), 
    COLOR(26, 26, 26), 
    COLOR(27, 27, 27), 
    COLOR(28, 28, 28), 
    COLOR(29, 29, 29), 
    COLOR(30, 30, 30), 
    COLOR(31, 31, 31), 
};


typedef struct _vertex_t
{
    /* Coordinates. */
    short x, x_target;
    short y, y_target;

    /* Luminosity (0:too dark, 15: normal, 31:too bright). */
    unsigned char lum, lum_target;
} vertex_t;

typedef struct _tile_t
{
    vertex_t* v[4];

    unsigned long tex_offset;

    unsigned short grad_offset;
    unsigned short* grad_ptr;
} tile_t;

typedef struct _logo_t
{
    vertex_t v[LOGO_TILECNT_X+1][LOGO_TILECNT_Y+1];
    tile_t   t[LOGO_TILECNT_X  ][LOGO_TILECNT_Y  ];

    /* Application type : Pseudo Saturn Kai or Save Data Manager.
     * Used in order to adapt behavior from an application to another.
     */
    int app_type;

    /* Transition pattern : LOGO_PATTERN_*
     * Backup from logo_set_pattern function's input parameter.
     */
    int pattern;

    /* Animation type : LOGO_ANIM_*
     * Backup from logo_set_pattern function's input parameter.
     */
    int anim_type;

    /* Logo enabled/disabled flag. */
    int disabled;

    /* Update counter, used to adjust luminosity fade speed. */
    unsigned long cntr;

    /* Transition counter divider, used to tune
     * luminosity fade speed.
     *
     * Examples :
     *  - 0 : setting prohibited.
     *  - 1 : fastest fade speed.
     *  - 8 : relatively slow fade speed.
     */
    int fade_divisor;

    /* Gouraud related flags :
     *  - gouraud_use      : enable Gouraud shading.
     *  - gouraud_auto_off : automatically disable Gouraud shading at end of fade.
     *  - lum_diff_total   : total luminosity difference with target.
     */
    int gouraud_use;
    int gouraud_auto_off;
    long lum_diff_total;

    /* Mesh usage flag.
     * Used in order to reduce luminosity by half
     * when using Gouraud shading isn't enough.
     */
    int mesh_use;

    /* Set to 1 when transparent logo image is used.
     *
     * When transparent image is used, starfield is displayed
     * first, and then logo image, so that starfield animation
     * is shown only on transparent regions.
     *
     * When opaque image is used, logo is displayed first, and
     * and then starfield animation, so that both are visible.
     */
    int transparency_use;
} logo_t;


/*
 * Can't use HRAM, because logo is displayed
 * in all menus, not only during credits page.
 */
//#define LOGO_INFO_PTR (logo_t*)0x06004000;
logo_t _logo_global;
#define LOGO_INFO_PTR (logo_t*)(&_logo_global);


void logo_init(int app_type)
{
    int i, j;
    logo_t* logo = LOGO_INFO_PTR;

    memset((void*)logo, 0, sizeof(logo_t));
    logo->app_type = app_type;
    logo->disabled = 1;

    /* Initialize logo load parameters that
     * can't be initialized at compile time.
     */
    _logo_load_params[LOGO_DEFAULT_LOAD_PARAMS].data = (unsigned char*)(&_logo_lite_dat);
    _logo_load_params[LOGO_DEFAULT_LOAD_PARAMS].datalen = (unsigned long )(&_logo_lite_end - &_logo_lite_dat);


    /* Initialize vertex luminosity. */
    for(j=0; j<(LOGO_TILECNT_Y+1); j++)
    {
        for(i=0; i<(LOGO_TILECNT_X+1); i++)
        {
            vertex_t* v = &(logo->v[i][j]);

            v->lum = 0;
            v->lum_target = 15;
        }
    }
}

/* Logo load parameters.
 * First entry is default one for lite version of the firmware.
 * Second entry is custom logo loaded from SD card, for full version of the firmware.
 * Next entries are alternate logos loaded from RomFs, for full version of the firmware.
 */
logo_load_params_t _logo_load_params[] =
{
    { /* User specified logo, loaded from SD card. */
/* Full path to custom logo file. */
#define LOGO_FOLDER "/PSKAI"
#define LOGO_PATH   "/PSKAI/LOGO.BIN"
        .dev_id = SC_FILE_SDCARD, 
        .file_path = LOGO_PATH, 

        .data = NULL, 
        .datalen = 0, 

        .compressed = 1, 
        .default_logo = 0, 
    }, 

    { /* Default logo by Alien, directly embedded with executable. */
        .dev_id = 0, 
        .file_path = NULL, 

        /* Will be initialized in logo_clear_internals function. */
        .data = NULL, 
        .datalen = 0, 

        .compressed = 1, 
        .default_logo = 1, 
    }, 

    { /* Terminating entry. */
        .dev_id = 0, 
        .file_path = NULL, 
        .data = NULL, 
        .datalen = 0, 
        .compressed = 0, 
        .default_logo = 0, 
    }, 

};

#ifdef MAKE_BUPMNGR
    /* Include the definition of _extract_buff.
     * This is the same memory usage as in credits
     * for Save Data Manager, so that LRAM can be totally
     * used for stuff other than credits and logo.
     */
#   include "BackupMngr/bup_menus.h"
#endif


int logo_load(logo_load_params_t* params)
{
    int i;
    logo_t* logo = LOGO_INFO_PTR;

    unsigned short* logo_ptr = (unsigned short*)LOGO_DATAPTR;

    /* Temporary location where to load compressed logo data.
     * It differs from an application to another:
     *  - Pseudo Saturn Kai : start of HRAM
     *  - Save Data Manager : save data load buffer, in HRAM (331KB)
     */
#ifdef MAKE_PSKAI
    void* logo_data_buff = (void*)0x06004000;
#endif
#ifdef MAKE_BUPMNGR
    void* logo_data_buff = (void*)_extract_buff;
#endif

#ifdef LOGO_DEBUG_ENABLE
    scd_logout("[logo_load]start (datalen:%u path:%s)", params->datalen, params->file_path);
#endif // LOGO_DEBUG_ENABLE
    logo->disabled = 1;

    unsigned long logo_data_len;

    if(params->file_path != NULL)
    {
        /* If needed, load logo from specified media. */
        logo_data_len = scf_size(params->dev_id, params->file_path);
#ifdef LOGO_DEBUG_ENABLE
        scd_logout("[logo_load]data_len:%u, expected:%u", logo_data_len, LOGO_DATALEN);
#endif // LOGO_DEBUG_ENABLE

        if(logo_data_len)
        {
            if(logo_data_len > (SC_COMP_HEADER_MAXSIZE+LOGO_DATALEN))
            {
                logo_data_len = SC_COMP_HEADER_MAXSIZE+LOGO_DATALEN;
            }

            /* If compressed, read to temporary buffer.
             * Else, read directly to VDP1 buffer.
             */
            if(!params->compressed)
            {
                logo_data_buff = (void*)logo_ptr;
            }

            logo_data_len = scf_read(params->dev_id, params->file_path, 0/*offset*/, logo_data_len, (void*)logo_data_buff);
        }
    }
    else
    {
        logo_data_buff = (void*)params->data;
        logo_data_len = params->datalen;
#ifdef LOGO_DEBUG_ENABLE
        scd_logout("[logo_load]load from memory (0x%08X %u)", logo_data_buff, logo_data_len);
#endif // LOGO_DEBUG_ENABLE
    }

    /* Decompress data if needed. */
    if(params->compressed)
    {
        if(logo_data_len > SC_COMP_HEADER_MAXSIZE)
        {
            logo_data_len = decomp_exec((unsigned char*)logo_data_buff, logo_data_len, (unsigned char*)logo_ptr);
#ifdef LOGO_DEBUG_ENABLE
            scd_logout("[logo_load]decomp_exec:%u", logo_data_len);
#endif // LOGO_DEBUG_ENABLE
        }
    }
    else
    {
        if((logo_data_buff != ((void*)logo_ptr))
        && (logo_data_len == LOGO_DATALEN))
        {
            memcpy((void*)logo_ptr, logo_data_buff, logo_data_len);
        }
    }

    /* Verify read and/or decompressed data length.
     * If not as expected, read default logo instead.
     */
#ifdef LOGO_DEBUG_ENABLE
    scd_logout("[logo_load]logo_data_len:%u (%u)", logo_data_len, LOGO_DATALEN);
#endif // LOGO_DEBUG_ENABLE
    if(logo_data_len != LOGO_DATALEN)
    {
        if(!params->default_logo)
        {
#ifdef LOGO_DEBUG_ENABLE
            scd_logout("[logo_load]load default logo");
#endif // LOGO_DEBUG_ENABLE
            return logo_load(_logo_load_params+LOGO_DEFAULT_LOAD_PARAMS);
        }
        else
        {
            /* Default logo load error : this may happen in case of
             * corrupted ROM or executable, etc.
             * In that case, indicate that logo couldn't be loaded.
             */
            return 0;
        }
    }

    /* Enable logo successfully loaded. */
    logo->disabled = 0;
#ifdef LOGO_DEBUG_ENABLE
    scd_logout("[logo_load]logo load success");
#endif // LOGO_DEBUG_ENABLE


    /* Initialize vertex coordinates. */
    logo_anim_start();


    /* Verify if logo uses transparent color or not. */
    logo->transparency_use = 0;
    logo_ptr = (unsigned short*)LOGO_DATAPTR;
    for(i=0; i<LOGO_DATALEN; i+=sizeof(unsigned short))
    {
        /* It would be better to check for higher bit
         * only, but data zero-ness is surely faster to check.
         */
        if((*logo_ptr++) == 0x0000)
        {
            logo->transparency_use = 1;
            break;
        }
    }

    /* Initialize sometimes misc. stuff.
     * The bitfield check below is basically
     * to differentiate behavior between
     * menus and credits ...
     */
    if(logo->pattern & LOGO_FORCE_GOURAUD_USE)
    {
        logo->gouraud_use = 0;
        logo->gouraud_auto_off = 0;
        logo->mesh_use = 0;
        logo->cntr = 0;

        logo_set_pattern(LOGO_DEFAULT_PATTERN);
    }

    return 1;
}




void logo_set_enable(int enable_flag)
{
    logo_t* logo = LOGO_INFO_PTR;
    logo->disabled = (enable_flag ? 0 : 1);
}

int logo_is_enabled(void)
{
    logo_t* logo = LOGO_INFO_PTR;
    return (logo->disabled ? 0 : 1);
}

int logo_is_transparent(void)
{
    logo_t* logo = LOGO_INFO_PTR;
    return logo->transparency_use;
}

int logo_is_mesh_enabled(void)
{
    logo_t* logo = LOGO_INFO_PTR;
    return logo->mesh_use;
}

long logo_get_lum_diff_total(void)
{
    logo_t* logo = LOGO_INFO_PTR;
    return logo->lum_diff_total;
}

void logo_set_pattern(unsigned long pattern)
{
    int i, j;
    logo_t* logo = LOGO_INFO_PTR;

    logo->pattern = pattern & LOGO_PATTERN_MASK;
    logo->anim_type = (pattern >> LOGO_ANIM_BITSHIFT) & LOGO_ANIM_MASK;

#ifdef LOGO_DEBUG_ENABLE
    scd_logout("[logo_set_pattern]logo->pattern:0x%08X (pattern:0x%02X, anim_type:0x%02X)", logo->pattern, pattern, anim_type);
#endif // LOGO_DEBUG_ENABLE

    unsigned char lum_target = 0;

    if(logo->pattern != LOGO_PATTERN_NOCHANGE)
    {
        switch(logo->pattern)
        {
        default:
        case(LOGO_PATTERN_NOMESH):
#ifdef LOGO_DEBUG_ENABLE
            scd_logout("[logo_set_pattern]LOGO_PATTERN_NOMESH");
#endif // LOGO_DEBUG_ENABLE
            logo->gouraud_use = 1;
            logo->gouraud_auto_off = 1;
            logo->mesh_use = 0;
            lum_target = 15;
            break;

        case(LOGO_PATTERN_DARK):
#ifdef LOGO_DEBUG_ENABLE
            scd_logout("[logo_set_pattern]LOGO_PATTERN_DARK");
#endif // LOGO_DEBUG_ENABLE
            logo->gouraud_use = 1;
            logo->gouraud_auto_off = 0;
            logo->mesh_use = 1;
            lum_target = 0;
            break;

        case(LOGO_PATTERN_SHADOW):
#ifdef LOGO_DEBUG_ENABLE
            scd_logout("[logo_set_pattern]LOGO_PATTERN_SHADOW");
#endif // LOGO_DEBUG_ENABLE
            logo->gouraud_use = 1;
            logo->gouraud_auto_off = 0;
            logo->mesh_use = 0;
            lum_target = 0;
            break;
        }
    }

    /* Force usage of Gouraud luminosity variation.
     *
     * This is a tweak to make credits working while
     * keeping behavior of usual logo (in menus, etc)
     * working as until now : I'm too tired to re-do
     * verification of logo transition for all menus ...
     */
    if(pattern & LOGO_FORCE_GOURAUD_USE)
    {
        logo->gouraud_use = 1;
    }

    int force_target = (pattern & LOGO_FORCE_TARGET ? 1 : 0);
    for(j=0; j<(LOGO_TILECNT_Y+1); j++)
    {
        for(i=0; i<(LOGO_TILECNT_X+1); i++)
        {
            vertex_t* v = &(logo->v[i][j]);

            if(logo->pattern != LOGO_PATTERN_NOCHANGE)
            {
                v->lum_target = lum_target;
            }

            if(force_target)
            {
                v->lum = v->lum_target;
                v->x = v->x_target;
                v->y = v->y_target;
            }
        }
    }


    /* Reset luminosity difference to a value large enough.
     * This is required because luminosity may not be updated on each frame.
     */
    if(!force_target)
    {
        logo->lum_diff_total = (1 << 30);
    }

    /* Retrieve fade speed divisor value. */
    if(!force_target)
    {
        logo->fade_divisor = (pattern >> LOGO_FADEDIV_BITSHIFT) & 0xFF;
    }

    /* Fade speed divisor value sanity check. */
    if(logo->fade_divisor == 0)
    {
        logo->fade_divisor = 1;
    }

#ifdef LOGO_DEBUG_ENABLE
    scd_logout("[logo_set_pattern]gouraud_use:%d, auto_off:%d, mesh_use:%d, lum_target:%d, force_target:%d, fade_divisor:%d", 
        logo->gouraud_use, 
        logo->gouraud_auto_off, 
        logo->mesh_use, 
        lum_target, 
        force_target, 
        logo->fade_divisor);
#endif // LOGO_DEBUG_ENABLE
}


void logo_anim_start(void)
{
    int i, j;
    logo_t* logo = LOGO_INFO_PTR;
    unsigned short* logo_ptr = (unsigned short*)LOGO_DATAPTR;
    unsigned short* grad_ptr = (unsigned short*)LOGO_GRADPTR;


    if(logo->anim_type == LOGO_ANIM_POP_CENTER)
    { /* Pop from screen center. */
        for(j=0; j<(LOGO_TILECNT_Y+1); j++)
        {
            for(i=0; i<(LOGO_TILECNT_X+1); i++)
            {
                vertex_t* v = &(logo->v[i][j]);

                v->x = LOGO_WIDTH  / 2;
                v->y = LOGO_HEIGHT / 2;
            }
        }
    }
    else if(logo->anim_type == LOGO_ANIM_NOCHANGE)
    { /* Keep same vertex position. */
    }


    /* Initialize vertex target coordinates. */
    for(j=0; j<(LOGO_TILECNT_Y+1); j++)
    {
        for(i=0; i<(LOGO_TILECNT_X+1); i++)
        {
            vertex_t* v = &(logo->v[i][j]);

            v->x_target = i*LOGO_TILE_W;
            v->y_target = j*LOGO_TILE_H;
        }
    }

    /* Initialize tiles. */
    for(j=0; j<LOGO_TILECNT_Y; j++)
    {
        for(i=0; i<LOGO_TILECNT_X; i++)
        {
            tile_t* t = &(logo->t[i][j]);

            t->v[0] = &(logo->v[i+0][j+0]);
            t->v[1] = &(logo->v[i+1][j+0]);
            t->v[2] = &(logo->v[i+1][j+1]);
            t->v[3] = &(logo->v[i+0][j+1]);
            t->tex_offset = LOGO_TILE_VDP1OFFS(LOGO_TILE_XY_TO_ID(i, j));

            t->grad_ptr = grad_ptr + LOGO_TILE_XY_TO_ID(i, j) * 4;
            t->grad_offset = ((unsigned long)(t->grad_ptr) >> 3) & 0xFFFF;
        }
    }
}


void logo_animate(void)
{
    int i, j;
    logo_t* logo = LOGO_INFO_PTR;

#ifdef LOGO_DEBUG_ENABLE
    scd_msgout( 8, "logo->disabled:%d", logo->disabled);
    scd_msgout( 9, " | gouraud_use:%d", logo->gouraud_use);
    scd_msgout(10, " | gouraud_auto_off:%d", logo->gouraud_auto_off);
    scd_msgout(11, " | fade_divisor:%d", logo->fade_divisor);
#endif // LOGO_DEBUG_ENABLE

    if(logo->disabled)
    {
        return;
    }

#ifdef LOGO_DEBUG_ENABLE
    {// DBG
        vertex_t* v = &(logo->v[0][0]);
        scd_msgout(15, "v->lum_target:%d v->lum:%d", v->lum_target, v->lum);
    }
#endif // LOGO_DEBUG_ENABLE

    /* Adjust fade luminosity speed. */
    int update_luminosity = ((logo->cntr % logo->fade_divisor) == 0 ? 1 : 0);
    long lum_diff_total = 0;

    for(j=0; j<(LOGO_TILECNT_Y+1); j++)
    {
        for(i=0; i<(LOGO_TILECNT_X+1); i++)
        {
            vertex_t* v = &(logo->v[i][j]);
            long d, e;

            /* Update vertex coordinates. */
            d = v->x_target - v->x;
            if(d > 0)
            {
                e = d/8; if(e == 0) { e =  1; }
                v->x += e;
            }
            else if(d < 0)
            {
                e = d/8; if(e == 0) { e = -1; }
                v->x += e;
            }

            d = v->y_target - v->y;
            if(d > 0)
            {
                e = d/8; if(e == 0) { e =  1; }
                v->y += e;
            }
            else if(d < 0)
            {
                e = d/8; if(e == 0) { e = -1; }
                v->y += e;
            }

            /* Update vertex luminosity. */
            if(update_luminosity)
            {
                d = v->lum_target - v->lum;
                if(d > 0)
                {
                    e = d/32; if(e == 0) { e =  1; }
                    v->lum += e;
                }
                else if(d < 0)
                {
                    e = d/32; if(e == 0) { e = -1; }
                    v->lum += e;
                }
                lum_diff_total += ABS(d);
            }
        }
    }

    /* Check if gradient stuff can be disabled. */
    if(update_luminosity)
    {
        logo->lum_diff_total = lum_diff_total;

#ifdef LOGO_DEBUG_ENABLE
        if(lum_diff_total != 0)
        {
            scd_logout("lum_diff_total=%d", lum_diff_total);
        }
#endif // LOGO_DEBUG_ENABLE

        if((lum_diff_total == 0) && (logo->gouraud_auto_off))
        {
#ifdef LOGO_DEBUG_ENABLE
            if(logo->gouraud_use)
            {
                scd_logout("Gouraud automatically disabled.");
            }
#endif // LOGO_DEBUG_ENABLE
            logo->gouraud_use = 0;
        }
    }
    //scd_msgout(1, "logo->gouraud_use=%d", logo->gouraud_use);

    logo->cntr++;
    //scd_msgout(2, "logo->cntr=%d", logo->cntr);
}

void logo_draw(void)
{
    int i, j;
    logo_t* logo = LOGO_INFO_PTR;

    if(logo->disabled)
    {
        return;
    }

    //scd_msgout(4, "[logo]grd:%d mesh:%d", logo->gouraud_use, logo->mesh_use);

    for(j=0; j<LOGO_TILECNT_Y; j++)
    {
        for(i=0; i<LOGO_TILECNT_X; i++)
        {
            tile_t* t = &(logo->t[i][j]);

            if(logo->gouraud_use)
            {
                /* Set Gouraud table for current tile. */
                t->grad_ptr[0] = tbl_lum2col[t->v[0]->lum];
                t->grad_ptr[1] = tbl_lum2col[t->v[1]->lum];
                t->grad_ptr[2] = tbl_lum2col[t->v[2]->lum];
                t->grad_ptr[3] = tbl_lum2col[t->v[3]->lum];
            }

            vdp1_add_distorted_sprite(
                t->v[0]->x  , t->v[0]->y  , 
                t->v[1]->x-1, t->v[1]->y  , 
                t->v[2]->x-1, t->v[2]->y-1, 
                t->v[3]->x  , t->v[3]->y-1, 
                VDP_SPRITE_SIZE(LOGO_TILE_W, LOGO_TILE_H), 
                t->tex_offset, 
                (logo->gouraud_use ? t->grad_offset : 0)/*grad*/, 
                logo->mesh_use);
        }
    }
}


void logo_shadow(short f1x, short f1y, short f2x, short f2y, short radius)
{
    logo_t* logo = LOGO_INFO_PTR;
    int i, j;
    short x, y;
    long d, tmp;
    long radius_square = radius*radius;

    if(logo->disabled)
    {
        return;
    }

    for(j=0; j<(LOGO_TILECNT_Y+1); j++)
    {
        for(i=0; i<(LOGO_TILECNT_X+1); i++)
        {
            vertex_t* v = &(logo->v[i][j]);
            x = i*LOGO_TILE_W;
            y = j*LOGO_TILE_H;

            d = 0;
            tmp = f1x - x; d += tmp*tmp;
            tmp = f1y - y; d += tmp*tmp;
            tmp = f2x - x; d += tmp*tmp;
            tmp = f2y - y; d += tmp*tmp;

            tmp = 15*d;
            tmp = tmp / radius_square;
            if(tmp > 20) {tmp = 20;}

            //v->lum_target = (d<radius_square ? 0 : 20);
            v->lum_target = tmp;
        }
    }

    logo->gouraud_use = 1;
    logo->gouraud_auto_off = 0;
}

void logo_shadow_disable(void)
{
    logo_t* logo = LOGO_INFO_PTR;
    int i, j;

    if(logo->disabled)
    {
        return;
    }

    for(j=0; j<(LOGO_TILECNT_Y+1); j++)
    {
        for(i=0; i<(LOGO_TILECNT_X+1); i++)
        {
            vertex_t* v = &(logo->v[i][j]);
            v->lum_target = 15;
        }
    }

    logo->gouraud_use = 1;
    logo->gouraud_auto_off = 0;
}



#ifdef MAKE_PSKAI

unsigned long bswap_ul(unsigned char* in)
{
    unsigned long out;

    out  = (in[0]) <<  0;
    out |= (in[1]) <<  8;
    out |= (in[2]) << 16;
    out |= (in[3]) << 24;

    return out;
}
unsigned short bswap_us(unsigned char* in)
{
    unsigned short out;

    out  = (in[0]) <<  0;
    out |= (in[1]) <<  8;

    return out;
}
#define RGB555_SET(R,G,B) (((B)<<10) | ((G)<<5) | (R) | 0x8000)


unsigned long customlogo_import(unsigned char dev_id, char* filename)
{
    unsigned char* logo_bmp = (unsigned char*)0x06004000;
    unsigned long datalen;

    /* First, delete the previous logo binary file. */
    scf_delete(SC_FILE_SDCARD, LOGO_PATH);

    /* Read specified BMP file. */
    datalen = scf_size(dev_id, filename);
#ifdef LOGO_DEBUG_ENABLE
    scd_logout("[cl]datalen = %d", datalen);
#endif // LOGO_DEBUG_ENABLE
    if(datalen == 0)
    {
        return 0;
    }
    //scd_logout("[cl]dev_id=%d", dev_id);
    //scd_logout("[cl]filename=\"%s\"", filename);
    datalen = scf_read(dev_id, filename, 0/*offset*/, datalen, (void*)logo_bmp);

    //scd_logout("[cl]datalen = %d", datalen);
    //scd_logout("[cl]logo_bmp %02X %02X %02X %02X %02X %02X %02X %02X", 
    //    logo_bmp[0], logo_bmp[1], logo_bmp[2], logo_bmp[3], 
    //    logo_bmp[4], logo_bmp[5], logo_bmp[6], logo_bmp[7]);

    unsigned char* bmp_header = logo_bmp;
    unsigned long  bmp_w, bmp_h;
    unsigned short bmp_bitcount;

    bmp_w = bswap_ul(bmp_header+18);
    bmp_h = bswap_ul(bmp_header+22);
    bmp_bitcount = bswap_us(bmp_header+28);

    //scd_logout("[cl]bmp_w = %d", bmp_w);
    //scd_logout("[cl]bmp_h = %d", bmp_h);
    //scd_logout("[cl]bmp_bitcount = %d", bmp_bitcount);

    unsigned short* logo_crop = (void*)0x060C0000;
    int i, j, ii, jj;

    if(bmp_bitcount == 8)
    { /* Palette 256 colors. */
        unsigned char* bmp_pallette = logo_bmp + 54;
        unsigned char* bmp_data = bmp_pallette + 256*4;

        /* Read 8 bits DIB image */
        jj=0;
        for(j=bmp_h-1; j>=0; j--)
        {
            ii=0;
            for(i=0; i<(int)bmp_w; i++)
            {
                if((ii < LOGO_WIDTH) && (jj < LOGO_HEIGHT))
                {
                    unsigned char r, g, b;
                    unsigned long index;

                    index = 4*bmp_data[j*bmp_w+i];

                    b = bmp_pallette[index + 0];
                    g = bmp_pallette[index + 1];
                    r = bmp_pallette[index + 2];

                    if((r == 0xFF) && (g == 0x00) && (b == 0xFF))
                    {
                        /* Set pink pixel (#FF00FF) as transparent. */
                        logo_crop[jj*bmp_w+ii] = 0x0000;
                    }
                    else
                    {
                        logo_crop[jj*bmp_w+ii] = RGB555_SET(r>>3, g>>3, b>>3);
                    }
                }
                ii++;
            }
            jj++;
        }
        
    }
    else
    { /* 24/32 bits. */
        unsigned char* bmp_data = logo_bmp + 54;

        jj=0;
        for(j=bmp_h-1; j>=0; j--)
        {
            ii=0;
            for(i=0; i<(int)bmp_w; i++)
            {
                if((ii < LOGO_WIDTH) && (jj < LOGO_HEIGHT))
                {
                    unsigned char r, g, b;
                    unsigned long index;

                    index = (j*bmp_w+i) * (bmp_bitcount == 24 ? 3 : 4);

                    b = bmp_data[index + 0];
                    g = bmp_data[index + 1];
                    r = bmp_data[index + 2];

                    if((r == 0xFF) && (g == 0x00) && (b == 0xFF))
                    {
                        /* Set pink pixel (#FF00FF) as transparent. */
                        logo_crop[jj*bmp_w+ii] = 0x0000;
                    }
                    else
                    {
                        logo_crop[jj*bmp_w+ii] = RGB555_SET(r>>3, g>>3, b>>3);
                    }
                }
                ii++;
            }
            jj++;
        }
    
    }


    /* Tile cropped logo image. */
    unsigned short* logo_ptr = (unsigned short*)LOGO_DATAPTR;
    for(j=0; j<LOGO_HEIGHT; j++)
    {
        for(i=0; i<LOGO_WIDTH; i++)
        {
            unsigned long tile_id = (j/LOGO_TILE_H) * (LOGO_WIDTH/LOGO_TILE_W) + (i/LOGO_TILE_W);
            unsigned long offset = tile_id * LOGO_TILE_W * LOGO_TILE_H;
            unsigned long tile_i = i % LOGO_TILE_W;
            unsigned long tile_j = j % LOGO_TILE_H;

            logo_ptr[offset + tile_j*LOGO_TILE_W + tile_i] = logo_crop[j*LOGO_WIDTH + i];
        }
    }

    /* Compress tiled logo data. */
    unsigned char* comp_ptr = (unsigned char*)0x06004000;
    datalen = comp_exec((unsigned char*)logo_ptr, LOGO_DATALEN, comp_ptr);

    //scd_logout("[cl]comp_datalen = %d", datalen);

    /* Save to SD card. Don't forget to create PSKAI folder before. */
    scf_mkdir(SC_FILE_SDCARD, LOGO_FOLDER);
    scf_write((void*)comp_ptr, datalen, SC_FILE_SDCARD, LOGO_PATH);

    /* Re-init logo stuff. */
    logo_load(_logo_load_params+LOGO_CUSTOM_LOAD_PARAMS);

    return datalen;
}

void customlogo_delete(void)
{
    scf_delete(SC_FILE_SDCARD, LOGO_PATH);

    /* Re-init logo stuff. */
    logo_load(_logo_load_params+LOGO_CUSTOM_LOAD_PARAMS);
}

#endif // MAKE_PSKAI

