/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef LOGO_H
#define LOGO_H

#include "vdp.h"


/* Note : this module is shared between Pseudo Saturn Kai and
 *        Save Data Manager, and application-specific code
 *        is managed with the following #definitions :
 *          - MAKE_BUPMNGR : Pseudo Saturn Kai specific code
 *          - MAKE_PSKAI   : Save Data Manager specific code
 */


/* 16x16 pxl per logo tile. */
#define LOGO_WIDTH  320
#define LOGO_HEIGHT 224
#define LOGO_TILE_W 16
#define LOGO_TILE_H 16

#define LOGO_TILECNT_X (LOGO_WIDTH  / LOGO_TILE_W)
#define LOGO_TILECNT_Y (LOGO_HEIGHT / LOGO_TILE_H)


/* Store logo tile data (RGB555) just before starfield image data.
 * 
 * Note : allocate memory for 352*240 image on VDP1 side, in order
 *        to re-use logo VRAM for Dezaemon 2 SGM screenshot image.
 */
#define LOGO_DATALEN  (LOGO_WIDTH*LOGO_HEIGHT*sizeof(unsigned short))
#define LOGO_ALLOCLEN (352*240*sizeof(unsigned short))
#define LOGO_DATAPTR  (unsigned short*)((unsigned char*)(STARFIELD_DATAPTR) - LOGO_ALLOCLEN)

#define LOGO_TILE_XY_TO_ID(_X_, _Y_) ((_Y_)*LOGO_TILECNT_X + (_X_))
#define LOGO_TILE_VDP1OFFS(_ID_) (((((unsigned long)logo_ptr)+(_ID_)*LOGO_TILE_W*LOGO_TILE_H*sizeof(unsigned short)) & 0xFFFFF)>>3)

/* Pointer to start of gradient data.
 * One gradient array (4 colors, 8 bytes) defined per tile.
 * Gradient data is stored just before logo tile data.
 *
 * Maximum data size for gradient data is set to 4KB so that
 * unused data can be allocated for future use.
 * Note : At 320x224, 16x16 tiles, gradient data size is 0x8C0 bytes, (~2KB).
 */
#define LOGO_GRAD_MAXDATALEN (4*1024)
//#define LOGO_GRADPTR (unsigned short*)((unsigned char*)(LOGO_DATAPTR) - (LOGO_TILECNT_X*LOGO_TILECNT_Y*4*sizeof(unsigned short)))
#define LOGO_GRADPTR (unsigned short*)((unsigned char*)(LOGO_DATAPTR) - LOGO_GRAD_MAXDATALEN)



/**
 * Logo internals initialization.
 *
 * Must be called at the very beginning of the application.
 * (Can be called before initialization of satcom_lib)
 *
 * Application ID is used in order to separate logo behavior
 * from an application to another, and must be specified accordingly.
**/
#define LOGO_APPTYPE_PSKAI   0
#define LOGO_APPTYPE_BUPMNGR 1
void logo_init(int app_type);


/**
 * Logo load parameters, used in logo_load function.
**/
typedef struct _logo_load_params_t
{
    /* File access parameters.
     * Set file_path to NULL when loading from memory.
     */
    unsigned char dev_id;
    char* file_path;

    /* Memory access parameters.
     * Ignored when file_path is not NULL.
     */
    unsigned char* data;
    unsigned long datalen;

    /* Compression flag.
     * Zero: raw image data, else: compressed data.
     * Applies to both file and memory access.
     *
     * Note : temporary buffer used to read compressed data
     *        from file differs from an application to another:
     *         - Pseudo Saturn Kai : HRAM
     *         - Save Data Manager : LRAM
     */
    int compressed;

    /* Indicates if default logo or not. */
    int default_logo;
} logo_load_params_t;


/* List of logo load parameters.
 * Terminating entry has NULL set at both file name and data pointer.
 * First entry is custom logo, read from SD card.
 * Second entry is default logo by Alien.
 */
#define LOGO_CUSTOM_LOAD_PARAMS  0
#define LOGO_DEFAULT_LOAD_PARAMS 1
extern logo_load_params_t _logo_load_params[];


/**
 * Logo data load function.
 *
 * Read, and if needed, decompress logo image data from memory or media.
 * Must be called after initializing SatCom library.
 *
 * In the case specified logo can't be loaded, default one is loaded instead.
 *
 * Returns 1 when logo successfully loaded, zero else.
**/
int logo_load(logo_load_params_t* params);


/**
 * Logo enable/disable stuff.
 *
 * For use when temporary disabling display of logo.
**/
void logo_set_enable(int enable_flag);
int logo_is_enabled(void);


/**
 * Logo shading/shape/etc transition setup.
 * Usage example : greyout logo in order to show text, etc.
**/
#define LOGO_PATTERN_NOCHANGE   0           /* Keep previous pattern.                      */
#define LOGO_PATTERN_DARK       1           /* Logo greyout : show foreground text.        */
#define LOGO_PATTERN_SHADOW     2           /* Shadow on specific portion.                 */
#define LOGO_PATTERN_NOMESH     3           /* Disable polygon mesh : text hardly visible. */
#define LOGO_PATTERN_MASK       0x0000000F
#define LOGO_FORCE_GOURAUD_USE  (1<< 4)     /* Force Gouraud usage : luminosity changes.   */
#define LOGO_FORCE_TARGET       (1<<31)     /* Skip shading/shape transition.              */

/* Logo fade divisor :
 * It is located on pattern's bits 8-15.
 */
#define LOGO_FADEDIV_BITSHIFT 8
#define LOGO_SET_FADEDIV(_VAL_) ((_VAL_) << LOGO_FADEDIV_BITSHIFT)


/* Logo animation type :
 * It is located on pattern's bits 16-23.
 */
#define LOGO_ANIM_BITSHIFT 16
#define LOGO_ANIM_MASK     0x000000FF
#define LOGO_SET_ANIM(_VAL_) ((_VAL_) << LOGO_ANIM_BITSHIFT)

/**
 * The values for animation types.
**/
#define LOGO_ANIM_POP_CENTER    0           /* Pop from screen center.                     */
#define LOGO_ANIM_NOCHANGE      1           /* Keep same vertex position.                  */


/* Pattern redefinitions, for safe use in menus. */
#define LOGOPAT_DISP_SLOW      (LOGO_PATTERN_NOMESH | LOGO_SET_FADEDIV(8) | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))
#define LOGOPAT_DISP_MEDIUM    (LOGO_PATTERN_NOMESH | LOGO_SET_FADEDIV(4) | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))
#define LOGOPAT_DISP           (LOGO_PATTERN_NOMESH | LOGO_SET_FADEDIV(2) | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))
#define LOGOPAT_FORCE_DISP     (LOGO_PATTERN_NOMESH | LOGO_FORCE_TARGET   | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))

#define LOGOPAT_GREYOUT_SLOW   (LOGO_PATTERN_DARK   | LOGO_SET_FADEDIV(8) | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))
#define LOGOPAT_GREYOUT_MEDIUM (LOGO_PATTERN_DARK   | LOGO_SET_FADEDIV(4) | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))
#define LOGOPAT_GREYOUT        (LOGO_PATTERN_DARK   | LOGO_SET_FADEDIV(2) | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))
#define LOGOPAT_FORCE_GREYOUT  (LOGO_PATTERN_DARK   | LOGO_FORCE_TARGET   | LOGO_SET_ANIM(LOGO_ANIM_POP_CENTER))

/* The default pattern. */
#define LOGO_DEFAULT_PATTERN   LOGO_PATTERN_NOMESH

void logo_set_pattern(unsigned long pattern);


/**
 * Logo animation start.
 *
 * Animates logo according to type specified in 
 * `logo_set_pattern' function.
 *
 * This is automatically called in `logo_load' function.
**/
void logo_anim_start(void);


/**
 * Image transparency related stuff.
 *
 * When transparent image is used, starfield is displayed
 * first, and then logo image, so that starfield animation
 * is shown only on transparent regions.
 *
 * When opaque image is used, logo is displayed first, and
 * and then starfield animation, so that both are visible.
**/
int logo_is_transparent(void);


/**
 * Retrieve logo mesh usage flag.
 *
 * Used in order to couple logo and starfield displays.
**/
int logo_is_mesh_enabled(void);

/**
 * Retrieve total luminosity difference with target.
 *
 * Used to wait until luminosity transition finished.
**/
long logo_get_lum_diff_total(void);

/**
 * Logo animation and drawing related function.
 *
 * logo_animate : move each vertex, change gradient, etc
 * logo_draw    : add VDP1 draw commands for logo
**/
void logo_animate(void);
void logo_draw(void);


/**
 * Custom logo import/delete stuff.
 * Used in Pseudo Saturn Kai only.
**/
#ifdef MAKE_PSKAI
unsigned long customlogo_import(unsigned char dev_id, char* filename);
void customlogo_delete(void);
#endif // MAKE_PSKAI

#endif // LOGO_H
