/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MAIN_H
#define MAIN_H

/**
 * Firmware memory usage :
 *
 * HRAM :
 *  06000000 - 06001FFF [    8KB] : Reserved for system.
 *  06002000 - 06003FFF [    8KB] : Stack.
 *  06004000 - 060FFFFF [ 1008KB] : ~~ Multi-purpose buffer #1 ~~
 *                                : > Used at startup to run firmware decompressor code.
 *                                : > Used at startup for backup library buffers (16KB + 8KB).
 *                                : > Used at startup for internal backup memory autodump.
 *                                : > Used when loading demo/flasher/etc from SD card.
 *                                : > Cheat codes configuration (see below for details).
 *                                : > Decompressed credits text.
 * When booting a program :
 *  060FF000 - 060FFFFF [    4KB] : Cheat codes configuration memory (when cheat codes used)
 *                                : This location is always initially used, even when another
 *                                : location is specified.
 *                                : Then, when applying cheat codes this configuration memory
 *                                : is relocated if needed.
 *
 * HRAM is used in order to store cheat codes configuration data :
 *  - 0x060FF000 : default location.
 *  - 0x06002800 : alternate location, in the case default one can't be used.
 * Configuration data is stored into ct_config_t structure (cheat_codes/cheats_data.h).
 * Maximum size for configuration data structure is 4KB (0x1000 bytes), but the smaller
 * the better it is regarding game compatibility.
 * When setting up configuration memory (from Pseudo Saturn Kai, before loading a game), default
 * location is used, and if alternate location is specified in master code, then configuration data
 * is moved from default location to alternate one during soft reset sequence.
 *
 * LRAM :
 *  00200000 - 002EFFFF [  960KB] : Firmware execution area, work RAM, etc.
 *  002F0000 - 002FB7FF [   46KB] : Cheats ROM-less code area. (*1)
 *  002FB800 - 002FF7FF [   16KB] : Stack (see CRT0.S).
 *  002FF800 - 002FFBFF [ 1024 B] : Copystub exec area. (*2) (*3)
 *  002FFC00 - 002FFF9F [  928 B] : Bootstub ROM-less code area. (*1)
 *  002FFFA0 - 002FFFCF [   48 B] : Reserved for future use. (former cheat codes config area)
 *  002FFFD0 - 002FFFDF [   16 B] : Bootstub parameters.
 *  002FFFE0 - 002FFFF7 [   24 B] : Vmem parameters. (*4)
 *  002FFFF8 - 002FFFFF [    8 B] : Debugger global data (pointer, then magic number).
 *
 * (*1) : Used in cartridge ROM-less mode.
 *        Typical application : Pseudo Saturn Kai demo application, alternate rmenu, etc.
 * (*2) : Copystub is declared in satcom_lib and used by Pseudo Saturn Kai
 *        when executing code from SD card or uploaded from USB.
 * (*3) : Data after copystub exec area (0x002FF800), including vmem parameters
 *        and debugger global data aren't modified by copystub.
 * (*4) : Contains Vmem global data (0x002FFFF0, 0x002FFFF4), so care is needed when moving
 *        or modifying this structure.
 *
 * Flash ROM mapping for full firmware :
 * Note : flash ROM available size is 1024KB, from 02000000 to 020FFFFF
 *  02000000 - 0203F3FF [  253KB] : IP, loader, decompressor, compressed firmware (mkfirm tool used).
 *  0203F400 - 0203FBFF [    2KB] : USB dev cart ROM code aka "The Blinking Screen". (64 bytes header, then code itself)
 *  0203FC00 - 0203FEFF [  768 B] : Boostub ROM code. (64 bytes header, then code itself)
 *  0203FF00 - 0203FFFF [  256 B] : Unused flash ROM page for write test before flashing firmware. (256 bytes required for legacy flash carts)
 * 
 *  02040000 - 0205BFFF [  112KB] : Vmem ROM code. (64 bytes header, then code itself)
 *  0205C000 - 02065FFF [   40KB] : Cheat codes ROM code (64 bytes header, then code itself) and data (address/code pairs only).
 *                                  Also includes in-game reset callback and in-game USB connectivity routines.
 *  02066000 - 0209BFFF [  216KB] : ROM code for homebrew applications.
 *  0209C000 - 0209FFFF [   16KB] : Mini Pseudo Saturn launcher. (64 bytes header, then code itself)
 *  020A0000 - 020FDFFF [  376KB] : RomFs data (Backup Memory Manager + etc).
 *                                  Note : RomFs start address is #defined as PSKAI_ROMFS_ADDRESS
 *  020FE000 - 020FFFFF [    8KB] : Unused flash ROM page for write test before flashing firmware. (8KB required for SST39SF0x0 flash chips)
 *
 * Flash ROM mapping for lite firmware :
 * Note : flash ROM available size is 256KB, from 02000000 to 0203FFFF
 *  02000000 - 0203EBFF [  251KB] : IP, loader, decompressor, compressed firmware (mkfirm tool used).
 *  0203EC00 - 0203F3FF [    2KB] : Cheat codes ROM code (64 bytes header, then code itself)
 *                                  Includes in-game reset callback but doesn't includes list of address/code pairs or debug features etc.
 *  0203F400 - 0203FBFF [    2KB] : USB dev cart ROM code aka "The Blinking Screen". (64 bytes header, then code itself)
 *  0203FC00 - 0203FEFF [  768 B] : Boostub ROM code. (64 bytes header, then code itself)
 *  0203FF00 - 0203FFFF [  256 B] : Unused flash ROM page for write test before flashing firmware. (256 bytes required for legacy flash carts)
 * 
 * VDP1 RAM :
 *  05C00000 - 05C1FFFF [  128KB] : VDP1 commands list. 2000 commands x 2 (dual buffering) x 32B/command + some spare room
 *  (33KB spare area, available for future usage.)
 *  05C28400 - 05C34BFF [   50KB] : DZ2SGM entry thumbnails (16 x 16, RGB555, max 100 thumbnails).          [SGM_ENTRY_THBPTR  SGM_ENTRY_THBMAXLEN ]
 *  05C34C00 - 05C54BFF [  128KB] : DZ2SGM list thumbnails (16 x 16, RGB555, max 256 thumbnails).           [SGM_LIST_THBPTR   SGM_LIST_THBMAXLEN  ]
 *  05C54C00 - 05C55BFF [    4KB] : Gradient data for logo. 4KB allocated, but 0x8C0 (~2KB) actually used.  [LOGO_GRADPTR      LOGO_GRAD_MAXDATALEN]
 *  05C55C00 - 05C7EFFF [  165KB] : Logo image (320 x 224, RGB555, tiled), or DZ2SGM screenshot (352 x 240) [LOGO_DATAPTR      LOGO_DATALEN        ]
 *  05C7F000 - 05C7FFFF [    4KB] : Starfield images (4 images, 8*3 pixels each + extra room).              [STARFIELD_DATAPTR STARFIELD_DATALEN   ]
**/

#include <sc_saturn.h>

#include "menu.h"

#include "soft_reset.h"

/* IP header verify/log function. */
#include "ip_header.h"
void ip_bin_display(ip_bin_t* ip);

/* Vmem stuff. */
#include "vmem/vmem_rom_exec.h"

/* Cheat codes stuff. */
#include "cheat_codes/cheat_codes_rom_exec.h"


/* Settings stored in SD card.
 * Contains both SatCom library settings
 * and Pseudo Saturn Kai settings, whose
 * structure is defined in pskai_settings.h .
 */
#include "pskai_settings.h"
extern sc_settings_t _satcom_settings;


/* Compatibility reporter stuff. */
#include "compat_report.h"


/**
 * Log output setting for CD-R loader.
 *  - 0 : disable log output
 *  - 1 : enable log output
 *
 * Enabling log output wastes space on ROM, so this
 * should be only when needed.
**/
#define CD_LOADER_DEBUG 0


/* Various flags to adapt Pseudo Saturn Kai
 * behavior to version of the ROM and/or
 * hardware used.
 */
extern int _pskai_lite;
extern int _pskai_romless;


/* Lite firmware internally enables ROM-less flag, so that this
 * macro below should be used when checking if Pseudo Saturn Kai
 * is running without cartridge ROM or not.
 */
#ifdef PSKAI_LITE_BUILD
#   define PSKAI_IS_ROM_LESS() (0)
#else // !PSKAI_LITE_BUILD
#   define PSKAI_IS_ROM_LESS() (_pskai_romless)
#endif // !PSKAI_LITE_BUILD


/* Yabause debug usage flag. */
extern int _pskai_yabause_debug;

/* Globals for CD-ROM access.
 * Theses are used by Pseudo Saturn exploit
 * and ODE menu modules.
 */
extern u8 _cd_workbuf[4096];
extern u8 *_sect_buffer;
extern ip_bin_t _ip_backup;


// void ipprog(void);


/**
 * show_pskai_title
 *
 * Show application title and version number.
**/
void show_pskai_title(void);

/**
 * show_pskai_license_whatever
 *
 * Show some kind of license/whatever message on screen.
**/
void show_pskai_license_whatever(void);

/**
 * start_game
 * Unlock CDROM and start game.
 *  - START_GAME_NORMAL :
 *     Prompt for ROM hacks, if available.
 *  - START_GAME_QUICK :
 *     Don't use ROM hacks, only show loader selection menu.
 *  - START_GAME_USBDEV :
 *     Don't use ROM hacks, use CWX loader and show USB dev mode menu.
 *  - START_GAME_FIRSTMENU :
 *     Specify that we are loading a game just after Pseudo Saturn Kai boot.
 *     This is used to decide if should display license/whatever message.
 *  - START_GAME_ENABLE_CT :
 *     Start game and enable cheat codes without prompting user about it.
 *     This is used to enable cheat codes from main menu without having
 *     to go through loader selection menu.
 *  - START_GAME_FORCE_CWX :
 *     Force loader to CWX. This is used to avoid boot loader selection menu.
 *  - START_GAME_FORCE_JHL :
 *     Force loader to JHL. This is used to avoid boot loader selection menu.
**/
#define START_GAME_NORMAL    (0 <<  0)
#define START_GAME_QUICK     (1 <<  0)
#define START_GAME_USBDEV    (1 <<  8)
#define START_GAME_FIRSTMENU (1 <<  9)
#define START_GAME_ENABLE_CT (1 << 16)
#define START_GAME_FORCE_CWX (1 << 17)
#define START_GAME_FORCE_JHL (1 << 18)
void start_game(unsigned long sg_flags);

/*
 * Boot type definitions.
 */
#define BOOTLOADER_CWX  0
#define BOOTLOADER_AUTO 1
#define BOOTLOADER_JHL  2
#define BOOTLOADER_SDM  3


/*
 * ROM Hacks definitions.
 */
#define ROMHACK_OFF       0
#define ROMHACK_ARCODES   (1 << 0)
#define ROMHACK_VMEM_ASK  (1 << 1)
#define ROMHACK_VMEM_AUTO (1 << 2)


/**
 * is_audio_cd
 *
 * Return 1 if audio CD is detected, zero else.
**/
int is_audio_cd(int onscreen_log);

void start_homebrew(void);
void settings(void);
void extra(void);
void bup_mngr(int vmem_dbg_mode);
void exit_menu(void);

void usb_firmware(void);

void debugger_mode(void);
void dump_system_info(void);
void system_info(void);

void gui_clear_window(void);


/**
 * wait_for_start_delta.
 * Prompt user to push start button.
 * If msg parameter is set to NULL, 
 * defaut prompt message is displayed.
 *
 * Returns code of button(s) pushed.
 *
 * Example :
 * | wait_for_start_delta(NULL) && PAD_C
 * |  -> pushed C button
**/
unsigned short wait_for_start_delta(char* msg);

/**
 * wait_for_start.
 * Prompt user to push start button.
 * If msg parameter is set to NULL, 
 * defaut prompt message is displayed.
 *
 * Returns 1 if A/C/Start button pushed, zero else.
**/
int wait_for_start(char* msg);

/* Background animation, used in menu display. */
void background_vblank_clbk(void);

/* Show timeout in loader/ROM hack menus.
 * Shared with ODE image selection menu.
 */
extern int _boot_timeout_lite_disp;
extern int _boot_timeout_help_disp;
void bootmenu_timeout_clbk(int second);

/* Screen manipulation sub routines. */
extern char* empty_line;
void clear_screen(void);

// dirty fix for using rand() function.
int rand (void);


/**
 * Display menu and ask user which ROM Hack to use.
 *
 * Additionally, a "dummy" function is provided when cheat
 * codes feature is not available, i.e. when using JHL loader.
 * This is used to prompt user to continue booting the CD-ROM or 
 * returning to main menu, in a similar way as when using CWX loader.
 *
 * Returns 1 if can continue booting CD-ROM, zero else.
**/
int romhack_prompt(ip_bin_t* ip, int* rh_vmem_use, int rh_prompt_type);
int dummy_romhack_prompt(ip_bin_t* ip, int rh_prompt_type);

/**
 * Apply vmem ROM Hack.
 *
 * VMEM_FLAG_DEFAULT : 
 *  Uses vmem with "replace" type initialization.
 * VMEM_FLAG_EXTDEVICE : 
 *  Enable extended (SD card & BRAM) backup devices.
 * VMEM_FLAG_NO_FDD : 
 *  Disable usage of virtual FDD.
**/
#define VMEM_FLAG_DEFAULT   0x00
#define VMEM_FLAG_EXTDEVICE 0x01
#define VMEM_FLAG_NO_FDD    0x02
int vmem_init(ip_bin_t* ip, unsigned long vmem_init_settings);


#endif
