/**
 *  Mini Pseudo Saturn Loader
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _MINIPSEUDO_RC_
#define _MINIPSEUDO_RC_

/* Required in order to get definition of sc_rom_code_header_t structure. */
#include <stdio.h>
#include "../../satcom_lib/sc_rom_code.h"


/* ROM code header and entry point definitions.
 * How to use :
 *  1. Verify that ROM code is available with rc_minipseudo_romcode_check macro
 *  2. (Option) verify CRC32 integrity of ROM code data
 *  3. Copy code from ROM to RAM
 *      FROM: MINIPSEUDO_DATA_START
 *      TO  : MINIPSEUDO_EXEC_START
 *      LEN : MINIPSEUDO_DATA_LEN
 *  4. Execute copied data from MINIPSEUDO_EXEC_START address
 * 
 * Sample code :
 * |     if(rc_minipseudo_romcode_check())
 * |     {
 * |         unsigned char* minips_src = (unsigned char*)MINIPSEUDO_DATA_START;
 * |         unsigned char* minips_dst = (unsigned char*)MINIPSEUDO_EXEC_START;
 * |         unsigned long  minips_len = MINIPSEUDO_DATA_LEN;
 * | 
 * |         if(minips_len <= MINIPSEUDO_MAX_LEN)
 * |         {
 * |             crc32 = crc32_calc(minips_src, minips_len);
 * | 
 * |             if(crc32 == MINIPSEUDO_DATA_CRC)
 * |             {
 * |                 memcpy((void*)minips_dst, (void*)minips_src, MINIPSEUDO_DATA_LEN);
 * | 
 * |                 unsigned int minipseudo_ep[RC_MINIPSEUDO_ENTRIES_COUNT];
 * |                 rc_minipseudo_get_entry_points(minipseudo_ep, RC_MINIPSEUDO_ENTRIES_COUNT);
 * | 
 * |                 int (*minipseudo_ep_go)(unsigned long);
 * |                 minipseudo_ep_go = (int (*)(unsigned long))(minipseudo_ep[MINIPSEUDO_GO_EP_ID]);
 * | 
 * |                 int ret = minipseudo_ep_go(MINIPSEUDO_COPY | MINIPSEUDO_EXEC | MINIPSEUDO_DISPLAY);
 * |             }
 * |         }
 * |     }
 * 
 * Notes :
 *  - Code location and length are designed to fit into two flash sectors
 *    for SST39SF040 device = 2 * 4KB bytes per sector.
 */
#define MINIPSEUDO_HDR_START (0x03800000)
#define MINIPSEUDO_HDR_ADDR ((sc_rom_code_header_t*)MINIPSEUDO_HDR_START)

#define MINIPSEUDO_MAX_LEN   (16*1024)
#define MINIPSEUDO_DATA_LEN (MINIPSEUDO_HDR_ADDR)->usedlen

#define MINIPSEUDO_DATA_CRC (MINIPSEUDO_HDR_ADDR)->crc32

#define MINIPSEUDO_DATA_START (MINIPSEUDO_HDR_START + sizeof(sc_rom_code_header_t))
#define MINIPSEUDO_EXEC_START (0x060F8000)


/**
 * Function     : rc_minipseudo_romcode_check
 * Description  : Verify if vmem ROM code is present or not.
 * Returns      : 0: can't use minipseudo.
 *                else: can use minipseudo.
**/
#define rc_minipseudo_romcode_check() ((MINIPSEUDO_HDR_ADDR->magic_42 == 0x42) && (MINIPSEUDO_HDR_ADDR->magic_53 == 0x53) ? 1 : 0)


/**
 * Function     : rc_minipseudo_get_entry_points
 * Description  : Retrieve ROM code entry point(s).
 * Note         : Before calling, it is required to copy ROM code to RAM.
 * Parameter #1 : unsigned int* entry_points_ptr : array used to store entry points
 * Parameter #2 : unsigned long entry_points_count : size of entry_points_ptr array
 * Note #1      : The entry points array should be sized to RC_MINIPSEUDO_ENTRIES_COUNT.
**/
#define RC_MINIPSEUDO_ENTRIES_COUNT 1
#define rc_minipseudo_get_entry_points(_EP_, _EPC_) (((void (*)(unsigned int* entry_points_ptr, unsigned long entry_points_count))MINIPSEUDO_EXEC_START)((_EP_), (_EPC_)))

/**
 * ROM code entry point #0 : minipseudo_ep_go
 * Execute minimalistic Pseudo Saturn.
 *
 * Input parameter #1: unsigned long config
 *      Configuration flags, for use in entry point first parameter.
 *       - Bit  0: copy main executable
 *       - Bit  1: load main executable
 *       - Bit 31: display information and eventually error message(s)
 *
 * Return parameter: int
 *      Error code
 *       - Zero: no error
**/
#define MINIPSEUDO_GO_EP_ID 0

/* Input parameter flags. */
#define MINIPSEUDO_COPY    (1 <<  0)
#define MINIPSEUDO_EXEC    (1 <<  1)
#define MINIPSEUDO_DISPLAY (1 << 31)

/* Return error codes. */
#define MINIPSEUDO_RET_OK  (0)


#endif /* _MINIPSEUDO_RC_ */
