/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#ifndef PSKAI_SETTINGS_H
#define PSKAI_SETTINGS_H

/* SC_BUP_EXTRA_LEN definition. */
#include <sc_saturn.h>


/* Pseudo Saturn Kai settings.
 * Theses are part of SatCom settings, 
 * read from SD card, and consequently set
 * to default value when it is not available.
 *
 * Settings are located into two locations :
 *  - SD card
 *  - Internal backup memory
 *
 * [SatCom Settings]
 * + link settings
 * + log to SD card
 * + [Pseudo Saturn Kai settings]
 *   + Autodump, etc settings
 *   + [Internal backup memory settings]
 *     + [ODE settings]
 *     + [ODE usage history]
 *     + [link settings] (in the case SD card is not available)
 *
 * Characteristics of each settings :
 *  - SatCom settings are used for all applications using
 *    SatCom, and not only Pseudo Saturn Kai.
 *    Example : "log to SD card", set on Pseudo Saturn Kai, and
 *              and used on vmem or Save Data Manager side.
 *
 *  - Pseudo Saturn Kai settings are used for cartridge specific
 *    settings, such as autodump.
 *
 *  - Internal backup memory settings are used for hardware
 *    specific settings, such as ODE type.
 */


 
/* Initialization status.
 * Prevents from using incompatible settings.
 *
 *  - 0 : unintialized
 *  - 1 : init done (pskai 6.037)
 *  - 2 : init done (pskai 6.064)
 *  - 3 : init done (pskai 6.210)
 *  - 4 : init done (pskai 6.227)
 *  - 5 : init done (pskai 6.240)
 *  - 6 : init done (latest version)
 */
#define PSKAI_SETTINGS_INITVAL 5




#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

#define BRAMSET_NAME "PSEUDO__SAT"
typedef struct _bram_settings_t
{
    /* Initialization status.
     * Prevents from using incompatible settings.
     */
    unsigned char init;

    /* Link and Action Replay types used.
     * Theses are usually read from SD card, but
     * may be initialized from internal backup memory
     * in the case SD card is not available, as on
     * Action Replay cartridge.
     * 
     *  - SC_LINK_NONE
     *  - SC_LINK_LEGACY
     *  - SC_LINK_USBDC
     *
     *  - SC_ARP_EMS
     *  - SC_ARP_DATEL
     */
    unsigned char link_type;
    unsigned char arp_type;

    /* Optical Disk type.
     *  - SET_ODTYPE_NORMAL  : the normal Saturn disk drive
     *  - SET_ODTYPE_RHOEBE  : Rhea/Phoebe ODE
     *  - SET_ODTYPE_FENRIR  : Fenrir ODE
     */
#define SET_ODTYPE_NORMAL  0
#define SET_ODTYPE_RHOEBE  1
#define SET_ODTYPE_FENRIR  2
#define SET_ODTYPE_CNT     3
    unsigned char od_type;

    /* ODE menu names prettify.
     *  - 0 : keep names as-is.
     *  - 1 : automatically convert to lower case when possible.
     */
    unsigned char ode_menu_prettify;

    /* ODE menu entries sort type.
     *  - Alphabetically sort titles.
     *  - Sort by release date.
     *  - Sort by folder ID.
     *
     * Note : to ease the implementation of ODE menu, it is required to
     *        define folder ID sort type as the last available one.
     */
#define SET_ODE_MENU_TITLE_SORT  0
#define SET_ODE_MENU_DATE_SORT   1
#define SET_ODE_MENU_FOLDER_SORT 2
#define SET_ODE_MENU_SORT_CNT    3
    unsigned char ode_menu_sort;

    /* Ascending/descending listing in ODE menu.
     * This is used in combination with sort type to
     * get user's favorite sort and titles listing.
     */
#define SET_ODE_MENU_LIST_ASCEND  0
#define SET_ODE_MENU_LIST_DESCEND 1
#define SET_ODE_MENU_LIST_CNT     2
    unsigned char ode_menu_list;

    /* Text justification in ODE menu.
     * It can be either centered or left-justfied.
     */
#define SET_ODE_MENU_JUSTIFY_CENTER 0
#define SET_ODE_MENU_JUSTIFY_LEFT   1
#define SET_ODE_MENU_JUSTIFY_CNT    2
    unsigned char ode_menu_justify;


    /* Last selected item information :
     *  - Folder ID
     *  - Description string CRC
     *  - (spare byte for padding and in case of)
     */
    unsigned long  ode_menu_last_crc;
    unsigned short ode_menu_last_imgid;


    /* Game information enable/disable setting.
     *  - 0 : show game information at the bottom of the screen.
     *  - 1 : hide game information, but instead image selection menu is a bit taller.
     */
    unsigned char ode_no_game_info;


    /* Style of image disc selection menu.
     *  - Folder : show only folder name
     *  - Title  : show only game title
     *  - Folder + Title : show both folder name and game title
     */
#define ODE_MENU_LIST_FOLDER_ONLY   0
#define ODE_MENU_LIST_TITLE_ONLY    1
#define ODE_MENU_LIST_FOLDER_TITLE  2
#define ODE_MENU_LIST_STYLE_CNT     3
    unsigned char ode_menu_list_style;


    /* Soft reset (A+B+C+Start) behavior.
     * SET_SOFTRES_MULTIPLAYER :
     *        Keep soft reset as-is to
     *        Saturn's multiplayer screen.
     * SET_SOFTRES_REBOOT :
     *        Patch soft reset in order
     *        reboot Saturn, hence returning
     *        to Pseudo Saturn Kai's main menu.
     */
#define SET_SOFTRES_REBOOT      0
#define SET_SOFTRES_MULTIPLAYER 1
    unsigned char softreset;

    /* Automatic boot usage flag :
     *  - 0 : don't use boot timeout feature.
     *  - 1 : automatically start game few seconds
     *        after loader selection menu display.
     */
    unsigned char autoboot;

    /* Display type in menu.
     *  - Monospace    : align characters of each image names.
     *  - Condensed    : variable width font and condensed item height.
     *  - Balanced     : variable width font, item height a bit taller
     *                   than condensed setting.
     *
     * This is currently used in ODE menu only.
     */
#define SET_MENU_DISP_BALANCED  0
#define SET_MENU_DISP_MONOSPACE 1
#define SET_MENU_DISP_CONDENSED 2
#define SET_MENU_DISP_CNT       3
    unsigned char menu_disp_type;

    /* Menu color theme : setup the colors used in menu.
     * Color theme IDs are defined in menu module.
     *
     * This is currently used in ODE menu only.
     */
    unsigned char menu_color_theme;


    /* Menu font ID : setup wich font to use in menu.
     *  - Zero : automatic font selection
     *  - Else : use "font ID - 1"
     *
     * This is currently used in ODE menu only.
     */
    unsigned char menu_font_id;

    /* Extra space for future use.
     * This structure must be packed to 28 bytes so
     * that it uses only one backup memory block.
     */
    unsigned char reserved9[28 - 21];
} bram_settings_t;


typedef struct _pskai_settings_t
{
    /* Initialization status.
     * Prevents from using incompatible settings.
     */
    unsigned char init;

    /* Compatibility reporter related.
     * Enable : enable/disable switch in settings
     * In use : set when temporary report data
     *          is written to SD card.
     */
    unsigned char compatrep_in_use;

    /* Internal memory autodump.
     * When enabled, archive saves from internal
     * memory to SD card on firmware startup.
     *  - 0 : disabled
     *  - 1 : enabled
     */
    unsigned char autodump;

    /* Extra space for future use. */
    unsigned char reserved9[SC_BUP_EXTRA_LEN - 3 - sizeof(bram_settings_t)];

    /* Settings from internal backup memory.
     * Theses are Saturn hardware specific (such as
     * mod settings, etc), and consequently stored
     * there instead of in cartridge' SD card.
     */
    bram_settings_t bram;
} pskai_settings_t;

#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


/**
 * pskai_settings_clear
 * pskai_settings_read
 * pskai_bram_settings_write
 * pskai_settings_write
 * 
 * Manage settings on both SD card and internal backup memory.
**/
void pskai_settings_clear(void);

void pskai_settings_read(void);

void pskai_bram_settings_write(void);
void pskai_settings_write(void);




/* Pseudo Saturn Kai extra settings.
 * As the name suggests, theses are extra settings whose
 * are not indispensable for Pseudo Saturn Kai normal behavior.
 *
 * Theses settings are currently stored into Saturn Internal
 * backup memory only, and any eventual extension to SD card
 * shall be done on each related features rather than this
 * settings access module.
 *
 * It currently gathers settings for the following features :
 *  - Custom loader settings
 *  - Manually set cheat codes
 */
#include "cheat_codes/cheats_codes.h"

#define BRAM_EXTRASET_NAME "PSEUDO_XTRA"

typedef struct _pskai_extra_settings_t
{
    /* Initialization status.
     * Prevents from using incompatible settings.
     *  - 0 : initialization done (latest version)
     */
    unsigned char init;
    /* Extra bytes for version related things. */
    unsigned char version_reserved[2];

    /* ------------------------------------------------------------------------
     * --- CUSTOM LOADER SETTINGS                                           ---
     * ------------------------------------------------------------------------
     */
    /* Custom loader entries count.
     * Memory usage : 4 bytes per entries, plus 8 bytes common to all entries
     * Maximum supported : 32 entries
     */
#define CL_ENTRIES_CNT 10
    /* ID where to write next custom loader settings. */
    unsigned char cl_write_id;

    /* Game loader bitfield.
     *  - 0 : Use JHL loader
     *  - 1 : Use CWX loader
     */
    unsigned long cl_cwx_loader_bf;

    /* Vmem usage bitfield.
     *  - 0 : Vmem disabled
     *  - 1 : Vmem enabled
     */
    unsigned long cl_vmem_use_bf;

    /* IP header CRC for each entries.
     * Set to IPHEADER_DUMMY_CRC in the case entry is unused.
     */
    unsigned long cl_ip_crc[CL_ENTRIES_CNT];


    /* ------------------------------------------------------------------------
     * --- USER CHEAT CODES                                                 ---
     * ------------------------------------------------------------------------
     */
    /* Code informations from master (`F') code :
     *  - Patch address
     *  - Patch opcode
     */
    unsigned long  ct_master_address;
    unsigned short ct_master_opc;

    /* User-defined enable ('D') code.
     * This code is initially stored as regular code (ct_code_t structure), 
     * but stored separately when using custom codes.
     */
    unsigned short ct_enable_val;
    unsigned long  ct_enable_addr;

    /* Configuration memory location from `B' code.
     * This defines address where to store cheat codes
     * interrupt handler code and codes cache.
     */
    unsigned long ct_cfg_location;

    /* Codes cache capacity, from `B' code.
     * Unit : 4 codes (32 bytes)
     * (0: no cache, 4: up to 4 codes, 2: up to 8 codes, etc)
     */
    unsigned char ct_cache_4_capacity;

    /* Reserved for data alignment. */
    unsigned char ct_reserved[3];

    /* Cheat codes count.
     * Memory usage : 8 bytes per code.
     */
#define BRAM_CT_CODES_CNT 16
    ct_code_t ct_list[BRAM_CT_CODES_CNT];


    /* ------------------------------------------------------------------------
     * --- RESERVED (SPARE) AREA                                            ---
     * ------------------------------------------------------------------------
     */

    /* Pseudo Saturn Kai extra settings is designed to use
     * 4 memory blocks, which is equal to 200 bytes available
     * on Pseudo Saturn Kai side.
     * Strictly speaking, it is equal to 202 bytes, but this is
     * rounded down to 4 bytes multiple in order to avoid
     * messy data alignment problems.
     * 
     * Main reason of this size being set like this is to make Pseudo Saturn Kai
     * settings (base ones and these extra ones) taking less than 1% of
     * Saturn internal backup memory, thus hopefully not causing problems with
     * games requiring nearly whole internal memory to work.
     * 
     * Block #1 :                       (22 user data bytes available)
     *  1) Header, Name, Language
     *  2) Comment, Date, Size (upper half)
     *  2) Size (lower half), Alloc Table (2x4 bytes), User Data (6 bytes)
     *  3) User Data (16 bytes)
     * 
     * Block #2~ :                      (60 user data bytes available)
     *  1) Header, User Data (12 bytes)
     *  2) user data (16 bytes)
     *  2) user data (16 bytes)
     *  3) user data (16 bytes)
     */
#define BRAM_EXTRASET_BKCNT 4
    unsigned char reserved9
    [
        200                                                 /* Total size.    */
         - 3                                                /* Version info.  */
         - 1 - (2 + CL_ENTRIES_CNT)*sizeof(unsigned long)   /* Custom loader. */
         - (20 + BRAM_CT_CODES_CNT * sizeof(ct_code_t))     /* User cheats.   */
    ];
} pskai_extra_settings_t;

extern pskai_extra_settings_t _extra_settings;


/**
 * pskai_extraset_read
 * pskai_extraset_write
 * 
 * Access extra settings on Saturn internal backup memory.
 * Implemented in extra_settings.c
**/
void pskai_extraset_read(void);
void pskai_extraset_write(void);


#endif // PSKAI_SETTINGS_H
