/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "soft_reset.h"

#include <string.h>
#include <sc_saturn.h>



#ifdef MAKE_PSKAI

/**
 * backup_vdp1_ram / restore_vdp1_ram
 *
 * Backup VDP1 RAM contents so that they can be restored when needed.
 * Reason to do that is to improve compatibility for games relying on VDP1 RAM
 * initialized by Saturn BIOS.
 * -> Because usage of this feature makes sense when loading a game, it is
 *    available only in Pseudo Saturn Kai.
 *
 * VDP1 RAM size is 512KB, which is too large to be backed up on Pseudo Saturn
 * Kai work memory. As a countermeasure, unused region of VDP2 RAM is used as
 * backup area. However available size there is "only" 256KB so remaining half
 * of VDP1 RAM is saved on work RAM (LRAM).
 *
 * This way of doing monopolizes 256KB of RAM for SH-2, which is quite a lot but
 * not a problem right now. In the future, a trick to optimize memory usage
 * could be the usage of LZF compression and use the space gained from that as
 * a memory pool.
 * As memory pool -> dynamic memory allocation -> no guarantee of allocation, 
 * this would be an optional memory storage, for example to cache data in disc
 * image selection menu etc.
**/
int _vdp1_ram_saved = 0;

/* Address of each VDP1 quarters. */
#define VDP1_Q1 (VDP1_VRAM + (  0 * 1024))
#define VDP1_Q2 (VDP1_VRAM + (128 * 1024))
#define VDP1_Q3 (VDP1_VRAM + (256 * 1024))
#define VDP1_Q4 (VDP1_VRAM + (384 * 1024))


/* Buffer handling Q1 and Q2. */
unsigned char _vdp1_ram2[256*1024];

/* The place where Q3 and Q4 are backed up. */
#define VDP2_BACKUP_ADDR (VDP2_VRAM + (256*1024))


void backup_vdp1_ram(void)
{
    /* Backup VDP1 RAM to SH-2 RAM. */
    memcpy((void*)_vdp1_ram2, (void*)VDP1_Q1, 256*1024);

    /* Backup VDP1 RAM to VDP2 RAM. */
    memcpy((void*)VDP2_BACKUP_ADDR, (void*)VDP1_Q3, 256*1024);

    /* Indicate that a backup of VDP1 RAM is available. */
    _vdp1_ram_saved = 1;
}

void restore_vdp1_ram(void)
{
    if(_vdp1_ram_saved)
    {
        /* Restore VDP1 RAM from SH-2 RAM. */
        memcpy((void*)VDP1_Q1, (void*)_vdp1_ram2, 256*1024);

        /* Restore VDP1 RAM from VDP2 RAM. */
        memcpy((void*)VDP1_Q3, (void*)VDP2_BACKUP_ADDR, 256*1024);
    }
}

#else // !MAKE_PSKAI

void reset_vdp1_ram(unsigned short* ptr, int count16)
{
    int i, j;

    for(i=(count16/16); i; i--)
    {
        *ptr++ = 0x8000;

        for(j=15; j; j--)
        {
            *ptr++ = 0x0000;
        }
    }
}

#endif // !MAKE_PSKAI


void soft_reset_init(void)
{
#ifdef MAKE_PSKAI
    /* Backup VDP1 RAM contents. */
    backup_vdp1_ram();
#endif // MAKE_PSKAI
}


void hram_reset(void)
{
    memset((void*)HRAM_BUFFER_ADR, 0, HRAM_BUFFER_LEN);
}

void soft_reset(unsigned long softres_flags)
{
    unsigned short *p;

    /* Don't forget to stop pad interrupts. */
    pad_shutdown();

    if(softres_flags & SC_SOFTRES_VDP2RAM)
    {
        /* Reset VDP2 VRAM/CRAM.
         * Note : as Pseudo Saturn Kai uses second half of VDP2 RAM to backup
         *        VDP1 RAM, only first half is cleared now.
         */
        memset((void*)VDP2_VRAM, 0, 256*1024); // 0x25E00000
        memset((void*)VDP2_CRAM, 0,   4*1024); // 0x25F00000
    }

    if(softres_flags & SC_SOFTRES_VDP2REGS)
    {
        /* Reset VDP2 regs.
         * Note : registers not modified by Pseudo Saturn Kai are commented out.
         */
        TVMD   = 0x8000;
        // EXTEN  = 0x0000;
        // TVSTAT = 0x0202;
        // VRSIZE = 0x0000;
        // HCNT   = 0x0000;
        // VCNT   = 0x00F1;
        RAMCTL = 0x0000;
        // CYCA0L = 0x0F44;
        // CYCA0U = 0xFFFF;
        // CYCA1L = 0xFFFF;
        // CYCA1U = 0xFFFF;
        // CYCB0L = 0xFFFF;
        // CYCB0U = 0xFFFF;
        // CYCB1L = 0xFFFF;
        // CYCB1U = 0xFFFF;
        BGON   = 0x0001;
        // MZCTL  = 0x0000;
        // SFSEL  = 0x0000;
        // SFCODE = 0x0000;
        CHCTLA = 0x0000;
        // CHCTLB = 0x0000;
        // BMPNA  = 0x0000;
        // BMPNB  = 0x0000;
        // PNCN0  = 0x8000;
        // PNCN1  = 0x0000;
        // PNCN2  = 0x0000;
        // PNCN3  = 0x0000;
        // PNCR   = 0x0000;
        // PLSZ   = 0x0000;
        MPOFN  = 0x0000;
        // MPOFR  = 0x0000;
        // MPABN0 = 0x0303;
        // MPCDN0 = 0x0303;
        // MPABN1 = 0x0000;
        // MPCDN1 = 0x0000;
        // MPABN2 = 0x0000;
        // MPCDN2 = 0x0000;
        // MPABN3 = 0x0000;
        // MPCDN3 = 0x0000;
        // MPABRA = 0x0000;
        // MPCDRA = 0x0000;
        // MPEFRA = 0x0000;
        // MPGHRA = 0x0000;
        // MPIJRA = 0x0000;
        // MPKLRA = 0x0000;
        // MPMNRA = 0x0000;
        // MPOPRA = 0x0000;
        // MPABRB = 0x0000;
        // MPCDRB = 0x0000;
        // MPEFRB = 0x0000;
        // MPGHRB = 0x0000;
        // MPIJRB = 0x0000;
        // MPKLRB = 0x0000;
        // MPMNRB = 0x0000;
        // MPOPRB = 0x0000;
        SCXIN0 = 0x0000;
        SCXDN0 = 0x0000;
        SCYIN0 = 0x0000;
        // SCYDN0 = 0x0000;
        // ZMXN0  = 0x00010000;
        // ZMYN0  = 0x00010000;
        // SCXIN1 = 0x0000;
        // SCXDN1 = 0x0000;
        // SCYIN1 = 0x0000;
        // SCYDN1 = 0x0000;
        // ZMXN1  = 0x00010000;
        // ZMYN1  = 0x00010000;
        // SCXN2  = 0x0000;
        // SCYN2  = 0x0000;
        // SCXN3  = 0x0000;
        // SCYN3  = 0x0000;
        // ZMCTL  = 0x0000;
        // SCRCTL = 0x0000;
        // VCSTA  = 0x00000000;
        // LSTA0  = 0x00000000;
        // LSTA1  = 0x00000000;
        // LCTA   = 0x00000000;
        // BKTAU  = 0x0000;
        // BKTAL  = 0x4000;
        // RPMD   = 0x0000;
        // RPRCTL = 0x0000;
        // KTCTL  = 0x0000;
        // KTAOF  = 0x0000;
        // OVPNRA = 0x0000;
        // OVPNRB = 0x0000;
        // RPTA   = 0x00000000;
        // WPSX0  = 0x0000;
        // WPSY0  = 0x0000;
        // WPEX0  = 0x0000;
        // WPEY0  = 0x0000;
        // WPSX1  = 0x0000;
        // WPSY1  = 0x0000;
        // WPEX1  = 0x0000;
        // WPEY1  = 0x0000;
        // WCTLA  = 0x0000;
        // WCTLB  = 0x0000;
        // WCTLC  = 0x0000;
        // WCTLD  = 0x0000;
        // LWTA0  = 0x00000000;
        // LWTA1  = 0x00000000;
        // SPCTL  = 0x0020;
        // SDCTL  = 0x0000;
        // CRAOFA = 0x0000;
        // CRAOFB = 0x0000;
        // LNCLEN = 0x0000;
        // SFPRMD = 0x0000;
        // CCCTL  = 0x0000;
        // SFCCMD = 0x0000;
        PRISA  = 0x0000;
        PRISB  = 0x0000;
        PRISC  = 0x0000;
        PRISD  = 0x0000;
        // PRINA  = 0x0007;
        // PRINB  = 0x0000;
        // PRIR   = 0x0000;
        // CCRSA  = 0x0000;
        // CCRSB  = 0x0000;
        // CCRSC  = 0x0000;
        // CCRSD  = 0x0000;
        // CCRNA  = 0x0000;
        // CCRNB  = 0x0000;
        // CCRR   = 0x0000;
        // CCRLB  = 0x0000;
        // CLOFEN = 0x0001;
        // CLOFSL = 0x0000;
        // COAR   = 0x0200;
        // COAG   = 0x0200;
        // COAB   = 0x0200;
        // COBR   = 0x0000;
        // COBG   = 0x0000;
        // COBB   = 0x0000;
    }

    if(softres_flags & SC_SOFTRES_VDP1REGS)
    {
        /* Reset VDP1 regs.
         * Note : registers not modified by Pseudo Saturn Kai are commented out.
         */
        TVMR = 0x0000;
        FBCR = 0x0000;
        PTMR = 0x0000;
        EWDR = 0x0000;
        EWLR = 0x0000;
        EWRR = 0x0000;
        // ENDR = 0x0000;
    }

    if(softres_flags & SC_SOFTRES_VDP1RAM)
    {
        /* Set first command to END, then force end of VDP1 drawing.
         * (= write to "Draw Forced Termination Register")
         */
        p = (unsigned short*)VDP1_VRAM; // 0x25c00000
        *p = 0x8000;
        ENDR = 0x0000;


        /* Reset all VDP1 commands table / VRAM :
         * If available, restore RAM to its original state.
         * Else, do a cheap clear of the RAM which is OK for normal usage, but
         * but problematic to correctly load some games.
         */
#ifdef MAKE_PSKAI
        restore_vdp1_ram();
#else
        reset_vdp1_ram((unsigned short*)VDP1_VRAM, (512*1024)/sizeof(unsigned short)); // 0x25c00000
#endif // MAKE_PSKAI

        /* Reset VDP1 frame buffer. */
        memset((void*)VDP1_FRAMEBUFFER, 0, 256*1024); // 0x25C80000
    }

    if(softres_flags & SC_SOFTRES_VDP2RAM)
    {
        /* Now that VDP1 RAM is restored, clear second half of VDP2 VRAM. */
        memset((void*)(VDP2_VRAM+(256*1024)), 0, 256*1024); // 0x25E00000

#ifdef MAKE_PSKAI
        /* Indicate that backup of VDP1 RAM is no longer available. */
        _vdp1_ram_saved = 0;
#endif // MAKE_PSKAI
    }
}


/* Extra wrapper for "exit to multiplayer" and "system reset" functions. */
#include "ode_fenrir.h"
#include "pskai_settings.h"

extern sc_settings_t _satcom_settings;

/* Exit to multiplayer. */
void pseudo_exit(void)
{
    sc_exit();
}

/* System reset. */
void pseudo_sysres(void)
{
#ifndef MAKE_WASCALOADER
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);

    if(set->bram.od_type == SET_ODTYPE_FENRIR)
    {
        /* Put the CD Block in a stable state before doing system reset. */
        fenrir_do_reset();
    }
#endif // !MAKE_WASCALOADER

    sc_sysres();
}
