/*

    Sega Saturn USB flash cart ROM
    Copyright © 2012, 2015 Anders Montonen
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
    Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdint.h>

#include "cpu.h"
#include "scu.h"
#include "vdp2.h"

#include "crc.h"

#define USB_FLAGS (*(volatile uint8_t*)(0x22200001))
#define USB_RXF     (1 << 0)
#define USB_TXE     (1 << 1)
#define USB_PWREN   (1 << 7)
#define USB_FIFO (*(volatile uint8_t*)(0x22100001))

#define WAIT_FOR_READ_FIFO()    do{while((USB_FLAGS&USB_RXF));}while(0)
#define WAIT_FOR_WRITE_FIFO()   do{while((USB_FLAGS&USB_TXE));}while(0)

#define USB_OUT_EP_SIZE 64

#define RGB(r, g, b) ((((b)&0x1f)<<10)|(((g)&0x1f)<<5)|((r)&0x1f))

enum
{
    USBDC_FUNC_DOWNLOAD      = 1,
    USBDC_FUNC_UPLOAD        = 2,
    USBDC_FUNC_EXEC          = 3,
    USBDC_FUNC_GET_BUFF_ADDR = 4,
    USBDC_FUNC_COPYEXEC      = 5,
};

const uint16_t ColorTable[] =
{
    RGB(   0,    0,    0), /* black  */
    RGB(0x1f,    0,    0), /* red    */
    RGB(0x19, 0x19,    0), /* yellow */
    RGB(0x10,    0, 0x1f), /* purple */
    RGB(   0, 0x0f,    0), /* green  */
    RGB(0x1f, 0x13,    0), /* orange */
    RGB(   0,    0, 0x0f), /* blue   */
};

typedef enum
{
    BLACK,
    RED,
    YELLOW,
    PURPLE,
    GREEN,
    ORANGE,
    BLUE,
} Color_e;

void mem_clear(void* ptr, uint32_t len)
{
    uint8_t* buff = (uint8_t*)ptr;

    while(len--)
    {
        *buff++ = 0;
    }
}
void mem_copy(void* dst, void* src, uint32_t len)
{
    uint8_t* d = (uint8_t*)dst;
    uint8_t* s = (uint8_t*)src;

    while(len--)
    {
        *d++ = *s++;
    }
}

__attribute__((section(".uncached"), noinline))
static void PurgeCache(void)
{
    uint8_t reg = CCR;
    reg &= ~CCR_CE;
    CCR = reg;
    reg |= CCR_CP;
    CCR = reg;
    reg |= CCR_CE;
    CCR = reg;
}

static void InitDma(void)
{
    (void)CHCR0;
    CHCR0 = 0;
    SAR0 = (uint32_t)&USB_FIFO;
    (void)DMAOR;
    DMAOR = DMAOR_DME;
}

static void ResetDma(void)
{
    (void)CHCR0;
    CHCR0 = 0;
    (void)DMAOR;
    DMAOR = 0;
}

static void ReceiveDma(uint8_t *pBuffer, uint32_t len)
{
    (void)CHCR0;
    CHCR0 = 0;
    DAR0 = (uint32_t)pBuffer;
    TCR0 = len;
    WAIT_FOR_READ_FIFO();
    CHCR0 = CHCR_DM0|CHCR_AR|CHCR_DE;
    while ((CHCR0 & CHCR_TE) == 0) ;
}

static void SetScreenColor(Color_e color)
{
    *(uint16_t*)VDP2_VRAM = ColorTable[color];
}

static void SignalError(Color_e col)
{
    /* Display a red screen for two seconds. */
    int delay = 60*2; /* Doesn't take 50Hz display into account.. */

    SetScreenColor(col);
    while(delay > 0)
    {
        WaitForVBLANKIn();
        WaitForVBLANKOut();
        --delay;
    }
}

#define COLOR_TOGGLE_COUNT (320*1024)
static uint8_t RecvByteAnimate(void)
{
    uint32_t i;
    Color_e color_table[] = {YELLOW, GREEN, BLUE};
    uint32_t color_id = 0;

    while(1)
    {
        SetScreenColor(color_table[color_id]);

        for(i=COLOR_TOGGLE_COUNT; i; i--)
        {
            /* Check I/O request from PC. */
            if((USB_FLAGS & USB_RXF) == 0)
                return USB_FIFO;
        }

        /* Toggle screen color. */
        color_id = (color_id + 1) % (sizeof(color_table) / sizeof(Color_e));
    }

    return 0; // For grumpy compiler
}

static uint8_t RecvByte(void)
{
    WAIT_FOR_READ_FIFO();
    return USB_FIFO;
}

static uint32_t RecvDword(void)
{
    uint32_t tmp = RecvByte();
    tmp = (tmp << 8) | RecvByte();
    tmp = (tmp << 8) | RecvByte();
    tmp = (tmp << 8) | RecvByte();

    return tmp;
}

static void SendByte(uint8_t byte)
{
    WAIT_FOR_WRITE_FIFO();
    USB_FIFO = byte;
}

static void DoDownload(void)
{
    uint8_t    *pData;
    uint32_t    len, ii;
    crc_t       checksum = crc_init();

    SetScreenColor(ORANGE);

    pData = (uint8_t*)RecvDword();

    len = RecvDword();

    for (ii = 0; ii < len; ++ii)
    {
        WAIT_FOR_WRITE_FIFO();
        USB_FIFO = pData[ii];
    }

    checksum = crc_update(checksum, pData, len);
    checksum = crc_finalize(checksum);

    SendByte(checksum);
}

static void DoDmaUpload(uint8_t *pBuffer, uint32_t len)
{
    while (len > USB_OUT_EP_SIZE)
    {
        ReceiveDma(pBuffer, USB_OUT_EP_SIZE);
        pBuffer += USB_OUT_EP_SIZE;
        len -= USB_OUT_EP_SIZE;
    }

    if (len > 0)
    {
        ReceiveDma(pBuffer, len);
    }
}

static void DoUpload(void)
{
    uint8_t    *pData;
    uint32_t    len;
    crc_t       readchecksum;
    crc_t       checksum = crc_init();

    SetScreenColor(ORANGE);

    pData = (uint8_t*)RecvDword();

    len = RecvDword();

    DoDmaUpload(pData, len);

    readchecksum = RecvByte();

    PurgeCache();
    checksum = crc_update(checksum, pData, len);
    checksum = crc_finalize(checksum);

    if (checksum != readchecksum)
    {
        SendByte(0x1);
        SignalError(RED);
    }
    else
    {
        SendByte(0);
    }
}

static void DoExecute(void)
{
    /* Read address, execute call. */
    void (*pFun)(void);

    pFun = (void(*)(void))RecvDword();
    (*pFun)();
}

/**
 * Send address to memory not used by this program.
 * Purpose : being able to upload & execute program over this program.
 * Example: if running from HRAM, should send LRAM address.
 * Since this program is running from uncommon RAM location, send null
 * address in order to inform that any place is OK for upload&exec.
**/
static void DoSendBufferAddress(void)
{
    SendByte(0x00);
    SendByte(0x00);
    SendByte(0x00);
    SendByte(0x00);
}

/**
 * Copy, then execute data.
 * Purpose : being able to upload & execute program over this program.
 * Since this program is running from uncommon RAM area, this function
 * isn't likely to be used.
 * Hence, receive parameters and do minimal required operations.
**/
static void DoCopyExec(void)
{
    void (*pFun)(void);
    void* src;
    uint32_t len;

    /* Address where to copy data from. */
    src = (void(*)(void))RecvDword();

    /* Data length. */
    len = RecvDword();

    /* Execution address where to copy data to. */
    pFun = (void(*)(void))RecvDword();

    mem_copy((void*)pFun, src, len);
    (*pFun)();
}

/* Setup back screen and color entries, turns video on. */
static void InitVideo(void)
{
    TVMD = 0x8110;
    BGON  = 0;
    LCTAU = 0;
    LCTAL = 0;
    BKTAL = 0;
    BKTAU = 0;
    SetScreenColor(BLACK);
}

int main(void)
{
    uint8_t command;

    /* Clear both H-RAM and L-RAM. */
    mem_clear((void*)0x00200000, 1024*1024);
    mem_clear((void*)0x06004000, 1024*1024 - 0x4000);

    InitVideo();

    while(1)
    {
        command = RecvByteAnimate();
        switch (command)
        {
        case USBDC_FUNC_DOWNLOAD:
            DoDownload();
            break;
        case USBDC_FUNC_UPLOAD:
            InitDma();
            DoUpload();
            ResetDma();
            break;
        case USBDC_FUNC_EXEC:
            DoExecute();
            InitVideo();
            break;
        case USBDC_FUNC_GET_BUFF_ADDR:
            DoSendBufferAddress();
            break;
        case USBDC_FUNC_COPYEXEC:
            DoCopyExec();
            break;
        default:
            SignalError(PURPLE); // Unknown command
            break;
        }

    }

    return 0;
}
