/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
/*
 * Sega Saturn cartridge flash tool
 * by Anders Montonen, 2012
 *
 * Original software by ExCyber
 * Graphics routines by Charles MacDonald
 *
 * Creative Commons Attribution-ShareAlike 3.0 Unported (CC BY-SA 3.0)
 */
#include <stdio.h>

#include "vdp.h"

/*
 * Global variables stuff.
 * Warning : globals are not directly usable from outside.
 */
unsigned short _list_index = 0; /* Offset into the command list */
void vdp_init(void)
{
    int i;
    volatile unsigned short *vdp2_vram = (volatile unsigned short *)VDP2_VRAM;
    volatile unsigned short *vdp2_cram = (volatile unsigned short *)VDP2_CRAM;

    TVMD = 0x0000;
    RAMCTL = RAMCTL & (~0x3000);

    /* Map Offset Register: Bitmap screen will be located at VRAM offset 0. */
    MPOFN = 0;

    /* Character Control Register: 256 colors, enable NBG0 as a bitmap.
     * |8 for 1024x256 bitmap.
     */
    CHCTLA = 0x0012|8;

    /* Screen Scroll Value Registers: No scroll. */
    SCXIN0 = 0;
    SCXDN0 = 0;
    SCYIN0 = 0;

    /* Screen Display Enable Register: Invalidate the transparency
     * code for NBG0 and display NBG0.
     */
    BGON = 0x0001;

    /* Clear VRAM.
     * Note : as Pseudo Saturn Kai uses second half of VDP2 RAM to
     *        backup VDP1 RAM, it's forbidden to clear it.
     */
#ifdef MAKE_PSKAI
    for(i=0; i<0x20000; i++)
    {
        vdp2_vram[i] = 0x0000;
    }
#else // !MAKE_PSKAI
    for(i=0; i<0x40000; i++)
    {
        vdp2_vram[i] = 0x0000;
    }
#endif // !MAKE_PSKAI

    /* Clear CRAM */
    for(i=0; i<0x0800; i++)
    {
        vdp2_cram[i] = 0x0000;
    }



    /* Set up VDP1 to appear under VDP2 graphics (0=hide, 1-7=visible) */
    PRISA = 0x0101;
    PRISB = 0x0101;
    PRISC = 0x0101;
    PRISD = 0x0101;

    /* Set up VDP1 registers */
    TVMR = 0x0000;
    FBCR = 0x0003;       /* VPD1 refresh : "Manual mode (change)"          */
    PTMR = 0x0002;       /* Starts drawing automatically with frame change */
    //EWDR = COLOR(7,7,7); /* Erase color: gray */
    EWDR = COLOR(1,1,1); /* Erase color: gray */
    EWLR = 0x0000;       /* Erase range: (0,0) -> (352,240) */
    EWRR = ((320 >> 3) << 9) | 224;

    /* Initialize display list */
    _list_index = 0;

    /* Start Filling VDP1 command list with buffer common stuff. */
    vdp1_add_system_clip(320, 224);
    vdp1_add_user_clip(0, 0, 319, 223);
    vdp1_add_local_coords(0, 0);

    /* Set VDP1 command list stuff. */
    vdp1_start(2/*double buffering*/, 2000/*max commands count.*/);

    /* Screen on */
    TVMD = 0x8000;
    TVMR = 0x0000;
    FBCR = 0x0001;
}




/**********************************************************************
 * VDP1 Commands stuff                                               *
**********************************************************************/


/* Macros */
/* Get the address of a command in the VDP1 VRAM */
#define CMD_ADDRESS(index) ((unsigned short *)(VDP1_VRAM + ((index) << 5)))




/* Global variables "get" functions */
unsigned short vdp1_list_get_index(void)
{
    return _list_index;
}
void vdp1_list_set_index(unsigned short i)
{
    _list_index = i;
}



/*
 * Set buffer count (1 or 2).
 * commands_count is the number of commands available
 * (i.e. depends of your memory management).
 * Note: command wrote until calling vdp1_start are considered
 *       as buffer common, and will not be overwritten after calling
 *       vdp1_flip (usefull for setting local coords, etc once for all).
 */
unsigned short _buffer_count    = 1;      /* Buffer count (user set). */
unsigned short _buffer_id       = 0;      /* Current buffer ID (0/1). */
unsigned short _jmp_cmd_id      = 0;      /* Jump command ID (used in double buffering). */
unsigned short _buffer_start[2] = {0, 0}; /* Buffer start address. */
/*
 * Example of double buffering configuration.
 * C C C C J b b b b b b E _ _ _ _ B B B B B B B E
 * | C    : common commands.
 * | J    : jump command in order to flip buffers.
 * | b(B) : Buffer 1(2) commands, start from _buffer_start[0(1)].
 * | E    : end commands.
 * | _    : unsused commands.
 */
void vdp1_start(int buffer_count, int commands_count)
{
    unsigned short *p;
    
    _buffer_count = (buffer_count == 1 ? 1 : 2);

    if(_buffer_count == 1)
    { /* Single buffering case. */

        _buffer_start[0] = _list_index;
        _buffer_start[1] = _buffer_start[0];

    }
    else if(_buffer_count == 2)
    { /* Double buffering case. */

        /* Next command: jump used for flipping. */
        _jmp_cmd_id = _list_index;
        /* Since no buffer are defined yet, don't jump. */
        p = CMD_ADDRESS(_jmp_cmd_id);
        p[CMDCTRL] = 0x8000; /* END */

        /* Set buffers start address. */
        _buffer_start[0] = _list_index+1;
        _buffer_start[1] = _buffer_start[0] + (commands_count - _buffer_start[0])/2;

    }

    /* Set list index to next command. */
    _buffer_id = 0;
    _list_index = _buffer_start[_buffer_id];
}

/*
 * End the current command list, and prepares the second buffer for writing.
 * (in single buffering, return to the begining of the current list, so that 
 *  next commands will overwrite the commands written so far)
 * Return polygons count.
 */
unsigned long vdp1_flip(void)
{
    unsigned long ret;
    unsigned short *p;

    /* Set polygon count. */
    ret = _list_index - _buffer_start[_buffer_id];

    /* Terminate the current command list. */
    p = CMD_ADDRESS(_list_index);
    p[CMDCTRL] = 0x8000; /* END */

    /* Double buffering stuff. */
    if(_buffer_count == 2)
    {
        /* Jump. */
        p = CMD_ADDRESS(_jmp_cmd_id);
        /* VDP1 "skip assign" command, ST-013-R3-061694.pdf, page72. */
        p[CMDCTRL] = 0x5000; /* Skip assign. */
        p[CMDLINK] = (_buffer_start[_buffer_id]) << 2;

        /* Flip. */
        _buffer_id = 1-_buffer_id;
    }

    /* Reset list index. */
    _list_index = _buffer_start[_buffer_id];

    return ret;
}



/* Terminate current command list. */
void vdp1_add_end(void)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x8000; /* END */
}


/* Add system clip to command list */
void vdp1_add_system_clip(short x, short y)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x0009; /* Set system clip */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDXC]   = x;
    p[CMDYC]   = y;

    _list_index++;
}


/* Add user clip to command list */
void vdp1_add_user_clip(short x1, short y1, short x2, short y2)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x0008; /* Set user clip */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXC]   = x2;
    p[CMDYC]   = y2;

    _list_index++;
}

/* Add local coords to command list */
void vdp1_add_local_coords(short x, short y)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x000A; /* Set local coords */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDXA]   = x;
    p[CMDYA]   = y;

    _list_index++;
}



/* Add local coords to command list */
void vdp1_add_polyline(short x1, short y1, short x2, short y2, short x3, short y3, short x4, short y4, unsigned short color)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x0005; /* Polyline */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDPMOD] = 0x00C0;
    p[CMDCOLR] = color;
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXB]   = x2;
    p[CMDYB]   = y2;
    p[CMDXC]   = x3;
    p[CMDYC]   = y3;
    p[CMDXD]   = x4;
    p[CMDYD]   = y4;

    _list_index++;
}

/* Add local coords to command list */
void vdp1_add_polygon(short x1, short y1, short x2, short y2, short x3, short y3, short x4, short y4, unsigned short color, unsigned long grda, int mesh_use)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x0004; /* Polygon */
    p[CMDLINK] = 0x0000; /* Jump address */
    p[CMDPMOD] = 0x00C0 | (mesh_use ? 0x0100 : 0) | 4;
    p[CMDCOLR] = color;
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXB]   = x2;
    p[CMDYB]   = y2;
    p[CMDXC]   = x3;
    p[CMDYC]   = y3;
    p[CMDXD]   = x4;
    p[CMDYD]   = y4;
    p[CMDGRDA] = grda>>3;

    _list_index++;
}

/* Add local coords to command list */
void vdp1_add_distorted_sprite(short x1, short y1, short x2, short y2, short x3, short y3, short x4, short y4, unsigned short spr_size, unsigned long source, unsigned short grad, int mesh_use)
{
    unsigned short *p;

    p = CMD_ADDRESS(_list_index);

    p[CMDCTRL] = 0x0002; /* Distorted sprite, RGB mode */
    p[CMDLINK] = 0x0000; /* Jump address */
    if(grad != 0)
    {
        p[CMDPMOD] = 0x00A8 | (1 << 2)/*gouraud_use*/;
    }
    else
    {
        p[CMDPMOD] = 0x00A8;
    }
    if(mesh_use)
    {
        p[CMDPMOD] |= 0x0100;
    }
    p[CMDCOLR] = 0;
    p[CMDSRCA] = source;
    p[CMDSIZE] = spr_size;
    p[CMDXA]   = x1;
    p[CMDYA]   = y1;
    p[CMDXB]   = x2;
    p[CMDYB]   = y2;
    p[CMDXC]   = x3;
    p[CMDYC]   = y3;
    p[CMDXD]   = x4;
    p[CMDYD]   = y4;
    p[CMDGRDA] = grad;

    _list_index++;
}



