/**
 *  Virtual Memory Cart : UBC debug tool
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "smpc.h"

/* SMPC direct I/O ports A and B */
#define PDR1        (*(volatile unsigned char  *)0x20100075)
#define PDR2        (*(volatile unsigned char  *)0x20100077)
#define DDR1        (*(volatile unsigned char  *)0x20100079)
#define DDR2        (*(volatile unsigned char  *)0x2010007B)
#define IOSEL       (*(volatile unsigned char  *)0x2010007D)
#define EXLE        (*(volatile unsigned char  *)0x2010007F)


/* Theses two dummy functions below are here to make satcom_lib
 * happy around current time acquisition from SMPC.
 *
 * Proper initialization and shutdown of pad acquisition internals
 * is required when :
 *  1. Acquiring current time from SPMC.
 *  2. Making V-Blank based pad status acquisition.
 *     This is mandatory to prevent status of some "3rd party"
 *     wireless pad from being incorrectly acquired.
 *
 * (1) may not be needed, because this modules runs from ROM and
 * consequently shall (not verified in details) retrieve "current"
 * time from a safe place in BUP library internals.
 *
 * (2) is completely not my concern as this module is just a debug
 * one for Pseudo Saturn Kai development needs.
 * Use properly crafted hardware from trusted places and everything will work fine.
 */
#ifdef SMPC_USE_POLLING
#else // !SMPC_USE_POLLING

void pad_init(void)
{
    // Do Nothing.
}

void pad_shutdown(void)
{
    // Do Nothing.
}

#endif // !SMPC_USE_POLLING


void lite_pad_shutdown(pad_t* p)
{
    DDR1    = 0;
    EXLE    = 0;
    IOSEL   = 0;
}

void lite_pad_init(pad_t* p)
{
    DDR1    = 0;
    EXLE    = 0;
    IOSEL   = 0; /* Set both ports to SMPC control mode. */

    /* Reset pad internals. */
    memset((void*)p, 0, sizeof(pad_t));
    p->repeat_th = 30;
}

void lite_pad_update_autorepeat(pad_t *p, int id)
{
    /* Button auto push feature. */
    if(p->repeat_th != 0)
    {
        int i;
        /* ID -> shift table. */
        unsigned short button_shift_tbl[13] =
        {
            SMPC_PAD_LBUTTON, 
            SMPC_PAD_RBUTTON, 
            SMPC_PAD_X      , 
            SMPC_PAD_Y      , 
            SMPC_PAD_Z      , 
            SMPC_PAD_START  , 
            SMPC_PAD_A      , 
            SMPC_PAD_C      , 
            SMPC_PAD_B      , 
            SMPC_PAD_RIGHT  , 
            SMPC_PAD_LEFT   , 
            SMPC_PAD_DOWN   , 
            SMPC_PAD_UP     
        };

        /* Count the number of times each button is consecutively hold. */
        for(i=0; i<13; i++)
        {
            if((p->new[id] & (button_shift_tbl[i])) != 0)
            {
                if(p->repeat_cnt[id][i] != 255)
                {
                    p->repeat_cnt[id][i]++;
                }
            }
            else
            {
                p->repeat_cnt[id][i] = 0;
            }
        }

        /* If hold enough, modify pad delta value. */
        for(i=0; i<13; i++)
        {
            if(p->repeat_cnt[id][i] >= p->repeat_th)
            {
                /* Mark this key as "pushed". */
                p->delta[id] |= button_shift_tbl[i];
            }
        }
    }
}



void lite_pad_read(pad_t* p)
{
    /* Waits for SF to be Reset. */
    smpc_wait_till_ready();

    /* Sets the Command Parameter in IREG. */
    SMPC_REG_IREG(0) = 0x00; // no intback status
    SMPC_REG_IREG(1) = 0x0A; // 15-byte mode, peripheral data returned, time optimized
    SMPC_REG_IREG(2) = 0xF0; // ???

    smpc_issue_command(SMPC_CMD_INTBACK);

    /* Wait SF Clear. */
    smpc_wait_till_ready();



    /* Port Status:
    0x04 - Sega-tap is connected
    0x16 - Multi-tap is connected
    0x21-0x2F - Clock serial peripheral is connected
    0xF0 - Not Connected or Unknown Device
    0xF1 - Peripheral is directly connected

    Peripheral ID:
    0x02 - Digital Device Standard Format
    0x13 - Racing Device Standard Format
    0x15 - Analog Device Standard Format
    0x23 - Pointing Device Standard Format
    0x23 - Shooting Device Standard Format
    0x34 - Keyboard Device Standard Format
    0xE1 - Mega Drive 3-Button Pad
    0xE2 - Mega Drive 6-Button Pad
    0xE3 - Saturn Mouse
    0xFF - Not Connected

    Special Notes:

    If a peripheral is disconnected from a port, SMPC only returns 1 byte for
    that port(which is the port status 0xF0), at the next OREG is the port
    status of the next port.

    e.g. If Port 1 has nothing connected, and Port 2 has a controller
        connected:

    OREG0 = 0xF0
    OREG1 = 0xF1
    OREG2 = 0x02
    etc.
    */

    unsigned short per_data;
    unsigned char port_con;
    unsigned char per_id = 0;
    unsigned char oreg_counter=0;
    unsigned char per_counter=0;
    int i, i2;

    // See what is actually connected to Port 1/2
    for (i2 = 0; i2 < 2; i2++)
    {
        port_con = SMPC_REG_OREG(oreg_counter);
        oreg_counter++;

        if ((port_con & 0xF0) != 0x20)
        {
            for (i = 0; i < (port_con & 0x0F); i++)
            {
                if((i+per_counter) == 0)
                { /* restrict polling to first pad only. */
                    p->per[i+per_counter].id = per_id = SMPC_REG_OREG(oreg_counter);
                }
                oreg_counter++;

                //            if (perid == 0x02)
                //            {
                //               oldperpush[i+percounter] = p->per[i+percounter].butpush;
                //               perdata = (SMPC_REG_OREG(oregcounter) << 8) | SMPC_REG_OREG(oregcounter+1);
                //               oregcounter += 2;
                //            }
                //            else
                //               oregcounter += (perid & 0xF);
                switch (per_id >> 4)
                {
                case 0x0: // Standard Pad type
                case 0x1: // Analog type(Racing wheel/Analog pad)
                case 0x3: // Keyboard type
                {
                    if((i+per_counter) == 0)
                    { /* restrict polling to first pad only. */
                        p->old_per_push[i+per_counter] = p->per[i+per_counter].but_push;
                    }
                    per_data = (SMPC_REG_OREG(oreg_counter) << 8) | SMPC_REG_OREG(oreg_counter+1);
                    oreg_counter += 2;

                    switch (per_id & 0xF)
                    {
                    case 0x3: // Racing
                        oreg_counter += 1;
                        break;
                    case 0x4: // Keyboard
                        //((keyboarddata_struct *)per)[i+per_counter].kbd_type = per_data & 0x7;
                        //per_data |= 0x7;
                        //((keyboarddata_struct *)per)[i+per_counter].flags = SMPC_REG_OREG(oreg_counter);
                        //((keyboarddata_struct *)per)[i+per_counter].key = SMPC_REG_OREG(oreg_counter+1);
                        //if (((keyboarddata_struct *)per)[i+per_counter].flags != 0x6)
                        //{
                        //    if (((keyboarddata_struct *)per)[i+per_counter].key == KEY_LEFTSHIFT
                        //    ||  ((keyboarddata_struct *)per)[i+per_counter].key == KEY_RIGHTSHIFT)
                        //    {
                        //        ((keyboarddata_struct *)per)[i+per_counter].extra_state[0] = (((keyboarddata_struct *)per)[i+per_counter].flags >> 3) & 0x1;
                        //    }
                        //}
                        oreg_counter += 2;
                        break;
                    case 0x6: // Analog Pad
                        oreg_counter += 4;
                        break;
                    default:
                        break;
                    }

                    if((i+per_counter) == 0)
                    { /* restrict polling to first pad only. */
                        p->per[i+per_counter].but_push = per_data ^ 0xFFFF;
                        p->per[i+per_counter].but_push_once = (p->old_per_push[i+per_counter] ^ p->per[i+per_counter].but_push) & p->per[i+per_counter].but_push;
                    }

                    break;
                }
                case 0x2: // Pointer type(Mouse/Gun)
                case 0xE: // Other type(Mega Drive 3/6 button pads/Shuttle Mouse)
                {
                    switch (per_id & 0xF)
                    {
                    case 0x2: // Mega Drive 6-button Pad
                    case 0x1: // Mega Drive 3-button Pad
                        break;
                    case 0x3: // Saturn/Shuttle Mouse
                    {
                    //    per_data = SMPC_REG_OREG(oreg_counter);
                    //    p->per[i+per_counter].but_push = (per_data & 0xF) << 8; // fix me
                    //    p->per[i+per_counter].but_push_once = (p->old_per_push[i+per_counter] ^ p->per[i+per_counter].but_push) & p->per[i+per_counter].but_push;
                    //    ((mousedata_struct *)per)[i+per_counter].flags = per_data >> 4;

                    //    ((mousedata_struct *)per)[i+per_counter].x = SMPC_REG_OREG(oreg_counter+1);
                    //    ((mousedata_struct *)per)[i+per_counter].y = SMPC_REG_OREG(oreg_counter+2);

                    //    // X Overflow
                    //    if (((mousedata_struct *)per)[i+per_counter].flags & 0x4)
                    //    ((mousedata_struct *)per)[i+per_counter].x++;

                    //    // X Sign
                    //    if (((mousedata_struct *)per)[i+per_counter].flags & 0x1)
                    //    ((mousedata_struct *)per)[i+per_counter].x = 0 - ((mousedata_struct *)per)[i+per_counter].x - 1;

                    //    // Y Overflow
                    //    if (((mousedata_struct *)per)[i+per_counter].flags & 0x8)
                    //    ((mousedata_struct *)per)[i+per_counter].y++;

                    //    // Y Sign
                    //    if (((mousedata_struct *)per)[i+per_counter].flags & 0x2)
                    //    ((mousedata_struct *)per)[i+per_counter].y = 0 - ((mousedata_struct *)per)[i+per_counter].y - 1;

                    //    break;
                    }
                    default:
                        break;
                    }

                    oreg_counter += (per_id & 0xF);
                    break;
                }
                default:
                    break;
                }

                if((i+per_counter) == 0)
                { /* restrict polling to first pad only. */
                    /* Update pad_t structure with peripheral data. */
                    p->old[per_counter] = p->new[per_counter];
                    p->new[per_counter] = p->per[i+per_counter].but_push;
                    p->delta[per_counter] = (p->new[per_counter] ^ p->old[per_counter]) & p->new[per_counter];
                    lite_pad_update_autorepeat(p, per_counter);
                }

                /* Update peripheral count. */
                per_counter++;
                //if(per_counter > p->per_count)
                //{
                //    p->per_count = per_counter;
                //}
            }
        }
    }

    // Issue break
    SMPC_REG_IREG(0) = 0x40;
}

