/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sc_saturn.h>

#include "vmem_main.h"

/* IP header related stuff. */
#include "../ip_header.h"


/* Define block size same as typical block size for good old memory cart.
 * Any other block size should be OK, but using a
 * memory cart-like value may help compatibility with some games.
 */
#define VMEM_FILE_BLOCK_SIZE 256

/**
 * get_id_from_filename
 * Function common to both dev_cart and dev_floppy.
 *
 * Return file ID from its filename.
 * Negative value is returned when file name don't matches expected pattern
 *
 * Example : file_get_id("DATA123.BUP") returns 123
**/
short get_id_from_filename(char* name)
{
    short ret = -1;
    char* bup_ext = BUP_DATA_EXT;

    if(sc_toupper(name[0]) != 'D') return ret;
    if(sc_toupper(name[1]) != 'A') return ret;
    if(sc_toupper(name[2]) != 'T') return ret;
    if(sc_toupper(name[3]) != 'A') return ret;
    if((name[4] < '0') || (name[4] > '9')) return ret;
    if((name[5] < '0') || (name[5] > '9')) return ret;
    if((name[6] < '0') || (name[6] > '9')) return ret;

    /* verify file extension (.BUP) and terminating character too. */
    if(name[7] != bup_ext[0]) return ret;
    if(sc_toupper(name[ 8]) != bup_ext[1]) return ret;
    if(sc_toupper(name[ 9]) != bup_ext[2]) return ret;
    if(sc_toupper(name[10]) != bup_ext[3]) return ret;
    if(name[11] != '\0') return ret;

    /* Retrieve save number from file name. */
    ret  =  name[6]-'0';
    ret += (name[5]-'0')*10;
    ret += (name[4]-'0')*100;

    /* Discard IDs too large. */
    if(ret >= VMEM_CART_INDEX_CNT)
    {
        ret = -1;
    }

    return ret;
}

/**
 * cart_set_filename_from_id
 * Format backup/header filename with full path.
 * Example :
 *  char name[BUP_FILE_NAMELEN];
 *  cart_set_filename_from_id(name, 123, FILENAME_BUP);
 * will output "/MEMCART/DATA123.BUP" to name
**/
#define FILENAME_BUP 0
#define FILENAME_TXT 1
void cart_set_filename_from_id(char* name, short id, short ext)
{
    if(ext == FILENAME_TXT)
    {
        sprintf(name, "/" CART_BUP_FOLDER "/" BUP_FILEPREFIX "%03d" ".TXT", id);
    }
    else
    {
        sprintf(name, "/" CART_BUP_FOLDER "/" BUP_FILEPREFIX "%03d" BUP_DATA_EXT, id);
    }
}


/**
 * File index related sub-routines.
**/
vmem_idxglobal_t* file_idxglobal_getptr(void)
{
    vmem_libram_t * l = VMEM_LIBRAM;
    return &(l->idx_global);
}
vmem_fileindex_t* fileindex_getptr(void)
{
    vmem_libram_t * l = VMEM_LIBRAM;
    return &(l->idx_data);
}

int cart_is_entry_used(vmem_fileindex_t* idx, short id)
{
    int ret;
    ret = (idx->header.file_used[id/32]) & (1 << (id%32));
    ret = ret == 0 ? 0 : 1;
    return ret;
}
void cart_used_setstatus(vmem_fileindex_t* idx, short id, int used)
{
    int i;

    if(used)
    {
        /* Mark file as used. */
        idx->header.file_used[id/32] |= 1 << (id%32);
    }
    else
    {
        /* Mark file as free. */
        idx->header.file_used[id/32] &= ~(1 << (id%32));
    }

    scd_logout("[RC_CART] idx_hdr->savecntr=%d", idx->header.savecntr);
    for(i=0; i<((VMEM_CART_INDEX_CNT / 32) + 1); i++)
    {
        scd_logout("[RC_CART] file_used[%2d]=0x%08X", i, idx->header.file_used[i]);
    }
}

/**
 * cart_get_free_id
 * Look for save with oldest date stamp.
 * Return negative value when all files are used.
**/
short cart_get_free_id(vmem_fileindex_t* idx)
{
    unsigned long save_id;
    short ret = -1;
    short id;

    for(id=0; id<VMEM_CART_INDEX_CNT; id++)
    {
        if(!cart_is_entry_used(idx, id))
        {
            if(ret == -1)
            {
                ret = id;
                save_id = idx->entries[id].save_id;
            }
            else
            {
                if(idx->entries[id].save_id < save_id)
                {
                    ret = id;
                    save_id = idx->entries[id].save_id;
                }
            }
        }
    }

    return ret;
}

/**
 * cart_idx_lookup
 * Look for corresponding backup filename from specified index value.
 * Return negative value in the case no entry matched.
**/
short cart_idx_lookup(vmem_fileindex_t* idx, char* filename, short id_stt)
{
    short id;

    for(id=id_stt; id<VMEM_CART_INDEX_CNT; id++)
    {
        if(cart_is_entry_used(idx, id))
        {
            char* filename_idx = (char*)(idx->entries[id].bup_filename);
            scd_logout("[RC_CART]ID[%3d] bup_filename_comp(filename=\"%s\", filename_idx=\"%s\") ...", id, filename, filename_idx);
            if(bup_filename_comp(filename, filename_idx) == 0)
            {
                scd_logout("[RC_CART]bup_filename_comp==0 return %d", id);
                return id;
            }
        }
    }

    return -1;
}


/**
 * Index data file name.
 * Contains vmem_fileindex_t structure.
**/
const char* _index_data_file = "/" CART_BUP_FOLDER "/CONTENTS.IDX";

/*
 * Index readme file name.
 * Written after (re)building index data.
 */
const char* _vmem_cart_readme_filename = "/" CART_BUP_FOLDER "/README.TXT";
const char* _vmem_cart_readme_text = "\
This folder contains Saturn save data files generated by\r\n\
Vmem module included in Pseudo Saturn Kai.\r\n\
Informations, support and updates about Vmem and Pseudo Saturn Kai\r\n\
can be found at the following address :\r\n\
http://ppcenter.free.fr/pskai/\r\n";


/**
 * cart_idx_save
 * Save index data to SD card.
**/
void cart_idx_save(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    scf_write((void*)(&(VMEM_LIBRAM->idx_data)), sizeof(vmem_fileindex_t), VMEM_MEDIA_ID, (char*)_index_data_file);
}

/**
 * cart_idx_populate
 * Populate indx related data.
 * In the case index file is found, read index data from file, 
 * Else, rebuild index data from all BUP files found in MEMCART folder.
 * After rebuilding index data, a small readme.txt file is also outpur to MEMCART folder.
**/
void cart_idx_populate(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    scdf_file_entry_t* buff = f->file_entries;
    unsigned long ret;
    unsigned long files_count, files_offs;

    vmem_idxglobal_t* global = file_idxglobal_getptr();
    vmem_fileindex_t* idx = fileindex_getptr();
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);

    int i;
    int done = 0;
    char* fn = global->name_buff;

    short id1, id2;
    unsigned long len;

    scd_logout("[RC_CART]cart_idx_populate ...");

    /* Try to read index file from SD card. */
    ret = scf_size(VMEM_MEDIA_ID, (char*)_index_data_file);
    scd_logout("[RC_CART]IDX file size = %d bytes (expected = %d)", ret, sizeof(vmem_fileindex_t));
    if(ret == sizeof(vmem_fileindex_t))
    {
        memset((void*)idx, 0, sizeof(vmem_fileindex_t));
        ret = scf_read(VMEM_MEDIA_ID, (char*)_index_data_file, 0/*offset*/, sizeof(vmem_fileindex_t), (void*)idx);

        if((ret == sizeof(vmem_fileindex_t))
        && (memcmp((void*)(idx->header.magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) == 0))
        {
            /* Index file read success !
             * Don't need to list files in MEMCART directory.
             */
            done = 1;
        }
        else
        {
            scd_logout("[RC_CART]IDX magic string (\"%c%c%c%c\") mismatch ! (read size=%d)", 
                char2pchar(idx->header.magic[0]), 
                char2pchar(idx->header.magic[1]), 
                char2pchar(idx->header.magic[2]), 
                char2pchar(idx->header.magic[3]), 
                ret);
        }
    }


    if(!done)
    {
        /* Reset Index Settings/Data. */
        memset((void*)idx, 0, sizeof(vmem_fileindex_t));
        memcpy((void*)(idx->header.magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);

        /* Read directory list by chunk of FILE_ENTRIES_CNT entries. */
        files_offs = 0;
        do
        {
            files_count = scf_filelist(buff, FILE_ENTRIES_CNT, files_offs/*list_offset*/, 0/*list_sort*/, BUP_DATA_EXT/*filter*/, VMEM_MEDIA_ID, "/" CART_BUP_FOLDER "/");
            scd_logout("[RC_CART]files_count(\"%s\", \"%s\") = %d ...", "/" CART_BUP_FOLDER "/", BUP_DATA_EXT, files_count);

            /* Analyze files in list. */
            for(i=0; i<FILE_ENTRIES_CNT; i++)
            {
                /* Analyze enabled file entries only. */
                if((buff[i].flags & FILEENTRY_FLAG_ENABLED)
                && (!(buff[i].flags & FILEENTRY_FLAG_FOLDER))
                && (i < files_count))
                {
                    char* filename = buff[i].filename;
                    short id = get_id_from_filename(filename);

                    scd_logout("[RC_CART]entry[%d/%d].filename = \"%s\", id=%d ...", i, files_count, buff[i].filename, id);

                    /* Is it a valid .BUP file ? */
                    if((id >= 0) && (id < VMEM_CART_INDEX_CNT))
                    {
                        cart_set_filename_from_id(fn, id, FILENAME_BUP);

                        len = scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);

                        if((len == sizeof(vmem_bup_header_t))
                        && (memcmp((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) == 0))
                        {
                            /* Mark file as used in index data. */
                            cart_used_setstatus(idx, id, 1/*stat*/);

                            /* Copy backup file name and index to index. */
                            int j;
                            idx->entries[id].save_id = bup_header->save_id;
                            idx->entries[id].date    = bup_header->date;
                            for(j=0; j<12; j++)
                            {
                                idx->entries[id].bup_filename[j] = bup_header->dir.filename[j];
                            }
                            scd_logout("[RC_CART] tb.filename = \"%s\" -> entries[id=%d]", 
                                bup_print_filename(bup_header->dir.filename), 
                                id);
                        }
                    }
                }
            }

            /* Update count of files listed so far. */
            if(files_count <= FILE_ENTRIES_CNT)
            {
                files_offs += files_count;
            }
            else
            {
                files_offs += FILE_ENTRIES_CNT;
            }
        } while(files_offs < files_count);

        /* Save index data to SD card. */
        cart_idx_save();

        /* Save readme file to SD card. */
        scf_write((void*)_vmem_cart_readme_text, strlen(_vmem_cart_readme_text), VMEM_MEDIA_ID, (char*)_vmem_cart_readme_filename);

        scd_logout("[RC_CART]cart_idx_populate END (files_count=%d)", files_count);
    }

    done = 0;
    /* Index integrity check#1 : verify maximum save ID. */
    for(id1=0; id1<VMEM_CART_INDEX_CNT; id1++)
    {
        if(cart_is_entry_used(idx, id1))
        {
            /* If needed, update save max ID. */
            if(idx->entries[id1].save_id > idx->header.savecntr)
            {
                idx->header.savecntr = idx->entries[id1].save_id;
                done = 1;
            }
        }
    }

    /* Index integrity check#2 : sort each save ID according to date. */
    for(id1=0; id1<(VMEM_CART_INDEX_CNT-1); id1++)
    {
        if(cart_is_entry_used(idx, id1))
        {
            for(id2=id1+1; id2<VMEM_CART_INDEX_CNT; id2++)
            {
                if(cart_is_entry_used(idx, id2))
                {
                    if((idx->entries[id1].save_id >= idx->entries[id2].save_id)
                    && (idx->entries[id1].date    <  idx->entries[id2].date   ))
                    {
                        unsigned long tmp;
                        tmp = idx->entries[id1].save_id;
                        idx->entries[id1].save_id = idx->entries[id2].save_id;
                        idx->entries[id2].save_id = tmp;
                    }
                    else
                    {
                        if((idx->entries[id1].save_id <= idx->entries[id2].save_id)
                        && (idx->entries[id1].date    >  idx->entries[id2].date   ))
                        {
                            unsigned long tmp;
                            tmp = idx->entries[id1].save_id;
                            idx->entries[id1].save_id = idx->entries[id2].save_id;
                            idx->entries[id2].save_id = tmp;
                        }
                    }
                }
            }
        }
    }

    /* Index integrity check#3 : discard duplicate saves. */
    for(id1=0; id1<(VMEM_CART_INDEX_CNT-1); id1++)
    {
        if(cart_is_entry_used(idx, id1))
        {
            for(id2=id1+1; id2<VMEM_CART_INDEX_CNT; id2++)
            {
                if((cart_is_entry_used(idx, id2))
                && (memcmp((void*)(idx->entries[id1].bup_filename), (void*)(idx->entries[id2].bup_filename), 12) == 0))
                {
                    scd_logout("[RC_CART][Duplicate] filename=\"%s\", IDs=%d/%d", 
                        bup_print_filename(idx->entries[id1].bup_filename), 
                        idx->entries[id1].save_id, idx->entries[id2].save_id);

                    if(idx->entries[id1].save_id > idx->entries[id2].save_id)
                    {
                        cart_used_setstatus(idx, id2, 0/*stat*/);
                    }
                    else
                    {
                        cart_used_setstatus(idx, id1, 0/*stat*/);
                    }
                    done = 1;
                }
            }
        }
    }

    /* Reset statistics informations on call to CartBUPInit function */
    if(VMEM_LIBRAM->idx_init == BUP_IDXINIT_START)
    {
        memset((void*)(&idx->header.stats), 0, sizeof(vmem_bup_stats_t));
    }


    if(done)
    {
        /* Save updated index data to SD card. */
        cart_idx_save();
    }
}










/**
 * Init file-based virtual memory devices.
 *
 * This function is common to both virtual memory cartridge
 * and floppy drives and is called just after BUP_Init function.
 *
 * Return value :
 *  0 : Couldn't init.
 *  1 : Init success.
**/
int FileBUPInit(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    unsigned long ret0, ret1, ret2;

    /* Create folder where to store cartridge saves files. */
    ret0 = scf_mkdir(VMEM_MEDIA_ID, "/" CART_BUP_FOLDER);
    scd_logout("mkdir(\"/" CART_BUP_FOLDER "\")=%d", ret0);

    /* Create one folder per partition, where to store FDD saves files. */
    ret1 = scf_mkdir(VMEM_MEDIA_ID, "/" FDD_BUP_FOLDER "1");
    scd_logout("mkdir(\"/" FDD_BUP_FOLDER "1\")=%d", ret1);
    ret2 = scf_mkdir(VMEM_MEDIA_ID, "/" FDD_BUP_FOLDER "2");
    scd_logout("mkdir(\"/" FDD_BUP_FOLDER "2\")=%d", ret2);

    if((!ret0) || (!ret1) || (!ret2))
    {
        /* Couldn't create folder(s), so disable virtual memory cart and virtual FDD. */
        return 0;
    }

    scd_logout("[RC_CART]Init success !");
    scd_logout("[RC_CART]sizeof(vmem_bup_header_t)=%d", sizeof(vmem_bup_header_t));
    scd_logout("[RC_CART]sizeof(vmem_idxheader_t) =%d", sizeof(vmem_idxheader_t));
    scd_logout("[RC_CART]sizeof(vmem_fileindex_t) =%d", sizeof(vmem_fileindex_t));
    scd_logout("[RC_CART]sizeof(vmem_workram_t)   =%d", sizeof(vmem_workram_t));
    scd_logout("[RC_CART]sizeof(vmem_libram_t)    =%d", sizeof(vmem_libram_t));

    /* Check structure sizes. */
    if((sizeof(vmem_workram_t) != ( 8*1024))
    || (sizeof(vmem_libram_t ) != (16*1024)))
    {
        while(1){;}
    }

    return 1;
}




/**
 * Init file-based memory cartridge device.
**/
void CartBUPInit(void)
{
    vmem_libram_t * l = VMEM_LIBRAM;

    /* Request index memory init on next call. */
    l->idx_init = BUP_IDXINIT_START;
}


//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPSelectPartition(Uint16 num)
{
    vmem_libram_t * l = VMEM_LIBRAM;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    /* Only one partition available. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPFormat(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_libram_t * l = VMEM_LIBRAM;
    scdf_file_entry_t* buff = f->file_entries;
    char* fn = file_idxglobal_getptr()->name_buff;
    unsigned long files_count, files_offs;
    int i;

    scd_logout("[RC_CART]format STT");

    /* Delete index/readme files. */
    scf_delete(VMEM_MEDIA_ID, (char*)_index_data_file);
    scf_delete(VMEM_MEDIA_ID, (char*)_vmem_cart_readme_filename);
    l->idx_init = BUP_IDXINIT_RESET;

    /* Delete all backup files, including the non indexed ones. */
    files_offs = 0;
    do
    {
        files_count = scf_filelist(buff, FILE_ENTRIES_CNT, files_offs/*list_offset*/, 0/*list_sort*/, BUP_DATA_EXT/*filter*/, VMEM_MEDIA_ID, "/" CART_BUP_FOLDER "/");

        for(i=0; i<FILE_ENTRIES_CNT; i++)
        {
            /* Delete enabled file entries only. */
            if((buff[i].flags & FILEENTRY_FLAG_ENABLED)
            && (!(buff[i].flags & FILEENTRY_FLAG_FOLDER))
            && (i < files_count))
            {
                short id = get_id_from_filename(buff[i].filename);

                if(id >= 0)
                {
                    cart_set_filename_from_id(fn, id, FILENAME_BUP);
                    scf_delete(VMEM_MEDIA_ID, fn);
                    cart_set_filename_from_id(fn, id, FILENAME_TXT);
                    scf_delete(VMEM_MEDIA_ID, fn);
                }
            }
        }

        /* Update count of files listed so far. */
        if(files_count <= FILE_ENTRIES_CNT)
        {
            files_offs += files_count;
        }
        else
        {
            files_offs += FILE_ENTRIES_CNT;
        }
    } while(files_offs < files_count);

    scd_logout("[RC_CART]format END");

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPStatus(Uint32 datasize, BupStat *tb)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_libram_t * l = VMEM_LIBRAM;
    vmem_fileindex_t* idx = fileindex_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    int freeblock;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    /* Size of block. */
    tb->blocksize  = VMEM_FILE_BLOCK_SIZE;
    /* Size of Backup Ram (in blocks) */
    tb->totalblock = (1024 * 1024) / VMEM_FILE_BLOCK_SIZE;
    /* Size of Backup Ram (in bytes) */
    tb->totalsize  = tb->totalblock * tb->blocksize;

    /* Free space (in bytes) */
    tb->freesize   = tb->totalsize;
    freeblock = 0;
    for(id=0; id<VMEM_CART_INDEX_CNT; id++)
    {
        if(cart_is_entry_used(idx, id))
        {
            cart_set_filename_from_id(fn, id, FILENAME_BUP);
            scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
            if(tb->freesize > bup_header->dir.datasize)
            {
                tb->freesize -= bup_header->dir.datasize;
            }
        }
        else
        {
            freeblock = 1;
        }
    }

    if(freeblock)
    {
        /* If at least one free block available, 
         * return device information with enough free
         * room for probably all Saturn games existing.
         * (According to DTS, recommended max save data is 256KB)
         */
        if(tb->freesize < (256*1024))
        {
            tb->freesize = 256*1024;
        }
    }
    else
    {
        /* If no block available, return zero free
         * space in device information.
         */
        tb->freesize = 0;
    }

    /* Free space (in blocks) */
    tb->freeblock  = tb->freesize / tb->blocksize;
    /* Not sure, but seems to be the same as Free Space (in blocks) */
    tb->datanum    = tb->freeblock;

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_libram_t * l = VMEM_LIBRAM;
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    vmem_fileindex_t* idx = fileindex_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id_prev, id;
    unsigned long len;
    BupDate sat_date;
    struct tm* t;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    /* Bump counter in statistics informations. */
    idx->header.stats.write_cnt++;

    /* Minimal sanity fix of information in BupDir structure. */
    tb->filename[11] = '\0';
    tb->comment [10] = '\0';
    /* Update size in block in BupDir structure.
     * (Some games may use it in the case write failed ?)
     */
    tb->blocksize = bup_convert_blocksize(VMEM_FILE_BLOCK_SIZE, tb->datasize);

    /* Look for save with same name in saves index. */
    id_prev = cart_idx_lookup(idx, (char*)(tb->filename), 0/*id_stt*/);

    /* If save data already exists, verify that we can erase it. */
    if((id_prev >= 0) && (wmode != 0))
    {
        /* Can't overwrite. */
        return BUP_FOUND;
    }

    /* Look for a file where to save backup data. */
    id = cart_get_free_id(idx);

    /* Special case where previous version exists, 
     * and no empty file where to save backup data :
     * overwrite previous version backup data.
     */
    if((id_prev >= 0) && (id < 0))
    {
        id = id_prev;
        id_prev = -1;
    }

    /* Not enough space to save backup data ? */
    if(id < 0)
    {
        /* Too bad, can't save. */
        return BUP_NOT_ENOUGH_MEMORY;
    }

    /* Write backup header to SD card. */
    cart_set_filename_from_id(fn, id, FILENAME_BUP);
    memset((void*)bup_header, 0, sizeof(vmem_bup_header_t));
    memcpy((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);
    idx->header.savecntr++;
    bup_header->save_id = idx->header.savecntr;
    /* Save current date to header. */
    t = sc_tim_smpcpoll();
    sat_date.year  = t->tm_year - 80;
    sat_date.month = t->tm_mon + 1;
    sat_date.day   = t->tm_mday;
    sat_date.time  = t->tm_hour;
    sat_date.min   = (t->tm_hour >= 60 ? 59 : t->tm_min);
    sat_date.week  = t->tm_wday;
    bup_header->date = bup_setdate_internal(&sat_date);
    memcpy((void*)(&(bup_header->stats)), (void*)(&(idx->header.stats)), sizeof(vmem_bup_stats_t));
    memcpy((void*)(&(bup_header->dir)), (void*)tb, sizeof(BupDir));

    len = scf_write((void*)bup_header, sizeof(vmem_bup_header_t), VMEM_MEDIA_ID, fn);
    if(len != sizeof(vmem_bup_header_t))
    {
        /* Write error : delete data file. */
        cart_set_filename_from_id(fn, id, FILENAME_BUP);
        scf_delete(VMEM_MEDIA_ID, fn);
        cart_set_filename_from_id(fn, id, FILENAME_TXT);
        scf_delete(VMEM_MEDIA_ID, fn);

        return BUP_NON;
    }

    /* Append backup data to SD card. */
    len = scf_append((void*)data, tb->datasize, VMEM_MEDIA_ID, fn);
    if(len != tb->datasize)
    {
        /* Write error : delete data file. */
        cart_set_filename_from_id(fn, id, FILENAME_BUP);
        scf_delete(VMEM_MEDIA_ID, fn);
        cart_set_filename_from_id(fn, id, FILENAME_TXT);
        scf_delete(VMEM_MEDIA_ID, fn);

        return BUP_NON;
    }

    /* Copy backup file name and ID and date to index. */
    int j;
    idx->entries[id].save_id = bup_header->save_id;
    idx->entries[id].date    = bup_header->date;
    for(j=0; j<12; j++)
    {
        idx->entries[id].bup_filename[j] = tb->filename[j];
    }

    /* Update index status. */
    cart_used_setstatus(idx, id, 1/*stat*/);

    /* If previous version of save data exists, 
     * mark it as "free".
     */
    if(id_prev >= 0)
    {
        cart_used_setstatus(idx, id_prev, 0/*stat*/);
    }

    /* Save index data to SD card. */
    cart_idx_save();

    /* Extra : save text some information in <save name>.txt file. */
#define TXTINFO_BUFF_LEN (FILE_ENTRIES_CNT * sizeof(scdf_file_entry_t))
    char* ptr = (char*)(f->file_entries);
    unsigned long offs;

    /* First lines : save filename and date.
     * (Comment may not be in ASCII format, so it is safer to reformat it.)
     */
    sprintf(ptr, "Save : %s:%s\r\nData size : %6d bytes, date=0x%08X\r\n", 
        tb->filename, 
        bup_print_comment(tb->comment), 
        (unsigned int)(tb->datasize), 
        (unsigned int)(tb->date));

    /* Then, save vmem specific informations. */
    offs = strlen(ptr);
    sprintf(ptr+offs, "Vmem ID/Date: %08X/%08X\r\n", (unsigned int)(bup_header->save_id), (unsigned int)(bup_header->date));
    offs = strlen(ptr);
    sprintf(ptr+offs, "\r\nStats :\r\n Bup_Dir   : %3d, BUP_Read   : %3d\r\n Bup_Write : %3d, BUP_Verify : %3d\r\n\r\n", 
        bup_header->stats.dir_cnt  , bup_header->stats.read_cnt  , 
        bup_header->stats.write_cnt, bup_header->stats.verify_cnt);


    /* Then, save IP header informations.
     *  char hwr_id[16]; // "SEGA SEGASATURN "
     *  char man_id[16]; // "SEGA TP KAISHA-A"
     *  char prd_nb[10]; // "T-053G    "
     *  char version[6]; // "V0.001"
     *  char date[8];    // "19940912"
     *  char dev_id[8];  // "CD-1/1  "
     *  char area[16];   // "JTUBKAEL        "
     *  char pers[16];   // "JAST            "
     *  char title[112]; // ...
     *
     * With FILE_ENTRIES_CNT set to 24, printf buffer is 
     * FILE_ENTRIES_CNT*sizeof(scdf_file_entry_t) = 768 bytes large, 
     * and first lines (see above) are less than 100 bytes
     * and IP header information is 208 bytes large
     * (carriage return characters are not taken in account, but it's OK anyway), 
     * so it is OK to write everything at once.
     *
     * Memo : typical text file size is around 570 bytes.
     */
    offs = strlen(ptr);
    offs += ip_header_logout((ip_bin_t*)0x06002000, ptr+offs);

    /* Write all text informations at once. */
    cart_set_filename_from_id(fn, id, FILENAME_TXT);
    scf_write((void*)ptr, offs, VMEM_MEDIA_ID, fn);

    /* Returns 0 if there's no error. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPRead(Uint8 *filename, volatile Uint8 *data)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_libram_t * l = VMEM_LIBRAM;
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    vmem_fileindex_t* idx = fileindex_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    unsigned long len;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    /* Bump counter in statistics informations. */
    idx->header.stats.read_cnt++;

    id = cart_idx_lookup(idx, (char*)(filename), 0/*id_stt*/);

    /* Save not found ? */
    if(id < 0)
    {
        return BUP_NOT_FOUND;
    }

    /* First, read backup header from SD card. */
    cart_set_filename_from_id(fn, id, FILENAME_BUP);
    len = scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
    if(len != sizeof(vmem_bup_header_t))
    {
        /* Read error ? */
        return BUP_BROKEN;
    }
    if(memcmp((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) != 0)
    {
        /* Version check. */
        return BUP_BROKEN;
    }

    /* As we now know backup data size, read data from SD card. */
    scd_logout("[RC_CART][BUP_Read]scf_read(\"%s\", size=%d) ...", fn, bup_header->dir.datasize);
    len = scf_read(VMEM_MEDIA_ID, fn, sizeof(vmem_bup_header_t)/*offset*/, bup_header->dir.datasize, (void*)data);
    if(len != bup_header->dir.datasize)
    {
        /* Read error ? */
        return BUP_BROKEN;
    }

    /* Returns 0 if there's no error. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPDelete(Uint8 *filename)
{
    vmem_libram_t * l = VMEM_LIBRAM;
    vmem_fileindex_t* idx = fileindex_getptr();
    short id;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    id = cart_idx_lookup(idx, (char*)(filename), 0/*id_stt*/);

    /* Save not found ? */
    if(id < 0)
    {
        return BUP_NOT_FOUND;
    }

    /* In order to allow previous save data recovery, 
     * don't delete save file itself.
     * Instead, mark it as unused in index structure.
     */

    /* Update index status. */
    cart_used_setstatus(idx, id, 0/*stat*/);

    /* Save index data to SD card. */
    cart_idx_save();

    /* Returns 0 if there's no error. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

/**
 * CartBUPDirectory : tbsize_bram parameter.
 * Used when listing saves from both BRAM and SD card.
 * First tbsize_bram items in tb are already set from caller.
 * In the case SD card save has same name as saves in BRAM, SD card save is not listed.
 *
 * Return value : number of saves in SD card device.
 * If listing can't fit in output parameter, returns negative value of number of saves in SD card device.
**/
Sint32 CartBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb, Uint16 tbsize_bram)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_libram_t * l = VMEM_LIBRAM;
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    vmem_fileindex_t* idx = fileindex_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    BupDir* dir;
    long dir_cnt, cnt;
    short id;
    int i, duplicate;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    /* Bump counter in statistics informations. */
    idx->header.stats.dir_cnt++;

    id = -1;
    dir_cnt = 0;
    do
    {
        id++;
        scd_logout("[RC_CART][BupDir]cart_idx_lookup(\"%s\", %d) ...", (char*)(filename), id);
        id = cart_idx_lookup(idx, (char*)(filename), id/*id_stt*/);
        scd_logout("[RC_CART][BupDir]cart_idx_lookup = %d", id);

        if(id >= 0)
        {
            /* Verify that curent save isn't duplicate with saves in BRAM. */
            duplicate = 0;
            for(i=0; i<tbsize_bram; i++)
            {
                dir = tb+i;
                if(memcmp((void*)(idx->entries[id].bup_filename), (void*)(dir->filename), 12) == 0)
                {
                    duplicate = 1;
                    break;
                }
            }

            if(!duplicate)
            {
                scd_logout("[RC_CART][BupDir] dir_cnt=%d, tbsize=%d", dir_cnt, tbsize);
                if((tbsize_bram+dir_cnt) < (long)tbsize)
                {
                    /* Still enough space to save result data : read header file from SD card. */
                    dir = tb+(tbsize_bram+dir_cnt);

                    cart_set_filename_from_id(fn, id, FILENAME_BUP);
                    scd_logout("[RC_CART][BupDir] fread(fn=\"%s\")", fn);

                    scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);

                    /* Copy the Backup Informations Record to output buffer. */
                    memcpy((void*)dir, (void*)(&(bup_header->dir)), sizeof(BupDir));

                    /* Fix size in block unit. */
                    dir->blocksize = bup_convert_blocksize(VMEM_FILE_BLOCK_SIZE, dir->datasize);
                }
                dir_cnt++;
            }
        }
    } while(id >= 0);

    if((tbsize_bram+dir_cnt) > (long)tbsize)
    {
        /* There was not enough space to save all result data. */
        dir_cnt = -dir_cnt;
    }

    /* Bubble-sort entries in list. */
    cnt = (dir_cnt >= 0 ? dir_cnt : tbsize); 
    if(cnt > 1)
    {
        BupDir* tb_temp = &(f->tb_temp);
        unsigned short i, j;
        for(i=0; i<cnt-1; i++)
        {
            for(j=i+1; j<cnt; j++)
            {
                if(bup_filename_is_greater(tb+i, tb+j))
                {
                    memcpy((void*)tb_temp, (void*)(tb+i) , sizeof(BupDir));
                    memcpy((void*)(tb+i) , (void*)(tb+j) , sizeof(BupDir));
                    memcpy((void*)(tb+j) , (void*)tb_temp, sizeof(BupDir));
                }
            }
        }
    }

    /* Returns the number of successfully read dir entries. */
    scd_logout("[RC_CART]return dir_cnt = %d", dir_cnt);
    return dir_cnt;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 CartBUPVerify(Uint8 *filename, volatile Uint8 *data)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_libram_t * l = VMEM_LIBRAM;
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    vmem_fileindex_t* idx = fileindex_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    unsigned long len, offs, tmp;

    if(l->idx_init != BUP_IDXINIT_DONE)
    {
        cart_idx_populate();
        l->idx_init = BUP_IDXINIT_DONE;
    }

    /* Bump counter in statistics informations. */
    idx->header.stats.verify_cnt++;

    /*
     * (2016/02/05) Wow, 1+ years old code when I was still trying
     * to write save data to cartridge flash memory oO
     * I let this code here for memorial purpose :D
     */

//    unsigned char* buff;
//    BupDir* dir = NULL;
//    unsigned long save_id;
//    unsigned char* ptr;
//    int offs;
//    unsigned long o;
//
//    /* Test data integrity. */
//    if(cart_check_header() != 0)
//    {
//        return BUP_UNFORMAT;
//    }
//
//    /* First, let's look for our save data. */
//    save_id = cart_find_save((char*)filename, &dir);
//
//    /* Save not found ? */
//    if((save_id == SECTOR_SAVEID_EMPTY) || (dir == NULL))
//    {
//        return BUP_NOT_FOUND;
//    }
//
//    /**
//     * Fetch for specified save data.
//    **/
//    for(offs=SECTOR_SIZE; offs<FLASH_LEN; offs+=SECTOR_SIZE)
//    {
//        ptr = (unsigned char*)(FLASH_START_ADR + offs);
//        vmem_sector_t* bl = (vmem_sector_t*)ptr;
//
//        if(bl->save_id == save_id)
//        {
//            /* Found specified save ID : copy to destination buffer. */
//            if(bl->sector_id == 0)
//            {
//                /* First sector : skip BupDir. */
//                memcpy((void*)data, (void*)(ptr + sizeof(vmem_sector_t) + sizeof(BupDir)), bl->sector_len);
//            }
//            else
//            {
//                /* Second~ sectors : compute destination offset. */
//                o  = SECTOR_SIZE-sizeof(BupDir)-sizeof(vmem_sector_t);
//                o += (bl->sector_id - 1) * (SECTOR_SIZE-sizeof(vmem_sector_t));
//
//                memcpy((void*)(data+o), (void*)(ptr + sizeof(vmem_sector_t)), bl->sector_len);
//            }
//        }
//    }

    id = cart_idx_lookup(idx, (char*)(filename), 0/*id_stt*/);

    /* Save not found ? */
    if(id < 0)
    {
        scd_logout("[RC_CART][BUP_Verify]Save not found !");
        return BUP_NOT_FOUND;
    }

    /* Work buffer in order to read save data from SD card. */
#define VERIFY_BUFF_LEN (FILE_ENTRIES_CNT * sizeof(scdf_file_entry_t))
    char* ptr = (char*)(f->file_entries);

    /* First, read backup header from SD card. */
    cart_set_filename_from_id(fn, id, FILENAME_BUP);
    len = scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
    if(len != sizeof(vmem_bup_header_t))
    {
        /* Read error ? */
        scd_logout("[RC_CART][BUP_Verify]Save header read error ! (id=%d)", id);
        return BUP_BROKEN;
    }
    if(memcmp((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) != 0)
    {
        /* Version check. */
        scd_logout("[RC_CART][BUP_Verify]Save version check error ! (id=%d)", id);
        return BUP_BROKEN;
    }

    /* Header check completed. Just keep save
     * data length in memory for next steps.
     */
    len = bup_header->dir.datasize;
    scd_logout("[RC_CART][BUP_Verify]Data length=%d bytes", len);

    /* As we now know backup data size, verify data
     * little by little from SD card.
     */
    for(offs=0; offs<len; offs+=VERIFY_BUFF_LEN)
    {
        /* Compute verify chunk size. */
        if((offs + VERIFY_BUFF_LEN) > len)
        {
            tmp = len - offs;
        }
        else
        {
            tmp = VERIFY_BUFF_LEN;
        }

        if(scf_read(VMEM_MEDIA_ID, fn, sizeof(vmem_bup_header_t)+offs/*offset*/, tmp/*length*/, (void*)ptr) != tmp)
        {
            /* Read error ? */
            scd_logout("[RC_CART][BUP_Verify]Save data read error ! (offset=%d, len=%d, total_len=%d)", offs, tmp, len);
            return BUP_BROKEN;
        }
    
        if(memcmp((void*)ptr, (void*)(data+offs), tmp) != 0)
        {
            /* Data mismatch ? */
            scd_logout("[RC_CART][BUP_Verify]Data mismatch near offset=%d bytes, len=%d bytes", offs, tmp);
            //for(id=0; id<tmp; id++)
            //{
            //    scd_logout(" %04X:%02X %02X", id, ((unsigned char*)ptr)[id], ((unsigned char*)(data+offs))[id]);
            //}
            return BUP_NO_MATCH;
        }
    }

    /* Returns 0 if there's no error. */
    scd_logout("[RC_CART]BUP_Verify success ! (data length = %d bytes)", len);
    return 0;
}

