/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sc_saturn.h>

#include "vmem_main.h"


/**
 * Extended device related sub-routine.
**/
vmem_extglobal_t* ext_global_getptr(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    return &(f->ext_global);
}


//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPSelectPartition(Uint16 num)
{
    BiosBUPSelectPartition(num);

    return CartBUPSelectPartition(num);
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPFormat(void)
{
    //BiosBUPFormat();

    return CartBUPFormat();
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPStatus(Uint32 datasize, BupStat *tb)
{
    BiosBUPStatus(datasize, tb);

    if(tb->freeblock < ((90*tb->totalblock)/100))
    {
        tb->freeblock = (90*tb->totalblock)/100;
        tb->datanum   = tb->freeblock;

        tb->freesize = tb->freeblock * tb->blocksize;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode)
{
    Sint32 ret;

    ret = CartBUPWrite(tb, data, wmode);

    /* Update size in block in BupDir structure.
     * (Some games may use it in the case write failed ?)
     */
    tb->blocksize = bup_convert_blocksize(VMEM_BRAM_BLOCKSZ, tb->datasize);

    return ret;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPRead(Uint8 *filename, volatile Uint8 *data)
{
    Sint32 ret;

    /* First, try to read from BRAM. */
    ret = BiosBUPRead(filename, data);
    if(ret == 0)
    {
        vmem_extglobal_t* global = ext_global_getptr();
        BupDir* tb = &(global->tb);

        /* Move save data from BRAM to SD card. */
        ret = BiosBUPDirectory(filename, 1/*tbsize*/, tb);
        if(ret == 1)
        {
            ret = CartBUPWrite(tb, data, 0/* overwrite check */);
            if(ret == 0)
            {
                /* Could copy save data to SD card, can now delete save data in BRAM. */
                return BiosBUPDelete(filename);
            }
        }
    }

    /* Try to read from SD card. */
    return CartBUPRead(filename, data);
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPDelete(Uint8 *filename)
{
    Sint32 ret1, ret2;

    ret1 = BiosBUPDelete(filename);
    ret2 = CartBUPDelete(filename);

    if(ret2 == 0)
    {
        ret1 = 0;
    }

    return ret1;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb)
{
    Sint32 ret, tbsize_bram, tbsize_sd, i;

    tbsize_bram = BiosBUPDirectory(filename, tbsize, tb);

    if(tbsize_bram < 0)
    {
        tbsize_sd = CartBUPDirectory(filename, tbsize, tb, tbsize/*tbsize_bram*/);

        /* Directories output buffer is full, 
         * so don't care with the case tbsize_sd > 0.
         */
        ret = tbsize_sd+tbsize_bram;
        return ret;
    }

    tbsize_sd = CartBUPDirectory(filename, tbsize, tb, tbsize_bram);

    /* Fix size in block unit for entries in SD card device.
     * (Hybrid device block size is same as BRAM's, so fixing block count
     * for BRAM entries is not necessary.)
     */
    for(i=0; i<tbsize_sd; i++)
    {
        tb[tbsize_bram+i].blocksize = bup_convert_blocksize(VMEM_BRAM_BLOCKSZ, tb[tbsize_bram+i].datasize);
    }

    /* 
     * BRAM entries are sorted, SD device entries are sorted, but 
     * both in a same array are not necessarily sorted.
     * So sorting entries in list is needed.
     */
    /* First, compute the number of entries in output buffer. */
    ret = (tbsize_sd >= 0 ? tbsize_sd+tbsize_bram : tbsize); 
    /* Bubble-sort entries in list. */
    if(ret > 1)
    {
        BupDir* tb_temp = &(VMEM_WORKRAM->tb_temp);
        unsigned short i, j;
        for(i=0; i<ret-1; i++)
        {
            for(j=i+1; j<ret; j++)
            {
                if(bup_filename_is_greater(tb+i, tb+j))
                {
                    memcpy((void*)tb_temp, (void*)(tb+i) , sizeof(BupDir));
                    memcpy((void*)(tb+i) , (void*)(tb+j) , sizeof(BupDir));
                    memcpy((void*)(tb+j) , (void*)tb_temp, sizeof(BupDir));
                }
            }
        }
    }

    if(tbsize_sd < 0)
    {
        ret = tbsize_sd-tbsize_bram;
        return ret;
    }

    ret = tbsize_sd+tbsize_bram;
    return ret;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 ExtBUPVerify(Uint8 *filename, volatile Uint8 *data)
{
    Sint32 ret;

    /* First, try to read from BRAM. */
    ret = BiosBUPRead(filename, data);
    if(ret == 0)
    {
        vmem_extglobal_t* global = ext_global_getptr();
        BupDir* tb = &(global->tb);

        /* Move save data from BRAM to SD card. */
        ret = BiosBUPDirectory(filename, 1/*tbsize*/, tb);
        if(ret == 1)
        {
            ret = CartBUPWrite(tb, data, 0/* overwrite check */);
            if(ret == 0)
            {
                /* Could copy save data to SD card, can now delete save data in BRAM. */
                BiosBUPDelete(filename);
            }
        }
    }

    /* Verify with save on SD card. */
    return CartBUPVerify(filename, data);
}

