/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sc_saturn.h>

#include "vmem_main.h"

/* IP header related stuff. */
#include "../ip_header.h"


/* Block size used in FDD (4KB).
 * Any other block size should be OK, but using a
 * FDD-like value may help compatibility with some games.
 */
#define VMEM_FDD_BLOCK_SIZE (4*1024)

/*
 * FDD saves filename.
 * Written on creation of saves folder for each partitions.
*/
const char* _vmem_fdd_readme_text = "\
This folder contains Saturn save data files generated by\r\n\
Vmem module included in Pseudo Saturn Kai.\r\n\
Informations, support and updates about Vmem and Pseudo Saturn Kai\r\n\
can be found at the following address :\r\n\
http://ppcenter.free.fr/pskai/\r\n";


/**
 * Get pointer to FDD related buffer.
**/
vmem_floppy_saves_t* fdd_saves_getptr(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    return &(f->fdd_saves);
}


/**
 * fdd_set_filename_from_id
 * Format backup/header filename with full path.
 * Example :
 *  char name[BUP_FILE_NAMELEN];
 *  fdd_set_filename_from_id(name, 0, 123, FILENAME_BUP);
 * will output "/FLOPPY1/DATA123.BUP" to name
**/
#define FILENAME_BUP 0
#define FILENAME_TXT 1
void fdd_set_filename_from_id(char* name, unsigned char part_id, short id, short ext)
{
    if(ext == FILENAME_TXT)
    {
        sprintf(name, "/" FDD_BUP_FOLDER "%d/" BUP_FILEPREFIX "%03d" ".TXT", part_id+1, id);
    }
    else
    {
        sprintf(name, "/" FDD_BUP_FOLDER "%d/" BUP_FILEPREFIX "%03d" BUP_DATA_EXT, part_id+1, id);
    }
}


/**
 * fdd_get_free_id
 * Look for save with oldest date stamp.
 * Return negative value when all files are used.
**/
short fdd_get_free_id(void)
{
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    short ret = -1;
    short id;

    for(id=0; id<VMEM_FLOPPY_SAVES_CNT; id++)
    {
        if(fdd_saves->bup_filename[id][0] == '\0')
        {
            ret = id;
            break;
        }
    }

    return ret;
}


/**
 * fdd_save_lookup
 * Look for corresponding backup filename from specified index value.
 * Return negative value in the case no entry matched.
**/
short fdd_save_lookup(char* filename, short id_stt)
{
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    short id;

    for(id=id_stt; id<VMEM_FLOPPY_SAVES_CNT; id++)
    {
        if(fdd_saves->bup_filename[id][0] != '\0')
        { /* Save slot used : verify if its name matches or not. */
            char* filename_idx = (char*)(fdd_saves->bup_filename[id]);
            scd_logout("[RC_FDD]ID[%3d] bup_filename_comp(filename=\"%s\", filename_idx=\"%s\") ...", id, filename, filename_idx);
            if(bup_filename_comp(filename, filename_idx) == 0)
            {
                scd_logout("[RC_FDD]bup_filename_comp==0 return %d", id);
                return id;
            }
        }
    }

    return -1;
}


/**
 * If needed, update saves list into FDD related buffer.
**/
void fdd_saves_update_list(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    scdf_file_entry_t* buff = f->file_entries;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    int i;
    char* fn = global->name_buff;
    unsigned long files_count, files_offs;
    unsigned long len;
    unsigned char sel_part_id_tmp;

    if(fdd_saves->part_id == fdd_saves->sel_part_id)
    {
        /* No need to update list. */
        return;
    }

    /* Update partition ID.
     * Catch the occasion to clear everything in FDD
     * saves list, and to sanitize partition
     * ID to zero (Partition 1) or 1 (Partition 2).
     */
    sel_part_id_tmp = fdd_saves->sel_part_id % 2;
    memset(fdd_saves, 0, sizeof(vmem_floppy_saves_t));
    fdd_saves->part_id = sel_part_id_tmp;
    fdd_saves->sel_part_id = sel_part_id_tmp;


    /* Read save files in folder for specified partition.
     * (Read directory list by chunk of FILE_ENTRIES_CNT entries)
     */
    files_offs = 0;
    do
    {
        sprintf(fn, "/" FDD_BUP_FOLDER "%d/", fdd_saves->part_id+1);
        files_count = scf_filelist(buff, FILE_ENTRIES_CNT, files_offs/*list_offset*/, 0/*list_sort*/, BUP_DATA_EXT/*filter*/, VMEM_MEDIA_ID, fn);
        scd_logout("[RC_FDD]files_count(\"%s\", \"%s\") = %d ...", fn, BUP_DATA_EXT, files_count);

        /* Analyze files in list. */
        for(i=0; i<FILE_ENTRIES_CNT; i++)
        {
            /* Analyze enabled file entries only. */
            if((buff[i].flags & FILEENTRY_FLAG_ENABLED)
            && (!(buff[i].flags & FILEENTRY_FLAG_FOLDER))
            && (i < files_count))
            {
                char* filename = buff[i].filename;
                short id = get_id_from_filename(filename);

                scd_logout("[RC_FDD]entry[%d/%d].filename = \"%s\", id=%d ...", i, files_count, buff[i].filename, id);

                /* Is it a valid .BUP file ? */
                if((id >= 0) && (id < VMEM_FLOPPY_SAVES_CNT))
                {
                    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);

                    len = scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);

                    if((len == sizeof(vmem_bup_header_t))
                    && (memcmp((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) == 0))
                    {
                        /* Copy backup file name to saves list. */
                        int j;
                        for(j=0; j<12; j++)
                        {
                            fdd_saves->bup_filename[id][j] = bup_header->dir.filename[j];
                        }
                        scd_logout("[RC_FDD] tb.filename = \"%s\" -> entries[id=%d]", 
                            bup_print_filename(bup_header->dir.filename), 
                            id);
                    }
                }
            }
        }

        /* Update count of files listed so far. */
        if(files_count <= FILE_ENTRIES_CNT)
        {
            files_offs += files_count;
        }
        else
        {
            files_offs += FILE_ENTRIES_CNT;
        }
    } while(files_offs < files_count);

    /* If needed, save readme file to SD card. */
    sprintf(fn, "/" FDD_BUP_FOLDER "%d/README.TXT", fdd_saves->part_id+1);
    if(scf_size(VMEM_MEDIA_ID, fn) != strlen(_vmem_fdd_readme_text))
    {
        scd_logout("[RC_FDD]write readme text (file=%s)", fn);
        scf_write((void*)_vmem_fdd_readme_text, strlen(_vmem_fdd_readme_text), VMEM_MEDIA_ID, fn);
    }

    scd_logout("[RC_FDD]fdd_saves_update_list END (files_count=%d)", files_count);
}




/**
 * Init file-based Saturn floppy drive device.
**/
void FloppyBUPInit(void)
{
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();

    /* Reset FDD related buffer. */
    memset(fdd_saves, 0, sizeof(vmem_floppy_saves_t));

    /* By default, select first partition, and don't load
     * FDD saves list (by seting part_id to invalid value).
     *
     * This is in order to load saves list only when game
     * requests it so that it won't slowdown other games.
     */
    fdd_saves->part_id = 0xFF;
    fdd_saves->sel_part_id = 0;
}



//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPSelectPartition(Uint16 num)
{
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();

    /* Update partition ID. */
    fdd_saves->sel_part_id = (unsigned char)(num % 2);

    /* Update saves list. */
    fdd_saves_update_list();

    /* Return success code. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPFormat(void)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    char* fn = file_idxglobal_getptr()->name_buff;
    short id;

    scd_logout("[RC_FDD]format STT");

    /* Delete readme file. */
    sprintf(fn, "/" FDD_BUP_FOLDER "%d/README.TXT", fdd_saves->part_id+1);
    scf_delete(VMEM_MEDIA_ID, fn);

    /* Delete all backup files available in saves list. */
    for(id=0; id<VMEM_FLOPPY_SAVES_CNT; id++)
    {
        if(fdd_saves->bup_filename[id][0] != '\0')
        {
            fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
            scf_delete(VMEM_MEDIA_ID, fn);
            fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_TXT);
            scf_delete(VMEM_MEDIA_ID, fn);
        }
    }

    /* request saves list update on next FDD BUP operation. */
    fdd_saves->part_id = 0xFF;

    scd_logout("[RC_FDD]format END");

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPStatus(Uint32 datasize, BupStat *tb)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    int freeblock;

    /* If needed, prepare saves list. */
    fdd_saves_update_list();


    /* Size of block. */
    tb->blocksize  = VMEM_FDD_BLOCK_SIZE;
    /* Size of Backup Ram (in blocks) */
    tb->totalblock = (720 * 1024) / VMEM_FDD_BLOCK_SIZE;
    /* Size of Backup Ram (in bytes) */
    tb->totalsize  = tb->totalblock * tb->blocksize;

    /* Free space (in bytes) */
    tb->freesize   = tb->totalsize;
    freeblock = 0;
    for(id=0; id<VMEM_FLOPPY_SAVES_CNT; id++)
    {
        if(fdd_saves->bup_filename[id][0] != '\0')
        {
            fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
            scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
            if(tb->freesize > bup_header->dir.datasize)
            {
                tb->freesize -= bup_header->dir.datasize;
            }
        }
        else
        {
            freeblock = 1;
        }
    }

    if(freeblock)
    {
        /* If at least one free block available, 
         * return device information with enough free
         * room for probably all Saturn games existing.
         * (According to DTS, recommended max save data is 256KB)
         */
        if(tb->freesize < (256*1024))
        {
            tb->freesize = 256*1024;
        }
    }
    else
    {
        /* If no block available, return zero free
         * space in device information.
         */
        tb->freesize = 0;
    }

    /* Free space (in blocks) */
    tb->freeblock  = tb->freesize / tb->blocksize;
    /* Not sure, but seems to be the same as Free Space (in blocks) */
    tb->datanum    = tb->freeblock;

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    unsigned long len;
    BupDate sat_date;
    struct tm* t;

    /* If needed, prepare saves list. */
    fdd_saves_update_list();


    /* Minimal sanity fix of information in BupDir structure. */
    tb->filename[11] = '\0';
    tb->comment [10] = '\0';
    /* Update size in block in BupDir structure.
     * (Some games may use it in the case write failed ?)
     */
    tb->blocksize = bup_convert_blocksize(VMEM_FDD_BLOCK_SIZE, tb->datasize);

    /* Look for save with same name in saves list. */
    id = fdd_save_lookup((char*)(tb->filename), 0/*id_stt*/);

    /* If save data already exists, verify that we can erase it. */
    if((id >= 0) && (wmode != 0))
    {
        /* Can't overwrite. */
        return BUP_FOUND;
    }

    /* Initialize save header. */
    memset((void*)bup_header, 0, sizeof(vmem_bup_header_t));

    if(id >= 0)
    {
        /* Save not found in list ? Cool, let's re-use
         * its header in order to increment save ID.
         */
        fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
        scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
        bup_header->save_id++;
    }
    else
    {
        /* Save not found in list ? Look for empty save slot then. */
        id = fdd_get_free_id();
    }

    /* Not enough space to save backup data ? */
    if(id < 0)
    {
        /* Too bad, can't save. */
        return BUP_NOT_ENOUGH_MEMORY;
    }

    /* Write backup header to SD card. */
    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
    memcpy((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);
    /* Save current date to header. */
    t = sc_tim_smpcpoll();
    sat_date.year  = t->tm_year - 80;
    sat_date.month = t->tm_mon + 1;
    sat_date.day   = t->tm_mday;
    sat_date.time  = t->tm_hour;
    sat_date.min   = (t->tm_hour >= 60 ? 59 : t->tm_min);
    sat_date.week  = t->tm_wday;
    bup_header->date = bup_setdate_internal(&sat_date);
    memset((void*)(&(bup_header->stats)), 0, sizeof(vmem_bup_stats_t));
    memcpy((void*)(&(bup_header->dir)), (void*)tb, sizeof(BupDir));

    len = scf_write((void*)bup_header, sizeof(vmem_bup_header_t), VMEM_MEDIA_ID, fn);
    if(len != sizeof(vmem_bup_header_t))
    {
        /* Write error : delete data file. */
        fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
        scf_delete(VMEM_MEDIA_ID, fn);
        fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_TXT);
        scf_delete(VMEM_MEDIA_ID, fn);

        return BUP_NON;
    }

    /* Append backup data to SD card. */
    len = scf_append((void*)data, tb->datasize, VMEM_MEDIA_ID, fn);
    if(len != tb->datasize)
    {
        /* Write error : delete data file. */
        fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
        scf_delete(VMEM_MEDIA_ID, fn);
        fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_TXT);
        scf_delete(VMEM_MEDIA_ID, fn);

        return BUP_NON;
    }

    /* Copy backup file name to saves list. */
    int j;
    for(j=0; j<12; j++)
    {
        fdd_saves->bup_filename[id][j] = tb->filename[j];
    }

    /* Extra : save text some information in <save name>.txt file. */
#define TXTINFO_BUFF_LEN (FILE_ENTRIES_CNT * sizeof(scdf_file_entry_t))
    char* ptr = (char*)(f->file_entries);
    unsigned long offs;

    /* First lines : save filename and date.
     * (Comment may not be in ASCII format, so it is safer to reformat it.)
     */
    sprintf(ptr, "Save : %s:%s\r\nData size : %6d bytes, date=0x%08X\r\n", 
        tb->filename, 
        bup_print_comment(tb->comment), 
        (unsigned int)(tb->datasize), 
        (unsigned int)(tb->date));

    /* Then, save vmem specific informations.
     * Note : unlike virtual memory cart, statistics and save
     * revision ID aren't updated. Reason is that virtual FDD
     * usage concerns very few games, hence without the need to
     * log informations from theses.
     */
    offs = strlen(ptr);
    sprintf(ptr+offs, "Vmem Date: %08X\r\n", (unsigned int)(bup_header->date));


    /* Then, save IP header informations.
     *  char hwr_id[16]; // "SEGA SEGASATURN "
     *  char man_id[16]; // "SEGA TP KAISHA-A"
     *  char prd_nb[10]; // "T-053G    "
     *  char version[6]; // "V0.001"
     *  char date[8];    // "19940912"
     *  char dev_id[8];  // "CD-1/1  "
     *  char area[16];   // "JTUBKAEL        "
     *  char pers[16];   // "JAST            "
     *  char title[112]; // ...
     *
     * With FILE_ENTRIES_CNT set to 24, printf buffer is 
     * FILE_ENTRIES_CNT*sizeof(scdf_file_entry_t) = 768 bytes large, 
     * and first lines (see above) are less than 100 bytes
     * and IP header information is 208 bytes large
     * (carriage return characters are not taken in account, but it's OK anyway), 
     * so it is OK to write everything at once.
     *
     * Memo : typical text file size is less than 550 bytes.
     */
    offs = strlen(ptr);
    offs += ip_header_logout((ip_bin_t*)0x06002000, ptr+offs);

    /* Write all text informations at once. */
    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_TXT);
    scf_write((void*)ptr, offs, VMEM_MEDIA_ID, fn);

    /* Returns 0 if there's no error. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPRead(Uint8 *filename, volatile Uint8 *data)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    unsigned long len;

    /* If needed, prepare saves list. */
    fdd_saves_update_list();


    id = fdd_save_lookup((char*)(filename), 0/*id_stt*/);

    /* Save not found ? */
    if(id < 0)
    {
        return BUP_NOT_FOUND;
    }

    /* First, read backup header from SD card. */
    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
    len = scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
    if(len != sizeof(vmem_bup_header_t))
    {
        /* Read error ? */
        return BUP_BROKEN;
    }
    if(memcmp((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) != 0)
    {
        /* Version check. */
        return BUP_BROKEN;
    }

    /* As we now know backup data size, read data from SD card. */
    scd_logout("[RC_FDD][BUP_Read]scf_read(\"%s\", size=%d) ...", fn, bup_header->dir.datasize);
    len = scf_read(VMEM_MEDIA_ID, fn, sizeof(vmem_bup_header_t)/*offset*/, bup_header->dir.datasize, (void*)data);
    if(len != bup_header->dir.datasize)
    {
        /* Read error ? */
        return BUP_BROKEN;
    }

    /* Returns 0 if there's no error. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPDelete(Uint8 *filename)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    char* fn = file_idxglobal_getptr()->name_buff;
    short id;

    /* If needed, prepare saves list. */
    fdd_saves_update_list();


    id = fdd_save_lookup((char*)(filename), 0/*id_stt*/);

    /* Save not found ? */
    if(id < 0)
    {
        return BUP_NOT_FOUND;
    }

    /* Delete the save and its description files. */
    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
    scf_delete(VMEM_MEDIA_ID, fn);
    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_TXT);
    scf_delete(VMEM_MEDIA_ID, fn);

    /* Mark the save as unused in the list. */
    fdd_saves->bup_filename[id][0] = '\0';

    /* Returns 0 if there's no error. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

/**
 * FloppyBUPDirectory
 *
 * Return value : number of saves in SD card device.
 * If listing can't fit in output parameter, returns negative value of number of saves in SD card device.
**/
Sint32 FloppyBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    BupDir* dir;
    long dir_cnt, cnt;
    short id;

    /* If needed, prepare saves list. */
    fdd_saves_update_list();


    id = -1;
    dir_cnt = 0;
    do
    {
        id++;
        scd_logout("[RC_FDD][BupDir]fdd_save_lookup(\"%s\", %d) ...", (char*)(filename), id);
        id = fdd_save_lookup((char*)(filename), id/*id_stt*/);
        scd_logout("[RC_FDD][BupDir]fdd_save_lookup = %d", id);

        if(id >= 0)
        {
            scd_logout("[RC_FDD][BupDir] dir_cnt=%d, tbsize=%d", dir_cnt, tbsize);
            if(dir_cnt < (long)tbsize)
            {
                /* Still enough space to save result data : read header file from SD card. */
                dir = tb + dir_cnt;

                fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
                scd_logout("[RC_FDD][BupDir] fread(fn=\"%s\")", fn);

                scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);

                /* Copy the Backup Informations Record to output buffer. */
                memcpy((void*)dir, (void*)(&(bup_header->dir)), sizeof(BupDir));

                /* Fix size in block unit. */
                dir->blocksize = bup_convert_blocksize(VMEM_FDD_BLOCK_SIZE, dir->datasize);
            }
            dir_cnt++;
        }
    } while(id >= 0);

    if(dir_cnt > (long)tbsize)
    {
        /* There was not enough space to save all result data. */
        dir_cnt = -dir_cnt;
    }

    /* Bubble-sort entries in list. */
    cnt = (dir_cnt >= 0 ? dir_cnt : tbsize); 
    if(cnt > 1)
    {
        BupDir* tb_temp = &(f->tb_temp);
        unsigned short i, j;
        for(i=0; i<cnt-1; i++)
        {
            for(j=i+1; j<cnt; j++)
            {
                if(bup_filename_is_greater(tb+i, tb+j))
                {
                    memcpy((void*)tb_temp, (void*)(tb+i) , sizeof(BupDir));
                    memcpy((void*)(tb+i) , (void*)(tb+j) , sizeof(BupDir));
                    memcpy((void*)(tb+j) , (void*)tb_temp, sizeof(BupDir));
                }
            }
        }
    }

    /* Returns the number of successfully read dir entries. */
    scd_logout("[RC_FDD]return dir_cnt = %d", dir_cnt);
    return dir_cnt;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 FloppyBUPVerify(Uint8 *filename, volatile Uint8 *data)
{
    vmem_workram_t* f = VMEM_WORKRAM;
    vmem_floppy_saves_t* fdd_saves = fdd_saves_getptr();
    vmem_idxglobal_t* global = file_idxglobal_getptr();
    char* fn = global->name_buff;
    vmem_bup_header_t* bup_header = &(global->bup_hdr_buff);
    short id;
    unsigned long len, offs, tmp;

    /* If needed, prepare saves list. */
    fdd_saves_update_list();


    id = fdd_save_lookup((char*)(filename), 0/*id_stt*/);

    /* Save not found ? */
    if(id < 0)
    {
        scd_logout("[RC_FDD][BUP_Verify]Save not found !");
        return BUP_NOT_FOUND;
    }

    /* Work buffer in order to read save data from SD card. */
#define VERIFY_BUFF_LEN (FILE_ENTRIES_CNT * sizeof(scdf_file_entry_t))
    char* ptr = (char*)(f->file_entries);

    /* First, read backup header from SD card. */
    fdd_set_filename_from_id(fn, fdd_saves->part_id, id, FILENAME_BUP);
    len = scf_read(VMEM_MEDIA_ID, fn, 0/*offset*/, sizeof(vmem_bup_header_t), (void*)bup_header);
    if(len != sizeof(vmem_bup_header_t))
    {
        /* Read error ? */
        scd_logout("[RC_FDD][BUP_Verify]Save header read error ! (id=%d)", id);
        return BUP_BROKEN;
    }
    if(memcmp((void*)(bup_header->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) != 0)
    {
        /* Version check. */
        scd_logout("[RC_FDD][BUP_Verify]Save version check error ! (id=%d)", id);
        return BUP_BROKEN;
    }

    /* Header check completed. Just keep save
     * data length in memory for next steps.
     */
    len = bup_header->dir.datasize;
    scd_logout("[RC_FDD][BUP_Verify]Data length=%d bytes", len);

    /* As we now know backup data size, verify data
     * little by little from SD card.
     */
    for(offs=0; offs<len; offs+=VERIFY_BUFF_LEN)
    {
        /* Compute verify chunk size. */
        if((offs + VERIFY_BUFF_LEN) > len)
        {
            tmp = len - offs;
        }
        else
        {
            tmp = VERIFY_BUFF_LEN;
        }

        if(scf_read(VMEM_MEDIA_ID, fn, sizeof(vmem_bup_header_t)+offs/*offset*/, tmp/*length*/, (void*)ptr) != tmp)
        {
            /* Read error ? */
            scd_logout("[RC_FDD][BUP_Verify]Save data read error ! (offset=%d, len=%d, total_len=%d)", offs, tmp, len);
            return BUP_BROKEN;
        }
    
        if(memcmp((void*)ptr, (void*)(data+offs), tmp) != 0)
        {
            /* Data mismatch ? */
            scd_logout("[RC_FDD][BUP_Verify]Data mismatch near offset=%d bytes, len=%d bytes", offs, tmp);
            //for(id=0; id<tmp; id++)
            //{
            //    scd_logout(" %04X:%02X %02X", id, ((unsigned char*)ptr)[id], ((unsigned char*)(data+offs))[id]);
            //}
            return BUP_NO_MATCH;
        }
    }

    /* Returns 0 if there's no error. */
    scd_logout("[RC_FDD]BUP_Verify success ! (data length = %d bytes)", len);
    return 0;
}

