/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _DEV_FLOPPY_H_
#define _DEV_FLOPPY_H_

#include <stdio.h>
#include "dev_cart.h"

/**
 * File Based Virtual FDD.
 * Backup data are read/written from/to "FLOPPY<partition>" folder in SD card.
 *
 *
 * Example :
 * FDD1 (folder)
 *  + DATA001.BUP
 *  + DATA002.BUP
 *  + README.TXT
 *
 * Each .BUP files contain following data :
 *  - Header data, including BupDir structure.
 *  - Save data itself.
 *
 * File naming follows the conditions below :
 *  - 8.3 file name
 *  - .BUP extension
 *
 * Because of memory constraints, maximum save count
 * per FDD is limited to few (10) saves.
 * Since few saves per FDD are supported, indexing of
 * saves (as done in Virtual Memory Cartridge) is not
 * supported, and it's consequently not possible to
 * recover overwritten/deleted saves.
 *
 * Virtual FDD supports two partitions per FDD :
 *  - Saves from first partition  : FLOPPY1 folder.
 *  - Saves from second partition : FLOPPY2 folder.
 * Don't know which game(s) support two partitions, but
 * it's better to add from start instead of realizing
 * in the future that xyz game works better with two
 * partitions ...
 *
 *
 * Memory usage :
 * Refer to vmem_main.h file for details.
**/
#define FDD_BUP_FOLDER     "FLOPPY"

/* Backup file example : "/FLOPPY1/DATA999.XYZ"
 * -> 20 characters
 * + null terminating character
 * => round to 24 characters
 *  => Since "FLOPPYx" has same character count as "MEMCART", 
 *     re-use BUP_FILE_NAMELEN from virtual memory cart.
 */


// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif



/**
 * Maximum saves count for floppy device.
 *
 * Only few games use FDD, so there is probably no
 * need to support many saves at once.
 * Entry for one save takes around 16 bytes in BUP library
 * memory (16KB + 8KB - other vmem things), so increasing
 * saves count to a reasonable value in the future shouldn't
 * be a problem.
**/
#define VMEM_FLOPPY_SAVES_CNT (10)


/**
 * File index entry structure.
 * Associates SD card filename and backup filename
**/
typedef struct _vmem_floppy_saves_t
{
    /* Backup filename (null terminated), from BupDir structure. */
    Uint8 bup_filename[VMEM_FLOPPY_SAVES_CNT][12];

    /* Partition ID currently loaded.
     *  - 0 : partition 1
     *  - 1 : partition 2
     * If partition ID is invalid (different from 0 or 1), 
     * or different from selected partition ID, then 
     * save list is reloaded by using `sel_part_id' value.
     */
    unsigned char part_id;

    /* Partition ID selected with BUP_SelPart function.
     *  - 0 : partition 1
     *  - 1 : partition 2
     */
    unsigned char sel_part_id;

    /* Padding. */
    unsigned char unused[2];
} vmem_floppy_saves_t;


// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


void FloppyBUPInit(void);

Sint32 FloppyBUPSelectPartition(Uint16 num);
Sint32 FloppyBUPFormat(void);
Sint32 FloppyBUPStatus(Uint32 datasize, BupStat *tb);
Sint32 FloppyBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode);
Sint32 FloppyBUPRead(Uint8 *filename, volatile Uint8 *data);
Sint32 FloppyBUPDelete(Uint8 *filename);
Sint32 FloppyBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb);
Sint32 FloppyBUPVerify(Uint8 *filename, volatile Uint8 *data);

#endif // _DEV_FLOPPY_H_
