/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
/*  Copyright 2006-2007 Theo Berkau

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sc_saturn.h>

#include "vmem_main.h"



unsigned char MappedMemoryReadByte(unsigned long addr)
{
    return *((unsigned char*)addr);
}
void MappedMemoryWriteByte(unsigned long addr, unsigned char val)
{
    *((unsigned char*)addr) = val;
}
void MappedMemoryWriteWord(unsigned long addr, unsigned short val)
{
    *((unsigned short*)addr) = val;
}

//////////////////////////////////////////////////////////////////////////////

static int CheckHeader(void)
{
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

static int CalcSaveSize(unsigned long tableaddr, int blocksize)
{
    int numblocks=0;

    // Now figure out how many blocks this save is
    for(;;)
    {
        unsigned short block;

        if (((tableaddr-1) & ((blocksize << 1) - 1)) == 0)
        {
            tableaddr += 8;
        }

        block = (MappedMemoryReadByte(tableaddr) << 8) | MappedMemoryReadByte(tableaddr + 2);
        //scd_logout("[Bios::CalcSaveSize] tableaddr=0x%08X, block=0x%04X", tableaddr, block);
        if (block == 0)
        {
            break;
        }

        tableaddr += 4;

        numblocks++;
    }

    scd_logout("[Bios::CalcSaveSize] return numblocks=%d", numblocks);

    return numblocks;
}

//////////////////////////////////////////////////////////////////////////////

static unsigned long GetFreeSpace(unsigned long size, unsigned long addr, unsigned long blocksize)
{
    unsigned long i;
    unsigned long usedblocks=0;

    for (i = ((2 * blocksize) << 1); i < (size << 1); i += (blocksize << 1))
    {
        // Find a block with the start of a save
        if (((char)MappedMemoryReadByte(addr + i + 1)) < 0)
        {
            // Now figure out how many blocks this save is
            usedblocks += (CalcSaveSize(addr+i+0x45, blocksize) + 1);
        }
    }

    i = (size / blocksize) - 2 - usedblocks;

    scd_logout("[GetFreeSpace] ret=%d", i);

    return i;
}

//////////////////////////////////////////////////////////////////////////////

static unsigned long FindSave(unsigned long stringaddr, unsigned long blockoffset, unsigned long size, unsigned long addr, unsigned long blocksize)
{
    unsigned long i;

    for (i = ((blockoffset * blocksize) << 1); i < (size << 1); i += (blocksize << 1))
    {
        // Find a block with the start of a save
        if (((char)MappedMemoryReadByte(addr + i + 1)) < 0)
        {
            int i3;

            // See if string matches, or if there's no string to check, just copy
            // the data over
            for (i3 = 0; i3 < 11; i3++)
            {
                unsigned char data = MappedMemoryReadByte(stringaddr+i3);

                if (MappedMemoryReadByte(addr+i+0x9+(i3*2)) != data)
                {
                    if (data == 0)
                    {
                        // There's no string to match
                        return ((i / blocksize) >> 1);
                    }
                    else
                    {
                        // No Match, move onto the next block
                        i3 = 12;
                    }
                }
                else
                {
                    // Match
                    if (i3 == 10 || data == 0)
                    {
                        return ((i / blocksize) >> 1);
                    }
                }
            }
        }
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

static void DeleteSave(unsigned long addr, unsigned long blockoffset, unsigned long blocksize)
{
    MappedMemoryWriteByte(addr + (blockoffset * blocksize * 2) + 0x1, 0x00);
}

//////////////////////////////////////////////////////////////////////////////

static unsigned short *GetFreeBlocks(unsigned long addr, unsigned long blocksize, unsigned long numblocks, unsigned long size)
{
    unsigned char *blocktbl = VMEM_WORKRAM->bram_buff.blocktbl;
    unsigned short *freetbl = VMEM_WORKRAM->bram_buff.blockmap;
    unsigned long tableaddr;
    unsigned long i;
    unsigned long blockcount=0;

    // Create a table that tells us which blocks are free and used
    memset(blocktbl, 0, VMEM_BRAM_BLOCKCNT*sizeof(unsigned char));
    for (i = ((2 * blocksize) << 1); i < (size << 1); i += (blocksize << 1))
    {
        // Find a block with the start of a save
        if (((char)MappedMemoryReadByte(addr + i + 1)) < 0)
        {
            tableaddr = addr+i+0x45;
            blocktbl[i / (blocksize << 1)] = 1;

            // Now let's figure out which blocks are used
            for(;;)
            {
                unsigned short block;

                if (((tableaddr-1) & ((blocksize << 1) - 1)) == 0)
                {
                    tableaddr += 8;
                }

                block = (MappedMemoryReadByte(tableaddr) << 8) | MappedMemoryReadByte(tableaddr + 2);
                if (block == 0)
                {
                    break;
                }
                tableaddr += 4;

                // block is used
                blocktbl[block] = 1;
            }
        }
    }

    // Find some free blocks for us to use
    memset(freetbl, 0, VMEM_BRAM_BLOCKCNT*sizeof(unsigned short));
    for (i = 2; i < (size / blocksize); i++)
    {
        if (blocktbl[i] == 0)
        {
            freetbl[blockcount] = (unsigned short)i;
            blockcount++;

            if (blockcount >= numblocks)
            {
                break;
            }
        }
    }

    // Ok, we're all done
    return freetbl;
}

//////////////////////////////////////////////////////////////////////////////

static unsigned short *ReadBlockTable(unsigned long addr, unsigned long *tableaddr, int block, int blocksize, int *numblocks, int *blocksread)
{
    unsigned short *blocktbl = VMEM_WORKRAM->bram_buff.blockmap;
    int i=0;

    scd_logout("[Bios::ReadBlockTable] addr=0x%08X, tableaddr=0x%08X, block=0x%04X, blocksize=%d", 
            addr, 
            tableaddr, 
            block, 
            blocksize);

    tableaddr[0] = addr + (block * blocksize * 2) + 0x45;
    blocksread[0]=0;
    scd_logout("[Bios::ReadBlockTable] tableaddr[0] = 0x%08X", tableaddr[0]);

    // First of all figure out how large of buffer we need
    numblocks[0] = CalcSaveSize(tableaddr[0], blocksize);

    scd_logout("[Bios::ReadBlockTable] numblocks[0] = %d", numblocks[0]);

    memset(blocktbl, 0, VMEM_BRAM_BLOCKCNT*sizeof(unsigned char));

    /* Empty block table ? */
    if(numblocks[0] == 0)
    {
        scd_logout("[Bios::ReadBlockTable] Empty table (single block save data ?)");
        tableaddr[0] += 4;

        return blocktbl;
    }

    // Now read in the table
    for(i = 0; i < numblocks[0]; i++)
    {
        unsigned short block;
        block = (MappedMemoryReadByte(tableaddr[0]) << 8) | MappedMemoryReadByte(tableaddr[0] + 2);
        tableaddr[0] += 4;

        if (((tableaddr[0]-1) & ((blocksize << 1) - 1)) == 0)
        {
            tableaddr[0] = addr + (blocktbl[blocksread[0]] * blocksize * 2) + 9;
            blocksread[0]++;
        }
        blocktbl[i] = block;
    }

    tableaddr[0] += 4;

    scd_logout("[Bios::ReadBlockTable] Return blocktbl=0x%08X", blocktbl);

    return blocktbl;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPSelectPartition(Uint16 num)
{
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPFormat(void)
{
    unsigned char* mem = (unsigned char*)VMEM_BRAM_ADDR;
    unsigned long size = VMEM_BRAM_MEMSZ;

    int i, i2;
    unsigned long i3;
    unsigned char header[32] = {
        0xFF, 'B', 0xFF, 'a', 0xFF, 'c', 0xFF, 'k',
        0xFF, 'U', 0xFF, 'p', 0xFF, 'R', 0xFF, 'a',
        0xFF, 'm', 0xFF, ' ', 0xFF, 'F', 0xFF, 'o',
        0xFF, 'r', 0xFF, 'm', 0xFF, 'a', 0xFF, 't'
    };

    // Fill in header
    for(i2 = 0; i2 < 4; i2++)
    {
        for(i = 0; i < 32; i++)
        {
            mem[(i2 * 32) + i] = header[i];
        }
    }

    // Clear the rest
    for(i3 = 0x40; i3 < size; i3++)
    {
        mem[i3*2+1] = 0x00;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPStatus(Uint32 datasize, BupStat *tb)
{
    unsigned long size      = VMEM_BRAM_MEMSZ;
    unsigned long addr      = VMEM_BRAM_ADDR;
    unsigned long blocksize = VMEM_BRAM_BLOCKSZ;
    unsigned long freeblocks=0;

    // Make sure there's a proper header, and return if there's any other errors
    if (CheckHeader() != 0)
    {
        // Error
        return BUP_NON;
    }

    freeblocks = GetFreeSpace(size, addr, blocksize);

    tb->totalsize  = size; // Size of Backup Ram (in bytes)
    tb->totalblock = size / blocksize; // Size of Backup Ram (in blocks)
    tb->blocksize  = blocksize; // Size of block
    tb->freesize   = ((blocksize - 6) * freeblocks) - 30; // Free space(in bytes)
    tb->freeblock  = freeblocks; // Free space(in blocks)
    tb->datanum    = freeblocks; // Not sure, but seems to be the same as Free Space(in blocks)

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode)
{
    unsigned long size      = VMEM_BRAM_MEMSZ;
    unsigned long addr      = VMEM_BRAM_ADDR;
    unsigned long blocksize = VMEM_BRAM_BLOCKSZ;
    unsigned long block;
    unsigned long savesize;
    unsigned short *blocktbl;
    unsigned long workaddr;
    unsigned long blockswritten=0;
    unsigned long datasize;
    unsigned long i;

    scd_logout("[BiosBUPWrite] STT");

    /* Minimal sanity fix of information in BupDir structure. */
    tb->filename[11] = '\0';
    tb->comment [10] = '\0';

    datasize = tb->datasize;

    // Let's figure out how many blocks will be needed for the save
    // Consider available size of a block as "blocksize - 6", 
    // because 4 bytes at the beginning of each headers are
    // reserved to save file system, and because each blocks
    // require 2 bytes in allocation table.
    // First block doesn't requires its entry in allocation
    // table, but since allocation table ends with dummy zero
    // block ID, we can consider that data available in first
    // block is the same as in the other blocks.
    // 
    // Memo when used with internal memory :
    //  - Block size is 64 bytes
    //  - datasize=  1; -> savesize:  1 block
    //  - datasize= 28; -> savesize:  1 block
    //  - datasize= 29; -> savesize:  2 blocks
    //  - datasize= 86; -> savesize:  2 blocks
    //  - datasize= 87; -> savesize:  3 blocks
    savesize = 1 + ((datasize + 0x1D) / (blocksize - 6));

    scd_logout("[BiosBUPWrite] savesize=%d", savesize);
    /* Update size in block in BupDir structure.
     * (Some games may use it in the case write failed ?)
     */
    tb->blocksize = savesize;

    // See if save exists already
    if ((block = FindSave((unsigned long)tb, 2, size, addr, blocksize)) != 0)
    {
        // save exists

        // Should we be overwriting it?
        if (wmode != 0)
        {
            // Nope, let's bail instead
            return BUP_FOUND;
        }

        // Delete old save
        DeleteSave(addr, block, blocksize);
    }

    // Will it blend? Err... fit
    if (savesize > GetFreeSpace(size, addr, blocksize))
    {
        // Nope, time to bail
        return BUP_NOT_ENOUGH_MEMORY;
    }

    // Find free blocks for the save
    scd_logout("[BiosBUPWrite] GetFreeBlocks ...");
    if ((blocktbl = GetFreeBlocks(addr, blocksize, savesize, size)) == NULL)
    {
        // Just return an error that might make sense
        return BUP_BROKEN;
    }

    scd_logout("[BiosBUPWrite] blocktbl[0]=0x%04X", blocktbl[0]);

    // Create save
    workaddr = addr + (blocktbl[0] * blocksize * 2);

    MappedMemoryWriteByte(workaddr+0x1, 0x80);

    // Copy over filename
    unsigned long tbptr = (unsigned long)tb;
    for (i = workaddr+0x9; i < ((workaddr+0x9) + (11 * 2)); i+=2)
    {
        MappedMemoryWriteByte(i, MappedMemoryReadByte(tbptr));
        tbptr++;
    }

    tbptr++;

    // Copy over comment
    for (i = workaddr+0x21; i < ((workaddr+0x21) + (10 * 2)); i+=2)
    {
        MappedMemoryWriteByte(i, MappedMemoryReadByte(tbptr));
        tbptr++;
    }

    // Copy over language
    MappedMemoryWriteByte(workaddr+0x1F, MappedMemoryReadByte(tbptr));
    tbptr++;

    tbptr++;

    // Copy over date
    for (i = workaddr+0x35; i < ((workaddr+0x35) + (4 * 2)); i+=2)
    {
        MappedMemoryWriteByte(i, MappedMemoryReadByte(tbptr));
        tbptr++;
    }

    // Copy over data size
    for (i = workaddr+0x3D; i < ((workaddr+0x3D) + (4 * 2)); i+=2)
    {
        MappedMemoryWriteByte(i, MappedMemoryReadByte(tbptr));
        tbptr++;
    }

    // write the block table
    workaddr += 0x45;

    for (i = 1; i < savesize; i++)
    {
        scd_logout("[BiosBUPWrite] blocktbl[%3d]=0x%04X", i, blocktbl[i]);

        MappedMemoryWriteByte(workaddr, (unsigned char)(blocktbl[i] >> 8));
        workaddr+=2;
        MappedMemoryWriteByte(workaddr, (unsigned char)blocktbl[i]);
        workaddr+=2;

        if (((workaddr-1) & ((blocksize << 1) - 1)) == 0)
        {
            // Next block
            blockswritten++;
            workaddr = addr + (blocktbl[blockswritten] * blocksize * 2) + 9;
        }
    }

    // Write 2 blank bytes so we now how large the table size is next time
    MappedMemoryWriteByte(workaddr, 0);
    workaddr+=2;
    MappedMemoryWriteByte(workaddr, 0);
    workaddr+=2;

    // Lastly, write the actual save data
    while (datasize > 0)
    {
        if (((workaddr-1) & ((blocksize << 1) - 1)) == 0)
        {
            // Next block
            blockswritten++;
            workaddr = addr + (blocktbl[blockswritten] * blocksize * 2) + 9;
        }

        MappedMemoryWriteByte(workaddr, MappedMemoryReadByte((unsigned long)data));
        datasize--;
        data++;
        workaddr+=2;
    }

    scd_logout("[BiosBUPWrite]blockswritten=%d", blockswritten);

    scd_logout("[BiosBUPWrite] END OK");

    return 0; // returns 0 if there's no error
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPRead(Uint8 *filename, volatile Uint8 *data)
{
    unsigned long size      = VMEM_BRAM_MEMSZ;
    unsigned long addr      = VMEM_BRAM_ADDR;
    unsigned long blocksize = VMEM_BRAM_BLOCKSZ;
    unsigned long block;
    unsigned long tableaddr;
    unsigned short *blocktbl;
    int numblocks;
    int blocksread;
    unsigned long datasize;

    // See if save exists
    if ((block = FindSave((unsigned long)filename, 2, size, addr, blocksize)) == 0)
    {
        // save doesn't exist
        scd_logout("[BiosBUPRead] Save not found, exit.");
        return BUP_NOT_FOUND;
    }

    tableaddr = addr + (block * blocksize * 2) + 0x3D;
    datasize = (MappedMemoryReadByte(tableaddr) << 24) | (MappedMemoryReadByte(tableaddr + 2) << 16) |
              (MappedMemoryReadByte(tableaddr+4) << 8) | MappedMemoryReadByte(tableaddr + 6);
    scd_logout("[BiosBUPRead] datasize = %d", datasize);

    // Read in Block Table
    if ((blocktbl = ReadBlockTable(addr, &tableaddr, block, blocksize, &numblocks, &blocksread)) == NULL)
    {
        // Just return an error that might make sense
        scd_logout("[BiosBUPRead] ReadBlockTable returned NULL, exit.");
        return BUP_BROKEN;
    }

    // Now let's read in the data
    while (datasize > 0)
    {
        if (((tableaddr-1) & ((blocksize << 1) - 1)) == 0)
        {
            // Load up the next block
            tableaddr = addr + (blocktbl[blocksread] * blocksize * 2) + 9;
            blocksread++;
        }

        MappedMemoryWriteByte((unsigned long)data, MappedMemoryReadByte(tableaddr));
        datasize--;
        data++;
        tableaddr+=2;
    }

    return 0; // returns 0 if there's no error
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPDelete(Uint8 *filename)
{
    unsigned long size      = VMEM_BRAM_MEMSZ;
    unsigned long addr      = VMEM_BRAM_ADDR;
    unsigned long blocksize = VMEM_BRAM_BLOCKSZ;
    unsigned long block;

    // See if save exists
    if ((block = FindSave((unsigned long)filename, 2, size, addr, blocksize)) == 0)
    {
        // Since the save doesn't exist, let's bail with an error
        return BUP_NOT_FOUND;
    }

    DeleteSave(addr, block, blocksize);

    return 0; // returns 0 if there's no error
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb)
{
    unsigned long size      = VMEM_BRAM_MEMSZ;
    unsigned long addr      = VMEM_BRAM_ADDR;
    unsigned long blocksize = VMEM_BRAM_BLOCKSZ;
    unsigned long i, cnt;
    unsigned long blockoffset=2;

    // int findmatch = MappedMemoryReadByte(sh->regs.R[5]);

    unsigned long tbptr = (unsigned long)tb;
    for (i = 0; i < tbsize; i++)
    {
        unsigned long i4;
        unsigned long datasize=0;
        unsigned long block = FindSave((unsigned long)filename, blockoffset, size, addr, blocksize);

        if (block == 0)
        {
            break;
        }

        blockoffset = block+1;

        // Alright, we found a match :) Time to copy over some data
        block = addr + (blocksize * block * 2);

        // Copy over filename
        for (i4 = block+0x9; i4 < ((block+0x9) + (11 * 2)); i4+=2)
        {
            MappedMemoryWriteByte(tbptr, MappedMemoryReadByte(i4));
            tbptr++;
        }

        MappedMemoryWriteByte(tbptr, 0);
        tbptr++;

        // Copy over comment
        for (i4 = block+0x21; i4 < ((block+0x21) + (10 * 2)); i4+=2)
        {
            MappedMemoryWriteByte(tbptr, MappedMemoryReadByte(i4));
            tbptr++;
        }

        // Copy over language
        MappedMemoryWriteByte(tbptr, MappedMemoryReadByte(block+0x1F));
        tbptr++;

        MappedMemoryWriteByte(tbptr, 0);
        tbptr++;

        // Copy over date
        for (i4 = block+0x35; i4 < ((block+0x35) + (4 * 2)); i4+=2)
        {
            MappedMemoryWriteByte(tbptr, MappedMemoryReadByte(i4));
            tbptr++;
        }

        // Copy over data size
        for (i4 = block+0x3D; i4 < ((block+0x3D) + (4 * 2)); i4+=2)
        {
            unsigned char data;
            datasize <<= 8;
            data = MappedMemoryReadByte(i4);
            MappedMemoryWriteByte(tbptr, data);
            datasize |= data;
            tbptr++;
        }

        // Calculate block size from the data size, and then copy it over
        MappedMemoryWriteWord(tbptr, (unsigned short)(((datasize + 0x1D) / (blocksize - 6)) + 1));
        tbptr += 2;

        /* 2 more bytes for data alignment. */
        tbptr += 2;
    }

    /* Bubble-sort entries in list. */
    cnt = (i <= tbsize ? i : tbsize); 
    if(cnt > 1)
    {
        BupDir* tb_temp = &(VMEM_WORKRAM->tb_temp);
        unsigned short i, j;
        for(i=0; i<cnt-1; i++)
        {
            for(j=i+1; j<cnt; j++)
            {
                if(bup_filename_is_greater(tb+i, tb+j))
                {
                    memcpy((void*)tb_temp, (void*)(tb+i) , sizeof(BupDir));
                    memcpy((void*)(tb+i) , (void*)(tb+j) , sizeof(BupDir));
                    memcpy((void*)(tb+j) , (void*)tb_temp, sizeof(BupDir));
                }
            }
        }
    }

    return i; // returns the number of successfully read dir entries
}

//////////////////////////////////////////////////////////////////////////////

Sint32 BiosBUPVerify(Uint8 *filename, volatile Uint8 *data)
{
    unsigned long size      = VMEM_BRAM_MEMSZ;
    unsigned long addr      = VMEM_BRAM_ADDR;
    unsigned long blocksize = VMEM_BRAM_BLOCKSZ;
    unsigned long block;
    unsigned long tableaddr;
    unsigned long datasize;
    unsigned short *blocktbl;
    int numblocks;
    int blocksread;

    // See if save exists
    if ((block = FindSave((unsigned long)filename, 2, size, addr, blocksize)) == 0)
    {
        // Since the save doesn't exist, let's bail with an error
        return BUP_NOT_FOUND;
    }

    tableaddr = addr + (block * blocksize * 2) + 0x3D;
    datasize = (MappedMemoryReadByte(tableaddr) << 24) | (MappedMemoryReadByte(tableaddr + 2) << 16) |
               (MappedMemoryReadByte(tableaddr+4) << 8) | MappedMemoryReadByte(tableaddr + 6);

    // Read in Block Table
    if ((blocktbl = ReadBlockTable(addr, &tableaddr, block, blocksize, &numblocks, &blocksread)) == NULL)
    {
        // Just return an error that might make sense
        return BUP_BROKEN;
    }

    // Now let's read in the data, and check to see if it matches 
    while (datasize > 0)
    {
        if (((tableaddr-1) & ((blocksize << 1) - 1)) == 0)
        {
            // Load up the next block
            tableaddr = addr + (blocktbl[blocksread] * blocksize * 2) + 9;
            blocksread++;
        }

        if (MappedMemoryReadByte((unsigned long)data) != MappedMemoryReadByte(tableaddr))
        {
            // Ok, the data doesn't match
            return BUP_NO_MATCH;
        }

        datasize--;
        data++;
        tableaddr+=2;
    }

    return 0; // returns 0 if there's no error
}

