/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h>

#include "vmem_main.h"

/* Device#0 : Re-implementation of internal backup library functions. */
#include "dev_internal.h"
/* Device#1 : Virtual memory cartridge functions. */
#include "dev_cart.h"
/* Device#2 : Virtual Saturn floppy drive functions. */
#include "dev_floppy.h"

/* Vmem compatibility reporter. */
#include "compat_vmem.h"

/* UBC debugger stuff, no longer used, but UI stuff
 * kept in case needed for debugging other modules.
 */
#include "debug_ui/ubc_debug.h"

/**
 * Return this ROM code version.
**/
//#include "../../satcom_lib/sc_rom_code.h" -> included from sc_saturn.h
int vmem_rcf_version(vmem_version_t* ver)
{
    sc_rom_code_header_t* hdr = (sc_rom_code_header_t*)(ROM_VMEM_EXEC_ADDRESS - sizeof(sc_rom_code_header_t));

    memset((void*)ver->version1, '\0', sizeof(ver->version1));
    memset((void*)ver->version2, '\0', sizeof(ver->version2));
    memset((void*)ver->date    , '\0', sizeof(ver->date    ));
    memset((void*)ver->comment1, '\0', sizeof(ver->comment1));
    memset((void*)ver->comment2, '\0', sizeof(ver->comment2));

    /* Copy version strings from ROM code header. */
    strncpy(ver->version1, hdr->version_string1, sizeof(ver->version1)-1);
    strncpy(ver->version2, hdr->version_string2, sizeof(ver->version2)-1);

    /* Write build date. */
    strncpy(ver->date, __DATE__ ", " __TIME__, sizeof(ver->date)-1);

    /* Format additional comments. */
    vmem_workram_t* f = NULL;
    vmem_libram_t * l = NULL;
    snprintf(ver->comment1, sizeof(ver->comment1)-1, "sz_lib=%d, unused=%d", sizeof(vmem_libram_t ), sizeof(l->unused));
    snprintf(ver->comment2, sizeof(ver->comment2)-1, "sz_wk =%d, unused=%d", sizeof(vmem_workram_t), sizeof(f->unused));

    return 0;
}


/**
 * Convert BUP library device ID to Virtual Device ID.
**/
unsigned char vmem_get_virtual_device(Uint32 device)
{
    if((VMEM_PARAMS_PTR->magic_ca != 0xCA)
    || (VMEM_PARAMS_PTR->magic_fe != 0xFE))
    {
        const char default_devices[3] = {VMEM_DEV_BRAM_INT, VMEM_DEV_MEMCART | VMEM_DEV_SDCARD, VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD};
        return default_devices[device%3];
    }

    return (VMEM_PARAMS_PTR->dev_cfg[device%3]);
}
/**
 * Verify if debug mode or not.
**/
int vmem_dbg_use(void)
{
    if((VMEM_PARAMS_PTR->magic_ca != 0xCA)
    || (VMEM_PARAMS_PTR->magic_fe != 0xFE))
    {
        return 0;
    }

    return (VMEM_PARAMS_PTR->dbg_use);
}



/**
 * Stack backup restore stuff, used in each BUP* functions.
 *
 * Standard BUP library uses around ~300 bytes on the stack, 
 * so it's not necessary (and may cause unexpected problems)
 * to backup stack start area.  Hence, doing backup from 200th
 * byte should be fine.
 *
 * Note : stack backup size (VMEM_STACK_BACKUP_SIZE) is #defined in vmem_main.h
**/
#define VMEM_STACK_BACKUP_OFFSET 200

#include <iapetus/iapetus.h>

/* Stack backup macro. */
#define vmem_stack_backup()  unsigned long stack_addr;                      \
                             asm("mov r15,%0": "=r"(stack_addr));           \
                             memcpy(VMEM_WORKRAM->stack_backup, ((unsigned char*)stack_addr)-VMEM_STACK_BACKUP_OFFSET-VMEM_STACK_BACKUP_SIZE, VMEM_STACK_BACKUP_SIZE);

/* Stack restore macro. */
#define vmem_stack_restore() memcpy(((unsigned char*)stack_addr)-VMEM_STACK_BACKUP_OFFSET-VMEM_STACK_BACKUP_SIZE, VMEM_WORKRAM->stack_backup, VMEM_STACK_BACKUP_SIZE);

/* Old macros used for debug and kept here in case of.
 * It seems that temporarily disabling interrupts causes more harm than fixing
 * things so that only stack backup/restore is left in "production" macros.
 */
#if 0
#define vmem_stack_backup()  unsigned long stack_addr;                      \
                             int old_level_mask;                            \
                             old_level_mask = interrupt_get_level_mask();   \
                             interrupt_set_level_mask(0xF);                 \
                             asm("mov r15,%0": "=r"(stack_addr));           \
                             memcpy(VMEM_WORKRAM->stack_backup, ((unsigned char*)stack_addr)-VMEM_STACK_BACKUP_OFFSET-VMEM_STACK_BACKUP_SIZE, VMEM_STACK_BACKUP_SIZE); \
                             scd_where_here("STT");

#define vmem_stack_restore() scd_where_here("END");                         \
                             memcpy(((unsigned char*)stack_addr)-VMEM_STACK_BACKUP_OFFSET-VMEM_STACK_BACKUP_SIZE, VMEM_WORKRAM->stack_backup, VMEM_STACK_BACKUP_SIZE); \
                             interrupt_set_level_mask(old_level_mask);
#endif


/**
 * Functions for use when game calls BUP* functions.
**/


/*
 *  Format          Sint32 BUP_SelPart(Uint32 device, Uint16 num)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                  num : partition number
 *                  0 - (number of partitions - 1)
 *  Output          none
 *  Function value  0 : success
 *                  other : failure
 *  Function        Selects a partition. In the initial condition, partition 0 is selected.
 */
Sint32 bup_selpart(Uint32 device, Uint16 num)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_selpart STT device=%d num=%d", device, num);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPSelectPartition(num);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPSelectPartition(num);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPSelectPartition(num);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPSelectPartition(num);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    scd_logout("[vmem]bup_selpart END ret=%d", ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Format(Uint32 device)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *  Output          none
 *  Function value  0 : success
 *                  BUP_WRITE_PROTECT : write protected
 *                  other : failure
 *  Function        Initializes the backup storage device.
 *                  Formats only the current partition in a partitioned backup storage device.
 */
Sint32 bup_format(Uint32 device)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_format STT device=%d", device);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPFormat();
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPFormat();
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPFormat();
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPFormat();
        break;

    default:
        ret = BUP_NON;
        break;
    }

    scd_logout("[vmem]bup_format END ret=%d", ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Stat(Uint32 device, Uint32 datasize, BupStat *stat)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                  datasize : specify size of data to be written in byte units
 *  Output          stat : status information
 *  Function value  returns device status
 *                  0 : success
 *                  BUP_NON : not connected
 *                  BUP_UNFORMAT : not formatted
 *  Function        Gets status information.
 */
Sint32 bup_stat(Uint32 device, Uint32 datasize, BupStat *tb)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_stat STT device=%d", device);

    /* Reset output parameter. */
    memset((void*)tb, 0, sizeof(BupStat));

    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPStatus(datasize, tb);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPStatus(datasize, tb);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPStatus(datasize, tb);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPStatus(datasize, tb);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    scd_logout("[vmem]bup_stat.totalsize  = %d", tb->totalsize);
    scd_logout("[vmem]bup_stat.totalblock = %d", tb->totalblock);
    scd_logout("[vmem]bup_stat.blocksize  = %d", tb->blocksize);
    scd_logout("[vmem]bup_stat.freesize   = %d", tb->freesize);
    scd_logout("[vmem]bup_stat.freeblock  = %d", tb->freeblock);
    scd_logout("[vmem]bup_stat.datanum    = %d", tb->datanum);

    scd_logout("[vmem]bup_stat END ret=%d", ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Write(Uint32 device, BupDir *dir, Uint8 *data, Uint8 owsw)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                  dir : file control information (input other than dir.blocksize)
 *                  data : pointer for write data
 *                  owsw: overwrite check mode
 *                      ON: does not write if file of same name exists
 *                      OFF: writes on file if file of same name exists
 *  Output          none
 *  Function value  0 : success
 *                  BUP_NON : not connected
 *                  BUP_UNFORMAT : not formatted
 *                  BUP_WRITE_PROTECT : write protect exists
 *                  BUP_FOUND : file of same name exists
 *                  Other : failure
 *  Function        Writes data to the backup storage device.
 */
Sint32 bup_write(Uint32 device,BupDir *tb,volatile Uint8 *data,Uint8 wmode)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_write STT, device=%d datasize=%d", device, tb->datasize);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPWrite(tb, data, wmode);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPWrite(tb, data, wmode);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPWrite(tb, data, wmode);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPWrite(tb, data, wmode);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    compat_print("[vmem]bup_write(device=%d, filename=\"%s\" datasize=%d) END ret=%d", 
        device, 
        bup_print_filename(tb->filename), 
        tb->datasize, 
        ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Read(Uint32 device, Uint8 *fname, Uint8* data)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                  fname : file name specification (11 characters are required in ASCII)
 *  Output          data : pointer for load destination buffer
 *  Function value  0 : success
 *                  BUP_NON : not connected
 *                  BUP_UNFORMAT : not formatted
 *                  BUP_NOT_FOUND : file not found
 *                  BUP_BROKEN : file is damaged
 *                  Other : failure
 *  Function        Loads data from the backup storage device.
 */
Sint32 bup_read(Uint32 device, Uint8 *filename, volatile Uint8 *data)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_read(\"%s\") STT, device=%d", 
        bup_print_filename(filename), 
        device);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPRead(filename, data);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPRead(filename, data);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPRead(filename, data);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPRead(filename, data);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    compat_print("[vmem]bup_read(device=%d, filename=\"%s\") END ret=%d", 
        device, 
        bup_print_filename(filename), 
        ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Delete(Uint32 device, Uint8 *fname)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                  fname : file name (11 characters is required in ASCII)
 *  Output          none
 *  Function value  0 : success
 *                  BUP_NON : not connected
 *                  BUP_UNFORMAT : not formatted
 *                  BUP_NOT_FOUND : file not found
 *                  BUP_WRITE_PROTECT : write protect exists
 *                  Other : failure
 *  Function        Deletes data on the backup storage device.
 */
Sint32 bup_delete(Uint32 device, Uint8 *filename)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_delete STT, device=%d", device);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPDelete(filename);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPDelete(filename);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPDelete(filename);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPDelete(filename);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    compat_print("[vmem]bup_delete(device=%d, filename=\"%s\") END ret=%d", 
        device, 
        bup_print_filename(filename), 
        ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Dir(Uint32 device, Uint8 *fname, Uint16 dirsize, BupDir *dir)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                  fname : file name specification (within 11 ASCII characters)
 *                  dirsize : specifies the number of directory info secured
 *  Output          dir : stores directory information
 *  Function value  number of directory information hits
 *  Function        The file name is done by searching forward, and the directory information
 *                  is stored in a table. The number of directories hit by the search
 *                  is returned as the function value. If negative, it indicates that the table
 *                  is too small, and the number of hits can be confirmed by inverting the
 *                  sign (if the result is 11 when the tbsize is set to 10 and a search is
 *                  performed, -11 is returned).
 */
Sint32 bup_dir(Uint32 device, Uint8 *filename, Uint16 tbsize, BupDir *tb)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_dir(\"%s\")(tbsize=%d) STT, device=%d", 
        bup_print_filename(filename), 
        tbsize, device);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPDirectory(filename, tbsize, tb);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPDirectory(filename, tbsize, tb);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPDirectory(filename, tbsize, tb, 0/*tbsize_bram*/);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPDirectory(filename, tbsize, tb);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    compat_print("[vmem]bup_dir(device=%d, filename=\"%s\", tbsize=%d) END ret=%d", 
        device, 
        bup_print_filename(filename), 
        tbsize, 
        ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          Sint32 BUP_Verify(Uint32 device, Uint8 *filename, Uint8* data)
 *  Input           device : device number
 *                      0: built-in memory
 *                      1: memory cartridge or parallel interface
 *                      2: serial interface
 *                      filename : file name (11 characters are required in ASCII)
 *                      data : pointer for data verification
 *  Output          none
 *  Function value  0 : success
 *                  BUP_NON : not connected
 *                  BUP_UNFORMAT : not formatted
 *                  BUP_NO_MATCH : data do not agree
 *                  BUP_NOT_FOUND : file not found
 *                  BUP_BROKEN : file is damaged
 *  Function        Verifies data written to the backup file.
 */
Sint32 bup_verify(Uint32 device, Uint8 *filename, volatile Uint8 *data)
{
    Sint32 ret;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_verify STT, device=%d", device);
    if(vmem_dbg_use()) { scd_logflush(); }

    switch(vmem_get_virtual_device(device))
    {
    case(VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD):
        ret = ExtBUPVerify(filename, data);
        break;

    case(VMEM_DEV_BRAM_INT):
        ret = BiosBUPVerify(filename, data);
        break;

    case(VMEM_DEV_MEMCART | VMEM_DEV_SDCARD):
        ret = CartBUPVerify(filename, data);
        break;

    case(VMEM_DEV_FLOPPY | VMEM_DEV_SDCARD):
        ret = FloppyBUPVerify(filename, data);
        break;

    default:
        ret = BUP_NON;
        break;
    }

    scd_logout("[vmem]bup_verify END ret=%d", ret);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return ret;
}


/*
 *  Format          void BUP_GetDate(Uint32 pdate,BupDate *date)
 *  Input           pdate : data and time data of directory information
 *                  date : date and time table
 *  Output          none
 *  Function value  none
 *  Function        Expands the date and time data in the directory information table.
 */
void bup_getdate(Uint32 date, BupDate *tb)
{
    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_getdate STT, date=%d", date);
    if(vmem_dbg_use()) { scd_logflush(); }

    unsigned long div;

    /* Set minute count. */
    tb->min = (unsigned char)(date % 60);

    /* Set hour. */
    tb->time = (unsigned char)((date % (60*24)) / 60);

    /* Compute days count. */
    div = date / (60*24);

    /* Set of week. */
    if (div > 0xAB71)
    {
        tb->week = (unsigned char)((div + 1) % 7);
    }
    else
    {
        tb->week = (unsigned char)((div + 2) % 7);
    }

    /* To process leap year, simply consider a pack of 4 years whose first one
     * is a multiple of 4, and loop up to 48 months until getting remaining
     * days count fitting into one month.
     *
     * It's a bit kind of naive (not to say dumb) implementation, but it works, 
     * and as it's not the kind of routine that require heavy optimization
     * there's no plan to make this algorithm smarter.
     */
    unsigned long year_base   = div / ((365*4) + 1); year_base = year_base * 4;
    unsigned long days_remain = div % ((365*4) + 1);
    const char days_count[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

    unsigned char month = 0;
    int i;
    for(i=0; i<(4*12); i++)
    {
        unsigned char days_per_month = days_count[i % 12];
        if(i == 1)
        {
            days_per_month++;
        }

        if(days_remain < days_per_month)
        {
            break;
        }

        days_remain -= days_per_month;
        month++;

        if((i % 12) == 11)
        {
            month = 0;
            year_base++;
        }
    }

    tb->month = month + 1;
    tb->day   = days_remain + 1;
    tb->year  = year_base;


    /* Code below is here to reproduce the same bug as in Saturn BIOS, 
     * so please disable it if you prefer a valid date rather than a
     * completely equivalent re-implementation.
     * 
     * Description : 1st January following a leap year is incorrectly
     *               returned as December 32nd of this leap year.
     * 
     * Examples :
     *  >Error2 #  1, BIO:1980/12/32 00:42 4, B:00080AEA
     *  >           , VMM:1981/01/01 00:42 4
     *  >           , ORG:1981/01/01 00:42
     *  >Error2 #  2, BIO:1980/12/32 07:42 4, B:00080C8E
     *  >           , VMM:1981/01/01 07:42 4
     *  >           , ORG:1981/01/01 07:42
     *  >Error2 #  5, BIO:1984/12/32 00:42 2, B:0028250A
     *  >           , VMM:1985/01/01 00:42 2
     *  >           , ORG:1985/01/01 00:42
     *
     * Note : it is preferable that the implementation here (vmem) should
     *        return a valid date rather than pushing mimic of BIOS behavior
     *        including its bugs.
     */
#if 0
    if((tb->month      == 1)
    && (tb->day        == 1)
    && ((tb->year % 4) == 1))
    {
        tb->month = 12;
        tb->day   = 32;
        tb->year--;
    }
#endif

    scd_logout("[vmem]bup_getdate END");
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();
}


/*
 *  Format          Uint32 BUP_SetDate(BupDate *date)
 *  Input           date : date and time table
 *  Output          none
 *  Function value  Data in date and time data form in the directory information table.
 *  Function        Compresses the date and time data in the directory information table.
 */
Uint32 bup_setdate(BupDate *tb)
{
    unsigned long date;

    /* Backup the stack. */
    vmem_stack_backup();

    scd_logout("[vmem]bup_setdate STT");
    if(vmem_dbg_use()) { scd_logflush(); }

    date = bup_setdate_internal(tb);

    scd_logout("[vmem]bup_setdate END date=0x%08X", date);
    scd_logout("[vmem_test][%08X %08X]", VMEM_WORKRAM->end, VMEM_LIBRAM->end);
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();

    return date;
}
/* Note : dev_cart.c converts date format, 
 * hence bup_setdate in two internal and external functions.
 */
Uint32 bup_setdate_internal(BupDate *tb)
{
    /* Table of elapsed days per month. */
    const unsigned short monthtbl[11] =
    {
        31, 
        31+28, 
        31+28+31, 
        31+28+31+30, 
        31+28+31+30+31, 
        31+28+31+30+31+30, 
        31+28+31+30+31+30+31, 
        31+28+31+30+31+30+31+31, 
        31+28+31+30+31+30+31+31+30, 
        31+28+31+30+31+30+31+31+30+31, 
        31+28+31+30+31+30+31+31+30+31+30
    };
    unsigned long date;
    unsigned char data;
    unsigned long remainder;


    /* Easter Egg when specified time is set to invalid null values
     * 
     * Technical details on SegaXtreme forums :
     *  https://segaxtreme.net/threads/backup-memory-structure.16803/
     *
     * antime   : If you take the 1980/0/0 value and go backwards from
     *            1980/1/1 you get 1963/10/8. I wonder if that's an
     *            easter egg - the programmer's birthday or something similar. 
     * TascoDLX : The value in question, 0x008246A0, corresponds to
     *            1996/3/26 0:00, which pretty much coincides with the
     *            passing of Comet Hyakutake.
     *            As far as easter eggs go, that'd be my guess. 
     * antime   : The comet wasn't discovered until January 1996.
     *            That'd be hell of an easter egg...
     * 
     * Note : in this re-implementation, the value before adding this
     *        easter egg was : t:0x026CECE0 -> 2057/05/15 00:00
     *        Probably something caused by random value read outside range, 
     *        or indicating the passing of another new comet ?!
     */
    if((tb->year  == 0)
    && (tb->month == 0)
    && (tb->day   == 0)
    && (tb->time  == 0)
    && (tb->min   == 0))
    {
        return 0x008246A0;
    }

    /* Add year to result. */
    data = tb->year;
    date = (data / 4) * 0x5B5; // 0x5B5 = 365.25 * 4
    remainder = data % 4;
    if(remainder)
    {
        date += (remainder * 0x16D) + 1; // 0x16D = 365
    }

    /* Leap year fix.
     * Code from Yabause HLE BIOS seems broken regarding leap years support.
     *
     * Rather than making a "clean" fix, and also because I'm not smart enough
     * to understand the "365.25 * 4" code above, date is adjusted in cases
     * causing problems.
     * 
     * Additionally, I don't want to spend days in optimizing that : writing
     * accurate date manipulation routines is some kind of art that may chew
     * a lot of free time that would be better being dedicated in more
     * interesting projects.
     * 
     * Fix is verified under test bench implemented in Save Data Manager's
     * extra menu which compares result between BIOS and this custom BUP_SetDate
     * functions, so that it should behave fine in any possible case until 2199.
     */
    unsigned short year = 1980 + tb->year;
    int leap_year;
    if((year % 4) != 0)
    {
        leap_year = 0;
    }
    else if((year % 100) != 0)
    {
        leap_year = 1;
    }
    else if((year % 400) != 0)
    {
        leap_year = 0;
    }
    else
    {
        leap_year = 1;
    }

    date--;
    if((leap_year) && (tb->month == 2))
    {
        date--;
    }
    if((year > 2000) && ((year % 100) == 0) && (tb->month == 2))
    {
        date--;
    }

    /* Add month to result. */
    data = tb->month;
    if(data != 1 && data < 13)
    {
        date += monthtbl[data - 2];
        if(date > 2 && remainder == 0)
        {
            date++;
        }
    }

    /* Add day to result. */
    date += tb->day;
    date *= 0x5A0;

    /* Add hour to result. */
    date += (tb->time * 0x3C);

    /* Add minute to result. */
    date += tb->min;

    return date;
}



/*
 * [vmem_asm.c] Functions to replace BIOS BUP_Init function.
 */
void vmem_bup_init_replacement(volatile Uint32 *lib, Uint32 *work, BupConfig tp[3]);


/*
 * Set UBC parameters to trigger at the start of BIOS's BUP_Init function.
 */
int vmem_rcf_start(vmem_params_t* set)
{
    int ret = 0;

    /* Copy parameters to RAM. */
    memcpy((void*)VMEM_PARAM_ADDRESS, (void*)set, sizeof(vmem_params_t));

    *((unsigned long*)0x06000358) = (unsigned long)vmem_bup_init_replacement;
    ret = (int)vmem_bup_init_replacement;

    return ret;
}







/* Pointer to BupConfig memory area. */
//#define BUP_CONF_PTR ((BupConfig**)0x26000FF0)
#define BUP_CONF_PTR ((BupConfig**)0x002FFFF0)

/*
 * Init backup memory, then init virtual memory cart.
 * Then, patch address to BUP* functions.
 */
#include "../release.h"
void bup_init(Uint32* lib_param, Uint32* work_param, BupConfig* tp_param)
{
    BupConfig* tp;
    vmem_workram_t* f;
    vmem_libram_t * l;

    /* Backup the stack. */
    vmem_stack_backup();

    /**
     * Use "entry point replacement" init mode :
     * Pointer to BupConfig structure is directly taken
     * from functions parameters, and pointers to work/lib RAM are set
     * to their respective places in RAM.
    **/
    tp = tp_param;
    * (Uint32 **)(0x6000354) = work_param/*8KB*/;
    **(Uint32***)(0x6000354) = lib_param/*16KB*/;

    f = VMEM_WORKRAM;
    l = VMEM_LIBRAM;

    /* Set dummy value at the end of both RAM areas, 
     * in order to verify eventual wrong structure size, 
     * buffer overflow, etc.
     */
    f->end = 0x0808EEEE;
    l->end = 0x1616EEEE;


    /**
     * Init SatCom.
     *
     * Usually, workram need to be declared, 
     * but as this code executes in ROM, 
     * SatCom uses RAM from backup memory library.
     * Hence, pointer to workram is NULL.
     *
     * Also, CDROM isn't initialized by SatCom library, 
     * because it may be used by host program.
    **/
    sc_initcfg_reset(&f->init_cfg, NULL/*sc_workram*/);
    f->init_cfg.flags = SC_INITFLAG_NOCDROM | SC_INITFLAG_NOSMPCTIME;
    f->init_cfg.logfile_name = "/VMEM.LOG";

    sc_init_cfg(&f->init_cfg);

    scd_logout("[vmem]sc_main_init END");
    if(vmem_dbg_use()) { scd_logflush(); }

    scd_logout("[vmem]Release ID : V%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);
    scd_logout("[vmem] VMEM_WORKRAM f = 0x%08X", f);
    scd_logout("[vmem] VMEM_LIBRAM  l = 0x%08X", l);
    scd_logout("[vmem]sizeof(vmem_libram_t )=%5u[%5u] (unused:%u)", 
        sizeof(vmem_libram_t) , 16*1024, 
        sizeof(l->unused));
    scd_logout("[vmem]sizeof(vmem_workram_t)=%5u[%5u] (unused_2048:%u, unused_end:%u)", 
        sizeof(vmem_workram_t),  8*1024, 
        sizeof(f->unused_2048), 
        sizeof(f->unused));
    scd_logout(" bup_conf_ptr=0x%08X [BupConfig*=0x%08X 0x%08X]", 
        (unsigned int)(BUP_CONF_PTR), 
        (unsigned int)(*BUP_CONF_PTR), (unsigned int)tp);

    /* Extra log output, in order to avoid to beg for full logs from beta tester. */
    compat_print("[vmem_debug] bup_conf_ptr=0x%08X [BupConfig*=0x%08X 0x%08X]", 
        (unsigned int)(BUP_CONF_PTR), 
        (unsigned int)(*BUP_CONF_PTR), (unsigned int)tp);

    if(vmem_dbg_use()) { scd_logflush(); }

    /* Setup device ID where to access memory cartridge files :
     *  - SD card by default
     *  - PC Link (via USB dev cart) when SD card couldn't be initialized
     *
     * In the case cartridge is equiped with both SD card and USB interfaces, 
     * it is required to remove SD card in order to access backup memory
     * files on PC.
     */
    f->dev_id = SC_FILE_SDCARD;

    sc_device_t sdcard_dev;
    scf_dev_status(&sdcard_dev, f->dev_id);

    if(!sdcard_dev.available)
    {
        f->dev_id = SC_FILE_PC;
    }



    /* Retrieve parameters set by Pseudo Saturn Kai in LRAM. */
    vmem_params_t* params_global = VMEM_PARAM_ADDRESS;
    if((params_global->magic_ca == 0xCA)
    && (params_global->magic_fe == 0xFE))
    {
        /* Valid parameters, copy them from LRAM. */
        memcpy((void*)VMEM_PARAMS_PTR, (void*)params_global, sizeof(vmem_params_t));
    }
    else
    {
        /* Parameters not found, use default parameters :
         * Device 1 : internal backup memory
         * Device 2 : cartridge on SD card
         * Device 3 : FDD on SD card
         */
        memset((void*)(VMEM_PARAMS_PTR), 0, sizeof(vmem_params_t));
        VMEM_PARAMS_PTR->dbg_use = 0;
        VMEM_PARAMS_PTR->dev_cfg[0] = VMEM_DEV_BRAM_INT;
        VMEM_PARAMS_PTR->dev_cfg[1] = VMEM_DEV_MEMCART | VMEM_DEV_SDCARD;
        VMEM_PARAMS_PTR->dev_cfg[2] = VMEM_DEV_FLOPPY  | VMEM_DEV_SDCARD;
    }


    /* Save RAM area reserved to BUP library. */
    /* (well, not really needed, because BIOS is supposed to have set it just before) */
    //f->ram_backuplib = lib;

    /* Overwrite backup library entry points. */
    f->bup_entry_points[BUP_ID_SELPART]   = (unsigned int)bup_selpart;
    f->bup_entry_points[BUP_ID_FORMAT]    = (unsigned int)bup_format;
    f->bup_entry_points[BUP_ID_STATUS]    = (unsigned int)bup_stat;
    f->bup_entry_points[BUP_ID_WRITE]     = (unsigned int)bup_write;
    f->bup_entry_points[BUP_ID_READ]      = (unsigned int)bup_read;
    f->bup_entry_points[BUP_ID_DELETE]    = (unsigned int)bup_delete;
    f->bup_entry_points[BUP_ID_DIRECTORY] = (unsigned int)bup_dir;
    f->bup_entry_points[BUP_ID_VERIFY]    = (unsigned int)bup_verify;
    f->bup_entry_points[BUP_ID_GETDATE]   = (unsigned int)bup_getdate;
    f->bup_entry_points[BUP_ID_SETDATE]   = (unsigned int)bup_setdate;
    f->bup_entry_points[10] = 0xDEADCAFE; //Dummy
scd_logout("[vmem]f->bup_entry_points[BUP_ID_WRITE] = 0x%08X", f->bup_entry_points[BUP_ID_WRITE]);

    unsigned char devs = VMEM_PARAMS_PTR->dev_cfg[0] | VMEM_PARAMS_PTR->dev_cfg[1] | VMEM_PARAMS_PTR->dev_cfg[2];

    if(devs & VMEM_DEV_SDCARD)
    {
        if(vmem_dbg_use()) { scd_logflush(); }

        if(!FileBUPInit())
        {
            /* Couldn't init SD card device ? */
            VMEM_PARAMS_PTR->dev_cfg[0] &= ~VMEM_DEV_SDCARD;
            VMEM_PARAMS_PTR->dev_cfg[1] &= ~VMEM_DEV_SDCARD;
            VMEM_PARAMS_PTR->dev_cfg[2] &= ~VMEM_DEV_SDCARD;
        }
        else
        {
            if(devs & VMEM_DEV_MEMCART)
            {
                CartBUPInit();
            }
            if(devs & VMEM_DEV_FLOPPY)
            {
                FloppyBUPInit();
            }
        }

        if(vmem_dbg_use()) { scd_logflush(); }
    }

    /* Mark first (Saturn internal) device as available.
     * Don't know if there is/are case when internal device
     * is unavailable ... should verify what happens after
     * a reset of the backup RAM.
     */
    tp[0].unit_id   = 1;
    tp[0].partition = 1;

    /* Verify second (cartridge/SD card) device availability. */
    if(VMEM_PARAMS_PTR->dev_cfg[1] & VMEM_DEV_SDCARD)
    {
        /* Init and mark SD card device as available. */
        tp[1].unit_id   = 2; /* Second backup device, displayed as "cartridge" on games. */
        tp[1].partition = 1; /* Won't support more than one partition. */
    }
    else
    {
        /* Couldn't init SD card device ?
         * That's bad, but I can't do anything for you.
         */
        tp[1].unit_id   = 0;
        tp[1].partition = 0;
    }


    /*
     * Third backup device : Saturn floppy device.
     * Unfortunately, vmem overwrites original BUP functions, so they can't be reused.
     * And, I won't add Floppy support from scratch in vmem, because it concerns
     * something like 3 users in the world, and would waste memory for other users.
     *
     * So instead, virtual FDD is emulated on SD card side, so that user won't need
     * to use real FDD in order to play the (few) games that use FDD.
     */
    if(VMEM_PARAMS_PTR->dev_cfg[2] & VMEM_DEV_SDCARD)
    {
        /* Show third backup device (displayed as "Floppy" on games) to program.
         * 
         * Unit ID is expected to be 3 for FDD, but for unknown reasons, 
         * it is set to zero when original BUP functions and official FDD is used.
         * So in order not to break compatibility, unit ID is set to same value
         * as if a real FDD would be used.
         *
         * And, partitions count is set to 2 partitions (720KB each).
         * If a game supports doesn't works when double partition FDD is used,
         * I may think about supporting single partition FDD, but that would make
         * Pseudo Saturn Kai user interface even more messy around FDD configuration.
         */
        tp[2].unit_id   = 0;
        tp[2].partition = 2;
    }
    else
    {
        tp[2].unit_id   = 0;
        tp[2].partition = 0;
    }


    /* Compatibility reporter stuff. */
    compat_print("[vmem]bup_init END (unit_id=%d %d %d, part=%d %d %d) (workram=0x%08X, libram=0x%08X)", 
        tp[0].unit_id, tp[1].unit_id, tp[2].unit_id, 
        tp[0].partition, tp[1].partition, tp[2].partition, 
        VMEM_WORKRAM, VMEM_LIBRAM);

    /* Set dummy value at the end of both RAM areas, 
     * in order to verify eventual wrong structure size, 
     * buffer overflow, etc.
     */
    f->end = 0x0808FEFE;
    l->end = 0x1616FEFE;

    /* For debug. */
    if(vmem_dbg_use()) { scd_logflush(); }

    /* Restore the stack. */
    vmem_stack_restore();
}


/*
 * "Replacement" init type.
 * Simply call vmem's BUP_Init function.
 */
void vmem_bup_init_replacement(volatile Uint32 *lib, Uint32 *work, BupConfig tp[3])
{
    bup_init((Uint32*)lib, (Uint32*)work, (BupConfig*)tp);
}


/* Function common for all devices. */
int bup_filename_is_greater(BupDir* tb1, BupDir* tb2)
{
    int ret = 0;
    int i;
    char c1, c2;

    for(i=0; i<11; i++)
    {
        c1 = sc_tolower(tb1->filename[i]);
        c2 = sc_tolower(tb2->filename[i]);

        if(c1 != c2)
        {
            ret = (c1 > c2 ? 1 : 0);
            break;
        }
    }

    return ret;
}

/* Function common for all devices. */
int bup_filename_comp(char* filename1, char* filename2)
{
    int i;
    for(i=0; i<11; i++)
    {
        if(*filename1 == '\0')
        {
            return 0;
        }
        if(*filename2 == '\0')
        {
            return 1;
        }
        if(*filename1 != *filename2)
        {
            return 2;
        }

        filename1++;
        filename2++;
    }

    return 0;
}

/* Function common for all devices. */
unsigned short bup_convert_blocksize(unsigned long device_blocksize, unsigned long size)
{
    /*
     * Note : this function is not 100% correct, but should work
     * enough in order to be compatible with a majority of games.
     */
    unsigned long ret;
    unsigned long tmp;

    /* First block : backup header, and few allocation clusters. */
    ret = 1;

    /* Last blocks : save data. */
    ret += size / device_blocksize;
    if((size % device_blocksize) != 0)
    {
        ret++;
    }

    /* Middle blocks : allocation table. */
    tmp = (ret * sizeof(unsigned short)) / device_blocksize;
    ret += tmp;

    /* First 4 bytes of each blocks are reserved for cluster header, 
     * so compute size it would cost in block unit.
     */
    tmp = (ret * 4) / device_blocksize;
    ret += tmp;

    /* In the hypotetical case block size is larger than 64K blocks, 
     * return maximum size.
     */
    if(ret >= 0xFFFF)
    {
        ret = 0xFFFF;
    }
    return ((unsigned short)ret);
}

/* Function common for all devices. */
char* bup_print_filename(unsigned char* filename)
{
    char* ret = VMEM_WORKRAM->string_format_buff;
    int i;

    for(i=0; i<11; i++)
    {
        char c = ((char*)filename)[i];
        if(c != '\0')
        {
            ret[i] = char2pchar(c);
        }
        else
        {
            ret[i] = c;
            break;
        }
    }
    ret[i] = '\0';

    return ret;
}


/**
 * bup_kana_to_ut8.
 *
 * Convert and append kana character code to UTF-8 encoding.
 *
 * Note : one kana character takes three bytes when converted to UTF-8.
**/
unsigned char _kana_utf8_tbl[8*8/*row * col*/ * 3/*bytes*/] = 
{
     ' ', '\0', '\0', 0xEF, 0xBD, 0xA1, 0xEF, 0xBD, 0xA2, 0xEF, 0xBD, 0xA3, 0xEF, 0xBD, 0xA4, 0xEF, 0xBD, 0xA5, 0xEF, 0xBD, 0xA6, 0xEF, 0xBD, 0xA7, //  
    0xEF, 0xBD, 0xA8, 0xEF, 0xBD, 0xA9, 0xEF, 0xBD, 0xAA, 0xEF, 0xBD, 0xAB, 0xEF, 0xBD, 0xAC, 0xEF, 0xBD, 0xAD, 0xEF, 0xBD, 0xAE, 0xEF, 0xBD, 0xAF, // 
    0xEF, 0xBD, 0xB0, 0xEF, 0xBD, 0xA7, 0xEF, 0xBD, 0xA8, 0xEF, 0xBD, 0xA9, 0xEF, 0xBD, 0xAA, 0xEF, 0xBD, 0xAB, 0xEF, 0xBD, 0xB6, 0xEF, 0xBD, 0xB7, // 
    0xEF, 0xBD, 0xB8, 0xEF, 0xBD, 0xB9, 0xEF, 0xBD, 0xBA, 0xEF, 0xBD, 0xBB, 0xEF, 0xBD, 0xBC, 0xEF, 0xBD, 0xBD, 0xEF, 0xBD, 0xBE, 0xEF, 0xBD, 0xBF, // 
    0xEF, 0xBE, 0x80, 0xEF, 0xBE, 0x81, 0xEF, 0xBE, 0x82, 0xEF, 0xBE, 0x83, 0xEF, 0xBE, 0x84, 0xEF, 0xBE, 0x85, 0xEF, 0xBE, 0x86, 0xEF, 0xBE, 0x87, // 
    0xEF, 0xBE, 0x88, 0xEF, 0xBE, 0x89, 0xEF, 0xBE, 0x8A, 0xEF, 0xBE, 0x8B, 0xEF, 0xBE, 0x8C, 0xEF, 0xBE, 0x8D, 0xEF, 0xBE, 0x8E, 0xEF, 0xBE, 0x8F, // 
    0xEF, 0xBE, 0x90, 0xEF, 0xBE, 0x91, 0xEF, 0xBE, 0x92, 0xEF, 0xBE, 0x93, 0xEF, 0xBE, 0x94, 0xEF, 0xBE, 0x95, 0xEF, 0xBE, 0x96, 0xEF, 0xBE, 0x97, // 
    0xEF, 0xBE, 0x98, 0xEF, 0xBE, 0x99, 0xEF, 0xBE, 0x9A, 0xEF, 0xBE, 0x9B, 0xEF, 0xBE, 0x9C, 0xEF, 0xBE, 0x9D, 0x22, '\0', '\0', 0xEF, 0xBE, 0x9F  // "
};

int bup_kana_to_ut8(char* str, unsigned char c)
{
    int ret;

    if(c == '\0')
    {
        /* Terminate string. */
        str[0] = '\0';
        ret = 1;
    }
    else if(c < ' ')
    {
        /* Reserved character, converted to space. */
        str[0] = ' ';
        ret = 1;
    }
    else if(c < 0x80)
    {
        /* Regular ASCII character. */
        str[0] = char2pchar(c);
        ret = 1;
    }
    else
    {
        if((c < (0x80+32)) || (c >= (0x80+96)))
        {
            /* Undefined extended character : let's
             * consider as a space character.
             */
            str[0] = ' ';
            ret = 1;
        }
        else
        {
            /* Kana converted from table. */
            memcpy(str, _kana_utf8_tbl + ((c - (0x80+32)) * 3), 3);
            if(str[1] == '\0')
            {
                /* "Kana" converted to ASCII code. */
                ret = 1;
            }
            else
            {
                ret = 3;
            }
        }
    }

    //str[ret] = '\0';

    return ret;
}


/* Function common for all devices. */
char* bup_print_comment(unsigned char* comment)
{
    char* ret = VMEM_WORKRAM->string_format_buff;
    int i;

    int pos = 0;
    int null_char = 0;
    for(i=0; i<10; i++)
    {
        unsigned char c = comment[i];

        if(c == '\0')
        {
            null_char = 1;
        }

        if(null_char)
        {
            ret[pos] = ' ';
            pos++;
        }
        else
        {
            pos += bup_kana_to_ut8(ret + pos, c);
        }
    }

    ret[pos] = '\0';

    return ret;
}

