/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _VMEM_MAIN_H_
#define _VMEM_MAIN_H_

/* ROM code and public structures definitions. */
#include "vmem_rom_exec.h"


/* Definitions for each backup device. */
#include "dev_internal.h"
#include "dev_cart.h"
#include "dev_floppy.h"
#include "dev_ext.h"

#include <sc_saturn.h>


/**
 * ROM code entry points.
**/
int vmem_rcf_version(vmem_version_t* ver);
int vmem_rcf_start(vmem_params_t* set);





// MEM[0x0007D600] : BUP_Init start address
//                 : rts at addresss 0x0007D628
// BUP_Init BackUpLibrary=0x06017e40 -> R12
// BUP_Init BackUpRamWork=0x06015e34 -> R11
// BUP_Init CnTb         =0x06017e34 -> R10
// MEM[0x06000354] : NULL on startup, contains pointer to BackUpRamWork after calling BUP_Init

/* Define address to BUP_init function. */
#define	BUP_LIB_ADDRESS		(*(volatile Uint32 *)(0x6000350+8))

/* Backup library entry points ID. */
#define BUP_ID_SELPART   0
#define BUP_ID_FORMAT    1
#define BUP_ID_STATUS    2
#define BUP_ID_WRITE     3
#define BUP_ID_READ      4
#define BUP_ID_DELETE    5
#define BUP_ID_DIRECTORY 6
#define BUP_ID_VERIFY    7
#define BUP_ID_GETDATE   8
#define BUP_ID_SETDATE   9



/**
 * Virtual memory cart parameters stored in work/lib ram areas.
 *
 * Memory usage related informations :
 *  unsigned long _bupLibrary[4096];
 *  unsigned long _bupRamWork[2048];
 *  BupConfig     _bupConf[3];
 *
 *  >_bupLibrary=0x06029008 (16KB, 4 bytes aligned)
 *  >_bupRamWork=0x0602d008 ( 8KB, 4 bytes aligned)
 *  >_bupConf   =0x06034dc8
 *  >BUP_Init       =0x0007d600
 *  >BUP_SelPart(+4)=0x060292cc
 *  >BUP_Format (+8)=0x060293b2
 *  >BUP_Stat   (+C)=0x0602967a
 *
 *   // Pointer to _bupRamWork (8KB)
 *   ptr1 =  *(void  *)(0x6000354);
 *   // Pointer to _bupLibrary (16KB)
 *   ptr2 = **(void **)(0x6000354);
 *
 *  MEM[0x06000354]:
 *   +00: 0x0602D008 <- pointer to _bupRamWork (8KB) (*1)
 *   +04: 0x0007D600 <- pointer to BUP_Init (BIOS function)
 *   +08: 0x0007D660
 *   +0C: 0x0600083C
 *   +10: 0x0600083C
 *   +14: 0x0600083C
 *   +18: 0x0600083C
 *   +1C: 0x0600083C
 *
 *  (*1) MEM[0x0602D008]: <- contents of _bupRamWork
 *   +00: 0x06029008 <- pointer to _bupLibrary (16KB)
 *   +04: 0x060292CC <- pointer to BUP_SelPart
 *   +08: 0x060293B2 <- pointer to BUP_Format
 *   +0C: 0x0602967A <- pointer to BUP_Stat
 *   +10: 0x06029A80
 *   +14: 0x0602A362
 *   +18: 0x0602A6E4
 *   +1C: 0x0602A8BC
 *   +20: 0x0602AAD4
 *   +24: 0x0602AE78
 *   +28: 0x0602AF6C
 *   +2C: 0x0602D080
 *
 *
 *
 *  // #define	BUP_VECTOR_ADDRESS	(*(volatile Uint32 *)(0x6000350+4))
 *  // #define	VMEM_WORKRAM ((vmem_workram_t*)(BUP_VECTOR_ADDRESS))
 *  vmem_workram_t* f = VMEM_WORKRAM;
 *
 *
 *  // f->ram_work :
 *  // Total 8KB, 4 bytes aligned.
 *  // First 2KB used to store bup library
 *  //  entry points, and few backup library data.
 *  // Remaining 6KB used to store SatCom internal data.
 *
 *  // f->ram_library :
 *  // Total 16KB, 4 bytes aligned.
 *  // First half (4KB) used for backup index data
 *  // Second half (12KB) used by SD card library.
**/


// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

/* Structure for Work RAM (8KB, lower 6KB used by SatCom) */
typedef struct _vmem_workram_t
{
    /* Pointer to library workram (length is 16KB). */
    unsigned long* ram_backuplib;

    /**
     * Backup library entry points
     * entries 0-9 are entry points defined in SEGA_BUP.h
     * #define	BUP_VECTOR_ADDRESS	(*(volatile Uint32 *)(0x6000350+4))
     * #define	VMEM_WORKRAM ((vmem_workram_t*)(BUP_VECTOR_ADDRESS))
     * #define	BUP_SelPart	((Sint32 (*)(Uint32 device,Uint16 num)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+4)))
     * #define	BUP_Format	((Sint32 (*)(Uint32 device)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+8)))
     * #define	BUP_Stat	((Sint32 (*)(Uint32 device,Uint32 datasize,BupStat *tb)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+12)))
     * #define	BUP_Write	((Sint32 (*)(Uint32 device,BupDir *tb,volatile Uint8 *data,Uint8 wmode)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+16)))
     * #define	BUP_Read	((Sint32 (*)(Uint32 device,Uint8 *filename,volatile Uint8 *data)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+20)))
     * #define	BUP_Delete	((Sint32 (*)(Uint32 device,Uint8 *filename)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+24)))
     * #define	BUP_Dir 	((Sint32 (*)(Uint32 device,Uint8 *filename,Uint16 tbsize,BupDir *tb)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+28)))
     * #define	BUP_Verify	((Sint32 (*)(Uint32 device,Uint8 *filename,volatile Uint8 *data)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+32)))
     * #define	BUP_GetDate	((void (*)(Uint32 date,BupDate *tb)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+36)))
     * #define	BUP_SetDate	((Uint32 (*)(BupDate *tb)) (*(Uint32 *)(BUP_VECTOR_ADDRESS+40)))
     * 
     * Entry 10 is undocumented ?
    **/
    unsigned long bup_entry_points[11];

    /* Backup of vmem parameters. */
    vmem_params_t params;

    /* Extended device global data. */
    vmem_extglobal_t ext_global;

    /* Temporary data used when sorting directory entries in list. */
    BupDir tb_temp;

    /* Temporary buffer used when formatting save filename/comment. */
    char string_format_buff[32];

    /* Backup RAM related buffers. */
    vmem_bram_buffer_t bram_buff;

    /* FDD saves list. */
    vmem_floppy_saves_t fdd_saves;

    /* SatCom library initialization parameters.
     * It is not required to keep after initializing, 
     * but stored here to reduce stack memory usage.
     */
    sc_initcfg_t init_cfg;

    /* Pack size of data declared so far to 2KB, and 1.5KB of data
     * after that which reserved for cheat codes.
     * This is used for cheat codes support of games where it's not possible
     * to setup cheat codes cache location elsewhere (ex: Dracula-X).
     *
     * In order to setup cheat codes cache location, it is required to
     * use workram address (as reported by vmem logs), plus 0x800 (2KB).
     * Example :
     *  >[vmem] VMEM_WORKRAM f = 0x060B1FC4
     *  >[vmem] VMEM_LIBRAM  l = 0x060ADFC0
     *  -> Cheat codes location is then B60B27C4
     * Memory allocated for cheat codes is 1.5KB, hence 172 codes can be stored :
     * (1536 - 0x40 - 0x60) / 8 = 172 codes
     * 172 / 4 (unit for capacity value) = 43 -> 0x002B
     *  -> Cheat codes location+capacity is then B60B27C4 002B
     */
#define VMEMCT_DATALEN      1536
#define VMEMCT_WORKRAM_OFFS 2048
    unsigned char unused_2048[VMEMCT_WORKRAM_OFFS
        - sizeof(unsigned long*)
        - 11*sizeof(unsigned long)
        - sizeof(vmem_params_t)
        - sizeof(vmem_extglobal_t)
        - sizeof(BupDir)
        - 32 * sizeof(char)
        - sizeof(vmem_bram_buffer_t)
        - sizeof(vmem_floppy_saves_t)
        - sizeof(sc_initcfg_t)];


    /* Reserved for cheat codes : access from vmem is prohibited. */
    unsigned char reserved_ct[VMEMCT_DATALEN];


    /**
     * Buffer used when listing backup files in PC/SD card.
     * Note #1 : The larger, the faster when listing files.
     * Note #2 : sizeof(scdf_file_entry_t) = 32 bytes.
     *
     * This is also used as work buffer in the following functions :
     *  - dev_cart > BUP_Verify, when reading save data from SD card.
     *  - dev_cart > BUP_Write , when saving text informations to SD card.
     *  - vmem_main > key functions, when compatibility reporting things.
     * 
     * So verifying the behavior of the functions above
     * (especially BUP_Write)
     * is required in the case you want to set a value
     * lower than 24 to FILE_ENTRIES_CNT.
     *
     * Note : the larger, the significantly better when listing saves
     *        (during BUP_Dir function), hence file entries count is
     *        set to 100.
     */
#define FILE_ENTRIES_CNT 100
    scdf_file_entry_t file_entries[FILE_ENTRIES_CNT];

    /* Backup area for stack.
     * Standard BUP library uses around ~300 bytes on the stack, 
     * and consequently using more memory may cause crash some games.
     * Example : Linkle Liver
     * As a workaround, stack is saved when entering each BUP entry
     * points, and restored just before exiting them.
     */
#define VMEM_STACK_BACKUP_SIZE 1024
    unsigned char stack_backup[VMEM_STACK_BACKUP_SIZE];

    /* Device ID for use by vmem :
     *  - SD card : normal use
     *  - PC Link : when plain USB dev cart is used
     */
    unsigned char dev_id;

    /* Reserved for future use (pack whole structure size to 8KB). */
    unsigned char unused[8*1024
        - VMEMCT_WORKRAM_OFFS
        - VMEMCT_DATALEN
        - VMEM_STACK_BACKUP_SIZE
        - FILE_ENTRIES_CNT*sizeof(scdf_file_entry_t)
        - sizeof(unsigned char)/*dev_id*/
        - sizeof(unsigned long/*end*/)];

    /* Dummy value, used to verify buffer is not overflowed. */
    unsigned long end;
} SC_ALIGN_4_BYTES(vmem_workram_t);

/* Structure for Library RAM (16KB) */
typedef struct _vmem_libram_t
{
    /* File index data, for use by virtual memory cart. */
    vmem_idxglobal_t idx_global;
    vmem_fileindex_t idx_data;

    /* Index init flag.
     * BUP_IDXINIT_START : initial state, set in CartBUPInit function.
     * BUP_IDXINIT_RESET : initial state reset request.
     * BUP_IDXINIT_DONE  : init OK.
     */
#define BUP_IDXINIT_START 0
#define BUP_IDXINIT_RESET 1
#define BUP_IDXINIT_DONE  2
    unsigned char idx_init;

    /* Reserved for future use (pack whole structure size to 16KB). */
    unsigned char unused[16*1024
        - sizeof(vmem_idxglobal_t)
        - sizeof(vmem_fileindex_t)
        - sizeof(unsigned char)
        - SC_WORKRAM_SIZE
        - sizeof(unsigned long/*end*/)];

    /** Last 7KB (= SC_WORKRAM_SIZE) are used by SatCom, so use here is prohibited. **/
    unsigned char sc_workram[SC_WORKRAM_SIZE];

    /* Dummy value, used to verify buffer is not overflowed. */
    unsigned long end;
} SC_ALIGN_4_BYTES(vmem_libram_t);


// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif

/* Macros to get pointer to each RAM areas. */
#define VMEM_WORKRAM ((vmem_workram_t*)( *(void  **)(0x6000354)))
#define VMEM_LIBRAM  ((vmem_libram_t *)(**(void ***)(0x6000354)))

/* Additional macro to access parameters structure. */
#define VMEM_PARAMS_PTR ((vmem_params_t*)(&(VMEM_WORKRAM)->params))


/*
 * Device used for file input/output.
 *  - Virtual Memory Cartridge.
 *  - Virtual FDD.
 *
 * It requires the following below defined :
 * vmem_workram_t* f = VMEM_WORKRAM;
 */
//#define VMEM_MEDIA_ID (SC_FILE_SDCARD)
#define VMEM_MEDIA_ID (f->dev_id)


/* BUP_setDate internal function, used in dev_cart.c. */
Uint32 bup_setdate_internal(BupDate *tb);

/**
 * Compare filename of both BupDir entries given in parameter.
 * (Function common for all devices.)
 *
 * Return 1 if tb1's name is greater than tb2's name, zero else.
**/
int bup_filename_is_greater(BupDir* tb1, BupDir* tb2);

/**
 * bup_filename_comp
 * Compare 2 saves filenames.
 * Both filenames don't need to be 11 chars long.
 * Return value :
 *   0 : filenames match
 *   else : filenames don't match
**/
int bup_filename_comp(char* filename1, char* filename2);

/**
 * Convert size in bytes to size in blocks.
 * (Function common for all devices.)
**/
unsigned short bup_convert_blocksize(unsigned long device_blocksize, unsigned long size);

/**
 * Functions to cleanup save filename/comment for safe print in buffer.
 * (Function common for all devices.)
 * Warning : uses the same shared buffer, so calling both
 *           at once may not work as expected.
**/
char* bup_print_filename(unsigned char* filename);
char* bup_print_comment(unsigned char* comment);


#endif /* _VMEM_MAIN_H_ */
