/*  Copyright (c) 2020 James Wah

    This file is in the public domain.
*/

#include "satiator.h"
#include <string.h>

#ifndef SC_NO_SATIATOR

// command enums {{{
typedef enum
{
    c_get_status = 0x90,
    c_write_buffer,
    c_read_buffer,

    c_mkfs = 0x94,
    c_info,
    c_settime,

    c_open = 0xA0,
    c_close,
    c_seek,
    c_read,
    c_write,
    c_truncate,
    c_stat,
    c_rename,
    c_unlink,
    c_mkdir,
    c_opendir,
    c_readdir,
    c_chdir,
    c_emulate,
} satiator_cmd_t;

typedef enum
{
    i_fw_version = 0,
    i_bootloader_version,
    i_serial_number,
    i_sd_latency,
} satiator_info_cmd_t;
// }}}


// }}}

// I/O primitives   {{{

typedef u16 cmd_t[4];

void exec_cmd(cmd_t cr, u16 wait)
{
    // mask all irqs
    uint32_t sr;
    asm("stc sr, %0" : "=r"(sr));
    uint32_t new_sr = sr;
    new_sr |= 0xf0;
    asm("ldc %0,sr": : "r" (new_sr));

    CDB_REG_HIRQ = ~(HIRQ_CMOK | wait);
    CDB_REG_CR1 = cr[0];
    CDB_REG_CR2 = cr[1];
    CDB_REG_CR3 = cr[2];
    CDB_REG_CR4 = cr[3];
    while (!(CDB_REG_HIRQ & HIRQ_CMOK));
    if(wait)
    {
        while (!(CDB_REG_HIRQ & wait));
    }

    // put interrupt level mask back
    asm("ldc %0,sr": : "r" (sr));
}


/* Note : de-inlining the function saves around 2552 bytes. */
/*static inline*/ int buffer_xfer(unsigned char* buf, int len, int dir)
{
    u8 cmdbyte = dir ? c_write_buffer : c_read_buffer;
    cmd_t cmd = {cmdbyte<<8, 0, 0, len};
    exec_cmd(cmd, HIRQ_EHST);
    if(CDB_REG_CR1)
    { // error
        return -1;
    }

    while (!(CDB_REG_HIRQ & HIRQ_DRDY));

    /* Compute the number of 32 bits transfers required.
     * If input length is not within 32 bits boundary, 
     * then data is transferred in excess.
     */
    u16 n = (len+3)/4;

    /* Copy internally and then transfer data.
     * This is required to properly handle pointers not
     * aligned within 32 bits boundary.
     *
     * Size of internal buffer can be freely modified but
     * as a too large size may explode the stack, setting
     * an intermediate size is a good tradeoff between
     * speed and memory usage.
     *
     * Bad example  : block count set to 512 = 1 pass but 2KB on the stack.
     * Good example : block count set to 4 = 128 pass and 16 bytes on the stack.
     */
#define BLOCK_COUNT 4
    u32 p[BLOCK_COUNT];
    u16 i, j, count, offs = 0;

    for(i=0; i<n; i+=BLOCK_COUNT)
    {
        if((i+BLOCK_COUNT) <= n)
        {
            count = BLOCK_COUNT;
        }
        else
        {
            count = n-i;
        }

        if(dir)
        { /* buffer_write = SH-2 -> SH-1 */

            /* Internally copy data outside of the buffer.
             * It's not a problem since security is not
             * really a concern on Saturn platform.
             */
            memcpy(p, buf+offs, sizeof(p));

            for(j=0; j<count; j++)
            {
                CDB_REG_DATATRNS = p[j];
            }
        }
        else
        { /* buffer_read = SH-2 <- SH-1 */

            for(j=0; j<count; j++)
            {
                p[j] = CDB_REG_DATATRNS;
            }

            /* In order to not overwrite the outside of
             * read buffer it's necessary to copy the
             * right amount of bytes, so that the extra
             * quick maths below.
             */
            int block_len;
            if((offs+sizeof(p)) <= len)
            {
                block_len = sizeof(p);
            }
            else
            {
                block_len = len-offs;
            }

            memcpy(buf+offs, p, block_len);
        }

        offs += sizeof(p);
    }

    if(dir)
    { /* buffer_write = SH-2 -> SH-1 */
        // mandatory but mysterious
        CDB_REG_DATATRNS = 0;
        CDB_REG_DATATRNS = 0;
    }

    return 0;
}

#define buffer_read(buf, len)  buffer_xfer((unsigned char*)buf, len, 0)
#define buffer_write(buf, len) buffer_xfer((unsigned char*)buf, len, 1)

// }}}

// Convenience functions {{{
// Most calls use a similar pattern
/* Note : de-inlining the function doesn't optimizes anything. */
static inline void set_cmd(cmd_t cmd, int op, int fd, int flags, int len)
{
    cmd[0] = (op<<8) | (fd & 0xff);
    cmd[1] = flags;
    cmd[2] = len>>16;
    cmd[3] = len;
}

/* Note : de-inlining the function saves around 1012 bytes. */
/*static inline*/ void get_stat(u16* sat_result)
{
    cmd_t cmd = {c_get_status<<8, 0, 0, 0};
    exec_cmd(cmd, 0);
    sat_result[0] = CDB_REG_CR1;
    sat_result[1] = CDB_REG_CR2;
    sat_result[2] = CDB_REG_CR3;
    sat_result[3] = CDB_REG_CR4;
}

// Most calls use the same sequence
// Most API returns a negative number for errors
#define simplecall(...)                     \
    u16 sat_result[4];                      \
    do {                                    \
        cmd_t cmd;                          \
        set_cmd(cmd, __VA_ARGS__);          \
        exec_cmd(cmd, HIRQ_MPED);           \
        get_stat(sat_result);               \
        u8 __retval = sat_result[0] >> 8;   \
        if(__retval)                        \
        {                                   \
            return -__retval;               \
        }                                   \
    } while(0);

// }}}

#define get_length() ((sat_result[2]<<16) | sat_result[3])

// File API {{{
// Given a filename and some FA_xxx flags, return a file descriptor
// (or a negative error corresponding to FR_xxx)
int s_open(const char *filename, int flags)
{
    buffer_write(filename, strlen(filename));
    simplecall(c_open, 0, flags, strlen(filename));

    return sat_result[3];   // handle
}

// Close a file descriptor
int s_close(int fd)
{
    simplecall(c_close, fd, 0, 0);

    return 0;
}

// Seek to a byte on an fd. Returns the offset
int s_seek(int fd, int offset, int whence)
{
    simplecall(c_seek, fd, whence, offset);

    return (sat_result[2]<<16) | sat_result[3];
}

// Read some data. Returns bytes read
int s_read(int fd, void *buf, int len)
{
    if(len > S_MAXBUF || len < 0)
    {
        return -FR_INVALID_PARAMETER;
    }

    simplecall(c_read, fd, 0, len);

    len = get_length();
    buffer_read(buf, len);

    return len;
}

// Write some data. Returns bytes written
int s_write(int fd, const void *buf, int len)
{
    if(len > S_MAXBUF || len < 0)
    {
        return -FR_INVALID_PARAMETER;
    }
    buffer_write(buf, len);
    simplecall(c_write, fd, 0, len);

    return get_length();
}

// Flush any buffered data to file
int s_sync(int fd)
{
    return s_seek(fd, 0, C_SEEK_CUR);
}

// Truncate file at current pointer. Returns new length
int s_truncate(int fd)
{
    simplecall(c_truncate, fd, 0, 0);

    return get_length();
}

// Get info on a named file. Pass in a pointer to a buffer and
// its size - the filename may be truncated if the buffer is
// short.  Returns the length of the (truncated) filename.
// If the filename is NULL, reads the next file from the
// current directory (readdir).
int s_stat(const char *filename, s_stat_t *stat, int statsize)
{
    int len;
    if(statsize < 9)
    {
        return -FR_INVALID_PARAMETER;
    }

    if(filename)
    {
        len = strlen(filename);
        buffer_write(filename, len);
    }
    else
    {
        len = 0;
    }

    simplecall(filename ? c_stat : c_readdir, 0, 0, len);

    len = get_length();
    if(len > statsize)
    {
        len = statsize;
    }

    buffer_read(stat, len);

    return len - 9;
}

// Rename a file.
int s_rename(const char *old, const char *new)
{
    // Need both names, zero separated, for the write
    char namebuf[512];  // nasty huh
    int len1 = strlen(old);
    int len2 = strlen(new);

    memcpy(namebuf, old, len1+1);
    memcpy(namebuf+len1+1, new, len2);

    buffer_write(namebuf, len1+1+len2);
    simplecall(c_rename, 0, 0, len1+1+len2);

    return 0;
}

// Create a directory.
int s_mkdir(const char *filename)
{
    int len = strlen(filename);

    buffer_write(filename, len);
    simplecall(c_mkdir, 0, 0, len);

    return 0;
}

// Delete a file.
int s_unlink(const char *filename)
{
    int len = strlen(filename);

    buffer_write(filename, len);
    simplecall(c_unlink, 0, 0, len);

    return 0;
}

// Open a directory to read file entries.
int s_opendir(const char *filename)
{
    int len = strlen(filename);

    buffer_write(filename, len);
    simplecall(c_opendir, 0, 0, len);

    return 0;
}

// Change working directory.
int s_chdir(const char *filename)
{
    int len = strlen(filename);

    buffer_write(filename, len);
    simplecall(c_chdir, 0, 0, len);

    return 0;
}

// Get working directory. Adds terminating null.
int s_getcwd(char *filename, int buflen)
{
    buffer_write(".", 1);
    simplecall(c_chdir, 0, 0, 1);
    buflen--;

    if(buflen > get_length())
    {
        buflen = get_length();
    }

    buffer_read(filename, buflen);
    filename[buflen] = 0;

    return buflen;
}

// Set the Satiator RTC. Takes a FAT timestamp.
int s_settime(uint32_t time)
{
    simplecall(c_settime, 0, 0, time);

    return 0;
}
// }}}

// System API {{{
int is_satiator_present(void)
{
    // This checks the MPEG version field.
    // Real MPEG cards have version 1.
    // The Satiator has version 2.
    cmd_t cmd = {0x0100, 0, 0, 0};
    exec_cmd(cmd, 0);

    return (CDB_REG_CR3 & 0xff) == 2;
}

//static enum satiator_mode cur_mode = s_unknown;
static enum satiator_mode cur_mode = s_cdrom;

int s_mode(enum satiator_mode mode)
{
    /* Switch between emulating a CD drive and exposing the SD card API.
     * This function returns:
     *      1: already set to specified mode
     *      0: success
     *      -1: Satiator not detected
     */

    //if(cur_mode == s_unknown)
    //{
    //    cur_mode = s_api;
    //    s_mode(s_cdrom);
    //}

    if(mode == cur_mode)
    {
        return 1;
    }

    if(mode == s_cdrom)
    {
        cmd_t cmd = {0x9300, 1, 0, 0};
        exec_cmd(cmd, HIRQ_MPED);
    }
    else
    {
        cmd_t cmd = {0xe000, 0x0000, 0x00c1, 0x05e7};
        exec_cmd(cmd, HIRQ_EFLS);

        // is there actually a Satiator attached?
        if(!is_satiator_present())
        {
            return -1;
        }

        // stop CD "drive"
        // cmd_t cmd_stop = {0x0400, 0x0001, 0x0000, 0x040f};
        // exec_cmd(cmd_stop, 0);
        cd_stop_drive();
    }
    cur_mode = mode;
    return 0;
}

// Given the filename of a disc descriptor, load it into the emulated drive.
int s_emulate(const char *filename)
{
    int len = strlen(filename);

    buffer_write(filename, len);
    simplecall(c_emulate, 0, 0, len);

    return 0;
}

int s_get_fw_version(char *buf, int buflen)
{
    simplecall(c_info, i_fw_version, 0, 0);

    buflen--;
    if(buflen > get_length())
    {
        buflen = get_length();
    }

    buffer_read(buf, buflen);
    buf[buflen] = 0;

    return buflen;
}

int s_get_bootloader_version(uint32_t *version)
{
    simplecall(c_info, i_bootloader_version, 0, 0);
    buffer_read(version, 4);

    return 0;
}

int s_get_serial_number(uint32_t *serial)
{
    simplecall(c_info, i_serial_number, 0, 0);
    buffer_read(serial, 4);

    return 0;
}

int s_get_sd_latency(uint16_t latency_us[], int *errors, int count)
{
    simplecall(c_info, i_sd_latency, count, 0);
    buffer_read(latency_us, 2*count);

    *errors = sat_result[1];

    return 0;
}

// int s_format_sd_card(int flags)
// {
//     simplecall(c_mkfs, flags, 0xfeed, 0xdeadbeef);
// 
//     return 0;
// }

// relaunch the satiator menu
int s_reset_to_satiator(void)
{
    s_mode(s_api);
    //int (**bios_get_mpeg_rom)(uint32_t index, uint32_t size, uint32_t addr) = (void*)0x06000298;
    int ret = (*bios_get_mpeg_rom)(2, 2, 0x200000);

    if(ret < 0)
    {
        return ret;
    }

    // mask all irqs
    uint32_t sr;
    asm("stc sr, %0" : "=r"(sr));
    uint32_t new_sr = sr;
    new_sr |= 0xf0;
    asm("ldc %0,sr": : "r" (new_sr));

    void (*entry)(uint32_t) = (void*)0x200000;
    entry(S_BOOT_NO_AUTOLOAD);

    asm("ldc %0,sr": : "r" (sr));
    return -0x1000;
}

// }}}

// // Cartridge API {{{
// satiator_cart_header_t *s_find_cartridge(void)
// {
//     uint8_t *ptr = (void*)0x02000000;
//     uint8_t *end = ptr + 1048576;
// 
//     while (ptr < end)
//     {
//         if(!memcmp(ptr, "SatiatorCart", 12))
//         {
//             return (void*)ptr;
//         }
// 
//         ptr += 0x100;
//     }
// 
//     return NULL;
// }
// // }}}

#endif // ndef SC_NO_SATIATOR

