/*  Copyright 2020 James Wah

    This file is in the public domain.
*/

#ifndef _SATIATOR_H
#define _SATIATOR_H

#include <stdio.h>
#include <string.h>

#include "iapetus/iapetus.h"

#define C_SEEK_SET  0
#define C_SEEK_CUR  1
#define C_SEEK_END  2

// CD image descriptor
typedef struct {
    uint32_t start;             // FAD
    uint32_t length;            // sectors

    uint32_t file_offset;       // byte offset within file where segment data starts
    uint32_t filename_offset;   // byte offset within descfile where filename is. zero for no data

    uint16_t flags;             // future proofing
    uint16_t secsize;           // 2048 or 2352

    uint8_t track;              // 1-99 for disc tracks.
    uint8_t index;              // typically 0 for pregap, 1 for track, others not used. track MSF counts backwards in pregap
    uint8_t q_mode;             // typically 0x41 (data) or 0x01 (audio)

} __attribute__((packed)) seg_desc_t;

// Satiator cart descriptor header, version 1
// This is written somewhere on a top slot Flash cart (eg. Action Replay) on a 0x100 byte boundary
// All offsets are relative to the start of this header
typedef struct {
    uint8_t signature[12];  // ASCII: "SatiatorCart"
    int32_t header_version; // 1 for this struct
    uint8_t version_str[32];// ASCII

    // Jumping to this address should be equivalent to booting the cart directly
    void (*bootcode)(void);

    void (*install_soft_reset_hook)(void);
} __attribute__((packed)) satiator_cart_header_t;


// XXX fatfs bits {{{
/* File access control and file status flags (FIL.flag) */

#define	FA_READ				0x01
#define	FA_OPEN_EXISTING	0x00

#define	FA_WRITE			0x02
#define	FA_CREATE_NEW		0x04
#define	FA_CREATE_ALWAYS	0x08
#define	FA_OPEN_ALWAYS		0x10

/* File attribute bits for directory entry */

#define AM_RDO  0x01    /* Read only */
#define AM_HID  0x02    /* Hidden */
#define AM_SYS  0x04    /* System */
#define AM_VOL  0x08    /* Volume label */
#define AM_LFN  0x0F    /* LFN entry */
#define AM_DIR  0x10    /* Directory */
#define AM_ARC  0x20    /* Archive */
#define AM_MASK 0x3F    /* Mask of defined bits */

typedef enum {
	FR_OK = 0,				/* (0) Succeeded */
	FR_DISK_ERR,			/* (1) A hard error occurred in the low level disk I/O layer */
	FR_INT_ERR,				/* (2) Assertion failed */
	FR_NOT_READY,			/* (3) The physical drive cannot work */
	FR_NO_FILE,				/* (4) Could not find the file */
	FR_NO_PATH,				/* (5) Could not find the path */
	FR_INVALID_NAME,		/* (6) The path name format is invalid */
	FR_DENIED,				/* (7) Access denied due to prohibited access or directory full */
	FR_EXIST,				/* (8) Access denied due to prohibited access */
	FR_INVALID_OBJECT,		/* (9) The file/directory object is invalid */
	FR_WRITE_PROTECTED,		/* (10) The physical drive is write protected */
	FR_INVALID_DRIVE,		/* (11) The logical drive number is invalid */
	FR_NOT_ENABLED,			/* (12) The volume has no work area */
	FR_NO_FILESYSTEM,		/* (13) There is no valid FAT volume */
	FR_MKFS_ABORTED,		/* (14) The f_mkfs() aborted due to any parameter error */
	FR_TIMEOUT,				/* (15) Could not get a grant to access the volume within defined period */
	FR_LOCKED,				/* (16) The operation is rejected according to the file sharing policy */
	FR_NOT_ENOUGH_CORE,		/* (17) LFN working buffer could not be allocated */
	FR_TOO_MANY_OPEN_FILES,	/* (18) Number of open files > _FS_SHARE */
	FR_INVALID_PARAMETER	/* (19) Given parameter is invalid */
} FRESULT;
// }}}

// API
typedef struct {
    uint32_t size;
    uint16_t date, time;
    uint8_t attrib;
    char name[];
} __attribute__((packed)) s_stat_t;

enum satiator_mode {
    s_cdrom = 0,
    s_api,
    s_unknown = 0xff,
};

typedef u16 cmd_t[4];

void exec_cmd(cmd_t cr, u16 wait);

int s_open(const char *filename, int flags);
int s_close(int fd);
int s_seek(int fd, int offset, int whence);
int s_read(int fd, void *buf, int len);
int s_write(int fd, const void *buf, int len);
int s_sync(int fd);
int s_truncate(int fd);
int s_stat(const char *filename, s_stat_t *stat, int statsize);
int s_rename(const char *old, const char *new);
int s_mkdir(const char *filename);
int s_unlink(const char *filename);
int s_opendir(const char *filename);
int s_chdir(const char *filename);
int s_getcwd(char *filename, int buflen);
int s_settime(uint32_t fattime);
int s_emulate(const char *filename);
int is_satiator_present(void);
int s_mode(enum satiator_mode mode);
int s_get_fw_version(char *buf, int buflen);
int s_get_bootloader_version(uint32_t *version);
int s_get_serial_number(uint32_t *serial);
int s_get_sd_latency(uint16_t latency_us[], int *errors, int count);
// int s_format_sd_card(int flags);
int s_reset_to_satiator(void);

// // See if there is a cartridge in the top slot providing Satiator services
// satiator_cart_header_t *s_find_cartridge(void);

#define S_MAXBUF    2048

#define S_BOOT_NO_AUTOLOAD  0x2f9862ac

#endif // _SATIATOR_H
