/**
 *  SatCom Library
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/satcom/
 *
 *  See LICENSE file for details.
**/

#ifndef SATCOM_COMMON_H
#define SATCOM_COMMON_H

//----------------------------------------------------------------------
// Endian conversion macros.
// 
//----------------------------------------------------------------------
#ifndef BSWAP16
#   define BSWAP16(x) (((x) >> 8) | ((x) << 8))
#endif
#ifndef BSWAP32
#   define BSWAP32(x) (((x) >> 24) | (((x) & 0x00FF0000) >> 8) | (((x) & 0x0000FF00) << 8) | ((x) << 24))
#endif


//----------------------------------------------------------------------
// Maximum/Minimum macros.
// 
//----------------------------------------------------------------------
#ifndef MAX
#   define MAX( x, y ) ((x) > (y) ? (x) : (y))
#endif
#ifndef MIN
#   define MIN( x, y ) ((x) < (y) ? (x) : (y))
#endif


//----------------------------------------------------------------------
// Structure alignment macros.
// 
// For use when defining strutures, example :
//     typedef struct _sc_example_struct_t
//     {
//         unsigned short data_u16;
//         unsigned char  data_u8;
//         unsigned char  padding;
//         unsigned long  data_u32;
//     } SC_ALIGN_4_BYTES(sc_example_struct_t);
//----------------------------------------------------------------------
#ifdef _BCB
#define SC_ALIGN_4_BYTES(_STRUCT_NAME_) _STRUCT_NAME_
#define SC_ALIGN_2_BYTES(_STRUCT_NAME_) _STRUCT_NAME_
#else // !_BCB
#define SC_ALIGN_4_BYTES(_STRUCT_NAME_) _STRUCT_NAME_ __attribute__((aligned(4)))
#define SC_ALIGN_2_BYTES(_STRUCT_NAME_) _STRUCT_NAME_ __attribute__((aligned(2)))
#endif // !_BCB


#ifdef _BCB
#ifdef  __cplusplus
extern "C" {
#endif
#endif // _BCB


//----------------------------------------------------------------------
// Char manipulation functions.
// Used for filename sorting (scf_filelist function) in sc_saturn.c.
// 
//----------------------------------------------------------------------
char sc_toupper(char c);
char sc_tolower(char c);


//----------------------------------------------------------------------
// Convert char to printable char.
// 
//----------------------------------------------------------------------
unsigned char char2pchar(unsigned char c);


//----------------------------------------------------------------------
// Saturn debugger common stuff.
// 
//----------------------------------------------------------------------
#include "debugger/scd_common.h"


//----------------------------------------------------------------------
// CRC32, used by RomFs tool.
// 
//----------------------------------------------------------------------
unsigned long crc32_update(unsigned long crc, const void *buf, unsigned long size);
#define CRC32_INITVAL 0
#define crc32_calc(_BUFF_, _SIZE_) crc32_update(CRC32_INITVAL, _BUFF_, _SIZE_)

//----------------------------------------------------------------------
// 8 bits CRC, used by USB dev cart
// http://www.iki.fi/Anders.Montonen/sega/usbcart/
// 
//----------------------------------------------------------------------
unsigned char crc_usbdc_init(void);
unsigned char crc_usbdc_update(unsigned char crc, const unsigned char *data, unsigned long data_len);
unsigned char crc_usbdc_finalize(unsigned char crc);




//----------------------------------------------------------------------
// Compression routines.
// 
//----------------------------------------------------------------------
#include "sc_compress.h"






//----------------------------------------------------------------------
// Flash memory main executable program data arrangement.
// Used in mkfirm and Pseudo Saturn Kai firmware.
// 
//----------------------------------------------------------------------

/* Flash memory start address. */
#define FC_FLASH_ADDRESS 0x02000000

/* About firmware data:
 *   CONTENTS    ~~ SIZE     ~~ START OFFSET
 *---------------------------------------------
 *  IP.BIN       ~~ 3840  B  ~~ 0x0000 (    0)
 *  loader       ~~ 112   B  ~~ 0x0F00 ( 3840)
 *  parameters   ~~ 16    B  ~~ 0x0F70 ( 3952)
 *  decompressor ~~ 3840  B  ~~ 0x0F80 ( 3968)
 *  comp. data   ~~ Varies   ~~ 0x1E80 ( 7808)
 */


/* IP.BIN start offset. */
#define FC_IPBIN_OFFSET        0
/* Loader start offset. */
#define FC_LOADER_OFFSET       (FC_IPBIN_OFFSET + 0xF00)
/* Parameters start offset. (Max 4 parameters) */
#define FC_PARAMS_OFFSET       (FC_LOADER_OFFSET + 112)
/* Decompressor start offset and length. */
#define FC_DECOMPRESSOR_OFFSET (FC_PARAMS_OFFSET + 16)
#define FC_DECOMPRESSOR_LENGTH (3840)
/* Decompressor execute address */
#define FC_DECOMPRESSOR_EXEC   0x002F0000
/* Compressed data start offset. */
#define FC_DATA_OFFSET         (FC_DECOMPRESSOR_OFFSET + FC_DECOMPRESSOR_LENGTH)

/* Parameter 1: input data length. */
#define FC_PDATALEN_INDEX    0
/* Parameter 2: execution address. */
#define FC_EXEADDRS_INDEX    1

/* Parameter 3: compressed data length (output parameter). */
#define FC_COMPDATALEN_INDEX 2
/* Parameter 4: compressed data CRC (output parameter). */
#define FC_COMPDATACRC_INDEX 3


#ifdef _BCB
#ifdef  __cplusplus
}
#endif
#endif // _BCB


#endif //SATCOM_COMMON_H
