/**
 *  SatCom Library
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/satcom/
 *
 *  See LICENSE file for details.
**/

#include <string.h> /* For memset,memcpy */
#include <stdarg.h>
#include <stdio.h>
size_t strnlen(const char *s, size_t maxlen);
int rand (void);
void srand (unsigned int seed);



/**
 * SatCom-related include.
**/
#include "sc_saturn.h"



/**
 * Satregs definitions, for internal purpose.
**/
#include "satregs/satregs.h"



/*----------------------------------------------------------------------------*/
/*- Libyaul stuff.                                                          --*/
#ifdef SC_USE_LIBYAUL

/* Should do libyaul includes someday ... */
// #include "libyaul\kernel\lib\memb.h"
// #include "libyaul\kernel\mm\slob.h"
// #include "libyaul\kernel\sys\init.h"

// //#include "libyaul\kernel\vfs\component_name.h"
// //#include "libyaul\kernel\vfs\hash.h"
// //#include "libyaul\kernel\vfs\mount.h"
// //#include "libyaul\kernel\vfs\pathname_cache.h"
// //#include "libyaul\kernel\vfs\vfs.h"
// //#include "libyaul\kernel\vfs\vnode.h"
// //#include "libyaul\kernel\vfs\fs\iso9660\iso9660-internal.h"
// //#include "libyaul\kernel\vfs\fs\iso9660\iso9660.h"
// //#include "libyaul\kernel\vfs\fs\romdisk\romdisk.h"

#endif // SC_USE_LIBYAUL


/**
 * Debugger global data.
**/
#include "debugger/scd_circbuffer.inc.c"

/**
 * SatCom Global data definition.
 * Used for/by debugger
**/
#include "sc_global_data.h"


/* Set to 1 to enable logging when accessing to Satiator. */
#define SATIS_LOG 1



#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif


/**
 * Global data that need to be stored in RAM.
**/
typedef struct _sc_internal_t
{
    /*-----------------------------------------*/
    /* Pointer to global data. */
    sc_global_t* global;

    /* Pointer to debugger data. */
    scd_data_t* dbg;

    /*-----------------------------------------*/
    /* Link type/etc related settings (saved in backup memory). */
    sc_settings_t set;

    /*-----------------------------------------*/
    /* ARP's IO registers: flag, R, W */
    unsigned long address[3];

    /*-----------------------------------------*/
    /* Set to 1 when comms link was first used. */
    unsigned char use_comms;

    /*-----------------------------------------*/
    /* Set to 1 when SatCom has been intialized. */
    unsigned char init_ok;

    /* Indicates if first log to output to SD card or not. */
    unsigned char already_log_sd;

    /* Set to 1 when SD card access is in read-only access. */
    unsigned char sd_read_only;

    /* Set to 1 when use of SMPC time is allowed. */
    unsigned char smpctime_use;

    /* CDROM use and init status. */
#define SC_CDSTATUS_USE  0x01
#define SC_CDSTATUS_INIT 0x02
    unsigned char cdrom_status;

    /* Set to 1 when custom yabause is used. */
    unsigned char yabause_use;

    /*-----------------------------------------*/
    /* Satiator related internals. */
#define SATIS_INIT_WAIT 0x80
#define SATIS_SDCARD_OK 0x01
#define SATIS_CDROM_OK  0x02
    unsigned char satis_flags;
    /* Unused. */
    unsigned char padding[3];

    /*-----------------------------------------*/
    /* Extension devices (SC_EXT_* value). */
    unsigned char ext_dev;

    /* Pointer used when sending data to execute. */
    unsigned char* exec_buffer;

    /*-----------------------------------------*/
    /* Callback function for soft reset is required. */
    Fct_sc_soft_reset soft_reset;

    /*-----------------------------------------*/
    /* Callback in order to show SatCom logs on screen. */
    Fct_scd_screen dbg_logscreen;

    /*-----------------------------------------*/
    /* Callback in order to show link related logs on screen. */
    Fct_scd_screen dbg_linkdiag;

    /*-----------------------------------------*/
    /* Used for SatCom debugging (internal use only !) */
#define SCD_PRINT_LEN 256
    char dbg_print[SCD_PRINT_LEN];

    /* Log file name when logging to SD card. */
    char* logfile_name;

    /*-----------------------------------------*/
    /* File I/O stuff. */
    char filename_buff[SC_PATHNAME_MAXLEN];

#define SCDF_BUFFSIZE 128
    char scd_file_request_buff[SCDF_BUFFSIZE];

    /* Pointer to RomFs data. */
    unsigned char* romfs;

    /* SD card-related buffer. */
    sdcard_t sd;

    /*-----------------------------------------*/
    /* Reserved for future usage and 4 bytes alignment.           */
    /* Keep structure total size to SC_INTERNAL_SIZE bytes (4KB). */
    unsigned char reserved9[SC_INTERNAL_SIZE
            - sizeof(sc_settings_t)
            - SCD_PRINT_LEN
            - SC_PATHNAME_MAXLEN
            - SCDF_BUFFSIZE
            - sizeof(sdcard_t)
            - 56/* Total size of misc. declarations */];
} sc_internal_t;

#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif



/**
 * Pointer to SatCom internal data.
**/
#ifdef FBA_BUILD
extern sc_internal_t* _sc;
#else
sc_internal_t* _sc = NULL;
#endif

sc_internal_t* sc_get_internal(void)
{
#ifdef SC_EXEC_IN_ROM
    /**
     * Use memory for backup library.
     * Size allocated to backup library libram is 16KB, 
     * and satcom related stuff is put at the end of this area.
    **/
    unsigned long* ptr = **(unsigned long ***)(0x6000354);
    ptr += (16*1024) / sizeof(unsigned long);
    ptr -= 1/*unsigned long end*/ + SC_WORKRAM_U32SIZE;

    return (sc_internal_t*)ptr;
#else
    /* Return pointer to user area memory. */
    return _sc;
#endif
}


/**
 * Set debugger absolute data at the end of LRAM.
 * This is set at the initialization of SatCom library, but may be
 * overwritten by external module. Typical example :
 *  - Usage by vmem
 *  AND
 *  - Usage by another SatCom application, such as Save Data Manager
 *
 * Proper setting of abosolute data is mandatory when flushing logs, or
 * accessing file via PC link (which uses log flush function), because
 * incorrect setting would make SatLink polling wrong SatCom debugger context.
**/
void sc_set_absolute_data(sc_internal_t* sc)
{
    /* Set debugger absolute data. */
    sc->global->scd_magic = SCD_ABS_MAGIC;
    sc->global->scd_ptr   = (unsigned long)(sc->dbg);
}









#define CARTRAM_ID_8M       0x5a
#define CARTRAM_ID_32M      0x5c

#define CartRAM0    ((void *)0x02400000)
#define CartRAM1    ((void *)0x02600000)

//  RB_initRAMCart(unsigned char cs)
//  - checks for present RAM cart, inits and verifies it
//
//  returns:
//      cartridge id    -   on success
//                          (8Mb: 0x5a or 32Mb: 0x5c)
//      0xFF            -   no cart
//      0xFE            -   data verify failed 

/*
 * When verifying all the data in the memory cart, about 3~4 seconds
 * are required, so in order to speed-up memory cart verify process, 
 * only some data in the memory cart are checked.
 */
const char* _ram_status_str[] =
{
    "Not detected    ", 
    "1MB test failure", 
    "4MB test failure", 
    "1MB init OK     ", 
    "4MB init OK     ", 
};

unsigned char RB_CartRAM_init(sc_cartram_t* ram, unsigned long detect_flags)
{
#ifndef SC_CARTRAM_USE
    /* Touching at cartridge expansion RAM is prohibited, 
     * so return expansion RAM status as "not detected".
     */
    ram->verify_ok = 1;

    ram->detected = 0;
    ram->size = 0;
    ram->ptr  = NULL;

    ram->status_str = _ram_status_str[0];
    return 0xFF;

#else // SC_CARTRAM_USE

    //unsigned long setReg, refReg;
    unsigned char id;

    ram->verify_ok = 1;

    // 3. write 1 to initialize
    // is at the end of A-BUS Dummy area
    // !!! word format !!! (but what is meant?)
    *((unsigned short *)0x257efffe) = 1;

    // cartridge initialization: as in ST-TECH-47.pdf
    // 1. verify cartridge id
    // 5A - 8 Mb
    // 5C - 32 Mb
    //! is last byte of A-BUS CS1 area
    id = *((unsigned char *)0x24ffffff);
    if(id == 0x5a)
    {
        /*
         * 1MB RAM detected !
         * Physical location of RAM area is
         *  0x22400000 (512KB) and 0x22600000 (512KB)
         * But as RAM areas are mirrored every 512KB, 
         * last mirror of first RAM upper half (512KB) 
         * and RAM lower half (512KB) can be accessed
         * as a 1MB continous RAM block.
         */
        ram->detected = 1;
        ram->size = 1*1024*1024;
        ram->ptr  = (void*)0x22580000;
    }
    else if(id == 0x5c)
    {
        ram->detected = 1;
        ram->size = 4*1024*1024;
        ram->ptr  = (void*)0x22400000;
    }
    else
    {
        ram->detected = 0;
        ram->size = 0;
        ram->ptr  = NULL;
        // 2. prompt proper connection
        //if(quiet == 0) {
        //    slPrint("The extended RAM", slLocate(11,5));
        //    slPrint("cartridge is not", slLocate(11,6));
        //    slPrint("inserted properly.", slLocate(11,7));
        //
        //    slPrint("Please turn off", slLocate(11,9));
        //    slPrint("power and reinsert", slLocate(11,10));
        //    slPrint("the extended RAM", slLocate(11,11));
        //    slPrint("cartridge.", slLocate(11,12));
        //}    
        ram->status_str = _ram_status_str[0];
        return 0xFF; // Error, no cart
    }
    
    // // 4. set A-bus set and refresh registers
    // setReg = refReg = 0;
    // if(cs == 0)
    // {
    //     // set cs0 for 32MBit
    //     setReg |= 1 << 29;  // After-READ precharge insert bit
    //     setReg |= 3 << 24;  // Burst cycle wait number
    //     setReg |= 3 << 20;  // Normal cycle wait number
    // }
    // else
    // {
    //     // set cs1 for 16MBit ??
    //     setReg |= 1 << 28;  // external wait effective bit
    //     setReg |= 15 << 24; // Burst cycle wait number
    //     setReg |= 15 << 20; // Normal cycle wait number
    // }
    // *((unsigned long *)0x25fe00B0) = setReg;
    // // A-Bus refresh register
    // refReg |= 1 << 4;       // A-Bus refresh enable bit
    // // I've just take this value
    // // see SCU user manual for value range
    // refReg |= 1;            // A-Bus refresh cycle wait
    // *((unsigned long *)0x25fe00B8) = refReg;
    (*(unsigned long *)0x25FE00B0) = 0x23301FF0;
    (*(unsigned long *)0x25FE00B8) = 0x00000013;


    /* When performing quick verify, verify only
     * "some" bytes in order to shorten verify time.
     *
     * 17 bytes increment is choosen in order to avoid
     * weird side effects leading to false positive
     * detection.
     */
    unsigned long verify_incr = (detect_flags & SCM_CARTRAM_FULL_VERIFY ? 1 : 17);
    unsigned long *DRAM0, *DRAM1, DRAMsize, i, ok;

    // 5. verify data before proceeding
    // cache-through DRAM0 and DRAM1
    DRAMsize = ram->size >> 3; // byte length per cart -> longword length per chip
    if(ram->size == (1*1024*1024))
    {
        /* Use 1st DRAM chip's last mirror. */
        DRAM0 = (unsigned long *)0x22580000;
    }
    else
    {
        DRAM0 = (unsigned long *)0x22400000;
    }
    DRAM1 = (unsigned long *)0x22600000;

    /*
     * Pre-fill RAM with FFh, then 00h.
     * Don't ask me why Action Replay carts want
     * this to be done in order to get correct
     * access after that ...
     */
    memset(DRAM0, 0xFF, ram->size/2);
    memset(DRAM1, 0xFF, ram->size/2);
    memset(DRAM0, 0x00, ram->size/2);
    memset(DRAM1, 0x00, ram->size/2);

    // write
    for(i = 0; i < DRAMsize; i+=verify_incr)
    {
        *(DRAM0 + i) = i;
        *(DRAM1 + i) = DRAMsize - 1 - i;
    }    

    // read
    ok = 1;
    for(i = 0; i < DRAMsize; i+=verify_incr)
    {
        if(*(DRAM0 + i) != i)
        {
            ok = 0;
        }    
        if(*(DRAM1 + i) != (DRAMsize - 1 - i))
        {
            ok = 0;
        }    
    }

    if(ok == 0)
    {
        //if(quiet == 0)
        //    slPrint("verifying RAM cart FAILED!", slLocate(8,7));
        ram->size = 0;
        ram->ptr  = NULL;
        ram->verify_ok = 0;
        ram->status_str = _ram_status_str[id == 0x5a ? 1 : 2];
        return 0xFE; // Error, verify failed
    }

    //if(quiet == 0)
    //    slPrint("verifying RAM cart OK!", slLocate(8,7));
    ram->status_str = _ram_status_str[id == 0x5a ? 3 : 4];
    return id;

#endif // SC_CARTRAM_USE
}


void scm_cartram_detect(sc_cartram_t* ram, unsigned long detect_flags)
{
    /* Initialize output parameter. */
    memset((void*)ram, 0, sizeof(sc_cartram_t));

    /* Look for official 1MB/4MB RAM. */
    ram->id = RB_CartRAM_init(ram, detect_flags);

    /* Look for RAM on custom cartridge (TBD). */


    /* Some RAM detected ? Don't forget to clear it. */
    if(ram->ptr)
    {
        memset(ram->ptr, 0, ram->size);
    }
}




/* ------------------------------------------------ */
/* ROM code/data/whatever loader. */


/**
 * Copy ROM code data to RAM.
 * Pointer to ROM data must point to ROM code header structure.
 * Length of data to copy, compression options whatever
 * are verified from ROM code header data.
 *
 * Return loaded data length if operation success, 0 else.
**/
#include "sc_compress.h"
unsigned long sc_rom_code_load(void* rom_adr, void* ram_adr)
{
    sc_rom_code_header_t* hdr = (sc_rom_code_header_t*)rom_adr;
    unsigned long load_len = 0;

    /* Valid header ? */
    if((hdr->magic_42 != 0x42)
    || (hdr->magic_53 != 0x53))
    {
        return load_len;
    }

    /* According to type, copy or decompress data. */
    switch(hdr->type)
    {
    case(RC_TYPE_ROM_CODE):
    case(RC_TYPE_DATA_RAW):
    default:
        memcpy(ram_adr, rom_adr+sizeof(sc_rom_code_header_t), hdr->usedlen);
        load_len = hdr->usedlen;
        break;

    case(RC_TYPE_DATA_COMP):
        load_len = decomp_exec((unsigned char*)(rom_adr+sizeof(sc_rom_code_header_t)), hdr->usedlen, (unsigned char*)ram_adr);
        /* If problem during decompression, return error. */
        if(load_len != hdr->decomplen)
        {
            load_len = 0;
        }
        break;
    }

    return load_len;
}




/**
 * Exit to multiplayer.
**/
void sc_exit(void)
{
    (**(void(**)(void))0x0600026C)();
    while(1){;} /* <- Needed in the case yabause & internal bios is used. */
}


/**
 * Execute specified SMPC command.
**/
#define SC_SMPC_CMD_TIMEOUT (5000*1000)
void sc_smpc_command(unsigned char cmd)
{
    unsigned long cntr = 0;

    /* Waits for SF to be Reset. */
    do
    {
        cntr++;
        if(cntr > SC_SMPC_CMD_TIMEOUT)
        {
            return;
        }
    } while(SF == 1);

    /* Writes Command Code in COMREG. */
    COMREG = cmd;

    /* Set SF. */
    SF = 1;

    /* Wait until command is applied.
     * (not really needed ?)
     */
    cntr = 0;
    do
    {
        cntr++;
        if(cntr > SC_SMPC_CMD_TIMEOUT)
        {
            return;
        }
    } while(OREG31 != cmd);
}


/**
 * System Reset.
**/
/* Acquisition timeout value, unit is poll count. */
void sc_sysres(void)
{
    /* Execute system reset command. */
    sc_smpc_command(0x0D);

    /* Final wait. */
    while(1){;}
}




/**
 * SatCom "log to screen" callback.
**/
void scd_logscreen_callback(Fct_scd_screen cons)
{
    sc_internal_t* sc = sc_get_internal();
    sc->dbg_logscreen = cons;
}


/**
 * Show link related logs on screen.
**/
void scd_linkdiag_callback(Fct_scd_screen cons)
{
    sc_internal_t* sc = sc_get_internal();
    sc->dbg_linkdiag = cons;
}




/* Get link name. */
char* _links_name[SC_LINKS_COUNT] =
{
    "None        ", 
    "USB dev cart"
};
char* sc_get_linkname(unsigned char p)
{
    if(p < SC_LINKS_COUNT)
    {
        return _links_name[p];
    }

    return _links_name[0]; /*NONE*/
}




/* ------------------------------------------------ */
/* Copy/execute stub. */

/* Start address, maximum length and parameters address */
#include "copystub/copystub.h"
/* The stub code itself */
#include "copystub/copystub_inc.c"


void copystub_execute(unsigned long softres_flags, unsigned long tmp_address, unsigned long length, unsigned long exec_address)
{
    /* Copy stub to memory and call it (-> quit SatCom library).
     * Code scrap from antime's satclient.c, http://www.iki.fi/~antime/
     */
    int (*stub)(unsigned long, unsigned char*, unsigned long, unsigned char*);
    stub = (int(*)(unsigned long, unsigned char*, unsigned long, unsigned char*))COPYSTUB_START_ADRESS;

    if(COPYSTUB_BIN_SIZ > COPYSTUB_LENGTH) return;

    memcpy((void*)stub, (void*)copystub_bin_dat, COPYSTUB_BIN_SIZ);

    (*stub)(softres_flags, (unsigned char*)(tmp_address & 0x0FFFFFFF), length, (unsigned char*)(exec_address & 0x0FFFFFFF));
}

void sc_execute(unsigned long softres_flags, unsigned long tmp_address, unsigned long length, unsigned long exec_address)
{
    sc_internal_t* sc = sc_get_internal();

    /* If callback available, reset VDP settings. */
    if(sc->soft_reset)
    {
        sc->soft_reset(softres_flags);
    }

    /* Set magic numbers to zero so that functions associated to SatCom debugger are disabled. */
    sc->global->scd_ptr   = 0;
    sc->global->scd_magic = 0;

    copystub_execute(softres_flags, tmp_address, length, exec_address);
}




/* ------------------------- */
/* Wait for VBlank function. */
void sc_wait_vblank(void)
{
    /* First, verify connection with PC. */
    sc_check();

    sc_wait_vblank_in();
    sc_wait_vblank_out();
}






/* ------------------------------------------------------------ */
/* SBL timer library is used in order to measure time interval. */
#include "SBL/INCLUDE/SEGA_TIM.H"

void sc_tmr_start(void)
{
// From sega_tim.h :
 // * NAME:    TIM_FRT_INIT            - FRT ("init free run timer")
 // *
 // * PARAMETERS :
 // *      (1) Uint32 mode             - <i>    (parameter1: clock divider)
// #define TIM_B_CKS0      (0)                             /* CKS0              */
//  #define TIM_CKS_8       ( 0 << TIM_B_CKS0) -> 0             /* 8 JEg        */
//  #define TIM_CKS_32      ( 1 << TIM_B_CKS0) -> 1             /* 32JEg        */
//  #define TIM_CKS_128     ( 2 << TIM_B_CKS0) -> 2             /* 128JEg       */
//  #define TIM_CKS_OUT     ( 3 << TIM_B_CKS0) -> 3             /* OJEg      */


    TIM_FRT_INIT(TIM_CKS_32); // TIM_CKS_8, TIM_CKS_32 or TIM_CKS_128
    TIM_FRT_SET_16(0);
}

unsigned short sc_tmr_lap(void)
{
    return (TIM_FRT_GET_16());
}
unsigned long sc_tmr_lap_usec(void)
{
    return sc_tmr_tick2usec(TIM_FRT_GET_16());

    // Float32 f = TIM_FRT_CNT_TO_MCR(TIM_FRT_GET_16());
    // f = f*10; // ???
    // return ((unsigned short)(f));
}


// Note : In order to avoid overflows, tick can't be greater than 1300000.
unsigned long sc_tmr_tick2usec(unsigned long tick)
{
    /* Old stuff before it was using floating point computation. */
    //Float32 f = TIM_FRT_CNT_TO_MCR(tick);
    //return ((unsigned long)(f));

    /* Older stuff that correctly cheched HW clock and PAL/NTSC rate.
     * Q&D : use fixed values to convert tick to usec. (-> timing may change with HW region)
     *     sc_tmr_tick2usec(60000)=71444
     *     sc_tmr_tick2usec(10000)=11907
     *     sc_tmr_tick2usec(2^14)=19508
     * Note : In order to avoid overflows, tick can't be greater than 155000.
     */
    //unsigned long ret = tick;
    //ret = ret*19508;
    //ret = ret >> 14;
    //return ret;

    /* New stuff that uses fixed point computation. */
    return ((unsigned long)(TIM_FRT_CNT_TO_MCR(tick)));
}

void sc_tick_sleep(unsigned short tick)
{
    sc_tmr_start();
    while(TIM_FRT_GET_16() < tick);
}
void sc_usleep(unsigned long usec)
{
    /* Convert delay to tick value. */
    unsigned long delay_tick;

    if(usec <= 50*1000)
    { /* Compute tick value at good precision for delay lower than 50 msec. */

        delay_tick = (usec*60000) / (sc_tmr_tick2usec(60000));

        // Changed to delay_tick = (usec*55040) / (sc_tmr_tick2usec(55040)); where sc_tmr_tick2usec(55040) = 64K
        // delay_tick = (usec*55040);
        // delay_tick = delay_tick >> 16;
    }
    else
    { /* Poor precision, but no overflow (up to 42 seconds) for higher values. */
        delay_tick = (usec*100) / (sc_tmr_tick2usec(100));

        // Changed to delay_tick = (usec*108) / (sc_tmr_tick2usec(108)); where sc_tmr_tick2usec(108) = 128
        // delay_tick = (usec*108);
        // delay_tick = delay_tick >> 7;
    }

    /* Sleep at most 60000 ticks. */
    unsigned long i;
    for(i=0; i<delay_tick; i+=60000)
    {
        unsigned long s = ((i+60000) < delay_tick ? 60000 : delay_tick - i);
        sc_tick_sleep(s);
    }
}




/* -------------------------------------------------------- */
/* Retrieve current date and time from SMPC internal clock. */


/**
 * SMPC clock acquisition test.
 * Retrieve Saturn internal clock value.
 * Note: Doesn't uses SMPC interrupt. Instead, SMPCregister is polled during acquisition.
 * 
 * For details about clock value acquisition, see information below (taken from Sega docs).
 * --- ST-169-R1-072694.pdf pages 13-16: ---
 *
 *  Table 2.4 Command Issue Types
 *    Type    Command
 *      A   MSHON, SYSRES, NMIREQ, CKCHG352, CKCHG320
 *      B   SSHON, SSHOFF, SNDON, SNDOFF, CDON, CDOFF, RESENAB, RESDISA
 *      C   SETTIME, SETSMEM
 *      D   INTBACK
 *
 *  Figure 2.4 Type D Command Flow
 *      1. START
 *      2. Waits for SF to be Reset
 *      3. Sets the Command Parameter in IREG
 *      4. Writes Command Code in COMREG
 *      5. Set SF
 *      6. END
 *
 *      1. SMPC Interrupt Processing START
 *      2. Clear Interrupt Flag
 *      3. Look at SR and Receive OREG Result Parameter
 *      4. Depending on Peripheral Status,
 *         Issue Continue or Break to SMPC
 *         (Refer to Chapter 3 for Details)
 *      5. Interrupt Processing END
 *  Note:
 *   A routine similar to the SMPC interrupt routine can be executed by masking the
 *   SMPC routine (SMPC interrupt is not used) and conducting polling after the SMPC
 *   interrupt flag or SF clear is conducted.
**/
/* Acquisition timeout value, unit is poll count. */
#define SC_RTC_TIMEOUT (5000*1000)
/* macro BCD_INT : convert bcd to int */
#define BCD_INT(bcd) (((bcd & 0xf0) >> 4) * 10 + (bcd &0x0f))


/* Small (but dirty) tweak to restore back
 * pad acquisition when it uses SMPC interrupt.
 *
 * This is required to correctly retrieve current
 * time from SMPC.
 */
#ifdef SMPC_USE_POLLING
#else // !SMPC_USE_POLLING
void pad_init(void);
void pad_shutdown(void);
#endif // !SMPC_USE_POLLING


struct tm* sc_tim_smpcpoll(void)
{
    sc_internal_t* sc = sc_get_internal();
    unsigned long cntr = 0;
    unsigned char oreg_tmp[7];
    int tmp;
    struct tm* t;

    t = &(sc->set.time);

    /* First, verify if should reset last date in buffer. */
    if((t->tm_sec  > 61)
    || (t->tm_min  > 59)
    || (t->tm_hour > 23)
    || (t->tm_mday == 0)
    || (t->tm_mday > 31)
    || (t->tm_mon  > 11)
    || (t->tm_wday >  6))
    {
        memset((void*)t, 0, sizeof(struct tm));

        /* Set some dummy time in the case SMPC use is forbidden, 
         * or in the case acquisition failed.
         * "Dummy time" is set as follow :
         *  "The value in question, 0x008246A0, corresponds to 
         *  1996/3/26 0:00, which pretty much coincides with
         *  the passing of Comet Hyakutake. As far as easter
         *  eggs go, that'd be my guess."
         */
        /* The number of years since 1900. */
        t->tm_year  = 1996-1900;
        t->tm_mon   = 3-1;
        t->tm_mday  = 26;
        t->tm_hour  = 0;
        t->tm_min   = 0;
        t->tm_sec   = 0;
        t->tm_isdst = -1;
    }

    /* Can't use SMPC time. Okay, don't go further. */
    if(!sc->smpctime_use)
    {
        return t;
    }

    /* Avoid interference with pad when it uses SMPC interrupt. */
#ifdef SMPC_USE_POLLING
#else // !SMPC_USE_POLLING
    pad_shutdown();

    /* Small wait to be sure that pad interrupt acquisition is off ... */
    sc_wait_vblank_in();
    sc_wait_vblank_out();
    sc_wait_vblank_in();
    sc_wait_vblank_out();
#endif // !SMPC_USE_POLLING

    /* Waits for SF to be Reset. */
    do { cntr++; if(cntr > SC_RTC_TIMEOUT) return t; } while(SF == 1);

    /* Sets the Command Parameter in IREG. */
    IREG0 = 0x01;
    IREG1 = 0x08;
    IREG2 = 0xF0;

    /* Writes Command Code in COMREG. */
    COMREG = 0x10;

    /* Set SF. */
    SF = 1;

    /* Wait SF Clear. */
    do { cntr++; if(cntr > SC_RTC_TIMEOUT) return t; } while(SF == 1);

    /* Get register values from SMPC. */
    oreg_tmp[0] = OREG1;
    oreg_tmp[1] = OREG2;
    oreg_tmp[2] = OREG3;
    oreg_tmp[3] = OREG4;
    oreg_tmp[4] = OREG5;
    oreg_tmp[5] = OREG6;
    oreg_tmp[6] = OREG7;

    /* Restore back pad acquisition when it uses SMPC interrupt. */
#ifdef SMPC_USE_POLLING
#else // !SMPC_USE_POLLING
    pad_init();
#endif // !SMPC_USE_POLLING

    /* The number of years since 1900. */
    tmp = (BCD_INT(oreg_tmp[0]) * 100 + BCD_INT(oreg_tmp[1])) - 1900;
    t->tm_year = tmp;

    /* The number of months since January, in the range 0 to 11. */
    tmp = (oreg_tmp[2] & 0x0F) - 1;
    t->tm_mon  = tmp;

    /* The day of the month, in the range 1 to 31. */
    tmp = BCD_INT(oreg_tmp[3]);
    t->tm_mday = tmp;

    /* The number of hours past midnight, in the range 0 to 23. */
    tmp = BCD_INT(oreg_tmp[4]);
    t->tm_hour = tmp;

    /* The number of minutes after the hour, in the range 0 to 59. */
    tmp = BCD_INT(oreg_tmp[5]);
    t->tm_min  = tmp;

    /* The number of seconds after the minute, normally in the range 0 to 59, but can be up to 60 to allow for leap seconds. */
    tmp = BCD_INT(oreg_tmp[6]);
    t->tm_sec  = tmp;

    /* No information about daylight saving. */
    t->tm_isdst = -1;

    /* tm_wday and tm_tm_yday fields will be set by mktime function. */

    return t;
}


void sc_bup_conv_date(struct tm* t, BupDate* sat_date)
{
//typedef struct BupDate {
//	Uint8	year;		/* Year - 1980 */
//	Uint8	month;		/* Month (1~12) */
//	Uint8	day;		/* Day (1~31) */
//	Uint8	time;		/* Hour (0~23) */
//	Uint8	min;		/* Minute (0~59) */
//	Uint8	week;		/* Weekday: (Sunday:0 - Saturday:6) */
//}
    sat_date->year  = t->tm_year - 80;
    sat_date->month = t->tm_mon + 1;
    sat_date->day   = t->tm_mday;
    sat_date->time  = t->tm_hour;
    sat_date->min   = (t->tm_hour >= 60 ? 59 : t->tm_min);
    sat_date->week  = t->tm_wday;
}




void sc_initcfg_reset(sc_initcfg_t* cfg, unsigned long* workram)
{
    /* Clear SatCom work RAM. */
    memset((void*)workram, 0, SC_WORKRAM_U32SIZE*sizeof(unsigned long));

    /* Reset initialization settings. */
    memset((void*)cfg, 0, sizeof(sc_initcfg_t));
    cfg->workram   = workram;
    cfg->flags     = SC_INITFLAG_DEFAULT;
}

/* Default file name when logging to SD card. */
char* _logfile_default = "satcom.log";

void sc_init_cfg(sc_initcfg_t* cfg)
{
    sc_internal_t* sc;
    sdcard_t* sd;

    /* First, test structures size. */
    if((sizeof(sc_internal_t) != SC_INTERNAL_SIZE)
    || (sizeof(scd_data_t)    != SC_DEBUGGER_SIZE))
    {
        while(1) {;}
    }

    /* Set pointer to internal data. */
    _sc = (sc_internal_t*)(cfg->workram);
    sc = sc_get_internal();
    memset((void*)sc, 0, /*sizeof(sc_internal_t)*/SC_WORKRAM_U32SIZE*sizeof(unsigned long));

    /* Set pointer to LRAM global data. */
    sc->global = (sc_global_t*)SC_GLOBAL_DATA_ADR;

    /* Set pointer to debugger data. */
    sc->dbg = (scd_data_t*)(((unsigned char*)sc) + sizeof(sc_internal_t));
    /* Small tweak in order to get debug data start address 4bytes alignment. */
    if((((unsigned long)sc->dbg) & 0x3) != 0)
    {
        sc->dbg = (scd_data_t*)((((unsigned long)sc->dbg) + 4) & ~0x3);
    }

    /* Init debugger stuff. */
    memset((void*)sc->dbg, 0, sizeof(scd_data_t));
    sc->dbg_logscreen = NULL;
    sc->logfile_name  = (cfg->logfile_name ? cfg->logfile_name : _logfile_default);

    /* Set debugger structure size and version */
    sc->dbg->version[0] = SCD_VERSION_MAJOR;
    sc->dbg->version[1] = SCD_VERSION_MINOR;
    sc->dbg->structsz   = sizeof(scd_data_t);
    /* Init circular buffer-related stuff. */
    scd_cbinit(sc->dbg);

    /* Output location informations. */
    sc->dbg->wh_on_log = 0;
    sc->dbg->dbg_logflush = 0;

    /* Set soft reset callback. */
    sc->soft_reset = cfg->soft_reset;

    /* If needed, set debug callback. */
    if(cfg->logscreen_callback)
    {
        scd_logscreen_callback(cfg->logscreen_callback);
    }

    /* Set debugger absolute data.
     * // sc->global->scd_magic = SCD_ABS_MAGIC;
     * // sc->global->scd_ptr   = (unsigned long)(sc->dbg);
     */
    sc_set_absolute_data(sc);

    /* Set buffer where to upload saturn executables. */
#ifdef SC_EXEC_IN_LRAM
    /* Execute satcom in LRAM -> use HRAM as a temporary buffer. */
    sc->exec_buffer = (unsigned char*)0x06004000;
#else
    /* Execute satcom in HRAM -> use LRAM as a temporary buffer. */
    sc->exec_buffer = (unsigned char*)0x00200000;
#endif

    /* Indicate if we can use CDROM or not. */
    sc->cdrom_status = (cfg->flags & SC_INITFLAG_NOCDROM ? 0 : SC_CDSTATUS_USE);

    /* Indicate if we can retrieve time from SMPC or not. */
    sc->smpctime_use = (cfg->flags & SC_INITFLAG_NOSMPCTIME ? 0 : 1);

    /* Indicate if we need to delete log file (SD card)
     * on first log output or not.
     */
    sc->already_log_sd = (cfg->flags & SC_INITFLAG_KEEPLOG ? 1 : 0);

    /* Check if custom yabause is used or not. */
    unsigned char ver_major = REG_CART_CPLD_VER >> 4;
    if((ver_major != 0x0) && (ver_major != 0xF))
    {
        sc->yabause_use = (REG_CART_BETA_ID & 0x80 ? 1 : 0);
    }
    else
    {
        sc->yabause_use = 0;
    }

    /* Set SatCom init flag to true */
    sc->init_ok = 1;

    /* Look for extension device. */
    sc->ext_dev = SC_EXT_NONE;

    if(memcmp((void*)(0x02000000), (void*)"SEGA SEGASATURN ", 16) == 0)
    { /* Boot cartridge (Action Replay or equivalent) detected. */
        sc->ext_dev = SC_EXT_BOOTCART;
    }

    if(sc_usb_is_port_available())
    { /* USB dev cart detected. */
        sc->ext_dev = SC_EXT_USBDC;
    }

    if((REG_CPLD_55 == 0x55)
    && (REG_CPLD_AA == 0xAA)
    && ((REG_CART_CPLD_VER & 0xF0) == 0x10))
    {
        /* SatCart v1 detected. */
        sc->ext_dev = SC_EXT_SATCART;
    }
    else
    {
        /* Todo : Add Wasca detection stuff here. */
    }


#ifndef SC_NO_SDCARD
    /* Init SD card library. */
    sdc_fat_reset();
    sdc_fat_init();
#endif // SC_NO_SDCARD

    /* Set "log to SD card" flag. */
    sd = sc_get_sdcard_buff();
    sd->log_sdcard = sc->set.log_sdcard;

    scd_logout(" sd->init_status=%d", sd->init_status);
    scd_logout(" hw_ver=0x%02X, yabause_use=%d", REG_CART_CPLD_VER, sc->yabause_use);

    /* Get link type from backup memory to SatCom internal buffer. */
    sc_settings_read(NULL);

    /* Check if the SD card should be accessed in read-only mode.
     *
     * Rather than using the physical lock switch not available
     * on micro SD card, read-only mode is enabled when a special
     * file is found on SD card.
     *
     * Note : READ_ONLY.TXT file must contain something to enable
     *        read-only mode. It won't do anything is file size
     *        is zero bytes.
     */
    sc->sd_read_only = (scf_size(SC_FILE_SDCARD, "/READ_ONLY.TXT") == 0 ? 0 : 1);


    /* Initialize access to Satiator. */
    sc->satis_flags = SATIS_INIT_WAIT;
#ifndef SC_NO_SATIATOR
    if((cfg->flags & SC_INITFLAG_NOSATIATOR) == 0)
    {
        /* Enable Satiator API, then check if it is available or not.
         *
         * Note about s_mode return values :
         *  -  1 : already set to specified mode
         *  -  0 : success
         *  - -1 : Satiator not detected
         */
#if SATIS_LOG == 1
        scd_logout("[SATIS]Test 1 : s_mode(s_api) ...");
#endif // SATIS_LOG == 1
        if(s_mode(s_api) >= 0)
        {
#if SATIS_LOG == 1
            scd_logout("[SATIS]Test 2 : s_mode(s_api) OK");
#endif // SATIS_LOG == 1
            sc->satis_flags |= SATIS_SDCARD_OK;
        }
    }
#if SATIS_LOG == 1
    scd_logout("[SATIS]Test 9 : END flags:0x%02X", sc->satis_flags);
#endif // SATIS_LOG == 1
#endif // ndef SC_NO_SATIATOR


    /* Init rand seed, in a Q&D way, but better than nothing.
     * Note : calling sc_tim_smpcpoll function is MANDATORY here.
     * The reason is that calling this function will initialize
     * time field in settings, and this field will be used by vmem
     * (or any other module where SMPC access is forbidden) when
     * creating files on SD card.
     */
    struct tm* t = sc_tim_smpcpoll();
    unsigned long rand_seed = mktime(t);
    srand(rand_seed);
    scd_logout("* date: %4d/%02d/%02d %02d:%02d:%02d (0x%08X)", 
        1900+t->tm_year, 1+t->tm_mon, t->tm_mday, 
        t->tm_hour, t->tm_min, t->tm_sec, 
        rand_seed);
    scd_logout("* size:sdcard=%d", sizeof(sdcard_t));
    scd_logout("* size:dbgr=%d, intrnl=%d(free=%d), total=%d[%d]", 
        sizeof(scd_data_t), 
        sizeof(sc_internal_t), 
        sizeof(sc->reserved9), 
        sizeof(scd_data_t)+sizeof(sc_internal_t), 
        SC_WORKRAM_SIZE);
    scd_logout("* sc->global=0x%08X, dbg=0x%08X, sd_ro:%d", sc->global, sc->dbg, sc->sd_read_only);

#ifdef SC_EXEC_IN_LRAM
    scd_logout("* Welcome to SatCom debugger (LRAM)", 0);
#else
    scd_logout("* Welcome to SatCom debugger", 0);
#endif
    scd_logout("* Build: " __DATE__ ", " __TIME__, 0);
    scd_logout(" --- ~ ~ ~ ~ --- ", 0);
}


void sc_init(unsigned long* workram)
{
    sc_initcfg_t cfg;

    /* Init SatCom with default settings. */
    sc_initcfg_reset(&cfg, workram);
    sc_init_cfg(&cfg);
}

unsigned char sc_extdev_getid(void)
{
    sc_internal_t* sc = sc_get_internal();

    return sc->ext_dev;
}

const char* _ext_dev_names[SC_EXTDEV_CNT] = 
{
    "Not Found   ", 
    "Boot Cart   ", 
    "USB dev cart", 
    "Gamer's Cart", 
    "Wasca       ", 
};
const char* sc_extdev_getname(void)
{
    sc_internal_t* sc = sc_get_internal();

    return _ext_dev_names[sc->ext_dev];
}

/**
 * Get/set SatCom settings from/to backup memory.
**/
void sc_settings_read(sc_settings_t* dat_out)
{
    sc_internal_t* sc = sc_get_internal();
    unsigned long ret;

    /* Read settings from SD card. */
    memset((void*)(&(sc->set)), 0, sizeof(sc_settings_t));
    ret = scf_read(SC_FILE_SDCARD, SC_SETTINGS_NAME/*filename*/, 0/*offset*/, sizeof(sc_settings_t)/*size*/, (void*)(&(sc->set)));

    if((ret != sizeof(sc_settings_t))
    || (sc->set.struct_version != SC_SET_VER1))
    {
        /* If couldn't read settings, delete invalid file and use parameters default values. */
        scf_delete(SC_FILE_SDCARD, SC_SETTINGS_NAME/*filename*/);

        memset((void*)(&(sc->set)), 0, sizeof(sc_settings_t));
        sc->set.struct_size    = sizeof(sc_settings_t);
        sc->set.struct_version = SC_SET_VER1;
    }

    if(sc->yabause_use)
    {
        /* Dirty hack when yabause is used : don't use any link. */
        sc->set.link_type = SC_LINK_NONE;
    }
    else
    {
        /* Automatically enable USB dev cart transfer
         * when USB port is detected.
         */
        sc->set.link_type = sc_usb_is_port_available() ? SC_LINK_USBDC : SC_LINK_NONE;
    }

    /* Copy the resulting data to the output buffer. */
    if(dat_out != NULL)
    {
        memcpy((void*)dat_out, (void*)(&(sc->set)), sizeof(sc_settings_t));
    }
}

void sc_settings_write(sc_settings_t* dat_in)
{
    sc_internal_t* sc = sc_get_internal();
    sdcard_t* sd = sc_get_sdcard_buff();

    if(dat_in == NULL) return;

    /* Update satcom library internal settings. */
    memcpy((void*)(&(sc->set)), (void*)dat_in, sizeof(sc_settings_t));

    /* Indicate SD card library if we save logs
     * to SD card or not.
     * This is needed in order to avoid SD card library
     * outputing logs that will be saved to SD card, 
     * that will output more logs, etc.
     */
    sd->log_sdcard = sc->set.log_sdcard;

    /* Save settings to SD card. */
    scf_write((void*)(&(sc->set)), sizeof(sc_settings_t)/*size*/, SC_FILE_SDCARD, SC_SETTINGS_NAME/*filename*/);
}

void sc_settings_clear(void)
{
    scf_delete(SC_FILE_SDCARD, SC_SETTINGS_NAME/*filename*/);
    sc_settings_read(NULL);
}

sc_settings_t* sc_settings_getptr(void)
{
    sc_internal_t* sc = sc_get_internal();
    return &(sc->set);
}



/* USB_FLAGS, USB_FIFO etc definitions, 
 * for USB dev cart.
 */
#include "satcart.h"

int sc_usb_is_port_available(void)
{
    unsigned char tmp = USB_FLAGS;

    /* USB flag's bit 6-2 are always low, and other 
     * cartridges (verified on Action Replay)
     * return FFh when reading this address, so the following
     * stuff below should detect if USB port is present or not.
     */
    if((tmp & 0x7C) == 0)
    {
        return 1;
    }

    return 0;
}

int sc_usb_is_connected(void)
{
    unsigned char tmp = USB_FLAGS;

    /* Verify bit 7 (PWREN#) which is low when
     * FTDI chip is USB powered.
     */
    if((tmp & 0xFC) == 0)
    {
        return 1;
    }

    return 0;
}


#ifndef SC_RELEASE

/*----------------------------------------------------------------------------*/
/*- Interrupt related functions (scrapped from iapetus). ---------------------*/
/*- Note: sh2int is also used in order to access CDROM. ----------------------*/
#include "iapetus/iapetus.h"
#include "iapetus/sh2/sh2int.h"


void sc_check_usbdc(void);


/* Small macro (and its best friend doing all the dirty jobs) in
 * order to show link related log messages on screen when
 * link diagnostic mode is turned on.
 * This uses a couple of local variables, hence won't compile
 * if not declared on each functions where used.
 */
void _sc_internal_linkdiag(sc_internal_t* sc, char *string, ...)
{
    va_list argptr;
    char buff[40];

    /* Print the log message to the memory. */
    va_start(argptr, string);
    vsnprintf(buff, sizeof(buff), string, argptr); buff[sizeof(buff)-1] = '\0';
    va_end(argptr);

    /* Send back to user program. */
    sc->dbg_linkdiag(buff);
}
#define LINKDIAG_LOGOUT(_STR_, ...) if(linkdiag_use){_sc_internal_linkdiag(sc, _STR_, __VA_ARGS__);}


void _sc_internal_check(void)
{
    sc_internal_t* sc = sc_get_internal();
    int old_level_mask;

    unsigned char l = sc->set.link_type;

    if(l == SC_LINK_USBDC)
    {
        /* Mask any interrupts.
         *
         * Reason to mask interrupts is that Pseudo Saturn Kai cheat
         * codes provides in-game USB connectivity, which may interfer
         * with this "source level USB connectivity" provided
         * by sc_check_usbdc function.
         */
        old_level_mask = interrupt_get_level_mask();
        interrupt_set_level_mask(0xF);

        sc_check_usbdc();

        /* Return interrupts back to normal. */
        interrupt_set_level_mask(old_level_mask);
    }
}


static unsigned char RecvByte(void)
{
    while ((USB_FLAGS & USB_RXF) != 0) ;
    return USB_FIFO;
}

static void SendByte(unsigned char byte)
{
    while ((USB_FLAGS & USB_TXE) != 0) ;
    USB_FIFO = byte;
}


static unsigned long RecvDword(void)
{
    unsigned long tmp = RecvByte();
    tmp = (tmp << 8) | RecvByte();
    tmp = (tmp << 8) | RecvByte();
    tmp = (tmp << 8) | RecvByte();

    return tmp;
}

static void SendDword(unsigned long data)
{
    SendByte((data >> 24) & 0xFF);
    SendByte((data >> 16) & 0xFF);
    SendByte((data >>  8) & 0xFF);
    SendByte((data >>  0) & 0xFF);
}


const unsigned char _usbdc_version_dat[USBDC_VERSION_LEN] =
{
    0x31, /* version */
    0x00, 0x00, 0x00,       /* reserved */
    0x00, 0x00, 0x00, 0x00, /* reserved */
    0x00, 0x00, 0x00, 0x00, /* reserved */
    0x00, 0x00, 0x00, 0x00  /* reserved */
};


static void DoDownload(void)
{
    sc_internal_t* sc = sc_get_internal();
    int linkdiag_use = (sc->dbg_linkdiag ? 1 : 0);

    unsigned char *pData;
    unsigned long len, ii;
    unsigned long rw4_buffer;
    unsigned char checksum = crc_usbdc_init();

    pData = (unsigned char*)RecvDword();
    len = RecvDword();
    LINKDIAG_LOGOUT(" ptr=0x%08X len=0x%08X", pData, len);

    if((len == 2) || (len == 4))
    { /* 1,2,4 bytes length data access. */
        if(len == 2)
        {
            unsigned short* ptr_in2  = (unsigned short*)(((unsigned long)pData) & 0xFFFFFFFE);
            unsigned short* ptr_out2 = (unsigned short*)&rw4_buffer;
            *ptr_out2 = *ptr_in2;
        }
        else if(len == 4)
        {
            unsigned long* ptr_in4  = (unsigned long*)(((unsigned long)pData) & 0xFFFFFFFC);
            unsigned long* ptr_out4 = (unsigned long*)&rw4_buffer;
            *ptr_out4 = *ptr_in4;
        }
        pData = (unsigned char*)&rw4_buffer;
    }

    /* USB dev cart firmware version check ? */
    if(((((unsigned long)pData) & 0x0FFFFFFF) == (USBDC_VERSION_ADR & 0x0FFFFFFF))
    && (len == USBDC_VERSION_LEN))
    {
        pData = (unsigned char*)_usbdc_version_dat;
    }


    for (ii = 0; ii < len; ++ii)
    {
        while ((USB_FLAGS & USB_TXE) != 0) ;
        USB_FIFO = pData[ii];
    }

    checksum = crc_usbdc_update(checksum, pData, len);
    checksum = crc_usbdc_finalize(checksum);

    SendByte(checksum);
}



static void DoUpload(void)
{
    sc_internal_t* sc = sc_get_internal();
    int linkdiag_use = (sc->dbg_linkdiag ? 1 : 0);

    unsigned char *pData;
    unsigned long len, ii;
    unsigned long rw4_buffer;
    unsigned char readchecksum;
    unsigned char checksum = USBDC_DUMMY_CRC;

    if(linkdiag_use)
    {
        /* Verify checksum only in link diagnostic mode. */
        checksum = crc_usbdc_init();
    }

    pData = (unsigned char*)RecvDword();
    len = RecvDword();
    LINKDIAG_LOGOUT(" ptr=0x%08X len=0x%08X", pData, len);

    if((len == 2) || (len == 4))
    { /* 1,2,4 bytes length data access. */
        unsigned char *pBuffer = (unsigned char*)&rw4_buffer;

        /* Receive data to write. */
        for (ii = 0; ii < len; ++ii)
        {
            // inlining is 20K/s faster
            while ((USB_FLAGS & USB_RXF) != 0) ;
            pBuffer[ii] = USB_FIFO;
        }

        if(linkdiag_use)
        {
            /* Verify checksum only in link diagnostic mode. */
            checksum = crc_usbdc_update(checksum, pBuffer, len);
            checksum = crc_usbdc_finalize(checksum);
        }

        if(len == 2)
        {
            unsigned short* ptr_in2  = (unsigned short*)pBuffer;
            unsigned short* ptr_out2 = (unsigned short*)(((unsigned long)pData) & 0xFFFFFFFE);
            *ptr_out2 = *ptr_in2;
        }
        else if(len == 4)
        {
            unsigned long* ptr_in4  = (unsigned long*)pBuffer;
            unsigned long* ptr_out4 = (unsigned long*)(((unsigned long)pData) & 0xFFFFFFFC);
            *ptr_out4 = *ptr_in4;
        }
    }
    else
    {
        for (ii = 0; ii < len; ++ii)
        {
            // inlining is 20K/s faster
            while ((USB_FLAGS & USB_RXF) != 0) ;
            pData[ii] = USB_FIFO;
        }

        if(linkdiag_use)
        {
            /* Verify checksum only in link diagnostic mode. */
            checksum = crc_usbdc_update(checksum, pData, len);
            checksum = crc_usbdc_finalize(checksum);
        }
    }

    if(linkdiag_use)
    {
        /* Verify checksum only in link diagnostic mode. */
        readchecksum = RecvByte();
    }
    else
    {
        /* Read, then ignore checksum from PC in non-debug mode. */
        RecvByte();
        readchecksum = USBDC_DUMMY_CRC;
    }

    if (checksum != readchecksum)
    {
        LINKDIAG_LOGOUT(" Checksum error ! 0x%02X 0x%02X", checksum, readchecksum);
        SendByte(0x1);
    }
    else
    {
        SendByte(0);
    }
}

static void DoExecute(void)
{
    sc_internal_t* sc = sc_get_internal();
    int linkdiag_use = (sc->dbg_linkdiag ? 1 : 0);

    /* Read address, execute call. */
    void (*pFun)(void);

    pFun = (void(*)(void))RecvDword();
    LINKDIAG_LOGOUT(" ptr=0x%08X", pFun);

    /* If callback available, perform soft reset. */
    if(sc->soft_reset)
    {
        /* Don't reset HRAM and LRAM, in order to avoid
         * problem with program currently executing.
         */
        sc->soft_reset(SC_SOFTRES_ALL & (~(SC_SOFTRES_HRAM | SC_SOFTRES_LRAM)));
    }

    (*pFun)();
}

static void DoExecExt(void)
{
    sc_internal_t* sc = sc_get_internal();
    int linkdiag_use = (sc->dbg_linkdiag ? 1 : 0);

    unsigned char *pData;
    unsigned char *pExec;
    unsigned long len, ii;
    unsigned long softres_flags;
    unsigned char readchecksum;
    unsigned char checksum = USBDC_DUMMY_CRC;

    /* Verify checksum only in link diagnostic mode. */
    if(linkdiag_use)
    {
        checksum = crc_usbdc_init();
    }

    /*
     * PC sends execute address (4 bytes)
     * PC sends data length     (4 bytes)
     * PC sends reset flags     (4 bytes)
     * PC sends data            (variable length)
     * PC sends CRC             (1 byte)
     * Saturn sends CRC check   (1 byte, 0: CRC match, else: CRC mismatch)
     */

    pExec = (unsigned char*)RecvDword();

    /* Verify if can receive data directly to execution address or not. */
    if((((unsigned long)(pExec)) & 0x0F000000) != ((unsigned long)(sc->exec_buffer) & 0x0F000000))
    {
        pData = sc->exec_buffer;
    }
    else
    {
        pData = pExec;
    }
    LINKDIAG_LOGOUT(" ptr=0x%08X 0x%08X", pExec, pData);

    len = RecvDword();

    softres_flags = RecvDword();
    LINKDIAG_LOGOUT(" len=0x%08X, flags=0x%08X", len, softres_flags);

    for (ii = 0; ii < len; ++ii)
    {
        // inlining is 20K/s faster
        while ((USB_FLAGS & USB_RXF) != 0) ;
        pData[ii] = USB_FIFO;
    }

    /* Verify checksum only in link diagnostic mode. */
    if(linkdiag_use)
    {
        checksum = crc_usbdc_update(checksum, pData, len);
        checksum = crc_usbdc_finalize(checksum);

        readchecksum = RecvByte();
    }
    else
    {
        RecvByte();
        readchecksum = USBDC_DUMMY_CRC;
    }

    if (checksum != readchecksum)
    {
        LINKDIAG_LOGOUT(" Checksum error ! 0x%02X 0x%02X", checksum, readchecksum);
        SendByte(0x1);
    }
    else
    { /* CRC matched, can execute program. */
        SendByte(0);

        /* If callback available, perform soft reset. */
        if(sc->soft_reset)
        {
            sc->soft_reset(softres_flags);
        }

        sc_execute(softres_flags, (unsigned long)pData, len, (unsigned long)pExec);
    }
}



static void DoSendBufferAddress(void)
{
    sc_internal_t* sc = sc_get_internal();
    unsigned long tmp = (unsigned long)sc->exec_buffer;

    SendDword(tmp);
}

static void DoCopyExec(void)
{
    unsigned long tmp_address  = RecvDword();
    unsigned long length       = RecvDword();
    unsigned long exec_address = RecvDword();

    sc_execute(SC_SOFTRES_ALL, tmp_address, length, exec_address);
}


void sc_check_usbdc(void)
{
    sc_internal_t* sc = sc_get_internal();
    int linkdiag_use = (sc->dbg_linkdiag ? 1 : 0);

    int can_exit = 1;
    unsigned char command;

    do
    {
        unsigned char usb_flags = USB_FLAGS;

        if((usb_flags & USB_RXF) == 0)
        {
            command = USB_FIFO;
            LINKDIAG_LOGOUT("CMD=0x%02X", command);

            // // Configure SCU A-bus settings
            // // Disabling A-bus refresh adds ~20KB/s
            // AREF = 0;
            // // Removing wait states doubles speed
            // ASR0 = 0;

            switch(command)
            {
            case USBDC_FUNC_DOWNLOAD:
                DoDownload();
                break;

            case USBDC_FUNC_UPLOAD:
                DoUpload();
                // can_exit = 0; //TMP
                break;

            case USBDC_FUNC_EXEC:
                DoExecute();
                can_exit = 1;
                break;

            case USBDC_FUNC_GET_BUFF_ADDR:
                DoSendBufferAddress();

                /* PC will soon send data for upload & exec, so it's better
                 * to stay here until the end.
                 */
                can_exit = 0;
                break;

            case USBDC_FUNC_COPYEXEC:
                DoCopyExec();
                can_exit = 1;
                break;

            case USBDC_FUNC_EXEC_EXT:
                DoExecExt();
                can_exit = 0;
                break;

            default:
                break;
            }

            //LINKDIAG_LOGOUT("CMD=0x%02X END can_exit:%d", command, can_exit);
        }

    } while(can_exit == 0);
}


#endif // !SC_RELEASE








/*----------------------------------------------------------------------------*/
/*- Debugger functions. ------------------------------------------------------*/

scd_data_t* sc_get_dbg_buff(void)
{
    sc_internal_t* sc = sc_get_internal();

    return sc->dbg;
}


#ifndef SC_RELEASE
void _scd_log_append(unsigned char type, unsigned char id, char *buff, unsigned long bufflen, int usbonly)
{
    sc_internal_t* sc = sc_get_internal();

    /* Set length and type to data */
    unsigned long len = strnlen(buff+3, bufflen-3);
    if(len == 0) return;
    buff[0] = 3 + len;
    buff[1] = type;
    buff[2] = id; /* 0: console, 1~: string grid */

    /* Append the log message to the debug data circular buffer. */
    scd_cbwrite(sc->dbg, (unsigned char*)buff, 3+len);

    /* If requested by user, display log message to screen. */
    if(sc->dbg_logscreen)
    {
        sc->dbg_logscreen(buff+3);
    }

    /* If requested by user, save log message to SD card. */
    if((!usbonly) && (sc->set.log_sdcard))
    {
        /*
         * Message stored in global buff array will be overwritten
         * next time logout function will be called.
         * This happens when SD card module call logout function during
         * scf_append function call.
         * Hence, data is copied into sc->dbg->prompt_answer. This works
         * well at the condition that scd_prompt function is not called
         * from SD card library.
         */
        char* buff_sd = (char*)(sc->dbg->prompt_answer);
#if 0
        /* Prefix something to easily find where_here messages in log file. */
        strcpy(buff_sd, "[LOC]");
        memcpy(buff_sd+5, buff+3, MIN(SCD_PRINT_LEN-3, SCD_PROMPT_SIZE-5-1));
#else
        memcpy(buff_sd, buff+3, MIN(SCD_PRINT_LEN-3, SCD_PROMPT_SIZE-1));
#endif

        /* Don't forget to append line break before writing to SD card. */
        buff_sd[SCD_PROMPT_SIZE-2] = '\0';
        len = strlen(buff_sd);
        buff_sd[len++] = '\r';
        buff_sd[len++] = '\n';

        /* If needed, delete previous log file, then append log message to file.
         * This feature just helps to avoid super long log file growing on SD card, 
         * but on the other hand wipes log after re-init of SatCom library, hence is
         * disabled by default.
         */
        if(!sc->already_log_sd)
        {
            //scf_delete(SC_FILE_SDCARD, sc->logfile_name); // <- un-comment me if needed
            sc->already_log_sd = 1;
        }
        scf_append((void*)(buff_sd), len, SC_FILE_SDCARD, sc->logfile_name/*filename*/);
    }

    /* Circular buffer auto-flush. */
    if(sc->dbg->dbg_on)
    {
        /* Update circular buffer memory usage. */
        sc->dbg->cb_memuse = scd_cbmemuse(sc->dbg);

        /* If requested by user, or if circular buffer nearly full (75%), automatically flush it to the PC. */
        if((sc->dbg->dbg_logflush)
        || (sc->dbg->cb_memuse >= 96/*75% x 128*/))
        {
            scd_logflush();
        }
    }
}

int sc_vsnprintf(int log_sdcard, char* buff, int buff_sz, const char* fmt, va_list args)
{
#ifndef SC_FAST_LOGOUT_DISABLE
    int len = 0;
    const char *w;
    char c;

    if(log_sdcard)
#endif // SC_FAST_LOGOUT_DISABLE
    {
        /* Log on SD card -> need to format everything on SH-2 side to make log readable. */
        return vsnprintf(buff, buff_sz, fmt, args);
    }

#ifndef SC_FAST_LOGOUT_DISABLE
    if(buff_sz <= 0)
    {
        return 0;
    }

    /* Process format string. */
    w = fmt;
    while ((c = *w++) != 0)
    {
        /* If not a format escape character, just print  */
        /* character.  Otherwise, process format string. */
        if (c != '%')
        {
            buff[len++] = c;
            if(len >= buff_sz)
            {
                return buff_sz;
            }
        }
        else
        {
            /* Get format character.  If none     */
            /* available, processing is complete. */
            if ((c = *w++) != 0)
            {
                if (c == '%')
                {
                    /* Process "%" escape sequence. */
                    buff[len++] = c;
                    if(len >= buff_sz)
                    {
                        return buff_sz;
                    }
                } 
                else if (c == 'c')
                {
                    int v = va_arg(args, int);
                    buff[len++] = v;
                    if(len >= buff_sz)
                    {
                        return buff_sz;
                    }
                }
                else if (c == 's')
                {
                    /* Process string format. */
                    char *s = va_arg(args, char *);

                    while(*s)
                    {
                        buff[len++] = *s++;
                        if(len >= buff_sz)
                        {
                            return buff_sz;
                        }
                    }
                }
                else
                {
                    /* Process hexadecimal number format.
                     * If not '%x' format, provide format and value in
                     * output string.
                     * Example : "%u[CAFE]", which should be displayed as
                     *           "51966" on PC side.
                     */
                    int direct_print = ((c == 'x') || (c == 'X') ? 1 : 0);
                    unsigned long v = va_arg(args, unsigned long);
                    unsigned long digit;
                    int digit_shift;

                    /* If not direct print, provide format in output buffer. */
                    if(!direct_print)
                    {
                        buff[len++] = '%';
                        if(len >= buff_sz)
                        {
                            return buff_sz;
                        }
                        while(1)
                        {
                            buff[len++] = c;
                            if(len >= buff_sz)
                            {
                                return buff_sz;
                            }

                            /* Don't forget to provide any eventual
                             * length specificator.
                             */
                            if((c < '0') || (c > '9'))
                            {
                                break;
                            }

                            c = *w++;
                        }

                        /* Add delimiter for value. */
                        buff[len++] = '[';
                        if(len >= buff_sz)
                        {
                            return buff_sz;
                        }
                    }

                    if (v == 0)
                    {
                        /* If the number value is zero, just print and continue. */
                        buff[len++] = '0';
                        if(len >= buff_sz)
                        {
                            return buff_sz;
                        }
                    }
                    else
                    {
                        /* Find first non-zero digit. */
                        digit_shift = 28;
                        while (!(v & (0xF << digit_shift)))
                        {
                            digit_shift -= 4;
                        }

                        /* Print digits. */
                        for (; digit_shift >= 0; digit_shift -= 4)
                        {
                            digit = (v & (0xF << digit_shift)) >> digit_shift;
                            if (digit <= 9)
                            {
                                c = '0' + digit;
                            }
                            else
                            {
                                c = 'A' + digit - 10;
                            }
                            buff[len++] = c;
                            if(len >= buff_sz)
                            {
                                return buff_sz;
                            }
                        }
                    }

                    /* If not direct print, don't forget to close value delimiter. */
                    if(!direct_print)
                    {
                        buff[len++] = ']';
                        if(len >= buff_sz)
                        {
                            return buff_sz;
                        }
                    }
                }
            }
            else
            {
                break;
            }
        }
    }

    /* Don't forget the terminating null character. */
    buff[len] = '\0';

    return len;
#endif // SC_FAST_LOGOUT_DISABLE
}

void _scd_internal_logout(const char* file, int line, const char* func, int usbonly, char *string, ...)
{
    sc_internal_t* sc = sc_get_internal();
    va_list argptr;
    char* buff = sc->dbg_print;

    /* If debug is disabled, don't do anything. */
    if(sc->dbg->dbg_disable) return;

    /* If needed, indicate from where we sent log message. */
    if(sc->dbg->wh_on_log)
    {
        _scd_internal_wh(file, line, func, "^-^", usbonly);
    }

    /* Print the log message to the memory. */
    va_start(argptr, string);

    int log_sdcard = ((!usbonly) && (sc->set.log_sdcard) ? 1 : 0);
    sc_vsnprintf(log_sdcard, buff+3, SCD_PRINT_LEN-3, string, argptr);
    buff[SCD_PRINT_LEN-1] = '\0';

    va_end(argptr);

    _scd_log_append(SCD_TYPE_MSG, 0, buff, SCD_PRINT_LEN, usbonly);
}

void _scd_internal_msgout(const char* file, int line, const char* func, int usbonly, unsigned char id, char *string, ...)
{
    sc_internal_t* sc = sc_get_internal();
    va_list argptr;
    char* buff = sc->dbg_print;

    /* If debug is disabled, don't do anything. */
    if(sc->dbg->dbg_disable) return;

    /* If needed, indicate from where we sent log message. */
    if(sc->dbg->wh_on_log)
    {
        _scd_internal_wh(file, line, func, "^-^", usbonly);
    }

    /* Print the log message to the memory. */
    va_start(argptr, string);

    int log_sdcard = ((!usbonly) && (sc->set.log_sdcard) ? 1 : 0);
    sc_vsnprintf(log_sdcard, buff+3, SCD_PRINT_LEN-3, string, argptr);
    buff[SCD_PRINT_LEN-1] = '\0';

    va_end(argptr);

    _scd_log_append(SCD_TYPE_MSG, id, buff, SCD_PRINT_LEN, usbonly);
}


char* _dummy_answer = "";
char* scd_prompt(char* prompt_str)
{
    sc_internal_t* sc = sc_get_internal();
    char* buff = sc->dbg_print;

    /* Exit when debugger is disabled, as release version would do. */
    if(sc->dbg->dbg_disable) return _dummy_answer;

    strncpy(buff+3, prompt_str, SCD_PRINT_LEN-3); buff[SCD_PRINT_LEN-1] = '\0';

    /* Send prompt message.
     * Note about usbonly parameter set to '1' :
     * Log to SD card use "sc->dbg->prompt_answer" global buffer that
     * is also used by this function.
     * Hence log to SD card is disabled in "scd_prompt" case.
     */
    _scd_log_append(SCD_TYPE_PROMPT, 0/*to console*/, buff, SCD_PRINT_LEN, 1/*usbonly*/);

    /* Wait until the PC finishes to send the prompt answer. */
    sc->dbg->proc_complete = 0;
    while(sc->dbg->proc_complete == 0)
    {
        /* Re-set absolute absolute values at the end of LRAM so that
         * SatLink would correctly check the proper debugger context, 
         * and consequently flush logs.
         * This re-set is mandatory when SatCom is used in two distinct
         * modules, and that first module (with its global values overwritten
         * by first one) tries to flush logs.
         * Typical example : Save Data Manager and vmem : scd_logflush calls
         *                   from vmem are "hidden" by Save Data Manager.
         */
        sc_set_absolute_data(sc);

        /* Wait for PC to consume all debugger data. */
        sc_check();
    }

    /* Return the prompt answer. */
    sc->dbg->prompt_answer[SCD_PROMPT_SIZE-1] = '\0';
    return (char*)sc->dbg->prompt_answer;
}



void scd_logflush(void)
{
    sc_internal_t* sc = sc_get_internal();

    /* Exit when debugger is disabled, as release version would do. */
    if(sc->dbg->dbg_disable) return;

    while(sc->dbg->writeptr != sc->dbg->readptr)
    {
        /* Re-set absolute absolute values at the end of LRAM so that
         * SatLink would correctly check the proper debugger context, 
         * and consequently flush logs.
         * This re-set is mandatory when SatCom is used in two distinct
         * modules, and that first module (with its global values overwritten
         * by first one) tries to flush logs.
         * Typical example : Save Data Manager and vmem : scd_logflush calls
         *                   from vmem are "hidden" by Save Data Manager.
         */
        sc_set_absolute_data(sc);

        /* Wait for PC to consume all debugger data. */
        sc_check();
    }
}


void _scd_internal_wh(const char* file, int line, const char* func, char* msg, int usbonly)
{
    sc_internal_t* sc = sc_get_internal();
    char* buff = sc->dbg_print;

    /* If debug is disabled, don't do anything. */
    if(sc->dbg->dbg_disable) return;

    /* Print the log message to the memory. */
    snprintf(buff+3, SCD_PRINT_LEN-3, "%s::%d [%s] (%s)", file, line, func, msg);
    buff[SCD_PRINT_LEN-1] = '\0';

    _scd_log_append(/*SCD_TYPE_MSG*/SCD_TYPE_WHEREHERE, 0/*id (nonzero for grid)*/, buff, SCD_PRINT_LEN, usbonly);
}


#endif // !SC_RELEASE


#ifdef SC_STACK_MEASURE_USE

void _scd_stack_measure_internal(const char* func, unsigned long stack_addr)
{
    unsigned long stack_addr_here;

    asm("mov r15,%0": "=r"(stack_addr_here));
    stack_addr_here = stack_addr_here & 0x0FFFFFFC;

    stack_addr = stack_addr & 0x0FFFFFFC;

    scd_logout("[STACK]%4u bytes [ADR:0x%08X] in function \"%s\""
        , stack_addr-stack_addr_here
        , stack_addr
        , func);
}

unsigned long _scd_stack_absolute_measure_internal(const char* func, char* str)
{
    /* Measure how far stack was modified. */
#define STACK_MAX_DEPTH (2048/sizeof(unsigned long))
#define STACK_INIT_VAL  0xA5A55A5A
    unsigned long stack_usage = 0;
    int i;
    for(i=(STACK_MAX_DEPTH-1); i>0; i--)
    {
        if(*((unsigned long*)(0x06004000 - i*sizeof(unsigned long))) != STACK_INIT_VAL)
        {
            stack_usage = i*sizeof(unsigned long);
            break;
        }
    }

    scd_logout("[STACK]ABS:%4u (0x%08X) in function \"%s\" (%s)"
        , stack_usage
        , 0x06004000 - i*sizeof(unsigned long)
        , func
        , str);

    return stack_usage;
}
#endif // SC_STACK_MEASURE_USE



/*----------------------------------------------------------------------------*/
/*- File I/O functions --> PC stuff. -----------------------------------------*/


unsigned long _scd_file_request(
    unsigned char type, 
    unsigned long foffset, 
    unsigned long frw_address, 
    unsigned long frw_len, 
    char*         fname)
{
#ifndef SC_RELEASE
    sc_internal_t* sc = sc_get_internal();
    char* buff = sc->scd_file_request_buff;

    /* If debug is disabled, don't do anything. */
    if(sc->dbg->dbg_disable) return 0;

    /* Write the parameters. */
    unsigned char offs = 2;
    buff[offs++] = ((foffset) >> 24) & 0xFF;
    buff[offs++] = ((foffset) >> 16) & 0xFF;
    buff[offs++] = ((foffset) >>  8) & 0xFF;
    buff[offs++] = ((foffset) >>  0) & 0xFF;

    buff[offs++] = ((frw_address) >> 24) & 0xFF;
    buff[offs++] = ((frw_address) >> 16) & 0xFF;
    buff[offs++] = ((frw_address) >>  8) & 0xFF;
    buff[offs++] = ((frw_address) >>  0) & 0xFF;

    buff[offs++] = ((frw_len) >> 24) & 0xFF;
    buff[offs++] = ((frw_len) >> 16) & 0xFF;
    buff[offs++] = ((frw_len) >>  8) & 0xFF;
    buff[offs++] = ((frw_len) >>  0) & 0xFF;

    /* Print the filename to the memory. */
    snprintf(buff+offs, SCDF_BUFFSIZE-offs, fname);

    /* Set length and type to data */
    unsigned long len = offs + strnlen(buff+offs, SCDF_BUFFSIZE-offs);
    buff[0] = 2 + len;
    buff[1] = type;

    /*
     * In the case circular buffer is full, file request data
     * will not appended to it, hence PC will never set the
     * proc_complete flag, which results in infinite loop.
     *
     * So that flushing circular buffer is needed before
     * appending file request data.
     */
    scd_logflush();

    /* Append the file request to the debug data circular buffer. */
    scd_cbwrite(sc->dbg, (unsigned char*)buff, 2+len);

    /* Wait until the PC processes the file data request. */
    sc->dbg->proc_complete = 0;
    sc->dbg->fileio_size   = 0;
    do
    {
        /* Re-set absolute absolute values at the end of LRAM so that
         * SatLink would correctly check the proper debugger context, 
         * and consequently flush logs.
         * This re-set is mandatory when SatCom is used in two distinct
         * modules, and that first module (with its global values overwritten
         * by first one) tries to flush logs.
         * Typical example : Save Data Manager and vmem : scd_logflush calls
         *                   from vmem are "hidden" by Save Data Manager.
         */
        sc_set_absolute_data(sc);

        sc_check();
    } while(sc->dbg->proc_complete == 0);

    /* Return size actually read/written/appended. */
    return(sc->dbg->fileio_size);
#else
    /* Don't do anything in release mode. */
    return 0;
#endif
}



/*----------------------------------------------------------------------------*/
/*- File I/O functions --> SD card stuff. ------------------------------------*/
sdcard_t* sc_get_sdcard_buff(void)
{
    sc_internal_t* sc = sc_get_internal();
    return &(sc->sd);
}


unsigned char sc_get_sdcard_devid(void)
{
    unsigned char dev_id = SC_FILE_INVALID;

    /* In the case both top cartridge and Satiator are detected, priority is put
     * on on SD card from top cartridge like Gamer's Cartridge or Wasca etc.
     * Reason is that user is more likely to remove top cartridge than Satiator.
     */
    sc_device_t dev;

    scf_dev_status(&dev, SC_FILE_SDCARD);

    if(dev.available)
    {
        dev_id = SC_FILE_SDCARD;
    }
#ifndef SC_NO_SATIATOR
    else
    {
        scf_dev_status(&dev, SC_FILE_SATIATOR);

        if(dev.available)
        {
            dev_id = SC_FILE_SATIATOR;
        }
    }
#endif // ndef SC_NO_SATIATOR

    return dev_id;
}


/*----------------------------------------------------------------------------*/
/*- File I/O functions --> RomFs stuff. --------------------------------------*/
#include "sc_romfs.h"

int sc_romfs_setdata(unsigned char* romfs)
{
    sc_internal_t* sc = sc_get_internal();

    /* Update pointer to RomFs data only if header and CRC verified succesfully. */
    if(romfs_check_header(romfs, 1/*crc_check*/))
    {
        sc->romfs = romfs;

        return 1;
    }

    return 0;
}

unsigned char* sc_romfs_get_dataptr(char* filename, unsigned long* datalen)
{
    sc_internal_t* sc = sc_get_internal();
    romfs_entry_t entry;
    romfs_entry_t* e;
    unsigned char* ret;

    if(datalen)
    {
        *datalen = 0;
    }

    e = romfs_get_first_entry(sc->romfs, filename);

    if(e == NULL)
    {
        ret = NULL;
    }
    else
    {
        if(e->flags & FILEENTRY_FLAG_FOLDER)
        {
            ret = NULL;
        }
        else
        {
            memcpy((void*)&entry, (void*)e, sizeof(romfs_entry_t));
            ret = romfs_get_entry_ptr(sc->romfs, &entry, datalen);
        }
    }

    return ret;
}



/*----------------------------------------------------------------------------*/
/*- File I/O functions --> Extra device stuff. -------------------------------*/
#ifdef SC_EXTRA_MEDIA

/* Pointer to extra media structure.
 * The body of this structure must be declared on user program side and
 * registered with this (scfx_register) function.
 *
 * If this pointer is set to NULL, extra media is not used.
 * If this points to something, then access to extra media is enabled
 * via scf_* functions.
 */
sc_extra_media_t* _sc_extra_media_ptr = NULL;

void scfx_register(sc_extra_media_t* extra_media)
{
    _sc_extra_media_ptr = extra_media;
}

#endif // SC_EXTRA_MEDIA



void scf_filename_bufferize(char* filename, char* filename_buff)
{
    int i;

    memset((void*)filename_buff, '\0', SC_PATHNAME_MAXLEN);

    for(i=0; i<SC_PATHNAME_MAXLEN; i++)
    {
        if(filename[i] == '\\')
        {
            /* Convert backslashes to slashes. */
            filename_buff[i] = '/';
        }
        else
        {
            filename_buff[i] = filename[i];
        }
        if(filename[i] == '\0')
        {
            break;
        }
    }

    filename_buff[SC_PATHNAME_MAXLEN-1] = '\0';
}


/*----------------------------------------------------------------------------*/
/*- File I/O functions --> CDROM stuff (scrapped from iapetus). --------------*/
//#include "cwx_cdrom/iapetus_cdrom.h"
#include "iapetus/iapetus.h"
#include "iapetus/cd/cd.h"
#include "iapetus/cd/cdfs.h"
#include "iapetus/sh2/sh2int.h"


/* Workram required by CDROM library. */
unsigned long _cd_workbuff[SC_CDWORKBUFF_SIZE/sizeof(unsigned long)];
file_struct _cdrom_file;

void cdrom_fs_init(void)
{
    sc_internal_t* sc = sc_get_internal();

    cdfs_init((void *)_cd_workbuff, sizeof(_cd_workbuff));

    /* Mark CDROM library as initialized.
     * Forcing SC_CDSTATUS_USE bit to be set is a tweak to enable
     * override CD-ROM access even if CD-ROM was initially disabled.
     */
    sc->cdrom_status |= (SC_CDSTATUS_USE|SC_CDSTATUS_INIT);
}

void cdrom_init_check(void)
{
    sc_internal_t* sc = sc_get_internal();

    /* Need to init CD-ROM ? */
    if( (sc->cdrom_status & SC_CDSTATUS_USE )
    &&(!(sc->cdrom_status & SC_CDSTATUS_INIT)))
    {
        cdrom_fs_init();
    }
}


/* CDROM file open/close functions, for SatCom library internal use. */
int cdrom_open_file(char* filename)
{
    char filename_buff[SC_PATHNAME_MAXLEN] = {'\0'};

    /* File name is normally sanitized into a global buffer, 
     * but in the case of CD-ROM access being overridden by
     * another device this global buffer can't be used twice.
     *
     * As a workaround, local buffer is used. Assuming that
     * applications critical regarding stack usage (vmem)
     * don't use CD-ROM access, this shouldn't be a problem.
     */
    scf_filename_bufferize(filename, filename_buff);

    return cdfs_open(filename_buff, &_cdrom_file);
}

void cdrom_close_file(void)
{
    cdfs_close(&_cdrom_file);
}

/* CD-ROM access retry count.
 * This is a desesperate attempt to get file access
 * working with dying disc drives.
 * It is used directly into scf_size and scf_read
 * functions so that Pseudo Saturn Kai isn't affected.
 */
#define CDROM_RETRY_CNT 5



const char* _dev_names[SC_BASE_DEVICES_CNT * 2] = 
{
    /* 6 letters , Max 12 letters  */
    "CD-ROM"     , "CD-ROM"      , 
    "RomFs "     , "RomFs"       , 
    "PClink"     , "PC Link"     , 
    "SDcard"     , "SD card"     , 
    "Satiat"     , "Satiator"    , 
};

char* scf_dev_getname_ex(unsigned char dev_id, int long_name)
{
    int id_offset = (long_name == 0 ? 0 : 1);
    char* ret = (char*)(_dev_names[id_offset]);

    switch(dev_id)
    {
    case(SC_FILE_CDROM):
    case(SC_FILE_ROMFS):
    case(SC_FILE_PC):
    case(SC_FILE_SDCARD):
    case(SC_FILE_SATIATOR):
        ret = (char*)(_dev_names[(2*dev_id) + id_offset]);
        break;

#ifdef SC_EXTRA_MEDIA
    case(SC_FILE_EXTRA):
        if(_sc_extra_media_ptr)
        { /* Consider other IDs as extra media(s). */
            if(long_name)
            {
                ret = _sc_extra_media_ptr->long_name;
            }
            else
            {
                ret = _sc_extra_media_ptr->short_name;
            }
        }
        break;
#endif // SC_EXTRA_MEDIA

    default:
        break;
    }

    return ret;
}

void scf_dev_status(sc_device_t* dev, unsigned char dev_id)
{
    sc_internal_t* sc = sc_get_internal();
    sdcard_t* sd;

    /* Reset device list.
     * Note : because contents are reset to zero, subsequent settings
     *        after that are commented out.
     *        This saves around 20 bytes on executable size.
     */
    memset((void*)dev, 0, sizeof(sc_device_t));

    /* Set status for device with specified ID. */
    dev->dev_id = dev_id;
    dev->name   = scf_dev_getname(dev_id);

    switch(dev_id)
    {
    case(SC_FILE_CDROM):
        dev->available  = (sc->cdrom_status & SC_CDSTATUS_USE ? 1 : 0);
        dev->read_only  = 1;
        //dev->model      = 0; /* Unused. */
        //dev->speed_info = 0x00;
        break;

    case(SC_FILE_ROMFS):
        dev->available  = romfs_check_header(sc->romfs, 0/*crc_check*/);
        dev->read_only  = 1;
        //dev->model      = 0; /* Unused. */
        //dev->speed_info = 0x00;
        break;

    case(SC_FILE_PC):
        dev->available  = (sc->dbg->dbg_on ? 1 : 0);
        //dev->read_only  = 0;
        //dev->model      = 0; /* Unused. */
        //dev->speed_info = 0x00;
        break;

    case(SC_FILE_SDCARD):
        /* Note : possible SD card models are :
         *  SC_EXT_NONE    (no SD card detected)
         *  SC_EXT_SATCART (Gamer's Cartridge)
         *  SC_EXT_WASCA
         * (Defined in sc_saturn.h file)
         */
        dev->model = sc_extdev_getid();
        sd = sc_get_sdcard_buff();
        dev->available  = (dev->model != SC_EXT_NONE ? (sd->init_ok ? 1 : 0) : 0);
        dev->read_only  = sc->sd_read_only;
        dev->speed_info = SC_DEVFLAG_FASTWRITE;
        break;

#ifndef SC_NO_SATIATOR
    case(SC_FILE_SATIATOR):
        dev->available  = (sc->satis_flags & SATIS_SDCARD_OK ? 1 : 0);
        //dev->read_only  = 0;
        //dev->model      = 0; /* Unused. */
        dev->speed_info = SC_DEVFLAG_FASTWRITE;
        break;
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    case(SC_FILE_EXTRA):
        if(_sc_extra_media_ptr)
        { /* Consider other IDs as extra media(s). */
            if(_sc_extra_media_ptr->fct_dev_status)
            {
                _sc_extra_media_ptr->fct_dev_status(dev, dev_id);
            }
        }
        break;
#endif // SC_EXTRA_MEDIA

    default:
        /* Unknown device. */
        //dev->available  = 0;
        //dev->read_only  = 0;
        //dev->model      = 0; /* Unused. */
        //dev->speed_info = 0x00;
        break;
    }
}


int scf_dev_list(sc_device_t* devlist)
{
    unsigned char i, j;

    /* If needed, enable PC link. */
    sc_check();

    j=0;
    for(i=0; i<SC_DEVICES_CNT; i++)
    {
        /* Set devices properties. */
        scf_dev_status(devlist+j, i/*dev_id*/);

        /* If device is available, update count. */
        if(devlist[j].available)
        {
            j++;
        }
    }

    return j;
}


#ifndef SC_NO_SATIATOR

/* Small wait before requesting Satiator to do something.
 * This is a temporary solution in the hope that a firmware fix will allow SD
 * card access without the need of artificial delay between operations.
 */
#define SATIATOR_SMALL_WAIT()  sc_usleep(10000)
#define SATIATOR_MIDDLE_WAIT() sc_usleep(30000)


/* Access override structure, used in order to read
 * "CD-ROM" contents via SD card.
 */
cdfs_over_t _satis_over = {0};

/* File descriptor to the iso file of this application. */
int _satis_over_fd = -1;

void satis_over_end(void)
{
    if(_satis_over_fd >= 0)
    {
        s_close(_satis_over_fd);
        _satis_over_fd = -1;
    }
}

int satis_over_init(void)
{
    /* Just in case of, terminate ongoing override access. */
    satis_over_end();

    /* In SD card current folder, look for a file with same IP header
     * as our application. If found, then open it with read-only access.
     */
    char* path = "";

    if(s_opendir(path) == 0)
    {
        char statbuf[280];
        s_stat_t *st = (s_stat_t*)statbuf;

        while(1)
        {
            int len = s_stat(NULL, st, sizeof(statbuf)-1);
            if((len <= 0) || (len >= (sizeof(statbuf)-9/*size, attrib etc*/)))
            {
                break;
            }

            st->name[len] = '\0';

            /* Skip special directories such as ".", ".." or ".system" etc.
             *
             * Also, avoid to access hidden or other exotic entry.
             *
             * And, avoid folders, because we are looking for an iso file.
             */
            if((st->name[0] < ' ') || (st->attrib & (AM_HID|AM_SYS|AM_VOL | AM_DIR)))
            {
                continue;
            }

            /* Don't try to read non-iso files. */
            if(!scf_compare_extension(st->name, "iso"))
            {
                continue;
            }

            /* Iso file found : verify its IP header. */
            int fd = s_open(st->name, FA_READ);

            if(fd >= 0)
            {
                unsigned char ip_header_buff[0xD0];

                if(s_read(fd, ip_header_buff, sizeof(ip_header_buff)) == sizeof(ip_header_buff))
                {
                    /* In order to bypass region check, Satiator loader
                     * utility tweaks region codes in IP header copy in RAM.
                     * As a consequence, it is necessary to skip region code
                     * (16 bytes from offset 0x40) when comparing IP headers.
                     */
                    unsigned char* ip_hdr = (unsigned char*)0x06002000;
                    if((memcmp(ip_header_buff+0x10, ip_hdr+0x10, 0x30) == 0)
                    && (memcmp(ip_header_buff+0x50, ip_hdr+0x50, 0x80) == 0))
                    {
                        /* IP header matches with ours ! */
                        _satis_over_fd = fd;
                        return IAPETUS_ERR_OK;
                    }
                }

                s_close(fd);
            }
        }
    }

    return IAPETUS_ERR_FILENOTFOUND;
}

int satis_over_read(void* buffer, u32 FAD, u32 num_bytes)
{
    if(_satis_over_fd >= 0)
    {
        /* Support only simple iso files because I don't have any plan
         * to develop an application in another format.
         */
        unsigned long file_offset = (FAD - 150) * 2048;
        s_seek(_satis_over_fd, file_offset, SEEK_SET);

        unsigned long read_offset;
        for(read_offset=0; read_offset < num_bytes; /*nothing*/)
        {
            int chunk_len = MIN(num_bytes - read_offset, S_MAXBUF);

            int sret = s_read(_satis_over_fd, ((unsigned char*)buffer) + read_offset, chunk_len);
            if(sret != chunk_len)
            {
                return IAPETUS_ERR_UNEXPECTDATA;
            }

            read_offset += chunk_len;
        }
    }

    return IAPETUS_ERR_OK;
}


void scf_satiator_access_init(unsigned char dev_id, int set_time)
{
    sc_internal_t* sc = sc_get_internal();


    /* On first access to emulated CD-ROM or SD card, do some initializations :
     *  1. Override access to CD-ROM with sectors read from SD card.
     *      This is done in order to avoid a voodoo bug which freezes the
     *      application when trying to access emulated CD-ROM after setting
     *      s_api mode.
     *
     *  2. Set current directory to SD card root folder
     *      Menu utility (which launches this applications) sets current
     *      directory to the directory containing the disc image, so it's
     *      necessary to set it to root folder in order to be able to access
     *      whole SD card.
     */
    if(sc->satis_flags == (SATIS_INIT_WAIT | SATIS_SDCARD_OK))
    {
        /* Setup override access. */
        memset(&_satis_over, 0, sizeof(cdfs_over_t));
        _satis_over.init = satis_over_init;
        _satis_over.read = satis_over_read;
        _satis_over.end  = satis_over_end;
        cdfs_override_start(&_satis_over);


        /* Satiator menu doesn't sets current directory to root one, 
         * so let's do it now.
         * Changing directory only once on startup should do the job
         * but anyway, as s_chdir function takes a negligible time
         * let's call it just before every functions.
         */
        s_chdir("/");


        /* Indicate that CD-ROM initialization went fine. */
        sc->satis_flags = SATIS_SDCARD_OK | SATIS_CDROM_OK;
    }



    /* For access modifying file time stamp (write, mkdir etc), 
     * it is recommended to provide current time to Satiator right now.
     *
     * Note : as date and time functions from fat_io_lib are re-used, 
     *        it won't set anything special in the case Gamer's Cartridge
     *        SD card access is disabled.
     *        Also, it will set a dummy timestamp in the case time SMPC
     *        time acquisition is disabled; that's not a bug but a feature.
     */
#ifndef SC_NO_SDCARD
    if((dev_id == SC_FILE_SATIATOR)
    && (set_time)
    && (sc->satis_flags & SATIS_SDCARD_OK))
    {
        struct tm * time_info;
        uint32 fat_time;
        uint32 fat_date;
        uint32_t timestamp;

        /* Get system time. */
        time_info = sc_tim_smpcpoll();

        /* Convert time to FAT format. */
        fat_time = fatfs_convert_to_fat_time(time_info->tm_hour, time_info->tm_min, time_info->tm_sec);

        /* Convert date to FAT format. */
        fat_date = fatfs_convert_to_fat_date(time_info->tm_mday, time_info->tm_mon + 1, time_info->tm_year + 1900);

        /* Put both date and time together. */
        timestamp = (fat_date << 16) | (fat_time);
        s_settime(timestamp);
    }
#endif // SC_NO_SDCARD

}
#endif // ndef SC_NO_SATIATOR


unsigned long scf_size(unsigned char dev_id, char* filename)
{
    sc_internal_t* sc = sc_get_internal();

    unsigned long size = 0;

    romfs_entry_t romfs_entry;
    romfs_entry_t* e;

    int retry;

    scf_filename_bufferize(filename, sc->filename_buff);
#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 0/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Use file from PC. */
        _scd_file_request(
            /*type       */SCD_TYPE_SFILE, 
            /*foffset    */0, 
            /*frw_address*/(unsigned long)(&size), 
            /*frw_len    */sizeof(unsigned long), 
            /*fname      */sc->filename_buff);
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Use file from RomFs. */
        e = romfs_get_first_entry(sc->romfs, sc->filename_buff);
        if(e == NULL)
        {
            size = 0;
        }
        else
        {
            if(e->flags & FILEENTRY_FLAG_FOLDER)
            {
                size = 0;
            }
            else
            {
                memcpy((void*)&romfs_entry, (void*)e, sizeof(romfs_entry_t));
                size = romfs_get_entry_size(sc->romfs, &romfs_entry);
            }
        }
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* Use file from CDROM. */
        if(sc->cdrom_status & SC_CDSTATUS_USE)
        {
            for(retry=0; retry<CDROM_RETRY_CNT; retry++)
            {
                cdrom_init_check();

                if(cdrom_open_file(filename) != IAPETUS_ERR_OK)
                {
                    continue;
                }

                size = _cdrom_file.size;
                cdrom_close_file();
            }
        }
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Use file from SD card. */
        sdc_fat_file_size(sc->filename_buff, &size);
        // scd_usbsd_logout(1/*usbonly*/, "@@@[SD]sdc_fat_file_size(filename=%s, size=%u) ...", sc->filename_buff, size);
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* Use file from Satiator. */
        SATIATOR_MIDDLE_WAIT();

        /* Extract folder and file name from full path. */
        char* path = sc->filename_buff;
        int path_len = 0;
        for(int i=0; i<SC_PATHNAME_MAXLEN; i++)
        {
            char c = sc->filename_buff[i];
            if(c == '\0')
            {
                break;
            }
            if(c == '/')
            {
                path_len = i;
            }
        }
        path[path_len] = '\0';
        char* name = sc->filename_buff + path_len+1;

#if SATIS_LOG == 1
        scd_logout("[SATIS]size path[%s] name[%s]", path, name);
#endif // SATIS_LOG == 1

        if(s_opendir(path))
        {
            size = 0;
        }
        else
        {
            char statbuf[280];
            s_stat_t *st = (s_stat_t*)statbuf;

            /* Folder found. List files inside it. */
            while(1)
            {
                int len = s_stat(NULL, st, sizeof(statbuf)-1);
                if((len <= 0) || (len >= (sizeof(statbuf)-9/*size, attrib etc*/)))
                {
                    break;
                }

                st->name[len] = '\0';

#if SATIS_LOG == 1
                scd_logout("[SATIS]size size:%6u st->name[%s]", st->size, st->name);
#endif // SATIS_LOG == 1

                /* Skip special directories such as ".", ".." or ".system" etc.
                 *
                 * Also, avoid to return the "size" of a directory entry, 
                 * or any kind of other exotic entry.
                 */
                if((st->name[0] < ' ') || (st->attrib & (AM_HID|AM_SYS|AM_VOL|AM_DIR)))
                {
                    continue;
                }

                if(strcmp(st->name, name) == 0)
                {
                    size = st->size;
                    break;
                }
            }
        }

#if SATIS_LOG == 1
        scd_logout("[SATIS]size ret:%u", size);
#endif // SATIS_LOG == 1
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_size)
        {
            size = _sc_extra_media_ptr->fct_size(dev_id, sc->filename_buff);
        }
    }
#endif // SC_EXTRA_MEDIA

    return size;
}



unsigned long scf_read(unsigned char dev_id, char* filename, unsigned long offset, unsigned long size, void* ptr)
{
    sc_internal_t* sc = sc_get_internal();

    unsigned long retsize = 0;

    romfs_entry_t romfs_entry;
    romfs_entry_t* e;

    int retry;

    scf_filename_bufferize(filename, sc->filename_buff);
#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

    // scd_usbsd_logout(1/*usbonly*/, "@@@scf_fread(ptr=%08X, 0ffset=%d, size=%d, dev_id=%d, name=\"%s\")", ptr, offset, size, dev_id, sc->filename_buff);

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 0/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Use file from PC. */
        unsigned long file_size = scf_size(dev_id, filename);

        /* Check parameters. */
        retsize = size;
        if(offset > file_size)
        {
            offset = 0;
        }
        if((offset + retsize) > file_size)
        {
            retsize = file_size - offset;
        }

        if(retsize)
        {
            retsize = _scd_file_request(
                        /*type       */SCD_TYPE_RFILE, 
                        /*foffset    */offset, 
                        /*frw_address*/(unsigned long)(ptr), 
                        /*frw_len    */retsize, 
                        /*fname      */sc->filename_buff);
        }
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Use file from RomFs. */
        unsigned long file_size = scf_size(dev_id, filename);

        /* Check parameters. */
        retsize = size;
        if(offset > file_size)
        {
            offset = 0;
        }
        if((offset + retsize) > file_size)
        {
            retsize = file_size - offset;
        }

        if(retsize)
        {
            e = romfs_get_first_entry(sc->romfs, sc->filename_buff);

            if(e == NULL)
            {
                retsize = 0;
            }
            else
            {
                if(e->flags & FILEENTRY_FLAG_FOLDER)
                {
                    retsize = 0;
                }
                else
                {
                    memcpy((void*)&romfs_entry, (void*)e, sizeof(romfs_entry_t));
                    romfs_get_entry_data(sc->romfs, &romfs_entry, (unsigned char*)ptr, offset, retsize);
                }
            }
        }
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* Use file from CD-ROM. */
        if(sc->cdrom_status & SC_CDSTATUS_USE)
        {
            for(retry=0; retry<CDROM_RETRY_CNT; retry++)
            {
                cdrom_init_check();

                if(cdrom_open_file(filename) != IAPETUS_ERR_OK)
                {
                    continue;
                }

                /* Check parameters. */
                if(offset > _cdrom_file.size)
                {
                    offset = 0;
                }
                if((offset + size) > _cdrom_file.size)
                {
                    size = _cdrom_file.size - offset;
                }

                if(cdfs_seek(&(_cdrom_file), offset, CDFS_SEEK_SET) != IAPETUS_ERR_OK)
                {
                    continue;
                }

                if(cdfs_read(ptr, 1, size, &(_cdrom_file)) != IAPETUS_ERR_OK)
                {
                    continue;
                }
                else
                {
                    retsize = size;
                }

                cdrom_close_file();
            }
        }
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Use file from SD card. */
        sdc_ret_t ret;
        unsigned long file_size = scf_size(dev_id, filename);

        /* Check parameters. */
        retsize = size;
        if(offset > file_size)
        {
            offset = 0;
        }
        if((offset + retsize) > file_size)
        {
            retsize = file_size - offset;
        }

        if(retsize)
        {
            scd_usbsd_logout(1/*usbonly*/, "@@@[SD]sdc_fat_read_file(offset=%d, size=%d) ...", offset, retsize);
            ret = sdc_fat_read_file(sc->filename_buff, offset, retsize, (unsigned char*)ptr, &retsize);
            if(ret != SDC_OK) retsize = 0;
        }
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* Use file from Satiator. */
        unsigned long file_size = scf_size(dev_id, filename);

        SATIATOR_MIDDLE_WAIT();

        /* Check parameters. */
        retsize = size;
        if(offset > file_size)
        {
            offset = 0;
        }
        if((offset + retsize) > file_size)
        {
            retsize = file_size - offset;
        }

        if(retsize)
        {
            /* As file size retrieval modifies filename_buff contents, it is
             * necessary to re-bufferize the path to file now.
             */
            scf_filename_bufferize(filename, sc->filename_buff);

            int fd = s_open(sc->filename_buff, FA_READ);
#if SATIS_LOG == 1
            scd_logout("[SATIS]read(%s) fd:%d", sc->filename_buff, fd);
#endif // SATIS_LOG == 1

            if(fd >= 0)
            {
                unsigned long read_size = retsize;
                unsigned long read_offset;
                s_seek(fd, offset, SEEK_SET);

                retsize = 0;
                for(read_offset=0; read_offset < read_size; /*nothing*/)
                {
                    int chunk_len = MIN(read_size - read_offset, S_MAXBUF);

                    int sret = s_read(fd, ((unsigned char*)ptr) + read_offset, chunk_len);
                    if(sret > 0)
                    {
                        retsize += sret;
                    }
#if SATIS_LOG == 1
                    scd_logout("[SATIS]read(OFS:%d, LEN:%d) ret:%d", read_offset, chunk_len, sret);
#endif // SATIS_LOG == 1


                    read_offset += chunk_len;
                }

                s_close(fd);
            }
        }
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_read)
        {
            unsigned long file_size = scf_size(dev_id, filename);

            /* Check parameters. */
            retsize = size;
            if(offset > file_size)
            {
                offset = 0;
            }
            if((offset + retsize) > file_size)
            {
                retsize = file_size - offset;
            }

            retsize = _sc_extra_media_ptr->fct_read(dev_id, sc->filename_buff, offset, retsize, ptr);
        }
    }
#endif // SC_EXTRA_MEDIA

    // scd_usbsd_logout(1/*usbonly*/, "@@@scf_read end retsize=%d", retsize);

    return retsize;
}


unsigned long scf_write(void* ptr, unsigned long size, unsigned char dev_id, char* filename)
{
    unsigned long retsize = 0;
    sc_internal_t* sc = sc_get_internal();

    scf_filename_bufferize(filename, sc->filename_buff);
#if SATIS_LOG == 1
    scd_logout("[SATIS]scf_write(%s, dev:%d, sc->satis_flags:0x%02X)", sc->filename_buff, dev_id, sc->satis_flags);
#endif // SATIS_LOG == 1

#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 1/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Modify file on PC. */
        retsize = _scd_file_request(
                    /*type       */SCD_TYPE_WFILE, 
                    /*foffset    */0, 
                    /*frw_address*/(unsigned long)(ptr), 
                    /*frw_len    */size, 
                    /*fname      */sc->filename_buff);
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Can't write to RomFs -> do nothing. */
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* Can't write to CD-ROM -> do nothing. */
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Use file from SD card. */
        if(sc->sd_read_only)
        {
            retsize = 0;
        }
        else
        {
            sdc_ret_t ret;
            ret = sdc_fat_write_file(sc->filename_buff, size, (unsigned char*)ptr, &retsize);
            if(ret != SDC_OK) retsize = 0;
        }
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* Use file from Satiator. */
        SATIATOR_MIDDLE_WAIT();

        int fd = s_open(sc->filename_buff, FA_WRITE | FA_CREATE_ALWAYS);

        if(fd >= 0)
        {
            int write_offset, chunk_len;
            for(write_offset=0; write_offset<size; write_offset+=chunk_len)
            {
                chunk_len = MIN(size - write_offset, S_MAXBUF);

                int sret = s_write(fd, ((unsigned char*)ptr) + write_offset, chunk_len);
                if(sret > 0)
                {
#if SATIS_LOG == 1
                    scd_logout("[SATIS]write(OFS:%d, LEN:%d) ret:%d", write_offset, chunk_len, sret);
#endif // SATIS_LOG == 1
                    retsize += sret;
                    s_sync(fd);

                    /* Let's assume that flushing data to SD card requires a
                     * little time and consequently let's wait for a while to
                     * avoid freeze of Saturn.
                     */
                    SATIATOR_SMALL_WAIT();
                }
                else
                {
#if SATIS_LOG == 1
                    scd_logout("[SATIS]write(OFS:%d, LEN:%d) ERROR ! ret:%d", write_offset, chunk_len, sret);
#endif // SATIS_LOG == 1
                    break;
                }
            }

            s_close(fd);
        }
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_write)
        {
            /* Because subfolders are not supported, Fenrir file write module
             * have to simplify path of the file to write.
             * 
             * Because of that, it is necessary to feedback the simplified
             * file path so that caller function can eventually log or display
             * it correctly after that.
             * 
             * As a consequence, "name bufferizing" is not used and input
             * pointer is specified directly to Fenrir file I/O module.
             * "Name bufferizing" is recommended to correct the syntax of path
             * to file, and shunting it in the particular case of Fenrir (of
             * dumping save data or cartridge ROM) shouldn't cause any harm.
             */
            //retsize = _sc_extra_media_ptr->fct_write(ptr, size, dev_id, sc->filename_buff);
            retsize = _sc_extra_media_ptr->fct_write(ptr, size, dev_id, filename);
        }
    }
#endif // SC_EXTRA_MEDIA

    return retsize;
}


unsigned long scf_append(void* ptr, unsigned long size, unsigned char dev_id, char* filename)
{
    unsigned long retsize = 0;
    sc_internal_t* sc = sc_get_internal();

    scf_filename_bufferize(filename, sc->filename_buff);
#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 1/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Use file from PC. */
        retsize = _scd_file_request(
                    /*type       */SCD_TYPE_AFILE, 
                    /*foffset    */0, 
                    /*frw_address*/(unsigned long)(ptr), 
                    /*frw_len    */size, 
                    /*fname      */sc->filename_buff);
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Can't write to RomFs -> do nothing. */
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* Can't write to CD-ROM -> do nothing. */
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Use file from SD card. */
        if(sc->sd_read_only)
        {
            retsize = 0;
        }
        else
        {
            sdc_ret_t ret;
            ret = sdc_fat_append_file(sc->filename_buff, size, (unsigned char*)ptr, &retsize);
            if(ret != SDC_OK) retsize = 0;
        }
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* Use file from Satiator. */
        unsigned long file_size = scf_size(dev_id, filename);

        SATIATOR_MIDDLE_WAIT();

#if SATIS_LOG == 1
        scd_logout("[SATIS]append name[%s] size[%u]", filename, file_size);
        char str_tmp[20];
        memcpy(str_tmp, ptr, sizeof(str_tmp)); str_tmp[sizeof(str_tmp)-1] = '\0';
        scd_logout("[SATIS]append STR[%s]", str_tmp);
#endif // SATIS_LOG == 1

        /* As file size retrieval modifies filename_buff contents, it is
         * necessary to re-bufferize the path to file now.
         */
        scf_filename_bufferize(filename, sc->filename_buff);

        int fd = s_open(sc->filename_buff, (file_size == 0 ? FA_WRITE | FA_CREATE_ALWAYS : FA_WRITE));
        if(fd >= 0)
        {
            if(file_size != 0)
            {
                s_seek(fd, file_size, SEEK_SET);
            }

            int write_offset, chunk_len;
            for(write_offset=0; write_offset<size; write_offset+=chunk_len)
            {
                chunk_len = MIN(size - write_offset, S_MAXBUF);

                int sret = s_write(fd, ((unsigned char*)ptr) + write_offset, chunk_len);
                if(sret > 0)
                {
#if SATIS_LOG == 1
                    scd_logout("[SATIS]append(OFS:%d, LEN:%d) ret:%d", write_offset, chunk_len, sret);
#endif // SATIS_LOG == 1
                    retsize += sret;
                    s_sync(fd);

                    /* Let's assume that flushing data to SD card requires a
                     * little time and consequently let's wait for a while to
                     * avoid freeze of Saturn.
                     */
                    SATIATOR_SMALL_WAIT();
                }
                else
                {
#if SATIS_LOG == 1
                    scd_logout("[SATIS]append(OFS:%d, LEN:%d) ERROR ! ret:%d", write_offset, chunk_len, sret);
#endif // SATIS_LOG == 1
                    break;
                }
            }

            s_close(fd);
        }
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_append)
        {
            retsize = _sc_extra_media_ptr->fct_append(ptr, size, dev_id, sc->filename_buff);
        }
    }
#endif // SC_EXTRA_MEDIA

    return retsize;
}


unsigned long scf_mkdir(unsigned char dev_id, char* folder)
{
    unsigned long retval = 0;
#ifndef SC_NO_SATIATOR
    sc_internal_t* sc = sc_get_internal();
#endif // ndef SC_NO_SATIATOR

#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 1/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Create folder on PC. */
        retval  = _scd_file_request(
                    /*type       */SCD_TYPE_MKDIR, 
                    /*foffset    */0, 
                    /*frw_address*/0, 
                    /*frw_len    */0, 
                    /*fname      */folder);
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Can't write to RomFs -> do nothing. */
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* Can't write to CD-ROM -> do nothing. */
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Create folder on SD card. */
        sc_internal_t* sc = sc_get_internal();

        if(sc->sd_read_only)
        {
            retval = 0;
        }
        else
        {
            sdc_ret_t ret;
            ret = sdc_fat_mkdir(folder);
            retval = (ret == SDC_OK ? 1 : 0);
        }
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* Create folder on Satiator. */
        SATIATOR_MIDDLE_WAIT();

        s_mkdir(folder);

        /* Check if the directory exist or not.
         * 
         * Note : s_mkdir may return an error because directory was previously
         *        existing, but this must be considered as normal behavior.
         *        For this reason, s_mkdir return code is not checked and
         *        opening of the folder is tested instead.
         */
        if(s_opendir(folder) == 0)
        {
            retval = 1;
        }
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_mkdir)
        {
            retval = _sc_extra_media_ptr->fct_mkdir(dev_id, folder);
        }
    }
#endif // SC_EXTRA_MEDIA

    return retval;
}


/**
 * Remove file specified in path.
 *
 * Return 1 on success, 0 else.
**/
unsigned long scf_delete(unsigned char dev_id, char* path)
{
    unsigned long retval = 0;
#ifndef SC_NO_SATIATOR
    sc_internal_t* sc = sc_get_internal();
#endif // ndef SC_NO_SATIATOR

#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 0/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Use file from PC. */
        retval  = _scd_file_request(
                    /*type       */SCD_TYPE_REMOVE, 
                    /*foffset    */0, 
                    /*frw_address*/0, 
                    /*frw_len    */0, 
                    /*fname      */path);
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Can't write to RomFs -> do nothing. */
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* Can't write to CD-ROM -> do nothing. */
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Delete file on SD card. */
        sc_internal_t* sc = sc_get_internal();

        if(sc->sd_read_only)
        {
            retval = 0;
        }
        else
        {
            sdc_ret_t ret;
            ret = sdc_fat_unlink(path);
            if(ret == SDC_OK) retval = 1;
        }
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* Delete file on Satiator. */
        SATIATOR_MIDDLE_WAIT();

        if(s_unlink(path) == 0)
        {
            retval = 1;
        }
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_delete)
        {
            retval = _sc_extra_media_ptr->fct_delete(dev_id, path);
        }
    }
#endif // SC_EXTRA_MEDIA

    return retval;
}



int scf_compare_extension(char* str, char* ext)
{
    int ret = 0;
    int i;
    char e1[4];
    char e2[sizeof(e1)];

    for(i=0; i<sizeof(e1); i++)
    {
        e1[i] = '\0';
        e2[i] = '\0';
    }

    i = strlen(str);
    if(i > 4)
    {
        str += i-4;
    }
    for(i=0; i<4; i++)
    {
        if((*str) == '.')
        {
            str++;
            break;
        }
        str++;
    }
    for(i=0; i<3; i++)
    {
        if(str[i] == '\0') break;
        e1[i] = sc_toupper(str[i]);
    }

    if(ext[0] == '.')
    {
        for(i=0; i<3; i++)
        {
            if(ext[i+1] == '\0') break;
            e2[i] = sc_toupper(ext[i+1]);
        }
    }
    else
    {
        for(i=0; i<3; i++)
        {
            if(ext[i] == '\0') break;
            e2[i] = sc_toupper(ext[i]);
        }
    }

    if((e1[0] == e2[0]) && (e1[1] == e2[1]) && (e1[2] == e2[2]))
    {
        ret = 1;
    }

    return ret;
}


char* scf_get_filename(char* path)
{
    char* ret = path;
    int n, i;

    n = strlen(path);

    /* Fix backslash in path. */
    for(i=0; i<(n-1); i++)
    {
        if((path[i] == '/' )
        || (path[i] == '\\'))
        {
            ret = path+i+1;
        }
    }

    return ret;
}

void scf_parent_path(char* path)
{
    int n, i;

    n = strlen(path);

    /* Fix backslash in path. */
    for(i=0; i<n; i++)
    {
        if((path[i] == '/' )
        || (path[i] == '\\'))
        {
            path[i] = '/';
        }
    }

    /* Remove eventual final slash. */
    if(n != 0)
    {
        if(path[n-1] == '/' )
        {
            n--;
            path[n] = '\0';
        }
    }

    if(n != 0)
    {
        for(i=n-1; i>=0; i--)
        {
            if(path[i] == '/')
            {
                path[i] = '\0';
                break;
            }
            /* Back to first folder -> root folder. */
            if(i == 0) path[0] = '\0';
        }
    }
}


/**
 * Return 1 if f1's name is greater than f2's name, zero else.
 * Note : case insensitive.
 * (for use in scf_filelist function only.)
**/
int scf_file_is_greater(scdf_file_entry_t* f1, scdf_file_entry_t* f2)
{
    int ret = 0;
    int i;
    char c1, c2;

    if((f1->flags & FILEENTRY_FLAG_FOLDER) != (f2->flags & FILEENTRY_FLAG_FOLDER))
    { /* Put folder on top of list. */
        ret = (f1->flags & FILEENTRY_FLAG_FOLDER ? 0 : 1);
    }
    else
    {
        for(i=0; i<SC_FILENAME_MAXLEN; i++)
        {
            c1 = sc_tolower(f1->filename[i]);
            c2 = sc_tolower(f2->filename[i]);

            if(c1 != c2)
            {
                ret = (c1 > c2 ? 1 : 0);
                break;
            }
        }
    }

    return ret;
}

/**
 * List files in specified folder.
 *
 * Return the count of files in the specified folder.
**/
unsigned long scf_filelist(scdf_file_entry_t* list_ptr, unsigned long list_count, unsigned long list_offset, unsigned char list_sort, char* filter, unsigned char dev_id, char* path)
{
    sc_internal_t* sc = sc_get_internal();

    unsigned long retval = 0;
    long file_index;
    unsigned long items_count;

    romfs_entry_t* e;

#ifdef SC_RELEASE
    /* In release mode, always use CD-ROM instead of PC. */
    if(dev_id == SC_FILE_PC)
    {
        dev_id = SC_FILE_CDROM;
    }
#endif

    if((list_ptr)
    && (list_count != 0))
    {
        /* Reset file list data. */
        memset((void*)list_ptr, 0x00, list_count*sizeof(scdf_file_entry_t));
    }

#ifndef SC_NO_SATIATOR
    /* Properly initialize Satiator SD card API or CD-ROM. */
    scf_satiator_access_init(dev_id, 0/*set_time*/);
#endif // ndef SC_NO_SATIATOR

    if(dev_id == SC_FILE_PC)
    { /* Use file from PC. */
        retval  = _scd_file_request(
                    /*type       */SCD_TYPE_LISTDIR, 
                    /*foffset    */list_offset, 
                    /*frw_address*/(unsigned long)list_ptr, 
                    /*frw_len    */list_count, 
                    /*fname      */path);
    }
    else if(dev_id == SC_FILE_CDROM)
    { /* List files in CD-ROM. */
        if(sc->cdrom_status & SC_CDSTATUS_USE)
        {
            cdrom_init_check();

            cdfs_filelist(list_ptr, list_count, list_offset, path, &retval);
        }
    }
    else if(dev_id == SC_FILE_ROMFS)
    { /* Use file from RomFs. */
        e = romfs_get_first_entry(sc->romfs, path);

        if(e == NULL)
        {
            retval = 0;
        }
        else
        {
            if(e->flags & FILEENTRY_FLAG_FOLDER)
            {
                /* Folder found. List files inside it. */
                file_index = 0;
                retval = 0;
                while(1)
                {
                    /* List files in selected range. */
                    if((file_index >= list_offset)
                    && (retval < list_count))
                    {
                        /* File within range : copy filename and flags to ouput parameter. */
                        char* name = (char*)(e) + sizeof(romfs_entry_t);

                        list_ptr[retval].flags = e->flags;
                        memcpy((void*)(list_ptr[retval].filename), (void*)name, e->filename_len);

                        retval++;
                    }

                    if(e->flags & FILEENTRY_FLAG_LASTENTRY)
                    {
                        break;
                    }

                    e =  romfs_get_next_entry(sc->romfs, e);
                    file_index++;
                }
            }
            else
            {
                retval = 0;
            }
        }
    }
#ifndef SC_NO_SDCARD
    else if(dev_id == SC_FILE_SDCARD)
    { /* Use file from SD card. */
        sdc_fat_filelist(list_ptr, list_count, list_offset, path, &retval);
    }
#endif // SC_NO_SDCARD

#ifndef SC_NO_SATIATOR
    else if((dev_id == SC_FILE_SATIATOR) && (sc->satis_flags & SATIS_SDCARD_OK))
    { /* List files on Satiator. */
        SATIATOR_MIDDLE_WAIT();

        if(s_opendir(path))
        {
            retval = 0;
        }
        else
        {
            char statbuf[280];
            s_stat_t *st = (s_stat_t*)statbuf;

            /* Folder found. List files inside it. */
            file_index = 0;
            retval = 0;
            while(1)
            {
                int len = s_stat(NULL, st, sizeof(statbuf)-1);
                if((len <= 0) || (len >= (sizeof(statbuf)-9/*size, attrib etc*/)))
                {
                    break;
                }

                st->name[len] = '\0';

                /* Skip special directories such as ".", ".." or ".system" etc.
                 *
                 * Also, avoid to access hidden or other exotic entry.
                 */
                if((st->name[0] < ' ') || (st->attrib & (AM_HID|AM_SYS|AM_VOL)))
                {
                    continue;
                }

                /* List files in selected range. */
                if((file_index >= list_offset)
                && (retval < list_count))
                {
                    /* File within range : copy filename and flags to ouput parameter. */
                    strncpy(list_ptr[retval].filename, st->name, SC_FILENAME_MAXLEN-1);
                    list_ptr[retval].filename[SC_FILENAME_MAXLEN-1] = '\0';

                    list_ptr[retval].flags = FILEENTRY_FLAG_ENABLED;
                    if(st->attrib & AM_DIR)
                    {
                        list_ptr[retval].flags |= FILEENTRY_FLAG_FOLDER;
                    }

                    /* Mark files with names too long to be handled here.
                     * Another way to do regarding LFN is to discard them from
                     * listing, but as it will make these files not being
                     * listed after that, it's better to list them and to show
                     * an error message when trying to handle a name marked
                     * as being too long.
                     */
                    if(strlen(st->name) > (SC_FILENAME_MAXLEN-1))
                    {
                        list_ptr[retval].flags |= FILEENTRY_FLAG_NAMETOOLONG;
                    }

                    retval++;
                }

                file_index++;
            }
        }
    }
#endif // ndef SC_NO_SATIATOR

#ifdef SC_EXTRA_MEDIA
    else if((dev_id >= SC_FILE_EXTRA) && (_sc_extra_media_ptr != NULL))
    { /* Use extra media. */
        if(_sc_extra_media_ptr->fct_filelist)
        {
            retval = _sc_extra_media_ptr->fct_filelist(list_ptr, list_count, list_offset, list_sort, filter, dev_id, path);
        }
    }
#endif // SC_EXTRA_MEDIA


    /* Get number if items in list. */
    items_count = retval;
    if(items_count > list_count)
    {
        items_count = list_count;
    }

    if(list_ptr)
    {
        /* Sanitize folder names : append slash on begining. */
        for(file_index=0; file_index<list_count; file_index++)
        {
            if((list_ptr[file_index].flags & FILEENTRY_FLAG_ENABLED)
            && (list_ptr[file_index].flags & FILEENTRY_FLAG_FOLDER ))
            {
                /* Append slash on begining of folder names. */
                if((list_ptr[file_index].filename[0] != '/' )
                && (list_ptr[file_index].filename[0] != '\\'))
                {
                    int i;
                    for(i=SC_FILENAME_MAXLEN-2; i>=0; i--)
                    {
                        list_ptr[file_index].filename[i+1] = list_ptr[file_index].filename[i];
                    }

                    /* Add slash. */
                    list_ptr[file_index].filename[0] = '/';

                    /* Don't forget terminating null character. */
                    list_ptr[file_index].filename[SC_FILENAME_MAXLEN-1] = '\0';
                }

                /* Remove "." and ".." folders. */
                if((list_ptr[file_index].filename[1] ==  '.')
                && (list_ptr[file_index].filename[2] == '\0'))
                {
                    list_ptr[file_index].flags &= ~FILEENTRY_FLAG_ENABLED;
                }
                if((list_ptr[file_index].filename[1] ==  '.')
                && (list_ptr[file_index].filename[2] ==  '.')
                && (list_ptr[file_index].filename[3] == '\0'))
                {
                    list_ptr[file_index].flags &= ~FILEENTRY_FLAG_ENABLED;
                }
            }
        }

        /* Reset sort index. */
        for(file_index=0; file_index<list_count; file_index++)
        {
            list_ptr[file_index].sort_index = file_index;
        }

        /* If required, apply file filtering. */
        if(filter != NULL)
        {
            char ext[5];
            int n = strlen(filter);
            int i, j;

            /* Mark all folders as selected by filtering. */
            for(file_index=0; file_index<items_count; file_index++)
            {
                if(list_ptr[file_index].flags & FILEENTRY_FLAG_FOLDER)
                {
                    list_ptr[file_index].flags |= FILEENTRY_FLAG_FILTERED;
                }
            }

            for(i = 0; i<n; i++)
            {
                if((i == 0)
                || (filter[i] == ',')
                || (filter[i] == '\0'))
                {
                    if(filter[i] == ',')
                    {
                        i++;
                    }

                    for(j=0; j<sizeof(ext); j++)
                    {
                        ext[j] = '\0';
                    }

                    for(j=0; j<(sizeof(ext)-1); j++)
                    {
                        if(filter[i] == '\0') break;
                        ext[j] = filter[i];
                        if(filter[i+1] == ',') break;
                        i++;
                    }

                    for(file_index=0; file_index<items_count; file_index++)
                    {
                        /* Don't compare already selected files. */
                        if((list_ptr[file_index].flags & FILEENTRY_FLAG_FILTERED) == 0)
                        {
                            if(scf_compare_extension(list_ptr[file_index].filename, ext))
                            {
                                list_ptr[file_index].flags |= FILEENTRY_FLAG_FILTERED;
                            }
                        }
                    }
                }
            }

            /* Disable non filtered files. */
            for(file_index=0; file_index<items_count; file_index++)
            {
                if((list_ptr[file_index].flags & FILEENTRY_FLAG_FILTERED) == 0)
                {
                    list_ptr[file_index].flags &= ~FILEENTRY_FLAG_ENABLED;
                }
                list_ptr[file_index].flags &= ~FILEENTRY_FLAG_FILTERED;
            }
        }

        /* Sort entries */
        if(list_sort)
        {
            // Sort faces. (from mic's flatcube)
            // Bubble-sort the whole lot.. yeehaw!
            if(items_count > 1)
            {
                unsigned short i, j, idi, idj;
                unsigned short temp;
                for(i=0; i<items_count-1; i++)
                {

                    for(j=i+1; j<items_count; j++)
                    {
                        idi = list_ptr[i].sort_index;
                        idj = list_ptr[j].sort_index;

                        if(scf_file_is_greater(list_ptr+idi, list_ptr+idj))
                        {
                            temp                   = list_ptr[i].sort_index;
                            list_ptr[i].sort_index = list_ptr[j].sort_index;
                            list_ptr[j].sort_index = temp;
                        }
                    }
                }
            }
        } // if(list_sort)
    }

    return retval;
}


/**
 * Return the count of files/folders in specified search result data.
**/
unsigned long scf_filecount(scdf_file_entry_t* list_ptr, unsigned long list_count)
{
    unsigned short list_index;
    unsigned long ret = 0;

    for(list_index=0; list_index<list_count; list_index++)
    {
        if(list_ptr[list_index].flags & FILEENTRY_FLAG_ENABLED)
        {
            ret++;
        }
    }

    return ret;
}

